#include "avoidObs/inc/AvoidTrajGen.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

#include "common/inc/Timer.hpp"

#define DFLT_DS_MAXDIST_POINTS 0.5
#define DFLT_MINNBITV_TRAJGEN 7

#define COEF_TIME_STOP 1.3

#define COEF_MINDIFF_STRANG_POINTS 5.0

#define CHECK_SAFETY false

#define DISPLAY_TIME false

                                    /**************************************************/
                                    /*************** Class AvoidTrajGen ***************/
                                    /**************************************************/

int AvoidTrajGen::version = 1;                      /**< AvoidTrajGen version used (1=list version, 2=graph version) */

// dem treatment parameters
double AvoidTrajGen::SLOPEMAP_DIST = 0.3;           /**< Maximal distance considered between two cells of the slope map to compute the slope */
double AvoidTrajGen::MAX_SLOPE = 0.35*M_PI;         /**< Maximal slope that the robot can climb */
double AvoidTrajGen::MAX_VAR_SLOPE = 2.0*M_PI;      /**< Maximal slope derivative that the robot can climb */
double AvoidTrajGen::SAFE_DIST = 1.0;               /**< Distance that the robot must keep between its center of rotation and the obstacle to be safe */

// corridor treatment parameters
double AvoidTrajGen::MAX_LENGTH_PATH_TTT = 30.0;    /**< maximal length of the path that will be analized for the trajectory computation */

  // general options
double AvoidTrajGen::ds_maxDist_points_ = DFLT_DS_MAXDIST_POINTS;
int AvoidTrajGen::minNbItv_trajgen_ = DFLT_MINNBITV_TRAJGEN;
double AvoidTrajGen::coefTime_ = 1.5;

bool AvoidTrajGen::POSTER_DEBUG_ATG = false;

float AvoidTrajGen::maxDistTgtReached_ = ATF_DFLT_MAX_DIST_TGT_REACHED;
float AvoidTrajGen::maxDistWpReached_ = ATF_DFLT_MAX_DIST_WP_REACHED;
float AvoidTrajGen::maxAngleWpReached_ = ATF_DFLT_MAX_ANGLE_WP_REACHED;


/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
AvoidTrajGen::AvoidTrajGen()
  : graphPsgwy_(0, 0)
{
  // general inputs
  sys_ = NULL;
  path_ = NULL;
  width_ = 0;
  max_length_ = 0;
  fov_ = 0;
  
  // stage inputs
  dem_ = NULL;
  loc_ = NULL;
  
  // stage outputs
  pctraj_ = NULL;
//   pstraj_ = NULL;
  straj_ = NULL;

  // general temp data
  border1_ = NULL;
  border2_ = NULL;

  trajgen_ = NULL; 

  treeMDP_ = NULL;
  tempGoal_ = NULL;
  nearPathPos_ = NULL;

//   // stage temp data
  curS_ = 0;

  distmap_ = NULL;

  graphTopoPsgwy_ = NULL;
  treePsgwy_ = NULL;
  simplifiedGraphPsgwy_ = NULL;

#if DEBUG_ATG
    image = NULL;
    image2 = NULL;
    image3 = NULL;
    image4 = NULL;
#endif
}

/**
 *  \brief default constructor
 */
AvoidTrajGen::AvoidTrajGen(System* sys, XYThS_Path* path, double max_length, double width, double fov)
  : graphPsgwy_(0.5*width, width)
{
  // general inputs
  sys_ = sys;
  path_ = path;
  width_ = width;
  max_length_ = max_length;
  fov_ = fov;
  
  // stage inputs
  dem_ = NULL;
  loc_ = NULL;
  
  // stage outputs
  pctraj_ = NULL;
//   pstraj_ = NULL;
  straj_ = NULL;

  // general temp data
  border1_ = NULL;
  border2_ = NULL;

  trajgen_ = NULL;

  treeMDP_ = NULL;
  tempGoal_ = NULL;
  nearPathPos_ = NULL;

  this->computeBorderPaths();

//   // stage temp data
  curS_ = 0;

  distmap_ = NULL;

  graphTopoPsgwy_ = NULL;
  treePsgwy_ = NULL;
  simplifiedGraphPsgwy_ = NULL;

#if DEBUG_ATG
    image = NULL;
    image2 = NULL;
    image3 = NULL;
    image4 = NULL;
#endif
}

/**
 *  \brief destructor
 */
AvoidTrajGen::~AvoidTrajGen()
{
  // stage inputs
  if(dem_ != NULL) delete dem_;
  if(loc_ != NULL) delete loc_;
  
  // stage outputs
  if(pctraj_ != NULL) delete pctraj_;
//   if(pstraj_ != NULL) delete pstraj_;
  if(straj_ != NULL) delete straj_;
  
  // stage temp data
  maxDistPoints_.clear();
  strangPoints_.clear();
  waypoints_.clear();

  // general temp data
  if(border1_ != NULL) delete border1_;
  if(border2_ != NULL) delete border2_;
  if(treeMDP_ != NULL) delete treeMDP_;
  if(tempGoal_ != NULL) delete tempGoal_;
#if DEBUG_ATG
  if(nearPathPos_ != NULL) delete nearPathPos_;
  vecNearPathPos_.clear();
#endif

//   if(graphTopoPsgwy_ != NULL) delete graphTopoPsgwy_;
//   if(simplifiedGraphPsgwy_ != NULL) delete simplifiedGraphPsgwy_;

  if(POSTER_DEBUG_ATG)
  {
    boost::mutex::scoped_lock lock(mutexDistMap_);
    if(distmap_ != NULL) delete distmap_;
  }

  if(graphTopoPsgwy_ != NULL) delete graphTopoPsgwy_;
  if(treePsgwy_ != NULL) delete treePsgwy_;
  if(simplifiedGraphPsgwy_ != NULL) delete simplifiedGraphPsgwy_;

#if DEBUG_ATG
  if(image != NULL) delete image;
  if(image2 != NULL) delete image2;
#endif
}


/*************** Getters and setters **************/

/**
 *  \brief set the distance between max dist points
 */
void AvoidTrajGen::setDSMaxDistPoints(double ds)
{
  AvoidTrajGen::ds_maxDist_points_ = ds;
}

/**
 *  \brief set the distance between max dist points
 */
void AvoidTrajGen::setMinNbItvTrajgen(int min_nb_itv)
{
  AvoidTrajGen::minNbItv_trajgen_ = min_nb_itv;
}

/**
 *  \brief set the time coef used to select the waypoints arriving time
 *  \param coef_time : the new time coefficient
 *
 *  coef_time = 1  -> maximal velocity used
 *  coef_time > 1  -> multiply the time needed to reach the goal
 */
void AvoidTrajGen::setCoefTime(double coef_time)
{
  AvoidTrajGen::coefTime_ = coef_time;
}

/**
 *  \brief return a copy of the list of waypoints found in the reference corridor
 *  \return a copy of the list of waypoints found in the reference corridor
 */
// std::vector<XYThS_Config> AvoidTrajGen::waypoints()
std::vector<WayPoint> AvoidTrajGen::waypoints()
{
//   std::vector<XYThS_Config> wp; 
  std::vector<WayPoint> wp; 
  
  int N = waypoints_.size();
  for(int i=0; i<N; i++)
    {
      wp.push_back(waypoints_.at(i));
    }
  return wp;
}

/**
 *  \brief return a copy of the current distmap
 *  \return a copy of the current distmap
 */
DistObsMap* AvoidTrajGen::distmap()
{
  boost::mutex::scoped_lock lock(mutexDistMap_);
  if(distmap_ != NULL) return new DistObsMap(*distmap_);
  else return NULL;
}

/**
 *  \brief return the current position of the robot
 *  \return the current position of the robot
 */
PositionXYTh AvoidTrajGen::curPos() const
{
  if(this->loc_ != NULL) return this->loc_->posXYTh();
  else
    {
      cerr << "AvoidTrajGen::curPos() : The current location has not been defined!" << endl;
      return PositionXYTh();
    }
}

/**
 *  \brief return the goal position (final position of the followed trajectory)
 *  \return the goal position (final position of the followed trajectory)
 */
PositionXYTh AvoidTrajGen::goalPos() const
{
  if(this->path_ != NULL)
    {
      XYThS_Config lastCfg = this->path_->lastCfg();
      return PositionXYTh(lastCfg.x(), lastCfg.y(), lastCfg.th());
    }
  else
    {
      cerr << "AvoidTrajGen::goalPos() : The followed path has not been defined!" << endl;
      return PositionXYTh();
    }
}

/**
 *  \brief return the goal configuration (final configuration of the followed trajectory)
 *  \return the goal configuration (final configuration of the followed trajectory)
 */
XYThS_Config AvoidTrajGen::goalCfg() const
{
  if(this->path_ != NULL) return this->path_->lastCfg();
  else
    {
      cerr << "AvoidTrajGen::goalCfg() : The followed path has not been defined!" << endl;
      return XYThS_Config();
    }
}


/********** Corridor borders computation **********/

/*
 *  \brief compute the corridor border paths
 *  \return 0 if the process succeded, -1 else
 */
int AvoidTrajGen::computeBorderPaths()
{
  if(border1_ != NULL) delete border1_;
  if(border2_ != NULL) delete border2_;

  // check if the reference path has been defined
  if(path_ == NULL) { cerr << "AvoidTrajGen::computeBorderPaths : Cannot compute the reference corridor borders. The reference path is null."<< endl;  return -1; }
  // check if the width of the corridor is valid
  if(width_ <= 0) { cerr << "AvoidTrajGen::computeBorderPaths : Cannot compute the reference corridor borders. The corridor width must be > 0."<< endl;  return -1; }

  // init the new border paths
  border1_ = new XYThS_Path();
  border2_ = new XYThS_Path();  

  int N = path_->size();
  for(int i=0; i<N; i++)
    {
      // get back the current configuration
      XYThS_Config pathCfg = path_->at(i);

      XYThS_Config border1Cfg(pathCfg.x() - width_*sin(pathCfg.th()),
			      pathCfg.y() + width_*cos(pathCfg.th()),
			      pathCfg.th(), 
			      pathCfg.s());
      border1_->append(border1Cfg);

      XYThS_Config border2Cfg(pathCfg.x() + width_*sin(pathCfg.th()),
			      pathCfg.y() - width_*cos(pathCfg.th()),
			      pathCfg.th(), 
			      pathCfg.s());
      border2_->append(border2Cfg);
    }

//   // copy the path into the borders
//   border1_ = new XYThS_Path(*path_);
//   border2_ = new XYThS_Path(*path_);
  
//   // shift them to the desired positions
//   border1_->shiftLat(-width_);
//   border2_->shiftLat(width_);

//   cout << "Border1 : " << endl;
//   border1_->display();
//   cout << endl;
//   cout << "Border2 : " << endl;
//   border2_->display();
//   cout << endl;
}


/************* Initialization functons ************/

/**
 *  \brief Initialize the standard parameters of the trajectory generator
 *  \param dbg : the trajectory generator debugging mode
 */
void AvoidTrajGen::initTrajGen(int dbg)
{
  if(trajgen_ != NULL)
    {
      /// common trajgen options
      trajgen_->setTcStatus(TIME_FREE,0,999999999.9);
      
      // init options of the trajectory generation
      trajgen_->setMaxNbStep(20);                 // default = 20
      trajgen_->setMaxResidual(0.05);             // default = 0.05
      
      // init the debug level mode
      trajgen_->setDBG(dbg);                      // default = 0
    }
}


/************* Trajectory computation *************/

/*
 *  \brief check for the AvoidTrajGen input parameters consistency
 *  \return 0 if all parameters are valid
 *  \return -2 if an I/O problem occured
 *  \return -3 if the avoid trajgen settings is not valid
 */
int AvoidTrajGen::checkSettings()
{
  // check if the system has been initialized
  if(sys_ == NULL) { cerr << "AvoidTrajGen::computeAvoidTraj : system is null" << endl;  return -3; }

  // check if the path has been initialized
  if(path_ == NULL) { cerr << "AvoidTrajGen::computeAvoidTraj : path is null" << endl;  return -2; }
  
  // check if the DEM has been initialized
  {
    boost::mutex::scoped_lock lock(mutexDEM_);
    if(dem_ == NULL) { cerr << "AvoidTrajGen::computeAvoidTraj : DEM is null" << endl;  return -2; }
  }

  // check if the localization has been initialized
  if(loc_ == NULL) { cerr << "AvoidTrajGen::computeAvoidTraj : loc is null" << endl;  return -2; }
  
  // check if the corridor width is valid
  if(width_ < 0) { cerr << "AvoidTrajGen::computeAvoidTraj : width not valid" << endl;  return -3; }

  return 0;
}

/*
 *  \brief reinit the avoidance trajectory computation process
 */
void AvoidTrajGen::reinit()
{
  // stage temp data
  maxDistPoints_.clear();
  strangPoints_.clear();
  waypoints_.clear();

  if(tempGoal_ != NULL)
    {
      delete tempGoal_;
      tempGoal_ = NULL;
    }

#if DEBUG_ATG
  if(nearPathPos_ != NULL)
    {
      delete nearPathPos_;
      nearPathPos_ = NULL;
    }
  vecNearPathPos_.clear();
#endif

  if(treeMDP_ != NULL)
    {
      delete treeMDP_;
      treeMDP_ = NULL;
    }

#if DEBUG_ATG
  if(image != NULL)
    {
      delete image;
      image = NULL;
    }
  if(image2 != NULL)
    {
      delete image2;
      image2 = NULL;
    }
  if(image3 != NULL)
    {
      delete image3;
      image3 = NULL;
    }
  if(image4 != NULL)
    {
      delete image4;
      image4 = NULL;
    }
#endif
}

/*
 *  \brief get back indices of the path containing the input curve abscica interval
 *  \param minS : input minimal curve abscica
 *  \param maxS : input maximal curve abscica
 *  \param indStart : output start index
 *  \param indEnd : output end index
 */
void AvoidTrajGen::getIndPathItv_from_SItv(double minS, double maxS, int& indStart, int& indEnd)
{
  // search for the closest configuration of the path from the robot position
  int N = path_->size();
  
  bool startFound = false;
  indStart=0;
  while(!startFound && indStart < N) 
    {
      XYThS_Config cfg = path_->at(indStart);
      if(cfg.s() > minS) startFound = true;
      else indStart++;
    }
  if(indStart != 0) indStart--;
  
  bool endFound = false;
  indEnd=indStart;
  while(!endFound && indEnd < N) 
    {
      XYThS_Config cfg = path_->at(indEnd);
      if(cfg.s() > maxS) endFound = true;
      else indEnd++;
    }
  if(indEnd == N) indEnd--;

//   cout << "minS : " << minS << "  maxS : " << maxS << "  indStart : " << indStart << "  indEnd : " << indEnd << endl;
}


/*
 *  \brief compute the current curve abscica of the path from the robot location
 *  \param pos : the robot location
 *  \param oldS : the previous curve abscica reached
 *  \param maxDS : the maximum distance traveled from the previous recorded curve abscica
 *  \return the current curve abscica
 */
double AvoidTrajGen::computeCurS(const PositionXY& pos, double oldS, double maxDS)
{
  // compute the upper bound of the curve absica interval on which the reached curve abscica must be searched
  double maxS = oldS + maxDS;

  int indStart, indEnd;
  this->getIndPathItv_from_SItv(oldS, maxS, indStart, indEnd);
//    cout << "computeCurS oldS : " << oldS << "  maxS : " << maxS << "  indStart : " << indStart << "  indEnd : " << indEnd << endl;

  // check for the trivial case where both start and end points are equals
  if(indStart == indEnd)
    {
      curS_ = path_->at(indStart).s();
      return curS_;
    }

  nearPathPos_ = new PositionXY(-15,-15);
  vecNearPathPos_.clear();

  double minDist = 999999999;
  for(int i=indStart; i<indEnd-1; i++)     // -1 added to remove a bug... need to be checked
    {
      XYThS_Config curCfg = path_->at(i);
      XYThS_Config nextCfg = path_->at(i+1);

      PositionXY p1(curCfg.x(), curCfg.y());
      PositionXY p2(nextCfg.x(), nextCfg.y());

//       PositionXY proj = mpc::orthoProjSgmt(pos, p1, p2);

//       // compute the distance from pos to proj
//       double dx = proj.x() - pos.x();
//       double dy = proj.y() - pos.y();
//       double dist = sqrt(dx*dx+dy*dy);

      double dist;
      PositionXY proj = mpc::closePosFromSgmt(pos, p1, p2, dist);

#if DEBUG_ATG
//       cout << "projection -> p1 : (" << p1.x() << "," << p1.y() << ")  p2 : (" << p2.x() << "," << p2.y() << ")  pos : (" << pos.x() << "," << pos.y() << ")  proj : (" << proj.x() << "," << proj.y() << ")" << endl;
      vecNearPathPos_.push_back(proj);
#endif

      if(i == indStart || dist < minDist)
	{
	  double dx2 = proj.x() - p1.x();
	  double dy2 = proj.y() - p1.y();
	  double dist2 = sqrt(dx2*dx2+dy2*dy2);

	  minDist = dist;
	  curS_ = curCfg.s() + dist2;

#if DEBUG_ATG
	  nearPathPos_->x() = proj.x();
	  nearPathPos_->y() = proj.y();
#endif
	}
    }
//   cout << "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< cur S : " << curS_ << "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<" << endl;

  return curS_;
}

/*
 *  \brief set a border path as an obstacle on the obstacle map
 *  \param obsmap : the obstacle map
 *  \param border : the border path
 *  \param curS : the current curve abscica
 *  \param maxDS : the distance on which the border path must be computed
 */
void AvoidTrajGen::setBorderOnObsMap(BoolObsMap* obsmap, XYThS_Path* border, double curS, double maxDS)
{
  // compute the upper bound of the curve absica interval on which the reached curve abscica must be searched
  double maxS = curS + maxDS;

  int indStart, indEnd;
  this->getIndPathItv_from_SItv(curS, maxS, indStart, indEnd);
//   cout << "setBorderOnObsMap curS : " << curS << "  maxS : " << maxS << "  indStart : " << indStart << "  indEnd : " << indEnd << endl;

  // check if both indices are consistent
  if(indStart < 0 || indStart >= border->size()) { cerr << "AvoidTrajGen::setBorderOnObsMap : indStart=" << indStart << "  is not consistent. border.size()=" << border->size() << endl;  exit(1); }
  if(indEnd < 0 || indEnd >= border->size()) { cerr << "AvoidTrajGen::setBorderOnObsMap : indEnd=" << indEnd << "  is not consistent. border.size()=" << border->size() << endl;  exit(1); }

//   cout << "indStart : " << indStart << "  indEnd : " << indEnd << endl;
//   cout << "border->size() : " << border->size() << endl;

//   cout << "Border : " << endl;
//   border->display(); 
//   cout << endl;

  for(int i=indStart; i<indEnd; i++)
    {
//       cout << "i : " << i << endl;
      // get back the current configuration of the border
      XYThS_Config curCfg = border->at(i);
//       cout << "curCfg : "; curCfg.display(); cout << endl;
      XYThS_Config nextCfg = border->at(i+1);
//       cout << "nextCfg : "; nextCfg.display(); cout << endl;

      // set the corresponding line as an obstacle on the map
      obsmap->addLineObs(curCfg.x(), curCfg.y(), nextCfg.x(), nextCfg.y());
    }
}

/*
 *  \brief compute the obstacle map from the reference corridor and the input DEM
 *  \param dem : the input DEM
 *
 *  The input DEM is supposed to ber not null
 *  The width of the corridor is supposed to be > 0
 */
BoolObsMap* AvoidTrajGen::buildObsMap(const DEM* dem)
{
  // compute the var slope map
  VarSlopeMap varslopemap(dem, AvoidTrajGen::SLOPEMAP_DIST);

  // compute the obstacle map from the slope map
  BoolObsMap* obsmap_vsm = new BoolObsMap(&varslopemap, AvoidTrajGen::MAX_SLOPE, AvoidTrajGen::MAX_VAR_SLOPE, AvoidTrajGen::SAFE_DIST);
//   BoolObsMap* obsmap_vsm = new BoolObsMap(&varslopemap, AvoidTrajGen::MAX_SLOPE, AvoidTrajGen::MAX_VAR_SLOPE, AvoidTrajGen::SAFE_DIST);

  // compute the current curve abscica of the path from the robot location
//   cout << "buildObsMap -> oldS (curS_) : " << curS_ << endl;
  
  double curS;
//   curS = this->computeCurS(loc_->posXYTh(), curS_, AvoidTrajGen::MAX_LENGTH_PATH_TTT); 
  curS = this->computeCurS(loc_->posXYTh(), curS_, max_length_); 
//   cout << "buildObsMap -> newS (curS_) : " << curS_ << endl;

//   this->setBorderOnObsMap(obsmap_vsm, border1_, curS, AvoidTrajGen::MAX_LENGTH_PATH_TTT); 
//   this->setBorderOnObsMap(obsmap_vsm, border2_, curS, AvoidTrajGen::MAX_LENGTH_PATH_TTT); 

//   cout << "AvoidTrajGen::buildObsMap -> border1 : " << endl;
//   border1_->display();
//   cout << endl;
//   cout << "AvoidTrajGen::buildObsMap -> border2 : " << endl;
//   border2_->display();
//   cout << endl;

  this->setBorderOnObsMap(obsmap_vsm, border1_, curS, max_length_); 
  this->setBorderOnObsMap(obsmap_vsm, border2_, curS, max_length_); 

  // ... shall I increase the borders by AvoidTrajGen::SAFE_DIST ??

  // return the computed obstacle map
  return obsmap_vsm;
}

// /*
//  *  \brief search for points maximizing the distance to obstacles in the navigation corridor
//  *  \param obsmap : the obstacle map
//  */
// void AvoidTrajGen::computeMaxDistPoints(const BoolObsMap* obsmap)
// {
//   // compute a distance map from the input obstacle map
//   DistObsMap distmap(obsmap);

//   distmap.saveIntoPPM("distmap_with_borders2.ppm");

//   RGBImage* image = distmap.rgbImage();
//   image->drawCross(50, 100, 6, RGBColor(255,255,255));

//   int iImage, jImage;
//   int fail = distmap.getIndexPPMImagePos(5.0, -5.0, iImage, jImage);
//   cout << "iImage : " << iImage << "  jImage : " << jImage << endl;
//   image->drawCross(iImage, jImage, 10, RGBColor(255,255,255));

//   int iImage2, jImage2;
//   int fail2 = distmap.getIndexPPMImagePos(-5.0, 2.0, iImage2, jImage2);
//   int iImage3, jImage3;
//   int fail3 = distmap.getIndexPPMImagePos(7.0, -3.0, iImage3, jImage3);
//   image->drawLine(iImage2, jImage2, iImage3, jImage3, RGBColor(0,255,0));

//   image->saveIntoPPM("distmap_with_borders2.ppm");
//   delete image;
// }

/*
 *  \brief search for the max dist point at curve abscica s
 *  \brief distMap : distance to obstacles map
 *  \param s : the current curve abscica
 *  \param pos : resulting cell position
 *  \param dist : resulting cell minimal distance to obstacles
 *  \param near : resulting cell nearest obstacle position
 *  \return true if a free position has been found, false else
 */
bool AvoidTrajGen::searchMaxDistPoint(const DistObsMap& distMap, double s, PositionXY& pos, float& dist, PositionXY& near)
{
  // get back border positions at curve abscica s
  XYThS_Config cfgB1; 
  border1_->getCfgAtCAbs(s, cfgB1);
  XYThS_Config cfgB2; 
  border2_->getCfgAtCAbs(s, cfgB2);

  obs2D_cell_t cellSt;
  bool foundFree = distMap.getCellMaxDistOnLine(cfgB1.x(), cfgB1.y(), cfgB2.x(), cfgB2.y(), pos, cellSt, dist, near);

  return foundFree;
}

/*
 *  \brief search for points maximizing the distance to obstacles in the navigation corridor
 *  \param obsmap : the obstacle map
 */
void AvoidTrajGen::computeMaxDistPoints(const BoolObsMap* obsmap)
{
  // compute a distance map from the input obstacle map
  DistObsMap distmap(obsmap);

  if(POSTER_DEBUG_ATG)
  {
    boost::mutex::scoped_lock lock(mutexDistMap_);
    if(distmap_ != NULL) delete distmap_;
    distmap_ = new DistObsMap(distmap);
  }

#if DEBUG_ATG
  if(image != NULL) delete image;
  image = distmap.rgbImage();

  this->drawXYThS_PathOnImage(*path_, image, RGBColor(0,255,255));

  int iImageNPP, jImageNPP;
  int failNPP = distmap.getIndexPPMImagePos(nearPathPos_->x(), nearPathPos_->y(), iImageNPP, jImageNPP);
  image->drawCross(iImageNPP, jImageNPP, 4, RGBColor(0,255,255));

  for(int num=0; num<vecNearPathPos_.size(); num++)
    {
      int iImageNP, jImageNP;
      int failNP = distmap.getIndexPPMImagePos(vecNearPathPos_.at(num).x(), vecNearPathPos_.at(num).y(), iImageNP, jImageNP);
      image->drawCross(iImageNP, jImageNP, 2, RGBColor(0,125,125));
    }
#endif

  // compute the total distance of the path from the current position to the goal
  double ds_tot = path_->lastCfg().s() - curS_;

  // compute the real ds used between max dist points
  int nbItv = ((int)(ds_tot / AvoidTrajGen::ds_maxDist_points_)) + 1;
  double real_ds = ds_tot / ((double)(nbItv));

  int itv = 1;
  bool colFound = false;

//   while(itv <= nbItv && !colFound && curS_ + itv*real_ds <= AvoidTrajGen::MAX_LENGTH_PATH_TTT)
//   while(itv <= nbItv && !colFound && curS_ + itv*real_ds <= max_length_)
  while(itv <= nbItv && !colFound && itv*real_ds <= max_length_)
    {
      // compute the current curve abscica
      double s = curS_ + itv*real_ds;

      // search for the current max dist point
      MaxDistPoint mdp;
      bool foundFree = this->searchMaxDistPoint(distmap, s, mdp.pos, mdp.dist, mdp.near);
      
      if(foundFree)
	{
	  mdp.ind = itv;
	  mdp.s = s;

	  maxDistPoints_.push_back(mdp);
	  itv++;
	}
      else colFound = true;
    }

  if(maxDistPoints_.size() == 0) 
    {
      cout << "=========================================" << endl;
      cout << "nbItv : " << nbItv << endl;
      cout << "colFound : " << colFound << endl;
      cout << "real_ds : " << real_ds << endl;
      obsmap->saveIntoPPM("obsmap_fail.ppm"); 
      distmap.saveIntoPPM("distmap_fail.ppm"); 
      cout << "=========================================" << endl;
      exit(1);
    }
}

/*
 *  \brief compute the current temporary goal
 */
void AvoidTrajGen::computeTempGoal()
{
  if(maxDistPoints_.size() == 0) { cerr << "AvoidTrajGen::computeTempGoal : maxDistPoints_.size() = 0!" << endl;  exit(1); }
  
  // get back the last free max dist point found
  MaxDistPoint lastMDP = maxDistPoints_.at(maxDistPoints_.size()-1);

  // get back the configuration of the followed path defined at curve absica of the last MDP
  XYThS_Config cfg;
  path_->getCfgAtCAbs(lastMDP.s, cfg);

  // compute the current goal position
  if(tempGoal_ != NULL) delete tempGoal_;
  tempGoal_ = new XYThS_Config(lastMDP.pos.x(), lastMDP.pos.y(), cfg.th(), cfg.s());

#if DEBUG_ATG
  int iImage, jImage, iImage2, jImage2;
  int fail = dem_->getIndexPPMImagePos(tempGoal_->x(), tempGoal_->y(), iImage, jImage);
  int fail2 = dem_->getIndexPPMImagePos(tempGoal_->x() + 2.5*cos(tempGoal_->th()), tempGoal_->y() + 2.5*sin(tempGoal_->th()), iImage2, jImage2);
  image->drawCross(iImage, jImage, 2, RGBColor(0,0,255));
  image->drawLine(iImage, jImage, iImage2, jImage2, RGBColor(0,0,255));
#endif
}

/*
 *  \brief search for a new strangling point
 *  \param indStart : the starting index of the search
 *  \param indEnd : the ending index of the search
 *  \param strang_point : the resulting strangling point
 */
void AvoidTrajGen::searchForStrangPoint(int indStart, int indEnd, MaxDistPoint& strang_point)
{
  float minDist = 999999999;
  int minInd = -1;
  
  for(int ind=indStart; ind<=indEnd; ind++)
    {
      if(ind == indStart || maxDistPoints_.at(ind).dist < minDist)
	{
	  minInd = ind;
	  minDist = maxDistPoints_.at(ind).dist;
	}
    }

  strang_point.ind = minInd;
  strang_point.s = maxDistPoints_.at(minInd).s;
  strang_point.pos.x() = maxDistPoints_.at(minInd).pos.x();
  strang_point.pos.y() = maxDistPoints_.at(minInd).pos.y();
  strang_point.dist = maxDistPoints_.at(minInd).dist;
  strang_point.near.x() = maxDistPoints_.at(minInd).near.x();
  strang_point.near.y() = maxDistPoints_.at(minInd).near.y();
}

/*
 *  \brief compute the tree of strangling points recursively
 */
void AvoidTrajGen::computeRecStrangPoints(BinNode<MaxDistPath>* nodeMDP)
{
  // check if both end points of the current node are close enough
  if(nodeMDP->val().mdp_end.ind - nodeMDP->val().mdp_beg.ind < 2*AvoidTrajGen::minNbItv_trajgen_)
    {
      // this branch is divided enough
      return;
    }
  else
    {
      // search for a new strangling point
      MaxDistPoint strang_point;
      this->searchForStrangPoint(nodeMDP->val().mdp_beg.ind + AvoidTrajGen::minNbItv_trajgen_, 
				 nodeMDP->val().mdp_end.ind - AvoidTrajGen::minNbItv_trajgen_, 
				 strang_point);

      // divide the current branch into two sub-branches
      MaxDistPath MDPath1;
      MDPath1.mdp_beg.ind = nodeMDP->val().mdp_beg.ind;
      MDPath1.mdp_beg.s = nodeMDP->val().mdp_beg.s;
      MDPath1.mdp_beg.pos.x() = nodeMDP->val().mdp_beg.pos.x();
      MDPath1.mdp_beg.pos.y() = nodeMDP->val().mdp_beg.pos.y();
      MDPath1.mdp_beg.dist = nodeMDP->val().mdp_beg.dist;
      MDPath1.mdp_beg.near.x() = nodeMDP->val().mdp_beg.near.x();
      MDPath1.mdp_beg.near.y() = nodeMDP->val().mdp_beg.near.y();

      MDPath1.mdp_end.ind = strang_point.ind;
      MDPath1.mdp_end.s = strang_point.s;
      MDPath1.mdp_end.pos.x() = strang_point.pos.x();
      MDPath1.mdp_end.pos.y() = strang_point.pos.y();
      MDPath1.mdp_end.dist = strang_point.dist;
      MDPath1.mdp_end.near.x() = strang_point.near.x();
      MDPath1.mdp_end.near.y() = strang_point.near.y();

      MaxDistPath MDPath2;
      MDPath2.mdp_beg.ind = strang_point.ind;
      MDPath2.mdp_beg.s = strang_point.s;
      MDPath2.mdp_beg.pos.x() = strang_point.pos.x();
      MDPath2.mdp_beg.pos.y() = strang_point.pos.y();
      MDPath2.mdp_beg.dist = strang_point.dist;
      MDPath2.mdp_beg.near.x() = strang_point.near.x();
      MDPath2.mdp_beg.near.y() = strang_point.near.y();

      MDPath2.mdp_end.ind = nodeMDP->val().mdp_end.ind;
      MDPath2.mdp_end.s = nodeMDP->val().mdp_end.s;
      MDPath2.mdp_end.pos.x() = nodeMDP->val().mdp_end.pos.x();
      MDPath2.mdp_end.pos.y() = nodeMDP->val().mdp_end.pos.y();
      MDPath2.mdp_end.dist = nodeMDP->val().mdp_end.dist;
      MDPath2.mdp_end.near.x() = nodeMDP->val().mdp_end.near.x();
      MDPath2.mdp_end.near.y() = nodeMDP->val().mdp_end.near.y();

      nodeMDP->setLeft(MDPath1);      
      nodeMDP->setRight(MDPath2);      

      this->computeRecStrangPoints(nodeMDP->left());
      this->computeRecStrangPoints(nodeMDP->right());
    }
}

/*
 *  \brief compute the list of strangling points from the tree
 */
void AvoidTrajGen::computeListStrangPoints()
{
  std::vector<MaxDistPath> vecLeaves;
  treeMDP_->getLeavesVal(vecLeaves);

  int N = vecLeaves.size();
  for(int i=0; i<N; i++) 
    {
      strangPoints_.push_back(vecLeaves.at(i).mdp_end);
//       cout << "Strangling point (" << i << ") -> x : " << strangPoints_.at(i).pos.x() << "  y : " << strangPoints_.at(i).pos.y() << endl;
    }
}

/*
 *  \brief add a first waypoint to the list of strangling points
 *  \param ind : index of the first strangling point
 *  \param (x,y) : location of the strangling point
 *  \param s : curve abscica reached at the initial location
 *  \param dist : distance to the nearest obstacle
 *  \param (nearX,nearY) : location of the nearest obstacle
 */
void AvoidTrajGen::addFirstStrangPoint(int ind, double x, double y, double s, double dist, double nearX, double nearY)
{
  // compute the MPD structure corresponding to the first point
  MaxDistPoint firstMPD;
  firstMPD.ind = ind;
  firstMPD.s = s;
  firstMPD.pos.x() = x;
  firstMPD.pos.y() = y;
  firstMPD.dist = dist;
  firstMPD.near.x() = nearX;
  firstMPD.near.y() = nearY;
  
  // add it to the list of strangling points
  strangPoints_.push_front(firstMPD);
}

/*
 *  \brief search for strangling points in the navigation corridor (from max dist points)
 *  \param obsmap : the obstacle map
 */
void AvoidTrajGen::computeStrangPoints()
{
  // compute the current goal position
  this->computeTempGoal();
  
  // reinit the tree if needed
  if(treeMDP_ != NULL) delete treeMDP_;

  // compute the root of the tree
  MaxDistPath root;
  root.mdp_beg.ind = 0;
  root.mdp_beg.s = curS_;
  root.mdp_beg.pos.x() = loc_->posXYTh().x();
  root.mdp_beg.pos.y() = loc_->posXYTh().y();
  root.mdp_beg.dist = width_;

  root.mdp_end.ind = maxDistPoints_.size() + 1;
  root.mdp_end.s = tempGoal_->s();
  root.mdp_end.pos.x() = tempGoal_->x();
  root.mdp_end.pos.y() = tempGoal_->y();
  root.mdp_end.dist = width_;

  treeMDP_ = new BinNode<MaxDistPath>(root);

  // compute the tree of strangling points recursively
  this->computeRecStrangPoints(treeMDP_);

  // compute the list of strangling points from the tree
  this->computeListStrangPoints();

  delete treeMDP_;
  treeMDP_ = NULL;
}

/*
 *  \brief search for strangling points in the navigation corridor (from max dist points) (2nd version)
 *  \param obsmap : the obstacle map
 */
void AvoidTrajGen::computeStrangPointsV2()
{
  // compute the current goal position
  this->computeTempGoal();
  
  // reinit the tree if needed
  if(treeMDP_ != NULL) delete treeMDP_;

  // compute the root of the tree
  MaxDistPath root;

  MaxDistPoint first_strang_point;
  if(maxDistPoints_.size() > 2.0*AvoidTrajGen::minNbItv_trajgen_)
    {
      // compute the strangling point from the initial position
      this->searchForStrangPoint(2, 
				 AvoidTrajGen::minNbItv_trajgen_, 
				 first_strang_point);  
      
      root.mdp_beg.ind = first_strang_point.ind;
      root.mdp_beg.s = first_strang_point.s;
      root.mdp_beg.pos.x() = first_strang_point.pos.x();
      root.mdp_beg.pos.y() = first_strang_point.pos.y();
      root.mdp_beg.dist = first_strang_point.dist;
      root.mdp_beg.near.x() = first_strang_point.near.x();
      root.mdp_beg.near.y() = first_strang_point.near.y();
    }
  else
    {
      root.mdp_beg.ind = 0;
      root.mdp_beg.s = curS_;
      root.mdp_beg.pos.x() = loc_->posXYTh().x();
      root.mdp_beg.pos.y() = loc_->posXYTh().y();
      root.mdp_beg.dist = width_;
    }

  root.mdp_end.ind = maxDistPoints_.size() + 1;
  root.mdp_end.s = tempGoal_->s();
  root.mdp_end.pos.x() = tempGoal_->x();
  root.mdp_end.pos.y() = tempGoal_->y();
  root.mdp_end.dist = width_;

  treeMDP_ = new BinNode<MaxDistPath>(root);

  // compute the tree of strangling points recursively
  this->computeRecStrangPoints(treeMDP_);

  // compute the list of strangling points from the tree
  this->computeListStrangPoints();

  // add the current location to the list of strangling points
  if(maxDistPoints_.size() > 2.0*AvoidTrajGen::minNbItv_trajgen_)
    {
      this->addFirstStrangPoint(first_strang_point.ind, 
				first_strang_point.pos.x(), 
				first_strang_point.pos.y(), 
				first_strang_point.s, 
				first_strang_point.dist, 
				first_strang_point.near.x(), 
				first_strang_point.near.y());
    }

  delete treeMDP_;
  treeMDP_ = NULL;
}

/*
 *  \brief compute the heading of the current waypoint
 *  \param posWP : position of the current waypoint
 *  \param posPrev : position of the previous waypoint
 *  \param posObs : position of the nearest obstacle
 */
double AvoidTrajGen::computeHeading(const PositionXY& posWP, const PositionXY& posPrev, const PositionXY& posObs)
{
  PositionXY prev_wp(posWP.x() - posPrev.x(), posWP.y() - posPrev.y());
  PositionXY wp_obs(posObs.x() - posWP.x(), posObs.y() - posWP.y());

  PositionXY wp_obs_plus_demiPI(-wp_obs.y(), wp_obs.x());
  PositionXY wp_obs_moins_demiPI(wp_obs.y(), -wp_obs.x());

  if(prev_wp.x()*wp_obs_plus_demiPI.x() + prev_wp.y()*wp_obs_plus_demiPI.y() >= 0) return atan2(wp_obs_plus_demiPI.y(), wp_obs_plus_demiPI.x());
  else return atan2(wp_obs_moins_demiPI.y(), wp_obs_moins_demiPI.x());
}

// /*
//  *  \brief search for waypoints in the navigation corridor (from strangling points)
//  */
// void AvoidTrajGen::computeWaypoints()
// {
//   // init the previous pos to the current state
//   PositionXY prevPos(loc_->posXYTh().x(), loc_->posXYTh().y());

//   int N = strangPoints_.size();
//   for(int i=0; i<N; i++) 
//     {
//       // compute the heading of the current waypoint
//       double th = this->computeHeading(strangPoints_.at(i).pos, prevPos, strangPoints_.at(i).near);

//       // build the new waypoint
//       WayPoint wp;
//       wp.cfg = XYThS_Config(strangPoints_.at(i).pos.x(), strangPoints_.at(i).pos.y(), th, strangPoints_.at(i).s);
//       wp.dist = strangPoints_.at(i).dist;

//       waypoints_.push_back(wp);

//       prevPos = PositionXY(wp.cfg.x(), wp.cfg.y());

// #if DEBUG_ATG
//       int iImage, jImage, iImage2, jImage2;
//       int fail = dem_->getIndexPPMImagePos(waypoints_.at(i).cfg.x(), waypoints_.at(i).cfg.y(), iImage, jImage);
//       int fail2 = dem_->getIndexPPMImagePos(waypoints_.at(i).cfg.x() + 2.5*cos(waypoints_.at(i).cfg.th()), waypoints_.at(i).cfg.y() + 2.5*sin(waypoints_.at(i).cfg.th()), iImage2, jImage2);
//       image->drawCross(iImage, jImage, 2, RGBColor(255,0,255));
//       image->drawLine(iImage, jImage, iImage2, jImage2, RGBColor(255,0,255));
// #endif
//     }
// }

/*
 *  \brief search for waypoints in the navigation corridor (from strangling points)
 */
void AvoidTrajGen::computeWaypoints()
{
  // init the previous pos to the current state
  PositionXY prevPos(loc_->posXYTh().x(), loc_->posXYTh().y());

#if DEBUG_ATG
  int i=0;
#endif

  for(list<MaxDistPoint>::iterator it = strangPoints_.begin(); it != strangPoints_.end(); it++) 
    {
      MaxDistPoint curMDP = *it;

      // compute the heading of the current waypoint
      double th = this->computeHeading(curMDP.pos, prevPos, curMDP.near);

      // build the new waypoint
      WayPoint wp;
      wp.cfg = XYThS_Config(curMDP.pos.x(), curMDP.pos.y(), th, curMDP.s);
      wp.dist = curMDP.dist;

      waypoints_.push_back(wp);

      prevPos = PositionXY(wp.cfg.x(), wp.cfg.y());

#if DEBUG_ATG
      int iImage, jImage, iImage2, jImage2;
      int fail = dem_->getIndexPPMImagePos(waypoints_.at(i).cfg.x(), waypoints_.at(i).cfg.y(), iImage, jImage);
      int fail2 = dem_->getIndexPPMImagePos(waypoints_.at(i).cfg.x() + 2.5*cos(waypoints_.at(i).cfg.th()), waypoints_.at(i).cfg.y() + 2.5*sin(waypoints_.at(i).cfg.th()), iImage2, jImage2);
      image->drawCross(iImage, jImage, 2, RGBColor(255,0,255));
      image->drawLine(iImage, jImage, iImage2, jImage2, RGBColor(255,0,255));

      i++;
#endif
    }
}

/**
 *  \brief compute the obstacle avoidance trajectory
 *  \return 1 if the goal has been reached
 *  \return 0 if the process succeeded,
 *  \return -1 if the avoid trajgen computation failed
 *  \return -2 if an I/O problem occured
 *  \return -3 if the avoid trajgen settings is not valid
 *  \return -4 if a possible collision or ICS may occur
 */
int AvoidTrajGen::computeAvoidTraj()
{
#if DISPLAY_TIME
  Timer timer;
  timer.start();
#endif

  int res = this->checkSettings();
  if(res != 0) return res;

  // reinit the avoidance trajectory computation process
  this->reinit();

  // check if the goal has been reached
  if(this->goalReached())
    {
      cout << "AvoidTrajGen::computeAvoidTraj : The goal position has been reached" << endl;
      return 1;
    }

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> Init : " << timer.elapsedTime() << endl;
#endif

  // compute the obstacle map from the reference corridor and the input DEM
  BoolObsMap* obsMap = NULL;
  {
    boost::mutex::scoped_lock lock(mutexDEM_);
    obsMap = this->buildObsMap(dem_);
  }
  if(obsMap == NULL) { cerr << "AvoidTrajGen::computeAvoidTraj : obstacle map computation failed" << endl;  return 2; }
#if DEBUG_ATG
  obsMap->saveIntoPPM("obsmap_with_borders2.ppm");
#endif

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> ObsMap computation : " << timer.elapsedTime() << endl;
#endif

  // search for points maximizing the distance to obstacles in the navigation corridor
  this->computeMaxDistPoints(obsMap);
  if(maxDistPoints_.size() == 0) { cerr << "AvoidTrajGen::computeAvoidTraj : Max dist points computation failed" << endl; return 2; }

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> Max dist points computation : " << timer.elapsedTime() << endl;
#endif

  // search for strangling points in the navigation corridor (from max dist points)
//   this->computeStrangPoints();
  this->computeStrangPointsV2();

//   // display the strangling points
//   this->displayStrangPoints();

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> Strang points computation : " << timer.elapsedTime() << endl;
#endif

  // search for waypoints in the navigation corridor (from strangling points)
  this->computeWaypoints();

//   // display the strangling points
//   this->displayWayPoints();

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> Waypoints computation : " << timer.elapsedTime() << endl;
#endif

  // select states from waypoints
  this->selectStatesFromWPs();

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> State selection : " << timer.elapsedTime() << endl;
#endif

  // generate a trajectory from the found waypoints
  int resTrajGen = this->computeTrajBtwnStates();
  if(resTrajGen != 0) { 
    cerr << "AvoidTrajGen::computeAvoidTraj : No collision free trajectory has been found" << endl; 
    res = -1;
  }

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> Trajectory generation : " << timer.elapsedTime() << endl;
#endif

#if DEBUG_ATG
  if(resTrajGen == 0) this->drawPSTrajOnImage(this->image);
#endif

#if CHECK_SAFETY
  // check for collision safety
  bool isSafe = this->checkSafety(*obsMap, *straj_);
  if(isSafe) res = 0;
  else res = -4;
#else
  res = 0;
#endif

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> Safety check : " << timer.elapsedTime() << endl;
#endif

  // delete the obstacle map
  delete obsMap;

#if DEBUG_ATG
  image->saveIntoPPM("distmap_avoidTG_debug.ppm");  
#endif

#if DISPLAY_TIME
  cout << "computeAvoidTraj -> End : " << timer.elapsedTime() << endl;
  cout << "computeAvoidTraj -> Total elapsed time : " << timer.totalElapsedTime() << endl;
#endif

  return res;
}


/****** Trajectory computation (2nd version) ******/

/*
 *  \brief reinit the avoidance trajectory computation process
 */
void AvoidTrajGen::reinitV2()
{
  // clear the vector of free passageways
  int N = vecFreeItv_.size();
  for(int i=0; i<N; i++) delete vecFreeItv_.at(i);
  vecFreeItv_.clear();

  // clear the graph of free passageways
  graphPsgwy_.clear();

  // clear the topological graph
  if(graphTopoPsgwy_ != NULL) 
    { 
      delete graphTopoPsgwy_;
      graphTopoPsgwy_ = NULL;
    }

  // clear the simplified graph
  if(simplifiedGraphPsgwy_ != NULL) 
    { 
      delete simplifiedGraphPsgwy_;
      simplifiedGraphPsgwy_ = NULL;
    }

  // clear the tree of free passageways
  if(treePsgwy_ != NULL) 
    {
      delete treePsgwy_;
      treePsgwy_ = NULL;
    }

#if DEBUG_ATG
  if(image != NULL)
    {
      delete image;
      image = NULL;
    }
  if(image2 != NULL)
    {
      delete image2;
      image2 = NULL;
    }
  if(image3 != NULL)
    {
      delete image3;
      image3 = NULL;
    }
  if(image4 != NULL)
    {
      delete image4;
      image4 = NULL;
    }
#endif
}

/**
 *  \brief return true if both passageways borders match
 *  \return true if both passageways borders match
 */
bool AvoidTrajGen::pasagewaysMatch(const Passageway& p1, const Passageway& p2)
{
  if((p1.minWidthB <= p2.maxWidthB) && (p2.minWidthB <= p1.maxWidthB)) return true;
  else return false;
}

/*
 *  \brief compute the free passageway intervals on a single discrete segment of the corridor
 *  \param distMap : distance to obstacles map
 *  \param s : current curve absica wrt. the followed trajectory
 *  \return a vector containing the current passageway intervals
 */
std::vector<Passageway>* AvoidTrajGen::computeSegmentItvs(const DistObsMap& distMap, float s)
{
//   cout << "AvoidTrajGen::computeSegmentItvs 1" << endl;

  // get back border positions at curve abscica s
  XYThS_Config cfgB1; 
  border1_->getCfgAtCAbs(s, cfgB1);
  XYThS_Config cfgB2; 
  border2_->getCfgAtCAbs(s, cfgB2);

  // get back the free segments on the distmap between these points
  vector<FreeSegment> vecSeg;
  int nbSgmts = distMap.getFreeSegmentsOnLine(cfgB1.x(), cfgB1.y(), cfgB2.x(), cfgB2.y(), vecSeg);

//   cout << "AvoidTrajGen::computeSegmentItvs 2 -> nbSgmts : " << nbSgmts << endl;

  // convert the free segments into passageways
  std::vector<Passageway>* vecPwy = new std::vector<Passageway>;

  for(int i=0; i<nbSgmts; i++)
    {
      if(vecSeg.at(i).length >= 2.0*sys_->getRadius()*COEF_SAFETY)
	{
	  Passageway pwy;
	  pwy.s = s;
	  pwy.minWidthB = vecSeg.at(i).minB;
	  pwy.maxWidthB = vecSeg.at(i).maxB;
	  pwy.maxDistPoint = vecSeg.at(i).maxDistPos;
	  pwy.dist = vecSeg.at(i).dist;
	  pwy.nearObs = vecSeg.at(i).nearObs;

	  vecPwy->push_back(pwy);
	}
    }

  // return the vector of passageways
  return vecPwy;
}

/*
 *  \brief compute the free passageway intervals on each discrete segment of the corridor
 *  \param distCol : if a collision has been found, contains the free distance in the front of the robot up to the obstacle
 *  \return 0 if the process succeeded, -1 if no passageway has been found
 */
int AvoidTrajGen::computePassagewayItv(float* distCol)
{
#if DEBUG_ATG
  if(image != NULL) delete image;
  image = distmap_->rgbImage();

  if(image2 != NULL) delete image2;
  image2 = distmap_->rgbImage();

  if(image3 != NULL) delete image3;
  image3 = distmap_->rgbImage();

  if(image4 != NULL) delete image4;
  image4 = distmap_->rgbImage();

  this->drawXYThS_PathOnImage(*path_, image, RGBColor(0,255,255));
  this->drawXYThS_PathOnImage(*path_, image2, RGBColor(0,255,255));
  this->drawXYThS_PathOnImage(*path_, image3, RGBColor(0,255,255));
  this->drawXYThS_PathOnImage(*path_, image4, RGBColor(0,255,255));
#endif

  // compute the total distance of the path from the current position to the goal
  double ds_tot = path_->lastCfg().s() - curS_;

  // compute the real ds used between max dist points
  int nbItv = ((int)(ds_tot / AvoidTrajGen::ds_maxDist_points_)) + 1;
  double real_ds = ds_tot / ((double)(nbItv));

  int itv = 1;
  bool colFound = false;

//   cout << "AvoidTrajGen::computePassagewayItv -> ds_tot : " << ds_tot << "  nbItv : " << nbItv << "  real_ds : " << real_ds << endl;

  while(itv <= nbItv && !colFound && itv*real_ds <= max_length_)
    {
      // compute the current curve abscica
      double s = curS_ + itv*real_ds;

      // search for the current passageway intervals
      std::vector<Passageway>* curSegmentItvs = this->computeSegmentItvs(*distmap_, s);

//       cout << "AvoidTrajGen::computePassagewayItv -> s : " << s << "  curSegmentItvs->size() : " << curSegmentItvs->size() << endl;
      
      if(curSegmentItvs->size() != 0)
	{
	  this->vecFreeItv_.push_back(curSegmentItvs);
	  itv++;
	}
      else colFound = true;
    }

  if(colFound) 
    {
      if(distCol != NULL) *distCol = (itv-1)*real_ds;
      return -1;
    }
  else return 0;
}

/*
 *  \brief compute the passageway graph
 *  \return 0 if the process succeeded, -1 if no passageway has been found
 */
int AvoidTrajGen::computePassagewayGraph()
{
  // consider initially the way to the goal as free
  bool noWay = false;
  float obstacleMinS = -1;

  // create a graph node from the current location of the robot
  Passageway initVertex_psgwy;
  initVertex_psgwy.s = this->curS_;
  initVertex_psgwy.minWidthB = 0;
  initVertex_psgwy.maxWidthB = 2.0*width_;
  initVertex_psgwy.maxDistPoint = PositionXY(loc_->posXYTh().x(), loc_->posXYTh().y());
  initVertex_psgwy.dist = width_;
  initVertex_psgwy.nearObs = PositionXY(loc_->posXYTh().x(), loc_->posXYTh().y());

  DGVertex_iterator<Passageway> initVertex_it = graphPsgwy_.addVertex(initVertex_psgwy);

  // create a vector of the current graph vertices iterators
  vector< DGVertex_iterator<Passageway> > curVerticesIt;
  // create a vector of the previous graph vertices iterators
  vector< DGVertex_iterator<Passageway> > prevVerticesIt;

  // add the init vertex to the previous vertices vector
  prevVerticesIt.push_back(initVertex_it);

  float prevS = 0;
  float curS = 0;

  // 
  int N = vecFreeItv_.size();
  for(int i=0; i<N && !noWay; i++)
    {
      // search for the current passageways
      std::vector<Passageway>* curVecPassageways = vecFreeItv_.at(i);

      int edgesCpt = 0;
      
      int N2 = curVecPassageways->size();
      for(int j=0; j<N2; j++)
	{
	  // create a node for each passageway found on the current segment
	  Passageway curPassageway = curVecPassageways->at(j);
	  DGVertex_iterator<Passageway> curVertex_it = graphPsgwy_.addVertex(curPassageway);
	  
	  // add it to the vector of current vertices
	  curVerticesIt.push_back(curVertex_it);

	  // save the current curve abscica
	  if(j == 0) curS = curPassageway.s;
	  
	  // check for each passageway, if it is reachable from the previous passagegways
	  // and add an edge for every reachable couple
	  int N3 = prevVerticesIt.size();
	  for(int k=0; k<N3; k++)
	    {
	      DGVertex_iterator<Passageway> prevVertex_it = prevVerticesIt.at(k);
	      Passageway prevPassageway = (*prevVertex_it)->val();
	      if(prevS == 0) prevS = prevPassageway.s;
	      
	      // check if the current passageway is reachable from the previous one
	      if(pasagewaysMatch(prevPassageway, curPassageway))
		{
		  edgesCpt++;
		  graphPsgwy_.addEdge(prevVertex_it, curVertex_it);
		}
	    }	  
	}

      // check if at least an edge has been added
      if(edgesCpt == 0 && !noWay)
	{
	  noWay = true;
	  obstacleMinS = prevS;
	}
      
      // replace the vector of previous vertices with the vector of current ones
      prevVerticesIt.clear();
      int N4 = curVerticesIt.size();
      for(int l=0; l<N4; l++) prevVerticesIt.push_back(curVerticesIt.at(l));
      curVerticesIt.clear();

      // replace the previous cruve absica with the current one;
      prevS = curS;
    }
  
  // check if a way has been found up to the goal
  if(noWay)
    {
      // compute the minimal distance to stop
      double minDistToStop = this->distToStop();

      // check if the robot will be able to stop before colliding an obstacle
      if(minDistToStop < (obstacleMinS - this->curS_)) return 0;
      else return -1;
    }
  else return 0;
}

/*
 *  \brief compute the topological graph
 */
void AvoidTrajGen::computeTopologicalGraph()
{
  graphTopoPsgwy_ = new GraphTopoPassageway(graphPsgwy_);
}

/*
 *  compute the strangling points on each node of the topological graph
 */
void AvoidTrajGen::computeStrangPointsTopoGraph()
{
//   cout << "min dist strang points : " << COEF_MINDIFF_STRANG_POINTS * ds_maxDist_points_ << endl;
  graphTopoPsgwy_->simplifyIntoSPGraph(COEF_MINDIFF_STRANG_POINTS * ds_maxDist_points_);
}

/*
 *  \brief convert the topological graph back to a simple passageway graph
 */
void AvoidTrajGen::convertTopoToSimpleGraph()
{
  simplifiedGraphPsgwy_ = graphTopoPsgwy_->toGraphPassageway();
}

/*
 *  remove the shortest and non collision-free edges
 */
void AvoidTrajGen::removeShortAndColEdges()
{
  DistObsMap* copyDistMap = this->distmap();
  if(copyDistMap == NULL) 
    {
      cout << "AvoidTrajGen::removeShortAndColEdges : copyDistMap is NULL" << endl;
      exit(1);
    }
//   simplifiedGraphPsgwy_->removeShortAndColEdges(COEF_MINDIFF_STRANG_POINTS * ds_maxDist_points_, *distmap_);
  simplifiedGraphPsgwy_->removeShortAndColEdges(COEF_MINDIFF_STRANG_POINTS * ds_maxDist_points_, copyDistMap);
  delete copyDistMap;
}

/*
 *  \brief simplify the passageway graph into a strangling points graph
 */
void AvoidTrajGen::simplifyPassagewayGraph()
{
  // compute the topological graph
  this->computeTopologicalGraph();

  // compute the strangling points on each node of the topological graph
  this->computeStrangPointsTopoGraph();

  // convert the topological graph back to a simple passageway graph
  this->convertTopoToSimpleGraph();

  // remove the shortest and non collision-free edges
  this->removeShortAndColEdges();
}

/*
 *  \brief search for the best state-time sequence (minimizing the final time)
 */
void AvoidTrajGen::searchBestBranch()
{
  cerr << "AvoidTrajGen::searchBestBranch : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief compute the obstacle avoidance trajectory (2nd version : Waypoint graph)
 *  \return 0 if the process succeeded, -1 else
 */
int AvoidTrajGen::computeAvoidTrajV2()
{
#if DISPLAY_TIME
  Timer timer;
  timer.start();
#endif

  int res = this->checkSettings();
  if(res != 0) return res;

  // reinit the avoidance trajectory computation process
  this->reinitV2();

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> Init : " << timer.elapsedTime() << endl;
#endif

  // compute the obstacle map from the reference corridor and the input DEM
  BoolObsMap* obsMap = NULL;
  {
    boost::mutex::scoped_lock lock(mutexDEM_);
    obsMap = this->buildObsMap(dem_);
  }
  if(obsMap == NULL) { cerr << "AvoidTrajGen::computeAvoidTraj : obstacle map computation failed" << endl;  return 2; }
#if DEBUG_ATG
  obsMap->saveIntoPPM("obsmap_with_borders_v2.ppm");
#endif

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> ObsMap computation : " << timer.elapsedTime() << endl;
#endif

  // compute a distance map from the input obstacle map
  {
    boost::mutex::scoped_lock lock(mutexDistMap_);
    if(distmap_ != NULL) delete distmap_;
    distmap_ = new DistObsMap(obsMap);
  }

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> DistMap computation : " << timer.elapsedTime() << endl;
#endif

  // compute the free passageway intervals on each discrete segment of the corridor
  float distCol;
  res = this->computePassagewayItv(&distCol);
  if(res == -1 && distCol < this->distToStop()) 
    {
      cerr << "AvoidTrajGen::computeAvoidTrajV2 : No free passageway has been found. maximal free distance : " << distCol << endl;
      return -1;
    }

// #if DEBUG_ATG                   
//   cout << "computeAvoidTrajV2 -> Draw passageway graph" << endl;
//   this->drawPassagewayItvOnImage(image, RGBColor(255,255,0));
//   image->saveIntoPPM("distmap_passagewayItv.ppm");  
// #endif  

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> Passageway itv computation : " << timer.elapsedTime() << endl;
#endif

  // compute the passageway graph
  this->computePassagewayGraph();

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> Passageway graph computation : " << timer.elapsedTime() << endl;
#endif

#if DEBUG_ATG
  cout << "computeAvoidTrajV2 -> Draw passageway graph" << endl;
  this->drawPassagewayGraphOnImage(graphPsgwy_, image, RGBColor(255,255,255));
  image->saveIntoPPM("distmap_passagewayGraph_1.ppm");  
#endif
  
  // simplify the passageway graph into a strangling points graph
  this->simplifyPassagewayGraph();

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> Passageway graph simplification : " << timer.elapsedTime() << endl;
#endif

#if DEBUG_ATG
  cout << "Topological graph : " << endl;
  graphTopoPsgwy_->display();

  cout << "computeAvoidTrajV2 -> Draw topological graph" << endl;
  this->drawTopoPwyGraphOnImage(image2, RGBColor(255,255,255), RGBColor(0,0,255));
  image2->saveIntoPPM("distmap_topoGraph_2.ppm");  
#endif

#if DEBUG_ATG
  cout << "computeAvoidTrajV2 -> Draw simplified graph" << endl;
  simplifiedGraphPsgwy_->display();
  this->drawPassagewayGraphOnImage(*simplifiedGraphPsgwy_, image3, RGBColor(255,255,255));
  image3->saveIntoPPM("distmap_simplifiedGraph_3.ppm");  
#endif

  // compute the state-time tree corresponding to each possible trajectories
  this->computeStateTimeTree();

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> State-time tree computation : " << timer.elapsedTime() << endl;
#endif

  // search for the best state-time sequence (minimizing the final time)
  this->searchBestBranch();

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> Best branch search : " << timer.elapsedTime() << endl;
#endif

  // generate a trajectory along the best branch found
  int resTrajGen = this->computeTrajBtwnStates();
  if(resTrajGen != 0) { 
    cerr << "AvoidTrajGen::computeAvoidTrajV2 : No collision free trajectory has been found" << endl; 
    res = -1;
  }

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> Trajectory generation : " << timer.elapsedTime() << endl;
#endif

#if DEBUG_ATG
  if(resTrajGen == 0) this->drawPSTrajOnImage(image4);
  image4->saveIntoPPM("distmap_avoidTG_4.ppm");  
#endif

#if CHECK_SAFETY
  // check for collision safety
  bool isSafe = this->checkSafety(*obsMap, *straj_);
  if(isSafe) res = 0;
  else res = 1; 
#else
  res = 0;
#endif

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> Safety check : " << timer.elapsedTime() << endl;
#endif

  // delete the obstacle map
  delete obsMap;

// #if DEBUG_ATG
//   image->saveIntoPPM("distmap_avoidTG_debug_v2.ppm");  
// #endif

#if DISPLAY_TIME
  cout << "computeAvoidTrajV2 -> End : " << timer.elapsedTime() << endl;
  cout << "computeAvoidTrajV2 -> Total elapsed time : " << timer.totalElapsedTime() << endl;
#endif

    return res;
}


/***************** Collision check ****************/

/*
 *  \brief check for collision safety
 *  \param map : the obstacle map
 *  \param pstraj : the input state trajectory
 *  \return true if the trajectory is safe up to its end
 */
bool AvoidTrajGen::checkSafety(const BoolObsMap& map, const STraj& straj)
{
  // compute the maximal time until which the computed trajectory is safe
  double maxSafeTime = CollisionCheck::checkCol(map, straj);

  // compute the time required by the robot to stop
  double timeStop = this->timeToStop();

  if(maxSafeTime == INFINITY) return true;
  else 
    {
      if(maxSafeTime < COEF_TIME_STOP * timeStop) return false;
      else return true;
    }
}


/****************** I/O functions *****************/

/*
 *  \brief update the localization and speed of the robot
 *  \param s : the current localization and speed
 */
void AvoidTrajGen::updateState(const State& s)
{
  if(loc_ != NULL) delete loc_;
  loc_ = s.copy();

//   Timer::absoluteTime(dateLoc_);
}

/**
 *  \brief update the Data Elevation Map
 *  \param dem : the input DEM
 *  \return 0 if the process succeed, -1 else
 */
int AvoidTrajGen::updateDEM(const DEM* dem)
{
  boost::mutex::scoped_lock lock(mutexDEM_);
  if(dem_ != NULL) delete dem_;

  if(dem == NULL) 
    {
      dem_ = NULL;
      return -1;
    }
  else 
    {
      dem_ = dem->copy();
      return 0;
    }     
}

/**
 *  \brief return true if the goal has been reached
 *  \return true if the goal has been reached
 */
bool AvoidTrajGen::goalReached() const
{
  // check if the current location has been defined
  if(this->loc_ == NULL) return false;
  // check if the followed path has been defined
  if(this->path_ == NULL) return false;

  // get back the current location
  PositionXYTh curPos = this->curPos();
  // get back the current curve abscica
  double curS = this->curS_;
  // get back the goal position to reach
  XYThS_Config goalCfg = this->goalCfg();

  // check if the path already travelled is long enough to possibly have reached the goal
  if(curS >= goalCfg.s() - AvoidTrajGen::maxDistWpReached_)
    {
      // compute the distance between the current and goal positions
      float dx = curPos.x() - goalCfg.x();
      float dy = curPos.y() - goalCfg.y();
      float sqDist = dx*dx + dy*dy;
      
      float goalToCur_head = atan2(dy,dx);
      double diffAngle = fabs(computeMinDiffAngle(goalToCur_head, goalCfg.th()));
      
      if(sqDist < AvoidTrajGen::maxDistTgtReached_ * AvoidTrajGen::maxDistTgtReached_)
	{
      return true;
	}
      else if(sqDist < AvoidTrajGen::maxDistWpReached_ * AvoidTrajGen::maxDistWpReached_ && diffAngle < AvoidTrajGen::maxAngleWpReached_)
	{
	  return true;
	}
      else return false;
    }
  else return false;
}


/************ Debug / Display functions ***********/

/*
 *  \brief display the list of max dist points
 */
void AvoidTrajGen::displayMaxDistPoints() const
{
  cout << "MaxDistPoints : { ";

  int N = maxDistPoints_.size();
  for(int i=0; i<N; i++)
    {
      MaxDistPoint mdp = maxDistPoints_.at(i);
      cout << "(ind=" << mdp.ind << ", s=" << mdp.s << ", pos=(" << mdp.pos.x() << "," <<  mdp.pos.y() << "), dist=" << mdp.dist << ", near=(" << mdp.near.x() << "," << mdp.near.y() << ")); ";
    }

  cout << "}" << endl;
}

/*
 *  \brief display the list of strangling points
 */
void AvoidTrajGen::displayStrangPoints() const
{
  cout << "StrangPoints : { ";

  for(list<MaxDistPoint>::const_iterator it = strangPoints_.begin(); it != strangPoints_.end(); it++)
    {
      MaxDistPoint strangPoint = *it;
      cout << "(ind=" << strangPoint.ind << ", s=" << strangPoint.s << ", pos=(" << strangPoint.pos.x() << "," <<  strangPoint.pos.y() << "), dist=" << strangPoint.dist << ", near=(" << strangPoint.near.x() << "," << strangPoint.near.y() << ")); ";      
    }

  cout << "}" << endl;
}

/*
 *  \brief display the list of waypoints
 */
void AvoidTrajGen::displayWayPoints() const
{
  cout << "WayPoints : { ";
  
  int N = waypoints_.size();
  for(int i=0; i<N; i++)
    {
      WayPoint wp = waypoints_.at(i);
      cout << "(cfg="; wp.cfg.display(); cout << ", dist=" << wp.dist << "); ";
    }

  cout << "}" << endl;
}

#if DEBUG_ATG
/*
 *  \brief draw a path on the image
 *  \param path : the path to draw
 *  \param img : a pointer to the image to modify
 *  \param col : the color of the path
 */
void AvoidTrajGen::drawXYThS_PathOnImage(const XYThS_Path& path, RGBImage* img, const RGBColor& col)
{
  if(img == NULL) { cerr << "AvoidTrajGen::drawXYThS_PathOnImage : The input image is null" << endl; return; }

  int N = path.size();
  for(int i=0; i<N-1; i++)
    {
      XYThS_Config curCfg = path.at(i);
      XYThS_Config nextCfg = path.at(i+1);

      int iImage, jImage, iImage2, jImage2;
      int fail = dem_->getIndexPPMImagePos(curCfg.x(), curCfg.y(), iImage, jImage);
      int fail2 = dem_->getIndexPPMImagePos(nextCfg.x(), nextCfg.y(), iImage2, jImage2);
      img->drawLine(iImage, jImage, iImage2, jImage2, col);
    }
}

/*
 *  \brief draw the passageway intervals found on the image
 *  \param img : a pointer to the image to modify
 *  \param col : the color of the intervals
 */
void AvoidTrajGen::drawPassagewayItvOnImage(RGBImage* img, const RGBColor& col)
{
  // check if the image is not null
  if(img == NULL) { cerr << "AvoidTrajGen::drawPassagewayItvOnImage : The output image is null. Cannot draw the graph." << endl; return; }

  // check if the dem is not null
  if(dem_ == NULL) { cerr << "AvoidTrajGen::drawPassagewayItvOnImage : The data elevation map is null. Cannot draw the graph." << endl; return; }

  int M = vecFreeItv_.size();
  for(int i=0; i<M; i++)
    {
      std::vector<Passageway>* vecPswy = vecFreeItv_.at(i);

      if(vecPswy == NULL) { cerr << "AvoidTrajGen::drawPassagewayItvOnImage : i=" << i << " -> vecFreeItv_.at(i) is null." << endl; }
      else
	{
	  int N = vecFreeItv_.at(i)->size();
	  for(int j=0; j<N; j++)
	    {
	      PositionXY pos = vecFreeItv_.at(i)->at(j).maxDistPoint;

	      // get back the coordinates of pos on the image
	      int iImage, jImage;
	      int fail = dem_->getIndexPPMImagePos(pos.x(), pos.y(), iImage, jImage);
	      
	      // draw a cross on the image at position pos
	      img->drawCross(iImage, jImage, 1, col);
	    }
	}
    }
}

/*
 *  \brief draw the passageway graph computed on the image
 *  \param graph : the passageway graph to display
 *  \param img : a pointer to the image to modify
 *  \param col : the color of the graph
 */
void AvoidTrajGen::drawPassagewayGraphOnImage(GraphPassageway& graph, RGBImage* img, const RGBColor& col)
{
  // check if the image is not null
  if(img == NULL) { cerr << "AvoidTrajGen::drawPassagewayGraphOnImage : The output image is null. Cannot draw the graph." << endl; return; }

  // check if the dem is not null
  if(dem_ == NULL) { cerr << "AvoidTrajGen::drawPassagewayGraphOnImage : The data elevation map is null. Cannot draw the graph." << endl; return; }

  // draw each vertex of the graph
  for(std::list< DGVertex<Passageway>* >::const_iterator it = graph.vertices().begin(); it != graph.vertices().end(); it++)
    {
      // get back the maxDistPoint of the current vertex
      PositionXY pos = (*it)->val().maxDistPoint;

//       cout << "graph vertex : (" << pos.x() << "," << pos.y() << ")" << endl;

      // get back the coordinates of pos on the image
      int iImage, jImage;
      int fail = dem_->getIndexPPMImagePos(pos.x(), pos.y(), iImage, jImage);

      // draw a cross on the image at position pos
      img->drawCross(iImage, jImage, 1, col);
    }

//   cout << "AvoidTrajGen::drawPassagewayGraphOnImage -> draw edges" << endl;


//   for(typename std::list< DGEdge<Passageway>* >::iterator it2 = graph.edges().begin(); it2 != graph.edges().end(); it2++) 
//     {
//       (*it2)->startVertex()->val().maxDistPoint.display(); 
//       cout << " -> "; 
//       (*it2)->endVertex()->val().maxDistPoint.display(); 
//       cout << endl;
//     }


  // draw each edge of the graph
  for(std::list< DGEdge<Passageway>* >::const_iterator it = graph.edges().begin(); it != graph.edges().end(); it++)
    {
      // get back both maxDistPoint of the end points of the current edge
      PositionXY pos1 = (*it)->startVertex()->val().maxDistPoint;
      PositionXY pos2 = (*it)->endVertex()->val().maxDistPoint;

      // get back the coordinates of pos1 and pos2 on the image
      int iImage1, jImage1, iImage2, jImage2;
      int fail1 = dem_->getIndexPPMImagePos(pos1.x(), pos1.y(), iImage1, jImage1);
      int fail2 = dem_->getIndexPPMImagePos(pos2.x(), pos2.y(), iImage2, jImage2);
      
      img->drawLine(iImage1, jImage1, iImage2, jImage2, col);
    }
}

/*
 *  \brief draw the topological passageway graph computed on the image
 *  \param img : a pointer to the image to modify
 *  \param colVertices : the color of the vertices of the graph
 *  \param colEdges : the color of the edges of the graph
 */
void AvoidTrajGen::drawTopoPwyGraphOnImage(RGBImage* img, const RGBColor& colVertices, const RGBColor& colEdges)
{
  // check if the image is not null
  if(img == NULL) { cerr << "AvoidTrajGen::drawTopoPwyGraphOnImage : The output image is null. Cannot draw the graph." << endl; return; }

  // check if the dem is not null
  if(dem_ == NULL) { cerr << "AvoidTrajGen::drawTopoPwyGraphOnImage : The data elevation map is null. Cannot draw the graph." << endl; return; }

  // draw each vertex of the graph
  for(std::list< DGVertex<VecPassageway*>* >::const_iterator it = graphTopoPsgwy_->vertices().begin(); it != graphTopoPsgwy_->vertices().end(); it++)
    {
      // get back the current passageway vector
      VecPassageway* vecPwy = (*it)->val();

      // check if the passageway vector is not null
      if(vecPwy == NULL)
	{
	  cerr << "AvoidTrajGen::drawTopoPwyGraphOnImage : Warning. The passageway vector is null" << endl;
	  return;
	}

      int N = vecPwy->size();
      PositionXY prev;

      for(int i=0; i<N; i++)
	{
	  // get back the maxDistPoint of the current vertex
	  PositionXY pos = vecPwy->at(i)->maxDistPoint;
	  
	  // get back the coordinates of pos on the image
	  int iImage1, jImage1;
	  int fail = dem_->getIndexPPMImagePos(pos.x(), pos.y(), iImage1, jImage1);
	  
	  // draw a cross on the image at position pos
	  img->drawCross(iImage1, jImage1, 1, colVertices);

	  if(i != 0)
	    {
	      // get back the coordinates of pos on the image
	      int iImage2, jImage2;
	      int fail2 = dem_->getIndexPPMImagePos(prev.x(), prev.y(), iImage2, jImage2);
	      
	      img->drawLine(iImage1, jImage1, iImage2, jImage2, colVertices);
	    }

	  prev = pos;
	}
    }

  // draw each edge of the graph
  for(std::list< DGEdge<VecPassageway*>* >::const_iterator it = graphTopoPsgwy_->edges().begin(); it != graphTopoPsgwy_->edges().end(); it++)
    {
      // get back both maxDistPoint of the end points of the current edge
      PositionXY pos1 = (*it)->startVertex()->val()->last()->maxDistPoint;
      PositionXY pos2 = (*it)->endVertex()->val()->first()->maxDistPoint;

      // get back the coordinates of pos1 and pos2 on the image
      int iImage1, jImage1, iImage2, jImage2;
      int fail1 = dem_->getIndexPPMImagePos(pos1.x(), pos1.y(), iImage1, jImage1);
      int fail2 = dem_->getIndexPPMImagePos(pos2.x(), pos2.y(), iImage2, jImage2);
      
      img->drawLine(iImage1, jImage1, iImage2, jImage2, colEdges);
    }

}
#endif
