#include "avoidObs/inc/DD2AvoidTrajGen.hpp"
using namespace mpc;

#include "gnuplot/inc/TrajPlot.hpp"

#define COEF_T 1.5

#define VARYING_SPEED false

                                    /**************************************************/
                                    /************** Class DD2AvoidTrajGen *************/
                                    /**************************************************/

double DD2AvoidTrajGen::coef_min_available_speed = 0.1;
double DD2AvoidTrajGen::coef_max_available_speed = 1.0;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2AvoidTrajGen::DD2AvoidTrajGen() : AvoidTrajGen()
{
  builtPCTraj_ = NULL;
  builtPSTraj_ = NULL;

  builtCTraj_ = NULL;
  builtSTraj_ = NULL;
}

/**
 *  \brief default constructor
 */
DD2AvoidTrajGen::DD2AvoidTrajGen(ManaSys* sys, XYThS_Path* path, double max_length, double width, double fov) : AvoidTrajGen(sys, path, max_length, width, fov)
{
  builtPCTraj_ = NULL;
  builtPSTraj_ = NULL;

  builtCTraj_ = NULL;
  builtSTraj_ = NULL;

  this->trajgen_ = new ManaTrajGen(sys);

  this->initTrajGen();
  this->initDD2TrajGen();
}

/**
 *  \brief destructor
 */
DD2AvoidTrajGen::~DD2AvoidTrajGen()
{
  if(builtPCTraj_ != NULL) delete builtPCTraj_;
  if(builtPSTraj_ != NULL) delete builtPSTraj_;

  if(builtCTraj_ != NULL) delete builtCTraj_;
  if(builtSTraj_ != NULL) delete builtSTraj_;

  if(trajgen_ != NULL) delete trajgen_;

  wp_statet_.clear();
}


/*************** Getters and setters **************/

/**
 *  \brief return a pointer to the Mana last registered localization
 *  \return a pointer to the Mana last registered localization
 */
DD2State* DD2AvoidTrajGen::loc() const 
{
  return ((DD2State*)(this->loc_));
}

/**
 *  \brief return a pointer to the Mana trajectory generator
 *  \return a pointer to the Mana trajectory generator
 */
ManaTrajGen* DD2AvoidTrajGen::trajgen()
{
  return ((ManaTrajGen*)(this->trajgen_));
}


/************* Initialization functons ************/

/**
 *  \brief Initialize the parameters of the Mana trajectory generator
 */
void DD2AvoidTrajGen::initDD2TrajGen()
{
  // Differential drive constraints
  trajgen()->setUseVelCst(false);               // default = true
  trajgen()->setUseAccCst(false);               // default = true
  trajgen()->setUseAVelCst(false);              // default = true
  trajgen()->setUseAAccCst(false);              // default = true

  // final state constraints
  trajgen()->fixFinalXY(true);                  // default = true
  trajgen()->fixFinalTH(true);                  // default = true
  trajgen()->fixFinalV(false);                  // default = true
  trajgen()->fixFinalOM(false);                 // default = true
}


/************* Trajectory computation *************/

/**
 *  \brief compute the time coefficient used to select the time to reach waypoints
 *  \param minDistToObs : the current minimal distance to obstacles
 *
 *  The return coefficient belongs to [1.0/coef_max_available_speed ; 1.0/coef_min_available_speed]
 */
double DD2AvoidTrajGen::computeCoefTime(double minDistToObs)
{
#if VARYING_SPEED
//   double vmin = coef_min_available_speed * sys()->getVMax();
//   double vmax = coef_max_available_speed * sys()->getVMax();

  double dmin = AvoidTrajGen::SAFE_DIST;
  double dmax = this->width_;

//   double v_allowed = 0.0;
  double coef_allowed = 1.0;
//   if(minDistToObs < dmin) v_allowed = vmin;
  if(minDistToObs < dmin) coef_allowed = coef_min_available_speed;
  else if(minDistToObs <= dmax)
    {
//       v_allowed = vmin + (vmax - vmin) * (minDistToObs - dmin) / (dmax - dmin);
      coef_allowed = coef_min_available_speed + (coef_max_available_speed - coef_min_available_speed) * (minDistToObs - dmin) / (dmax - dmin);
    }
  else 
    {
      cerr << "!!!!!!!!!!!!!!!!!!!!!!!!!!!! DD2AvoidTrajGen::computeCoefTime  minDistToObs > dmax. This case should not appear !!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
//       v_allowed = vmax; // this case should not appear
      coef_allowed = coef_max_available_speed; // this case should not appear
    }

//   return (1.0 / v_allowed);
  return (1.0 / coef_allowed);

#else
  return AvoidTrajGen::coefTime_;
#endif
}

/*
 *  \brief select states from waypoints
 */
void DD2AvoidTrajGen::selectStatesFromWPs()
{
  double prevT = 0;
  double curV = 0;
  double prevS = curS_;

  int N = waypoints_.size();
  for(int i=0; i<N; i++) 
    {
//       double ds = waypoints_.at(i).s() - prevS;
      double ds = waypoints_.at(i).cfg.s() - prevS;

      // compute the minimum time to reach the next waypoint
      double minTime = minTravelTime(ds, curV, sys()->getVMax(), sys()->getAMax());

//       double dt = AvoidTrajGen::coefTime_ * minTime;
      // get back the minimal distance to obstacles
      double minDistToObs = waypoints_.at(i).dist;
      double dt = this->computeCoefTime(minDistToObs) * minTime;

//       cout << "i : " << i << "  minDistToObs : " << minDistToObs << "  computeCoefTime(minDistToObs) : " << computeCoefTime(minDistToObs) << endl;

      if(dt != 0) curV = ds / dt;
      else curV = 0;
      
      // compute the state-time to reach
      DD2StateT statet(waypoints_.at(i).cfg.x(),
		       waypoints_.at(i).cfg.y(), 
		       waypoints_.at(i).cfg.th(), 
		       curV,
		       0.0,
		       prevT + dt);

//       cout << "statet : "; statet.display(); cout << endl;

      wp_statet_.push_back(statet);

      prevT += dt;
      prevS += ds;
    }
}

/*
 *  \brief generate a trajectory from the found waypoints
 *  \return 0 if a free trajectory has been found, -1 else              ------------------- actually, the process does never fail...
 */
int DD2AvoidTrajGen::computeTrajBtwnStates()
{
  // reinit the temporary trajectories
  if(builtPCTraj_ != NULL) { delete builtPCTraj_;  builtPCTraj_ = NULL; }
  if(builtPSTraj_ != NULL) { delete builtPSTraj_;  builtPSTraj_ = NULL; }

  if(builtCTraj_ != NULL) { delete builtCTraj_;  builtCTraj_ = NULL; }
  if(builtSTraj_ != NULL) { delete builtSTraj_;  builtSTraj_ = NULL; }
  if(pctraj_ != NULL) { delete pctraj_;  pctraj_ = NULL; }

  // init the current state-time
  DD2StateT initST(*((DD2State*)loc_), 0);
  DD2StateT curST = initST;

  int N = wp_statet_.size();
  for(int i=0; i<N; i++) 
//   for(int i=0; i<=0; i++) 
    {
      // get back the current goal waypoint
      DD2StateT goalST = wp_statet_.at(i);

      goalST.t() -= curST.t();
      curST.t() = 0;

      // generation of a trajectory between s0 and sf
      // return 0 if the process succeed, -1 else
      int res = trajgen()->genTraj(curST,goalST);

//       cout << "curST : "; curST.display(); cout << "  goalST : "; goalST.display(); cout << "  res : " << res << endl;

//       if(pctraj_ == NULL) pctraj_ = trajgen()->getPCTraj();
      if(builtPCTraj_ == NULL) builtPCTraj_ = trajgen()->getPCTraj();
      else
	{
	  DD2PCTraj* curPCTraj = trajgen()->getPCTraj();
	  builtPCTraj_->concat(*curPCTraj);
	  delete curPCTraj;
	}

      curST = wp_statet_.at(i);
    }

//   cout << "PCTraj : " << endl;
//   builtPCTraj_->display();
//   TrajPlot::plotDD2PCTraj(*builtPCTraj_);

  // build a state trajectory from the control trajectory
  builtPSTraj_ = new DD2PSTraj(*((DD2State*)loc_), *builtPCTraj_);

//   // check for collision safety
//   double maxSafeTime;
//   bool isSafeUntilEnd = this->checkSafety(builtPSTraj_, maxSafeTime);

  // replace the resulting control and state trajectories
  {
    boost::mutex::scoped_lock lock(mutexTraj_);
    if(pctraj_ != NULL) delete pctraj_;
    pctraj_ = new DD2PCTraj(*builtPCTraj_);

//     if(pstraj_ != NULL) delete pstraj_;
//     pstraj_ = new DD2PSTraj(*builtPSTraj_);
  }

  // replace the resulting control and state trajectories
  {
    boost::mutex::scoped_lock lock(mutexTraj_);
    if(straj_ != NULL) delete straj_;
    straj_ = builtPSTraj_->sample(100);
  }

  return 0;
}

// /*
//  *  \brief generate a trajectory from the found waypoints
//  *  \return 0 if a free trajectory has been found, -1 else
//  */
// int DD2AvoidTrajGen::computeTrajBtwnStates_old()
// {
//   // reinit the temporary trajectories
//   if(builtCTraj_ != NULL) { delete builtCTraj_;  builtCTraj_ = NULL; }
//   if(builtSTraj_ != NULL) { delete builtSTraj_;  builtSTraj_ = NULL; }
//   if(pctraj_ != NULL) { delete pctraj_;  pctraj_ = NULL; }

//   // init the current state-time
//   DD2StateT initST(*((DD2State*)loc_), 0);
//   DD2StateT curST = initST;

//   int N = wp_statet_.size();
// //   for(int i=0; i<N; i++) 
//   for(int i=0; i<=0; i++) 
//     {
//       // get back the current goal waypoint
//       DD2StateT goalST = wp_statet_.at(i);

//       goalST.t() -= curST.t();
//       curST.t() = 0;

//       // generation of a trajectory between s0 and sf
//       // return 0 if the process succeed, -1 else
//       int res = trajgen()->genTraj(curST,goalST);

//       if(pctraj_ == NULL) pctraj_ = trajgen()->getPCTraj();

//       DD2PCTraj* curPCTraj = trajgen()->getPCTraj();
//       DD2PSTraj* curPSTraj = new DD2PSTraj(DD2State(curST.x(), curST.y(), curST.th(), curST.v(), curST.w()), *curPCTraj);
//       DD2STraj* curSTraj = curPSTraj->sample(20);
//       curSTraj->shift(curST.t());
//       delete curPCTraj;
//       delete curPSTraj;

//       if(builtSTraj_ == NULL) builtSTraj_ = curSTraj;
//       else
// 	{
// 	  builtSTraj_->concat(*curSTraj);
// 	  delete curSTraj;
// 	}

//       curST = wp_statet_.at(i);
//     }

//   // replace the resulting control and state trajectories
//   {
//     boost::mutex::scoped_lock lock(mutexTraj_);
//     if(straj_ != NULL) delete straj_;
//     straj_ = new DD2STraj(*builtSTraj_);
//   }

//   return 0;
// }


/****** Trajectory computation (2nd version) ******/

/*
 *  \brief compute the state-time tree corresponding to each possible trajectories
 */
void DD2AvoidTrajGen::computeStateTimeTree()
{
  

  cerr << "DD2AvoidTrajGen::computeStateTimeTree : Not implemented yet!" << endl;
  exit(1);
}


/***************** Collision check ****************/

/*
 *  \brief return the time needed by the robot to stop
 *  \return the time needed by the robot to stop
 */
double DD2AvoidTrajGen::timeToStop() const
{
  // get back the current velocity
  double curVel = loc()->v();

  // get back the maximal deceleration
  double maxDecel = 3.0 * sys()->getAMax();

  return fabs(curVel) / maxDecel;
}

/*
 *  \brief compute the minimal distance traveled by the robot from its current to stop
 *  \return the minimal distance traveled by the robot from its current to stop
 */
double DD2AvoidTrajGen::distToStop() const
{
  double vinit = loc()->v();
  double vgoal = 0;
  double amax = 3.0 * sys()->getAMax();
  double tres;
  double dres;
  
  minTDtoReachSpd(vinit, vgoal, amax, tres, dres);

  return dres;
}


/****************** I/O functions *****************/

/**
 *  \brief compute the obstacle avoidance trajectory
 *  \param state : the current state (position and speed) of the robot
 *  \return 1 if the goal has been reached
 *  \return 0 if the process succeeded,
 *  \return -1 if the avoid trajgen computation failed
 *  \return -2 if an I/O problem occured
 *  \return -3 if the avoid trajgen settings is not valid
 *  \return -4 if a possible collision or ICS may occur
 */
int DD2AvoidTrajGen::computeAvoidTraj(const DD2State& s)
{
  // update the robot localization
  this->updateState(s);

  // clear the vector of waypoints
  wp_statet_.clear();

  // compute the obstacle avoidance trajectory
  int res;
  if(AvoidTrajGen::version == 1) 
    res = AvoidTrajGen::computeAvoidTraj();
  else // if(AvoidTrajGen::version == 2) 
    res = AvoidTrajGen::computeAvoidTrajV2();

  return res;
}

// /**
//  *  \brief get back the resulting state trajectory
//  *  \return the resulting state trajectory
//  */
// DD2PSTraj* DD2AvoidTrajGen::getSTraj()
// {
//   if(pstraj_ != NULL)
//     {
//       boost::mutex::scoped_lock lock(mutexTraj_);
//       DD2PSTraj* copyTraj = new DD2PSTraj(*pstraj());
//       return copyTraj;
//     }
//   else return NULL;
// }

/**
 *  \brief get back the resulting control trajectory
 *  \return the resulting control trajectory
 */
DD2PCTraj* DD2AvoidTrajGen::getPCTraj()
{
  if(pctraj_ != NULL)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      DD2PCTraj* copyTraj = new DD2PCTraj(*pctraj());
      return copyTraj;
    }
  else return NULL;
}

/**
 *  \brief get back the resulting state trajectory
 *  \return the resulting state trajectory
 */
DD2STraj* DD2AvoidTrajGen::getSTraj()
{
  if(straj_ != NULL)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      DD2STraj* copyTraj = new DD2STraj(*straj());
      return copyTraj;
    }
  else return NULL;
}

/**
 *  \brief get back the resulting shift trajectory
 *  \return a new instance of the resulting shift trajectory
 */
ShiftRecord* DD2AvoidTrajGen::getShiftRecord()
{
}


#if DEBUG_ATG
/***************** Debug functions ****************/

/*
 *  \brief draw the state trajectory found on the image
 *  \param img : a pointer to the image to modify
 */
void DD2AvoidTrajGen::drawPSTrajOnImage(RGBImage* img)
{
//   if(pstraj() == NULL) return;

//   double dt_spg = 0.5;
//   double maxTime = pstraj()->endTime();
//   int nbP = ((int)(maxTime / dt_spg)) + 1;
//   double real_dt_spg = maxTime / nbP;

//   for(int i=0; i<nbP; i++)
//     {
//       double curT = i*real_dt_spg;
//       double nextT = (i+1)*real_dt_spg;
//       DD2StateT curS = pstraj()->at(curT);
//       DD2StateT nextS = pstraj()->at(nextT);

//       int iImage, jImage, iImage2, jImage2;
//       int fail = dem_->getIndexPPMImagePos(curS.x(), curS.y(), iImage, jImage);
//       int fail2 = dem_->getIndexPPMImagePos(nextS.x(), nextS.y(), iImage2, jImage2);
//       img->drawLine(iImage, jImage, iImage2, jImage2, RGBColor(255,255,0));
//     }

  if(straj() == NULL) return;

  int N = straj()->size();
  for(int i=0; i<N-1; i++)
    {
      DD2StateT curS = straj()->at(i);
      DD2StateT nextS = straj()->at(i+1);

      int iImage, jImage, iImage2, jImage2;
      int fail = dem_->getIndexPPMImagePos(curS.x(), curS.y(), iImage, jImage);
      int fail2 = dem_->getIndexPPMImagePos(nextS.x(), nextS.y(), iImage2, jImage2);
      img->drawLine(iImage, jImage, iImage2, jImage2, RGBColor(255,255,0));
    }
}
#endif
