#include "avoidObs/inc/GraphPassageway.hpp"
using namespace mpc;


                                    /**************************************************/
                                    /************** Class GraphPassageway *************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
GraphPassageway::GraphPassageway() : DGraph<Passageway>()
{
  this->tildY_origin_ = 0;
  this->fixed_width_ = 0;
}

/**
 *  \brief default / init constructor
 *  \param tildY_origin : lateral shift origin of the graph
 *  \param tildY_origin : fixed width of the graph (in the case of fixed width mode)
 */
GraphPassageway::GraphPassageway(float tildY_origin, float fixed_width) : DGraph<Passageway>()
{
  this->tildY_origin_ = tildY_origin;
  this->fixed_width_ = fixed_width;
}

/**
 *  \brief default constructor
 */
GraphPassageway::GraphPassageway(const GraphPassageway& g) : DGraph<Passageway>(g)
{
  this->tildY_origin_ = g.tildY_origin_;
  this->fixed_width_ = g.fixed_width_;
}

/**
 *  \brief destructor
 */
GraphPassageway::~GraphPassageway()
{
}


/******************** Searches ********************/

/**
 *  \brief search for a passageway in the graph
 *  \param pwy : the searched passageway
 *  \param it_res : the resulting vertex iterator
 *  \return true if the passageway has been found in the graph, false otherwise
 */
bool GraphPassageway::searchForPassageway(const Passageway& pwy, DGVertex_iterator<Passageway>& it_res)
{
  bool found = false;
  std::list< DGVertex<Passageway>* >::iterator it = this->vertices_.begin();

  while(it != this->vertices_.end() && !found) 
    {
      // get back the current passageway
      Passageway curPwy = (*it)->val();

      if(pwy.s == curPwy.s &&
	 pwy.s == curPwy.s &&
	 pwy.s == curPwy.s)
	{
	  it_res = DGVertex_iterator<Passageway>(it);
	  found = true;
	}
      else it++;
    }
  return found;
}


/**************** Utility functions ***************/

/*
 *  \brief check if the input edge is too short
 *  \param edge : the input edge
 *  \param minDiff : minimal curve abscica difference between both endpoints of the edge
 *  \return true if the input edge is too short, false otherwise
 */
bool GraphPassageway::isShortEdge(const DGEdge<Passageway>& edge, float minDiff) const
{
  if((edge.endVertex())->val().s - (edge.startVertex())->val().s < minDiff) return true;
  else return false;
}

/*
 *  \brief check if the input edge is in collision with obstacles on the input distmap
 *  \param edge : the input edge
 *  \param distmap : the distance to obstacles map
 *  \return true if the input edge is in collision with obstacles, false otherwise
 */
bool GraphPassageway::isColEdge(const DGEdge<Passageway>& edge, const DistObsMap& distmap) const
{
  PositionXY pos_start_mdp = (edge.startVertex())->val().maxDistPoint;
  PositionXY pos_end_mdp = (edge.endVertex())->val().maxDistPoint;
  return (!(distmap.isFreeSegmentsOnLine(pos_start_mdp.x(), pos_start_mdp.y(), pos_end_mdp.x(), pos_end_mdp.y())));
}

/*
 *  \brief check if the edge defined from v1 to v2 is in collision with obstacles on the input distmap
 *  \param edge : the input edge
 *  \param distmap : the distance to obstacles map
 *  \return true if the edge defined from v1 to v2 is in collision with obstacles, false otherwise
 */
bool GraphPassageway::isColEdge(const DGVertex<Passageway>& v1, const DGVertex<Passageway>& v2, const DistObsMap& distmap) const
{
  PositionXY pos_start_mdp = v1.val().maxDistPoint;
  PositionXY pos_end_mdp = v2.val().maxDistPoint;
  return (!(distmap.isFreeSegmentsOnLine(pos_start_mdp.x(), pos_start_mdp.y(), pos_end_mdp.x(), pos_end_mdp.y())));
}

/*
 *  \brief select the best way to replace the current edge
 *  \param it_edge : iterator to the edge of the graph to replace
 *  \return GRAPH_INPUT if the best option is to link its ending vertex to the previous vertices of the graph
 *  \return GRAPH_OUTPUT if the best option is to link its starting vertex to the next vertices of the graph
 */
graph_io_t GraphPassageway::selectBestEdgeRep(const DGEdge_iterator<Passageway>& it_edge)
{
  DGVertex_iterator<Passageway> vertexBeg_it = this->startVertex(it_edge);
  float YtildBeg = (*vertexBeg_it)->val().latShift(this->tildY_origin_);
  DGVertex_iterator<Passageway> vertexEnd_it = this->endVertex(it_edge);
  float YtildEnd = (*vertexEnd_it)->val().latShift(this->tildY_origin_);

  float minInputDY = -1;
  int N1 = this->nbPredecessors(vertexBeg_it);
  for(int i=0; i<N1; i++)
    {
      // get back the current predecessor of the starting vertex
      DGVertex_iterator<Passageway> pred_it = this->predecessor(vertexBeg_it, i);

      // compute the lateral shift between the current vertex and vertexBeg
      float DY = fabs((*pred_it)->val().latShift(this->tildY_origin_) - YtildBeg);

      // check if it is the minimal lateral shift found
      if(i==0 || DY < minInputDY) minInputDY = DY;
    }

  float minOutputDY = -1;
  int N2 = this->nbSuccessors(vertexEnd_it);
  for(int i=0; i<N2; i++)
    {
      // get back the current successor of the ending vertex
      DGVertex_iterator<Passageway> suc_it = this->successor(vertexEnd_it, i);

      // compute the lateral shift between the current vertex and vertexEnd
      float DY = fabs((*suc_it)->val().latShift(this->tildY_origin_) - YtildEnd);

      // check if it is the minimal lateral shift found
      if(i==0 || DY < minOutputDY) minOutputDY = DY;
    }

  if(N1 == 0 && N2 == 0) return GRAPH_IO_ND;
  else if(N1 == 0) return GRAPH_OUTPUT;
  else if(N2 == 0) return GRAPH_INPUT;
  else
    {
      if(minInputDY <= minOutputDY) return GRAPH_INPUT;
      else return GRAPH_OUTPUT;
    }
}

/*
 *  \brief reconnect an edge from the graph and try to reconnect it to the previous nodes of the graph
 *  \param vertexBeg_it : iterator to the starting vertex of the edge to replace
 *  \param vertexEnd_it : iterator to the ending vertex of the edge to replace
 *  \param distmap : distance to obstacles map
 */
// void GraphPassageway::reconnectEdgeToPrevNodes(const DGEdge_iterator<Passageway>& it_edge, const DistObsMap& distmap)
void GraphPassageway::reconnectEdgeToPrevNodes(DGVertex_iterator<Passageway>& vertexBeg_it, DGVertex_iterator<Passageway>& vertexEnd_it, const DistObsMap& distmap)
{
//   DGVertex_iterator<Passageway> vertexBeg_it = this->startVertex(it_edge);
//   DGVertex_iterator<Passageway> vertexEnd_it = this->endVertex(it_edge);

  int N = this->nbPredecessors(vertexBeg_it);
  for(int i=0; i<N; i++)
    {
      // get back the current predecessor of the starting vertex
      DGVertex_iterator<Passageway> pred_it = this->predecessor(vertexBeg_it, i);

      // check if an edge from pred_it to vertexEnd_it will not be in collision
      bool collision = this->isColEdge(**pred_it, **vertexEnd_it, distmap);

      // add a new edge from pred_it to vertexEnd_it
      if(!collision) this->addEdge(pred_it, vertexEnd_it);
    }

  // check if vertexBeg is still connected to the next part of the graph
  if(this->nbSuccessors(vertexBeg_it) == 0) this->removeVertex(vertexBeg_it);
}

/*
 *  \brief reconnect an edge from the graph and try to reconnect it to the next nodes of the graph
 *  \param vertexBeg_it : iterator to the starting vertex of the edge to replace
 *  \param vertexEnd_it : iterator to the ending vertex of the edge to replace
 *  \param distmap : distance to obstacles map
 */
// void GraphPassageway::reconnectEdgeToNextNodes(const DGEdge_iterator<Passageway>& it_edge, const DistObsMap& distmap)
void GraphPassageway::reconnectEdgeToNextNodes(DGVertex_iterator<Passageway>& vertexBeg_it, DGVertex_iterator<Passageway>& vertexEnd_it, const DistObsMap& distmap)
{
//   DGVertex_iterator<Passageway> vertexBeg_it = this->startVertex(it_edge);
//   DGVertex_iterator<Passageway> vertexEnd_it = this->endVertex(it_edge);

  int N = this->nbSuccessors(vertexEnd_it);
  for(int i=0; i<N; i++)
    {
      // get back the current successor of the ending vertex
      DGVertex_iterator<Passageway> suc_it = this->successor(vertexEnd_it, i);

      // check if an edge from vertexBeg_it to suc_it will not be in collision
      bool collision = this->isColEdge(**vertexBeg_it, **suc_it, distmap);

      // add a new edge from vertexBeg_it to suc_it
      if(!collision) this->addEdge(vertexBeg_it, suc_it);
    }

  // check if vertexEnd is still connected to the previous part of the graph
  if(this->nbPredecessors(vertexEnd_it) == 0) this->removeVertex(vertexEnd_it);
}

/*
 *  \brief remove an edge from the graph and try to reconnect the graph
 *  \param it_edge : iterator to the edge to replace
 *  \param side : the side of the graph to which the nodes of the edge must be reconnected
 *  \param distmap : distance to obstacles map
 *  \return an iterator to the next edge in the list of edges
 */
DGEdge_iterator<Passageway> GraphPassageway::replaceEdge(const DGEdge_iterator<Passageway>& it_edge, graph_io_t side, const DistObsMap& distmap)
{
  if(side == GRAPH_IO_ND)
    {
      cerr << "GraphPassageway::replaceEdge : The current edge is not connected to the graph. Unable to replace it." << endl; exit(1);
      return it_edge;
    }

  DGVertex_iterator<Passageway> vertexBeg_it = this->startVertex(it_edge);
  DGVertex_iterator<Passageway> vertexEnd_it = this->endVertex(it_edge);

  // remove the current edge from the graph/
//   DGEdge_iterator<Passageway> nextIt = it_edge; nextIt++;
  DGEdge_iterator<Passageway> nextIt = this->removeEdge(it_edge);

  // check the side of the graph to which the current edge must be reconnected
  if(side == GRAPH_INPUT) this->reconnectEdgeToPrevNodes(vertexBeg_it, vertexEnd_it, distmap);
  else this->reconnectEdgeToNextNodes(vertexBeg_it, vertexEnd_it, distmap);

  return nextIt;
}

/**
 *  \brief remove edges that are either too short or in collision with obstacles
 *  \param minDiff : the minimal distance between two successives vertices of the graph
 *  \param distmap : distance to obstacles map
 */
void GraphPassageway::removeShortAndColEdges(float minDiff, const DistObsMap& distmap)
{
  std::list< DGEdge<Passageway>* >::iterator it = this->edges_.begin();

//   for(std::list< DGEdge<Passageway>* >::iterator it = this->edges_.begin(); it != this->edges_.end(); it++) 
  while(it != this->edges_.end()) 
    {
      if(this->isShortEdge(**it, minDiff) || this->isColEdge(**it, distmap))
	{
	  // create a DGEdge_iterator from the current iterator
	  DGEdge_iterator<Passageway> edge_it(it);

	  // select the best way to replace the current edge
	  graph_io_t side = this->selectBestEdgeRep(edge_it);

	  // remove the current edge from the graph and try to reconnect the graph afterwards
	  it = this->replaceEdge(edge_it, side, distmap);
	}
      else it++;
    }
}

/**
 *  \brief convert the passageway graph into a passageway tree
 *  \param sys : the robotic system constraints
 *  \return a pointer to the new passageway tree
 */
PassagewayNode* GraphPassageway::toPassawagewayTree(const System& sys)
{
  // check if the current passageway graph is not empty
  if(this->nbVertices() == 0) 
    {
      cerr << "GraphPassageway::toPassawagewayTree : The graph is empty. Unable to convert it into a tree" << endl;
      return NULL;
    }

  // convert the passageway graph into a simple passageway tree
  Node<Passageway>* simplePwyTree = this->toTree();
}


/**************** Display functions ***************/

/**
 *  \brief display the graphe
 */
void GraphPassageway::display()
{
      std::cout << "Graphe vertices (" << this->vertices_.size() << ") : " << std::endl;
      for(std::list< DGVertex<Passageway>* >::iterator it = this->vertices_.begin(); it != this->vertices_.end(); it++) 
	{
	  (*it)->val().maxDistPoint.display(); 
	  cout << endl;
	}

      std::cout << "Graphe edges (" << this->edges_.size() << ") " << std::endl;      
      for(std::list< DGEdge<Passageway>* >::iterator it2 = this->edges_.begin(); it2 != this->edges_.end(); it2++) 
	{
	  (*it2)->startVertex()->val().maxDistPoint.display(); 
	  cout << " -> "; 
	  (*it2)->endVertex()->val().maxDistPoint.display(); 
	  cout << endl;
	}
}

// /**
//  *  \brief draw the current graph on an image
//  *  \param image : the image on which the graph must be drawn
//  *  \param col : the color of the path
//  */
// void GraphPassageway::drawOnImage(const RGBImage& image, const RGBColor& col)
// {
//   // draw each vertex of the graph
//   for(typename std::list< DGVertex<Passageway>* >::iterator it = vertices_.begin(); it != vertices_.end(); it++)
//     {
//       // get back the maxDistPoint of the current vertex
//       PositionXY pos = (*it)->val().maxDistPoint;

//       // get back the coordinates of pos on the image
//       int iImage, jImage;
//       int fail = dem_->getIndexPPMImagePos(pos.x(), pos.y(), iImage, jImage);

//       // draw a cross on the image at position pos
//       image.drawCross(iImage, jImage, 3, col);
//     }

//   // draw each edge of the graph
//   for(typename std::list< DGEdge<Passageway>* >::iterator it = edges_.begin(); it != edges_.end(); it++)
//     {
//       // get back both maxDistPoint of the end points of the current edge
//       PositionXY pos1 = (*it)->startVertex()->val().maxDistPoint;
//       PositionXY pos2 = (*it)->endVertex()->val().maxDistPoint;

//       // get back the coordinates of pos1 and pos2 on the image
//       int iImage1, jImage1, iImage2, jImage2;
//       int fail1 = dem_->getIndexPPMImagePos(pos1.x(), pos1.y(), iImage1, jImage1);
//       int fail2 = dem_->getIndexPPMImagePos(pos2.x(), pos2.y(), iImage2, jImage2);
      
//       image.drawLine(iImage1, jImage1, iImage2, jImage2, col);
//     }
// }


                                    /**************************************************/
                                    /*************** Class VecPassageway **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
VecPassageway::VecPassageway()
{
}

/**
 *  \brief copy constructor
 */
VecPassageway::VecPassageway(const VecPassageway& v)
{
  int N = v.size();
  for(int i=0; i<N; i++)
    {
      this->push_back(new Passageway(*(v.at(i))));
    }
}

/**
 *  \brief destructor
 */
VecPassageway::~VecPassageway()
{
  this->clearVec();
}


/******************* Accessors ********************/

/**
 *  \brief get back the first element of the vector
 *  \return a pointer to the first element of the vector or NULL if the list is empty
 */
const Passageway* VecPassageway::first() const
{
  if(this->empty()) return NULL;
  else return this->at(0);
}

/**
 *  \brief get back the last element of the vector
 *  \return a pointer to the last element of the vector or NULL if the list is empty
 */
const Passageway* VecPassageway::last() const
{
  if(this->empty()) return NULL;
  else return this->at(this->size()-1);
}

 
/************** Vector modifications **************/

/**
 *  \brief append an element at the end of the vector
 *  \param p : the new passageway to append
 */
void VecPassageway::append(const Passageway& p)
{
  this->push_back(new Passageway(p));
}

/**
 *  \brief prepend an element at the beginning of the vector
 *  \param p : the new passageway to prepend
 */
void VecPassageway::prepend(const Passageway& p)
{
  if(this->size() == 0) this->push_back(new Passageway(p));
  else this->insert(this->begin(), new Passageway(p));
}

/**
 *  \brief concatenate a vector at the end of the current one
 *  \param v : the new vector to concat
 */
void VecPassageway::concat(const VecPassageway* v)
{
  if(v == NULL) return;

  int N=v->size();
  for(int i=0; i<N; i++) this->append(*(v->at(i)));
}

/**
 *  \brief clear all the values of the current vector
 */
void VecPassageway::clearVec()
{
  int N = this->size();
  for(int i=0; i<N; i++)
    {
      if(this->at(i) != NULL) 
	{ 
	  delete this->at(i);
	  this->at(i) = NULL;
	}
    }
  this->clear();
}


/******************** Searches ********************/

/**
 *  \brief search for a strangling point in the current vector between indices i_min and i_max
 *  \param i_min : index of the starting point of the search
 *  \param i_max : index of the ending point of the search
 *  \param i_strang : index of the resulting strangling point
 *  \return a pointer to the strangling point between indices i_min and i_max
 */
const Passageway* VecPassageway::searchForStrangPoint(int i_min, int i_max, int& i_strang) const
{
  float distMin=0;
  const Passageway* pwy = NULL;
  i_strang = -1;

  for(int i=i_min; i<=i_max; i++)
    {
      const Passageway* curPwy = this->at(i);
      if(pwy == NULL || curPwy->dist < distMin)
	{
	  distMin = curPwy->dist;
	  pwy = curPwy;
	  i_strang = i;
	}
    }

  return pwy;
}

/**
 *  \brief search for the vector of strangling points of the current vector (recusrive method)
 *  \param minDiffInd : the minimum number of indices between two successives strangling points
 *  \param i_min : index of the starting point of the search
 *  \param i_max : index of the ending point of the search
 *  \return a vector of the strangling points of the current vector
 *
 *  A new instance of VecPassageway is returned
 */
VecPassageway* VecPassageway::searchForStrangPointsRec(int minDiffInd, int i_min, int i_max) const
{
  VecPassageway* vecRes = new VecPassageway();

  // search for the current strangling point on the current interval
  int i_strang;
  const Passageway* strangPoint = this->searchForStrangPoint(i_min, i_max, i_strang);

  // search for new strangulation points before the current found if needed
  if(i_strang - i_min > minDiffInd)
    {
      VecPassageway* vecBeg = this->searchForStrangPointsRec(minDiffInd, i_min, i_strang-1);
      vecRes->concat(vecBeg);
      delete vecBeg;
    }

  // append the current strangulation point found
  vecRes->append(*strangPoint);
  delete strangPoint;

  // search for new strangulation points after the current found if needed
  if(i_max - i_strang > minDiffInd)
    {
      VecPassageway* vecEnd = this->searchForStrangPointsRec(minDiffInd, i_strang+1, i_max);
      vecRes->concat(vecEnd);
      delete vecEnd;
    }

  return vecRes;
}

/**
 *  \brief search for the vector of strangling points of the current vector (recusrive method)
 *  \param minDiff : the minimal distance between two successives strangling points
 *  \param i_min : index of the starting point of the search
 *  \param i_max : index of the ending point of the search
 *  \return a vector of the strangling points of the current vector
 *
 *  A new instance of VecPassageway is returned
 */
VecPassageway* VecPassageway::searchForStrangPointsRec(float minDiff, int i_min, int i_max) const
{
  VecPassageway* vecRes = new VecPassageway();

  // search for the current strangling point on the current interval
  int i_strang;
  const Passageway* strangPoint = this->searchForStrangPoint(i_min, i_max, i_strang);

  // search for new strangulation points before the current found if needed
  if(this->at(i_strang)->s - this->at(i_min)->s > minDiff)
    {
      int i_newmax = i_strang-1;
      while(this->at(i_strang)->s - this->at(i_newmax)->s < minDiff && i_newmax != i_min) i_newmax--;

//       cout << "i_min : " << i_min << "  i_newmax : " << i_newmax << "  i_strang : " << i_strang << "  this->at(i_min)->s : " << this->at(i_min)->s << "  this->at(i_strang)->s : " << this->at(i_strang)->s << "  diff : " << this->at(i_strang)->s - this->at(i_min)->s << "  minDiff : " << minDiff << "  newDiff : " << this->at(i_strang)->s - this->at(i_newmax)->s  << endl;

//       VecPassageway* vecBeg = this->searchForStrangPointsRec(minDiff, i_min, i_strang-1);
      VecPassageway* vecBeg = this->searchForStrangPointsRec(minDiff, i_min, i_newmax);
      vecRes->concat(vecBeg);
      delete vecBeg;
    }

  // append the current strangulation point found
  vecRes->append(*strangPoint);
//   delete strangPoint;

  // search for new strangulation points after the current found if needed
  if(this->at(i_max)->s - this->at(i_strang)->s > minDiff)
    {
      int i_newmin = i_strang+1;
      while(this->at(i_newmin)->s - this->at(i_strang)->s < minDiff && i_newmin != i_max) i_newmin++;

//       cout << "i_strang : " << i_strang << "  i_newmin : " << i_newmin << "  i_max : " << i_max << "  this->at(i_strang)->s : " << this->at(i_strang)->s << "  this->at(i_max)->s : " << this->at(i_max)->s << "  diff : " << this->at(i_max)->s - this->at(i_strang)->s << "  minDiff : " << minDiff << "  newDiff : " << this->at(i_max)->s - this->at(i_newmin)->s << endl;

//       VecPassageway* vecEnd = this->searchForStrangPointsRec(minDiff, i_strang+1, i_max);
      VecPassageway* vecEnd = this->searchForStrangPointsRec(minDiff, i_newmin, i_max);
      vecRes->concat(vecEnd);
      delete vecEnd;
    }

  return vecRes;
}

/**
 *  \brief search for the vector of strangling points of the current vector
 *  \param minDiffInd : the minimum number of indices between two successives strangling points
 *  \return a vector of the strangling points of the current vector
 *
 *  A new instance of VecPassageway is returned
 */
VecPassageway* VecPassageway::searchForStrangPoints(int minDiffInd) const
{
  // check if the vector is not empty
  if(this->empty())
    {
      cerr << "VecPassageway::searchForStrangPoint : the vector is empty!" << endl;
      return NULL;
    }

  return this->searchForStrangPointsRec(minDiffInd, 0, this->size()-1);
}

/**
 *  \brief search for the vector of strangling points of the current vector
 *  \param minDiff : the minimal distance between two successives strangling points
 *  \return a vector of the strangling points of the current vector
 *
 *  A new instance of VecPassageway is returned
 */
VecPassageway* VecPassageway::searchForStrangPoints(float minDiff) const
{
  // check if the vector is not empty
  if(this->empty())
    {
      cerr << "VecPassageway::searchForStrangPoint : the vector is empty!" << endl;
      return NULL;
    }

  return this->searchForStrangPointsRec(minDiff, 0, this->size()-1);
}

/**
 *  \brief search for the vector of strangling points of the current vector and includes its first and last values
 *  \param minDiff : the minimal distance between two successives strangling points
 *  \return a vector of the strangling points of the current vector
 *
 *  A new instance of VecPassageway is returned
 */
VecPassageway* VecPassageway::searchForStrangPointsAndBounds(float minDiff) const
{
  // search for the strangling points of the current vector
  VecPassageway* vecRes = this->searchForStrangPoints(minDiff);

  // includes the first value of the current vector if it doesn't 
  // correspond to the first value of the vector of strangling points
  const Passageway* vecRes_first = vecRes->first();
  const Passageway* this_first = this->first();

  if(vecRes_first->s != this_first->s ||
     vecRes_first->minWidthB != this_first->minWidthB ||
     vecRes_first->maxWidthB != this_first->maxWidthB)
    vecRes->prepend(*this_first);

  // includes the last value of the current vector if it doesn't 
  // correspond to the last value of the vector of strangling points
  const Passageway* vecRes_last = vecRes->last();
  const Passageway* this_last = this->last();

  if(vecRes_last->s != this_last->s ||
     vecRes_last->minWidthB != this_last->minWidthB ||
     vecRes_last->maxWidthB != this_last->maxWidthB)
    vecRes->append(*this_last);

  return vecRes;
}


                                    /**************************************************/
                                    /************ Class GraphTopoPassageway ***********/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
GraphTopoPassageway::GraphTopoPassageway() : DGraph<VecPassageway*>()
{
  this->tildY_origin_ = 0;
  this->fixed_width_ = 0;
}

/**
 *  \brief copy constructor
 */
GraphTopoPassageway::GraphTopoPassageway(const GraphTopoPassageway& g) : DGraph<VecPassageway*>(g)
{
  this->tildY_origin_ = g.tildY_origin_;
  this->fixed_width_ = g.fixed_width_;
}

/**
 *  \brief init constructor (from a passageway graph)
 *  \param gp : input passageway graph
 */
GraphTopoPassageway::GraphTopoPassageway(GraphPassageway& gp) : DGraph<VecPassageway*>()
{
  this->tildY_origin_ = gp.tildY_origin();
  this->fixed_width_ = gp.fixed_width();

  // get back the "first" vertex of the graph
  DGVertex_iterator<Passageway> gp_it = gp.firstVertex();

  // check if the graph is not empty
  if(*gp_it == NULL) 
    {
      cerr << "GraphTopoPassageway::GraphTopoPassageway : Warning. The input passageway graph is empty" << endl;
      return;
    }

  // create the first vertex of the topo graph
  this->buildTopoVertex(gp, gp_it);
}

/**
 *  \brief destructor
 */ 
GraphTopoPassageway::~GraphTopoPassageway()
{
  this->clear();
}


/*************** Graph modifications **************/

/*
 *  \brief recursive function used to build the topo graph from a passageway graph
 *  \param gp : reference to the passageway graph
 *  \param gp_it : iterator to a vertex of the passageway graph
 *  \return an iterator to the new node created in the topo graph
 */
DGVertex_iterator<VecPassageway*> GraphTopoPassageway::buildTopoVertex(const GraphPassageway& gp, DGVertex_iterator<Passageway>& gp_it)
{
  // copy the input graph passageway iterator
  DGVertex_iterator<Passageway> cur_gp_it = gp_it;

  // create the new passageway vector corresponding to the current branch
  VecPassageway* vec_p = new VecPassageway();

  // create the vector of the first nodes of the next branches
  vector< DGVertex_iterator<Passageway> > vec_next_branches;

  bool end = false;
  while(!end)
    {
      if(vec_p->size() != 0 && (*cur_gp_it)->nbInputEdges() > 1)
	{
	  // stops the current branch and create a new one with the current node
	  vec_next_branches.push_back(cur_gp_it);
	  end = true;
	}
      else
	{
	  // add the current vertex to the current branch
	  vec_p->append((*cur_gp_it)->val());

	  // check the number of output edges of the current vertex
	  if((*cur_gp_it)->nbOutputEdges() == 0)
	    {
	      // "leaf" of the graph (no outpout edges from the current vertex)
	      end = true;
	    }
	  else if((*cur_gp_it)->nbOutputEdges() == 1)
	    {
	      // moving on the current branch
	      cur_gp_it = gp.successor(cur_gp_it, 0);
	    }
	  else // if((*cur_gp_it)->nbOutputEdges() >= 2)
	    {
	      // creating a new branch from the end points of every output edges
	      int N = (*cur_gp_it)->nbOutputEdges();
	      for(int i=0; i<N; i++)
		{
		  vec_next_branches.push_back(gp.successor(cur_gp_it, i));
		}
	      end = true;
	    }
	}
    }

  // create a vertex of the topological graph from the built passageway vector
  DGVertex_iterator<VecPassageway*> cur_gtp_vertex = this->addVertex(vec_p);

  // create each new branch if needed
  int N = vec_next_branches.size();
  for(int i=0; i<N; i++)
    {
      // get back the current iterator to the first passageway of the next branch
      DGVertex_iterator<Passageway> next_branch = vec_next_branches.at(i);

      // check if this branch has already been built
      DGVertex_iterator<VecPassageway*> next_gtp_vertex;
      bool includePwy = this->includePassageway((*next_branch)->val(), next_gtp_vertex);

      if(!includePwy) next_gtp_vertex = buildTopoVertex(gp, next_branch);

//       if(includePwy) cout << "GraphTopoPassageway::buildTopoVertex ... passageway found in the graph" << endl;
//       else cout << "GraphTopoPassageway::buildTopoVertex ... passageway not found" << endl;

      // add an edge between the current and the next topo vertices
      this->addEdge(cur_gtp_vertex, next_gtp_vertex);
    }

  return cur_gtp_vertex;
}

/**
 *  \brief clear the graph content
 */
void GraphTopoPassageway::clear()
{
  for(std::list< DGVertex<VecPassageway*>* >::iterator it = vertices_.begin(); it != vertices_.end(); it++) 
    {
      // delete the passageway vector 
      delete (*it)->val();
      // delete the vertex
      delete (*it);
    }
  vertices_.clear();
  
  for(std::list< DGEdge<VecPassageway*>* >::iterator it = edges_.begin(); it != edges_.end(); it++) delete (*it);
  edges_.clear();
}

/**
 *  \brief simplfy the graph into a graph of strangling points
 *  \param minDiff : the minimal distance between two successives strangling points
 */
void GraphTopoPassageway::simplifyIntoSPGraph(float minDS_StrPoints)
{
  int i = 0;
  int N = vertices_.size();

  for(std::list< DGVertex<VecPassageway*>* >::iterator it = vertices_.begin(); it != vertices_.end(); it++) 
    {
      // get back the vector of strangling points from the current passageway vector
      VecPassageway* strPoints = NULL;
      if(i == 0 || i == N-1) strPoints = (*it)->val()->searchForStrangPointsAndBounds(minDS_StrPoints);
      else strPoints = (*it)->val()->searchForStrangPoints(minDS_StrPoints);

      // clear the current passageway vector
      delete (*it)->val();
//       (*it)->val()->clearVec();

      // replace it by the corresponding vector of strangling points
      (*it)->val() = strPoints;

      i++;
    }
}


/**************** Utility functions ***************/

/*
 *  \brief search for a node in the topo graph containing the input passageway
 *  \param pwy : the input passageway
 *  \param gtp_vertex : the output node of the topo graph corresponding to the passageway (if it has been found)
 *  \return true if the input passageway has been found in the topo graph
 */
bool GraphTopoPassageway::includePassageway(const Passageway& pwy, DGVertex_iterator<VecPassageway*>& gtp_vertex)
{
  bool found = false;
  std::list< DGVertex<VecPassageway*>* >::iterator it = this->vertices_.begin();

  while(!found && it != this->vertices_.end()) 
    {
      // get back the first passageway of the current passageway vector
      const Passageway* curPwy = (*it)->val()->first();

      // check if the current passageway corresponds to the input one
      if(curPwy == NULL) it++;
      else if(curPwy->s == pwy.s && 
	      curPwy->minWidthB == pwy.minWidthB && 
	      curPwy->maxWidthB == pwy.maxWidthB) 
	{
	  found = true;
	  gtp_vertex = DGVertex_iterator<VecPassageway*>(it);
	}
      else it++;
    }

  return found;
}

/**
 *  \brief convert the topo graph into a simple passageway graph
 *  \return a pointer to the new passageway graph
 *
 *  A new instance of GraphPassageway is returned
 */
GraphPassageway* GraphTopoPassageway::toGraphPassageway()
{
  GraphPassageway* graphRes = new GraphPassageway(this->tildY_origin_, this->fixed_width_);

  for(std::list< DGVertex<VecPassageway*>* >::iterator it = this->vertices_.begin(); it != this->vertices_.end(); it++) 
    {
      // get back the current passageway vector
      VecPassageway* vec = (*it)->val();
      int N = vec->size();
      
      DGVertex_iterator<Passageway> prev_vertex_it;
      for(int i=0; i<N; i++) 
	{
	  // add the current vertex to the new passageway graph
	  DGVertex_iterator<Passageway> vertex_it = graphRes->addVertex(*(vec->at(i)));

	  // add an edge from the previous vertex
	  if(i != 0) graphRes->addEdge(prev_vertex_it, vertex_it);

	  prev_vertex_it = vertex_it;
	}
    }
  
  for(std::list< DGEdge<VecPassageway*>* >::iterator it2 = this->edges_.begin(); it2 != this->edges_.end(); it2++) 
    {
      const Passageway* startPwy = (*it2)->startVertex()->val()->last();
      const Passageway* endPwy = (*it2)->endVertex()->val()->first(); 

      DGVertex_iterator<Passageway> it_startPwy;
      DGVertex_iterator<Passageway> it_endPwy;
      bool foundStartPwy = graphRes->searchForPassageway(*startPwy, it_startPwy);
      bool foundEndPwy = graphRes->searchForPassageway(*endPwy, it_endPwy);

      if(foundStartPwy && foundEndPwy) graphRes->addEdge(it_startPwy, it_endPwy);
      else cerr << "GraphTopoPassageway::toGraphPassageway : Unable to find both passageway iterators" << endl;
    }

  return graphRes;
}


/**************** Display functions ***************/

/**
 *  \brief display the graphe
 */
void GraphTopoPassageway::display()
{
  std::cout << "Graphe vertices (" << this->vertices_.size() << ") : " << std::endl;

  int k=0;
  for(std::list< DGVertex<VecPassageway*>* >::iterator it = this->vertices_.begin(); it != this->vertices_.end(); it++) 
    {
      cout << "Vertex " << k << " : " << endl;
      
      // get back the current passageway vector
      VecPassageway* vec = (*it)->val();
      int N = vec->size();
      
      for(int i=0; i<N; i++) 
	{
	  cout << i << " : "; 
	  vec->at(i)->display();
	}
      
      k++;
    }
  
  std::cout << "Graphe edges (" << this->edges_.size() << ") " << std::endl;      
  for(std::list< DGEdge<VecPassageway*>* >::iterator it2 = this->edges_.begin(); it2 != this->edges_.end(); it2++) 
    {
      (*it2)->startVertex()->val()->first()->display(); 
      cout << " -> "; 
      (*it2)->endVertex()->val()->first()->display(); 
      cout << endl;
    }
}
