#include "avoidObs/inc/Passageway.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /**************** Class Passageway ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Passageway::Passageway()
{
  s = 0;
  minWidthB = 0;
  maxWidthB = 0;
  maxDistPoint = PositionXY(0,0);
  dist = 0;
  nearObs = PositionXY(0,0);
}

/**
 *  \brief copy constructor
 */
Passageway::Passageway(const Passageway& p)
{
  this->s = p.s;
  this->minWidthB = p.minWidthB;
  this->maxWidthB = p.maxWidthB;
  this->maxDistPoint = PositionXY(p.maxDistPoint);
  this->dist = p.dist;
  this->nearObs = PositionXY(p.nearObs);
}

/**
 *  \brief init constructor
 *  \param s : curve abscica
 *  \param minWidthB : lateral minimal bound
 *  \param maxWidthB : lateral maximal bound
 *  \param maxDistPoint : position of the interval maximizing its distance to obstacles
 *  \param dist : distance to the closest obstacle
 *  \param nearObs : position of the closest obstacle
 */
Passageway::Passageway(float s, float minWidthB, float maxWidthB, const PositionXY& maxDistPoint, float dist, const PositionXY& nearObs)
{
  this->s = s;
  this->minWidthB = minWidthB;
  this->maxWidthB = maxWidthB;
  this->maxDistPoint = PositionXY(maxDistPoint);
  this->dist = dist;
  this->nearObs = PositionXY(nearObs);
}

/**
 *  \brief destructor
 */
Passageway::~Passageway()
{
}


/******************* Accessors ********************/

/**
 *  \brief return the average lateral shift of the passageway wrt. the reference corridor
 *  \param tildY_origin : origin of the lateral shift on the corridor
 *  \return the average lateral shift of the passageway wrt. the reference corridor
 */
float Passageway::latShift(float tildY_origin) const
{
  return (minWidthB + maxWidthB)*0.5 - tildY_origin;
}


/**************** Display functions ***************/

/**
 *  \brief display the passageway
 */
void Passageway::display() const
{
  cout << "s : " << s << "  minWidthB : " << minWidthB << "  maxWidthB : " << maxWidthB;
  cout << "  maxDistPoint : "; this->maxDistPoint.display();
  cout << "  dist : " << dist << "  nearObs : "; this->nearObs.display(); 
  cout << endl;
}


                                    /**************************************************/
                                    /*************** Class PassagewayWP ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/
    
/**
 *  \brief default constructor
 */
PassagewayWP::PassagewayWP() : Passageway()
{
  this->cfg = XYThS_Config();
  this->v = 0.0;
  this->t = 0.0;
}

/**
 *  \brief copy constructor
 */
PassagewayWP::PassagewayWP(const PassagewayWP& p) : Passageway(p)
{
  this->cfg = p.cfg;
  this->v = p.v;
  this->t = p.t;
}

/**
 *  \brief init constructor
 *  \param s : curve abscica
 *  \param minWidthB : lateral minimal bound
 *  \param maxWidthB : lateral maximal bound
 *  \param maxDistPoint : position of the interval maximizing its distance to obstacles
 *  \param dist : distance to the closest obstacle
 *  \param nearObs : position of the closest obstacle
 *  \param cfg : the configuration that must be reach
 *  \param v : linear velocity at which it must be reached
 *  \param t : time at which it must be reached
 */
PassagewayWP::PassagewayWP(float s, float minWidthB, float maxWidthB, const PositionXY& maxDistPoint, float dist, const PositionXY& nearObs, XYThS_Config cfg, double v, double t) : Passageway(s, minWidthB, maxWidthB, maxDistPoint, dist, nearObs)
{
  this->cfg = cfg;
  this->v = v;
  this->t = t;
}

/**
 *  \brief destructor
 */
PassagewayWP::~PassagewayWP()
{
}

/**
 *  \brief return a copy of the current PassagewayWP instance
 *  \return a copy of the current PassagewayWP instance
 */
PassagewayWP* PassagewayWP::copy() const
{
  return new PassagewayWP(*this);
}


/******************* Accessors ********************/

/**
 *  \brief return the differential drive state-time corresponding to the current PassagewayWP
 *  \return the differential drive state-time corresponding to the current PassagewayWP
 */
DD2StateT PassagewayWP::toDD2StateT() const
{
  return DD2StateT(this->cfg.x(),
		   this->cfg.x(),
		   this->cfg.x(),
		   this->v, 
		   0.0,
		   this->t);
}

/**
 *  \brief return the waypoint corresponding to the current PassagewayWP
 *  \return the waypoint corresponding to the current PassagewayWP
 */
WayPoint PassagewayWP::toWayPoint() const
{
  WayPoint wp;
  wp.cfg = this->cfg;
  wp.dist = this->dist;
  return wp;
}
