#ifndef BOUNDINGSHAPE_HPP
#define BOUNDINGSHAPE_HPP

#include "common/inc/Position.hpp"

#include <vector>
#include <cmath>

namespace mpc
{

  /**
   * \enum bshape_t
   * \brief Enumeration of the different bounding boxes
   *
   * UNDEFINED_BSHAPE : undefined bounding shape
   * POINT_BSHAPE     : ponctual bounding shape
   * SPHERE_BSHAPE    : spherical bounding shape
   * POLYGON_BSHAPE   : polygonal bounding shape
   * BOX_BSHAPE       : rectangular bounding shape
   */
  typedef enum {
    UNDEFINED_BSHAPE     =  0,
    POINT_BSHAPE         =  1,
    SPHERE_BSHAPE        =  2,
    POLYGON_BSHAPE       =  3,
    BOX_BSHAPE           =  4
  } bshape_t;

                                    /**************************************************/
                                    /************** Class BoundingShape ***************/
                                    /**************************************************/

  /** \class BoundingShape
   *  \brief Bounding shape of an object
   */
  class BoundingShape
  {
  protected : 
    bshape_t type_;                       /**< type of the bounding box considered */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    BoundingShape();
    /**
     *  \brief copy constructor
     *  \param bs : the bounding shape to copy
     */
    BoundingShape(const BoundingShape& bs);
    /**
     *  \brief destructor
     */
    virtual ~BoundingShape();

    /******************** Accessors *******************/
  
    /**
     *  \brief Type getter
     *  \return the type of the bounding shape
     */
    inline bshape_t type() const { return this->type_; }

    //   /************* Obstruction functions **************/

    //   /**
    //    *  \brief compute the nearest position of the shape from position p
    //    *  \param p : position from which we want to compute the closest point
    //    *  \param near : resulting nearest position
    //    *  \param dist : distance from p to near
    //    *  \return true if p is in the box, false else
    //    */
    //   virtual bool nearPosFrom(const PositionXY& p, PositionXY& near, double& dist) const = 0;

    /*************** Display functions ****************/
  
    /**
     *  \brief display the bounding shape's features
     */
    virtual void display() const = 0;

    /**
     *  \brief draw the bounding shape in an openGL viewer
     */
    virtual void toGL() const = 0;
  };

};
#endif // BOUNDINGSHAPE_HPP


#ifndef POINTSHAPE_HPP
#define POINTSHAPE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class PointShape ****************/
                                    /**************************************************/

  /** \class PointShape
   *  \brief Point shape
   */
  class PointShape : public BoundingShape
  {
  protected :

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PointShape();
    /**
     *  \brief copy constructor
     *  \param bs : the bounding shape to copy
     */
    PointShape(const PointShape& bs);
    /**
     *  \brief destructor
     */
    virtual ~PointShape();

    /*************** Display functions ****************/
  
    /**
     *  \brief display the point shape's features
     */
    virtual void display() const;

    /**
     *  \brief draw the bounding shape in an openGL viewer
     */
    virtual void toGL() const;
  };

};
#endif // POINTSHAPE_HPP


#ifndef BSPHERE_HPP
#define BSPHERE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class BDisk *******************/
                                    /**************************************************/

  /** \class BDisk
   *  \brief Spherical bounding shape
   */
  class BDisk : public PointShape
  {
  protected : 
    double radius_;                     /**< radius of the spherical bounding shape */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    BDisk();

  public : 
    /**
     *  \brief copy constructor
     *  \param bs : the spherical bounding shape to copy
     */
    BDisk(const BDisk& bb);
    /**
     *  \brief init constructor
     *  \param r : radius of the spherical bounding shape
     */
    BDisk(double r);
    /**
     *  \brief destructor
     */
    virtual ~BDisk();

    /******************** Accessors *******************/
  
    /**
     *  \brief Radius getter
     *  \return the radius of the spherical bounding shape
     */
    inline double radius() const { return this->radius_; }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the bounding sphere's features
     */
    virtual void display() const;

    /**
     *  \brief draw the bounding sphere in an openGL viewer
     */
    virtual void toGL() const;
  };

};
#endif // BSPHERE_HPP


#ifndef BPOLYGON_HPP
#define BPOLYGON_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class BPolygon *****************/
                                    /**************************************************/

  /** \class BPolygon
   *  \brief Polygonal bounding shape
   */
  class BPolygon : public BoundingShape
  {
  protected : 
    std::vector<Position*> bounds_;                /**< positions defining the bounds of the polygon */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    BPolygon();
    /**
     *  \brief copy constructor
     *  \param bs : the spherical bounding shape to copy
     */
    BPolygon(const BPolygon& bp);
    /**
     *  \brief destructor
     */
    virtual ~BPolygon();

    /******************* Accessors ********************/

    /**
     *  \brief getter : return the i-th position defining the bounds 
     *  \param i : index of the bound to return
     *  \return a reference to the i-th position defining the bounds 
     */
    Position& getBound(int i) const;

    /**
     *  \brief getter : return the number of bounds of the polygon
     *  \return the number of bounds of the polygon 
     */
    int getNbBounds() const;

    /**
     *  \brief add a bound to the polygon
     *  \param the new position defining a bound of the polygon
     */
    void addBound(const Position& p);

    /**
     *  \brief delete all bounds of the polygon
     */
    void clearBounds();

    /*************** Display functions ****************/
  
    /**
     *  \brief display the bounding polygon's features
     */
    virtual void display() const;

    /**
     *  \brief draw the bounding polygon in an openGL viewer
     */
    virtual void toGL() const;
  };

};
#endif // BPOLYGON_HPP


#ifndef BBOX_HPP
#define BBOX_HPP

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class BBox *******************/
                                    /**************************************************/

  /** \class BBox
   *  \brief bounding box
   */
  class BBox : public PointShape
  {
  protected : 
    double lg_;                     /**< length of the bounding box */
    double wd_;                     /**< width of the bounding box */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    BBox();

  public : 
    /**
     *  \brief copy constructor
     *  \param bs : the spherical bounding shape to copy
     */
    BBox(const BBox& bb);
    /**
     *  \brief init constructor
     *  \param l : length of the bounding box
     *  \param w : width of the bounding box
     */
    BBox(const double& l,const double& w);
    /**
     *  \brief destructor
     */
    virtual ~BBox();

    /******************** Accessors *******************/

    /**
     *  \brief getter : return the length of the box
     *  \return the length of the box
     */
    inline double& lg() { return this->lg_; }

    /**
     *  \brief getter : return the width of the box
     *  \return the width of the box
     */
    inline double& wd() { return this->wd_; }

    /**
     *  \brief getter : return the length of the box
     *  \return the length of the box
     */
    inline double lg() const { return this->lg_; }

    /**
     *  \brief getter : return the width of the box
     *  \return the width of the box
     */
    inline double wd() const { return this->wd_; }

    /**
     *  \brief getter : return the coord of the upper left corner of the box in frame f
     *  \param f : frame of the box
     *  \return the coord of the upper left corner of the box
     */
    inline PositionXY getULCorner(const PositionXYTh& f) const 
    { return PositionXY(f.x() - 0.5*lg()*cos(f.th()) - 0.5*wd()*sin(f.th()),
			f.y() - 0.5*lg()*sin(f.th()) + 0.5*wd()*cos(f.th())); }

    /**
     *  \brief getter : return the coord of the upper right corner of the box in frame f
     *  \param f : frame of the box
     *  \return the coord of the upper right corner of the box
     */
    inline PositionXY getURCorner(const PositionXYTh& f) const 
    { return PositionXY(f.x() + 0.5*lg()*cos(f.th()) - 0.5*wd()*sin(f.th()),
			f.y() + 0.5*lg()*sin(f.th()) + 0.5*wd()*cos(f.th())); }

    /**
     *  \brief getter : return the coord of the lower left corner of the box in frame f
     *  \param f : frame of the box
     *  \return the coord of the lower left corner of the box
     */
    inline PositionXY getLLCorner(const PositionXYTh& f) const 
    { return PositionXY(f.x() - 0.5*lg()*cos(f.th()) + 0.5*wd()*sin(f.th()),
			f.y() - 0.5*lg()*sin(f.th()) - 0.5*wd()*cos(f.th())); }

    /**
     *  \brief getter : return the coord of the lower right corner of the box in frame f
     *  \param f : frame of the box
     *  \return the coord of the lower right corner of the box
     */
    inline PositionXY getLRCorner(const PositionXYTh& f) const 
    { return PositionXY(f.x() + 0.5*lg()*cos(f.th()) + 0.5*wd()*sin(f.th()),
			f.y() + 0.5*lg()*sin(f.th()) - 0.5*wd()*cos(f.th())); }

    /*************** Growing functions ****************/

    /**
     *  \brief grows the dimension of the bouding box
     *  \param v : dimension to grow the box
     */  
    virtual void grows(const double v);

  
    /*************** Display functions ****************/
  
    /**
     *  \brief display the bounding sphere's features
     */
    virtual void display() const;

    /**
     *  \brief draw the bounding sphere in an openGL viewer
     */
    virtual void toGL() const;
  };

};
#endif // BBOX_HPP


#ifndef COLLISION_HPP
#define COLLISION_HPP

namespace mpc
{
                                    /**************************************************/
                                    /*********** Static collision functions ***********/
                                    /**************************************************/

  /**
   *  \brief collision test between two arbitrary bounding shapes
   *  \param p1 : absolute position of the first bounding shape
   *  \param bs1 : the first bounding shape
   *  \param p2 : absolute position of the second bounding shape
   *  \param bs2 : the second bounding shape
   *  \param nearP1 : nearest position from bs1 to bs2
   *  \param nearP2 : nearest position from bs2 to bs1
   *  \param dist : distance between nearP1 and nearP2
   *  \return true if both bounding shapes are in collision
   */
  bool areInCollision(const PositionXY& p1,const BoundingShape& bs1,const PositionXY& p2,const BoundingShape& bs2,
		      PositionXY& nearP1, PositionXY& nearP2, double& dist);

  /**
   *  \brief collision test between two point shapes
   *  \param p1 : absolute position of the first point shape
   *  \param bs1 : the first point shape
   *  \param p2 : absolute position of the second point shape
   *  \param bs2 : the second point shape
   *  \param nearP1 : nearest position from ps1 to ps2
   *  \param nearP2 : nearest position from ps2 to ps1
   *  \param dist : distance between nearP1 and nearP2
   *  \return true if both point shapes are in collision
   */
  bool areInCollision(const PositionXY& p1,const PointShape& ps1,const PositionXY& p2,const PointShape& ps2,
		      PositionXY& nearP1, PositionXY& nearP2, double& dist);

  /**
   *  \brief collision test between a point shape and a bouding sphere
   *  \param p1 : absolute position of the first point shape
   *  \param bs : the bounding sphere
   *  \param p2 : absolute position of the second point shape
   *  \param ps : the point shape
   *  \param nearP1 : nearest position from bs to ps
   *  \param nearP2 : nearest position from ps to bs
   *  \param dist : distance between nearP1 and nearP2
   *  \return true if the point shape and the bouding sphere are in collision
   */
  bool areInCollision(const PositionXY& p1,const BDisk& bs,const PositionXY& p2,const PointShape& ps,
		      PositionXY& nearP1, PositionXY& nearP2, double& dist);


  /**
   *  \brief collision test between two bounding spheres
   *  \param p1 : absolute position of the first bounding sphere
   *  \param bs1 : the first bounding sphere
   *  \param p2 : absolute position of the second bounding sphere
   *  \param bs2 : the second bounding sphere
   *  \param nearP1 : nearest position from bs1 to bs2
   *  \param nearP2 : nearest position from bs2 to bs1
   *  \param dist : distance between nearP1 and nearP2
   *  \return true if both bounding spheres are in collision
   */
  bool areInCollision(const PositionXY& p1,const BDisk& bs1,const PositionXY& p2,const BDisk& bs2,
		      PositionXY& nearP1, PositionXY& nearP2, double& dist);

  /**
   *  \brief collision test between a point shape and a bouding box
   *  \param p1 : absolute position of the bounding box
   *  \param bb : the bounding box
   *  \param p2 : absolute position of the point shape
   *  \param ps : the point shape
   *  \param nearP1 : nearest position from bb to ps
   *  \param nearP2 : nearest position from ps to bb
   *  \param dist : distance between nearP1 and nearP2
   *  \return true if the point shape and the bouding box are in collision
   */
  bool areInCollision(const PositionXYTh& p1,const BBox& bb,const PositionXYTh& p2,const PointShape& ps,
		      PositionXY& nearP1, PositionXY& nearP2, double& dist);

  /**
   *  \brief collision test between a bounding box and a bounding sphere
   *  \param p1 : absolute position of the bounding box
   *  \param bb : the bounding box
   *  \param p1 : absolute position of the bounding sphere
   *  \param bs : the bounding sphere
   *  \param nearP1 : nearest position from bb to bs
   *  \param nearP2 : nearest position from bs to bb
   *  \param dist : distance between nearP1 and nearP2
   *  \return true if the bounding box and the bouding sphere are in collision
   */
  bool areInCollision(const PositionXYTh& p1,const BBox& bb,const PositionXYTh& p2,const BDisk& bs,
		      PositionXY& nearP1, PositionXY& nearP2, double& dist);

  /**
   *  \brief collision test between two bounding boxes
   *  \param p1 : absolute position of the first bounding box
   *  \param bb1 : the first bounding box
   *  \param p2 : absolute position of the second bounding box
   *  \param bb2 : the second bounding box
   *  \param nearP1 : nearest position from bb1 to bb2
   *  \param nearP2 : nearest position from bb2 to bb1
   *  \param dist : distance between nearP1 and nearP2
   *  \return true if both bounding boxes are in collision
   */
  bool areInCollision(const PositionXYTh& p1,const BBox& bb1,const PositionXYTh& p2,const BBox& bb2,
		      PositionXY& nearP1, PositionXY& nearP2, double& dist);


  /**
   *  \brief collision test between a bounding disk and a segment
   *  \param p1 : absolute position of the bounding disk
   *  \param bd : the bounding disk
   *  \param p1 : first end position of the segment
   *  \param p2 : second end position of the segment
   *  \param itsc1 : first intersection between the disk and the segment
   *  \param itsc2 : second intersection between the disk and the segment
   *  \param dist : closest distance from the segment to the disk (negative distance if they are in collision)
   *  \return the number of intersections (0, 1 or 2)
   */
  int areInCollision(const PositionXYTh& pos,const BDisk& bd,const PositionXY& p1,const PositionXY& p2,
		     PositionXY& itsc1, PositionXY& itsc2, double& dist);

  /**
   *  \brief collision test between a bounding box and a segment
   *  \param p1 : absolute position of the bounding box
   *  \param bd : the bounding box
   *  \param p1 : first end position of the segment
   *  \param p2 : second end position of the segment
   *  \param itsc1 : first intersection between the disk and the segment
   *  \param itsc2 : second intersection between the disk and the segment
   *  \param dist : closest distance from the segment to the box (negative distance if they are in collision)
   *  \return the number of intersections (0, 1 or 2)
   */
  int areInCollision(const PositionXYTh& pos,const BBox& bb,const PositionXY& p1,const PositionXY& p2,
		     PositionXY& itsc1, PositionXY& itsc2);

  /**
   *  \brief collision test between a bounding box and a segment
   *  \param p1 : absolute position of the bounding box
   *  \param bd : the bounding box
   *  \param p1 : first end position of the segment
   *  \param p2 : second end position of the segment
   *  \param itsc1 : first intersection between the disk and the segment
   *  \param itsc2 : second intersection between the disk and the segment
   *  \param dist : closest distance from the segment to the box (negative distance if they are in collision)
   *  \return the number of intersections (0, 1 or 2)
   */
  int areInCollision(const PositionXYTh& pos,const BBox& bb,const PositionXY& p1,const PositionXY& p2,
		     PositionXY& itsc1, PositionXY& itsc2, double& dist);
};
#endif // COLLISION_HPP
