#ifndef MPC_OBJECT_HPP
#define MPC_OBJECT_HPP

#include "common/inc/Position.hpp"
#include "common/inc/BoundingShape.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class Object ******************/
                                    /**************************************************/

  /** \class Object
   *  \brief Physical entity in the workspace (W)
   *
   *  Parent of all entities constituing or interacting 
   *  with a robotic system
   */
  class Object
  {
  protected : 
    //   Position* pos_;                                 /**< position of the object in its global workspace */
    BoundingShape* bshape_;                         /**< bounding box enclosing the object */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    Object();

  public : 
    /**
     *  \brief copy constructor
     *  \param o : the object to copy
     */
    Object(const Object& o);
    /**
     *  \brief destructor
     */
    virtual ~Object();

    /******************** Accessors *******************/
  
    //   /**
    //    *  \brief Position accessor
    //    *  \return the position of the object in its workspace (W)
    //    */
    //   virtual inline Position& getPos() const { return *pos_; }

    //   /**
    //    *  \brief Position accessor
    //    *  \return the position of the object in its workspace (W)
    //    */
    //   inline Position pos() const { return *pos_; }

  public : 
    /**
     *  \brief Bounding shape accessor
     *  \return the bounding shape of the object
     */
    inline BoundingShape& bshape() const { return *bshape_; }

  protected : 
    /**
     *  \brief setter : set the bounding shape of the object
     *  \param bs : the new bounding shape of the object
     */
    void setBShape(const BoundingShape& bs);

    //   /*************** Collision methods ****************/

    //   /**
    //    *  \brief collision test
    //    *  \param o : the object to test the collision with
    //    *  \return true if the current object in in collision with the object o
    //    *
    //    *  This method must be later modified to consider the relative 
    //    *  position of the bounding shape to the object
    //    */
    //   inline bool isInCollisionWith(Object& o,PositionXY& nearP1,PositionXY& nearP2,double& dist) const 
    //   { return areInCollision( *(this->pos_), *(this->bshape_), *(o.pos_), *(o.bshape_) ,nearP1, nearP2, dist); }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the object's features
     */
    virtual void display() const = 0;

    /**
     *  \brief draw the object in an openGL viewer
     */
    virtual void toGL() const = 0;
  };

};
#endif // MPC_OBJECT_HPP


#ifndef MPC_OBJECT2D_HPP
#define MPC_OBJECT2D_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class Object2D *****************/
                                    /**************************************************/

  /** \class Object2D
   *  \brief Physical entity in the workspace (W = R^2)
   */
  class Object2D : public Object
  {
  protected : 

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    Object2D();

    /**
     *  \brief init constructor
     *  \param bs : the bounding shape of the object
     */
    Object2D(const BoundingShape& bs);

  public : 
    /**
     *  \brief copy constructor
     *  \param o : the object to copy
     */
    Object2D(const Object2D& o);
    /**
     *  \brief destructor
     */
    virtual ~Object2D();

    /******************** Accessors *******************/
  
  public : 
    /**
     *  \brief Position accessor
     *  \return the position of the object in its workspace (W)
     */
    virtual const PositionXYTh& pos() const = 0; // const { return *((PositionXYTh*)pos_); }

    /**
     *  \brief Position accessor
     *  \param t : time at which the position of the object is evaluated
     *  \return the position of the object in its workspace (W) at time t
     */
    virtual const PositionXYTh& pos(double t) const = 0;

  protected : 
    /**
     *  \brief Position accessor
     *  \return the position of the object in its workspace (W)
     */
    virtual PositionXYTh& pos() = 0; // const { return *((PositionXYTh*)pos_); }

    //   /**
    //    *  \brief setter : set the current position of the system component
    //    *  \param p : the new position of the system component
    //    */
    //   inline void setPos(const PositionXYTh& p) = 0; //{ this->pos_ = new PositionXYTh(p); }

    /**************** Motion functions ****************/
  
  protected : 
    /**
     *  \brief translate the system component
     *  \param p : vector of the translation
     */
    void translate(const PositionXY& p);
  
    /**
     *  \brief rotate the system component around p about an angle a
     *  \param p : center of the rotation
     *  \param a : angle of the rotation
     */
    void rotate(const PositionXY& p,const double& a);

    /*************** Collision methods ****************/

  public : 
    /**
     *  \brief collision test
     *  \param o : the object to test the collision with
     *  \param nearP1 : nearest position of the current object to the second object o
     *  \param nearP2 : nearest position of the object o to the current object
     *  \param dist : distance between nearP1 and nearP2
     *  \return true if the current object in in collision with the object o
     *
     *  This method must be later modified to consider the relative 
     *  position of the bounding shape to the object
     */
    inline bool isInCollisionWith(Object2D& o,PositionXY& nearP1,PositionXY& nearP2,double& dist) const 
    { return areInCollision(this->pos(), *(this->bshape_), o.pos(), *(o.bshape_) ,nearP1, nearP2, dist); }

    /**
     *  \brief collision test
     *  \param o : the object to test the collision with
     *  \param t : time at which positions of the objects are evaluated
     *  \param nearP1 : nearest position of the current object to the second object o
     *  \param nearP2 : nearest position of the object o to the current object
     *  \param dist : distance between nearP1 and nearP2
     *  \return true if the current object in in collision with the object o
     *
     *  This method must be later modified to consider the relative 
     *  position of the bounding shape to the object
     */
    inline bool isInCollisionWith(Object2D& o,double t,PositionXY& nearP1,PositionXY& nearP2,double& dist) const 
    { return areInCollision(this->pos(t), *(this->bshape_), o.pos(t), *(o.bshape_) ,nearP1, nearP2, dist); }


    /*************** Display functions ****************/
  
    /**
     *  \brief display the object's features
     */
    virtual void display() const = 0;

    /**
     *  \brief draw the object in an openGL viewer
     */
    virtual void toGL() const = 0;
  };

};
#endif // MPC_OBJECT2D_HPP
