#ifndef MPC_POSITION_HPP
#define MPC_POSITION_HPP

#include "common/inc/Vector.hpp"
#include "common/inc/Utils.hpp"

#include <vector>

                                    /**************************************************/
                                    /***************** Class Position *****************/
                                    /**************************************************/

namespace mpc
{

  /** \class Position
   *  \brief Position in the workspace (W)
   */
  class Position : public Vector
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Position();
    /**
     *  \brief copy constructor
     *  \param p : the position to copy
     */
    Position(const Position& p);
    /**
     *  \brief destructor
     */
    virtual ~Position();

    /*************** Distance functions ***************/
  
    /**
     *  \brief distance computation between two positions
     *  \param p : the 2nd position to compute the distance to
     *  \return the distance between the current position and p
     */
    virtual double dist(const Position& p) const = 0;

    /**
     *  \brief square distance computation between two positions
     *  \param p : the 2nd position to compute the square distance to
     *  \return the square distance between the current position and p
     */
    virtual double sqDist(const Position& p) const = 0;

    /**************** Motion functions ****************/

    /**
     *  \brief translate the current position
     *  \param p : vector of the translation
     */
    void translate(const Position& p);

    /*************** Display functions ****************/
  
    /**
     *  \brief display the position's coordinates
     */
    virtual void display() const;
  };

};
#endif // MPC_POSITION_HPP


#ifndef MPC_POSITIONXY_HPP
#define MPC_POSITIONXY_HPP

                                    /**************************************************/
                                    /**************** Class PositionXY ****************/
                                    /**************************************************/

namespace mpc
{

  /** \class PositionXY
   *  \brief 2D Position in the workspace (W = XxY)
   */
  class PositionXY : public Position
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PositionXY();
    /**
     *  \brief copy constructor
     *  \param p : the position to copy
     */
    PositionXY(const PositionXY& p);
    /**
     *  \brief init constructor
     *  \param x : the first position coord
     *  \param y : the second position coord
     */
    PositionXY(double x,double y);
    /**
     *  \brief destructor
     */
    virtual ~PositionXY();

    /******************** Accessors *******************/
  
    /**
     *  \brief accessor of the first coordinate
     *  \return the first coordinate of the position
     */
    inline double& x() { return (*this)[0]; } // { return this->at(0); }

    /**
     *  \brief accessor of the first coordinate
     *  \return the first coordinate of the position
     */
    inline double& y() { return (*this)[1]; } // { return this->at(1); }

    /**
     *  \brief accessor of the first coordinate
     *  \return the first coordinate of the position
     */
    inline double x() const { return this->at(0); } // { return this->at(0); }

    /**
     *  \brief accessor of the first coordinate
     *  \return the first coordinate of the position
     */
    inline double y() const { return this->at(1); } // { return this->at(1); }


    /*************** Distance functions ***************/
  
    /**
     *  \brief distance computation between two positions
     *  \param p : the 2nd position to compute the distance to
     *  \return the distance between the current position and p
     */
    virtual double dist(const Position& p) const;

    /**
     *  \brief square distance computation between two positions
     *  \param p : the 2nd position to compute the square distance to
     *  \return the square distance between the current position and p
     */
    virtual double sqDist(const Position& p) const;

    /**************** Motion functions ****************/

    /**
     *  \brief translate the current position
     *  \param p : vector of the translation
     */
    virtual void translate(const PositionXY& p);

    /**
     *  \brief rotate the current position around p about an angle a
     *  \param p : center of the rotation
     *  \param a : angle of the rotation
     */
    virtual void rotate(const PositionXY& p,const double& a);

    /**
     *  \brief rotate the current position around (0,0) about an angle a
     *  \param a : angle of the rotation
     */
    virtual void rotate(const double& a);

    /******************* Operators ********************/

    /**
     *  \brief operator = : assignation
     *  \param p2 : the position to assign to the current one
     */
    void operator=(const PositionXY& p2);

    /**
     *  \brief operator + : sumation
     *  \param p2 : the second vector to add
     *  \return the sum of both positions
     */
    PositionXY operator+(const PositionXY& p2) const;

    /**
     *  \brief operator + : substraction
     *  \param p2 : the second position to substract
     *  \return the substraction of both positions
     */
    PositionXY operator-(const PositionXY& p2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current position and the scalar
     */
    PositionXY operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    PositionXY operator/(double d) const;

    /***************** Math functions *****************/
  
    /**
     *  \brief compute the normalized position vector from the current one
     *  \return the normalized position vector
     */
    PositionXY normalized() const;
  };

};
#endif // MPC_POSITIONXY_HPP


#ifndef MPC_POSITIONXYTH_HPP
#define MPC_POSITIONXYTH_HPP

                                    /**************************************************/
                                    /*************** Class PositionXYTh ***************/
                                    /**************************************************/

namespace mpc
{

  /** \class PositionXYTh
   *  \brief 3D Position in the workspace (W = XxYxTH)
   */
  class PositionXYTh : public PositionXY
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PositionXYTh();
    /**
     *  \brief copy constructor
     *  \param p : the position to copy
     */
    PositionXYTh(const PositionXYTh& p);
    /**
     *  \brief init constructor
     *  \param x : the first position coord
     *  \param y : the second position coord
     *  \param th : the third position coord (orientation)
     */
    PositionXYTh(double x,double y,double th);
    /**
     *  \brief destructor
     */
    virtual ~PositionXYTh();

    /******************** Accessors *******************/
  
    /**
     *  \brief accessor of the first coordinate
     *  \return the third coordinate of the position
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief accessor of the first coordinate
     *  \return the third coordinate of the position
     */
    inline double th() const { return this->at(2); }

    /**************** Motion functions ****************/

    /**
     *  \brief rotate the current position around p about an angle a
     *  \param p : center of the rotation
     *  \param a : angle of the rotation
     */
    virtual void rotate(const PositionXY& p,const double& a);

    /**
     *  \brief rotate the current position around (0,0) about an angle a
     *  \param a : angle of the rotation
     */
    virtual void rotate(const double& a);

    /******************* Operators ********************/

    /**
     *  \brief operator = : assignation
     *  \param p2 : the position to assign to the current one
     */
    void operator=(const PositionXYTh& p2);

    /**
     *  \brief operator + : sumation
     *  \param p2 : the second vector to add
     *  \return the sum of both positions
     */
    PositionXYTh operator+(const PositionXYTh& p2) const;

    /**
     *  \brief operator + : substraction
     *  \param p2 : the second position to substract
     *  \return the substraction of both positions
     */
    PositionXYTh operator-(const PositionXYTh& p2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current position and the scalar
     */
    PositionXYTh operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    PositionXYTh operator/(double d) const;

    /***************** Math functions *****************/
  
    /**
     *  \brief compute the normalized position vector from the current one
     *  \return the normalized position vector
     */
    PositionXYTh normalized() const;

    /*************** Direction functions **************/
  
    /**
     *  \brief return the cardinal direction of a position p from the current position
     *  \param p : the position to evaluate
     *  \return cardinal direction of a position p from the current position
     */
    cd::carddir_t cardDir(const PositionXY& p) const;

    /**
     *  \brief return true if the position p is in the front of the current position
     *  \param p : the position to evaluate
     *  \return true if the position p is in the front of the current position
     */
    bool isInTheFront(const PositionXY& p) const;

    /**
     *  \brief return true if the position p is in the back of the current position
     *  \param p : the position to evaluate
     *  \return true if the position p is in the back of the current position
     */
    bool isInTheBack(const PositionXY& p) const;

    /**
     *  \brief return true if the position p is on the left of the current position
     *  \param p : the position to evaluate
     *  \return true if the position p is on the left of the current position
     */
    bool isOnTheLeft(const PositionXY& p) const;

    /**
     *  \brief return true if the position p is on the right of the current position
     *  \param p : the position to evaluate
     *  \return true if the position p is on the right of the current position
     */
    bool isOnTheRight(const PositionXY& p) const;
  };

};
#endif // MPC_POSITIONXYTH_HPP


#ifndef MPC_POSITIONXYZ_HPP
#define MPC_POSITIONXYZ_HPP

                                    /**************************************************/
                                    /**************** Class PositionXYZ ***************/
                                    /**************************************************/

namespace mpc
{

  /** \class PositionXYZ
   *  \brief 3D Position in the workspace (W = XxYxZ)
   */
  class PositionXYZ : public PositionXY
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PositionXYZ();
    /**
     *  \brief copy constructor
     *  \param p : the position to copy
     */
    PositionXYZ(const PositionXYZ& p);
    /**
     *  \brief init constructor
     *  \param x : the first position coord
     *  \param y : the second position coord
     *  \param z : the third position coord
     */
    PositionXYZ(double x,double y,double z);
    /**
     *  \brief destructor
     */
    virtual ~PositionXYZ();

    /******************** Accessors *******************/
  
    /**
     *  \brief accessor of the first coordinate
     *  \return the third coordinate of the position   */
    inline double& z() { return (*this)[2]; }

    /**
     *  \brief accessor of the first coordinate
     *  \return the third coordinate of the position   */
    inline double z() const { return this->at(2); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : assignation
     *  \param p2 : the position to assign to the current one
     */
    void operator=(const PositionXYZ& p2);

    /**
     *  \brief operator + : sumation
     *  \param p2 : the second vector to add
     *  \return the sum of both positions
     */
    PositionXYZ operator+(const PositionXYZ& p2) const;

    /**
     *  \brief operator + : substraction
     *  \param p2 : the second position to substract
     *  \return the substraction of both positions
     */
    PositionXYZ operator-(const PositionXYZ& p2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current position and the scalar
     */
    PositionXYZ operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    PositionXYZ operator/(double d) const;

    /**
     *  \brief operator ^ : cross product
     *  \param p2 : the 3D position vector to do the cross product with
     *  \return cross product of both position vectors
     */  
    PositionXYZ operator^(const PositionXYZ& p2) const;

    /*************** Distance functions ***************/
  
    /**
     *  \brief distance computation between two positions
     *  \param p : the 2nd position to compute the distance to
     *  \return the distance between the current position and p
     */
    virtual double dist(const PositionXYZ& p) const;

    /**
     *  \brief square distance computation between two positions
     *  \param p : the 2nd position to compute the square distance to
     *  \return the square distance between the current position and p
     */
    virtual double sqDist(const PositionXYZ& p) const;

    /***************** Math functions *****************/
  
    /**
     *  \brief compute the normalized position vector from the current one
     *  \return the normalized position vector
     */
    PositionXYZ normalized() const;
  };

};
#endif // MPC_POSITIONXYTH_HPP


#ifndef MPC_POSITION6D_HPP
#define MPC_POSITION6D_HPP

                                    /**************************************************/
                                    /**************** Class Position6D ****************/
                                    /**************************************************/

namespace mpc
{

  /** \class Position6D
   *  \brief Position in the workspace (W=R^3 * S^3)
   */
  class Position6D : public PositionXYZ
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Position6D();
    /**
     *  \brief init constructor
     */
    Position6D(double x,double y,double z,double yaw,double pitch,double roll);
    /**
     *  \brief copy constructor
     *  \param p : the position to copy
     */
    Position6D(const Position6D& p);
    /**
     *  \brief destructor
     */
    virtual ~Position6D();

    /******************** Accessors *******************/
  
    /**
     *  \brief accessor of the sixth coordinate
     *  \return the sixth coordinate of the position   */
    inline double& yaw() { return (*this)[3]; }
  
    /**
     *  \brief accessor of the sixth coordinate
     *  \return the sixth coordinate of the position   */
    inline double yaw() const { return this->at(3); }

    /**
     *  \brief accessor of the fifth coordinate
     *  \return the fifth coordinate of the position   */
    inline double& pitch() { return (*this)[4]; }
  
    /**
     *  \brief accessor of the fifth coordinate
     *  \return the fifth coordinate of the position   */
    inline double pitch() const { return this->at(4); }

    /**
     *  \brief accessor of the fourth coordinate
     *  \return the fourth coordinate of the position   */
    inline double& roll() { return (*this)[5]; }
  
    /**
     *  \brief accessor of the fourth coordinate
     *  \return the fourth coordinate of the position   */
    inline double roll() const { return this->at(5); }
  };

};
#endif // MPC_POSITION6D_HPP


#ifndef MPC_MANAPOSE_HPP
#define MPC_MANAPOSE_HPP

                                    /**************************************************/
                                    /**************** Class Position6D ****************/
                                    /**************************************************/

namespace mpc
{

  /** \class Position6D
   *  \brief Position in the workspace (W=R^3 * S^3)
   */
  class ManaPose : public Position6D
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    ManaPose();
    /**
     *  \brief init constructor
     */
    ManaPose(double x,double y,double z,double yaw,double pitch,double roll);
    /**
     *  \brief copy constructor
     *  \param p : the position to copy
     */
    ManaPose(const ManaPose& p);
    /**
     *  \brief destructor
     */
    virtual ~ManaPose();
  };
};
#endif // MPC_MANAPOSE_HPP


#ifndef MPC_FRAMECHANGE_HPP
#define MPC_FRAMECHANGE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /************* Intersections functions ************/
                                    /**************************************************/

  /**
   *  \brief compute the intersection between a straight line and a circle
   *  \param posL : starting point of the straight line
   *  \param vecL : vector defining the straight line
   *  \param posC : center of the circle
   *  \param radC : radius of the circle
   *  \param itsc1 : first intersection found
   *  \param itsc2 : second intersection found
   *  \return the number of intersections found
   */
  int intersecLineCircle(const PositionXY& posL,const PositionXY& vecL,const PositionXY& posC,double radC,PositionXY& itsc1,PositionXY& itsc2);

  /**
   *  \brief compute the intersection between a straight line and a circle
   *  \param posL : starting point of the straight line
   *  \param vecL : vector defining the straight line
   *  \param posC : center of the circle
   *  \param radC : radius of the circle
   *  \param itsc1 : first intersection found
   *  \param itsc2 : second intersection found
   *  \param dist1 : distance between posL and the first intersection (negative if the intersection is behind the obstacle)
   *  \param dist2 : distance between posL and the second intersection (negative if the intersection is behind the obstacle)
   *  \return the number of intersections found
   */
  int intersecLineCircle(const PositionXY& posL,const PositionXY& vecL,const PositionXY& posC,double radC,
			 PositionXY& itsc1,PositionXY& itsc2,double& dist1,double& dist2);

  /**
   *  \brief compute the intersection between a segment and a circle
   *  \param posL : starting point of the straight line
   *  \param vecL : vector defining the straight line
   *  \param posC : center of the circle
   *  \param radC : radius of the circle
   *  \param itsc1 : first intersection found
   *  \param itsc2 : second intersection found
   *  \return the number of intersections found
   *
   *  Warning : This function won't find any intersection if the segment is completely included in the circle
   */
  int intersecSgmtCircle(const PositionXY& p1,const PositionXY& p2,const PositionXY& posC,double radC,PositionXY& itsc1,PositionXY& itsc2);

  /**
   *  \brief compute the intersection between a straight line and a circle
   *  \param pos1 : first end position of the first segment
   *  \param vec1 : second end position of the first segment
   *  \param pos2 : first end position of the second segment
   *  \param vec2 : second end position of the second segment
   *  \param itsc : the intersection found (if exists)
   *  \return true if an intersection has been found, false else (both lines are colinear)
   */
  bool intersecLines(const PositionXY& pos1,const PositionXY& vec1,const PositionXY& pos2,const PositionXY& vec2,PositionXY& itsc);

  /**
   *  \brief compute the intersection between a straight line and a circle
   *  \param pos11 : first end position of the first segment
   *  \param pos12 : second end position of the first segment
   *  \param pos21 : first end position of the second segment
   *  \param pos22 : second end position of the second segment
   *  \param itsc  : the intersection found (if exists)
   *  \return true if an intersection has been found, false else (both segments are colinear)
   */
  bool intersecSgmts(const PositionXY& pos11,const PositionXY& pos12,const PositionXY& pos21,const PositionXY& pos22,PositionXY& itsc);


                                    /**************************************************/
                                    /************** Projections functions *************/
                                    /**************************************************/

  /**
   *  \brief orthogonal projection of a 2D point on a line defined by a point and a vector
   *  \param pos : the position to project
   *  \param pline : the position defining the line
   *  \param vline : the vector defining the line
   *  \return the orthogonal projection of pos on the line
   */
  PositionXY orthoProjLine(const PositionXY& pos,const PositionXY& pline,const PositionXY& vline);

  /**
   *  \brief orthogonal projection of a 2D point on a segment defined by its end points
   *  \param pos : the position to project
   *  \param p1 : the first end point of the segment
   *  \param p2 : the second end point of the segment
   *  \return the orthogonal projection of pos on the segment
   */
  PositionXY orthoProjSgmt(const PositionXY& pos,const PositionXY& p1,const PositionXY& p2);


                                    /**************************************************/
                                    /*************** Distance functions ***************/
                                    /**************************************************/

  /**
   *  \brief compute the closest distance from a point to a line
   *  \param pos : the position to evaluate
   *  \param pline : the position defining the line
   *  \param vline : the vector defining the line
   *  \return the closest distance from a point to a line
   */
  double closeDistLine(const PositionXY& pos,const PositionXY& pline,const PositionXY& vline);

  /**
   *  \brief compute the closest distance from a point to a segment
   *  \param pos : the position to evaluate
   *  \param p1 : the first end point of the segment
   *  \param p2 : the second end point of the segment
   *  \return the closest distance from a point to a line
   */
  double closeDistSgmt(const PositionXY& pos,const PositionXY& p1,const PositionXY& p2);

  /**
   *  \brief compute the closest distance from a point to a segment
   *  \param pos : the position to evaluate
   *  \param p1 : the first end point of the segment
   *  \param p2 : the second end point of the segment
   *  \param minDist : the closest distance from a point to a line
   *  \return the closest position of segment p1-p2 to position pos
   */
  PositionXY closePosFromSgmt(const PositionXY& pos,const PositionXY& p1,const PositionXY& p2, double& minDist);


                                    /**************************************************/
                                    /************* Frame changes functions ************/
                                    /**************************************************/

  /**
   *  \brief frame change function
   *  \param p : coord of the position in frame (0;0)
   *  \param f : coord of the new frame
   *  \return the coordinates of position p in the new frame f from the frame (0;0)
   */  
  PositionXYTh coordInFrame(const PositionXYTh& p,const PositionXYTh& f);

  /**
   *  \brief frame change function
   *  \param p : coord of the position in frame f
   *  \param f : coord of the old frame
   *  \return the coordinates of position p in the new frame (0;0) from the frame f
   */  
  PositionXYTh coordFromFrame(const PositionXYTh& p,const PositionXYTh& f);

  /**
   *  \brief homothetic transformation
   *  \param p : coord of the input position
   *  \param hCenter : center of the homothetic transformation
   *  \param hCoef : coefficients of the homothetic transformation
   *  \return the homothetic transformation of the position p
   */  
  PositionXY homothety(const PositionXY& p,const PositionXY& hCenter,const PositionXY& hCoef);
};
#endif // MPC_FRAMECHANGE_HPP
