#ifndef MPC_STATETIMETREE_HPP
#define MPC_STATETIMETREE_HPP

#include "common/inc/Tree.hpp"
#include "common/inc/StateTime.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class StateTimeNode **************/
                                    /**************************************************/

  /*  \class StateTimeNode
   *  \brief node class for construction of a state-time tree
   */
  class StateTimeNode : public Node<StateTime*>
  {
  public : 
    static int displayPrecision_;                      /**< display precision used */
    static int displayTab_;                            /**< display tabulation used */

    /********** Constructors and destructors **********/

  protected :
    /**
     *  \brief default constructor
     */  
    StateTimeNode();

  public : 
    /**
     *  \brief copy constructor
     *  \param n : the node to copy
     */  
    StateTimeNode(const StateTimeNode& n);

    /**
     *  \brief init constructor
     *  \param val : the state-time contained in the node
     */  
    StateTimeNode(const StateTime& s);
  
    /**
     *  \brief destructor
     */  
    virtual ~StateTimeNode();


    /*************** Getters and Setters **************/

  public : 
//     /**
//      *  \brief return the father of the current node
//      *  \return a reference to the father of the current node
//      */
//     inline StateTimeNode*& father() { return ((StateTimeNode*)(this->father_)); }

    /**
     *  \brief return the father of the current node
     *  \return a pointer to the father of the current node
     */
    inline StateTimeNode* father() const { return ((StateTimeNode*)(this->father_)); }

    /**
     *  \brief return the i-th son of the current node
     *  \param i : the index of the son
     *  \return the i-th son of the current node
     */
    inline StateTimeNode* son(int i) const { return ((StateTimeNode*)(this->sons_.at(i))); }

    /**
     *  \brief add a son to the current node
     *  \param s : a pointer to the new son
     */
    void addSon(StateTimeNode* s);

    /**
     *  \brief add a son to the current node
     *  \param s : the state-time value of the son
     */
    void addSon(const StateTime& s);


    /**************** Utility functions ***************/

    /**
     *  \brief return the branch of the tree reaching the position pos and minimizing the final time
     *  \param targetPos : the target position
     *  \param minTimeBranch : the resulting vector containing all the nodes of the best branch
     */
    void computeMinTimeBranch(const PositionXY& targetPos, std::list<StateTime*> minTimeBranch);


    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the state-time tree from the current root
     *  \param level : input level from the root of the tree used to shift the current display
     */
    void display(int level=0) const;
  };


                                    /**************************************************/
                                    /************* Class DD2StateTimeNode *************/
                                    /**************************************************/

  /*  \class DD2StateTimeNode
   *  \brief node class for construction of a differential drive state-time tree
   */
  class DD2StateTimeNode : public StateTimeNode
  {
    /********** Constructors and destructors **********/

  protected :
    /**
     *  \brief default constructor
     */  
    DD2StateTimeNode();

  public : 
    /**
     *  \brief copy constructor
     *  \param n : the node to copy
     */  
    DD2StateTimeNode(const DD2StateTimeNode& n);

    /**
     *  \brief init constructor
     *  \param val : the state-time contained in the node
     */  
    DD2StateTimeNode(const DD2StateT& s);
  
    /**
     *  \brief destructor
     */  
    virtual ~DD2StateTimeNode();


    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the state-time of the current node
     *  \return the state-time of the current node
     */
    DD2StateT getStateT() const;
  };


                                    /**************************************************/
                                    /************** Class CLStateTimeNode *************/
                                    /**************************************************/

  /*  \class CLStateTimeNode
   *  \brief node class for construction of a car-like state-time tree
   */
  class CLStateTimeNode : public StateTimeNode
  {
    /********** Constructors and destructors **********/

  protected :
    /**
     *  \brief default constructor
     */  
    CLStateTimeNode();

  public : 
    /**
     *  \brief copy constructor
     *  \param n : the node to copy
     */  
    CLStateTimeNode(const CLStateTimeNode& n);

    /**
     *  \brief init constructor
     *  \param val : the state-time contained in the node
     */  
    CLStateTimeNode(const CLStateT& s);
  
    /**
     *  \brief destructor
     */  
    virtual ~CLStateTimeNode();


    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the state-time of the current node
     *  \return the state-time of the current node
     */
    CLStateT getStateT() const;
  };

};

#endif // MPC_STATETIMETREE_HPP
