#ifndef MPC_TIMER_HPP
#define MPC_TIMER_HPP

#include <time.h>
#include <sys/time.h>

namespace mpc
{
  /**
   * \enum timerState_t
   * \brief Enumeration of the different timer state types
   *
   * TIMER_NONINIT : non initialized timer
   * TIMER_STOPPED : stopped timer
   * TIMER_STARTED : started timer
   */
  typedef enum {
    TIMER_NONINIT  =  0,
    TIMER_STOPPED  =  1,
    TIMER_STARTED  =  2
  } timerState_t;

                                    /**************************************************/
                                    /******************* Class Timer ******************/
                                    /**************************************************/

  /** \class Timer
   *  \brief Timer to compute the time elasped during two processes
   */
  class Timer
  {
  protected : 
    struct timeval initTime_;                          /*< initial time at which we started the timer */
    struct timeval lastTime_;                          /*< last starting time saved by the timer */
    struct timeval lastRequestedTime_;                 /*< last requested time saved by the timer */
    struct timeval stopTime_;                          /*< time at which the timer has been stopped */
    timerState_t timerSt_;                             /*< state of the timer (started, stopped, or non initialized) */
    double sumTotElapsedTime_;                         /*< sum of elapsed time from the initial time */
    double sumElapsedTime_;                            /*< sum of elapsed time from the last time requested */
    double sumTotRequestedTime_;                       /*< total elapsed time until the last requested time saved by the timer */
    int tic_;                                          /*< number of time requests */

  public  : 
  
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Timer();

    /**
     *  \brief destructor
     */
    ~Timer();


    /******************** Accessors *******************/

  public : 

    /**
     *  \brief compute and return the last elapsed time
     *  \return the last elapsed time (in sec.)
     */
    double elapsedTime();

    /**
     *  \brief compute and return the total elapsed time
     *  \return the total elapsed time (in sec.)
     */
    double totalElapsedTime() const;

    /**
     *  \brief test if the timer is running
     *  \return true if the timer is running
     */
    inline bool isRunning() const { return (timerSt_ == TIMER_STARTED); }

    /**
     *  \brief return the average period
     *  \return the average period
     */
    double avgPeriod() const;

    /**
     *  \brief return the average frequency
     *  \return the average frequency
     */
    double avgFrequency() const;

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief init the timer with the current time
     */
    void init();

  public : 
    /**
     *  \brief start the timer
     */
    void start();

    /**
     *  \brief stop the timer
     */
    void stop();
  
    /**
     *  \brief reinitialize the timer
     *  \return 0 if the reinitialization succeed, -1 else
     */
    int reinit();

    /**
     *  \brief return the absolute init time
     *  \param time : the absolute init time
     */
    void absoluteInitTime(struct timeval& time);

    /**
     *  \brief return the absolute init time
     *  \param time : the absolute init time
     */
    void absoluteInitTime(double& time);
    

    /***************** Static functions ***************/

    /**
     *  \brief compute the elapsed time between two struct timeval t1 and t2
     *  \param t1 : the start timeval
     *  \param t2 : the end timeval
     *  \return the elapsed time
     */
    static double computeElapsedTime(const struct timeval& t1,const struct timeval& t2);

    /**
     *  \brief get back the absolute time
     *  \param time : resulting structure containing the current absolute time
     */
    static void absoluteTime(struct timeval& time);

    /**
     *  \brief get back the absolute time
     *  \param time : resulting current absolute time
     */
    static void absoluteTime(double& time);
  };


                                    /**************************************************/
                                    /************* struct timeval operators ***********/
                                    /**************************************************/

  /**
   *  \brief compare two struct timeval
   *  \param t1 : the first time to evaluate
   *  \param t2 : the second time to evaluate
   *  \return true if t1 < t2
   */
  bool operator<(const struct timeval& t1,const struct timeval& t2);

  /**
   *  \brief compare two struct timeval
   *  \param t1 : the first time to evaluate
   *  \param t2 : the second time to evaluate
   *  \return true if t1 > t2
   */
  bool operator>(const struct timeval& t1,const struct timeval& t2);

  /**
   *  \brief compare two struct timeval
   *  \param t1 : the first time to evaluate
   *  \param t2 : the second time to evaluate
   *  \return true if t1 <= t2
   */
  bool operator<=(const struct timeval& t1,const struct timeval& t2);

  /**
   *  \brief compare two struct timeval
   *  \param t1 : the first time to evaluate
   *  \param t2 : the second time to evaluate
   *  \return true if t1 >= t2
   */
  bool operator>=(const struct timeval& t1,const struct timeval& t2);

  /**
   *  \brief compare two struct timeval
   *  \param t1 : the first time to evaluate
   *  \param t2 : the second time to evaluate
   *  \return true if t1 == t2
   */
  bool operator==(const struct timeval& t1,const struct timeval& t2);

  // /**
  //  *  \brief copy a struct timeval
  //  *  \param t1 : the resulting time
  //  *  \param t2 : the time to copy
  //  */
  // void operator=(struct timeval& t1,const struct timeval& t2);

};
#endif // MPC_TIMER_HPP
