#include "common/inc/Config.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /******************* Class Config *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Config::Config() : Vector()
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
Config::Config(const Config& c) : Vector(c)
{
}

/**
 *  \brief init constructor
 *  \param n : dimension of the configuration
 */
Config::Config(int n) : Vector(n)
{
}

/**
 *  \brief destructor
 */
Config::~Config()
{
}


                                    /**************************************************/
                                    /***************** Class PMConfig *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PMConfig::PMConfig() : Config(2)
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
PMConfig::PMConfig(const PMConfig& c) : Config(c)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 */
PMConfig::PMConfig(double x,double y) : Config(2)
{
  this->x()  = x;
  this->y()  = y;
}

/**
 *  \brief destructor
 */
PMConfig::~PMConfig()
{
}

/**
 *  \brief init constructor
 *  \param n : dimension of the configuration
 *
 *  Only used by subclasses
 */
PMConfig::PMConfig(int n) : Config(n)
{
}


                                    /**************************************************/
                                    /***************** Class DPMConfig ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DPMConfig::DPMConfig() : PMConfig(3)
{
}

/**
 *  \brief copy constructor
 *  \param p : the configuration to copy
 */
DPMConfig::DPMConfig(const DPMConfig& c) : PMConfig(c)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 */
DPMConfig::DPMConfig(double x,double y,double th) : PMConfig(3)
{
  this->x()  = x;
  this->y()  = y;
  this->th()  = th;  
}

/**
 *  \brief destructor
 */
DPMConfig::~DPMConfig()
{
}

/**
 *  \brief init constructor
 *  \param n : dimension of the configuration
 *
 *  Only used by subclasses
 */
DPMConfig::DPMConfig(int n) : PMConfig(n)
{
}



                                    /**************************************************/
                                    /*************** Class XYThS_Config ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
XYThS_Config::XYThS_Config() : DPMConfig(4)
{
}

/**
 *  \brief copy constructor
 *  \param p : the configuration to copy
 */
XYThS_Config::XYThS_Config(const XYThS_Config& c) : DPMConfig(c)
{
}

/**
 *  \brief init constructor
 *  \param x  : position x
 *  \param y  : position y
 *  \param th : heading th
 */
XYThS_Config::XYThS_Config(double x,double y,double th,double s) : DPMConfig(4)
{
  this->x()  = x;
  this->y()  = y;
  this->th()  = th;  
  this->s()  = s;  
}

/**
 *  \brief destructor
 */
XYThS_Config::~XYThS_Config()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input configuration
 *  \param s2 : the configuration to copy
 */
void XYThS_Config::operator=(const XYThS_Config& c2)
{
  this->x()  = c2.x();
  this->y()  = c2.y();
  this->th()  = c2.th();  
  this->s()  = c2.s();  
}

/**
 *  \brief operator += : add the input configuration
 *  \param s2 : the configuration to add
 */
void XYThS_Config::operator+=(const XYThS_Config& c2)
{
  this->x()  += c2.x();
  this->y()  += c2.y();
  this->th()  += c2.th();  
  this->s()  += c2.s();  
}

/**
 *  \brief operator -= : substract the input configuration
 *  \param s2 : the configuration to substract
 */
void XYThS_Config::operator-=(const XYThS_Config& c2)
{
  this->x()  -= c2.x();
  this->y()  -= c2.y();
  this->th()  -= c2.th();  
  this->s()  -= c2.s();  
}

/**
 *  \brief operator += : multiply with the input coefficient
 *  \param d : the coefficient to mutilpy with
 */
void XYThS_Config::operator*=(double d)
{
  this->x()  *= d;
  this->y()  *= d;
  this->th()  *= d;  
  this->s()  *= d;  
}

/**
 *  \brief operator /= : divide with the input coefficient
 *  \param d : the coefficient to divide with
 */
void XYThS_Config::operator/=(double d)
{
  this->x()  /= d;
  this->y()  /= d;
  this->th()  /= d;  
  this->s()  /= d;  
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
XYThS_Config XYThS_Config::operator+(const XYThS_Config& c2) const
{
  return XYThS_Config(this->x()  + c2.x(),
		      this->y()  + c2.y(),
		      normalizedAngle(this->th() + c2.th()),
		      this->s()  + c2.s());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
XYThS_Config XYThS_Config::operator-(const XYThS_Config& c2) const
{
  return XYThS_Config(this->x()  - c2.x(),
		      this->y()  - c2.y(),
		      normalizedAngle(this->th() - c2.th()),
		      this->s()  - c2.s());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
XYThS_Config XYThS_Config::operator*(double d) const
{
  return XYThS_Config(this->x()  * d,
		      this->y()  * d,
		      normalizedAngle(this->th() * d),
		      this->s()  * d);
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
XYThS_Config XYThS_Config::operator/(double d) const
{
  return XYThS_Config(this->x()  / d,
		      this->y()  / d,
		      normalizedAngle(this->th() / d),
		      this->s()  / d);
}


                                    /**************************************************/
                                    /***************** Class CLConfig *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CLConfig::CLConfig() : DPMConfig()
{
}

/**
 *  \brief copy constructor
 *  \param p : the configuration to copy
 */
CLConfig::CLConfig(const CLConfig& c) : DPMConfig(c)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 */
CLConfig::CLConfig(double x,double y,double th) : DPMConfig(x,y,th)
{
}

/**
 *  \brief destructor
 */
CLConfig::~CLConfig()
{
}


                                    /**************************************************/
                                    /***************** Class SSConfig *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SSConfig::SSConfig() : DPMConfig()
{
}

/**
 *  \brief copy constructor
 *  \param p : the configuration to copy
 */
SSConfig::SSConfig(const SSConfig& c) : DPMConfig(c)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 */
SSConfig::SSConfig(double x,double y,double th) : DPMConfig(x,y,th)
{
}

/**
 *  \brief destructor
 */
SSConfig::~SSConfig()
{
}

                                    /**************************************************/
                                    /***************** Class DDConfig *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DDConfig::DDConfig() : DPMConfig()
{
}

/**
 *  \brief copy constructor
 *  \param p : the configuration to copy
 */
DDConfig::DDConfig(const DDConfig& c) : DPMConfig(c)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 */
DDConfig::DDConfig(double x,double y,double th) : DPMConfig(x,y,th)
{
}

/**
 *  \brief destructor
 */
DDConfig::~DDConfig()
{
}
