#include "Position.hpp"
using namespace mpc;

#include "Utils.hpp"

#include <cmath>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /***************** Class Position *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
Position::Position() : Vector()
{}

/*
 *  \brief copy constructor
 *  \param p : the position to copy
 */
Position::Position(const Position& p) : Vector(p)
{}

/*
 *  \brief destructor
 */
Position::~Position(){}


/**************** Motion functions ****************/

/*
 *  \brief translate the current position
 *  \param p : vector of the translation
 */
void Position::translate(const Position& p)
{
  if(this->size() == p.size())
    {
      for(unsigned int i=0; i<p.size(); i++)
	{
	  this->coord_.at(i) += p.at(i);
	}
    }
  else
    {
      cout << "Position::translate : Warning. Both position do not have the same vector size" << endl;
    }
}


/*************** Display functions ****************/

/*
 *  \brief display the position's coordinates
 */
void Position::display() const
{
  int N = this->size();

  cout << "(";
  for(int i=0; i<N-1; i++)
    cout << this->at(i) << ",";
  if(N != 0) cout << this->at(N-1);
  cout << ")";
}


                                    /**************************************************/
                                    /**************** Class PositionXY ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
PositionXY::PositionXY() : Position()
{
  this->coord_ = vector<double>(2);
}

/*
 *  \brief copy constructor
 *  \param p : the position to copy
 */
PositionXY::PositionXY(const PositionXY& p) : Position(p)
{
}

/*
 *  \brief init constructor
 *  \param x : the first position coord
 *  \param y : the second position coord
 */
PositionXY::PositionXY(double x,double y) : Position()
{
  this->coord_ = vector<double>(2);
  this->x() = x;
  this->y() = y;
}

/*
 *  \brief destructor
 */
PositionXY::~PositionXY()
{
}


/*************** Distance functions ***************/

/*
 *  \brief distance computation between two positions
 *  \param p : the 2nd position to compute the distance to
 *  \return the distance between the current position and p
 */
double PositionXY::dist(const Position& p) const
{
  if(p.size() >= 2)
    {
      double dx = this->at(0) - p.at(0);
      double dy = this->at(1) - p.at(1);
      return sqrt(dx*dx + dy*dy);
    }
  else return 0;
}

/*
 *  \brief square distance computation between two positions
 *  \param p : the 2nd position to compute the square distance to
 *  \return the square distance between the current position and p
 */
double PositionXY::sqDist(const Position& p) const
{
  if(p.size() >= 2)
    {
      double dx = this->at(0) - p.at(0);
      double dy = this->at(1) - p.at(1);
      return (dx*dx + dy*dy);
    }
  else return 0;
}


/**************** Motion functions ****************/

/*
 *  \brief translate the current position
 *  \param p : vector of the translation
 */
void PositionXY::translate(const PositionXY& p)
{
  this->x() += p.x();
  this->y() += p.y();
}

/*
 *  \brief rotate the current position around p about an angle a
 *  \param c : center of the rotation
 *  \param a : angle of the rotation
 */
void PositionXY::rotate(const PositionXY& c,const double& a)
{
  // compute the distance between both positions
  double sqNorm = this->sqDist(c);
  if(sqNorm == 0)
    {
      // nothing to do
    }
  else
    {
      double norm = sqrt(sqNorm);
     
      // compute the direction defined by the vector \overrightarrow{P this}
      double b = atan2(this->y() - c.y(), this->x() - c.x());

      // normalize the total angle
      double angle = normalizedAngle(a + b);

      // compute the new position
      this->x() = c.x() + norm * cos(angle);
      this->y() = c.y() + norm * sin(angle);
    }
}

/*
 *  \brief rotate the current position around (0,0) about an angle a
 *  \param a : angle of the rotation
 */
void PositionXY::rotate(const double& a)
{
  // compute the norm of the current vector
  double norm = sqrt(x()*x() + y()*y());

  if(norm == 0)
    {
      // nothing to do
    }
  else
    {
      // compute the direction defined by the vector \overrightarrow{P this}
      double b = atan2(this->y(),this->x());

      // normalize the total angle
      double angle = normalizedAngle(a + b);

      // compute the new position
      this->x() = norm * cos(angle);
      this->y() = norm * sin(angle);
    }
}


/******************* Operators ********************/

/*
 *  \brief operator = : assignation
 *  \param p2 : the position to assign to the current one
 */
void PositionXY::operator=(const PositionXY& p2)
{
  this->x() = p2.x();
  this->y() = p2.y();
}

/*
 *  \brief operator + : sumation
 *  \param p2 : the second position to add
 *  \return the sum of both positions
 */
PositionXY PositionXY::operator+(const PositionXY& p2) const
{
  PositionXY pres;
  pres.x() = this->x() + p2.x();
  pres.y() = this->y() + p2.y();
  return pres;
}

/*
 *  \brief operator + : substraction
 *  \param p2 : the second position to substract
 *  \return the substraction of both positions
 */
PositionXY PositionXY::operator-(const PositionXY& p2) const
{
  PositionXY pres;
  pres.x() = this->x() - p2.x();
  pres.y() = this->y() - p2.y();
  return pres;
}

/*
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current position and the scalar
 */
PositionXY PositionXY::operator*(double d) const
{
  PositionXY pres;
  pres.x() = this->x() * d;
  pres.y() = this->y() * d;
  return pres;
}

/*
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
PositionXY PositionXY::operator/(double d) const
{
  PositionXY pres;
  if(d != 0)
    {
      pres.x() = this->x() / d;
      pres.y() = this->y() / d;
    }
  return pres;
}


/***************** Math functions *****************/

/*
 *  \brief compute the normalized position vector from the current one
 *  \return the normalized position vector
 */
PositionXY PositionXY::normalized() const
{
  double norm = this->norm();
  
  if(norm != 0)
    {
      return ((*this) / norm);
    }
  else return PositionXY();
}


                                    /**************************************************/
                                    /*************** Class PositionXYTh ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
PositionXYTh::PositionXYTh()
{
  this->coord_ = vector<double>(3);
}

/*
 *  \brief copy constructor
 *  \param p : the position to copy
 */
PositionXYTh::PositionXYTh(const PositionXYTh& p) : PositionXY(p)
{
}

/*
 *  \brief init constructor
 *  \param x  : the first position coord
 *  \param y  : the second position coord
 *  \param th : the third position coord (orientation)
 */
PositionXYTh::PositionXYTh(double x,double y,double th)
{
  this->coord_ = vector<double>(3);
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
}

/*
 *  \brief destructor
 */
PositionXYTh::~PositionXYTh()
{
}


/**************** Motion functions ****************/

/*
 *  \brief rotate the current position around p about an angle a
 *  \param p : center of the rotation
 *  \param a : angle of the rotation
 */
void PositionXYTh::rotate(const PositionXY& p,const double& a)
{
  this->PositionXY::rotate(p,a);
  this->th() = normalizedAngle(this->th() + a);
}

/*
 *  \brief rotate the current position around (0,0) about an angle a
 *  \param a : angle of the rotation
 */
void PositionXYTh::rotate(const double& a)
{
  this->PositionXY::rotate(a);
  this->th() = normalizedAngle(this->th() + a);
}


/******************* Operators ********************/

/*
 *  \brief operator = : assignation
 *  \param p2 : the position to assign to the current one
 */
void PositionXYTh::operator=(const PositionXYTh& p2)
{
  this->x()  = p2.x();
  this->y()  = p2.y();
  this->th() = p2.th();
}

/*
 *  \brief operator + : sumation
 *  \param p2 : the second vector to add
 *  \return the sum of both positions
 */
PositionXYTh PositionXYTh::operator+(const PositionXYTh& p2) const
{
  PositionXYTh pres;
  pres.x()  = this->x() + p2.x();
  pres.y()  = this->y() + p2.y();
  pres.th() = this->th() + p2.th();
  return pres;
}

/*
 *  \brief operator + : substraction
 *  \param p2 : the second position to substract
 *  \return the substraction of both positions
 */
PositionXYTh PositionXYTh::operator-(const PositionXYTh& p2) const
{
  PositionXYTh pres;
  pres.x()  = this->x() - p2.x();
  pres.y()  = this->y() - p2.y();
  pres.th() = this->th() - p2.th();
  return pres;
}

/*
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current position and the scalar
 */
PositionXYTh PositionXYTh::operator*(double d) const
{
  PositionXYTh pres;
  pres.x()  = this->x() * d;
  pres.y()  = this->y() * d;
  pres.th() = this->th() * d;
  return pres;
}

/*
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
PositionXYTh PositionXYTh::operator/(double d) const
{
  PositionXYTh pres;
  if(d != 0)
    {
      pres.x()  = this->x() / d;
      pres.y()  = this->y() / d;
      pres.th() = this->th() / d;
    }
  return pres;
}


/***************** Math functions *****************/

/*
 *  \brief compute the normalized position vector from the current one
 *  \return the normalized position vector
 */
PositionXYTh PositionXYTh::normalized() const
{
  double norm = this->norm();
  
  if(norm != 0)
    {
      return ((*this) / norm);
    }
  else return PositionXYTh();
}

 
/*************** Direction functions **************/
  
/*
 *  \brief return the cardinal direction of a position p from the current position
 *  \param p : the position to evaluate
 *  \return cardinal direction of a position p from the current position
 */
cd::carddir_t PositionXYTh::cardDir(const PositionXY& p) const
{
  PositionXYTh P(p.x(),p.y(),0.0);

  // compute the coordinates of P in the frame defined by the current position
  PositionXYTh PinF = coordInFrame(P,*this);

  // compute the polar coordinates of P in F
  double angle = atan2(PinF.y(),PinF.x());

  if(angle >= -0.875*M_PI && angle < -0.625*M_PI) return cd::SW;
  else if(angle >= -0.625*M_PI && angle < -0.375*M_PI) return cd::S;
  else if(angle >= -0.375*M_PI && angle < -0.125*M_PI) return cd::SE;
  else if(angle >= -0.125*M_PI && angle <  0.125*M_PI) return cd::E;
  else if(angle >=  0.125*M_PI && angle <  0.375*M_PI) return cd::NE;
  else if(angle >=  0.375*M_PI && angle <  0.625*M_PI) return cd::N;
  else if(angle >=  0.625*M_PI && angle <  0.875*M_PI) return cd::NW;
  else return cd::W;
}

/*
 *  \brief return true if the position p is in the front of the current position
 *  \param p : the position to evaluate
 *  \return true if the position p is in the front of the current position
 */
bool PositionXYTh::isInTheFront(const PositionXY& p) const
{
  PositionXYTh P(p.x(),p.y(),0.0);

  // compute the coordinates of P in the frame defined by the current position
  PositionXYTh PinF = coordInFrame(P,*this);

  // compute the polar coordinates of P in F
  double angle = atan2(PinF.y(),PinF.x());

  if(angle >= 0.0 && angle < M_PI) return true;
  else return false;
}

/*
 *  \brief return true if the position p is in the back of the current position
 *  \param p : the position to evaluate
 *  \return true if the position p is in the back of the current position
 */
bool PositionXYTh::isInTheBack(const PositionXY& p) const
{
  return !isInTheFront(p);
}

/*
 *  \brief return true if the position p is on the left of the current position
 *  \param p : the position to evaluate
 *  \return true if the position p is on the left of the current position
 */
bool PositionXYTh::isOnTheLeft(const PositionXY& p) const
{
  PositionXYTh P(p.x(),p.y(),0.0);

  // compute the coordinates of P in the frame defined by the current position
  PositionXYTh PinF = coordInFrame(P,*this);

  // compute the polar coordinates of P in F
  double angle = atan2(PinF.y(),PinF.x());

  if(angle >= -0.5*M_PI && angle < 0.5*M_PI) return true;
  else return false;
}

/*
 *  \brief return true if the position p is on the right of the current position
 *  \param p : the position to evaluate
 *  \return true if the position p is on the right of the current position
 */
bool PositionXYTh::isOnTheRight(const PositionXY& p) const
{
  return !isOnTheLeft(p);
}


                                    /**************************************************/
                                    /**************** Class PositionXYZ ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
PositionXYZ::PositionXYZ()
{
  this->coord_ = vector<double>(3);
}

/*
 *  \brief copy constructor
 *  \param p : the position to copy
 */
PositionXYZ::PositionXYZ(const PositionXYZ& p) : PositionXY(p)
{
}

/*
 *  \brief init constructor
 *  \param x : the first position coord
 *  \param y : the second position coord
 *  \param z : the third position coord
 */
PositionXYZ::PositionXYZ(double x,double y,double z)
{
  this->coord_ = vector<double>(3);
  this->x()  = x;
  this->y()  = y;
  this->z()  = z;
}

/*
 *  \brief destructor
 */
PositionXYZ::~PositionXYZ()
{
}


/******************* Operators ********************/

/*
 *  \brief operator = : assignation
 *  \param p2 : the position to assign to the current one
 */
void PositionXYZ::operator=(const PositionXYZ& p2)
{
  this->x() = p2.x();
  this->y() = p2.y();
  this->z() = p2.z();
}

/*
 *  \brief operator + : sumation
 *  \param p2 : the second vector to add
 *  \return the sum of both positions
 */
PositionXYZ PositionXYZ::operator+(const PositionXYZ& p2) const
{
  PositionXYZ pres;
  pres.x() = this->x() + p2.x();
  pres.y() = this->y() + p2.y();
  pres.z() = this->z() + p2.z();
  return pres;
}

/*
 *  \brief operator + : substraction
 *  \param p2 : the second position to substract
 *  \return the substraction of both positions
 */
PositionXYZ PositionXYZ::operator-(const PositionXYZ& p2) const
{
  PositionXYZ pres;
  pres.x() = this->x() - p2.x();
  pres.y() = this->y() - p2.y();
  pres.z() = this->z() - p2.z();
  return pres;
}

/*
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current position and the scalar
 */
PositionXYZ PositionXYZ::operator*(double d) const
{
  PositionXYZ pres;
  pres.x() = this->x() * d;
  pres.y() = this->y() * d;
  pres.z() = this->z() * d;
  return pres;
}

/*
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
PositionXYZ PositionXYZ::operator/(double d) const
{
  PositionXYZ pres;
  if(d != 0)
    {
      pres.x() = this->x() / d;
      pres.y() = this->y() / d;
      pres.z() = this->z() / d;
    }
  return pres;
}

/*
 *  \brief operator ^ : cross product
 *  \param p2 : the 3D position vector to do the cross product with
 *  \return cross product of both position vectors
 */  
PositionXYZ PositionXYZ::operator^(const PositionXYZ& p2) const
{
  return PositionXYZ(this->at(1)*p2.at(2) - this->at(2)*p2.at(1),
		     this->at(2)*p2.at(0) - this->at(0)*p2.at(2),
		     this->at(0)*p2.at(1) - this->at(1)*p2.at(0));

}


/*************** Distance functions ***************/

/*
 *  \brief distance computation between two positions
 *  \param p : the 2nd position to compute the distance to
 *  \return the distance between the current position and p
 */
double PositionXYZ::dist(const PositionXYZ& p) const
{
  double dx = this->x() - p.x();
  double dy = this->y() - p.y();
  double dz = this->z() - p.z();
  return sqrt(dx*dx + dy*dy + dz*dz);
}

/*
 *  \brief square distance computation between two positions
 *  \param p : the 2nd position to compute the square distance to
 *  \return the square distance between the current position and p
 */
double PositionXYZ::sqDist(const PositionXYZ& p) const
{
  double dx = this->x() - p.x();
  double dy = this->y() - p.y();
  double dz = this->z() - p.z();
  return (dx*dx + dy*dy + dz*dz);
}


/***************** Math functions *****************/

/*
 *  \brief compute the normalized position vector from the current one
 *  \return the normalized position vector
 */
PositionXYZ PositionXYZ::normalized() const
{
  double norm = this->norm();
  
  if(norm != 0)
    {
      return ((*this) / norm);
    }
  else return PositionXYZ();
}


                                    /**************************************************/
                                    /**************** Class Position6D ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Position6D::Position6D()
{
  this->coord_ = vector<double>(6);
}

/**
 *  \brief init constructor
 */
Position6D::Position6D(double x,double y,double z,double yaw,double pitch,double roll)
{
  this->coord_ = vector<double>(6);
  this->x()  = x;
  this->y()  = y;
  this->z()  = z;
  this->yaw()  = yaw;
  this->pitch()  = pitch;
  this->roll()  = roll;
}

/**
 *  \brief copy constructor
 *  \param p : the position to copy
 */
Position6D::Position6D(const Position6D& p) : PositionXYZ(p)
{
}

/**
 *  \brief destructor
 */
Position6D::~Position6D()
{
}


                                    /**************************************************/
                                    /**************** Class Position6D ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
ManaPose::ManaPose() : Position6D()
{
}

/**
 *  \brief init constructor
 */
ManaPose::ManaPose(double x,double y,double z,double yaw,double pitch,double roll) : Position6D(x,y,z,yaw,pitch,roll)
{
}

/**
 *  \brief copy constructor
 *  \param p : the position to copy
 */
ManaPose::ManaPose(const ManaPose& p) : Position6D(p)
{
}

/**
 *  \brief destructor
 */
ManaPose::~ManaPose()
{
}


                                    /**************************************************/
                                    /************* Intersections functions ************/
                                    /**************************************************/

/*
 *  \brief compute the intersection between a straight line and a circle
 *  \param posL : starting point of the straight line
 *  \param vecL : vector defining the straight line
 *  \param posC : center of the circle
 *  \param radC : radius of the circle
 *  \param itsc1 : first intersection found
 *  \param itsc2 : second intersection found
 *  \return the number of intersections found
 */
int mpc::intersecLineCircle(const PositionXY& posL,const PositionXY& vecL,const PositionXY& posC,double radC,PositionXY& itsc1,PositionXY& itsc2)
{
  // compute the vector PLPC, its direction and norm
  PositionXY posLposC = posC - posL;
  double dirPLPC = atan2(posLposC.y(),posLposC.x());
  double distPLPC = posLposC.norm();

  // compute the difference of angle between dirVecL and dirPLPC
  double dirVecL = atan2(vecL.y(),vecL.x());
  double diffAng = normalizedAngle(dirPLPC - dirVecL);

  // compute the projection of posC on the straight line defined by posL and vecL
  double distPLProj = cos(diffAng) * distPLPC;
  PositionXY nvecL = vecL.normalized();
  PositionXY proj = posL + (nvecL * distPLProj);

  // compute the distance between proj and posC
  PositionXY posCproj= proj - posC;
  double distPCProj = posCproj.norm();

  if(distPCProj > radC) // no intersection
    {
      return 0;
    }
  else if(distPCProj == radC) // one intersection
    {
      itsc1 = proj;
      return 1;
    }
  else // two intersections
    {
      // compute the distance between the projection and each intersection
      double distProjItsc = sqrt(radC*radC - distPCProj*distPCProj);
      itsc1 = proj - (nvecL * distProjItsc);
      itsc2 = proj + (nvecL * distProjItsc);
      return 2;
    }
}

/*
 *  \brief compute the intersection between a straight line and a circle
 *  \param posL : starting point of the straight line
 *  \param vecL : vector defining the straight line
 *  \param posC : center of the circle
 *  \param radC : radius of the circle
 *  \param itsc1 : first intersection found
 *  \param itsc2 : second intersection found
 *  \param dist1 : distance between posL and the first intersection (negative if the intersection is behind the obstacle)
 *  \param dist2 : distance between posL and the second intersection (negative if the intersection is behind the obstacle)
 *  \return the number of intersections found
 */
int mpc::intersecLineCircle(const PositionXY& posL,const PositionXY& vecL,const PositionXY& posC,double radC,
		       PositionXY& itsc1,PositionXY& itsc2,double& dist1,double& dist2)
{
  // compute the vector PLPC, its direction and norm
  PositionXY posLposC = posC - posL;
  double dirPLPC = atan2(posLposC.y(),posLposC.x());
  double distPLPC = posLposC.norm();

  // compute the difference of angle between dirVecL and dirPLPC
  double dirVecL = atan2(vecL.y(),vecL.x());
  double diffAng = normalizedAngle(dirPLPC - dirVecL);

  // compute the projection of posC on the straight line defined by posL and vecL
  double distPLProj = cos(diffAng) * distPLPC;
  PositionXY nvecL = vecL.normalized();
  PositionXY proj = posL + (nvecL * distPLProj);

  // compute the distance between proj and posC
  PositionXY posCproj= proj - posC;
  double distPCProj = posCproj.norm();

  if(distPCProj > radC) // no intersection
    {
      return 0;
    }
  else if(distPCProj == radC) // one intersection
    {
      itsc1 = proj;
      dist1 = 0.0;
      itsc2 = proj;
      dist2 = 0.0;
      return 1;
    }
  else // two intersections
    {
      // compute the distance between the projection and each intersection
      double distProjItsc = sqrt(radC*radC - distPCProj*distPCProj);
      itsc1 = proj - (nvecL * distProjItsc);
      dist1 = distPLProj - distProjItsc;
      itsc2 = proj + (nvecL * distProjItsc);
      dist2 = distPLProj + distProjItsc;
      return 2;
    }
}

/*
 *  \brief compute the intersection between a segment and a circle
 *  \param posL : starting point of the straight line
 *  \param vecL : vector defining the straight line
 *  \param posC : center of the circle
 *  \param radC : radius of the circle
 *  \param itsc1 : first intersection found
 *  \param itsc2 : second intersection found
 *  \return the number of intersections found
 *
 *  Warning : This function won't find any intersection if the segment is completely included in the circle
 */
int mpc::intersecSgmtCircle(const PositionXY& p1,const PositionXY& p2,const PositionXY& posC,double radC,PositionXY& itsc1,PositionXY& itsc2)
{
  // compute the vector p1p2
  PositionXY p1p2(p2.x()-p1.x(),p2.y()-p1.y());
  PositionXY itscL1;
  PositionXY itscL2;

  int nbItsc = mpc::intersecLineCircle(p1,p1p2,posC,radC,itscL1,itscL2);
  if(nbItsc == 0)
    {
      return 0;
    }
  else if(nbItsc == 1)
    {
      // compute the distance P1P2 (square)
      double dxP1P2 = p2.x() - p1.x();
      double dyP1P2 = p2.y() - p1.y();
      double sdDistP1P2 = (dxP1P2*dxP1P2 + dyP1P2*dyP1P2);

      // compute the distance P1pres (square)
      double dxP1itsc = itscL1.x() - p1.x();
      double dyP1itsc = itscL1.y() - p1.y();
      double sdDistP1itsc = (dxP1itsc*dxP1itsc + dyP1itsc*dyP1itsc);

      // compute the distance P2pres (square)
      double dxP2itsc = itscL1.x() - p2.x();
      double dyP2itsc = itscL1.y() - p2.y();
      double sdDistP2itsc = (dxP2itsc*dxP2itsc + dyP2itsc*dyP2itsc);

      if(sdDistP1itsc <= sdDistP1P2 && sdDistP2itsc <= sdDistP1P2)
	{
	  itsc1 = itscL1;
	  return 1;
	}
      else return 0;
    }
  else if(nbItsc == 2)
    {
      // check if one of the end points of the segment is completely included in the circle

      // compute the distance P1PC (square)
      double dxP1PC = posC.x() - p1.x();
      double dyP1PC = posC.y() - p1.y();
      double sdDistP1PC = (dxP1PC*dxP1PC + dyP1PC*dyP1PC);
      bool P1inside = (sdDistP1PC < (radC*radC));
      
      // compute the distance P2PC (square)
      double dxP2PC = posC.x() - p2.x();
      double dyP2PC = posC.y() - p2.y();
      double sdDistP2PC = (dxP2PC*dxP2PC + dyP2PC*dyP2PC);
      bool P2inside = (sdDistP2PC < (radC*radC));

      // check if both end points are included in the circle
      if(P1inside && P2inside) return 0;
      
      // check if both end points are outside of the circle
      if(!P1inside && !P2inside)
	{
	  // compute the distance P1itsc1
	  double dxP1I1 = itscL1.x() - p1.x();
	  double dyP1I1 = itscL1.y() - p1.y();
	  double sdDistP1I1 = (dxP1I1*dxP1I1 + dyP1I1*dyP1I1);
	  // compute the distance P1itsc2
	  double dxP1I2 = itscL2.x() - p1.x();
	  double dyP1I2 = itscL2.y() - p1.y();
	  double sdDistP1I2 = (dxP1I2*dxP1I2 + dyP1I2*dyP1I2);
	  // compute the distance P2itsc1
	  double dxP2I1 = itscL1.x() - p2.x();
	  double dyP2I1 = itscL1.y() - p2.y();
	  double sdDistP2I1 = (dxP2I1*dxP2I1 + dyP2I1*dyP2I1);
	  // compute the distance P2itsc2
	  double dxP2I2 = itscL2.x() - p2.x();
	  double dyP2I2 = itscL2.y() - p2.y();
	  double sdDistP2I2 = (dxP2I2*dxP2I2 + dyP2I2*dyP2I2);

	  // check if both P1 and P2 are on the same side of the circle
	  if((sdDistP1I1 <= sdDistP1I2 && sdDistP2I1 <= sdDistP2I2) || (sdDistP1I1 > sdDistP1I2 && sdDistP2I1 > sdDistP2I2))
	    {
	      // both P1 and P2 are on the same side of the circle
	      //   -> no intersection
	      return 0;
	      
	    }
	  else
	    {
	      // P1 and P2 are on each side of the circle
	      //   -> 2 intersections
	      itsc1 = itscL1;
	      itsc2 = itscL2;
	      return 2;
	    }
	}
      else if(P1inside)
	{
	  // compute the distance P2itsc1
	  double dxP2I1 = itscL1.x() - p2.x();
	  double dyP2I1 = itscL1.y() - p2.y();
	  double sdDistP2I1 = (dxP2I1*dxP2I1 + dyP2I1*dyP2I1);
	  // compute the distance P2itsc2
	  double dxP2I2 = itscL2.x() - p2.x();
	  double dyP2I2 = itscL2.y() - p2.y();
	  double sdDistP2I2 = (dxP2I2*dxP2I2 + dyP2I2*dyP2I2);

	  if(sdDistP2I1 <= sdDistP2I2) itsc1 = itscL1;
	  else itsc1 = itscL2;
	  return 1;
	}
      else // if(P2inside)
	{
	  // compute the distance P1itsc1
	  double dxP1I1 = itscL1.x() - p1.x();
	  double dyP1I1 = itscL1.y() - p1.y();
	  double sdDistP1I1 = (dxP1I1*dxP1I1 + dyP1I1*dyP1I1);
	  // compute the distance P1itsc2
	  double dxP1I2 = itscL2.x() - p1.x();
	  double dyP1I2 = itscL2.y() - p1.y();
	  double sdDistP1I2 = (dxP1I2*dxP1I2 + dyP1I2*dyP1I2);

	  if(sdDistP1I1 <= sdDistP1I2) itsc1 = itscL1;
	  else itsc1 = itscL2;
	  return 1;
	}
    }
  else
    {
      cerr << "mpc::intersecSgmtCircle : Should never arrive here..." << endl;
      return -1;
    }
}

/*
 * \brief compute the coefficients (a,b,c) of the equation of a line defined by a point p and a vector v
 *
 * !!!!!! I don't understand this function anymore!!!!
 */
void coefEq(PositionXY p,PositionXY v,double& a,double& b,double& c)
{
  a = v.y();
  b = -v.x();
  c = v.x() * p.y() - v.y() * p.x();
}


/*
 *  \brief compute the intersection between a straight line and a circle
 *  \param pos1 : first end position of the first segment
 *  \param vec1 : second end position of the first segment
 *  \param pos2 : first end position of the second segment
 *  \param vec2 : second end position of the second segment
 *  \param itsc : the intersection found (if exists)
 *  \return true if an intersection has been found, false else (both lines are colinear)
 */
bool mpc::intersecLines(const PositionXY& pos1,const PositionXY& vec1,const PositionXY& pos2,const PositionXY& vec2,PositionXY& itsc)
{
  if(vec1.x()*vec2.y() - vec1.y()*vec2.x() != 0)
    {
      double a1,b1,c1;
      double a2,b2,c2;
      coefEq(pos1,vec1,a1,b1,c1);
      coefEq(pos2,vec2,a2,b2,c2);
		
      // defining the result
      itsc.x() = -(c1*b2 - c2*b1)/(a1*b2 - a2*b1);
      itsc.y() = -(c1*a2 - c2*a1)/(b1*a2 - b2*a1);
      return true;
    }
  else
    {
      return false;
    }
}

/*
 *  \brief compute the intersection between a straight line and a circle
 *  \param p1 : first end position of the first segment
 *  \param p2 : second end position of the first segment
 *  \param p3 : first end position of the second segment
 *  \param p4 : second end position of the second segment
 *  \param itsc : the intersection found (if exists)
 *  \return true if an intersection has been found, false else (both segments are colinear)
 */
bool mpc::intersecSgmts(const PositionXY& p1,const PositionXY& p2,const PositionXY& p3,const PositionXY& p4,PositionXY& itsc)
{
  // compute the vector p1p2
  PositionXY vecP1P2(p2.x() - p1.x(),
		     p2.y() - p1.y());

  // compute the vector p3p4
  PositionXY vecP3P4(p4.x() - p3.x(),
		     p4.y() - p3.y());

  bool noncolinear = intersecLines(p1, vecP1P2, p3, vecP3P4, itsc);

  if(!noncolinear) return false;
  else
    {
      // compute the distance P1P2 (square)
      double dxP1P2 = p2.x() - p1.x();
      double dyP1P2 = p2.y() - p1.y();
      double sdDistP1P2 = (dxP1P2*dxP1P2 + dyP1P2*dyP1P2);

      // compute the distance P1pres (square)
      double dxP1itsc = itsc.x() - p1.x();
      double dyP1itsc = itsc.y() - p1.y();
      double sdDistP1itsc = (dxP1itsc*dxP1itsc + dyP1itsc*dyP1itsc);

      // compute the distance P2pres (square)
      double dxP2itsc = itsc.x() - p2.x();
      double dyP2itsc = itsc.y() - p2.y();
      double sdDistP2itsc = (dxP2itsc*dxP2itsc + dyP2itsc*dyP2itsc);
      
      // compute the distance P3P4 (square)
      double dxP3P4 = p4.x() - p3.x();
      double dyP3P4 = p4.y() - p3.y();
      double sdDistP3P4 = (dxP3P4*dxP3P4 + dyP3P4*dyP3P4);

      // compute the distance P3pres (square)
      double dxP3itsc = itsc.x() - p3.x();
      double dyP3itsc = itsc.y() - p3.y();
      double sdDistP3itsc = (dxP3itsc*dxP3itsc + dyP3itsc*dyP3itsc);

      // compute the distance P4pres (square)
      double dxP4itsc = itsc.x() - p4.x();
      double dyP4itsc = itsc.y() - p4.y();
      double sdDistP4itsc = (dxP4itsc*dxP4itsc + dyP4itsc*dyP4itsc);

      // check if intersec in on the segment P1P2
      if(sdDistP1itsc <= sdDistP1P2 && sdDistP2itsc <= sdDistP1P2 && sdDistP3itsc <= sdDistP3P4 && sdDistP4itsc <= sdDistP3P4)
	{
	  return true;
	}
      else return false;
    }
}


                                    /**************************************************/
                                    /************** Projections functions *************/
                                    /**************************************************/

/**
 *  \brief orthogonal projection of a 2D point on a line defined by a point and a vector
 *  \param pos : the position to project
 *  \param pline : the position defining the line
 *  \param vline : the vector defining the line
 *  \return the orthogonal projection of pos on the line
 */
PositionXY mpc::orthoProjLine(const PositionXY& pos,const PositionXY& pline,const PositionXY& vline)
{
  // compute a perpendicular vector to vline
  PositionXY vec(vline.y(),-vline.x());

  // compute the intersection between the lines defined vy (pos,vec) and (pline,vline)
  PositionXY itsc;
  bool noncolinear = mpc::intersecLines(pos,vec,pline,vline,itsc);
  if(!noncolinear)
    {
//       itsc.x() = pos.x();
//       itsc.y() = pos.y();
      cerr << "mpc::orthoProjLine : vec and vline colinear??? -> pos : (" << pos.x() << "," << pos.y() << ")  pline : (" << pline.x() << "," << pline.y() << "  vline : (" << vline.x() << "," << vline.y() << ")" << endl;
    }
  return itsc;
}

/**
 *  \brief orthogonal projection of a 2D point on a segment defined by its end points
 *  \param pos : the position to project
 *  \param p1 : the first end point of the segment
 *  \param p2 : the second end point of the segment
 *  \return the orthogonal projection of pos on the segment
 */
PositionXY mpc::orthoProjSgmt(const PositionXY& pos,const PositionXY& p1,const PositionXY& p2)
{
  // compute the vector p1p2
  PositionXY p1p2(p2.x()-p1.x(),
		  p2.y()-p1.y());

  // compute the projection of pos on the line defined by (p1,p1p2)
  return mpc::orthoProjLine(pos,p1,p1p2);
}


                                    /**************************************************/
                                    /*************** Distance functions ***************/
                                    /**************************************************/

/**
 *  \brief compute the closest distance from a point to a line
 *  \param pos : the position to evaluate
 *  \param pline : the position defining the line
 *  \param vline : the vector defining the line
 *  \return the closest distance from a point to a line
 */
double mpc::closeDistLine(const PositionXY& pos,const PositionXY& pline,const PositionXY& vline)
{
  // compute the orthogonal projection of pos on the line
  PositionXY itsc = orthoProjLine(pos,pline,vline);

  // return the distance from pos to itsc
  double dx = itsc.x() - pos.x();
  double dy = itsc.y() - pos.y();
  return sqrt(dx*dx + dy*dy);
}

/**
 *  \brief compute the closest distance from a point to a segment
 *  \param pos : the position to evaluate
 *  \param p1 : the first end point of the segment
 *  \param p2 : the second end point of the segment
 *  \return the closest distance from a point to a line
 */
double mpc::closeDistSgmt(const PositionXY& pos,const PositionXY& p1,const PositionXY& p2)
{
  // compute the orthogonal projection of pos on the segment
  PositionXY itsc = orthoProjSgmt(pos,p1,p2);
   
  // compute the distance P1P2 (square)
  double dxP1P2 = p2.x() - p1.x();
  double dyP1P2 = p2.y() - p1.y();
  double sdDistP1P2 = (dxP1P2*dxP1P2 + dyP1P2*dyP1P2);
  
  // compute the distance P1itsc (square)
  double dxP1itsc = itsc.x() - p1.x();
  double dyP1itsc = itsc.y() - p1.y();
  double sdDistP1itsc = (dxP1itsc*dxP1itsc + dyP1itsc*dyP1itsc);
  
  // compute the distance P2itsc (square)
  double dxP2itsc = itsc.x() - p2.x();
  double dyP2itsc = itsc.y() - p2.y();
  double sdDistP2itsc = (dxP2itsc*dxP2itsc + dyP2itsc*dyP2itsc);

  // check if itsc belongs to the segment
  if(sdDistP1itsc <= sdDistP1P2 && sdDistP2itsc <= sdDistP1P2)
    {
      // itsc belongs to the segment : 
      // return the distance from pos to itsc
      double dx = pos.x() - itsc.x();
      double dy = pos.y() - itsc.y();
      return sqrt(dx*dx + dy*dy);
    }
  else if(sdDistP1itsc <= sdDistP2itsc)
    {
      // itsc does not belongs to the segment and is closer from p1 than p2
      // return the distance from pos to p1
      double dx = pos.x() - p1.x();
      double dy = pos.y() - p1.y();
      return sqrt(dx*dx + dy*dy);
    }
  else
    {
      // itsc does not belongs to the segment and is closer from p2 than p1
      // return the distance from pos to p2
      double dx = pos.x() - p2.x();
      double dy = pos.y() - p2.y();
      return sqrt(dx*dx + dy*dy);
    }
}

/**
 *  \brief compute the closest distance from a point to a segment
 *  \param pos : the position to evaluate
 *  \param p1 : the first end point of the segment
 *  \param p2 : the second end point of the segment
 *  \param minDist : the closest distance from a point to a line
 *  \return the closest position of segment p1-p2 to position pos
 */
PositionXY mpc::closePosFromSgmt(const PositionXY& pos,const PositionXY& p1,const PositionXY& p2, double& minDist)
{
  // compute the orthogonal projection of pos on the segment
  PositionXY itsc = orthoProjSgmt(pos,p1,p2);
   
  // compute the distance P1P2 (square)
  double dxP1P2 = p2.x() - p1.x();
  double dyP1P2 = p2.y() - p1.y();
  double sdDistP1P2 = (dxP1P2*dxP1P2 + dyP1P2*dyP1P2);
  
  // compute the distance P1itsc (square)
  double dxP1itsc = itsc.x() - p1.x();
  double dyP1itsc = itsc.y() - p1.y();
  double sdDistP1itsc = (dxP1itsc*dxP1itsc + dyP1itsc*dyP1itsc);
  
  // compute the distance P2itsc (square)
  double dxP2itsc = itsc.x() - p2.x();
  double dyP2itsc = itsc.y() - p2.y();
  double sdDistP2itsc = (dxP2itsc*dxP2itsc + dyP2itsc*dyP2itsc);

  // check if itsc belongs to the segment
  if(sdDistP1itsc <= sdDistP1P2 && sdDistP2itsc <= sdDistP1P2)
    {
      // itsc belongs to the segment : 
      // return the distance from pos to itsc
      double dx = pos.x() - itsc.x();
      double dy = pos.y() - itsc.y();
      minDist = sqrt(dx*dx + dy*dy);

      return itsc;
    }
  else if(sdDistP1itsc <= sdDistP2itsc)
    {
      // itsc does not belongs to the segment and is closer from p1 than p2
      // return the distance from pos to p1
      double dx = pos.x() - p1.x();
      double dy = pos.y() - p1.y();
      minDist = sqrt(dx*dx + dy*dy);

      return p1;
    }
  else
    {
      // itsc does not belongs to the segment and is closer from p2 than p1
      // return the distance from pos to p2
      double dx = pos.x() - p2.x();
      double dy = pos.y() - p2.y();
      minDist = sqrt(dx*dx + dy*dy);

      return p2;
    }
}


                                    /**************************************************/
                                    /************* Frame changes functions ************/
                                    /**************************************************/

/*
 *  \brief frame change function
 *  \param p : coord of the position in frame (0;0)
 *  \param f : coord of the new frame
 *  \return the coordinates of position p in the new frame f from the frame (0;0)
 */  
PositionXYTh mpc::coordInFrame(const PositionXYTh& p,const PositionXYTh& f)
{
  // compute the vector OP where O is the origin of the frame F
  PositionXYTh vecOP = p - f;
//   cout << "vecOP : "; vecOP.display(); cout << endl;

  if(vecOP.x() == 0 && vecOP.y() == 0)
    {
      return PositionXYTh(0.0,0.0,normalizedAngle(p.th() - f.th()));
    }

  // compute the direction of vector OP in the current frame
  double dirOP = atan2(vecOP.y(),vecOP.x());

  // compute the direction of vector OP in the frame F
  double dirOPinF = normalizedAngle(dirOP - f.th());

  // compute the distance between O and P
  double distOP = f.dist(p);

//   cout << "dirOP : " << dirOP << endl;
//   cout << "dirOPinF : " << dirOPinF << endl;
//   cout << "distOP : " << distOP << endl;

  // compute the coordinate of the point P in the frame F
  return PositionXYTh(distOP * cos(dirOPinF),
		      distOP * sin(dirOPinF),
		      normalizedAngle(p.th() - f.th()));
}

/*
 *  \brief frame change function
 *  \param p : coord of the position in frame f
 *  \param f : coord of the old frame
 *  \return the coordinates of position p in the new frame (0;0) from the frame f
 */  
PositionXYTh mpc::coordFromFrame(const PositionXYTh& p,const PositionXYTh& f)
{
  // compute the distance between (0,0) and the origin O of F
  double dx = f.x();
  double dy = f.y();
  double distO0 = sqrt(dx*dx+dy*dy);
  
  if(distO0 == 0)
    {
//       return PositionXYTh(0.0,0.0,p.th() + f.th());
      return coordInFrame(p,PositionXYTh(0.0,0.0,-f.th()));
    }

  // compute the direction of the vector between (0,0) and O
  double dirO0 = atan2(-dy,-dx);

  // compute the direction of (0,0) from F
  double dir0fromF = normalizedAngle(dirO0 - f.th());

  // compute the coord of the frame (0,0) from frame F
  PositionXYTh coord00fromF(distO0 * cos(dir0fromF),
			    distO0 * sin(dir0fromF),
			    -f.th());

  // return the coordinate of the point P in the frame (0.0)
  return coordInFrame(p, coord00fromF);
}

/**
 *  \brief homothetic transformation
 *  \param p : coord of the input position
 *  \param hCenter : center of the homothetic transformation
 *  \param hCoef : coefficients of the homothetic transformation
 *  \return the homothetic transformation of the position p
 */  
PositionXY mpc::homothety(const PositionXY& p,const PositionXY& hCenter,const PositionXY& hCoef)
{
  PositionXY vec(p.x() - hCenter.x(),
		 p.y() - hCenter.y());

  return PositionXY(hCenter.x() + hCoef.x()*vec.x(),
		    hCenter.y() + hCoef.y()*vec.y());
}
