#include "StateTime.hpp"
using namespace mpc;

#include "Utils.hpp"

#include <GL/gl.h>
#include <GL/glu.h>

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /***************** Class StateTime ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
StateTime::StateTime() : State()
{
}

/**
 *  \brief copy constructor
 *  \param s : the state-time to copy
 */
StateTime::StateTime(const StateTime& s) : State(s)
{
}

/**
 *  \brief init constructor
 *  \param n : dimension of the state-time
 */
StateTime::StateTime(int n) : State(n)
{
}

/**
 *  \brief destructor
 */
StateTime::~StateTime()
{
}


/*************** Display functions ****************/

/**
 *  \brief display the state-time in an openGL viewer
 */
void StateTime::toGL() const
{
  this->toGL(1.,0.,0.,0.25);
}

/**
 *  \brief display the state-time in an openGL viewer
 *  \param R : the red color
 *  \param G : the green color
 *  \param B : the blue color
 *  \param size : the size of the object displayed   
 */
void StateTime::toGL(double R,double G,double B,double size) const
{
  PositionXYT pos = this->posXYT();
  
  const GLfloat material[]={R,G,B,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, material);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, material);

  glPushMatrix();
  glTranslated(pos.x(),pos.y(),pos.t());
  
  gluDisk(gluNewQuadric(), 0, size, 20, 1);
  gluCylinder(gluNewQuadric(), size, size, 0.1*size, 20, 5);
  glTranslated(0,0,0.02);
  gluDisk(gluNewQuadric(), 0, size, 20, 1);
  
  glPopMatrix();
}


                                    /**************************************************/
                                    /***************** Class PMStateT *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PMStateT::PMStateT() : StateTime(5)
{
}

/**
 *  \brief copy constructor
 *  \param p : the state-time to copy
 */
PMStateT::PMStateT(const PMStateT& s) : StateTime(5)
{
  this->x()  = s.x();
  this->y()  = s.y();
  this->sx() = s.sx();
  this->sy() = s.sy();
  this->t()  = s.t();
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param sx : speed sx
 *  \param sy : speed s
 *  \param t : time t
 */
PMStateT::PMStateT(double x,double y,double sx,double sy,double t) : StateTime(5)
{
  this->x()  = x;
  this->y()  = y;
  this->sx() = sx;
  this->sy() = sy;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
PMStateT::~PMStateT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state
 *  \param s2 : the state to copy
 */
void PMStateT::operator=(const PMStateT& s2)
{
  this->x()  = s2.x();
  this->y()  = s2.y();
  this->sx() = s2.sx();
  this->sy() = s2.sy();
  this->t()  = s2.t();
}

/**
 *  \brief operator += : add the input state
 *  \param s2 : the state to add
 */
void PMStateT::operator+=(const PMStateT& s2)
{
  this->x()  += s2.x();
  this->y()  += s2.y();
  this->sx() += s2.sx();
  this->sy() += s2.sy();
  this->t()  += s2.t();
}

/**
 *  \brief operator -= : substract the input state
 *  \param s2 : the state to substract
 */
void PMStateT::operator-=(const PMStateT& s2)
{
  this->x()  -= s2.x();
  this->y()  -= s2.y();
  this->sx() -= s2.sx();
  this->sy() -= s2.sy();
  this->t()  -= s2.t();
}

/**
 *  \brief operator += : multiply with the input coefficient
 *  \param d : the coefficient to mutilpy with
 */
void PMStateT::operator*=(double d)
{
  this->x()  *= d;
  this->y()  *= d;
  this->sx() *= d;
  this->sy() *= d;
  this->t()  *= d;
}

/**
 *  \brief operator /= : divide with the input coefficient
 *  \param d : the coefficient to divide with
 */
void PMStateT::operator/=(double d)
{
  this->x()  /= d;
  this->y()  /= d;
  this->sx() /= d;
  this->sy() /= d;
  this->t()  /= d;
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
PMStateT PMStateT::operator+(const PMStateT& s2) const
{
  return PMStateT(this->x()  + s2.x(),
		  this->y()  + s2.y(),
		  this->sx() + s2.sx(),
		  this->sy() + s2.sy(),
		  this->t()  + s2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
PMStateT PMStateT::operator-(const PMStateT& s2) const
{
  return PMStateT(this->x()  - s2.x(),
		  this->y()  - s2.y(),
		  this->sx() - s2.sx(),
		  this->sy() - s2.sy(),
		  this->t()  - s2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
PMStateT PMStateT::operator*(double d) const
{
  return PMStateT(this->x()  * d,
		  this->y()  * d,
		  this->sx() * d,
		  this->sy() * d,
		  this->t()  * d);
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
PMStateT PMStateT::operator/(double d) const
{
  return PMStateT(this->x()  / d,
		  this->y()  / d,
		  this->sx() / d,
		  this->sy() / d,
		  this->t()  / d);
}


/*************** Display functions ****************/

/**
 *  \brief display the vector's coordinates
 */
void PMStateT::display() const
{
  cout << "(" << this->x() << "," << this->y() << "," << this->sx() << "," << 
    this->sy() << "," << this->t() << ")";
}


                                    /**************************************************/
                                    /**************** Class PM3DStateT ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PM3DStateT::PM3DStateT() : StateTime(7)
{
}

/**
 *  \brief copy constructor
 *  \param p : the state-time to copy
 */
PM3DStateT::PM3DStateT(const PM3DStateT& s) : StateTime(7)
{
  this->x()  = s.x();
  this->y()  = s.y();
  this->z()  = s.z();
  this->sx() = s.sx();
  this->sy() = s.sy();
  this->sz() = s.sz();
  this->t()  = s.t();
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param z : position z
 *  \param sx : speed sx
 *  \param sy : speed sy
 *  \param sz : speed sz
 *  \param t : time t
 */
PM3DStateT::PM3DStateT(double x,double y,double z,double sx,double sy,double sz,double t) : StateTime(7)
{
  this->x()  = x;
  this->y()  = y;
  this->z()  = z;
  this->sx() = sx;
  this->sy() = sy;
  this->sz() = sz;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
PM3DStateT::~PM3DStateT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state-time
 *  \param s2 : the state-time to copy
 */
void PM3DStateT::operator=(const PM3DStateT& s2)
{
  this->x()  = s2.x();
  this->y()  = s2.y();
  this->z()  = s2.z();
  this->sx() = s2.sx();
  this->sy() = s2.sy();
  this->sz() = s2.sz();
  this->t()  = s2.t();
}

/**
 *  \brief operator += : add the input state-time
 *  \param s2 : the state-time to add
 */
void PM3DStateT::operator+=(const PM3DStateT& s2)
{
  this->x()  += s2.x();
  this->y()  += s2.y();
  this->z()  += s2.z();
  this->sx() += s2.sx();
  this->sy() += s2.sy();
  this->sz() += s2.sz();
  this->t()  += s2.t();
}

/**
 *  \brief operator -= : substract the input state-time
 *  \param s2 : the state-time to substract
 */
void PM3DStateT::operator-=(const PM3DStateT& s2)
{
  this->x()  -= s2.x();
  this->y()  -= s2.y();
  this->z()  -= s2.z();
  this->sx() -= s2.sx();
  this->sy() -= s2.sy();
  this->sz() -= s2.sz();
  this->t()  -= s2.t();
}

/**
 *  \brief operator += : multiply with the input coefficient
 *  \param d : the coefficient to mutilpy with
 */
void PM3DStateT::operator*=(double d)
{
  this->x()  *= d;
  this->y()  *= d;
  this->z()  *= d;
  this->sx() *= d;
  this->sy() *= d;
  this->sz() *= d;
  this->t()  *= d;
}

/**
 *  \brief operator /= : divide with the input coefficient
 *  \param d : the coefficient to divide with
 */
void PM3DStateT::operator/=(double d)
{
  this->x()  /= d;
  this->y()  /= d;
  this->z()  /= d;
  this->sx() /= d;
  this->sy() /= d;
  this->sz() /= d;
  this->t()  /= d;
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
PM3DStateT PM3DStateT::operator+(const PM3DStateT& s2) const
{
  return PM3DStateT(this->x()  + s2.x(),
		    this->y()  + s2.y(),
		    this->z()  + s2.z(),
		    this->sx() + s2.sx(),
		    this->sy() + s2.sy(),
		    this->sz() + s2.sz(),
		    this->t()  + s2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
PM3DStateT PM3DStateT::operator-(const PM3DStateT& s2) const
{
  return PM3DStateT(this->x()  - s2.x(),
		    this->y()  - s2.y(),
		    this->z()  - s2.z(),
		    this->sx() - s2.sx(),
		    this->sy() - s2.sy(),
		    this->sz() - s2.sz(),
		    this->t()  - s2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
PM3DStateT PM3DStateT::operator*(double d) const
{
  return PM3DStateT(this->x()  * d,
		    this->y()  * d,
		    this->z()  * d,
		    this->sx() * d,
		    this->sy() * d,
		    this->sz() * d,
		    this->t()  * d);
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
PM3DStateT PM3DStateT::operator/(double d) const
{
  return PM3DStateT(this->x()  / d,
		    this->y()  / d,
		    this->z()  / d,
		    this->sx() / d,
		    this->sy() / d,
		    this->sz() / d,
		    this->t()  / d);
}


/*************** Display functions ****************/

/**
 *  \brief display the vector's coordinates
 */
void PM3DStateT::display() const
{
  cout << "(" << this->x() << "," << this->y() << "," << this->z() << "," 
       << this->sx() << "," << this->sy() << "," << this->sz() << "," << this->t() << ")";
}


                                    /**************************************************/
                                    /***************** Class CLStateT *****************/
                                    /**************************************************/

// double CLStateT::maxNoiseXY_ = 2.0;
// double CLStateT::maxNoiseTH_ = 0.03*M_PI;
double CLStateT::maxNoiseXY_ = 300.0;
double CLStateT::maxNoiseTH_ = 0.3*M_PI;
double CLStateT::maxNoiseS_ = 0.0;
double CLStateT::maxNoisePH_ = 0.0;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CLStateT::CLStateT() : StateTime(6)
{
}

/**
 *  \brief copy constructor
 *  \param p : the state-time to copy
 */
CLStateT::CLStateT(const CLStateT& s) : StateTime(6)
{
  this->x()  = s.x();
  this->y()  = s.y();
  this->th() = s.th();
  this->ph() = s.ph();
  this->s()  = s.s();
  this->t()  = s.t();
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param ph : steering angle ph
 *  \param s : linear speed s
 *  \param t : time t 
 */
CLStateT::CLStateT(double x,double y,double th,double ph,double s,double t) : StateTime(6)
{
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
  this->ph() = ph;
  this->s()  = s;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
CLStateT::~CLStateT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state
 *  \param s2 : the state to copy
 */
void CLStateT::operator=(const CLStateT& s2)
{
  this->x()  = s2.x();
  this->y()  = s2.y();
  this->th() = s2.th();
  this->ph() = s2.ph();
  this->s()  = s2.s();
  this->t()  = s2.t();
}

/**
 *  \brief operator += : add the input state
 *  \param s2 : the state to add
 */
void CLStateT::operator+=(const CLStateT& s2)
{
  this->x()  += s2.x();
  this->y()  += s2.y();
  this->th() += s2.th();
  this->ph() += s2.ph();
  this->s()  += s2.s();
  this->t()  += s2.t();
}

/**
 *  \brief operator -= : substract the input state
 *  \param s2 : the state to substract
 */
void CLStateT::operator-=(const CLStateT& s2)
{
  this->x()  -= s2.x();
  this->y()  -= s2.y();
  this->th() -= s2.th();
  this->ph() -= s2.ph();
  this->s()  -= s2.s();
  this->t()  -= s2.t();

}

/**
 *  \brief operator += : multiply with the input coefficient
 *  \param d : the coefficient to mutilpy with
 */
void CLStateT::operator*=(double d)
{
  this->x()  *= d;
  this->y()  *= d;
  this->th() *= d;
  this->ph() *= d;
  this->s()  *= d;
  this->t()  *= d;
}

/**
 *  \brief operator /= : divide with the input coefficient
 *  \param d : the coefficient to divide with
 */
void CLStateT::operator/=(double d)
{
  this->x()  /= d;
  this->y()  /= d;
  this->th() /= d;
  this->ph() /= d;
  this->s()  /= d;
  this->t()  /= d;
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
CLStateT CLStateT::operator+(const CLStateT& s2) const
{
  return CLStateT(this->x()  + s2.x(),
		  this->y()  + s2.y(),
		  normalizedAngle(this->th() + s2.th()),
		  this->ph() + s2.ph(),
		  this->s()  + s2.s(),
		  this->t()  + s2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
CLStateT CLStateT::operator-(const CLStateT& s2) const
{
  return CLStateT(this->x()  - s2.x(),
		  this->y()  - s2.y(),
		  normalizedAngle(this->th() - s2.th()),
		  this->ph() - s2.ph(),
		  this->s()  - s2.s(),
		  this->t()  - s2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
CLStateT CLStateT::operator*(double d) const
{
  return CLStateT(this->x()  * d,
		  this->y()  * d,
		  normalizedAngle(this->th() * d),
		  this->ph() * d,
		  this->s()  * d,
		  this->t()  * d);
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
CLStateT CLStateT::operator/(double d) const
{
  return CLStateT(this->x()  / d,
		  this->y()  / d,
		  normalizedAngle(this->th() / d),
		  this->ph() / d,
		  this->s()  / d,
		  this->t()  / d);
}


/**************** Noise functions *****************/

/**
 *  \brief return a randomly noised state from the current one
 *  \param t : the elapsed time from the previous known state
 *  \return noised state
 */
CLStateT CLStateT::noisedState(double t)
{
  double noiseX = getDoubleRandom(-maxNoiseXY_*t,maxNoiseXY_*t);
  double noiseY = getDoubleRandom(-maxNoiseXY_*t,maxNoiseXY_*t);
  double noiseTH = getDoubleRandom(-maxNoiseTH_*t,maxNoiseTH_*t);
  double noiseS = getDoubleRandom(-maxNoiseS_*t,maxNoiseS_*t);
  double noisePH = getDoubleRandom(-maxNoisePH_*t,maxNoisePH_*t);

  //cout << "t : " << t << "  maxNoiseXY_ : " << maxNoiseXY_ << "  noiseX : " << noiseX << "  noiseY : " << noiseY << endl;
  
  return CLStateT(this->x() + noiseX,
		  this->y() + noiseY,
		  this->th() + noiseTH,
		  this->ph() + noisePH,
		  this->s() + noiseS,
		  this->t());
}


/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void CLStateT::display() const
{
  cout << "(" << this->x() << "," << this->y() << "," << this->th() << "," << 
    this->ph() << "," << this->s() << "," << this->t() << ")";
}


/*************** Utility functions ****************/

/**
 *  \brief compute the relative coordinates between both state-times
 *  \param initST : the initial state-time
 *  \param finalST : the final state-time
 *
 *  initST is center in position and heading (0,0,0)
 *  finalST is adjust in the same frame than initST
 */
void CLStateT::relativeStateT(CLStateT& initST,CLStateT& finalST)
{
  double dxWRot = finalST.x() - initST.x();
  double dyWRot = finalST.y() - initST.y();
  PositionXY posXY(dxWRot,dyWRot);
  posXY.rotate(PositionXY(0,0),-initST.th());
  double dx = posXY.x();
  double dy = posXY.y();

  double dth = computeMinDiffAngle(finalST.th(),initST.th());
  double dt = finalST.t() - initST.t();

  initST.x() = 0;
  initST.y() = 0;
  initST.th() = 0;
  initST.t() = 0;

  finalST.x() = dx;
  finalST.y() = dy;
  finalST.th() = dth;
  finalST.t() = dt;
}


                                    /**************************************************/
                                    /***************** Class SSStateT *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SSStateT::SSStateT() : StateTime(7)
{
}

/**
 *  \brief copy constructor
 *  \param p : the state-time to copy
 */
SSStateT::SSStateT(const SSStateT& s) : StateTime(7)
{
  this->x()  = s.x();
  this->y()  = s.y();
  this->th() = s.th();
  this->sx() = s.sx();
  this->sy() = s.sy();
  this->om() = s.om();
  this->t()  = s.t();
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param sx : speed sx
 *  \param sy : speed sy
 *  \param om : angular velocity om
 *  \param t : time t
 */
SSStateT::SSStateT(double x,double y,double th,double sx,double sy,double om,double t) : StateTime(7)
{
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
  this->sx() = sx;
  this->sy() = sy;
  this->om() = om;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
SSStateT::~SSStateT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state
 *  \param s2 : the state to copy
 */
void SSStateT::operator=(const SSStateT& s2)
{
  this->x()  = s2.x();
  this->y()  = s2.y();
  this->th() = s2.th();
  this->sx() = s2.sx();
  this->sy() = s2.sy();
  this->om() = s2.om();
  this->t()  = s2.t();
}

/**
 *  \brief operator += : add the input state
 *  \param s2 : the state to add
 */
void SSStateT::operator+=(const SSStateT& s2)
{
  this->x()  += s2.x();
  this->y()  += s2.y();
  this->th() += s2.th();
  this->sx() += s2.sx();
  this->sy() += s2.sy();
  this->om() += s2.om();
  this->t()  += s2.t();

}

/**
 *  \brief operator -= : substract the input state
 *  \param s2 : the state to substract
 */
void SSStateT::operator-=(const SSStateT& s2)
{
  this->x()  -= s2.x();
  this->y()  -= s2.y();
  this->th() -= s2.th();
  this->sx() -= s2.sx();
  this->sy() -= s2.sy();
  this->om() -= s2.om();
  this->t()  -= s2.t();
}

/**
 *  \brief operator += : multiply with the input coefficient
 *  \param d : the coefficient to mutilpy with
 */
void SSStateT::operator*=(double d)
{
  this->x()  *= d;
  this->y()  *= d;
  this->th() *= d;
  this->sx() *= d;
  this->sy() *= d;
  this->om() *= d;
  this->t()  *= d;
}

/**
 *  \brief operator /= : divide with the input coefficient
 *  \param d : the coefficient to divide with
 */
void SSStateT::operator/=(double d)
{
  this->x()  /= d;
  this->y()  /= d;
  this->th() /= d;
  this->sx() /= d;
  this->sy() /= d;
  this->om() /= d;
  this->t()  /= d;
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
SSStateT SSStateT::operator+(const SSStateT& s2) const
{
  return SSStateT(this->x()  + s2.x(),
		  this->y()  + s2.y(),
		  normalizedAngle(this->th() + s2.th()),
		  this->sx() + s2.sx(),
		  this->sy() + s2.sy(),
		  this->om() + s2.om(),
		  this->t()  + s2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
SSStateT SSStateT::operator-(const SSStateT& s2) const
{
  return SSStateT(this->x()  - s2.x(),
		  this->y()  - s2.y(),
		  normalizedAngle(this->th() - s2.th()),
		  this->sx() - s2.sx(),
		  this->sy() - s2.sy(),
		  this->om() - s2.om(),
		  this->t()  - s2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
SSStateT SSStateT::operator*(double d) const
{
  return SSStateT(this->x()  * d,
		  this->y()  * d,
		  normalizedAngle(this->th() * d),
		  this->sx() * d,
		  this->sy() * d,
		  this->om() * d,
		  this->t()  * d);
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
SSStateT SSStateT::operator/(double d) const
{
  return SSStateT(this->x()  / d,
		  this->y()  / d,
		  normalizedAngle(this->th() / d),
		  this->sx() / d,
		  this->sy() / d,
		  this->om() / d,
		  this->t()  / d);
}


/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void SSStateT::display() const
{
  cout << "(" << this->x() << "," << this->y() << "," << this->th() << "," << 
    this->sx() << "," << this->sy() << "," << this->om() << "," << this->t() << ")";
}


/*************** Utility functions ****************/

/**
 *  \brief compute the relative coordinates between both state-times
 *  \param initST : the initial state-time
 *  \param finalST : the final state-time
 *
 *  initST is center in position and heading (0,0,0)
 *  finalST is adjust in the same frame than initST
 */
void SSStateT::relativeStateT(SSStateT& initST,SSStateT& finalST)
{
  double dxWRot = finalST.x() - initST.x();
  double dyWRot = finalST.y() - initST.y();
  PositionXY posXY(dxWRot,dyWRot);
  posXY.rotate(PositionXY(0,0),-initST.th());
  double dx = posXY.x();
  double dy = posXY.y();

  double dth = computeMinDiffAngle(finalST.th(),initST.th());
  double dt = finalST.t() - initST.t();

  initST.x() = 0;
  initST.y() = 0;
  initST.th() = 0;
  initST.t() = 0;

  finalST.x() = dx;
  finalST.y() = dy;
  finalST.th() = dth;
  finalST.t() = dt;
}


                                    /**************************************************/
                                    /***************** Class DDStateT *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DDStateT::DDStateT() : StateTime(6)
{
}

/**
 *  \brief copy constructor
 *  \param p : the state-time to copy
 */
DDStateT::DDStateT(const DDStateT& s) : StateTime(6)
{
  this->x()  = s.x();
  this->y()  = s.y();
  this->th() = s.th();
  this->omr() = s.omr();
  this->oml() = s.oml();
  this->t()  = s.t();
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param omr : right wheel velocity omr
 *  \param oml : left wheel velocity oml
 *  \param t : time t
 */
DDStateT::DDStateT(double x,double y,double th,double omr,double oml,double t) : StateTime(6)
{
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
  this->omr() = omr;
  this->oml() = oml;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
DDStateT::~DDStateT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state
 *  \param s2 : the state to copy
 */
void DDStateT::operator=(const DDStateT& s2)
{
  this->x()   = s2.x();
  this->y()   = s2.y();
  this->th()  = s2.th();
  this->omr() = s2.omr();
  this->oml() = s2.oml();
  this->t()   = s2.t();
}

/**
 *  \brief operator += : add the input state
 *  \param s2 : the state to add
 */
void DDStateT::operator+=(const DDStateT& s2)
{
  this->x()   += s2.x();
  this->y()   += s2.y();
  this->th()  += s2.th();
  this->omr() += s2.omr();
  this->oml() += s2.oml();
  this->t()   += s2.t();
}

/**
 *  \brief operator -= : substract the input state
 *  \param s2 : the state to substract
 */
void DDStateT::operator-=(const DDStateT& s2)
{
  this->x()   -= s2.x();
  this->y()   -= s2.y();
  this->th()  -= s2.th();
  this->omr() -= s2.omr();
  this->oml() -= s2.oml();
  this->t()   -= s2.t();
}

/**
 *  \brief operator += : multiply with the input coefficient
 *  \param d : the coefficient to mutilpy with
 */
void DDStateT::operator*=(double d)
{
  this->x()   *= d;
  this->y()   *= d;
  this->th()  *= d;
  this->omr() *= d;
  this->oml() *= d;
  this->t()   *= d;
}

/**
 *  \brief operator /= : divide with the input coefficient
 *  \param d : the coefficient to divide with
 */
void DDStateT::operator/=(double d)
{
  if(d != 0)
    {
      this->x()   /= d;
      this->y()   /= d;
      this->th()  /= d;
      this->omr() /= d;
      this->oml() /= d;
      this->t()   /= d;
    }
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
DDStateT DDStateT::operator+(const DDStateT& s2) const
{
  return DDStateT(this->x()   + s2.x(),
		  this->y()   + s2.y(),
		  normalizedAngle(this->th() + s2.th()),
		  this->omr() + s2.omr(),
		  this->oml() + s2.oml(),
		  this->t()   + s2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
DDStateT DDStateT::operator-(const DDStateT& s2) const
{
  return DDStateT(this->x()   - s2.x(),
		  this->y()   - s2.y(),
		  normalizedAngle(this->th() - s2.th()),
		  this->omr() - s2.omr(),
		  this->oml() - s2.oml(),
		  this->t()   - s2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
DDStateT DDStateT::operator*(double d) const
{
  return DDStateT(this->x()   * d,
		  this->y()   * d,
		  normalizedAngle(this->th() * d),
		  this->omr() * d,
		  this->oml() * d,
		  this->t()   * d);
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
DDStateT DDStateT::operator/(double d) const
{
  return DDStateT(this->x()   / d,
		  this->y()   / d,
		  normalizedAngle(this->th() / d),
		  this->omr() / d,
		  this->oml() / d,
		  this->t()   / d);
}

/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void DDStateT::display() const
{
  cout << "(" << this->x() << "," << this->y() << "," << this->th() << "," << 
    this->omr() << "," << this->oml() << "," << this->t() << ")";
}


                                    /**************************************************/
                                    /***************** Class DD2StateT *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2StateT::DD2StateT() : StateTime(6)
{
}

/**
 *  \brief copy constructor
 *  \param p : the state-time to copy
 */
DD2StateT::DD2StateT(const DD2StateT& s) : StateTime(6)
{
  this->x()  = s.x();
  this->y()  = s.y();
  this->th() = s.th();
  this->s()  = s.s();
  this->om() = s.om();
  this->t()  = s.t();  
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param s : linear velocity s
 *  \param om : angular velocity om
 *  \param t : time t
 */
DD2StateT::DD2StateT(double x,double y,double th,double s,double om,double t) : StateTime(6)
{
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
  this->s()  = s;
  this->om() = om;
  this->t()  = t;
}

/**
 *  \brief init constructor
 *  \param s : state s
 *  \param t : time t
 */
DD2StateT::DD2StateT(const DD2State& s,double t) : StateTime(6)
{
  this->x()  = s.x();
  this->y()  = s.y();
  this->th() = s.th();
  this->s()  = s.s();
  this->om() = s.om();
  this->t()  = t;
}

/**
 *  \brief destructor
 */
DD2StateT::~DD2StateT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state
 *  \param s2 : the state to copy
 */
void DD2StateT::operator=(const DD2StateT& s2)
{
  this->x()  = s2.x();
  this->y()  = s2.y();
  this->th() = s2.th();
  this->s()  = s2.s();
  this->om() = s2.om();
  this->t()  = s2.t();
}

/**
 *  \brief operator += : add the input state
 *  \param s2 : the state to add
 */
void DD2StateT::operator+=(const DD2StateT& s2)
{
  this->x()  += s2.x();
  this->y()  += s2.y();
  this->th() += s2.th();
  this->s()  += s2.s();
  this->om() += s2.om();
  this->t()  += s2.t();
}

/**
 *  \brief operator -= : substract the input state
 *  \param s2 : the state to substract
 */
void DD2StateT::operator-=(const DD2StateT& s2)
{
  this->x()  -= s2.x();
  this->y()  -= s2.y();
  this->th() -= s2.th();
  this->s()  -= s2.s();
  this->om() -= s2.om();
  this->t()  -= s2.t();
}

/**
 *  \brief operator += : multiply with the input coefficient
 *  \param d : the coefficient to mutilpy with
 */
void DD2StateT::operator*=(double d)
{
  this->x()  *= d;
  this->y()  *= d;
  this->th() *= d;
  this->s()  *= d;
  this->om() *= d;
  this->t()  *= d;
}

/**
 *  \brief operator /= : divide with the input coefficient
 *  \param d : the coefficient to divide with
 */
void DD2StateT::operator/=(double d)
{
  if(d != 0)
    {
      this->x()  /= d;
      this->y()  /= d;
      this->th() /= d;
      this->s()  /= d;
      this->om() /= d;
      this->t()  /= d;
    }
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
DD2StateT DD2StateT::operator+(const DD2StateT& s2) const
{
  return DD2StateT(this->x()   + s2.x(),
		  this->y()   + s2.y(),
		  normalizedAngle(this->th() + s2.th()),
		  this->s()   + s2.s(),
		  this->om()  + s2.om(),
		  this->t()   + s2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
DD2StateT DD2StateT::operator-(const DD2StateT& s2) const
{
  return DD2StateT(this->x()   - s2.x(),
		  this->y()   - s2.y(),
		  normalizedAngle(this->th() - s2.th()),
		  this->s()   - s2.s(),
		  this->om()  - s2.om(),
		  this->t()   - s2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
DD2StateT DD2StateT::operator*(double d) const
{
  return DD2StateT(this->x()   * d,
		  this->y()   * d,
		  normalizedAngle(this->th() * d),
		  this->s()   * d,
		  this->om()  * d,
		  this->t()   * d);
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
DD2StateT DD2StateT::operator/(double d) const
{
  return DD2StateT(this->x()   / d,
		  this->y()   / d,
		  normalizedAngle(this->th() / d),
		  this->s()   / d,
		  this->om()  / d,
		  this->t()   / d);
}


/**************** Noise functions *****************/

/**
 *  \brief return a randomly noised state from the current one
 *  \param t : the elapsed time from the previous known state
 *  \return noised state
 */
DD2StateT DD2StateT::noisedState(double t)
{
  return DD2StateT(*this);
}


/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void DD2StateT::display() const
{
  cout << "(" << this->x() << "," << this->y() << "," << this->th() << "," << 
    this->s() << "," << this->om() << "," << this->t() << ")";
}


/*************** Utility functions ****************/

/**
 *  \brief compute the relative coordinates between both state-times
 *  \param initST : the initial state-time
 *  \param finalST : the final state-time
 *
 *  initST is center in position and heading (0,0,0)
 *  finalST is adjust in the same frame than initST
 */
void DD2StateT::relativeStateT(DD2StateT& initST,DD2StateT& finalST)
{
  double dxWRot = finalST.x() - initST.x();
  double dyWRot = finalST.y() - initST.y();
  PositionXY posXY(dxWRot,dyWRot);
  posXY.rotate(PositionXY(0,0),-initST.th());
  double dx = posXY.x();
  double dy = posXY.y();

  double dth = computeMinDiffAngle(finalST.th(),initST.th());
  double dt = finalST.t() - initST.t();

  initST.x() = 0;
  initST.y() = 0;
  initST.th() = 0;
  initST.t() = 0;

  finalST.x() = dx;
  finalST.y() = dy;
  finalST.th() = dth;
  finalST.t() = dt;
}


//                                     /**************************************************/
//                                     /**************** Class ManaStateT ****************/
//                                     /**************************************************/

// /********** Constructors and destructors **********/

// /**
//  *  \brief default constructor
//  */
// ManaStateT::ManaStateT() : WCStateT()
// {
// }

// /**
//  *  \brief copy constructor
//  *  \param p : the state-time to copy
//  */
// ManaStateT::ManaStateT(const ManaStateT& s) : WCStateT(s)
// {
// }

// /**
//  *  \brief init constructor
//  *  \param x : position x
//  *  \param y : position y
//  *  \param th : heading th
//  *  \param v : linear velocity v
//  *  \param w : angular velocity w
//  *  \param t : time t
//  */
// ManaStateT::ManaStateT(double x,double y,double th,double v,double w,double t) : WCStateT(x,y,th,v,w,t)
// {
// }

// /**
//  *  \brief init constructor
//  *  \param s : state s
//  *  \param t : time t
//  */
// ManaStateT::ManaStateT(const ManaState& s,double t) : WCStateT(s.x(),s.y(),s.th(),s.v(),s.w(),t)
// {
// }

// /**
//  *  \brief destructor
//  */
// ManaStateT::~ManaStateT()
// {
// }


// /******************* Operators ********************/

// /**
//  *  \brief operator = : copy the input state
//  *  \param s2 : the state to copy
//  */
// void ManaStateT::operator=(const ManaStateT& s2)
// {
//   this->x()  = s2.x();
//   this->y()  = s2.y();
//   this->th() = s2.th();
//   this->v()  = s2.v();
//   this->w()  = s2.w();
//   this->t()  = s2.t();
// }

// //     /**
// //      *  \brief operator += : add the input state
// //      *  \param s2 : the state to add
// //      */
// //     void operator+=(const ManaStateT& s2);

// //     /**
// //      *  \brief operator -= : substract the input state
// //      *  \param s2 : the state to substract
// //      */
// //     void operator-=(const ManaStateT& s2);

// /**
//  *  \brief operator + : sumation
//  *  \param s2 : the second vector to add
//  *  \return the sum of both vectors
//  */
// ManaStateT ManaStateT::operator+(const ManaStateT& s2) const
// {
//   return ManaStateT(this->x()   + s2.x(),
// 		    this->y()   + s2.y(),
// 		    normalizedAngle(this->th() + s2.th()),
// 		    this->v()   + s2.v(),
// 		    this->w()   + s2.w(),
// 		    this->t()   + s2.t());
// }

// /**
//  *  \brief operator + : substraction
//  *  \param s2 : the second vector to substract
//  *  \return the substraction of both vectors
//  */
// ManaStateT ManaStateT::operator-(const ManaStateT& s2) const
// {
//   return ManaStateT(this->x()   - s2.x(),
// 		    this->y()   - s2.y(),
// 		    normalizedAngle(this->th() - s2.th()),
// 		    this->v()   - s2.v(),
// 		    this->w()   - s2.w(),
// 		    this->t()   - s2.t());
// }

// /**
//  *  \brief operator * : scalar multiplication
//  *  \param d : the scalar to multiply with
//  *  \return the product of the current vector and the scalar
//  */
// ManaStateT ManaStateT::operator*(double d) const
// {
//   return ManaStateT(this->x()   * d,
// 		    this->y()   * d,
// 		    normalizedAngle(this->th() * d),
// 		    this->v()   * d,
// 		    this->w()   * d,
// 		    this->t()   * d);
// }

// //     /**
// //      *  \brief operator / : division by a scalar
// //      *  \param d : the scalar to divide with
// //      *  \return the division result
// //      */
// //     ManaStateT operator/(double d) const;
