#include "common/inc/StateTimeTree.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /*************** Class StateTimeNode **************/
                                    /**************************************************/

int StateTimeNode::displayPrecision_ = 4;
int StateTimeNode::displayTab_ = 4;


/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
StateTimeNode::StateTimeNode() : Node<StateTime*>()
{
  this->val() = NULL;
}

/**
 *  \brief copy constructor
 *  \param n : the node to copy
 */  
StateTimeNode::StateTimeNode(const StateTimeNode& n)
{
  this->father_ = n.father();
  int N = n.nbSons();
  for(int i=0; i<N; i++)
    this->addSon(n.son(i));

  this->val() = n.val()->copy();
}

/**
 *  \brief init constructor
 *  \param val : the state-time contained in the node
 */  
StateTimeNode::StateTimeNode(const StateTime& s)
{
  this->father_ = NULL;
  this->val() = s.copy();
}

/**
 *  \brief destructor
 */  
StateTimeNode::~StateTimeNode()
{
  if(val_ != NULL) delete val_;
}


/*************** Getters and Setters **************/

/**
 *  \brief add a son to the current node
 *  \param s : the state-time value of the son
 */
void StateTimeNode::addSon(const StateTime& s)
{
  StateTimeNode* newNode = new StateTimeNode(s);
  this->addSon(newNode);
}

/**
 *  \brief add a son to the current node
 *  \param s : a pointer to the new son
 */
void StateTimeNode::addSon(StateTimeNode* s)
{
  this->sons_.push_back(s); 
  s->father_ = this;
}


/**************** Utility functions ***************/

/**
 *  \brief return the branch of the tree reaching the position pos and minimizing the final time
 *  \param targetPos : the target position
 *  \param minTimeBranch : the resulting list containing all the nodes of the best branch
 */
void StateTimeNode::computeMinTimeBranch(const PositionXY& targetPos, std::list<StateTime*> minTimeBranch)
{
  // get back the leaves of the tree
  std::vector<Node<StateTime*>*> vecLeaves;
  this->getLeavesNodes(vecLeaves);

  // check if the leaves vector is not empty
  if(vecLeaves.size() == 0)
    {
      cerr << "StateTimeNode::computeMinTimeBranch : The vector of leaves is empty!" << endl;
      return;
    }

  // search for the leaf minimizing the final time
  double minT = 0;
  Node<StateTime*>* bestNode = NULL;
  int N = vecLeaves.size();

  for(int i=0; i<N; i++)
    {
      // get back the current position-time
      PositionXYT curPosT = vecLeaves.at(i)->val()->posXYT();

      if(bestNode == NULL || (curPosT.x() == targetPos.x() && curPosT.y() == targetPos.y() && curPosT.t() < minT))
	{
	  bestNode = vecLeaves.at(i);
	  minT = curPosT.t();
	}
    }

  // get back the whole branch going from the root of the tree up to the best node
  bestNode->getVecValuesFromRoot(minTimeBranch);
}


/**************** Display functions ***************/

/**
 *  \brief display the state-time tree from the current root
 *  \param level : input level from the root of the tree used to shift the current display
 */
void StateTimeNode::display(int level) const
{
  for(int i=0; i<level * displayTab_; i++) cout << " ";
  cout << " -> ";
  if(val_ != NULL) val_->display(StateTimeNode::displayPrecision_);
  else cout << "NULL";
  cout << endl;

  int N = this->nbSons();
  for(int i=0; i<N; i++)
    {
      this->son(i)->display(level+1);
    }
}


                                    /**************************************************/
                                    /************* Class DD2StateTimeNode *************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
DD2StateTimeNode::DD2StateTimeNode() : StateTimeNode()
{
}

/**
 *  \brief copy constructor
 *  \param n : the node to copy
 */  
DD2StateTimeNode::DD2StateTimeNode(const DD2StateTimeNode& n) : StateTimeNode(n)
{
}

/**
 *  \brief init constructor
 *  \param val : the state-time contained in the node
 */  
DD2StateTimeNode::DD2StateTimeNode(const DD2StateT& s) : StateTimeNode(s)
{
}

/**
 *  \brief destructor
 */  
DD2StateTimeNode::~DD2StateTimeNode()
{
}


/*************** Getters and Setters **************/

/**
 *  \brief return the state-time of the current node
 *  \return the state-time of the current node
 */
DD2StateT DD2StateTimeNode::getStateT() const
{
  return *((DD2StateT*)(val_));
}


                                    /**************************************************/
                                    /************** Class CLStateTimeNode *************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
CLStateTimeNode::CLStateTimeNode() : StateTimeNode()
{
}

/**
 *  \brief copy constructor
 *  \param n : the node to copy
 */  
CLStateTimeNode::CLStateTimeNode(const CLStateTimeNode& n) : StateTimeNode(n)
{
}

/**
 *  \brief init constructor
 *  \param val : the state-time contained in the node
 */  
CLStateTimeNode::CLStateTimeNode(const CLStateT& s) : StateTimeNode(s)
{
}

/**
 *  \brief destructor
 */  
CLStateTimeNode::~CLStateTimeNode()
{
}


/*************** Getters and Setters **************/

/**
 *  \brief return the state-time of the current node
 *  \return the state-time of the current node
 */
CLStateT CLStateTimeNode::getStateT() const
{
  return *((CLStateT*)(val_));
}
