#include "Vector.hpp"
using namespace mpc;

#include <cmath>
#include <iostream>
#include <iomanip>
using namespace std;

double eps = 1e-10;

                                    /**************************************************/
                                    /******************** Class Vec *******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Vector::Vector(){}

/**
 *  \brief copy constructor
 *  \param v : the vector to copy
 */
Vector::Vector(const Vector& v)
{
  //this->coord_.clear();
  for(unsigned int i=0; i<v.size(); i++)
    {
      this->coord_.push_back(v.at(i));
    }
}

/**
 *  \brief init constructor
 *  \param n : number of elements of the vector
 */
Vector::Vector(int n)
{
  //this->coord_.clear();
  this->coord_ = vector<double>(n);
}

/**
 *  \brief destructor
 */
Vector::~Vector(){}


/******************** Accessors *******************/

/**
 *  \brief equality test
 *  \param v2 : the second vector to test equality with
 *  \return true if both vectors are equals, false else
 */
bool Vector::equals(const Vector& v2) const
{
  return ((*this) == v2);
}


/******************* Operators ********************/

/**
 *  \brief operator = : assignation
 *  \param v2 : the vector to assign to the current one
 */
void Vector::operator=(const Vector& v2)
{
  this->coord_.clear();
  for(unsigned int i=0; i<v2.size(); i++)
    {
      this->coord_.push_back(v2.at(i));
    }
}

/**
 *  \brief operator + : sumation
 *  \param v2 : the second vector to add
 *  \return the sum of both vectors
 */
Vector Vector::operator+(const Vector& v2) const
{
  if(this->size() != v2.size())
    {
      return Vector();
    }
  else
    {
      Vector v(this->size());
      for(int i=0; i<v.size(); i++)
	{
	  v[i] = this->at(i) + v2.at(i);
	}
      return v;
    }
}

/**
 *  \brief operator + : substraction
 *  \param v2 : the second vector to substract
 *  \return the substraction of both vectors
 */
Vector Vector::operator-(const Vector& v2) const
{
  if(this->size() != v2.size())
    {
      return Vector();
    }
  else
    {
      Vector v(this->size());
      for(int i=0; i<v.size(); i++)
	{
	  v[i] = this->at(i) - v2.at(i);
	}
      return v;
    }
}

/**
 *  \brief operator == : equality test
 *  \param v2 : the second vector to test equality with
 *  \return true if both vectors are equals, false else
 */
bool Vector::operator==(const Vector& v2) const
{
  if(this->size() != v2.size())
    {
      return false;
    }
  else
    {
      int i=0;
      bool trouve = false;
      while(i<this->size() && !trouve)
	{
	  if(this->at(i) >= v2.at(i) - eps && this->at(i) <= v2.at(i) + eps)
	    i++;
	  else
	    trouve = true;
	}
      if(trouve) return false;
      else return true;
    }
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
Vector Vector::operator*(double d) const
{
  Vector res(this->size());
  for(int i=0; i<res.size(); i++)
    {
      res[i] = this->at(i) * d;
    }
  return res;
}

/**
 *  \brief operator * : dot product
 *  \param v2 : the vector to multiply with
 *  \return the dot product of both vectors
 */
double Vector::operator*(const Vector& v2) const
{
  // check if the size of both vector is the same
  if(this->size() == v2.size())
    {
      // compute the scalar product
      double res;
      for(int i=0; i<this->size(); i++)
	{
	  res += (this->at(i) * v2.at(i));
	}
      return res;
    }
  else
    {
      return 0;
    }
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
Vector Vector::operator/(double d) const
{
  if(d != 0)
    {
      Vector res(this->size());
      for(int i=0; i<res.size(); i++)
	{
	  res[i] = this->at(i) * d;
	}
      return res;
    }
  else
    {
      return Vector();
    }
}

/**
 *  \brief operator += : sumation
 *  \param v2 : the second vector to add
 */
void Vector::operator+=(const Vector& v2)
{
  if(this->size() != v2.size())
    {
      // nothing to do
    }
  else
    {
      for(int i=0; i<this->size(); i++)
	{
	  (*this)[i] += v2.at(i);
	}
    }
}

/**
 *  \brief operator += : substraction
 *  \param v2 : the second vector to substract
 */
void Vector::operator-=(const Vector& v2)
{
  if(this->size() != v2.size())
    {
      // nothing to do
    }
  else
    {
      for(int i=0; i<this->size(); i++)
	{
	  (*this)[i] -= v2.at(i);
	}
    }
}

/**
 *  \brief operator *= : scalar multiplication
 *  \param d : the scalar to multiply with
 */
void Vector::operator*=(double d)
{
  for(int i=0; i<this->size(); i++)
    {
      (*this)[i] *= d;
    }
}

/**
 *  \brief operator /= : division by a scalar
 *  \param d : the scalar to divide with
 */
void Vector::operator/=(double d)
{
  if(d == 0)
    {
      // nothing to do
    }
  else
    {
      for(int i=0; i<this->size(); i++)
	{
	  (*this)[i] /= d;
	}
    }
}


/***************** Math functions *****************/

/**
 *  \brief norm of the vector
 *  \return the norm of the vector
 */
double Vector::norm() const
{
  return sqrt(this->sqNorm());
}

/**
 *  \brief square norm of the vector
 *  \return the square norm of the vector
 */
double Vector::sqNorm() const
{
  double res = 0;
  for(int i=0; i<this->size(); i++)
    res += (this->at(i) * this->at(i));
  return res;
}

/**
 *  \brief compute the normalized vector from the current one
 *  \return the normalized vector
 */
Vector Vector::normalized() const
{
  double norm = this->norm();

  if(norm != 0)
    {
      return ((*this) / norm);
    }
  else return Vector(this->size());
}

/**
 *  \brief normalize the current vector
 */
void Vector::normalize()
{
  double norm = this->norm();

  if(norm != 0)
    {
      (*this) /= norm;
    }
}


/*************** Display functions ****************/

/**
 *  \brief display the vector's coordinates
 */
void Vector::display() const
{
  int N = this->size();

  cout << "(";
  for(int i=0; i<N-1; i++)
    { cout << coord_[i] << ","; }
  if(N != 0) cout << this->at(N-1);
  cout << ")";
}

/**
 *  \brief display the vector's coordinates
 *  \param number : the number of digit used to display each value
 */
void Vector::display(int number) const
{
  int N = this->size();

  cout << "(";
  for(int i=0; i<N-1; i++)
    { cout << setprecision(number) << coord_[i] << ","; }
  if(N != 0) cout << setprecision(number) << this->at(N-1);
  cout << ")";
}

/**
 *  \brief convert the vector into a string
 *  \return the resulting string
 */
const char* Vector::toString() const
{
  int N = this->size();
  std::string str("(");
  
  for(int i=0; i<N-1; i++)
    {
      char nbChar[50];
      double2String(this->at(i),nbChar,4);
      str += nbChar;
      str += ",";
    }
  if(N != 0)
    {
      char nbChar[50];
      double2String(this->at(N-1),nbChar,4);
      str += nbChar;
    }
  str += ")";
  
  return str.c_str();
}
