#include "controller/inc/ManaSimpleController.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /*********** Class ManaSimpleController ***********/
                                    /**************************************************/

bool ManaSimpleController::fixFinalTH = false;

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 *  Only used by subclasses
 */
ManaSimpleController::ManaSimpleController() : SimpleController()
{
}

/**
 *  \brief init constructor
 *  \param sys : the robotic system handled
 *  \param dt : the controller period
 *  \param debug : the debugging level
 */
ManaSimpleController::ManaSimpleController(ManaSys* sys, double dt, int debug) : SimpleController(sys,dt,debug)
{
  this->trajgen_ = new ManaTrajGen(sys);

  this->initTrajGen();
  this->initManaTrajGen();
}

/**
 *  \brief destructor
 */
ManaSimpleController::~ManaSimpleController()
{
  if(trajgen_ != NULL) delete trajgen_;
}


/*************** Getters and Setters **************/

/**
 *  \brief return a pointer to the Mana trajectory generator
 *  \return a pointer to the Mana trajectory generator
 */
ManaTrajGen* ManaSimpleController::trajgen()
{
  return ((ManaTrajGen*)(this->trajgen_));
}

/**
 *  \brief return the system constraints
 *  \return the system constraints
 */
const ManaSys* ManaSimpleController::sys() const
{
  return ((ManaSys*)(this->sys_));
}


/*************** Control Computation **************/

/**
 *  \brief Initialize the parameters of the Mana trajectory generator
 */
void ManaSimpleController::initManaTrajGen()
{
  /// Mana trajgen options

  // Mana constraints
  trajgen()->setUseVelCst(false);               // default = true
  trajgen()->setUseAccCst(false);               // default = true
  trajgen()->setUseAVelCst(false);              // default = true
  trajgen()->setUseAAccCst(false);              // default = true

  // final state constraints
  trajgen()->fixFinalXY(true);                              // default = true
  trajgen()->fixFinalTH(ManaSimpleController::fixFinalTH);  // default = true
  trajgen()->fixFinalV(false);                              // default = true
  trajgen()->fixFinalOM(false);                             // default = true
}

/**
 *  \brief compute the control trajectory to apply to go from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the goal state
 *  \return 0 if the process succeeded, -1 else
 */
int ManaSimpleController::computePCTraj(const ManaState& s1,const ManaState& s2)
{
  // compute the distance between the initial and final states
  double dx = s2.x() - s1.x();
  double dy = s2.y() - s1.y();
  double dist = sqrt(dx*dx + dy*dy);

  double tf = 1.0 * minTravelTime(dist,                     // distance to travel
				  s1.v(),                   // current velocity
				  sys()->getVMax(),         // maximal speed
				  sys()->getAMax());        // maximal acceleration
  
  ManaStateT st1(s1,0.0);
  ManaStateT st2(s2,tf);

  // generation of a trajectory between s0 and sf
  // return 0 if the process succeed, -1 else
  int res = trajgen()->genTraj(st1,st2);

  // get back the resulting control trajectory
  {
    boost::mutex::scoped_lock lock(mutexTraj_);
    if(ctraj_ != NULL) delete ctraj_;
    ctraj_ = trajgen()->getPCTraj();
  }

  if(ctraj_ != NULL)
    {
      if(debug_)
	{
	  boost::mutex::scoped_lock lock(mutexTraj_);
	  if(straj_ != NULL) delete straj_;
	  straj_ = trajgen()->getPSTraj();
	}

      return 0;
    }
  else return -1;
}

/**
 *  \brief compute the control to apply to go from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the goal state
 *  \return the resulting control
 */
ManaControl ManaSimpleController::computeControl(const ManaState& s1,const ManaState& s2)
{
  // Generation of a control trajectory up to the goal
  int resTrajGen = this->computePCTraj(s1, s2);

  // check if the trajectory generation process succeeded
  if(resTrajGen == 0)
    {
      // compute the control that should be reached at the end of a controller period
      //ManaControl ctrl = trajgen()->getManaControlReachedAt(dt_);
      ManaControl ctrl = ctraj()->at(dt_);
      
      // compute the average control to apply during the next period of the task
      ManaControl ctrlRes = ctrl;

      return ctrlRes;
    }
  else return ManaControl(0,0);
}

/**
 *  \brief compute the control trajectory to apply to go from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the goal state
 *  \return a new instance of the the control trajectory to apply, or NULL if the process failed
 */
ManaPCTraj* ManaSimpleController::computeCTraj(const ManaState& s1,const ManaState& s2)
{
  // Generation of a control trajectory up to the goal
  int resTrajGen = this->computePCTraj(s1, s2);

  // check if the trajectory generation process succeeded
  if(resTrajGen == 0)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      ManaPCTraj* copyTraj = new ManaPCTraj(*ctraj());
      return copyTraj;
    }
  else return NULL;
}


/***************** Debug / Display ****************/

/**
 *  \brief return a copy of the current control trajectory
 *  \return a copy of the current control trajectory
 * 
 *  A new instance of ManaPCTraj is returned
 */
const ManaPCTraj* ManaSimpleController::getPCTraj()
{
  if(ctraj_ != NULL)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      ManaPCTraj* copyTraj = new ManaPCTraj(*ctraj());
      return copyTraj;
    }
  else return NULL;
}

/**
 *  \brief return a copy of the current control trajectory
 *  \return a copy of the current control trajectory
 * 
 *  A new instance of ManaPSTraj is returned
 */
const ManaPSTraj* ManaSimpleController::getPSTraj()
{
  if(straj_ != NULL)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      ManaPSTraj* copyTraj = new ManaPSTraj(*straj());
      return copyTraj;
    }
  else return NULL;
}
