#ifndef MPC_TRAJDEFORM_HPP
#define MPC_TRAJDEFORM_HPP

#include "common/inc/PositionTime.hpp"
#include "common/inc/StateTime.hpp"
#include "common/inc/ControlTime.hpp"
#include "motion/inc/STrajectory.hpp"
#include "environment/inc/Environment.hpp"
#include "system/inc/System.hpp"
#include "common/inc/Timer.hpp"

namespace mpc
{

  /**
   * \enum modeDeform_t
   * \brief Different deformation modes
   *
   * SPATIOTEMP_DEFORM : spatio-temporal deformation
   * TEMPONLY_DEFORM   : temporal deformation only
   * SPATONLY_DEFORM   : spatial deformation only
   */
  typedef enum {
    SPATIOTEMP_DEFORM  = 1,
    TEMPONLY_DEFORM    = 2,
    SPATONLY_DEFORM    = 3
  } modeDeform_t;

#define T_FEA   0
#define T_FAR   1
#define T_UNFEA 2


  /**
   * \enum deformerState_t
   * \brief Enumeration of the different deformer status types
   *
   * DEFORMER_NULL : deformer not allocated
   * DEFORMER_DISCONNECTED : disconnected deformer
   * DEFORMER_STOPPED : stopped deformer
   * DEFORMER_STARTED : started deformer
   */
  typedef enum {
    DEFORMER_NULL          = -1,
    DEFORMER_DISCONNECTED  =  0,
    DEFORMER_STOPPED       =  1,
    DEFORMER_STARTED       =  2
  } deformerStatus_t;

                                    /**************************************************/
                                    /**************** Class TrajDeform ****************/
                                    /**************************************************/

  /** \class TrajDeform
   *  \brief Trajectory deformer TEDDY
   */
  class TrajDeform
  {
  protected : 
    STraj* traj_;                    /**< the trajectory which will be deformed */
    Environment* env_;                     /**< the environment by which the trajectory is deformed */
    System*      sys_;                     /**< the features of the robot system which is used */

    StateTime*   curStateT_;               /**< the current state-time of the system (localization + state informations) */

    double       K_INT;                    /**< internal forces coefficient */
    double       K_SPT_TMP;                /**< balance between spatial and temporal forces */
    double       K_EXT;                    /**< external forces coefficient (dynamic obstacles) */
    double       K_STEXT;                  /**< external forces coefficient (static obstacles) */

    modeDeform_t modeDeform_;              /**< mode of the deformation which will be applied */

    double       alphaMin;                 /**< minimal spatial coefficient (for the metric) */
    double       alphaMax;                 /**< maximal spatial coefficient (for the metric) */
    double       betaMin;                  /**< minimal temporal coefficient (for the metric) */
    double       betaMax;                  /**< maximal temporal coefficient (for the metric) */

    double       spMed;                    /**< spatial maximal effective distance */
    double       tmpMed;                   /**< temporal maximal effective distance */
    double       maxStEffectDist;          /**< maximum distance until which external forces take effects (static obstacles) */

    double       reliabilityTime;          /**< reliability time */
    double       epsilonRT;                /**< epsilon for reliability time */
    double       predHorizon_;             /**< prediction horizon during which we consider the previsional model to b completely reliable */

    bool         useTemporalConsistency_;  /**< true if the temporal consistency is used */
    bool         useResampling_;           /**< true if the resampling process is used */

    int          *typeState_;              /**< table of the type of the states of the trajectory */

    PositionXYT* *tableNearState_;         /**< table of nearest states of dynamic obstacles from the states of the trajectory */
    PositionXYT* *tableNearStState_;       /**< table of nearest states of static obstacles from the states of the trajectory */

    double       oldTime_;                 /**< time of the previous deformation which has been done */

    double       timeStep_;                /**< time step of the trajectory computation (used to adjust the number of state-time) */
    double       posStep_;                 /**< pos step of the trajectory computation (used to adjust the number of state-time) */

    int          I;

    double       lastDeformDuration_;      /**< duration of the last deformation process */

    //   Timer       timer_;                   /**< timer used for the trajectory deformer Teddy */

    int         debug_;                    /**< debug mode used for the deformation process */

    boost::mutex* mutexTraj2_;             /**< mutex used to prevent trajectory access conflicts */
    boost::mutex mutexTables_;             /**< mutex used to prevent display tables (forces, nearest obstacles) access conflicts */


    /********** Constructors and destructors **********/
  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    TrajDeform();

  public : 
    /**
     *  \brief init constructor
     *  \param traj : the initial trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \param sys : the features of the robot system used
     */
    TrajDeform(STraj* traj,Environment* env,System* sys);

    /**
     *  \brief init constructor
     *  \param traj : the initial trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \param sys : the features of the robot system used
     *  \param dbg : the debugging mode used
     */
    TrajDeform(STraj* traj,Environment* env,System* sys,int dbg);

    //   /**
    //    *  \brief init constructor
    //    *  \param traj : the initial trajectory to deform
    //    *  \param env : the environment in which the system evolve
    //    *  \param sys : the features of the robot system used
    //    *  \param typeS : pointer to the table of state types (for simulation display)
    //    *  \param tns : pointer to the table of nearest states (for simulation display)
    //    *  \param tfi : pointer to the table of internal forces (for simulation display)
    //    *  \param tfe : pointer to the table of external forces (for simulation display)
    //    */
    //   TrajDeform(STraj* traj,Environment* env,System* sys,int** typeS,PositionXYT*** tns,StateTime** tfi,StateTime** tfe);

    /**
     *  \brief destructor
     */
    virtual ~TrajDeform();


    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    virtual STraj* traj() const = 0;

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline STraj* getTraj() const { return this->traj_; }

    /**
     *  \brief set the current trajectory
     *  \param the input trajectory
     */
    inline void setTraj(STraj* traj) { this->traj_ = traj; }

    /**
     *  \brief return the current environment
     *  \return the current environment
     */
    inline Environment* getEnv() const { return this->env_; }

    /**
     *  \brief set the current environment
     *  \param env : the input environment
     */
    inline void setEnv(Environment* env) { this->env_ = env; }

    /**
     *  \brief return the current features of the system
     *  \return the current features of the system
     */
    inline System* getSys() const { return this->sys_; }

    /**
     *  \brief set the current features of the system
     *  \param sys : the input features of the system
     */
    inline void setSystem(System* sys) { this->sys_ = sys; }

    /**
     *  \brief set the deformation mode
     *  \param mode : the input deformation mode
     */
    inline void setModeDeform(modeDeform_t mode) { this->modeDeform_ = mode; }

    /**
     *  \brief set the temporal consistency mode
     *  \param b : true if the temporal consistency must be used
     */
    inline void setTemporalConst(bool b) { this->useTemporalConsistency_ = b; }

    /**
     *  \brief set the time step
     *  \param ts : average time that should be kept between two nodes of the trajectory
     */
    inline void setTimeStep(double ts) { this->timeStep_ = ts; }

    /**
     *  \brief set the position step
     *  \param ts : average distance that should be kept between two nodes of the trajectory
     */
    inline void setPosStep(double ps) { this->posStep_ = ps; }

    /**
     *  \brief set the coefficients of internal and external forces and the balance between external spatial and temporal forces
     *  \param kint : pointer to the internal forces coefficient
     *  \param kext : pointer to the external forces coefficient for dynamic obstacles
     *  \param kstext : pointer to the external forces coefficient for static obstacles
     *  \param amin : pointer to the minimal spatial coefficient
     *  \param amax : pointer to the maximal spatial coefficient
     *  \param bmin : pointer to the minimal temporal coefficient
     *  \param bmax : pointer to the maximal temporal coefficient
     */
    void setForcesCoef(double kint, double kext, double kstext, double amin, double amax, double bmin, double bmax); 
       
    /**
     *  \brief set the maximal effective distances
     *  \param spMed : pointer to the maximal effective spatial distance from dynamic obstacles
     *  \param tmpMed : pointer to the maximal effective temporal distance from dynamic obstacles
     *  \param staticMed : pointer to the maximal effective spatial distance from static obstacles
     */
    void setMaxEffectDist(double spMed,double tmpMed, double staticMed);

    /**
     *  \brief set the prediction horizon, the reliability time and the associated epsilon value
     *  \param predH : pointer to the prediction horizon time (time until which we consider the forecast model to be completely reliable)
     *  \param rtime : pointer to the reliability time (time until which we consider the forecast model to be partialy reliable)
     *  \param eps : pointer to the standard deviation of the gaussian model
     */
    void setReliability(double predH, double rtime, double eps);
    // 
    /**
     *  \brief set the duration of the last deformation process
     *  \param deformDur : the duration of the last deformation process
     *
     *    ???????????????????????????
     */
    inline void setLastDeformDuration(double deformDur) { lastDeformDuration_ = deformDur; }

    /**
     *  \brief enable/disable the resampling mode
     *  \param u : true if the resampling process is used
     */
    inline void setUseResampling(bool u) { useResampling_ = u; }

    /**
     *  \brief set the debugging level
     *  \param dbg : the new debugging level
     */
    inline virtual void setDebug(int dbg) { this->debug_ = dbg; }

    /*************** Forces computation ***************/

  protected : 
    //   /**
    //    *  \brief computing of the internal forces
    //    *  \param prev : the previous node of the trajectory
    //    *  \param cur : the current node of the trajectory
    //    *  \param next : the next node of the trajectory
    //    */
    //   StateTime computeFInt(const StateTime& prev, const StateTime& cur, const StateTime& next);

    /**
     *  \brief compute the total external forces
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     *  \param timeEnv : the current time of the environment
     */
    virtual PositionXYT computeFExt(const StateTime& cur, Environment* env, double time, bool last,double timeEnv=-1);

    /**
     *  \brief compute the total external forces (due to dynamic obstacles)
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     *  \param timeEnv : the current time of the environment
     *
     *  Compute the resulting force from the nearest obstacle
     */
    PositionXYT computeDynFExt(const StateTime& cur, Environment* env, double time, bool last,double timeEnv=-1);

    /**
     *  \brief compute the total external forces (due to dynamic obstacles)
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param last : true if the current node is the last node of the trajectory
     *  \param timeTraj : the current time on the followed trajectory
     *  \param timeEnv : the current time registered by the environment perception process
     *
     *  Compute the resulting force as the sum of all forces from the obstacles
     *  surrounding the robot in a close radius
     */
    PositionXYT computeDynFExt2(const StateTime& cur, Environment* env, bool last, double timeTraj, double timeEnv);

    /**
     *  \brief compute the total external forces due to static obstacles (1st version)
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     *
     *  Compute the resulting force from the nearest obstacle
     */
    PositionXYT computeStFExt(const StateTime& cur, Environment* env, bool last);

    /**
     *  \brief compute the total external forces due to static obstacles (2nd version -> force applied from each static obstacle)
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     *
     *  Compute the resulting force as the sum of all forces from the obstacles
     *  surrounding the robot in a close radius
     */
    PositionXYT computeStFExt2(const StateTime& cur, Environment* env, bool last);

    /**
     *  \brief compute the total external forces due to static obstacles (3rd version -> force applied from the closest obstacle)
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     *
     *  Compute the resulting force from the nearest obstacle
     */
    PositionXYT computeStFExt3(const StateTime& cur, Environment* env, bool last);


    //   /**
    //    *  \brief compute the total forces
    //    *  \param prev : the previous node of the trajectory
    //    *  \param cur : the current node of the trajectory
    //    *  \param next : the next node of the trajectory
    //    *  \param env : the environment of the system
    //    *  \param time : the current time
    //    *  \param last : true if the current node is the last node of the trajectory
    //    */
    //   StateTime computeForces(const StateTime& prev, const StateTime& cur, const StateTime& next, Environment* env, double time, bool last);

    //   /**
    //    *  \brief apply the internal forces on the current node
    //    *  \param i : index of the current node on the trajectory
    //    *  \param prev : the previous node of the trajectory
    //    *  \param cur : the current node of the trajectory
    //    *  \param next : the next node of the trajectory
    //    */
    //   virtual void applyFInt(int i,const StateTime& prev, StateTime& cur, const StateTime& next) = 0;

    //   /**
    //    *  \brief compute the total external forces
    //    *  \param i : index of the current node on the trajectory
    //    *  \param cur : the current node of the trajectory
    //    *  \param env : the environment of the system
    //    *  \param time : the current time
    //    *  \param last : true if the current node is the last node of the trajectory
    //    */
    //   virtual void applyFExt(int i,StateTime& cur, Environment* env, double time, bool last) = 0;

    /**
     *  \brief compute the reliability coefficient
     *  \return the reliability coefficient
     *
     *  return the reliability coefficient to apply on the forces 
     *  knowing the difference dt between the current time and the 
     *  time of the node on which we apply the force
     */
    double getCoefReliability(double dt) const;


    /******* Connectivity and Collision checking ******/

  protected : 
    //   /**
    //    *  \brief check the connectivity between two nodes of the trajectory
    //    *  \param i : 
    //    *  \param oldCur : 
    //    *  \param prev : the previous node of the trajectory
    //    *  \param cur : the current node of the trajectory
    //    *  \param force : the resulting force to apply to restore the connectivity
    //    */
    //   virtual void checkConnectivity(int i, const StateTime& oldCur, const StateTime& prev, const StateTime& cur, StateTime& force) = 0;

    /**
     *  \brief check if the still-to-be executed part of the trajectory is in collision with the environment
     *  \param traj : the trajectory to test with
     *  \param env : the environment of the system
     *  \param curTime : the already elapsed time
     *  \return the first time from the current time where a collision occurs. return -1 if there is no collision
     */
    virtual double checkCollision(STraj* traj,Environment* env,double curTime);


    /******************* Resampling *******************/
  
  protected : 
    /**
     *  \brief resample the trajectory
     *  \param curStep : index of the node already reached
     */
    virtual void resample(int curStep) = 0;


    /*************** Deformation process **************/

  protected : 
    /**
     *  \brief init the deformation and check the temporal space between two consecutive states
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual void initDeform(double time,int curStep) = 0;

    /**
     *  \brief finish the deformation
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual int finishDeform(double time,int curStep) = 0;

  public : 
    /**
     *  \brief deform the current trajectory
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     *  \return 0 if the trajectory generation process succeed without any problem
     *  \return -1 if no input trajectory has been defined
     *  \return -2 if a collision may occur during the prediction horizon
     *  \return -3 if the connectivity of the trajectory has been lost
     */
    virtual int deform(double time, int curStep) = 0;

    /**
     *  \brief deform the current trajectory
     *  \param time : the current time
     *  \return 0 if the trajectory generation process succeed without any problem
     *  \return -1 if no input trajectory has been defined
     *  \return -2 if a collision may occur during the prediction horizon
     *  \return -3 if the connectivity of the trajectory has been lost
     *  \return -4 if the final time of the trajectory has been already past
     */
    virtual int deform(double time);

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the predictional horizon profile on interval [0;dt]
     *  \param dt : the length of time interval to display
     */
    void displayPredHorizonProfile(double dt) const;

    /**
     *  \brief display the external forces (from dynamic obstacles) in an openGL viewer
     */
    virtual void dispFExt() = 0;

    /**
     *  \brief display the external forces (from static obstacles) in an openGL viewer
     */
    virtual void dispStFExt() = 0;

    /**
     *  \brief display the internal forces in an openGL viewer
     */
    virtual void dispFInt() = 0;

    /**
     *  \brief display the nearest position-time of the dynamic obstacles from each state-time of the trajectory
     */
    virtual void dispNearObs();

    /**
     *  \brief display the nearest position-time of the static obstacles from each state-time of the trajectory
     */
    virtual void dispNearStObs();
  
    //   /**************** Running functions ***************/

    // public : 
    //   /**
    //    *  \brief start the deformation process
    //    */
    //   virtual void start();

    //   /**
    //    *  \brief stop the deformation process
    //    */
    //   virtual void stop();

    //   /**
    //    *  \brief reset the deformation process
    //    */
    //   virtual void reset();


    //   /************* Access/Update functions ************/

    // public : 
    //   /**
    //    *  \brief update the environment informations
    //    *  \param env : the new input informations about the environment
    //    */
    //   void updateEnv(Environment* env);

    //   /**
    //    *  \brief update the system localization and state informations
    //    *  \param s : the new input informations about the system state
    //    */
    //   void updateSysState(StateTime* s);

    // //   /**
    // //    *  \brief Request for a new control to apply (at a specified time)
    // //    *  \param t : time at which is evaluated the control
    // //    *  \param dt : duration of the application of the control
    // //    *  \param ctrl : resulting control
    // //    */
    // //   virtual void getControl(double t,double dt,Control& ctrl) = 0;
  };

};
#endif // MPC_TRAJDEFORM_HPP
