#ifndef MPC_WCTRAJDEFORM3_HPP
#define MPC_WCTRAJDEFORM3_HPP

#include "deformation/inc/TrajDeform.hpp"
#include "motion/inc/STrajectory.hpp"
#include "system/inc/WheelChairSys.hpp"
#include "trajGen/inc/WCTrajGen.hpp"
//#include "WCTrajGen2.hpp"

#include <list>

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class WCTrajDeform3 **************/
                                    /**************************************************/

  /** \class WCTrajDeform3
   *  \brief Trajectory deformer TEDDY for a Wheelchair system
   */
  class WCTrajDeform3 : public TrajDeform
  {
  protected :
    WCStateT   *tableFInt_;                            /**< table of the internal forces */
    WCStateT   *tableFExt_;                            /**< table of the external forces (from dynamic obstacles) */
    WCStateT   *tableStFExt_;                          /**< table of the external forces (from static obstacles) */

    double maxResidual_;                               /**< the maximal residual allowed to stop the trajectory generation process */
//     double maxResidualAccepted_;                       /**< the maximal residual until which the trajectory generated is kept */
    int maxTGnbSteps_;                                 /**< the maximal number of steps allowed to the trajectory generator to find a solution */

    int lastDeformedNode_;                             /**< index of the last node on which forces are applied at the current step */
    double lastTimeShift_;                             /**< time shift applied on the last deformed node at the current step */

    /********** Constructors and destructors **********/
  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    WCTrajDeform3();

  public : 
    /**
     *  \brief init constructor
     *  \param traj : the initial trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \param sys : the features of the robot system used
     *  \param dbg : the debug level
     */
    WCTrajDeform3(WCSTraj* traj,Environment* env,WheelChairSys* sys,int dbg=0);

    /**
     *  \brief destructor
     */
    virtual ~WCTrajDeform3();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline WCSTraj* traj() const { return ((WCSTraj*)(this->traj_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline WheelChairSys* sys() const { return ((WheelChairSys*)(this->sys_)); }

    /**
     *  \brief set the maximal residual allowed to stop the trajectory generation process
     *  \param maxResd : the new maximal residual allowed
     */
    inline void setMaxResidual(double maxResd) { this->maxResidual_ = maxResd; }

//     /**
//      *  \brief set the maximal residual until which the trajectory generated is kept
//      *  \param maxResdA : the new maximal residual until which the trajectory generated is kept
//      */
//     inline void setMaxResidualAccepted(double maxResdA) { this->maxResidualAccepted_ = maxResdA; }

    /**
     *  \brief set the maximal number of steps allowed to the trajectory generator to find a solution
     *  \param maxNbS : the new maximal number of steps allowed to the trajectory generator to find a solution
     */
    inline void setMaxTGnbSteps(int maxNbS) { this->maxTGnbSteps_ = maxNbS; }

    /*************** Forces computation ***************/

  protected : 
    /**
     *  \brief computing of the internal forces
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     *  \param nextIsLast : true if next is the last node of the trajectory
     */
    WCStateT computeFInt(const WCStateT& prev, const WCStateT& cur, const WCStateT& next, bool nextIsLast);

    /**
     *  \brief compute the total external forces
     *  \param i : index of the current node on the trajectory
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     *  \param timeEnv : the current time of the environment
     */
    virtual PositionXYT computeFExt(int i,const StateTime& cur, Environment* env, double time, bool last,double timeEnv=-1);

    /**
     *  \brief apply the internal forces on the current node
     *  \param i : index of the current node on the trajectory
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     *  \param time : the current time
     */
    virtual void applyFInt(int i,const WCStateT& prev, WCStateT& cur, const WCStateT& next, double time);

    /**
     *  \brief apply the internal forces on the current node
     *  \param i : index of the current node on the trajectory
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param traj : the deformed trajectory
     *  \param time : the current time
     */
    virtual void applyFInt(int i,const WCStateT& prev, WCStateT& cur, const WCSTraj* traj, double time);

    /**
     *  \brief compute the total external forces
     *  \param i : index of the current node on the trajectory
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param timeEnv : the current time of the environment
     */
    virtual void applyFExt(int i,WCStateT& cur, Environment* env, double time,double timeEnv=-1);


    /******* Connectivity and Collision checking ******/

  protected : 
    /**
     *  \brief check the connectivity of the trajectory
     *  \param curStep : index of the node already reached
     *  \return the time to the first disconnection found (-1 if no disconnection has been found)
     */
    double checkConnectivity(int curStep);
  
    /**
     *  \brief adjust the time of the nodes of the trajectory to maintain its consistency
     *  \param curStep : index of the node already reached
     */
    void adjustStateTimes(double time,int curStep);


    /******************* Resampling *******************/
  
  protected : 
    /**
     *  \brief resample the trajectory
     *  \param curStep : index of the node already reached
     */
    virtual void resample(int curStep);

    /**
     *  \brief smooth the trajectory
     *  \param time : the current time
     *  \param curStep : index of the node already reached
     */
    virtual void smoothTraj(double time,int curStep);

    /*************** Deformation process **************/

  protected : 
    /**
     *  \brief init the deformation and check the temporal space between two consecutive states
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual void initDeform(double time,int curStep);
  
    /**
     *  \brief finish the deformation
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual int finishDeform(double time,int curStep);

  public : 
    /**
     *  \brief deform the current trajectory
     *  \param timeTraj : the current time (reached on the trajectory)
     *  \param curStep : index of the node of the trajectory that has been already reached
     *  \param timeEnv : the current time of the environment
     *  \return 0 if the trajectory generation process succeed without any problem
     *  \return -1 if no input trajectory has been defined
     *  \return -2 if a collision may occur during the prediction horizon
     *  \return -3 if the connectivity of the trajectory has been lost
     */
    inline virtual int deform(double timeTraj, int curStep) { return this->deform(timeTraj,curStep,-1); }

    /**
     *  \brief deform the current trajectory
     *  \param timeTraj : the current time (reached on the trajectory)
     *  \param curStep : index of the node of the trajectory that has been already reached
     *  \param timeEnv : the current time of the environment
     *  \return 0 if the trajectory generation process succeed without any problem
     *  \return -1 if no input trajectory has been defined
     *  \return -2 if a collision may occur during the prediction horizon
     *  \return -3 if the connectivity of the trajectory has been lost
     */
    virtual int deform(double timeTraj, int curStep,double timeEnv=-1);

    /**
     *  \brief deform the current trajectory
     *  \param time : the current time
     *  \return 0 if the trajectory generation process succeed without any problem
     *  \return -1 if no input trajectory has been defined
     *  \return -2 if a collision may occur during the prediction horizon
     *  \return -3 if the connectivity of the trajectory has been lost
     *  \return -4 if the final time of the trajectory has been already past
     */
    inline virtual int deform(double time) { return this->TrajDeform::deform(time); }

    /**
     *  \brief deform the current trajectory
     *  \param timeTraj : the current time (reached on the trajectory)
     *  \param timeEnv : the current time of the environment
     *  \return 0 if the trajectory generation process succeed without any problem
     *  \return -1 if no input trajectory has been defined
     *  \return -2 if a collision may occur during the prediction horizon
     *  \return -3 if the connectivity of the trajectory has been lost
     *  \return -4 if the final time of the trajectory has been already past
     */
    int deform(double timeTraj,double timeEnv);

    /**************** Trajectory modifications ***************/

  public : 
    /**
     *  \brief return the state of the deformed trajectory at time t
     *  \param t : the input time
     *  \return the state of the deformed trajectory at time t
     */
    WCStateT getStateTraj(double t) const;
  
  protected : 
//     /**
//      *  \brief insert a state in the deformed trajectory
//      *  \param s : the state to insert
//      *  \param pos : its position in the trajectory's nodes vector
//      *  \param param : the vector of parameters for the trajectory generation used to reach the new state
//      *  \param fea : the feasibility status of the new inserted node
//      */
//     void insertStateInTraj(const WCStateT& s,int pos,double* param,int fea);

    /**
     *  \brief insert a state in the deformed trajectory
     *  \param ind : the place to insert the new node
     */
    void insertStateInTraj(int ind);

    /**
     *  \brief delete one state-time from the trajectory
     *  \param ind : the node to delete
     */
    void deleteStateFromTraj(int ind);

    /**
     *  \brief delete one or several states from the trajectory including start but not including end
     *  \param start : the first node to delete
     *  \param end : the node after the last node to delete
     */
    void deleteStateFromTraj(int start,int end);


    /******************* Display functions *******************/
  
  public : 
    /**
     *  \brief display the refined trajectory in an openGL viewer
     */
    void displayTraj() const;

    /**
     *  \brief display the external forces (from dynamic obstacles) in an openGL viewer
     */
    virtual void dispFExt();

    /**
     *  \brief display the external forces (from static obstacles) in an openGL viewer
     */
    virtual void dispStFExt();

    /**
     *  \brief display the internal forces in an openGL viewer
     */
    virtual void dispFInt();
  };

};
#endif // MPC_WCTRAJDEFORM3_HPP
