#include "WCTrajDeform3.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

#define COEF_SMOOTH 0.5
#define DT_FINT 1.5

                                    /**************************************************/
                                    /*************** Class WCTrajDeform3 **************/
                                    /**************************************************/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
WCTrajDeform3::WCTrajDeform3() : TrajDeform()
{
  tableFInt_ = NULL;
  tableFExt_ = NULL;
  tableStFExt_ = NULL;

  maxResidual_ = 0.005;
  maxTGnbSteps_ = 20;

  lastDeformedNode_ = -1;
  lastTimeShift_ = 0.0;
}

/**
 *  \brief init constructor
 *  \param traj : the initial trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \param sys : the features of the robot system used
 *  \param dbg : the debug level
 */
WCTrajDeform3::WCTrajDeform3(WCSTraj* traj,Environment* env,WheelChairSys* sys,int dbg) : TrajDeform(traj,env,sys)
{
  tableFInt_ = NULL;
  tableFExt_ = NULL;
  tableStFExt_ = NULL;

  this->setDebug(dbg);

  maxResidual_ = 0.005;
  maxTGnbSteps_ = 20;

  lastDeformedNode_ = -1;
  lastTimeShift_ = 0.0;
}

/**
 *  \brief destructor
 */
WCTrajDeform3::~WCTrajDeform3()
{
//   if(tableFInt_ != NULL) delete [] tableFInt_;
//   if(tableFExt_ != NULL) delete [] tableFExt_;
//   if(tableStFExt_ != NULL) delete [] tableStFExt_;
}


/*************** Forces computation ***************/

/**
 *  \brief computing of the internal forces
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param next : the next node of the trajectory
 *  \param nextIsLast : true if next is the last node of the trajectory
 */
WCStateT WCTrajDeform3::computeFInt(const WCStateT& prev, const WCStateT& cur, const WCStateT& next, bool nextIsLast)
{
  WCTrajGen trajgen(sys());
  trajgen.fixFinalXY(true);
  trajgen.fixFinalTH(true);
//   trajgen.fixFinalV(nextIsLast);
//   trajgen.fixFinalOM(nextIsLast);
  trajgen.fixFinalV(false);
  trajgen.fixFinalOM(false);

  trajgen.setUseVelCst(true);
  trajgen.setUseAccCst(true);
  trajgen.setUseAVelCst(true);
  trajgen.setUseAAccCst(false);

//   cout << "computeFInt -> prev:"; prev.display(); 
//   cout << "  cur:"; cur.display(); 
//   cout << "  next:"; next.display(); 
//   cout << "  last:" << nextIsLast << endl;

  TrajGen::setMaxResidual(maxResidual_);
  TrajGen::setMaxNbStep(maxTGnbSteps_);

  int res;
  WCStateT stateAtCurT;
  try
    {
      res = trajgen.genTraj(prev,next);
      stateAtCurT = trajgen.getWCStateTReachedAt(cur.t() - prev.t());
    }
  catch(int val)
    {
      cerr << "WCTrajDeform3::computeFInt -> exception val : " << val << endl;
      cerr << "prev : "; prev.display(); cerr << endl;
      cerr << "cur : "; cur.display(); cerr << endl;
      cerr << "next : "; next.display(); cerr << endl;
      cerr << "nextIsLast : " << nextIsLast << endl;
//       exit(1);
    }

  WCStateT sres(stateAtCurT.x() - cur.x(),
		stateAtCurT.y() - cur.y(),
		computeMinDiffAngle(stateAtCurT.th(),cur.th()),
		stateAtCurT.s() - cur.s(),
		stateAtCurT.om() - cur.om(),
		0.0);

  return sres;
}

/**
 *  \brief compute the total external forces
 *  \param i : index of the current node on the trajectory
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 *  \param timeEnv : the current time of the environment
 */
PositionXYT WCTrajDeform3::computeFExt(int i,const StateTime& cur, Environment* env, double time, bool last,double timeEnv)
{
  PositionXYT dynFExt = this->computeDynFExt2(cur,env,last,time,timeEnv);
  PositionXYT stFExt = this->computeStFExt3(cur,env,last);
  PositionXYT FExt = (dynFExt * (K_EXT)) + (stFExt * (K_STEXT));

  if(debug_ >= 1)
    {
      // save the force exterted from dynamic obstacles for display
      tableFExt_[i] = WCStateT(3.0*dynFExt.x(),3.0*dynFExt.y(),0,0,0,dynFExt.t());
      // save the force exterted from static obstacles for display
      tableStFExt_[i] = WCStateT(stFExt.x(),stFExt.y(),0,0,0,stFExt.t());
    }

//   cout << "dynFExt : "; dynFExt.display();
//   cout << "  stFExt : "; stFExt.display();
//   cout << "  FExt : "; FExt.display();
//   cout << "  K_EXT : " << K_EXT << "  K_STEXT : " << K_STEXT;
//   cout << "  spMed : " << spMed;
//   cout << "  tmpMed : " << tmpMed;
//   cout << "  maxStEffectDist : " << maxStEffectDist;
//   cout << endl;

  return FExt;
}

/**
 *  \brief apply the internal forces on the current node
 *  \param i : index of the current node on the trajectory
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param next : the next node of the trajectory
 *  \param time : the current time
 */
void WCTrajDeform3::applyFInt(int i,const WCStateT& prev, WCStateT& cur, const WCStateT& next, double time)
{
  // check if the current node is the last state-time of the trajectory
  bool nextIsLast;
  const WCStateT& lastNode = this->traj()->lastST();
  if(lastNode.x() == next.x() && lastNode.y() == next.y()) nextIsLast = true;
  else nextIsLast = false;  

  // compute the internal force
  WCStateT force = this->computeFInt(prev,cur,next,nextIsLast);

  // compute the coefficient to weight the force
  double dt = cur.t() - time;
  double coefRT = this->getCoefReliability(dt);
  force = force * coefRT;
  
  // apply the force on the current node
//   cur = cur + force;
  cur.x() += force.x();
  cur.y() += force.y();
  cur.th() = normalizedAngle(cur.th() + force.th());
  cur.s() += force.s();
  cur.om() += force.om();
  cur.t() += force.t();

  if(debug_ >= 1)
    {
      // save the force applied for display
      tableFInt_[i] = force;
    }
}

/**
 *  \brief apply the internal forces on the current node
 *  \param i : index of the current node on the trajectory
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param traj : the deformed trajectory
 *  \param time : the current time
 */
void WCTrajDeform3::applyFInt(int i,const WCStateT& prev, WCStateT& cur, const WCSTraj* traj, double time)
{
  // get the state of the trajectory at time curT + DT_FINT
  WCStateT next;
  traj->getStateAtTime(cur.t() + DT_FINT,next);

  this->applyFInt(i,prev,cur,next,time);
}

/**
 *  \brief compute the total external forces
 *  \param i : index of the current node on the trajectory
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param timeEnv : the current time of the environment
 */
void WCTrajDeform3::applyFExt(int i,WCStateT& cur, Environment* env, double time,double timeEnv)
{
  bool last;
  if(i == this->traj()->getSize()-1) last = true;
  else last = false;
  
  // compute the external force
  PositionXYT pforce = this->computeFExt(i,cur,env,time,last,timeEnv);
  WCStateT force(pforce.x(),pforce.y(),0,0,0,pforce.t());
  
  // apply it on the current node
  cur = cur + force;

  // save the time shift if the current node is the last deformed node
  if(i == lastDeformedNode_) lastTimeShift_ += force.t();
}


/******* Connectivity and Collision checking ******/

/**
 *  \brief check the connectivity of the trajectory
 *  \param curStep : index of the node already reached
 *  \return the time to the first disconnection found (-1 if no disconnection has been found)
 */
double WCTrajDeform3::checkConnectivity(int curStep)
{
}
  
/**
 *  \brief adjust the time of the nodes of the trajectory to maintain its consistency
 *  \param curStep : index of the node already reached
 */
void WCTrajDeform3::adjustStateTimes(double time,int curStep)
{
  double vmax = sys()->getVMax();

//   cout << "------------------------ adjustStateTimes ------------------------" << endl;
//   sys()->display();

  for(int i=curStep+1; i<this->traj()->getSize()-1 && this->traj()->at(i).t() < time + reliabilityTime; i++)
    {
      WCStateT prev = this->traj()->at(i-1);
      WCStateT cur = this->traj()->at(i);
      WCStateT next = this->traj()->at(i+1);

//       cout << "i:" << i << "  prev:"; prev.display();
//       cout << "  cur:"; cur.display();
//       cout << "  next:"; cur.display();      

      // compute the distance from prev to next
      double dx = next.x() - prev.x();
      double dy = next.y() - prev.y();

      // compute the average heading
      double thmoy = atan2(dy,dx);
      (*(this->traj()))[i].th() = thmoy;

      // compute the distance from prev to cur
      double dx2 = cur.x() - prev.x();
      double dy2 = cur.y() - prev.y();

      double dist = sqrt(dx2*dx2 + dy2*dy2);

      // compute the average speed along this path
      double dt = cur.t() - prev.t();
      double vmoy = dist / dt;

      if(dt < 0 || vmoy > vmax) 
	{
	  // save the time shift if the current node is the last deformed node
	  if(i == lastDeformedNode_) lastTimeShift_ = prev.t() + (dist / vmax) - (*(this->traj()))[i].t();

	  (*(this->traj()))[i].s() = vmax;
	  (*(this->traj()))[i].t() = prev.t() + (dist / vmax);
	}
      else 
	{
	  (*(this->traj()))[i].s() = vmoy;
	}
    }

  // adjust the time of the non-deformed part of the trajectory
  for(int i=lastDeformedNode_+1; i<this->traj()->getSize(); i++)
    {
      (*(this->traj()))[i].t() += lastTimeShift_;
    }

  // adjust the time of the final node of the trajectory
  WCStateT prevLast = this->traj()->at(this->traj()->getSize()-2);
  WCStateT curLast = this->traj()->at(this->traj()->getSize()-1);
  if(curLast.t() <= time + reliabilityTime)
    {
      // compute the distance from prev to next
      double dx = curLast.x() - prevLast.x();
      double dy = curLast.y() - prevLast.y();
      double dist = sqrt(dx*dx + dy*dy);

      // compute the average speed along this path
      double dt = curLast.t() - prevLast.t();
      double vmoy = dist / dt;

      double tres;
      double dres;
      minTDtoReachSpd(prevLast.s(),curLast.s(),sys()->getAMax(),tres,dres);

      if(dt < tres)
	{
	  (*(this->traj()))[this->traj()->getSize()-1].t() = prevLast.t() + tres;
	}
    }

//   cout << "---------------------- fin adjustStateTimes ----------------------" << endl;
}


/******************* Resampling *******************/
  
/**
 *  \brief resample the trajectory 
 *  \param curStep : index of the node already reached
 */
void WCTrajDeform3::resample(int curStep)
{
  double minTime = 0.5*this->timeStep_;  
  double maxTime = 2*this->timeStep_;

  int i=curStep;
  while(i < this->traj_->getSize() - 1)
    {
      // get back the current state
      WCStateT curS  = this->traj()->at(i);

      // check if the temporal space is too short
      bool tooShort = false;
      if((i < this->traj()->getSize() - 2))
	{
	  WCStateT nnextS = this->traj()->at(i+2);
	  double dt2 = nnextS.t() - curS.t();
	  
	  if(dt2 < minTime)
	    {
	      this->deleteStateFromTraj(i+1,i+2);
	      tooShort = true;
	    }
	}
      
      // check if the spatial or temporal space is too large
      bool tooLarge = false;
      if(!tooShort)
	{
	  WCStateT nextS = this->traj()->at(i+1);
	  double dt1 = nextS.t() - curS.t();

	  if(dt1 > maxTime)
	    {
	      this->insertStateInTraj(i+1);
	      tooLarge = true;
	    }
	}

      if((!tooShort) && (!tooLarge)) i++;
    }
}

/**
 *  \brief smooth the trajectory
 *  \param time : the current time
 *  \param curStep : index of the node already reached
 */
void WCTrajDeform3::smoothTraj(double time,int curStep)
{
  for(int i=curStep+1; i<this->traj()->getSize()-1 && this->traj()->at(i).t() < time + (reliabilityTime); i++)
    {
      WCStateT prev = this->traj()->at(i-1);
      WCStateT cur = this->traj()->at(i);
      WCStateT next = this->traj()->at(i+1);

      WCStateT sBary((prev.x() + next.x())*0.5, 
		     (prev.y() + next.y())*0.5,
		     normalizedAngle(prev.th() + 0.5*computeMinDiffAngle(next.th(),prev.th())), 
		     (prev.s() + next.s())*0.5,
		     (prev.om() + next.om())*0.5,
		     (prev.t() + next.t())*0.5);
      
      WCStateT sMid(cur.x()*(1.0-COEF_SMOOTH) + sBary.x()*COEF_SMOOTH, 
		    cur.y()*(1.0-COEF_SMOOTH) + sBary.y()*COEF_SMOOTH,
		    normalizedAngle(cur.th() + COEF_SMOOTH*computeMinDiffAngle(sBary.th(),cur.th())), 
		    cur.s()*(1.0-COEF_SMOOTH) + sBary.s()*COEF_SMOOTH,
		    cur.om()*(1.0-COEF_SMOOTH) + sBary.om()*COEF_SMOOTH,
		    cur.t()*(1.0-COEF_SMOOTH) + sBary.t()*COEF_SMOOTH);

      // save the time shift if the current node is the last deformed node
      if(i == lastDeformedNode_) lastTimeShift_ += (sMid.t() - (*(this->traj()))[i].t());
      
      (*(this->traj()))[i].x() = sMid.x();
      (*(this->traj()))[i].y() = sMid.y();
      (*(this->traj()))[i].th() = sMid.th();
      (*(this->traj()))[i].s() = sMid.s();
      (*(this->traj()))[i].om() = sMid.om();
      (*(this->traj()))[i].t() = sMid.t();
    }
}


/*************** Deformation process **************/

/**
 *  \brief init the deformation and check the temporal space between two consecutive states
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
void WCTrajDeform3::initDeform(double time,int curStep)
{
//   // resample the trajectory if needed
//   this->resample(curStep);

  // clear the tables of forces, states types and nearest states
  if(debug_ >= 1)
    {
//       cout << "TrajDeform::computeDynFExt : waiting for mutex Tables" << endl;
      boost::mutex::scoped_lock lock(mutexTables_);
//       cout << "TrajDeform::computeDynFExt : mutex Tables locked" << endl;

      if(tableNearState_ != NULL) delete [] tableNearState_;
      if(tableNearStState_ != NULL) delete [] tableNearStState_;
      if(tableFInt_ != NULL) delete [] tableFInt_;
      if(tableFExt_ != NULL) delete [] tableFExt_;
      if(tableStFExt_ != NULL) delete [] tableStFExt_;

      tableNearState_ = (new PositionXYT* [this->traj()->getSize()]);
      for(int i=0; i<this->traj()->getSize(); i++) tableNearState_[i] = NULL;
      tableNearStState_ = (new PositionXYT* [this->traj()->getSize()]);
      for(int i=0; i<this->traj()->getSize(); i++) tableNearStState_[i] = NULL;

      tableFInt_ = new WCStateT [this->traj()->getSize()];  
      tableFExt_ = new WCStateT [this->traj()->getSize()];
      tableStFExt_ = new WCStateT [this->traj()->getSize()];
    }

  // search for the last deformed node at the current step
  lastDeformedNode_ = -1;
  int i=curStep+1;
  while(lastDeformedNode_ == -1)
    {
      if(i >= this->traj()->getSize() || this->traj()->at(i).t() >= time + reliabilityTime) lastDeformedNode_ = i-1;
      else i++;
    }

  lastTimeShift_ = 0.0;
}
  
/**
 *  \brief finish the deformation
 *  \param time : the current time
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
int WCTrajDeform3::finishDeform(double time,int curStep)
{
  return 0;
}

/**
 *  \brief deform the current trajectory
 *  \param time : the current time (reached on the trajectory)
 *  \param curStep : index of the node of the trajectory that has been already reached
 *  \param timeEnv : the current time of the environment
 *  \return 0 if the trajectory generation process succeed without any problem
 *  \return -1 if no input trajectory has been defined
 *  \return -2 if a collision may occur during the prediction horizon
 *  \return -3 if the connectivity of the trajectory has been lost
 */
int WCTrajDeform3::deform(double time, int curStep,double timeEnv)
{
//   cout << "///////////////////////////////////////////////////////////////////////////" << endl;
//   cout << "///////////////////////////////////////////////////////////////////////////" << endl;

//   cout << "traj to deform -> curStep : " << curStep << endl;
//   traj()->display(); 
//   cout << "///////////////////////////////////////////////////////////////////////////" << endl;

//   cout << "deform -> time: " << time << "  curStep: " << curStep << endl;
  
  if(timeEnv == -1) timeEnv = time;

  // check if the trajectory has been defined
  if(this->traj_ == NULL)
    {
      // if the trajectory hasn't been computed, the deformation must be stoped (this cas should'nt appear)
      return -1;
    }

  double maxSpd = this->sys()->getVMax();
  double elapsedTime = time - this->oldTime_;

  // init the deformation
  this->initDeform(time,curStep);

//   cout << "traj (depart) : " << endl;
//   this->traj()->display();
//   cout << endl;

  // apply the repulsive forces from the obstacles
  for(int i=curStep+1; i<this->traj()->getSize() && this->traj()->at(i).t() < time + (reliabilityTime); i++)
    {
      I = i;

      // compute and apply the external force on the current node
      this->applyFExt(i,(*(this->traj()))[i],env_,time,timeEnv);
    }

  // smooth the trajectory
  this->smoothTraj(time,curStep);

  // adjust the trajectory nodes to help the internal forces computation
  this->adjustStateTimes(time,curStep);

  // apply the internal forces (connectivity maintenance)
  for(int i=curStep+1; i<this->traj()->getSize()-1 && traj()->at(i).t() < time + (reliabilityTime); i++)
    {
      this->applyFInt(i,(*(this->traj()))[i-1],(*(this->traj()))[i],this->traj(),time);
    }

  // check for collisions on the trajectory
  double timeCol = this->checkCollision(traj_,env_,time);
  
  // check for connectivity loss on the trajectory
  double timeDisco = this->checkConnectivity(curStep);
  
  // finish the deformation
  return this->finishDeform(time,curStep);
}

/**
 *  \brief deform the current trajectory
 *  \param timeTraj : the current time (reached on the trajectory)
 *  \param timeEnv : the current time of the environment
 *  \return 0 if the trajectory generation process succeed without any problem
 *  \return -1 if no input trajectory has been defined
 *  \return -2 if a collision may occur during the prediction horizon
 *  \return -3 if the connectivity of the trajectory has been lost
 *  \return -4 if the final time of the trajectory has been already past
 */
int WCTrajDeform3::deform(double timeTraj,double timeEnv)
{
  // check if the input trajectory hazs been defined
  if(traj_ == NULL) return -1;

  // search for the current step of the trajectory already reached
  bool trouve = false;
  int i=0;
  while(!trouve && i<traj_->size())
    {
      if(traj_->at(i).t() > timeTraj) trouve = true;
      else i++;
    }
  
  // check if the final time of the trajectory has been already reached
  if(!trouve)
    {
      return -4;
    }
  else 
    {
      return this->deform(timeTraj,i,timeEnv);
    }
}


/**************** Trajectory modifications ***************/

/**
 *  \brief return the state of the deformed trajectory at time t
 *  \param t : the input time
 *  \return the state of the deformed trajectory at time t
 */
WCStateT WCTrajDeform3::getStateTraj(double t) const
{
  // check if the time is defined over the trajectory
  if(t<0)
    {
      cout << "WCTrajDeform3::getStateTraj : Warning. TTTTTTIIIIIIMMMMMMMEEEE UUUUNNNNDDDDIIIFFFIIINNNEEEDDD!!!!!!" << endl;
      return WCStateT();
    }
  else if(t>this->traj()->at(traj()->getSize()-1).t())
    {
      return this->traj()->at(traj_->getSize()-1);
    }

  bool trouve = false;
  int i=1; 
  while(!trouve && i < this->traj()->getSize())
    {
      if(this->traj()->at(i).t() > t)
	{
	  trouve = true;
	  i--;
	}
      else
	{
	  i++;
	}
    }

  // get back the corresponding state
  WCStateT cur = this->traj()->at(i);
  WCStateT next = this->traj()->at(i+1);

  double coef = (t - cur.t()) / (next.t() - cur.t());
  WCStateT res(cur.x()*(1.0-coef)   + next.x()*coef,
	       cur.y()*(1.0-coef)   + next.y()*coef,
	       normalizedAngle(cur.th() + coef * computeMinDiffAngle(next.th(),cur.th())),
	       cur.s()*(1.0-coef) + next.s()*coef,
	       cur.om()*(1.0-coef) + next.om()*coef,
	       cur.t()*(1.0-coef)   + next.t()*coef);

//   cout << "cur: "; cur.display(); cout << "  next: "; next.display(); cout << "  coef: " << coef; cout << "res: "; res.display(); cout << endl;

  return res;
}
  
// /**
//  *  \brief insert a state in the deformed trajectory
//  *  \param s : the state to insert
//  *  \param pos : its position in the trajectory's nodes vector
//  *  \param param : the vector of parameters for the trajectory generation used to reach the new state
//  *  \param fea : the feasibility status of the new inserted node
//  */
// void WCTrajDeform3::insertStateInTraj(const WCStateT& s,int pos,double* param,int fea)
// {
// }

/**
 *  \brief insert a state in the deformed trajectory
 *  \param ind : the place to insert the new node
 */
void WCTrajDeform3::insertStateInTraj(int ind)
{
  // check if the trajectory has been defined
  if(traj_ == NULL) throw string("WCTrajDeform3::insertStateInTraj : No trajectory has been defined"); 

  // check if the input index is valid
  if(ind <= 0 || ind >= traj()->getSize()) throw string("WCTrajDeform3::insertStateInTraj : input index not valid");

  // get back the previous and the following nodes
  WCStateT prev = this->traj()->at(ind-1);
  WCStateT next = this->traj()->at(ind);

  PositionXY pos1(prev.x(),prev.y());
  PositionXY pos2(next.x(),next.y());
  PositionXY vec1,vec2;
  if(ind == 1)
    {
      vec1.x() = cos(prev.th());
      vec1.y() = sin(prev.th());
    }
  else
    {
      WCStateT pprev = this->traj()->at(ind-2);
      vec1.x() = prev.x() - pprev.x();
      vec1.y() = prev.y() - pprev.y();
    }

  if(ind == traj()->getSize()-1)
    {
      vec2.x() = -cos(next.th());
      vec2.y() = -sin(next.th());
    }
  else
    {
      WCStateT nnext = this->traj()->at(ind+1);
      vec2.x() = next.x() - nnext.x();
      vec2.y() = next.y() - nnext.y();
    }

  PositionXY itsc;
  bool nonColinear = intersecLines(pos1,vec1,pos2,vec2,itsc);

  // compute the node to insert in the trajectory
  WCStateT itd;
  if(nonColinear)
    {
      PositionXY mid(0.5*(prev.x() + next.x()),
		     0.5*(prev.y() + next.y()));

      
      itd.x() = 0.5*(mid.x() + itsc.x());
      itd.y() = 0.5*(mid.y() + itsc.y());
      itd.th() = atan2(next.y()-prev.y(),next.x()-prev.x());
      itd.s() = 0.5*(prev.s() + next.s());
      itd.om() = 0.5*(prev.om() + next.om());
      itd.t() = 0.5*(prev.t() + next.t());      
    }
  else
    {
      itd.x() = 0.5*(prev.x() + next.x());
      itd.y() = 0.5*(prev.y() + next.y());
      itd.th() = prev.th();
      itd.s() = 0.5*(prev.s() + next.s());
      itd.om() = 0.5*(prev.om() + next.om());
      itd.t() = 0.5*(prev.t() + next.t());
    }

  // insert it in the trajectory
  traj()->insert(ind, itd);
}

/**
 *  \brief delete one state-time from the trajectory
 *  \param ind : the node to delete
 */
void WCTrajDeform3::deleteStateFromTraj(int ind)
{
  // check if the trajectory has been defined
  if(traj_ == NULL) throw string("WCTrajDeform3::deleteStateFromTraj : No trajectory has been defined"); 

  // check if the input index is valid
  if(ind < 0 || ind >= traj()->getSize()) throw string("WCTrajDeform3::deleteStateFromTraj : input index not valid");

  // remove the state from the trajectory
  traj()->remove(ind);
}

/**
 *  \brief delete one or several states from the trajectory including start but not including end
 *  \param start : the first node to delete
 *  \param end : the node after the last node to delete
 */
void WCTrajDeform3::deleteStateFromTraj(int start,int end)
{
  // check if the trajectory has been defined
  if(traj_ == NULL) throw string("WCTrajDeform3::deleteStateFromTraj : No trajectory has been defined"); 

  // check if the input index is valid
  if(start < 0 || start >= end || end >= traj()->getSize()) throw string("WCTrajDeform3::deleteStateFromTraj : input index not valid");

  // remove states from the trajectory
  traj()->remove(start,end);
}


/******************* Display functions *******************/

/**
 *  \brief display the refined trajectory in an openGL viewer
 */
void WCTrajDeform3::displayTraj() const
{
}

/**
 *  \brief display the external forces (from dynamic obstacles) in an openGL viewer
 */
void WCTrajDeform3::dispFExt()
{
  // check if the table to store external forces is not empty
  if(tableFExt_ == NULL) throw string("WCTrajDeform3::dispFExt : tableFExt_ has not been defined");

  const GLfloat orangeMaterial[]={1.,1.,0.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, orangeMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, orangeMaterial);

  int N = traj()->size();

  for(int i=0; i<N; i++)
    {
      glBegin(GL_LINES);
      glVertex3f(traj()->at(i).x(), 
		 traj()->at(i).y(), 
		 traj()->at(i).t());
      glVertex3f(traj()->at(i).x() + 30 * tableFExt_[i].x(),
		 traj()->at(i).y() + 30 * tableFExt_[i].y(),
		 traj()->at(i).t() + 30 * tableFExt_[i].t());
      glEnd();
    }
}

/**
 *  \brief display the external forces (from static obstacles) in an openGL viewer
 */
void WCTrajDeform3::dispStFExt()
{
  // check if the table to store external forces is not empty
  if(tableStFExt_ == NULL) throw string("WCTrajDeform3::dispStFExt : tableStFExt_ has not been defined");

  const GLfloat cyanMaterial[]={0.,1.,1.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, cyanMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, cyanMaterial);

  int N = traj()->size();

  for(int i=0; i<N; i++)
    {
      glBegin(GL_LINES);
      glVertex3f(traj()->at(i).x(), 
		 traj()->at(i).y(), 
		 traj()->at(i).t());
      glVertex3f(traj()->at(i).x() + 30 * tableStFExt_[i].x(),
		 traj()->at(i).y() + 30 * tableStFExt_[i].y(),
		 traj()->at(i).t() + 30 * tableStFExt_[i].t());
      glEnd();
    }
}

/**
 *  \brief display the internal forces in an openGL viewer
 */
void WCTrajDeform3::dispFInt()
{
}
