#ifndef MAP2D_HPP
#define MAP2D_HPP

/************************************************************************************
 *            Map2d saving frame :                   Map2d displaying frame :       *
 *                                                                                  *
 *     (0,0)    -- x -->                                                            *
 *       +---------------------------              ----------------------------     *
 *       |                          |              |                          |     *
 *       |                          |              |                          |     *
 *       |                          |              |                          |     *
 *     | |                          |            A |                          |     *
 *     | |                          |            | |                          |     *
 *     y |                          |            | |                          |     *
 *     | |                          |            y |                          |     *
 *     | |                          |            | |                          |     *
 *     V |                          |            | |                          |     *
 *       |                          |              |                          |     *
 *       |                          |              |                          |     *
 *       |                          |              |                          |     *
 *       ----------------------------              +---------------------------     *
 *                                               (0,0)    -- x -->                  *
 *                                                                                  *
 ************************************************************************************/

#include <vector>
#include "graphics/inc/RGBImage.hpp"

namespace mpc
{

/**
 * \enum map2D_cell_t
 * \brief Enumeration of the different map2D cell types
 *
 * CELL_EMPTY  : the cell hasn't been alredy filled
 * CELL_FILLED : the cell has been alredy filled
 */
typedef enum {
  CELL_EMPTY   =  0,
  CELL_FILLED  =  1
} map2D_cell_t;

typedef struct MAP2D_INDEX
{
  int col_x;
  int line_y;
} Map2D_index;

                                    /**************************************************/
                                    /******************* Class Map2D ******************/
                                    /**************************************************/

  /** \class Map2D
   * \brief map of a 2D+ robot environment
   */
  class Map2D
  {
  protected : 
    int nbCols_;              /**< number of columns of the DEM table */
    int nbLines_;             /**< number of lines of the DEM table */
    
    float xOrigin_;           /**< X coordinate of the top left corner of the (0,0) pixel in the global frame */
    float yOrigin_;	      /**< Y coordinate of the top left corner of the (0,0) pixel in the global frame */
    
    float xScale_;            /**< X size of a cell in meters */
    float yScale_;            /**< Y size of a cell in meters */

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    Map2D();
    
    /**
     *  \brief copy constructor
     *  \param map : the map to copy
     */
    Map2D(const Map2D& map);

    /**
     *  \brief init constructor
     *  \param nbCols : number of columns of the map table
     *  \param nbLines : number of lines of the map table
     *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param xScale : X size of a cell in meters
     *  \param yScale : Z size of a cell in meters
     */
    Map2D(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale);

    /**
     *  \brief destructor
     */
    virtual ~Map2D();

    /**
     *  \brief return a copy of the current map
     *  \return a pointer to the copy of the current map
     *
     *  A new instance of a subclass of Map2D is allocated
     */
    virtual Map2D* copy() const = 0;


    /***************** Initialization *****************/

  protected : 
//     /*
//      *  \brief init tables from their dimesions
//      *  \param number of columns of the table
//      *  \param number of lines of the table
//      */
//     virtual void initTables(int nbCols,int nbLines) = 0;
    
//     /*
//      *  \brief delete all the content of the DEM tables
//      */
//     virtual void clearTables() = 0;


    /******************** Accessors *******************/

  public : 
    /*
     *  \brief get back the index of the input coord in the grid
     *  \param x : input absica
     *  \param y : input ordinate
     *  \param i : output number of the column
     *  \param j : output number of the line
     *  return 0 if the input coordinates are inside of the DEM grid, -1 else
     */
    int getIndexPos(float x,float y,int& i,int& j) const;

    /*
     *  \brief get back the index of the input coord if it belongs to the grid, or the closest one if it doesn't belong to it
     *  \param x : input absica
     *  \param y : input ordinate
     *  \param i : output number of the column
     *  \param j : output number of the line
     *  return 0 if the input coordinates are inside of the DEM grid, -1 else
     */
    int getClosestIndexToPos(float x,float y,int& i,int& j) const;

    /*
     *  \brief get back the index of the input coord in a grid of ppm image
     *  \param x : input absica
     *  \param y : input ordinate
     *  \param i : output number of the column
     *  \param j : output number of the line
     *  return 0 if the input coordinates are inside of the DEM grid, -1 else
     */
    int getIndexPPMImagePos(float x,float y,int& i,int& j) const;

    /*
     *  \brief get back the list of indices of the grid corresponding to the line defined between the two input points
     *  \param x1 : first point absica
     *  \param y1 : first point ordinate
     *  \param x2 : second point absica
     *  \param y2 : second point ordinate
     *  \param res : the resulting list of indices
     */
    void getIndicesLine(float x1, float y1, float x2 ,float y2, std::vector<Map2D_index>& res) const;

    /*
     *  \brief get back the index of the input coord in the grid, and the weights of each surrounding nodes
     *  \param x : input absica
     *  \param y : input ordinate
     *  \param i : output number of the column
     *  \param j : output number of the line
     *  \param wUL : weight of the upper left corner
     *  \param wUR : weight of the upper right corner
     *  \param wBL : weight of the upper lower left corner
     *  \param wBR : weight of the upper lower right corner
     *  return 0 if the input coordinates are inside of the DEM grid, -1 else
     */
    int getIndexPosAndWeights(float x,float y,int& i,int& j,float& wUL,float& wUR,float& wBL,float& wBR) const;

  public : 
    /*
     *  \brief get back the coord corresponding to the input index of the grid
     *  \param i : input number of the column
     *  \param j : input number of the line
     *  \param x : output absica
     *  \param y : output ordinate
     *  return 0 if the input coordinates are inside of the DEM grid, -1 else
     */
    int getPosIndex(int i,int j,float& x,float& y) const;

    /**
     *  \brief Return the number of columns of the DEM table
     *  \return the number of columns of the DEM table
     */
    int nbCols() const;

    /**
     *  \brief Return the number of lines of the DEM table
     *  \return the number of lines of the DEM table
     */
    int nbLines() const;
    
    /**
     *  \brief Return the X coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \return the X coordinate of the top left corner of the (0,0) pixel in the global frame
     */
    float xOrigin() const;

    /**
     *  \brief Return the Y coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \return the Y coordinate of the top left corner of the (0,0) pixel in the global frame
     */
    float yOrigin() const;
    
    /**
     *  \brief Return the X size of a cell in meters
     *  \return the X size of a cell in meters
     */
    float xScale() const;

    /**
     *  \brief Return the Y size of a cell in meters
     *  \return the Y size of a cell in meters
     */
    float yScale() const;
    
//     /**
//      *  \brief set the content of a cell as empty
//      *  \param i : the column number
//      *  \param j : the line number
//      */
//     virtual void setEmptyCell(int i, int j);

//     /**
//      *  \brief return true is the cell corresponding to theinput index is empty
//      *  \param i : the column number
//      *  \param j : the line number
//      *  \return true is the cell corresponding to theinput index is empty
//      *
//      *  The user must have checked if the input coord are valid
//      */
//     virtual bool isEmptyCell(int i, int j) const;
    

    /************ Input / Output functions ************/

  protected : 
    /*
     *  \brief init the Z color parameters
     *  \param minColZ : output minimal Z value used to convert the DEM into an image
     *  \param maxColZ : output maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : output Z scale value used to convert the DEM into an image
     */
    virtual void initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const = 0;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \return the colour corresponding to a cell of the DEM
     */
    virtual int greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const = 0;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \param R : the output red colour
     *  \param G : the output green colour
     *  \param B : the output blue colour
     */
    virtual void rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const = 0;

  public : 
    /**
     *  \brief create a rgb image from the 2D map
     *  \param minZ : minimal Z value used to convert the DEM into an image
     *  \param maxZ : maximal Z value used to convert the DEM into an image
     *  \return a new instance of rgb image
     */
    RGBImage* rgbImage(float minZ=0, float maxZ=0);
    
    /**
     *  \brief save the DEM into an output PGM file
     *  \param filename : the output filename
     *  \param minZ : minimal Z value used to convert the DEM into an image
     *  \param maxZ : maximal Z value used to convert the DEM into an image
     *  \return 0 if the process succeed, -1 else
     */
    int saveIntoPGM(const char* filename, float minZ=0, float maxZ=0) const;

    /**
     *  \brief save the DEM into an output PPM file
     *  \param filename : the output filename
     *  \param minZ : minimal Z value used to convert the DEM into an image
     *  \param maxZ : maximal Z value used to convert the DEM into an image
     *  \return 0 if the process succeed, -1 else
     */
    int saveIntoPPM(const char* filename, float minZ=0, float maxZ=0) const;


    /**************** Utility functions ***************/

  protected : 
    /**
     *  \brief extract a sub-map of the current map
     *  \param iMin : the number of the first column to keep
     *  \param iMax : the number of the last column to keep
     *  \param jMin : the number of the first line to keep
     *  \param jMax : the number of the last line to keep
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of DEM is created
     */
    virtual Map2D* subMap(int iMin, int iMax, int jMin, int jMax) const = 0;

  public : 
    /**
     *  \brief extract a sub-map of the current map
     *  \param xMin : minimal bound over the X axis
     *  \param xMax : maximal bound over the X axis
     *  \param yMin : minimal bound over the Y axis
     *  \param yMax : maximal bound over the Y axis
     *  \param forceSquareMap : force the resulting map to be square
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of Map2D is created
     */
    Map2D* subMap(float xMin, float xMax, float yMin, float yMax, bool forceSquareMap=false) const;

    /**
     *  \brief extract a sub-map of the current map
     *  \param (xCenter,yCenter) : coordinates of the center of the sub-map
     *  \param size : size of the sub-map
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of Map2D is created
     */
    Map2D* subMap(float xCenter, float yCenter, float size) const;

  public : 
    /**
     *  \brief display the main features of the DEM
     *  \param dispGrid : display the grid if true
     */
    virtual void display() const;
  };


  

};

#endif // MAP2D_HPP
