#include "environment/inc/BoolObsMap.hpp"
using namespace mpc;

#include <stdlib.h>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /***************** Class BoolObsMap ***************/
                                    /**************************************************/

float BoolObsMap::colDepth_ = 255;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
BoolObsMap::BoolObsMap() : Map2D()
{
  this->cellState_ = NULL;

  this->thresSlope_ = 0.25*M_PI;
  this->thresVarSlope_ = 5.0*M_PI;
  this->safeDist_ = 0.5;
  this->maxNbCells_ = 0;
}
    
/**
 *  \brief copy constructor
 *  \param map : the map to copy
 */
BoolObsMap::BoolObsMap(const BoolObsMap& map) : Map2D(map)
{
  this->initTables(map.nbCols_,map.nbLines_);

  this->thresSlope_ = map.thresSlope_;
  this->thresVarSlope_ = map.thresVarSlope_;
  this->safeDist_ = map.safeDist_;
  this->maxNbCells_ = map.maxNbCells_;

  // copy the tables
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	this->setCellType(i,j,map.getCellType(i,j));
      }
}

/**
 *  \brief init constructor
 *  \param nbCols : number of columns of the map table
 *  \param nbLines : number of lines of the map table
 *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param xScale : X size of a cell in meters
 *  \param yScale : Z size of a cell in meters
 */
BoolObsMap::BoolObsMap(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale, float thresSlope, float thresVarSlope, float safeDist) : Map2D(nbCols,nbLines,xOrigin,yOrigin,xScale,yScale)
{
  this->initTables(nbCols,nbLines);

//   this->thresSlope_ = 0.25*M_PI;
//   this->thresVarSlope_ = 5.0*M_PI;
//   this->safeDist_ = 0.5;
  this->thresSlope_ = thresSlope;
  this->thresVarSlope_ = thresVarSlope;
  this->safeDist_ = safeDist;
  this->maxNbCells_ = 0;

  // set all the cells as unknown
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	this->setCellType(i,j,CELL_UNKWN);
      }
}

/**
 *  \brief init constructor (from a data elevation map)
 *  \param dem : the input dem
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 */
BoolObsMap::BoolObsMap(const SlopeMap* slopemap, float thresSlope, float safeDist) : Map2D()
{
  this->cellState_ = NULL;

  this->buildFromSM(slopemap,thresSlope,safeDist);
}

/**
 *  \brief init constructor (from a data elevation map)
 *  \param dem : the input dem
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param thresVarSlope : the maximal slope derivative that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 */
BoolObsMap::BoolObsMap(const VarSlopeMap* varslopemap, float thresSlope, float thresVarSlope, float safeDist)
{
  this->cellState_ = NULL;

  this->buildFromVSM(varslopemap, thresSlope, thresVarSlope, safeDist);
}

/**
 *  \brief destructor
 */
BoolObsMap::~BoolObsMap()
{
  this->clearTables();
}

/**
 *  \brief return a copy of the current map
 *  \return a pointer to the copy of the current map
 *
 *  A new instance of BoolObsMap is allocated
 */
BoolObsMap* BoolObsMap::copy() const
{
  return new BoolObsMap(*this);
}


/***************** Initialization *****************/

/*
 *  \brief init tables from their dimesions
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void BoolObsMap::initTables(int nbCols,int nbLines)
{
  nbCols_ = nbCols;
  nbLines_ = nbLines;

  // Type of each cell
  cellState_ = new obs2D_cell_t* [nbCols];
  for(int i=0; i<nbCols; i++) cellState_[i] = new obs2D_cell_t [nbLines];
}

/*
 *  \brief delete all the content of the slope map tables
 */
void BoolObsMap::clearTables()
{
  if(cellState_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(cellState_[i] != NULL) delete [] cellState_[i];
      delete [] cellState_;
    }
  cellState_ = NULL;
}


/******************** Accessors *******************/

/*
 *  \brief set the cell type
 *  \param (i,j) : index of the cell in the grid
 *  \param type : the new cell type
 */
void BoolObsMap::setCellType(int i, int j,obs2D_cell_t type)
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    return;
  else 
    cellState_[i][j] = type;
}

/**
 *  \brief get the type of cell (i,j)
 *  \param (i,j) : index of the cell in the grid
 *  \return the cell type
 */
obs2D_cell_t BoolObsMap::getCellType(int i, int j) const
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    return CELL_UNKWN;
  else 
    return cellState_[i][j];
}

/**
 *  \brief get the type of cell (i,j)
 *  \param (x,y) : position of the cell in the grid
 *  \return the cell type
 */
obs2D_cell_t BoolObsMap::getCellType(float x, float y) const
{
  int i, j;
  int outside = this->getIndexPos(x, y, i, j);

  if(outside) // (outside == -1)
      return CELL_UNKWN;

  else return this->getCellType(i, j);
}


/************** Obstacle map building *************/

/*
 *  \brief thresholding the input slope map to obstain an obstacle map
 *  \param sm : the input slope map
 *  \param thresSlope : the maximal slope that the robot can climb
 */
void BoolObsMap::thresSM(const SlopeMap* sm, float thresSlope)
{
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	// get back informations from the slope map about the current cell
	map2D_cell_t cellSt;
	float slope;
	sm->getCellInfo(i, j, cellSt, slope);

	if(cellSt == CELL_EMPTY) this->setCellType(i, j, CELL_UNKWN);
	else if(slope <= thresSlope) this->setCellType(i, j, CELL_FREE);
	else this->setCellType(i, j, CELL_OBS);
      }
}

/*
 *  \brief thresholding the input var slope map to obstain an obstacle map
 *  \param vsm : the input var slope map
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param thresVarSlope : the maximal slope derivative that the robot can climb
 */
void BoolObsMap::thresVSM(const VarSlopeMap* vsm, float thresSlope, float thresVarSlope)
{
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	// get back informations from the slope map about the current cell
	map2D_cell_t cellSt;
	float slope;
	float varSlope;
	vsm->getCellInfo(i, j, cellSt, slope, varSlope);

	if(cellSt == CELL_EMPTY) this->setCellType(i, j, CELL_UNKWN);
	else if(slope <= thresSlope && varSlope <= thresVarSlope) this->setCellType(i, j, CELL_FREE);
	else this->setCellType(i, j, CELL_OBS);
      }
}

/*
 *  \brief increase the obstacles with the input safe distance
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 */
void BoolObsMap::increaseObs(float safeDist)
{
  this->initMaxNbCells(safeDist);

  // copy the table of cell types
  obs2D_cell_t** copyCellState = new obs2D_cell_t* [nbCols_];
  for(int i=0; i<nbCols_; i++) copyCellState[i] = new obs2D_cell_t [nbLines_];

  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	copyCellState[i][j] = this->getCellType(i,j);
      }

  // reinit the slope map tables
  this->clearTables();
  this->initTables(nbCols_,nbLines_);  

  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	bool found = false;

	// check for each cell surrounding the current one if it is an obstacle
	int k=i-maxNbCells_;
	while(!found && k<=i+maxNbCells_)
	  {
	    int l=j-maxNbCells_;
	    while(!found && l<=j+maxNbCells_)
	      {
		if(k < 0 || k >= nbCols_ || l < 0 || l >= nbLines_ ) l++;
		else if(copyCellState[k][l] == CELL_OBS) found = true;
		else l++;
	      }
	    if(!found) k++;
	  }

	if(found) this->setCellType(i, j, CELL_OBS);
	else
	  {
	    if(copyCellState[i][j] == CELL_FREE) this->setCellType(i, j, CELL_FREE);
	    else this->setCellType(i, j, CELL_UNKWN);
	  }
      }

  // delete the copy of cell types
  for(int i=0; i<nbCols_; i++) delete [] copyCellState[i];
  delete [] copyCellState;
}

/*
 *  \brief initialize the maximal number of cells considered to increase the obstacles
 *  \param maxDist : the maximal distance considered to increase the obstacles
 */
void BoolObsMap::initMaxNbCells(float safeDist)
{
  if(safeDist == 0) maxNbCells_ = 0;
  else
    {
      maxNbCells_ = (int)(safeDist / xScale_);
      if(maxNbCells_ == 0) maxNbCells_ = 1;
    }

//   cout << "maxNbCells_ : " << maxNbCells_ << endl;
}

/**
 *  \brief build the obstacle map from an input slope map
 *  \param sm : the input slope map
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 *  \return 0 if the process succeed, -1 else
 */
int BoolObsMap::buildFromSM(const SlopeMap* sm, float thresSlope, float safeDist)
{
  if(sm == NULL) return -1;

  // initialize the slope map parameters
  this->nbCols_ = sm->nbCols();
  this->nbLines_ = sm->nbLines();
  
  this->xOrigin_ = sm->xOrigin();
  this->yOrigin_ = sm->yOrigin();
  
  this->xScale_ = sm->xScale();
  this->yScale_ = sm->yScale();
  
  // initialize the obstacle map tables
  this->clearTables();
  this->initTables(sm->nbCols(),sm->nbLines());

  // initialize the slope threshold and the safe distance
  this->thresSlope_ = thresSlope;
  this->safeDist_ = safeDist;

  // thresholding the input slope map to obtain an obstacle map
  this->thresSM(sm, thresSlope);

  // increasing the obstacles with the input safe distance
  this->increaseObs(safeDist);

  return 0;
}

/**
 *  \brief build the obstacle map from an input slope map
 *  \param vsm : the input var slope map
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param thresVarSlope : the maximal slope derivative that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 *  \return 0 if the process succeed, -1 else
 */
int BoolObsMap::buildFromVSM(const VarSlopeMap* vsm, float thresSlope, float thresVarSlope, float safeDist)
{
  if(vsm == NULL) return -1;

  // initialize the slope map parameters
  this->nbCols_ = vsm->nbCols();
  this->nbLines_ = vsm->nbLines();
  
  this->xOrigin_ = vsm->xOrigin();
  this->yOrigin_ = vsm->yOrigin();
  
  this->xScale_ = vsm->xScale();
  this->yScale_ = vsm->yScale();
  
  // initialize the obstacle map tables
  this->clearTables();
  this->initTables(vsm->nbCols(),vsm->nbLines());

  // initialize the slope threshold and the safe distance
  this->thresSlope_ = thresSlope;
  this->thresVarSlope_ = thresVarSlope;
  this->safeDist_ = safeDist;

  // thresholding the input slope map to obtain an obstacle map
  this->thresVSM(vsm, thresSlope, thresVarSlope);

  // increasing the obstacles with the input safe distance
  this->increaseObs(safeDist);

  return 0;
}


/************ Input / Output functions ************/

/*
 *  \brief init the Z color parameters
 *  \param minColZ : output minimal Z value used to convert the DEM into an image
 *  \param maxColZ : output maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : output Z scale value used to convert the DEM into an image
 */
void BoolObsMap::initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const
{
  // nothing to do
  minColZ = 0;
  maxColZ = 0;
  scaleColZ = 0;
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \return the colour corresponding to a cell of the DEM
 */
int BoolObsMap::greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const
{
  obs2D_cell_t cellSt = this->getCellType(i,j);

  if(cellSt == CELL_UNKWN)
    return ((int)(0.5*colDepth_));
  else if(cellSt == CELL_FREE)
    return colDepth_;
  else return 0;
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \param R : the output red colour
 *  \param G : the output green colour
 *  \param B : the output blue colour
 */
void BoolObsMap::rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const
{
  obs2D_cell_t cellSt = this->getCellType(i,j);

  if(cellSt == CELL_UNKWN)
    {
      R = (int)(0.5*colDepth_);
      G = (int)(0.5*colDepth_);
      B = (int)(0.5*colDepth_);
    }
  else if(cellSt == CELL_FREE)
    {
      R = colDepth_;
      G = colDepth_;
      B = colDepth_;
    }
  else
    {
      R = 0;
      G = 0;
      B = 0;
    }
}


/**************** Utility functions ***************/

/**
 *  \brief extract a sub-map of the current map
 *  \param iMin : the number of the first column to keep
 *  \param iMax : the number of the last column to keep
 *  \param jMin : the number of the first line to keep
 *  \param jMax : the number of the last line to keep
 *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
 *
 *  A new instance of BoolObsMap is created
 */
BoolObsMap* BoolObsMap::subMap(int iMin, int iMax, int jMin, int jMax) const
{
  // check if the input parameters are valid
  if(iMax <= iMin || jMax <= jMin) return NULL;

  BoolObsMap* subMap = new BoolObsMap(iMax - iMin, 
				      jMax - jMin, 
				      this->xOrigin_ + iMin*this->xScale_, 
				      this->yOrigin_ + jMin*this->yScale_, 
				      this->xScale_, 
				      this->yScale_, 
				      this->thresSlope_, 
				      this->thresVarSlope_, 
				      this->safeDist_);
  
  for(int i=iMin; i<iMax; i++)
    for(int j=jMin; j<jMax; j++)
      {
	subMap->cellState_[i-iMin][j-jMin] = this->cellState_[i][j];
      }

  return subMap;
}

/**
 *  \brief display the main features of the DEM
 *  \param dispGrid : display the grid if true
 */
void BoolObsMap::display(bool dispGrid) const
{
  // display the main features
  cout << "nbLines : " << nbLines_ <<  "  nbCols : " << nbCols_ << endl;
  cout << "xOrigin : " << xOrigin_ <<  "  yOrigin : " << yOrigin_ << endl;
  cout << "xScale : " << xScale_ <<  "  yScale : " << yScale_ << endl;
  cout << endl;

  // display the grid if needed
  if(dispGrid)
    {
      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      obs2D_cell_t cellSt = this->getCellType(i,j);
	      
// 	      if(cellSt == CELL_UNKWN) cout << "? ";
// 	      else if(cellSt == CELL_FREE) cout << "F ";
// 	      else cout << "O ";

	      if(cellSt == CELL_UNKWN) cout << "? ";
	      else if(cellSt == CELL_FREE) cout << "- ";
	      else cout << "O ";
	    }
	  cout << endl;
	}
      cout << endl;
    }
}

/**
 *  \brief add a linear obstacle on the map from its end points
 *  \param (x1,y1) : coordinates of the first end point on the map
 *  \param (x2,y2) : coordinates of the second end point on the map
 */
void BoolObsMap::addLineObs(float x1, float y1, float x2, float y2)
{
  // get back indices of the grid corresponding to the input line
  std::vector<Map2D_index> res;
  this->getIndicesLine(x1, y1, x2, y2, res);

  // set each index found as an obstacle on the grid
  for(int k=0; k<res.size(); k++)
    {
      this->setCellType(res.at(k).col_x, res.at(k).line_y, CELL_OBS);
    }
}
