#include "DEM.hpp"
using namespace mpc;

#include <stdlib.h>
#include <iostream>
#include <fstream>
using namespace std;

#include "Utils.hpp"

                                    /**************************************************/
                                    /******************** Class DEM *******************/
                                    /**************************************************/

float DEM::colDepth_ = 255;


/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DEM::DEM() : Map2D()
{  
  zOrigin_ = 0.0;	   // Elevation of the '0' value
  zScale_ = 0.1;           // Scale to store the elevation
  
  z_ = NULL;               // Altitude of each cell
  nb_points_ = NULL;       // Number of points fused in the dem
//   sigma_z_ = NULL;         // Standard deviation
//   confidence_ = NULL;      // Confidence of the altitude computed in each cell
}
    
/**
 *  \brief copy constructor
 *  \param dem : the dem to copy
 */
DEM::DEM(const DEM& dem) : Map2D(dem)
{
  this->zOrigin_ = dem.zOrigin_;  
  this->zScale_ =  dem.zScale_;

  this->initTables(dem.nbCols_,dem.nbLines_);

  // copy the altitude and nbpoints tables
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	this->setCell(i, j, dem.z_[i][j], dem.nb_points_[i][j]);
      }
}

/**
 *  \brief init constructor
 *  \param nbCols : number of columns of the DEM table
 *  \param nbLines : number of lines of the DEM table
 *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param zOrigin : Elevation of the '0' value 
 *  \param xScale : X size of a cell in meters
 *  \param yScale : Z size of a cell in meters
 *  \param zScale : Scale to store the elevation
 */
DEM::DEM(int nbCols, int nbLines, float xOrigin, float yOrigin, float zOrigin, float xScale, float yScale, float zScale) : Map2D(nbCols, nbLines, xOrigin, yOrigin, xScale, yScale)
{
  this->zOrigin_ = zOrigin;
  this->zScale_ = zScale;

  this->initTables(nbCols,nbLines);
}

/**
 *  \brief destructor
 */
DEM::~DEM()
{
  this->clearTables();
}

/**
 *  \brief return a copy of the current map
 *  \return a pointer to the copy of the current map
 *
 *  A new instance of DEM is allocated
 */
DEM* DEM::copy() const
{
  return new DEM(*this);
}


/***************** Initialization *****************/

/*
 *  \brief init tables from their dimesions
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void DEM::initTables(int nbCols,int nbLines)
{
  nbCols_ = nbCols;            // number of columns of the DEM table
  nbLines_ = nbLines;          // number of lines of the DEM table

  // Altitude of each cell
  z_ = new float* [nbCols];
  for(int i=0; i<nbCols; i++) z_[i] = new float [nbLines];
  
  // Number of points fused in the dem
  nb_points_ = new int* [nbCols];
  for(int i=0; i<nbCols; i++) nb_points_[i] = new int [nbLines];

  for(int i=0; i<nbCols; i++) 
    for(int j=0; j<nbLines; j++) nb_points_[i][j] = 0;      

//   // Standard deviation
//   sigma_z_ = new float* [nbCols];
//   for(int i=0; i<nbCols; i++) sigma_z_[i] = new float [nbLines];

//   // Confidence of the altitude computed in each cell
//   confidence_ = new float* [nbCols];
//   for(int i=0; i<nbCols; i++) confidence_[i] = new float [nbLines];
}

/*
 *  \brief delete all the content of the DEM tables
 */
void DEM::clearTables()
{
  if(z_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(z_[i] != NULL) delete [] z_[i];
      delete [] z_;
    }
  
  if(nb_points_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(nb_points_[i] != NULL) delete [] nb_points_[i];
      delete [] nb_points_;
    }

//   if(sigma_z_ != NULL)
//     {
//       for(int i=0; i<nbCols_; i++) if(sigma_z_[i] != NULL) delete [] sigma_z_[i];
//       delete [] sigma_z_;
//     }

//   if(confidence_ != NULL)
//     {
//       for(int i=0; i<nbCols_; i++) if(confidence_[i] != NULL) delete [] confidence_[i];
//       delete [] confidence_;
//     }
}


/******************** Accessors *******************/

/**
 *  \brief get back the altitude of a cell of the DEM
 *  \param i : number of the column
 *  \param j : number of the line
 *  \return the altitude of a cell on the DEM
 */
float DEM::getZ(int i,int j) const
{
  return z_[i][j];
}

/**
 *  \brief get back the altitude of a position (x,y) on the DEM
 *  \param x : absica
 *  \param y : ordinate
 *  \return the altitude of a position (x,y) on the DEM
 *  \return -1 if the input position is outside of the grid 
 *             or if all altitudes of the cells of the DEM surrounding the input position are undefined
 */
float DEM::getZ(float x,float y) const
{
  // get back indexes of the grid corresponding to the input position
  int i,j;
  float wUL, wUR, wBL, wBR;
  int resGIP = this->getIndexPosAndWeights(x,y,i,j,wUL,wUR,wBL,wBR);

  if(resGIP == 0)
    {
      double sumAlt = 0;
      double sumWeights = 0;

      if(nb_points_[i][j] > 0)
	{
	  sumAlt += wUL * z_[i][j];
	  sumWeights += wUL;
	}
      if(nb_points_[i+1][j] > 0)
	{
	  sumAlt += wUR * z_[i+1][j];
	  sumWeights += wUR;
	}
      if(nb_points_[i][j+1] > 0)
	{
	  sumAlt += wBL * z_[i][j+1];
	  sumWeights += wBL;
	}
      if(nb_points_[i+1][j+1] > 0)
	{
	  sumAlt += wBR * z_[i+1][j+1];
	  sumWeights += wBR;
	}

      sumAlt /= sumWeights;
      return sumAlt;
    }  
  else return -1;
}

/**
 *  \brief get back the number of points seen a cell of the DEM
 *  \param i : number of the column
 *  \param j : number of the line
 *  \return the number of points seen in a cell of the DEM
 */
float DEM::getNbP(int i,int j) const
{
  return nb_points_[i][j];
}
    
/**
 *  \brief Return the Elevation of the '0' value
 *  \return the Elevation of the '0' value
 */
float DEM::zOrigin() const { return this->zOrigin_; }

/**
 *  \brief Return the Scale to store the elevation
 *  \return the Scale to store the elevation
 */
float DEM::zScale() const { return this->zScale_; }

/**
 *  \brief set the content of a non-empty cell of the grid
 *  \param i : the column number
 *  \param j : the line number
 *  \param z : the altitude value
 *  \param nbPoints : the number of points seen corresponding to the cell
 *  \param sigma_z : the standard deviation
 *  \param confidence : the confidence over the previous data
 */
void DEM::setCell(int i, int j, float z, int nbPoints, float sigma_z, float confidence)
{
  // check if the input coord are defined
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      cerr << "DEM::setCell : the input coord are outside of the grid!" << endl; 
      return;
    }

  this->z_[i][j] = z;
  this->nb_points_[i][j] = nbPoints;
//   this->sigma_z_[i][j] = sigma_z;
//   this->confidence_[i][j] = confidence;
}

/**
 *  \brief set the content of a cell as empty
 *  \param i : the column number
 *  \param j : the line number
 */
void DEM::setEmptyCell(int i, int j)
{
  this->setCell(i, j, 0.0, 0);
}

/**
 *  \brief return true is the cell corresponding to theinput index is empty
 *  \param i : the column number
 *  \param j : the line number
 *  \return true is the cell corresponding to theinput index is empty
 *
 *  The user must have checked if the input coord are valid
 */
bool DEM::isEmptyCell(int i, int j) const
{
  return (this->nb_points_[i][j] == 0);
}


/**************** DEM modifications ***************/

/**
 *  \brief add a new altitude value in a cell
 *  \param i : the abscica of the cell
 *  \param j : the ordinate of the cell
 *  \param z : the altitude value
 *  \param nbP : the number of points corresponding to the new value
 *  \return 0 if the process succeed, -1 if the input coordinate are out of the grid
 */
int DEM::addValueInCell(int i,int j,float z,int nbP)
{
  // check if the number of input points is not null
  if(nbP == 0)
    {
      // do nothing
      return 0;
    }

  // check if the input coordinate are in the grid
  if(i < 0 || i >= nbCols_ || j < 0 || j > nbLines_) return -1;
  else
    {
      if(nb_points_[i][j] == 0)
	{
	  this->setCell(i, j, z, nbP);

// 	  z_[i][j] = z;
// 	  nb_points_[i][j] = nbP;
	}
      else 
	{
	  float newZ = ((nb_points_[i][j]-1) * z_[i][j] + nbP * z) / (nb_points_[i][j] + nbP);
	  int newNbP = (nb_points_[i][j]) + nbP;

	  this->setCell(i, j, newZ, newNbP);

// 	  z_[i][j] = ((nb_points_[i][j]-1) * z_[i][j] + nbP * z) / (nb_points_[i][j] + nbP);
// 	  (nb_points_[i][j]) += nbP;
	}

      return 0;
    }
}

/**
 *  \brief add a new altitude value in a cell
 *  \param i : the abscica of the cell
 *  \param j : the ordinate of the cell
 *  \param z : the altitude value
 *  \param nbP : the number of points corresponding to the new value
 *  \return 0 if the process succeed, -1 if the input coordinate are out of the grid
 */
int DEM::addValueInCoord(double x,double y,float z,int nbP)
{
  int i,j;
  int res = this->getIndexPos(x,y,i,j);

  if(res == 0) return this->addValueInCell(i,j,z,nbP);
  else return res;
}

/**
 *  \brief merge the current DEM with the input one (2nd version)
 *  \param dem2 : the input DEM
 *  \result return 0 if the process succeed, -1 else
 */
int DEM::mergeDEM(const DEM& dem2)
{
  // merge both DEMs
  for(int i=0; i<dem2.nbCols(); i++)
    {
      for(int j=0; j<dem2.nbLines(); j++)
	{
	  // check if the current cell of the second grid is defined
	  if(!dem2.isEmptyCell(i,j))
	    {
	      // get back the altitude and number of points corresponding to the current index
	      float z2 = dem2.getZ(i,j);
	      int nbP2 = dem2.getNbP(i,j);

	      // get back the coordinates corresponding to the current index
	      float x2,y2;
	      dem2.getPosIndex(i,j,x2,y2);

	      // add these values to the current grid
	      int resAdd = this->addValueInCoord(x2,y2,z2,nbP2);
	    }
	}
    }
  return 0;
}


// /***** Boolean obstacles building & evaluation ****/

// /**
//  *  \brief build boolean obstacles from the DEM
//  *  \param maxSlope : the maximal slope the robot is able to climb
//  *  \return 0 if the process succeed, -1 else
//  */
// int DEM::buildBoolObs(float maxSlope)
// {
// }

// /**
//  *  \brief increase the dimensions of the boolean obstacles
//  *  \param dist : the distance that the obstacles are inceased by
//  *  \return 0 if the process succeed, -1 else
//  */
// int DEM::increaseBoolObs(float dist)
// {
// }

// /**
//  *  \brief check if a position (x,y) on the DEM is free
//  *  \param x : absica
//  *  \param y : ordinate
//  */
// bool DEM::isFree(float x,float y) const
// {
// }

// /**
//  *  \brief check if a position (x,y) on the DEM is obstructed
//  *  \param x : absica
//  *  \param y : ordinate
//  */
// bool DEM::isObs(float x,float y) const
// {
// }


/******** Initialization functions (DEBUG) ********/

/**
 *  \brief initialize the DEM with a flat terrain
 *  \param nbCols : number rof columns
 *  \param nbLines : number rof lines
 *  \param xScale : X size of each cell
 *  \param yScale : Y size of each cell
 *  \param zMin : minimal Z value used to convert the DEM into an image
 *  \param zMax : maximal Z value used to convert the DEM into an image
 *  \param zPlane : altitude of the plane
 */
void DEM::initFlatDEM(int nbCols, int nbLines, float xScale, float yScale, float zMin, float zMax, float zPlane)
{
  // clear tables
  this->clearTables();

  // init tables from their dimesions
  this->initTables(nbCols,nbLines);

  // init the additionnal values of the DEM
  xOrigin_ = 0.0;
  yOrigin_ = 0.0;
  
  xScale_ = xScale;
  yScale_ = yScale;
  
  zOrigin_ = 0.5*(zMin+zMax);
  zScale_ = 0.1;
  
  // add a z value in each cell of the grid
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      this->addValueInCell(i,j,zPlane);
}

/**
 *  \brief initialize the DEM with a random terrain
 *  \param nbCols : number rof columns
 *  \param nbLines : number rof lines
 *  \param xScale : X size of each cell
 *  \param yScale : Y size of each cell
 *  \param zMin : minimal Z value used to convert the DEM into an image
 *  \param zMax : maximal Z value used to convert the DEM into an image
 *  \param probSeen : the probability of each cell to have be already seen
 */
void DEM::initRandDEM(int nbCols, int nbLines, float xScale, float yScale, float zMin, float zMax, float probSeen)
{
  // clear tables
  this->clearTables();

  // init tables from their dimesions
  this->initTables(nbCols,nbLines);

  // init the additionnal values of the DEM
  xOrigin_ = 0.0;
  yOrigin_ = 0.0;
  
  xScale_ = xScale;
  yScale_ = yScale;
  
  zOrigin_ = 0.5*(zMin+zMax);
  zScale_ = 0.1;
  
  // add a z value in each cell of the grid
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	double randZ = getDoubleRandom(zMin,zMax);
	double randSeen = getDoubleRandom(0.0,1.0);

	if(randSeen <= probSeen) this->addValueInCell(i,j,randZ);
      }
}


/************ Input / Output functions ************/

/*
 *  \brief init the Z color parameters
 *  \param minColZ : output minimal Z value used to convert the DEM into an image
 *  \param maxColZ : output maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : output Z scale value used to convert the DEM into an image
 */
void DEM::initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const 
{
  // search for the minimal and maximal altitudes over the map
  float min = 999999999;
  float max = -999999999;

  for(int i=0; i<nbCols_; i++)
    {
      for(int j=0; j<nbLines_; j++)
	{
	  if(nb_points_[i][j] > 0)
	    {
	      if(z_[i][j] < min) min = z_[i][j];
	      if(z_[i][j] > max) max = z_[i][j];
	    }
	}
    }

  // compute the minimal and maximal used values and the scale - heuristic values
  float diff = max - min;
  minColZ = min + 0.1*diff;
  maxColZ = max - 0.1*diff;
  scaleColZ = (maxColZ - minColZ) / colDepth_;
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 */
int DEM::greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const 
{
  if(nb_points_[i][j] <= 0) 
    return ((int)(0.5*colDepth_));

  else if(z_[i][j] <= minColZ) 
    return 0;

  else if(z_[i][j] >= maxColZ)
    return colDepth_;

  else 
    return (z_[i][j] - minColZ) / scaleColZ;
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \param R : the output red colour
 *  \param G : the output green colour
 *  \param B : the output blue colour
 */
void DEM::rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const
{
  if(nb_points_[i][j] <= 0) 
    {
      R = ((int)(0.5*colDepth_));
      G = ((int)(0.5*colDepth_));
      B = ((int)(0.5*colDepth_));
    }
  else if(z_[i][j] <= minColZ) 
    {
      R = 0;
      G = 0;
      B = colDepth_;
    }
  else if(z_[i][j] >= maxColZ)
    {
      R = 0;
      G = colDepth_;
      B = 0;
    }
  else 
    {
      R = 0;
      G = (int)((z_[i][j] - minColZ) / scaleColZ);
      B = (int)(colDepth_ - ((z_[i][j] - minColZ) / scaleColZ));
      return ;
    }
}

/**
 *  \brief save all information about the DEM into an output file
 *  \param filename : the output filename
 */
int DEM::save(const char* filename) const
{
  // open the output file
  ofstream fic(filename, ios::out | ios::trunc | ios::binary);

  // check if the opening succeeded
  if(fic)
    {
      // write the dimension of the DEM
      fic << nbCols_ << " " << nbLines_ << endl;

      // write the origin cooridnates of the DEM
      fic << xOrigin_ << " " << yOrigin_ << " " << zOrigin_ << endl;

      // write the scale of the DEM
      fic << xScale_ << " " << yScale_ << " " << zScale_ << endl;

      // write the height table
      for(int i=0; i<nbCols_; i++)
	for(int j=0; j<nbLines_; j++)
	  fic << z_[i][j] << " ";
      fic << endl;

      // write the nb_points table
      for(int i=0; i<nbCols_; i++)
	for(int j=0; j<nbLines_; j++)
	  fic << nb_points_[i][j] << " ";
      fic << endl;
      
      fic.close();

      return 0;
    }
  else
    {
      cerr << "DEM::save : cannot open the output file" << endl;
      return -1;
    }
}

/**
 *  \brief load the DEM from an input file
 *  \param filename : the input filename
 */
int DEM::load(const char* filename)
{
  // create an input file
  ifstream file;
  
  // open the file
  file.open(filename);

  if(file.fail())
    {
      cerr << "DEM::load : Unable to open the input file " << filename << endl;
      return -1;
    }

  // read the dimensions of the DEM
  int nbCols, nbLines;
  file >> nbCols;
  file >> nbLines;
  if(nbCols <= 0 || nbLines <= 0)
    {
      cerr << "DEM::load : DEM dimensions not valid (nbCols=" << nbCols << ", nbLines=" << nbLines << ")" << endl;
      return -1;
    }

  cout << "DEM::load -> nbCols : " << nbCols << "  nbLines : " << nbLines << endl;

  // write the origin cooridnates of the DEM
  file >> xOrigin_;
  file >> yOrigin_;
  file >> zOrigin_;
  
  // write the scale of the DEM
  file >> xScale_;
  file >> yScale_;
  file >> zScale_;

  // reinit the DEM tables
  this->clearTables();
  this->initTables(nbCols,nbLines);

  // read the height table
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      file >> z_[i][j];
  
  // write the nb_points table
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      file >> nb_points_[i][j];

  // close the file
  file.close();

  return 0;
}


/**************** Utility functions ***************/

/**
 *  \brief extract a sub-map of the current map
 *  \param iMin : the number of the first column to keep
 *  \param iMax : the number of the last column to keep
 *  \param jMin : the number of the first line to keep
 *  \param jMax : the number of the last line to keep
 *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
 *
 *  A new instance of DEM is created
 */
DEM* DEM::subMap(int iMin, int iMax, int jMin, int jMax) const
{
  // check if the input parameters are valid
  if(iMax <= iMin || jMax <= jMin) return NULL;

  DEM* subMap = new DEM(iMax - iMin, 
			jMax - jMin, 
			this->xOrigin_ + iMin*this->xScale_, 
			this->yOrigin_ + jMin*this->yScale_, 
			this->zOrigin_,
			this->xScale_, 
			this->yScale_,
			this->zScale_);
  
  for(int i=iMin; i<iMax; i++)
    for(int j=jMin; j<jMax; j++)
      {
	subMap->z_[i-iMin][j-jMin] = this->z_[i][j];
	subMap->nb_points_[i-iMin][j-jMin] = this->nb_points_[i][j];
      }

  return subMap;
}

/**
 *  \brief extract a sub-map of the current map
 *  \param xMin : the number of the first column to keep
 *  \param xMax : the number of the last column to keep
 *  \param yMin : the number of the first line to keep
 *  \param yMax : the number of the last line to keep
 *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
 *
 *  A new instance of DEM is created
 */
DEM* DEM::subMap(float xMin, float xMax, float yMin, float yMax) const
{
  int iMin, iMax, jMin, jMax;
  int resGIP1 = this->getClosestIndexToPos(xMin, yMin, iMin, jMin);
  int resGIP2 = this->getClosestIndexToPos(xMax, yMax, iMax, jMax);

  cout << "DEM::subMap -> iMin : " << iMin << "  iMax : " << iMax << "  jMin : " << jMin << "  jMax : " << jMax << endl;

  return this->subMap(iMin, iMax, jMin, jMax);
}

/**
 *  \brief extract a sub-map of the current map
 *  \param (xCenter,yCenter) : coordinates of the center of the sub-map
 *  \param size : size of the sub-map
 *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
 *
 *  A new instance of DEM is created
 */
DEM* DEM::subMap(float xCenter, float yCenter, float size) const
{
  float xMin = xCenter - 0.5*size;
  float xMax = xCenter + 0.5*size;
  float yMin = yCenter - 0.5*size;
  float yMax = yCenter + 0.5*size;

  return this->subMap(xMin, xMax, yMin, yMax);
}

/**
 *  \brief display the main features of the DEM
 *  \param dispGrid : display the grid if true
 */
void DEM::display(bool dispGrid) const
{
  // display the main features
//   cout << "DEM features : " << endl;
  cout << "nbLines : " << nbLines_ <<  "  nbCols : " << nbCols_ << endl;
  cout << "xOrigin : " << xOrigin_ <<  "  yOrigin : " << yOrigin_ << "  zOrigin : " << zOrigin_ << endl;
  cout << "xScale : " << xScale_ <<  "  yScale : " << yScale_ << "  zScale : " << zScale_ << endl;
  cout << endl;

  // display the grid if needed
  if(dispGrid)
    {
      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      cout << z_[i][j] << "(" << nb_points_[i][j] << ") ";
	    }
	  cout << endl;
	}
      cout << endl;
    }
}
