#include "environment/inc/SlopeMap.hpp"
using namespace mpc;

#include <stdlib.h>
#include <iostream>
#include <cmath>
using namespace std;

                                    /**************************************************/
                                    /****************** Class SlopeMap ****************/
                                    /**************************************************/

float SlopeMap::colDepth_ = 255;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SlopeMap::SlopeMap() : Map2D()
{
  this->slope_ = NULL;
  this->cellState_ = NULL;
}

/**
 *  \brief copy constructor
 *  \param map : the map to copy
 */
SlopeMap::SlopeMap(const SlopeMap& map) : Map2D(map)
{
  this->initTables(map.nbCols_,map.nbLines_);

  // copy the tables
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	map2D_cell_t cellSt;
	float slope;
	map.getCellInfo(i,j,cellSt,slope);

	if(cellSt == CELL_FILLED) this->setSlope(i,j,slope);
	else this->setUndefCell(i,j);
      }
}

/**
 *  \brief init constructor
 *  \param nbCols : number of columns of the map table
 *  \param nbLines : number of lines of the map table
 *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param xScale : X size of a cell in meters
 *  \param yScale : Z size of a cell in meters
 */
SlopeMap::SlopeMap(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale) : Map2D(nbCols,nbLines,xOrigin,yOrigin,xScale,yScale)
{
  this->initTables(nbCols,nbLines);

  // set all the cells as unknown
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	this->setUndefCell(i,j);
      }
}

/**
 *  \brief init constructor (from a data elevation map)
 *  \param dem : the input dem
 *  \param maxDist : the maximal distance considered between two cells to compute the slope
 */
SlopeMap::SlopeMap(const DEM* dem, float maxDist) : Map2D()
{
  this->slope_ = NULL;
  this->cellState_ = NULL;
  
  this->buildFromDEM(dem, maxDist);
}

/**
 *  \brief destructor
 */
SlopeMap::~SlopeMap()
{
  this->clearTables();
}

/**
 *  \brief return a copy of the current map
 *  \return a pointer to the copy of the current map
 *
 *  A new instance of SlopeMap is allocated
 */
SlopeMap* SlopeMap::copy() const
{
  return new SlopeMap(*this);
}


/***************** Initialization *****************/

/*
 *  \brief init tables from their dimesions
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void SlopeMap::initTables(int nbCols,int nbLines)
{
  nbCols_ = nbCols;
  nbLines_ = nbLines;

  // Slope of each cell
  slope_ = new float* [nbCols];
  for(int i=0; i<nbCols; i++) slope_[i] = new float [nbLines];

  // State of each cell
  cellState_ = new map2D_cell_t* [nbCols];
  for(int i=0; i<nbCols; i++) cellState_[i] = new map2D_cell_t [nbLines];
}

/*
 *  \brief delete all the content of the slope map tables
 */
void SlopeMap::clearTables()
{
  if(slope_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(slope_[i] != NULL) delete [] slope_[i];
      delete [] slope_;
    }
  slope_ = NULL;

  if(cellState_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(cellState_[i] != NULL) delete [] cellState_[i];
      delete [] cellState_;
    }
  cellState_ = NULL;
}


/******************** Accessors *******************/

/*
 *  \brief set the cell (i,j) as undefined (no knowledge about the slope)
 *  \param (i,j) : index of the cell in the grid     
 */
void SlopeMap::setUndefCell(int i,int j)
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      return;
    }
  else 
    {
      slope_[i][j] = 0;
      cellState_[i][j] = CELL_EMPTY;
    }
}

/*
 *  \brief set the slope of cell (i,j)
 *  \param (i,j) : index of the cell in the grid     
 *  \param slope : the new slope
 */
void SlopeMap::setSlope(int i,int j,float slope)
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      return;
    }
  else 
    {
      slope_[i][j] = slope;
      cellState_[i][j] = CELL_FILLED;
    }
}

/**
 *  \brief get informations about the cell (i,j)
 *  \param (i,j) : index of the cell in the grid
 *  \param cellSt : the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
 *  \param slope : the resulting slope
 */
void SlopeMap::getCellInfo(int i, int j, map2D_cell_t& cellSt, float& slope) const
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      cellSt = CELL_EMPTY;
      slope = 0;
    }
  else 
    {
      cellSt = cellState_[i][j];
      slope = slope_[i][j];
    }
}


/*************** Slope map building ***************/

/*
 *  \brief compute the slope between two cells (i,j) and (k,l)
 *  \param dem : the input dem
 *  \param (i,j) : index of the cell in the grid
 *  \param zIJ : altitude in the cell of index (i,j)
 *  \param (k,l) : index of the second cell in the grid
 *  \param slope : the resulting slope
 *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
 *
 *  The cell (i,j) is supposed to be filled
 */
map2D_cell_t SlopeMap::computeSlope(const DEM* dem, int i, int j, float zIJ, int k, int l, float& slope) const
{
//   // check if the cell (k,l) is filled
//   map2D_cell_t cellStKL; 
//   float slopeKL;
//   this->getCellInfo(k, l, cellStKL, slopeKL);

  if(k < 0 || k >= nbCols_ || l < 0 || l >= nbLines_)
    {
      slope = 0;
      return CELL_EMPTY;
    }
  else if(dem->isEmptyCell(k,l))
    {
      slope = 0;
      return CELL_EMPTY;
    }
  else 
    {
      // get back the altitude in cell (k,l)
      float zKL = dem->getZ(k,l);

      // compute the distance between both cells
      float dx = abs(k-i)*xScale_;
      float dy = abs(l-j)*yScale_;
      float dist = sqrt(dx*dx + dy*dy);
      
      // compute the slope between both cell
      if(dist == 0)
	{
	  slope = 0;
	  return CELL_FILLED;
	}
      else 
	{
	  slope = atan2(fabs(zKL - zIJ), dist);
	  return CELL_FILLED;
	}
    }
}

/*
 *  \brief compute the slope in a cell (i,j)
 *  \param dem : the input dem
 *  \param (i,j) : index of the cell in the grid
 *  \param slope : the resulting slope
 *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
 */
map2D_cell_t SlopeMap::computeSlope(const DEM* dem, int i, int j, float& slope) const
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      slope = 0;
      return CELL_EMPTY;
    }
  else if(dem->isEmptyCell(i,j))
    {
      slope = 0;
      return CELL_EMPTY;
    }
  else 
    {
      // get back the altitude of teh current cell
      float zIJ = dem->getZ(i,j);

      // initialize the maximale slope
      float maxSlope = 0;
      float curSlope;

      for(int k=i-maxNbCells_; k<=i+maxNbCells_; k++)
	{
	  for(int l=j-maxNbCells_; l<=j+maxNbCells_; l++)
	    {
	      map2D_cell_t cellStKL = this->computeSlope(dem, i, j, zIJ, k, l, curSlope);
	      if (curSlope > maxSlope) maxSlope = curSlope;
	    }
	}
      slope = maxSlope;
      return CELL_FILLED;
    }
}

/*
 *  \brief initialize the maximal number of cells considered to compute the slope
 *  \param maxDist : the maximal distance considered to compute the slope
 */
void SlopeMap::initMaxNbCells(float maxDist)
{
  if(maxDist == 0) maxNbCells_ = 1;
  else
    {
      maxNbCells_ = (int)(maxDist / xScale_);
      if(maxNbCells_ == 0) maxNbCells_ = 1;
    }

//   cout << "maxNbCells_ : " << maxNbCells_ << endl;
}

/**
 *  \brief build the slope map from an input data elevation map
 *  \param dem : the input dem
 *  \param maxDist : the maximal distance considered between two cells to compute the slope
 *  \return 0 if the process succeed, -1 else
 */
int SlopeMap::buildFromDEM(const DEM* dem, float maxDist)
{
  if(dem == NULL) return -1;

  // initialize the slope map parameters
  this->nbCols_ = dem->nbCols();
  this->nbLines_ = dem->nbLines();
  
  this->xOrigin_ = dem->xOrigin();
  this->yOrigin_ = dem->yOrigin();
  
  this->xScale_ = dem->xScale();
  this->yScale_ = dem->yScale();
  
  // initialize the slope map tables
  this->clearTables();
  this->initTables(dem->nbCols(),dem->nbLines());

  // initialize the maximal number of cells considered to compute the slope
  this->initMaxNbCells(maxDist);

  // compute the slope of each cell
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	float curSlope;
	map2D_cell_t cellStIJ = this->computeSlope(dem, i, j, curSlope);	

	if(cellStIJ == CELL_EMPTY) this->setUndefCell(i,j);
	else this->setSlope(i,j,curSlope);
      }

  return 0;
}


/************ Input / Output functions ************/

/*
 *  \brief init the Z color parameters
 *  \param minColZ : output minimal Z value used to convert the DEM into an image
 *  \param maxColZ : output maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : output Z scale value used to convert the DEM into an image
 */
void SlopeMap::initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const
{
  // nothing to do
  minColZ = 0;
  maxColZ = 0;
  scaleColZ = 0;
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \return the colour corresponding to a cell of the DEM
 */
int SlopeMap::greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const
{
  map2D_cell_t cellSt;
  float slope;
  getCellInfo(i, j, cellSt, slope);

  if(cellSt == CELL_EMPTY)
    return ((int)(0.5*colDepth_));
  else
    return ((int) (colDepth_ * slope / (0.5*M_PI)));
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \param R : the output red colour
 *  \param G : the output green colour
 *  \param B : the output blue colour
 */
void SlopeMap::rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const
{
  map2D_cell_t cellSt;
  float slope;
  getCellInfo(i, j, cellSt, slope);

  if(cellSt == CELL_EMPTY)
    {
      R = (int)(0.5*colDepth_);
      G = (int)(0.5*colDepth_);
      B = (int)(0.5*colDepth_);
    }
  else
    {
      R = (int) (colDepth_ * slope / (0.5*M_PI));
      G = (int) (colDepth_ - (colDepth_ * slope / (0.5*M_PI)));
      B = 0;
    }
}


/**************** Utility functions ***************/

/**
 *  \brief extract a sub-map of the current map
 *  \param iMin : the number of the first column to keep
 *  \param iMax : the number of the last column to keep
 *  \param jMin : the number of the first line to keep
 *  \param jMax : the number of the last line to keep
 *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
 *
 *  A new instance of SlopeMap is created
 */
SlopeMap* SlopeMap::subMap(int iMin, int iMax, int jMin, int jMax) const
{
  // check if the input parameters are valid
  if(iMax <= iMin || jMax <= jMin) return NULL;

  SlopeMap* subMap = new SlopeMap(iMax - iMin, 
				  jMax - jMin, 
				  this->xOrigin_ + iMin*this->xScale_, 
				  this->yOrigin_ + jMin*this->yScale_, 
				  this->xScale_, 
				  this->yScale_);
  
  for(int i=iMin; i<iMax; i++)
    for(int j=jMin; j<jMax; j++)
      {
	subMap->slope_[i-iMin][j-jMin] = this->slope_[i][j];
	subMap->cellState_[i-iMin][j-jMin] = this->cellState_[i][j];
      }

  return subMap;
}

/**
 *  \brief display the main features of the slope map
 *  \param dispGrid : display the grid if true
 */
void SlopeMap::display(bool dispGrid) const
{
  // display the main features
  cout << "nbLines : " << nbLines_ <<  "  nbCols : " << nbCols_ << endl;
  cout << "xOrigin : " << xOrigin_ <<  "  yOrigin : " << yOrigin_ << endl;
  cout << "xScale : " << xScale_ <<  "  yScale : " << yScale_ << endl;
  cout << endl;

  // display the grid if needed
  if(dispGrid)
    {
      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      map2D_cell_t cellSt;
	      float slope;
	      getCellInfo(i, j, cellSt, slope);

	      if(cellSt == CELL_EMPTY) cout << "(?) ";
	      else cout << slope/M_PI*180.0 << " ";
	    }
	  cout << endl;
	}
      cout << endl;
    }
}
