#include "environment/inc/VarSlopeMap.hpp"
using namespace mpc;

#include <iostream>
#include <cmath>
using namespace std;

#define MAX_VAR_SLOPE 5.0*M_PI

                                    /**************************************************/
                                    /**************** Class VarSlopeMap ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
VarSlopeMap::VarSlopeMap() : SlopeMap()
{
  this->varSlope_ = NULL;
}
    
/**
 *  \brief copy constructor
 *  \param map : the map to copy
 */
VarSlopeMap::VarSlopeMap(const VarSlopeMap& map) : SlopeMap(map)
{
  this->initTableVS(map.nbCols(),map.nbLines());

  for(int i=0; i<map.nbCols(); i++)
    for(int j=0; j<map.nbLines(); j++)
      varSlope_[i][j] = map.varSlope_[i][j];
}

/**
 *  \brief init constructor (from a data elevation map)
 *  \param dem : the input dem
 *  \param maxDist : the maximal distance considered between two cells to compute the slope
 */
VarSlopeMap::VarSlopeMap(const DEM* dem, float maxDist)
{
  this->varSlope_ = NULL;

  this->buildVSFromDEM(dem, maxDist);
}

/**
 *  \brief destructor
 */
VarSlopeMap::~VarSlopeMap()
{
  this->clearTables();
}

/**
 *  \brief return a copy of the current map
 *  \return a pointer to the copy of the current map
 *
 *  A new instance of SlopeMap is allocated
 */
VarSlopeMap* VarSlopeMap::copy() const
{
  return new VarSlopeMap(*this);
}


/***************** Initialization *****************/

/*
 *  \brief init the slope derivative table from its dimesions
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void VarSlopeMap::initTableVS(int nbCols,int nbLines)
{
  // Slope derivative of each cell
  varSlope_ = new float* [nbCols];
  for(int i=0; i<nbCols; i++) varSlope_[i] = new float [nbLines];
}

/*
 *  \brief delete all the content of the slope map tables
 */
void VarSlopeMap::clearTables()
{
  this->SlopeMap::clearTables();

  if(varSlope_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(varSlope_[i] != NULL) delete [] varSlope_[i];
      delete [] varSlope_;
    }
  varSlope_ = NULL;
}


/******************** Accessors *******************/

/*
 *  \brief set the slope of cell (i,j)
 *  \param (i,j) : index of the cell in the grid     
 *  \param varSlope : the new slope derivative
 */
void VarSlopeMap::setVarSlope(int i, int j, float varSlope)
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      return;
    }
  else 
    {
      varSlope_[i][j] = varSlope;
    }
}

/**
 *  \brief get informations about the cell (i,j)
 *  \param (i,j) : index of the cell in the grid
 *  \param cellSt : the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
 *  \param slope : the resulting slope
 *  \param varSlope : the resulting slope derivative
 */
void VarSlopeMap::getCellInfo(int i, int j, map2D_cell_t& cellSt, float& slope, float& varSlope) const
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      cellSt = CELL_EMPTY;
      slope = 0;
      varSlope = 0;
    }
  else 
    {
      cellSt = cellState_[i][j];
      slope = slope_[i][j];
      varSlope = varSlope_[i][j];
    }
}


/*************** Slope map building ***************/

/*
 *  \brief compute the slope between two cells (i,j) and (k,l)
 *  \param dem : the input dem
 *  \param (i,j) : index of the cell in the grid
 *  \param slopeIJ : altitude in the cell of index (i,j)
 *  \param (k,l) : index of the second cell in the grid
 *  \param varSlope : the resulting slope derivative
 *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
 *
 *  The cell (i,j) is supposed to be filled on the dem
 */
map2D_cell_t VarSlopeMap::computeVarSlope(int i, int j, float slopeIJ, int k, int l, float& varSlope) const
{
  if(k < 0 || k >= nbCols_ || l < 0 || l >= nbLines_)
    {
      varSlope = 0;
      return CELL_EMPTY;
    }
  
  // check if the cell (k,l) is filled
  map2D_cell_t cellStKL; 
  float slopeKL;
  this->SlopeMap::getCellInfo(k, l, cellStKL, slopeKL);

  if(cellStKL == CELL_EMPTY)
    {
      varSlope = 0;
      return CELL_EMPTY;
    }
  else
    {
      // compute the distance between both cells
      float dx = fabs(k-i)*xScale_;
      float dy = fabs(l-j)*yScale_;
      float dist = sqrt(dx*dx + dy*dy);

      if(dist == 0)
	{
	  varSlope = 0;
	  return CELL_FILLED;
	}
      else
	{
	  varSlope = fabs(slopeKL - slopeIJ) / dist;
	  return CELL_FILLED;
	}
    }
}

/*
 *  \brief compute the slope in a cell (i,j)
 *  \param dem : the input dem
 *  \param (i,j) : index of the cell in the grid
 *  \param varSlope : the resulting slope derivative
 *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
 */
map2D_cell_t VarSlopeMap::computeVarSlope(int i, int j, float& varSlope) const
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    {
      varSlope = 0;
      return CELL_EMPTY;
    }
  else 
    {
      // check if the cell (i,j) is filled
      map2D_cell_t cellStIJ; 
      float slopeIJ;
      this->SlopeMap::getCellInfo(i, j, cellStIJ, slopeIJ);

      if(cellStIJ == CELL_EMPTY)
	{
	  varSlope = 0;
	  return CELL_EMPTY;
	}

      // initialize the maximale slope derivative
      float maxVarSlope = 0;
      
      for(int k=i-maxNbCells_; k<=i+maxNbCells_; k++)
	{
	  for(int l=j-maxNbCells_; l<=j+maxNbCells_; l++)
	    {
	      float curVarSlope;
	      map2D_cell_t cellStKL = this->computeVarSlope(i, j, slopeIJ, k, l, curVarSlope);

	      if (curVarSlope > maxVarSlope) maxVarSlope = curVarSlope;
	    }
	}
      varSlope = maxVarSlope;
      return CELL_FILLED;
    }
}

/**
 *  \brief build the slope map from an input data elevation map
 *  \param dem : the input dem
 *  \param maxDist : the maximal distance considered between two cells to compute the slope
 *  \return 0 if the process succeed, -1 else
 */
int VarSlopeMap::buildVSFromDEM(const DEM* dem, float maxDist)
{
  // clear all tables
  this->clearTables();

  // build the slope map
  int resBuildSlope = this->SlopeMap::buildFromDEM(dem, maxDist);
  if(resBuildSlope == -1)
    return -1;

  // initialize the table of slope derivatives
  this->initTableVS(nbCols_, nbLines_);

  // compute the slope derivative of each cell
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	float curVarSlope;
	map2D_cell_t cellStIJ = this->computeVarSlope(i, j, curVarSlope);	

	if(cellStIJ != CELL_EMPTY) this->setVarSlope(i,j,curVarSlope);
      }

  return 0;  
}


/************ Input / Output functions ************/

/*
 *  \brief init the Z color parameters
 *  \param minColZ : output minimal Z value used to convert the DEM into an image
 *  \param maxColZ : output maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : output Z scale value used to convert the DEM into an image
 */
void VarSlopeMap::initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const
{
  // nothing to do
  minColZ = 0;
  maxColZ = MAX_VAR_SLOPE;
  scaleColZ = MAX_VAR_SLOPE / colDepth_;
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \return the colour corresponding to a cell of the DEM
 */
int VarSlopeMap::greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const
{
  map2D_cell_t cellSt;
  float slope;
  float varSlope;
  getCellInfo(i, j, cellSt, slope, varSlope);

  if(cellSt == CELL_EMPTY) return ((int)(0.5*colDepth_));

  else if(varSlope > MAX_VAR_SLOPE) return colDepth_;

  else return ((int) (colDepth_ * varSlope / MAX_VAR_SLOPE));
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \param R : the output red colour
 *  \param G : the output green colour
 *  \param B : the output blue colour
 */
void VarSlopeMap::rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const
{
  map2D_cell_t cellSt;
  float slope;
  float varSlope;
  getCellInfo(i, j, cellSt, slope, varSlope);

  if(cellSt == CELL_EMPTY) 
    {
      R = ((int)(0.5*colDepth_));
      G = ((int)(0.5*colDepth_));
      B = ((int)(0.5*colDepth_));
    }

  else if(varSlope > MAX_VAR_SLOPE)
    {
      R = colDepth_;
      G = 0;
      B = 0;
    }

  else 
    {
      R = ((int) (colDepth_ * varSlope / MAX_VAR_SLOPE));
      G = colDepth_ - ((int) (colDepth_ * varSlope / MAX_VAR_SLOPE));
      B = 0;
    }
}


/**************** Utility functions ***************/

/**
 *  \brief display the main features of the var slope map
 *  \param dispGrid : display the grid if true
 */
void VarSlopeMap::display(bool dispGrid) const
{
  // display the main features
  cout << "nbLines : " << nbLines_ <<  "  nbCols : " << nbCols_ << endl;
  cout << "xOrigin : " << xOrigin_ <<  "  yOrigin : " << yOrigin_ << endl;
  cout << "xScale : " << xScale_ <<  "  yScale : " << yScale_ << endl;
  cout << endl;

  // display the grid if needed
  if(dispGrid)
    {
      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      map2D_cell_t cellSt;
	      float slope;
	      float varSlope;
	      getCellInfo(i, j, cellSt, slope, varSlope);

	      if(cellSt == CELL_EMPTY) cout << "(?) ";
	      else cout << varSlope << " ";
	    }
	  cout << endl;
	}
      cout << endl;
    }
}
