#ifndef MPC_TRAJPLOT_HPP
#define MPC_TRAJPLOT_HPP

#include "common/inc/ControlTime.hpp"
#include "common/inc/StateTime.hpp"
#include "motion/inc/PSTraj.hpp"
#include "motion/inc/PCTraj.hpp"

#include "gnuplot/inc/gnuplot_i.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class TrajPlot ******************/
                                    /**************************************************/

  /** \class TrajPlot
   *  \brief Graphic tool to plot a control or state trajectory in a gnuplot window
   */
  class TrajPlot
  {
  protected : 

  public : 
    /***************** Plot functions *****************/

    /**
     *  \brief plot a math function
     *  \param plot : the gnuplot window
     *  \param fct : the function to plot
     *  \param title : the tite of the function
     *  \param sampling : the number of sampling points of the trajectory
     *  \param minBound : the minimal boundary to display
     *  \param maxBound : the maximal boundary to display
     */
    static void plotFct(Gnuplot& plot, const Function& fct, const char* title, int sampling=50, double minBound=INFINITY, double maxBound=-INFINITY);

    /**
     *  \brief plot a path from a Mana continuous state trajectory
     *  \param plot : the gnuplot window
     *  \param straj : the continuous state trajectory to display
     *  \param title : the tite of the function
     *  \param sampling : the number of sampling points of the trajectory
     */
    static void plotDD2Path(Gnuplot& plot, const DD2PSTraj& straj, const char* title, int sampling=50);

    /**
     *  \brief plot the control trajectory of a Mana robotic system
     *  \param ctraj : the discrete control trajectory to display
     *  \param sys : the system constraints
     *  \param dispBounds : true if the bounds defined over the control must be displayed
     */
    static void plotDD2CTraj(const DD2CTraj& ctraj, const DiffDrive2Sys* sys=NULL, bool dispBounds=true);

    /**
     *  \brief plot the control trajectory of a Mana robotic system
     *  \param ctraj : the discrete control trajectory to display
     *  \param sys : the system constraints
     *  \param dispBounds : true if the bounds defined over the control must be displayed
     */
    static void plotDD2AccCTraj(const DD2AccCTraj& ctraj, const DiffDrive2Sys* sys=NULL, bool dispBounds=true);

    /**
     *  \brief plot the control trajectory of a Mana robotic system
     *  \param ctraj : the discrete control trajectory to display
     *  \param sys : the system constraints
     *  \param dispBounds : true if the bounds defined over the control must be displayed
     */
    static void plotDD2STraj(const DD2STraj& straj, const DiffDrive2Sys* sys=NULL, bool dispBounds=true);

    /**
     *  \brief plot the state trajectory of a Mana robotic system
     *  \param straj1 : the first discrete state trajectory to display
     *  \param straj2 : the second discrete state trajectory to display
     *  \param name1 : name of the first discrete state trajectory
     *  \param name2 : name of the second discrete state trajectory
     *  \param sys : the system constraints (robot constraints are displayed if the system is not null)
     */
    static void plotDD2STraj(const DD2STraj& straj1, const DD2STraj& straj2, const char* name1, const char* name2, const DiffDrive2Sys* sys=NULL);

    /**
     *  \brief plot the control trajectory of a Mana robotic system
     *  \param ctraj : the continuous control trajectory to display
     *  \param sys : the system constraints
     *  \param dispBounds : true if the bounds defined over the control must be displayed
     *  \param sampling : the number of sampling points of the trajectory
     */
    static void plotDD2PCTraj(const DD2PCTraj& ctraj, const DiffDrive2Sys* sys=NULL, bool dispBounds=true, int sampling=50);

    /**
     *  \brief plot the control trajectory of a Mana robotic system
     *  \param straj : the continuous state trajectory to display
     *  \param sys : the system constraints
     *  \param dispBounds : true if the bounds defined over the control must be displayed
     *  \param sampling : the number of sampling points of the trajectory
     */
    static void plotDD2PSTraj(const DD2PSTraj& straj, const DiffDrive2Sys* sys=NULL, bool dispBounds=true, int sampling=50);

    /**
     *  \brief plot the control trajectory of a Mana robotic system
     *  \param ctraj : the continuous control trajectory to display
     *  \param sys : the system constraints
     *  \param dispBounds : true if the bounds defined over the control must be displayed
     *  \param sampling : the number of sampling points of the trajectory
     */
    static void plotDD2PCTraj(Gnuplot& plotVel, Gnuplot& plotAVel, Gnuplot& plotAcc, Gnuplot& plotAAcc, const DD2PCTraj& ctraj, const DiffDrive2Sys* sys=NULL, int sampling=50);

//     /**
//      *  \brief plot the control trajectory of a Mana robotic system
//      *  \param straj : the continuous state trajectory to display
//      *  \param sys : the system constraints
//      *  \param dispBounds : true if the bounds defined over the control must be displayed
//      *  \param sampling : the number of sampling points of the trajectory
//      */
//     static void plotDD2PSTraj(Gnuplot& plot, const DD2PSTraj& straj, const DiffDrive2Sys* sys=NULL, int sampling=50);
  };

};
#endif // MPC_TRAJPLOT_HPP
