#ifndef EQUATION_HPP
#define EQUATION_HPP

// #include "Variable.hpp"
#include <vector>

/**
 * \enum math_func_t
 * \brief Enumeration of the different mathematical function types
 *
 * ND_FUNC        : Undefined function
 * POLY_FUNC      : Polynomial function
 * QUAD_FUNC      : Quadratic function
 * MULTIITV_FUNC  : Piecewise function
 */
typedef enum {
  ND_FUNC        = 0,
  POLY_FUNC      = 1, 
  ITG_FUNC       = 2,
  MULTIITV_FUNC  = 3,
  MULT_FUNC      = 4,
  SIN_FUNC       = 5,
  COS_FUNC       = 6
} math_func_t;

typedef struct DISC_POINT
{
  double t;
  double v;
} Disc_Point;

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class Function *****************/
                                    /**************************************************/
  class Function
  {
  public : 
    double binf_;                                      // minimal bound of the input parameter
    double bsup_;                                      // maximal bound of the input parameter

    math_func_t type_;                                 // math function type

    //   std::vector<Variable> *lvar_;                      // list of variables of the function

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Function();

    /**
     *  \brief copy constructor
     *  \param f :  the function to copy 
     */
    Function(const Function& f);

    /**
     *  \brief destructor
     */
    virtual ~Function();

    /**
     *  \brief return a copy of the current function
     *  \return a copy of the current function
     *
     *  A new instance of function is allocated and returned here
     */
    virtual Function* copy() const = 0;

    /**************** Setters & Getters ***************/

    /**
     *  \brief return the value of the polynomial at a given parameter t
     *  \param t : the input parameter value
     *  \return the value of the polynomial at a given parameter t
     */
    virtual double getVal(double t) const = 0;

    /**
     *  \brief set the bounds on which the function is defined
     *  \param binf : the new lower bound
     *  \param bsup : the new upper bound
     */
    void setBounds(double binf,double bsup);

    /**
     *  \brief return the minimal bound of the function
     *  \return the minimal bound of the function
     */
    inline virtual double getBInf() const { return binf_; }

    /**
     *  \brief return the maximal bound of the function
     *  \return the maximal bound of the function
     */
    inline virtual double getBSup() const { return bsup_; }
  
    /**
     *  \brief return the math function type
     *  \return the math function type
     */
    inline math_func_t getType() const { return type_; }


    /******************** Operators *******************/

    /**
     *  \brief multiply the current function by a constant value
     *  \param v : the input multiplication coefficient
     */
    virtual void operator*=(double v) = 0;

    //   virtual void operator*=(const Polynomial& f2);     // multiply the current Function by an other Polynomial
    //   virtual void operator*=(const MultiItvFunction& f2); // multiply the current Function by a MultiItvFunction

    /************ Integration & Derivation ************/

    /**
     *  \brief return the derivation of the current function
     *  \return the derivation of the current function
     */
    virtual Function* derivation() const = 0;

    //   virtual Function* derivation(const char* nameVar) = 0; // return the derivation of the current polynome wrt the variable named nameVar

    /**
     *  \brief return the integration of the current function
     *  \return the integration of the current function
     */
    virtual Function* integration() const = 0;

    /**
     *  \brief return the integration of the current function
     *  \param cste : the integration constant at t=0
     *  \return the integration of the current function
     */
    virtual Function* integration(double cste) const = 0;

    /**
     *  \brief return the integration of the current function
     *  \param cste,binf : cste is the integration constant value defined at t=binf
     *  \return the integration of the current function
     */
    virtual Function* integration(double binf,double cste) const = 0;

    /****************** Roots search ******************/

//     /**
//      *  \brief return the real roots of this function
//      *  \param r : the list of roots of the function
//      *  \return the number of roots of the function
//      */
//     virtual int realRoots(std::vector<double>& r) const = 0;

    /**
     *  \brief return the roots of the equation eq - v = 0
     *  \param r : the list of roots of the function
     *  \param v : the input constant value
     *  \return the number of roots of the function
     */
    virtual int realRoots(std::vector<double>& r,double v=0.0) const = 0;

    /**************** Utility functions ***************/

    /**
     *  \brief display the polynomial
     */
    virtual void display() const = 0;

    /**
     *  \brief sample the function on interval [binf;bsup]
     *  \param binf : the lower bound from which the function must be sampled
     *  \param bsup : the upper bound from which the function must be sampled
     *  \param step : the sampling step
     *  \param sampling : the resulting sampling vector
     */
    virtual void sample(double binf,double bsup,int step,std::vector<double>& sampling) const;

    /**
     *  \brief return true if t \in [binf_;bsup_]
     *  \param t : the input parameter value
     *  \return true if t \in [binf_;bsup_]
     */
    bool isDefinedAt(double t) const;

    /**
     *  \brief shift the function on Y axis from value v
     *  \param v : the shifting value
     */
    virtual void shiftY(double v) = 0;

    /**
     *  \brief shift the time parameter of the function
     *  \param t : the shifting value
     */
    virtual void shiftTime(double t) = 0;


    /***************** Discontinuities ****************/

    /**
     *  \brief search for the discontinuities of the function
     *  \param discPoints : resulting discontinuity points
     */
    virtual void getVecDiscPoints(std::vector<Disc_Point>& discPoints) const = 0;

    /**
     *  \brief merges and sorts two vectors of discontinuity points
     *  \param discPoints1 : the first discontinuity points vector
     *  \param discPoints2 : the second discontinuity points vector
     *  \param discPointsRes : the resulting discontinuity points vector
     */
    static void mergeDiscPoints(const std::vector<Disc_Point>& discPoints1, 
				const std::vector<Disc_Point>& discPoints2, 
				std::vector<Disc_Point>& discPointsRes);

    /**
     *  \brief display a vector of discontinuity points
     */
    static void displayDiscPoints(const std::vector<Disc_Point>& discPoints);
  };

};
#endif
