#ifndef MPC_INTERVAL_HPP
#define MPC_INTERVAL_HPP

#include <vector>

/**
 * \enum bound_t
 * \brief Enumeration of the different bound types
 *
 * OPEN_B  : open bound
 * CLOSE_B : close bound
 */
typedef enum {
  OPEN_B  = 0,
  CLOSE_B = 1
} bound_t;

/**
 * \enum order_t
 * \brief Enumeration of the different order types
 *
 * BEFORE : before...
 * DURING : during...
 * AFTER  : after...
 */
typedef enum {
  BEFORE  = 0,
  INCLUDE = 1,
  AFTER   = 2
} order_t;

                                    /**************************************************/
                                    /***************** Class Interval *****************/
                                    /**************************************************/

/*  \class Interval
 *  \brief Interval
 */
class Interval
{
protected : 
  double bLow_;                            /*< lower bound of the interval */
  double bUp_;                             /*< upper bound of the interval */
  bound_t lowBT_;                          /*< type of the lower bound (open or closed) */
  bound_t upBT_;                           /*< type of the upper bound (open or closed) */

public : 
  /********** Constructors and destructors **********/

  /**
   *  \brief default constructor
   */
  Interval();

  /**
   *  \brief copy constructor
   *  \param i : the interval to copy
   */
  Interval(const Interval& i);

  /**
   *  \brief init constructor
   *  \param bLow : the lower bound of the interval
   *  \param bLowT : the type of the lower bound of the interval
   *  \param bUp : the upper bound of the interval
   *  \param bUpT : the type of the upper bound of the interval
   */
  Interval(double bLow,bound_t bLowT,double bUp,bound_t bUpT);

  /**
   *  \brief destructor
   */
  virtual ~Interval();


  /**************** Setters & Getters ***************/

  /**
   *  \brief getter : reference to the lower bound
   *  \return reference to the lower bound of the interval
   */
  inline double& bLow() { return this->bLow_; }

  /**
   *  \brief getter : lower bound
   *  \return lower bound of the interval
   */
  inline double bLow() const { return this->bLow_; }

  /**
   *  \brief getter : reference to the lower bound
   *  \return reference to the lower bound of the interval
   */
  inline double& bUp() { return this->bUp_; }

  /**
   *  \brief getter : lower bound
   *  \return lower bound of the interval
   */
  inline double bUp() const { return this->bUp_; }

  /**
   *  \brief getter : reference to the type of the lower bound
   *  \return reference to the type of the lower bound of the interval
   */
  inline bound_t& lowBT() { return this->lowBT_; }

  /**
   *  \brief getter : type of the lower bound
   *  \return type of the lower bound of the interval
   */
  inline bound_t lowBT() const { return this->lowBT_; }

  /**
   *  \brief getter : reference to the type of the upper bound
   *  \return reference to the type of the upper bound of the interval
   */
  inline bound_t& upBT() { return this->upBT_; }

  /**
   *  \brief getter : type of the upper bound
   *  \return type of the upper bound of the interval
   */
  inline bound_t upBT() const { return this->upBT_; }

  /**
   *  \brief return true if the current interval is empty
   *  \return true if the current interval is empty
   */
  bool isEmpty() const;

  /**
   *  \brief empty the current interval
   */
  void empty();


  /**************** Ordering functions **************/

  /**
   *  \brief return true if the input value is include in the current interval
   *  \param t : the value to test
   *  \return true if the input value is include in the current interval
   */
  bool include(double t) const;

  /**
   *  \brief return the ordering of the input value wrt. the current interval
   *  \param t : the input value to order
   *  \return ordering of the input value wrt. the current interval
   *
   *  ordering : the value t is before, after, or include in the interval
   */
  order_t order(double t) const;


  /*************** Shifting functions ***************/

  /**
   *  \brief shift the interval of a value t
   *  \param t : the value to shift the interval
   */
  void shift(double t);

  
  /**************** Display functions ***************/
  
  /**
   *  \brief display the current interval
   */
  void display() const;
};

#endif // MPC_INTERVAL_HPP


#ifndef MPC_PIECEWISEITV_HPP
#define MPC_PIECEWISEITV_HPP

                                    /**************************************************/
                                    /*************** Class PiecewiseItv ***************/
                                    /**************************************************/

/*  \class PiecewiseItv
 *  \brief PiecewiseItv
 */
class PiecewiseItv
{
protected : 
  std::vector<Interval> vecItv_;

public : 
  /********** Constructors and destructors **********/

  /**
   *  \brief default constructor
   */
  PiecewiseItv();

  /**
   *  \brief copy constructor
   *  \param i : the interval to copy
   */
  PiecewiseItv(const PiecewiseItv& itv);

  /**
   *  \brief init constructor (from a simple interval)
   *  \param i : the simple interval
   */
  PiecewiseItv(const Interval& itv);

  /**
   *  \brief int constructor (simple interval)
   *  \param bLow : the lower bound of the interval
   *  \param bLowT : the type of the lower bound of the interval
   *  \param bUp : the upper bound of the interval
   *  \param bUpT : the type of the upper bound of the interval
   */
  PiecewiseItv(double bLow,bound_t bLowT,double bUp,bound_t bUpT);
  
  /**
   *  \brief destructor
   */
  virtual ~PiecewiseItv();


  /**************** Setters & Getters ***************/

protected : 
  /**
   *  \brief add an interval to the current vector of intervals
   *  \param itv : the simple interval to add 
   *
   *  Before to insert it, must check if there is no intersection with the current intervals
   */
  void addSItv(const Interval& itv);

  /**
   *  \brief append an interval at the end of the current vector of intervals
   *  \param itv : the simple interval to append
   *
   *  Before to insert it, must check if there is no intersection with the current intervals
   */
  void appendSItv(const Interval& itv);

  /**
   *  \brief insert an interval in the current vector of intervals
   *  \param i : the index in the vector to insert the new simple interval
   *  \param itv : the simple interval to insert
   *
   *  Before to insert it, must check if there is no intersection with the current intervals
   */
  void insertSItv(int i,const Interval& itv);


public : 
  /**
   *  \brief access to the i-th simple interval of the piecewise current one
   *  \param i : index of the simple interval we wants to reach
   *  \return a const reference to the i-th simple interval of the piecewise current one
   */
  const Interval& at(int i) const;

  /**
   *  \brief return the number of simple intervals that belongs to the piecewise current one
   *  \return the number of simple intervals
   */
  inline int getNbSItv() const { return this->vecItv_.size(); }

  /**
   *  \brief return true if the current interval is empty
   *  \return true if the current interval is empty
   */
  bool isEmpty() const;

  /**
   *  \brief empty the current interval
   */
  void empty();


  /******************** Operators *******************/
  
  /**
   *  \brief access to the i-th simple interval of the piecewise current one
   *  \param i : index of the simple interval we wants to reach
   *  \return a reference to the i-th simple interval of the piecewise current one
   */
  Interval& operator[](int i);


  /**************** Ordering functions **************/

  /**
   *  \brief return true if the input value is include in the current interval
   *  \param t : the value to test
   *  \return true if the input value is include in the current interval
   */
  bool include(double t) const;


  /*************** Shifting functions ***************/

  /**
   *  \brief shift the interval of a value t
   *  \param t : the value to shift the interval
   */
  void shift(double t);


  /**************** Display functions ***************/

  /**
   *  \brief display the current interval
   */
  void display() const;


  /**************** Friend functions ****************/

  friend void intersecItv(const Interval& itv1,const Interval& itv2,Interval& res);
  friend void intersecItv(const PiecewiseItv& itv1,const PiecewiseItv& itv2,PiecewiseItv& res);
  friend void unionItv(const Interval& itv1,const Interval& itv2,PiecewiseItv& res);
  friend void unionItv(const PiecewiseItv& itv1,const PiecewiseItv& itv2,PiecewiseItv& res);
  friend void complementItv(const Interval& itv1,const Interval& itv2,PiecewiseItv& res);
  friend void complementItv(const PiecewiseItv& itv1,const Interval& itv2,PiecewiseItv& res);
};


                                    /**************************************************/
                                    /************* Intervals computations *************/
                                    /**************************************************/

/*
 *  \brief compute the intersection between two simple intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void intersecItv(const Interval& itv1,const Interval& itv2,Interval& res);

/*
 *  \brief compute the intersection between two piecewise intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 *
 *  Barbar method. Should be improved later.
 */
void intersecItv(const PiecewiseItv& itv1,const PiecewiseItv& itv2,PiecewiseItv& res);

/*
 *  \brief compute the intersection between two simple intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void unionItv(const Interval& itv1,const Interval& itv2,PiecewiseItv& res);

/*
 *  \brief compute the intersection between two piecewise intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void unionItv(const PiecewiseItv& itv1,const PiecewiseItv& itv2,PiecewiseItv& res);

// /*
//  *  \brief compute the complement of interval itv1 in interval itv2
//  *  \param itv1 : the first interval
//  *  \param itv2 : the second interval
//  *  \param res : the resulting interval
//  */
// void complementItv(const Interval& itv1,const Interval& itv2,PiecewiseItv& res);

/*
 *  \brief compute the complement of interval itv1 in interval itv2
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void complementItv(const PiecewiseItv& itv1,const Interval& itv2,PiecewiseItv& res);


#endif // MPC_PIECEWISEITV_HPP
