#ifndef POLYNOMIAL_HPP
#define POLYNOMIAL_HPP

#include "math/inc/Function.hpp"
#include "math/inc/MultiItvFunction.hpp"

#include <vector>

// class Function
// {
// public : 
//   /******************** Operators *******************/
//   virtual void operator*=(const Polynomial& f2) = 0; // multiply the current Polynomial by a MultiItvFunction
// };

// class mpc::MultiItvFunction;

namespace mpc
{
  class MultiItvFunction;

                                    /**************************************************/
                                    /**************** Class Polynomial ****************/
                                    /**************************************************/

  class Polynomial : public Function
  {
  protected : 
    std::vector <double> coef_;
//     double t0_;

    /********** Constructors and destructors **********/

  public : 
    Polynomial();                                      // default constructor
    Polynomial(const Polynomial& p);                   // copy constructor
    Polynomial(std::vector<double>& c);                // init constructor
    Polynomial(int deg);                               // init constructor (degree of the polynomial)
    Polynomial(double cste);                           // init constructor (constant value)
    Polynomial(double A,double B);                     // init constructor (1st order polynomial)
    Polynomial(double A,double B,double C);            // init constructor (2nd order polynomial)
    Polynomial(double A,double B,double C,double D);   // init constructor (3rd order polynomial)
    virtual ~Polynomial();                             // destructor

    /**
     *  \brief return a copy of the current function
     *  \return a copy of the current function
     *
     *  A new instance of function is allocated and returned here
     */
    virtual Polynomial* copy() const;

    /**************** Setters & Getters ***************/

  public : 
    double getCoef(int i) const;                       // return the i-th coefficient
    void setCoef(int i,double c);                      // set the i-th coefficient
    virtual double getVal(double t) const;             // return the value of the polynomial at a given parameter t
    inline int order() const { return this->coef_.size() - 1; }    // return the order of the polynomial

    /******************** Operators *******************/

  public : 
    virtual Function* operator+(double v) const;               // return the sum of a Polynomial and a constant
    virtual Function* operator+(const Polynomial& f2) const;   // return the sum of two Polynomials
    virtual Function* operator+(const MultiItvFunction& f2) const; // return the sum of a Polynomial and a MultiItvFunction
    virtual Function* operator*(double v) const;               // return the product of a Polynomial and a constant
    virtual Function* operator*(const Polynomial& f2) const;   // return the product of two Polynomials
    virtual Function* operator*(const MultiItvFunction& f2) const; // return the product of a Polynomial and a MultiItvFunction
    virtual Function* operator-(double v) const;               // return the difference of a Polynomial and a constant
    virtual Function* operator-(const Polynomial& f2) const;   // return the difference of two Polynomials  
    virtual Function* operator-(const MultiItvFunction& f2) const; // return the difference of a Polynomial and a MultiItvFunction
    virtual void operator*=(double v);                 // multiply the current Polynomial by a constante
    virtual void operator*=(const Polynomial& f2);     // multiply the current Polynomial by an other Polynomial
    virtual void operator*=(const MultiItvFunction& f2); // multiply the current Polynomial by a MultiItvFunction

    /************ Integration & Derivation ************/

  public : 
    virtual Function* derivation() const;                    // return the derivation of the current polynome
    //   virtual Function* derivation(const char* nameVar); // return the derivation of the current polynome wrt the variable named nameVar
    virtual Function* integration() const;                   // return the integration of the current polynome
    virtual Function* integration(double cste) const;        // return the integration of the current polynome with the integration constant
    virtual Function* integration(double binf,double cste) const; // return the integration of the current polynome with the integration constant

    /****************** Roots search ******************/

  public : 
    virtual int realRoots(std::vector<double>& r) const; // return the real roots of this function
    virtual int realRoots(std::vector<double>& r,double v) const; // return the roots of the equation eq - v = 0

    /**************** Utility functions ***************/
  public : 
    virtual void display() const;                      // display the polynomial
    virtual void shiftY(double v);                     // shift the function on Y axis from value v
    virtual void shiftTime(double t);                  // shift the reference time value (t0)

    /***************** Discontinuities ****************/

  public : 
    /**
     *  \brief search for the discontinuities of the function
     *  \param discPoints : resulting discontinuity points
     */
    virtual void getVecDiscPoints(std::vector<Disc_Point>& discPoints) const;
  };

};
#endif // POLYNOMIAL_HPP
