#ifndef SINFUNC_HPP
#define SINFUNC_HPP

#include "math/inc/Operation.hpp"

                                    /**************************************************/
                                    /****************** Class SinFunc *****************/
                                    /**************************************************/

namespace mpc
{
  class SinFunc : public Operation
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    SinFunc();

  public : 
    /**
     *  \brief init constructor
     *  \param f : the operand
     *
     *  Input pointer is directly assigned to the structure
     */
    SinFunc(Function* f);

//     /**
//      *  \brief init constructor
//      *  \param f : the first operand
//      *
//      *  A copy of the input functions is performed
//      */
//     SinFunc(const Function& f);

    /**
     *  \brief copy constructor
     */
    SinFunc(const SinFunc& op);

    /**
     *  \brief destructor
     */
    virtual ~SinFunc();

    /**
     *  \brief return a copy of the current function
     *  \return a copy of the current function
     *
     *  A new instance of function is allocated and returned here
     */
    virtual SinFunc* copy() const;

    /**************** Setters & Getters ***************/

  public : 
    /**
     *  \brief return the value of the polynomial at a given parameter t
     *  \param t : the input parameter value
     *  \return the value of the polynomial at a given parameter t
     */
    virtual double getVal(double t) const;
  
    /******************** Operators *******************/

  public : 
    /**
     *  \brief multiply the current function by a constant value
     *  \param v : the input multiplication coefficient
     */
    virtual void operator*=(double v);

    /************ Integration & Derivation ************/

  public : 
    /**
     *  \brief return the derivation of the current function
     *  \return the derivation of the current function
     */
    virtual Function* derivation() const;

    //   virtual Function* derivation(const char* nameVar) = 0; // return the derivation of the current polynome wrt the variable named nameVar

  protected : 
    /**
     *  \brief return the integration of the current function
     *  \return the integration of the current function
     */
    virtual Function* integration() const;

  public : 
    /**
     *  \brief return the integration of the current function
     *  \param cste : the integration constant at t=0
     *  \return the integration of the current function
     */
    virtual Function* integration(double cste) const;

    /**
     *  \brief return the integration of the current function
     *  \param cste,binf : cste is the integration constant value defined at t=binf
     *  \return the integration of the current function
     */
    virtual Function* integration(double binf,double cste) const;

    /****************** Roots search ******************/

  public : 
    /**
     *  \brief return the real roots of this function
     *  \param r : the list of roots of the function
     *  \return the number of roots of the function
     */
    virtual int realRoots(std::vector<double>& r) const;

    /**
     *  \brief return the roots of the equation eq - v = 0
     *  \param r : the list of roots of the function
     *  \param v : the input constant value
     *  \return the number of roots of the function
     */
    virtual int realRoots(std::vector<double>& r,double v) const;

    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief display the polynomial
     */
    virtual void display() const;

    /**
     *  \brief shift the function on Y axis from value v
     *  \param v : the shifting value
     */
    virtual void shiftY(double v);

    /**
     *  \brief shift the time parameter of the function
     *  \param t : the shifting value
     */
    virtual void shiftTime(double t);

    /***************** Discontinuities ****************/

    /**
     *  \brief search for the discontinuities of the function
     *  \param discPoints : resulting discontinuity points
     */
    virtual void getVecDiscPoints(std::vector<Disc_Point>& discPoints) const;
  };

};
#endif // SINFUNC_HPP
