#ifndef QUADRATURE_H_
#define QUADRATURE_H_

#include "nrtypes.h"

#include <iostream>
#include <cmath>
using namespace std;

double trapzd(double func(const double), const double a, const double b, const int n)
{
    double x,tnm,sum,del;
    static double s;
    int it,j;

    if (n == 1) {
        return (s=0.5*(b-a)*(func(a)+func(b)));
    } else {
        for (it=1,j=1;j<n-1;j++) it <<= 1;
        tnm=it;
        del=(b-a)/tnm;
        x=a+0.5*del;
        for (sum=0.0,j=0;j<it;j++,x+=del) sum += func(x);
        s=0.5*(s+(b-a)*sum/tnm);
        return s;
    }
}

double qsimp(double func(const double), const double a, const double b)
{
    const int JMAX=20;
    const double EPS=1.0e-10;
    int j;
    double s,st,ost=0.0,os=0.0;

    for (j=0;j<JMAX;j++) {
        st=trapzd(func,a,b,j+1);
        s=(4.0*st-ost)/3.0;
        if (j > 5)
            if (fabs(s-os) < EPS*fabs(os) ||
                (s == 0.0 && os == 0.0)) return s;
        os=s;
        ost=st;
    }
    fprintf(stderr,"Too many steps in routine qsimp \n");
    return 0.0;
}

void gauleg(const double x1, const double x2, double x[], double w[], int n)
{
    const double EPS=1.0e-14;
    int m,j,i;
    double z1,z,xm,xl,pp,p3,p2,p1;
    
    m=(n+1)/2;
    xm=0.5*(x2+x1);
    xl=0.5*(x2-x1);
    for (i=0;i<m;i++) {
        z=cos(3.141592654*(i+0.75)/(n+0.5));
        do {
            p1=1.0;
            p2=0.0;
            for (j=0;j<n;j++) {
                p3=p2;
                p2=p1;
                p1=((2.0*j+1.0)*z*p2-j*p3)/(j+1);
            }
            pp=n*(z*p1-p2)/(z*z-1.0);
            z1=z;
            z=z1-p1/pp;
        } while (fabs(z-z1) > EPS);
        x[i]=xm-xl*z;
        x[n-1-i]=xm+xl*z;
        w[i]=2.0*xl/((1.0-z*z)*pp*pp);
        w[n-1-i]=w[i];
    }
}


double qgaus(double func(const double), const double a, const double b)
{
    static const int n = 10;
    static double x[n];
    static double w[n];
    int j;
    double xr,xm,dx,s;
    
    gauleg(a, b, x, w, n);
    
    xm=0.5*(b+a);
    xr=0.5*(b-a);
    s=0;
    for (j=0;j<n;j++) {        
        s += w[j]*(func(x[j]));
    }
    return s;
}

double qgausParameter(double func(const double,const int), const double a, const double b,int p)
{
    static const int n = 10;
    static double x[n];
    static double w[n];
    int j;
    double xr,xm,dx,s;
    
    gauleg(a, b, x, w, n);
    
    xm=0.5*(b+a);
    xr=0.5*(b-a);
    s=0;
    for (j=0;j<n;j++) {        
        s += w[j]*(func(x[j],p));
    }
    return s;    
}


#endif /*QUADRATURE_H_*/
