#include "InvMatrix.hpp"

#include <iostream>
using namespace std;

using namespace NR;

/********** Standard operations on matrices **********/

// return the product of two matrices A and B
void prodMatDP(const Mat_DP& A, const Mat_DP& B, Mat_DP& AB)
{
  int M = A.nrows();
  int N = B.ncols();

  int M2 = A.ncols();
  int N1 = B.nrows();

  if(M2 != N1)
    {
      std::cerr << "The second dimension of A and the first dimension of B do not match. Cannot compute A*B." << std::endl;
      return;
    }

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  double sum = 0;
	  for(int k=0; k<M2; k++)
	    {
	      sum += (A[i][k] * B[k][j]);
	    }
	  if(fabs(sum) < 1e-10)
	    AB[i][j] = 0.0;
	  else
	    AB[i][j] = sum;
	}
    }

  return;
}

// return the transpose of a matrix A
void transpMatDP(const Mat_DP& A, Mat_DP& At)
{
  int M = A.nrows();
  int N = A.ncols();

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  At[j][i] = A[i][j];
	}
    }

  return;
}


/*************** Inversion of matrices ***************/

// return the inverse of a matrix A (LU decomposition)
void luInvMatDP(const Mat_DP& A, Mat_DP& Ainv)
{
  const int N=A.nrows();

  Mat_DP a2(N,N), prod(N,N);
  Vec_DP col(N);
  Vec_INT indx(N);
  double d;
  int i,j,k;

  // compute a copy of the matrix to not modify the input one
  for(i=0; i<N; i++)
    {
      for(j=0; j<N; j++)
	{
	  a2[i][j] = A[i][j];
	}
    }

  // compute the LU decomposition of the matrix
  ludcmp(a2,indx,d);
  for(j=0; j<N; j++)
    {
      for(i=0; i<N; i++) col[i] = 0.0;
      col[j] = 1.0;
      lubksb(a2,indx,col);
      for(i=0; i<N; i++) Ainv[i][j] = col[i];
    }

  return;
}

// return the pseudo inverse of a matrix A (SVD)
void svdInvMatDP(const Mat_DP& A, Mat_DP& Ainv)
{
  int M = A.nrows();
  int N = A.ncols();

  Mat_DP U(M,N);        // copie de la matrice initiale et matrice unitaire apres decomp
  Vec_DP E(N);          // vecteur de valeurs singulieres
  Mat_DP V(N,N);        // matrice adjointe

  double d;
  int i,j,k;

  // effectue une copie de la matrice initiale
  for(i=0; i<M; i++)
    {
      for(j=0; j<N; j++)
	{
	  U[i][j] = A[i][j];
	}
    }

  try{
    svdcmp(U, E, V);
  }
  catch(const std::exception& e)
    {
      cerr << "SVDCMP exception : " << e.what() << endl;
    }

  // calcule les transposees de U et V
  Mat_DP Ut(N,M); transpMatDP(U,Ut);
  Mat_DP Vt(N,N); transpMatDP(V,Vt);

  // cree la matrice contenant l'oppose de la matrice des valeurs singulieres
  Mat_DP Einv(N,N);
  for(i=0; i<N; i++)
    {
      for(j=0; j<N; j++)
	{
	  if(i == j) (Einv)[i][j] = 1.0/E[i];
	  else (Einv)[i][j] = 0;
	}
    }
  
  // calcule la matrice inverse de A
  Mat_DP prodVEinv(N,N);
  prodMatDP(V,Einv,prodVEinv);
  prodMatDP(prodVEinv,Ut,Ainv);

  return;
}

// return the inverse or pseudo inverse of a matrix A
void invMatDP(const Mat_DP& A, Mat_DP& Ainv)
{
  int M = A.nrows();
  int N = A.ncols();

  // check if one line or column is null
  bool linNul[M];
  bool colNul[N];
  int M2 = 0, N2 = 0;
  for(int i=0; i<M; i++)
    {
      bool trouve=false;
      int j=0;
      while(j<N && !trouve)
	{
	  if(A[i][j] != 0) trouve = true;
	  else j++;
	}
      if(trouve) { linNul[i] = false; M2++; }
      else linNul[i] = true;
    }
  for(int i=0; i<N; i++)
    {
      bool trouve=false;
      int j=0;
      while(j<M && !trouve)
	{
	  if(A[j][i] != 0) trouve = true;
	  else j++;
	}
      if(trouve) { colNul[i] = false; N2++; }
      else colNul[i] = true;
    }

  // construct the matrix with no line or column null
  Mat_DP A2(M2,N2);
  int i2 = 0;
  for(int i=0; i<M; i++)
    {
      if(linNul[i] == false)
	{
	  int j2 = 0;
	  for(int j=0; j<N; j++)
	    {
	      if(colNul[j] == false)
		{
		  A2[i2][j2] = A[i][j];
		  j2++;
		}
	    }
	  i2++;
	}
    }

  // check if the matrix is square or not to
  // choose which inversion method to use
  Mat_DP Ainv2(N2,M2);
  if(M2 > N2)
    {
      svdInvMatDP(A2,Ainv2);
    }
  else if(M2 == N2)
    {
      luInvMatDP(A2,Ainv2);
    }
  else // (M2 < N2)
    {
      // compute the transpose of the input matrix
      Mat_DP At(N2,M2); transpMatDP(A2,At);
      Mat_DP AtInv(M2,N2); svdInvMatDP(At,AtInv);
      transpMatDP(AtInv,Ainv2);
    }

  // add the null lines and column to the inverse matrix
  int i3 = 0;
  for(int i=0; i<N; i++)
    {
      if(colNul[i] == false)
	{
	  int j3 = 0;
	  for(int j=0; j<M; j++)
	    {
	      if(linNul[j] == false)
		{
		  Ainv[i][j] = Ainv2[i3][j3];
		  j3++;
		}
	      else
		{
		  Ainv[i][j] = 0;
		}
	    }
	  i3++;
	}
      else
	{
	  for(int j=0; j<M; j++)
	    {
	      Ainv[i][j] = 0;
	    }
	}
    }
}


/***************** Utility functions *****************/

// display a matrix A
void displayMatDP(const Mat_DP& A)
{
  int M = A.nrows();
  int N = A.ncols();

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  cout << A[i][j] << " ";
	}
      cout << endl;
    }
}
