#include "math/inc/ItgFunc.hpp"
using namespace mpc;

#include <iostream>
#include <stdlib.h>
#include <cmath>
using namespace std;

                                    /**************************************************/
                                    /****************** Class ItgFunc *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
ItgFunc::ItgFunc() : Operation(1)
{
  this->op_ = ITG_OP;
  this->cst_ = 0;

  this->setBounds(0.0,-1.0);

  dp_updated_ = false;

  this->type_ = ITG_FUNC;
}

/**
 *  \brief init constructor
 *  \param f : the operand
 *  \param cst : integration constant
 *
 *  Input pointer is directly assigned to the structure
 */
ItgFunc::ItgFunc(Function* f, double cst) : Operation(1)
{
  this->op_ = ITG_OP;
  this->tabOperand_[0] = f;
  this->cst_ = cst;

  this->setBounds(f->getBInf(),f->getBSup());

  this->updateDiscPoints();

  this->type_ = ITG_FUNC;
}

/**
 *  \brief copy constructor
 */
ItgFunc::ItgFunc(const ItgFunc& op) : Operation(op)
{
  this->cst_ = op.cst_;

  if(op.dp_updated_)
    {
      // copy the discontinuity points vector
      int N=op.disc_points_.size();
      for(int i=0; i<N; i++)
	{
	  Disc_Point newDP;
	  this->disc_points_.push_back(op.disc_points_.at(i));
	}
      this->dp_updated_ = true;
    }
  else this->updateDiscPoints();
}

/**
 *  \brief destructor
 */
ItgFunc::~ItgFunc()
{
  disc_points_.clear();
}

/**
 *  \brief return a copy of the current function
 *  \return a copy of the current function
 *
 *  A new instance of function is allocated and returned here
 */
ItgFunc* ItgFunc::copy() const
{
  return new ItgFunc(*this);
}


/**************** Quadrature tools ****************/

/*
 *  \brief Quadrature tools...
 */  
void ItgFunc::gauleg(const double x1, const double x2, double x[], double w[], int n) const
{
    const double EPS=1.0e-14;
    int m,j,i;
    double z1,z,xm,xl,pp,p3,p2,p1;
    
    m=(n+1)/2;
    xm=0.5*(x2+x1);
    xl=0.5*(x2-x1);
    for (i=0;i<m;i++) {
        z=cos(3.141592654*(i+0.75)/(n+0.5));
        do {
            p1=1.0;
            p2=0.0;
            for (j=0;j<n;j++) {
                p3=p2;
                p2=p1;
                p1=((2.0*j+1.0)*z*p2-j*p3)/(j+1);
            }
            pp=n*(z*p1-p2)/(z*z-1.0);
            z1=z;
            z=z1-p1/pp;
        } while (fabs(z-z1) > EPS);
        x[i]=xm-xl*z;
        x[n-1-i]=xm+xl*z;
        w[i]=2.0*xl/((1.0-z*z)*pp*pp);
        w[n-1-i]=w[i];
    }
}

/*
 *  \brief Quadrature tools...
 */
double ItgFunc::qgaus(double (ItgFunc::*func)(double) const, const double a, const double b) const
{
  static const int n = 20;
  double x[n];
  double w[n];
  int j;
  double s;
  
  gauleg(a, b, x, w, n);      
  s=0;
  for (j=0;j<n;j++) {        
    s += w[j]*((this->*func)(x[j]));
    }
  return s;
}


/**************** Setters & Getters ***************/

/*
 *  \brief return the value of the integrated function at a given parameter t
 *  \param t : the input parameter value
 *  \return the value of the integrated function at a given parameter t
 */
double ItgFunc::getIntegratedFuncVal(double t) const
{
  // check if the integrated function has been defined
  if(tabOperand_ == NULL || tabOperand_[0] == NULL)
    {
      cerr << "ItgFunc::getIntegratedFuncVal : Undefined integrated function" << endl;
      exit(1);
    }

  // check if the input parameter is inside of the bounds of the function
  if(t >= binf_ && t <= bsup_)
    {
      return tabOperand_[0]->getVal(t);
    }
  else return 0;
}

/**
 *  \brief return the value of the polynomial at a given parameter t
 *  \param t : the input parameter value
 *  \return the value of the polynomial at a given parameter t
 */
double ItgFunc::getVal(double t) const
{
//   // check if discontinuity points have been updated
//   if(!dp_updated_) this->updateDiscPoints();

  // search for the discontinuity point preceding the input time
  double prevT = this->binf_;
  double prevV = this->cst_;

  // check if a discontinuity point exists
  int N = disc_points_.size();
  if(N == 0)
    {
      prevT = this->binf_;
      prevV = this->cst_;
    }
  else
    {
      bool found=false;
      int i=N-1;
      while(!found && i >= 0)
	{
	  if(t >= disc_points_.at(i).t)
	    {
	      found = true;
	      prevT = disc_points_.at(i).t;
	      prevV = disc_points_.at(i).v;
	    }
	  else i--;
	}

      if(!found)
	{
	  prevT = this->binf_;
	  prevV = this->cst_;	  
	}
    }

  return (prevV + qgaus(&ItgFunc::getIntegratedFuncVal,prevT,t));
}


/******************** Operators *******************/

/**
 *  \brief multiply the current function by a constant value
 *  \param v : the input multiplication coefficient
 */
void ItgFunc::operator*=(double v)
{
  cerr << "ItgFunc::operator*= : Not defined yet!" << endl;
  exit(1);
}


/************ Integration & Derivation ************/

/**
 *  \brief return the derivation of the current function
 *  \return the derivation of the current function
 */
Function* ItgFunc::derivation() const
{
  cerr << "ItgFunc::derivation : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \return the integration of the current function
 */
Function* ItgFunc::integration() const
{
  cerr << "ItgFunc::integration : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \param cste : the integration constant at t=0
 *  \return the integration of the current function
 */
Function* ItgFunc::integration(double cste) const
{
  cerr << "ItgFunc::integration : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \param cste,binf : cste is the integration constant value defined at t=binf
 *  \return the integration of the current function
 */
Function* ItgFunc::integration(double binf,double cste) const
{
  cerr << "ItgFunc::integration : Not defined yet!" << endl;
  exit(1);
}


/****************** Roots search ******************/

/**
 *  \brief return the real roots of this function
 *  \param r : the list of roots of the function
 *  \return the number of roots of the function
 */
int ItgFunc::realRoots(std::vector<double>& r) const
{
  cerr << "ItgFunc::realRoots : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the roots of the equation eq - v = 0
 *  \param r : the list of roots of the function
 *  \param v : the input constant value
 *  \return the number of roots of the function
 */
int ItgFunc::realRoots(std::vector<double>& r,double v) const
{
  cerr << "ItgFunc::realRoots : Not defined yet!" << endl;
  exit(1);
}


/**************** Utility functions ***************/

/**
 *  \brief display the polynomial
 */
void ItgFunc::display() const
{
  cout << "[" << binf_ << ";" << bsup_ << "] -> Integration : " << endl;
  if(tabOperand_ == NULL) cout << "tabOperand nul" << endl;
  else
    {
      if(tabOperand_[0] == NULL) cout << "tabOperand[0] nul" << endl;
      else tabOperand_[0]->display();
    }
}

/**
 *  \brief shift the function on Y axis from value v
 *  \param v : the shifting value
 */
void ItgFunc::shiftY(double v)
{
  cerr << "ItgFunc::shiftY : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief shift the time parameter of the function
 *  \param t : the shifting value
 */
void ItgFunc::shiftTime(double t)
{
  if(tabOperand_ != NULL && tabOperand_[0] != NULL) tabOperand_[0]->shiftTime(t);
}


/***************** Discontinuities ****************/

/**
 *  \brief search for the discontinuities of the function
 *  \param discPoints : resulting discontinuity points
 */
void ItgFunc::getVecDiscPoints(std::vector<Disc_Point>& discPoints) const
{
  if(tabOperand_ != NULL && tabOperand_[0] != NULL) tabOperand_[0]->getVecDiscPoints(discPoints);
}

/**
 *  brief update the discontinuity points from the pointed function
 */
void ItgFunc::updateDiscPoints()
{
  // search for the discontinuities of the function of the pointed function
  // and fill the disc_points_ vector with the dicontinuity points found
  this->getVecDiscPoints(this->disc_points_);

  // set the value of the starting point of the function
  double oldT = this->binf_;
  double oldV = this->cst_;

  // compute for each disc points the value of the current integration function
  int N = disc_points_.size();
  for(int i=0; i<N; i++)
    {
      // compute the value of the integration until the current time
      double curT = disc_points_.at(i).t;
      double curV = oldV + qgaus(&ItgFunc::getIntegratedFuncVal,oldT,curT);
      disc_points_[i].v = curV;

      // switch to the current discontinuity point
      oldT = curT;
      oldV = curV;
    }

  dp_updated_ = true;
}
