#include "MathFunc.hpp"
#include "Utils.hpp"
using namespace mpc;

/*********** Geometric transformations ************/

// return the result of the translation of pos from vector v
Vec MathFunc::translate(Vec pos,Vec v)
{
  Vec pRes(2);
  pRes[0] = pos.getVal(0) + v.getVal(0);
  pRes[1] = pos.getVal(1) + v.getVal(1);
  return pRes;
}

// return the result of the rotation of pos around c from angle th
Vec MathFunc::rotate(Vec pos,Vec c,double th)
{
  Vec pRes(2);
  double dx = pos.getVal(0) - c.getVal(0);
  double dy = pos.getVal(1) - c.getVal(1);
  double n = sqrt(dx*dx + dy*dy);
  if(n == 0)
    {
      pRes.setVal(0,pos.getVal(0));
      pRes.setVal(1,pos.getVal(1));
    }
  else
    {
      double alpha = atan2(dy,dx);
      pRes.setVal(0,c.getVal(0) + n*cos(normalizedAngle(th + alpha)));
      pRes.setVal(1,c.getVal(1) + n*sin(normalizedAngle(th + alpha)));
    }
  return pRes;
}


/*************** Clothoid functions ***************/

// return x clothoid value from curve abscica lg and curve speed sig
double MathFunc::ClothX(double lg,double sigma)
{
  return (sqrt(M_PI / fabs(sigma)) * fresnel_c(sqrt(fabs(sigma) / M_PI) * lg));
}

// return x clothoid value from curve abscica lg and curve speed sig
double MathFunc::ClothY(double lg,double sigma)
{
  if(sigma >= 0)
    return (sqrt(M_PI / fabs(sigma)) * fresnel_s(sqrt(fabs(sigma) / M_PI) * lg));
  else
    return (-sqrt(M_PI / fabs(sigma)) * fresnel_s(sqrt(fabs(sigma) / M_PI) * lg));
}


/************** Hermite interpolation *************/

// compute the hermite interpolation from point and tangent (P1,T1) to point and tangent (P2,T2) at abscica S
double MathFunc::hermiteItp(double P1,double P2,double T1,double T2,double S)
{
  // compute the 4 hermite coefficients
  double h1 = 2*S*S*S - 3*S*S + 1;
  double h2 = -2*S*S*S + 3*S*S;
  double h3 = S*S*S - 2*S*S + S;
  double h4 = S*S*S - S*S;
  
  // compute the hermite interpolation value
  return (h1*P1 + h2*P2 + h3*T1 + h4*T2);
}


/*************** Distances computing **************/

// compute the maximal travelable distance from initial speed initSp during  
// a time dt with speed and acceleration constraints maxSp and maxAcc
double MathFunc::maxTravDist(double initSp,double maxSp,double maxAcc,double dt)
{
  double dtInt = (maxSp - initSp) / maxAcc;
  if(dt < dtInt)
    {
      return initSp*dt + 0.5*maxAcc*dt*dt;
    }
  else
    {
      return initSp*dtInt + 0.5*maxAcc*dtInt*dtInt + maxSp*(dt-dtInt);
    }
}

// compute the minimal travelable distance from initial speed initSp during  
// a time dt with speed and acceleration constraints minSp and minAcc
double MathFunc::minTravDist(double initSp,double minSp,double minAcc,double dt)
{
  double dtInt = (minSp - initSp) / minAcc;
  if(dt < dtInt)
    {
      return initSp*dt + 0.5*minAcc*dt*dt;
    }
  else
    {
      return initSp*dtInt + 0.5*minAcc*dtInt*dtInt + minSp*(dt-dtInt);
    }
}
