#include "MultiItvFunction.hpp"

#include "Polynomial.hpp"
#include "Utils.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

#include <stdio.h>

#define EPS 1e-6
#define EPS2 1e-10


//                  /**************************************************************************************/
//                  /********************************** ItvFunction class *********************************/
//                  /**************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
ItvFunction::ItvFunction()
{
  this->binf_ = 0;
  this->bsup_ = 0;
  this->fct_ = NULL;
}

// copy constructor
ItvFunction::ItvFunction(const ItvFunction& itvf)
{
  this->binf_ = itvf.binf_;
  this->bsup_ = itvf.bsup_;
  if(itvf.fct_ != NULL)
    {
      Function* copyFct = itvf.fct_->copy();
      this->fct_  = copyFct;
    }
  else this->fct_ = NULL;
}

// init constructor
ItvFunction::ItvFunction(double binf,double bsup,Function* f)
{
  this->binf_ = binf;
  this->bsup_ = bsup;
  this->fct_  = f;

  if(f != NULL)
    {
      f->binf_ = binf;
      f->bsup_ = bsup;
    }
}

// destructor
ItvFunction::~ItvFunction()
{
  if(fct_ != NULL) delete fct_;
}


/**************** Utility functions ***************/

/**
 *  \brief shift the time bounds of the current interval
 *  \param dt : the shifting time value
 */
void ItvFunction::shiftTime(double dt)
{
  this->binf_ += dt;
  this->bsup_ += dt;

  if(this->fct_ != NULL) this->fct_->shiftTime(dt);
}

// return true if t \in [binf_;bsup_]
bool ItvFunction::isDefinedAt(double t) const
{
  if(t >= this->binf_ && t <= this->bsup_)
    return true;
  else 
    return false;
}

// display the interval
void ItvFunction::display() const
{
  cout << "[" << this->binf_ << "," << this->bsup_ << "]";
  if(fct_ != NULL) 
    {
      cout << " : "; 
      fct_->display();
    }
  else cout << endl;
}

/**
 *  \brief get back the polynomial coefficients of the current interval if it corresponds to a ploynomial function
 *  \param polyCoef : the resulting coefficients of the polynomial function
 *  \param binf : the lower boudary of the interval
 *  \param bsup : the upper boudary of the interval
 *  \return the number of values in the resulting vector, or -1 if the current function is not a polynomial function
 */
int ItvFunction::getPolyCoef(vector<double>& polyCoef, double& binf, double& bsup) const
{
  if(fct_ != NULL) 
    {
      binf = fct_->getBInf();
      bsup = fct_->getBSup();
      
      if(fct_->getType() == POLY_FUNC)
	{
	  for(int i=0; i <= ((Polynomial*)fct_)->order(); i++)
	    {
	      polyCoef.push_back( ((Polynomial*)fct_)->getCoef(i) );
	    }
	  return ((Polynomial*)fct_)->order() + 1;
	}
    }
  else return -1;
}


/**************** Static functions ****************/

// return true if t is include in one of the intervals
bool ItvFunction::isIncludeIn(double t,std::list<ItvFunction>* l)
{
  list<ItvFunction>::iterator it = l->begin();
  while(it != l->end())
    {
      if(it->isDefinedAt(t)) return true;
      else it++;
    }
  return false;
}


                 /**************************************************************************************/
                 /******************************** MultiItvFunction class ******************************/
                 /**************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
MultiItvFunction::MultiItvFunction() : Function()
{
  this->vecItvFunc_ = new list<ItvFunction*>();
  this->setBounds(0.0,-1.0);

  this->type_ = MULTIITV_FUNC;
}

// copy constructor
MultiItvFunction::MultiItvFunction(const MultiItvFunction& f)
{
  this->vecItvFunc_ = new list<ItvFunction*>();
  if(f.vecItvFunc_ != NULL)
    {
      list<ItvFunction*>::iterator it = f.vecItvFunc_->begin();
      while(it != f.vecItvFunc_->end())
	{
	  this->vecItvFunc_->push_back(new ItvFunction(**it));
	  it++;
	}
    }

  this->setBounds(f.getBInf(),f.getBSup());

  this->type_ = MULTIITV_FUNC;
}

// init constructor
MultiItvFunction::MultiItvFunction(const Function* f,double binf,double bsup)
{
  Function* fcopy = f->copy();

  if(binf == 0 && bsup == -1)
    {
      binf = f->getBInf();
      bsup = f->getBSup();
    }
//   else
//     {
//       fcopy->binf_ = binf;
//       fcopy->bsup_ = bsup;
//     }

  this->vecItvFunc_ = new list<ItvFunction*>();
  ItvFunction* itv = new ItvFunction(binf, bsup, fcopy);
  this->vecItvFunc_->push_back(itv);

//   this->setBounds(f->getBInf(),f->getBSup());
  this->setBounds(binf,bsup);

  this->type_ = MULTIITV_FUNC;
}

/**
 *  \brief init constructor (from a vector of polynomial coefficients)
 *  \param vecPolys : the vector of polynomial coefficients
 *
 *  The structure of each interval vector must be described as follows : 
 *  vecPolys.at(i).at(0) -> lower boundary of the current interval
 *  vecPolys.at(i).at(1) -> upper boundary of the current interval
 *  vecPolys.at(i).at(2+) -> coefficient of degree 0+ of the current interval
 */
MultiItvFunction::MultiItvFunction(const std::vector<std::vector<double> >& vecPolys)
{
  this->vecItvFunc_ = new list<ItvFunction*>();

  for(int i=0; i<vecPolys.size(); i++)
    {
      Polynomial* curPoly;

      if(vecPolys.at(i).size() == 3)
	{
	  curPoly = new Polynomial(vecPolys.at(i).at(2));
	}
      else if(vecPolys.at(i).size() == 4)
	{
	  curPoly = new Polynomial(vecPolys.at(i).at(3), 
				   vecPolys.at(i).at(2));
	}
      else if(vecPolys.at(i).size() == 5)
	{
	  curPoly = new Polynomial(vecPolys.at(i).at(4), 
				   vecPolys.at(i).at(3), 
				   vecPolys.at(i).at(2));
	}
      else if(vecPolys.at(i).size() == 6)
	{
	  curPoly = new Polynomial(vecPolys.at(i).at(5), 
				   vecPolys.at(i).at(4), 
				   vecPolys.at(i).at(3), 
				   vecPolys.at(i).at(2));
	}
      else
	{
	  cerr << "MultiItvFunction::MultiItvFunction : Unhandled number of polynomial parameters" << endl;
	  exit(1);
	}

      ItvFunction* itv = new ItvFunction(vecPolys.at(i).at(0), 
					 vecPolys.at(i).at(1), 
					 curPoly);

      this->vecItvFunc_->push_back(itv);
    }

  this->setBounds(vecPolys.at(0).at(0),
		  vecPolys.at(vecPolys.size()-1).at(1));

  this->type_ = MULTIITV_FUNC;
}

// destructor
MultiItvFunction::~MultiItvFunction()
{
  if(this->vecItvFunc_ != NULL)
    {
      this->clearItv();

      delete this->vecItvFunc_;
      this->vecItvFunc_ = NULL;
    }
}

/**
 *  \brief return a copy of the current function
 *  \return a copy of the current function
 *
 *  A new instance of function is allocated and returned here
 */
MultiItvFunction* MultiItvFunction::copy() const
{
  MultiItvFunction* res = new MultiItvFunction();

  if(this->vecItvFunc_ != NULL)
    {
      list<ItvFunction*>::iterator it = this->vecItvFunc_->begin();
      while(it != this->vecItvFunc_->end())
	{
	  res->vecItvFunc_->push_back(new ItvFunction(**it));
	  it++;
	}
    }

  res->setBounds(this->getBInf(),this->getBSup());

  return res;
}


/**************** Setters & Getters ***************/

// return the value of the polynomial at a given parameter t
double MultiItvFunction::getVal(double t) const
{
  Function* f = this->getFunction(t);
  if(f == NULL)
    {
      cout << "AAAAAAAAAARRRRRRRRRRGGGGGGGGGGGGGHHHHHHHHHHHHHHHH"; // << endl;
      cout << "  t : " << t << endl;
      this->display();
      return 0.0;
    }
  else 
    return f->getVal(t);

//   double binf_func;
//   Function* f = this->getFunctionBInf(t, binf_func);
//   if(f == NULL)
//     {
//       cout << "AAAAAAAAAARRRRRRRRRRGGGGGGGGGGGGGHHHHHHHHHHHHHHHH"; // << endl;
//       cout << "  t : " << t << endl;
//       this->display();
//       return 0.0;
//     }
//   else 
//     return f->getVal(t - binf_func);
}

// return the function used at the input parameter t
Function* MultiItvFunction::getFunction(double t) const
{
  if(this->vecItvFunc_ == NULL)
    {
      cout << "AAAAAAAAAARRRRRRRRRRGGGGGGGGGGGGGHHHHHHHHHHHHHHHH111111111111" << endl;
      return NULL;
    }
  else
    {
      list<ItvFunction*>::iterator it = this->vecItvFunc_->begin();
      while(it != this->vecItvFunc_->end())
	{
	  ItvFunction* curItv = *it;
	  if(curItv->isDefinedAt(t))
	    {
	      return curItv->fct_;
	    }
	  else it++;
	}
      cout << "AAAAAAAAAARRRRRRRRRRGGGGGGGGGGGGGHHHHHHHHHHHHHHHH22222222222" << endl;      

      {
	it = this->vecItvFunc_->begin();
	while(it != this->vecItvFunc_->end())
	  {
	    ItvFunction* curItv = *it;
	    if(t < curItv->binf_) cout << "t = " << t  << " < curItv->binf_ = " << curItv->binf_ << endl;
	    else cout << "t = " << t  << " >= curItv->binf_ = " << curItv->binf_ << endl;

	    if(t > curItv->bsup_) cout << "t = " << t  << " > curItv->bsup_ = " << curItv->bsup_ << endl;
	    else cout << "t = " << t  << " <= curItv->bsup_ = " << curItv->bsup_ << endl;

	    printf("t = %.20lf      curItv->bsup_ = %.20lf \n", t, curItv->bsup_);

	    it++;
	  }
// 	throw 1;
      }

      return NULL;
    }
}

/**
 *  \brief return the function used at the input parameter t
 *  \param t : the input parameter
 *  \param binf : the resulting lower bound
 *  \return a pointer to the function defined at the value t if one is defined, return NULL else
 */
Function* MultiItvFunction::getFunctionBInf(double t, double& binf) const
{
  if(this->vecItvFunc_ == NULL)
    {
      cout << "AAAAAAAAAARRRRRRRRRRGGGGGGGGGGGGGHHHHHHHHHHHHHHHH111111111111" << endl;
      return NULL;
    }
  else
    {
      list<ItvFunction*>::iterator it = this->vecItvFunc_->begin();
      while(it != this->vecItvFunc_->end())
	{
	  ItvFunction* curItv = *it;
	  if(curItv->isDefinedAt(t))
	    {
	      binf = curItv->binf_;

	      return curItv->fct_;
	    }
	  else it++;
	}
      cout << "AAAAAAAAAARRRRRRRRRRGGGGGGGGGGGGGHHHHHHHHHHHHHHHH22222222222" << endl;      

      return NULL;
    }
}

// return the minimal bound of the function
double MultiItvFunction::getBInf() const
{
  if(this->vecItvFunc_ != NULL && this->vecItvFunc_->size() != 0)
    {
      list<ItvFunction*>::iterator it = this->vecItvFunc_->begin();
      return (*it)->binf_;
    }
  else 
    {
      return this->binf_;
    }
}

// return the maximal bound of the function
double MultiItvFunction::getBSup() const
{
  if(this->vecItvFunc_ != NULL && this->vecItvFunc_->size() != 0)
    {
      // go to the end of the list
      list<ItvFunction*>::iterator it = this->vecItvFunc_->end();
      // replace on the last interval
      it--;

      return (*it)->bsup_;
    }
  else return this->bsup_;
}

// destroy all intervals of the function
void MultiItvFunction::clearItv()
{
  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      delete (*it);
    }
  this->vecItvFunc_->clear();
}

/**
 *  \brief get back the polynomial coefficients of the current interval if it corresponds to a ploynomial function
 *  \param i : the input index of the resulting polynomial interval
 *  \param polyCoef : the resulting coefficients of the polynomial function
 *  \param binf : the lower boudary of the interval
 *  \param bsup : the upper boudary of the interval
 *  \return the number of values in the resulting vector, or -1 if the current function is not a polynomial function
 */
int MultiItvFunction::getPolyCoef(int i, std::vector<double>& polyCoef, double& binf, double& bsup) const
{
  list<ItvFunction*>::iterator it = this->vecItvFunc_->begin();
  for(int j=0; j<i; j++) it++;

  return (*it)->getPolyCoef(polyCoef, binf, bsup);
}

/**
 *  \brief get back the polynomial coefficients of the current interval if it corresponds to a ploynomial function
 *  \param polyCoef : the resulting coefficients of the polynomial function
 *  \return 0 if the process succeed, -1 else
 *
 *  The structure of each interval vector is described as follows : 
 *  vec.at(0) -> lower boundary of the current interval
 *  vec.at(1) -> upper boundary of the current interval
 *  vec.at(2+) -> coefficient of degree 0+ of the current interval
 */
int MultiItvFunction::getPolyCoef(std::vector<std::vector<double> >& polyCoef) const
{
  list<ItvFunction*>::iterator it;

  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      double curBInf;
      double curBSup;
      vector<double> curPolyCoef;
      int curRes = (*it)->getPolyCoef(curPolyCoef, curBInf, curBSup);
      
      if(curRes == -1) 
	{
	  cerr << "MultiItvFunction::getPolyCoef : Warning. The current interval does not contain a polynomial function" << endl;
	  return -1;
	}
      else 
	{	
	  vector<double> vecCurItv;
	  vecCurItv.push_back(curBInf);
	  vecCurItv.push_back(curBSup);
	  for(int j=0; j<curPolyCoef.size(); j++) vecCurItv.push_back(curPolyCoef.at(j));
	  
	  polyCoef.push_back(vecCurItv);
	}
    }

  return 0;
}

// /**
//  *  \brief get back the polynomial coefficients of the current interval if it corresponds to a ploynomial function
//  *  \param i : the input index of the resulting polynomial interval
//  *  \param polyCoef : the resulting coefficients of the polynomial function
//  *  \param binf : the lower boudary of the interval
//  *  \param bsup : the upper boudary of the interval
//  *  \return 0 if the process succeed, -1 else
//  */
// int MultiItvFunction::getPolyCoef(std::vector<std::vector<double> >& polyCoef, double& binf, double& bsup) const
// {
//   list<ItvFunction*>::iterator it;

//   for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
//     {
//       vector<double> curPolyCoef;
//       int curRes = (*it)->getPolyCoef(curPolyCoef, binf, bsup);
      
//       if(curRes == -1) 
// 	{
// // 	  throw std::string("MultiItvFunction::getPolyCoef : Warning. The current interval does not contain a polynomial function");
// 	  cerr << "MultiItvFunction::getPolyCoef : Warning. The current interval does not contain a polynomial function" << endl;
// 	  return -1;
// 	}
//       else polyCoef.push_back(curPolyCoef);
//     }

//   return 0;
// }


/******************** Operators *******************/

// multiply the current Polynomial by a constante
void MultiItvFunction::operator*=(double v)
{
  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      (*((*it)->fct_)) *= v;
    }
}

// multiply the current Polynomial by an other Polynomial
void MultiItvFunction::operator*=(const Polynomial& f2)
{
  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      *((Polynomial*)((*it)->fct_)) *= f2;
    }
}

// multiply the current Polynomial by a MultiItvFunction
void MultiItvFunction::operator*=(const MultiItvFunction& f2)
{
  // compute a copy of the current MultiItvFunction
  MultiItvFunction f1 = MultiItvFunction(*this);

  // clear the current function
  this->clearItv();

  list<ItvFunction*>::iterator it1 = f1.vecItvFunc_->begin();
  list<ItvFunction*>::iterator it2 = f2.vecItvFunc_->begin();
  
  // check if both lists are not empty
  if(it1 == f1.vecItvFunc_->end() || it2 == f2.vecItvFunc_->end())
    {
      this->binf_ = 0;
      this->bsup_ = 0;
      return;
    }

  // compute the minimal bound of the definition set of this function
  this->binf_ = min((*it1)->binf_,(*it2)->binf_);
  double bInf = this->binf_;
  
  while(it1 != f1.vecItvFunc_->end() && it2 != f2.vecItvFunc_->end())
    {
      // create the new function
      Function* newFct = *((Polynomial*) (*it1)->fct_) * *((Polynomial*) (*it2)->fct_);

      // get back the minimal superior bound from both intervals
      double bSup;
      if((*it1)->bsup_ == (*it2)->bsup_)
	{
	  bSup = (*it1)->bsup_;
	  it1++;
	  it2++;
	}
      else if((*it1)->bsup_ < (*it2)->bsup_)
	{
	  bSup = (*it1)->bsup_;
	  it1++;
	}
      else
	{
	  bSup = (*it2)->bsup_;
	  it2++;
	}

      // create a new interval
      ItvFunction* newItv = new ItvFunction(bInf,bSup,newFct);
      // append it to the list of intervals
      this->vecItvFunc_->push_back(newItv);

      bInf = bSup;
    }

  this->bsup_ = bInf;
}

// add the current MultiItvFunction to a MultiItvFunction
void MultiItvFunction::operator+=(const MultiItvFunction& f2)
{
  // compute a copy of the current MultiItvFunction
  MultiItvFunction f1 = MultiItvFunction(*this);

  // clear the current function
  this->clearItv();

  list<ItvFunction*>::iterator it1 = f1.vecItvFunc_->begin();
  list<ItvFunction*>::iterator it2 = f2.vecItvFunc_->begin();
  
  // check if both lists are not empty
  if(it1 == f1.vecItvFunc_->end() || it2 == f2.vecItvFunc_->end())
    {
      this->binf_ = 0;
      this->bsup_ = 0;
      return;
    }

  // compute the minimal bound of the definition set of this function
  this->binf_ = min((*it1)->binf_,(*it2)->binf_);
  double bInf = this->binf_;
  
  while(it1 != f1.vecItvFunc_->end() && it2 != f2.vecItvFunc_->end())
    {
      // create the new function
      Function* newFct = *((Polynomial*) (*it1)->fct_) + *((Polynomial*) (*it2)->fct_);

      // get back the minimal superior bound from both intervals
      double bSup;
      if((*it1)->bsup_ == (*it2)->bsup_)
	{
	  bSup = (*it1)->bsup_;
	  it1++;
	  it2++;
	}
      else if((*it1)->bsup_ < (*it2)->bsup_)
	{
	  bSup = (*it1)->bsup_;
	  it1++;
	}
      else
	{
	  bSup = (*it2)->bsup_;
	  it2++;
	}

      // create a new interval
      ItvFunction* newItv = new ItvFunction(bInf,bSup,newFct);
      // append it to the list of intervals
      this->vecItvFunc_->push_back(newItv);

      bInf = bSup;
    }

  this->bsup_ = bInf;
}

// add the current MultiItvFunction to a MultiItvFunction
void MultiItvFunction::operator-=(const MultiItvFunction& f2)
{
  // compute a copy of the current MultiItvFunction
  MultiItvFunction f1 = MultiItvFunction(*this);

  // clear the current function
  this->clearItv();

  list<ItvFunction*>::iterator it1 = f1.vecItvFunc_->begin();
  list<ItvFunction*>::iterator it2 = f2.vecItvFunc_->begin();
  
  // check if both lists are not empty
  if(it1 == f1.vecItvFunc_->end() || it2 == f2.vecItvFunc_->end())
    {
      this->binf_ = 0;
      this->bsup_ = 0;
      return;
    }

  // compute the minimal bound of the definition set of this function
  this->binf_ = min((*it1)->binf_,(*it2)->binf_);
  double bInf = this->binf_;
  
  while(it1 != f1.vecItvFunc_->end() && it2 != f2.vecItvFunc_->end())
    {
      // create the new function
      Function* newFct = *((Polynomial*) (*it1)->fct_) - *((Polynomial*) (*it2)->fct_);

      // get back the minimal superior bound from both intervals
      double bSup;
      if((*it1)->bsup_ == (*it2)->bsup_)
	{
	  bSup = (*it1)->bsup_;
	  it1++;
	  it2++;
	}
      else if((*it1)->bsup_ < (*it2)->bsup_)
	{
	  bSup = (*it1)->bsup_;
	  it1++;
	}
      else
	{
	  bSup = (*it2)->bsup_;
	  it2++;
	}

      // create a new interval
      ItvFunction* newItv = new ItvFunction(bInf,bSup,newFct);
      // append it to the list of intervals
      this->vecItvFunc_->push_back(newItv);

      bInf = bSup;
    }

  this->bsup_ = bInf;
}


/************ Integration & Derivation ************/

// return the derivation of the current polynome
MultiItvFunction* MultiItvFunction::derivation() const 
{
  MultiItvFunction* deriv = new MultiItvFunction();

  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      ItvFunction* curItv = *it;
      Function* curDeriv = curItv->fct_->derivation();
      curDeriv->setBounds(curItv->fct_->getBInf(),curItv->fct_->getBSup());

      ItvFunction* newItv = new ItvFunction(curItv->binf_,curItv->bsup_,curDeriv);

      deriv->vecItvFunc_->push_back(newItv);
    }

  deriv->setBounds(this->getBInf(),this->getBSup());
  return deriv;
}

// // return the derivation of the current polynome wrt the variable named nameVar
// Function* MultiItvFunction::derivation(const char* nameVar)
// {
//   return NULL;
// }

// return the integration of the current polynome
MultiItvFunction* MultiItvFunction::integration() const 
{
  MultiItvFunction* itg = new MultiItvFunction();

  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      ItvFunction* curItv = *it;
      Function* curItg = curItv->fct_->integration();
      curItg->setBounds(curItv->fct_->getBInf(),curItv->fct_->getBSup());

      ItvFunction* newItv = new ItvFunction(curItv->binf_,curItv->bsup_,curItg);

      itg->vecItvFunc_->push_back(newItv);
    }

  itg->setBounds(this->getBInf(),this->getBSup());
  return itg;
}

// return the integration of the current polynome with the integration constant
MultiItvFunction* MultiItvFunction::integration(double cste) const 
{
  MultiItvFunction* itg = new MultiItvFunction();

  int nbitv = this->vecItvFunc_->size();

  double lastCste = cste;

  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      ItvFunction* curItv = *it;
      Function* curItg = curItv->fct_->integration(lastCste);
      curItg->setBounds(curItv->fct_->getBInf(),curItv->fct_->getBSup());

      ItvFunction* newItv = new ItvFunction(curItv->binf_,curItv->bsup_,curItg);

      itg->vecItvFunc_->push_back(newItv);

      // change the constante to use for the next interval
      lastCste = curItg->getVal(curItv->bsup_);
    }

  itg->setBounds(this->getBInf(),this->getBSup());
  return itg;
}

// return the integration of the current polynome with the integration constant at time binf
MultiItvFunction* MultiItvFunction::integration(double binf,double cste) const
{
  MultiItvFunction* itg = new MultiItvFunction();

  int nbitv = this->vecItvFunc_->size();

  double lastCste = 0;

  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      ItvFunction* curItv = *it;
      Function* curItg;
      if(it == this->vecItvFunc_->begin()) curItg = curItv->fct_->integration();
      else curItg = curItv->fct_->integration(curItv->binf_,lastCste);
      curItg->setBounds(curItv->fct_->getBInf(),curItv->fct_->getBSup());

      ItvFunction* newItv = new ItvFunction(curItv->binf_,curItv->bsup_,curItg);

      itg->vecItvFunc_->push_back(newItv);

      // change the constante to use for the next interval
      lastCste = curItg->getVal(curItv->bsup_);
    }  

  // shift all the function to have f(binf) = cste
  double valBinf = itg->getVal(binf);
  itg->shiftY(cste - valBinf);

  itg->setBounds(this->getBInf(),this->getBSup());
  return itg;
}


/***************** Absolute value *****************/
// return the piecewise polynomial corresponding to the absolute value
// of the current polynomial
void MultiItvFunction::addAbsoluteValue()
{
  // compute a copy of the current MultiItvFunction
  MultiItvFunction f1 = MultiItvFunction(*this);

  list<ItvFunction*>::iterator it;
  for(it = f1.vecItvFunc_->begin(); it != f1.vecItvFunc_->end(); it++)
    {
      // get back the function on the current interval
      ItvFunction* curItv = *it;

      Function* curFct = curItv->fct_;
      
      double minB = curItv->binf_;
      double maxB = curItv->bsup_;

      // get back the roots of the current function
      vector<double> roots;
      curFct->realRoots(roots);

      // just keep the roots included between the bounds
      vector<double> rootsInItv;
      getValuesBetween(minB, maxB, roots, rootsInItv);

      // construct a vector composed of the roots and the bounds
      vector<double> chgPoints;
      chgPoints.push_back(minB);
      for(int i=0; i<rootsInItv.size(); i++)
	{
	  if(rootsInItv[i] != minB && rootsInItv[i] != maxB) chgPoints.push_back(rootsInItv[i]);
	}
      chgPoints.push_back(maxB);
      
      // compute the opposite function of the current function
      Polynomial opposite(*((Polynomial*)curFct));
      opposite *= -1.0;
//       Polynomial opposite = new Polynomial(*((Polynomial*)curFct));
//       (*opposite) *= -1.0;

      // check between each couple of values if the interval has to be changed
      int N = chgPoints.size();

      for(int i=0; i<N-1; i++)
	{
	  // get back the inferior and superior bounds
	  // Warning : these values must not be used to define the bounds
	  double vmin = chgPoints[i];
	  double vmax = chgPoints[i+1];
	  if(vmin == -INFINITY) vmin = -999999999.0;
	  if(vmax == INFINITY) vmax = 999999999.0;
	  
	  double valThis = this->getVal(0.5*(vmin + vmax));
	  
//  	  cout << "chgPoints[i] : " << chgPoints[i] << "  chgPoints[i+1] : " << chgPoints[i+1] << endl;

	  if(valThis < 0)
	    {
	      // case where the function is lesser than 0
	      // we change the interval with the opposite function
	      this->changeItv(chgPoints[i],chgPoints[i+1],&opposite);	  
	    }
	}
    }
}


/****************** Roots search ******************/

// // return the real roots of this function
// int MultiItvFunction::realRoots(std::vector<double>& r) const
// {
//   int res = 0;

//   list<ItvFunction*>::iterator it;
//   for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
//     {
//       ItvFunction* curItv = *it;

//       // get back the roots of the current interval
//       vector<double> curR;
//       int curRes = curItv->fct_->realRoots(curR);

//       double bi = curItv->binf_;
//       double bs = curItv->bsup_;

//       // append them to the global roots vector
//       int curValidR = 0;
//       for(int j=0; j<curRes; j++) 
// 	{
// 	  if(curR.at(j) >= bi && curR.at(j) <= bs)
// 	    {
// 	      if(fabs(curR.at(j)) < 1e-10) r.push_back(0.0);
// 	      else r.push_back(curR.at(j));
// 	      curValidR++;
// 	    }
// 	}
//       res += curValidR;
//     }

//   return res;
// }

// return the roots of the equation eq - v = 0
int MultiItvFunction::realRoots(std::vector<double>& r,double v) const
{
  int res = 0;

  double valPrevBSup = 0;

  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      ItvFunction* curItv = *it;

      double bi = curItv->binf_;
      double bs = curItv->bsup_;

      // check if there is a discontinuity at the inferior boundary of the interval
      if(it != this->vecItvFunc_->begin())
	{
	  // get back the value of the current interval at its inferior boundary
	  double valCurBinf = curItv->fct_->getVal(bi);

	  if((valPrevBSup < v && valCurBinf > v) || (valPrevBSup > v && valCurBinf < v))
	    {
	      r.push_back(bi);
	      res++;
	    }
	}
      // get back the value of the current interval at its superior bound for the next step
      valPrevBSup = curItv->fct_->getVal(bs);

      // get back the roots of the current interval
      vector<double> curR;
      int curRes = curItv->fct_->realRoots(curR,v);

      // append them to the global roots vector
      int curValidR = 0;
      for(int j=0; j<curRes; j++) 
	{
	  if(curR.at(j) >= bi && curR.at(j) <= bs)
	    {
	      if(fabs(curR.at(j)) < 1e-10) r.push_back(0.0);
	      else r.push_back(curR.at(j));
	      curValidR++;
	    }
	}
      res += curValidR;
    }

  return res;
}


/**************** Interval changes ****************/

// change one or several intervals of the multiItvFunction by f
void MultiItvFunction::changeItv(double bi,double bs,MultiItvFunction* f)
{
  if(bi >= bs) return;

  if(f == NULL) return;
  double fbinf = bi;
  double fbsup = bs;
  
  bool trouveDeb = false;
  bool trouveFin = false;

//   cout << "changeItv -> f : " << endl;
//   f->display();

  // compute the list of functions defined between bi and bs
  list<ItvFunction*> listNewFct = f->getListItvBetween(fbinf,fbsup);

//   cout << "listNewFct -> size(): " << listNewFct->size() << endl;
//   list<ItvFunction*>::iterator iter;
//   for(iter = listNewFct->begin(); iter != listNewFct->end(); iter++)
//     {
//       //cout << "argh" << endl;
//       cout << "fct : "; (*iter)->display();
//     }
//   cout << "end listNewFct" << endl;

  if(listNewFct.size() == 0) return;

  list<ItvFunction*>::iterator it = this->vecItvFunc_->begin();
  // search where the input function must be inserted
  while(!trouveFin && it != this->vecItvFunc_->end())
    {
      // get back the function in the current interval
      ItvFunction* curItv = *it;
      
      if(it == this->vecItvFunc_->begin() && fbinf < curItv->binf_)
	{
	  trouveDeb = true;
	  
	  ItvFunction* itv = new ItvFunction(fbinf,fbsup,new MultiItvFunction(*f));
	  it = this->vecItvFunc_->insert(it, itv);
	  list<ItvFunction*>::iterator itLNF;
	  for(itLNF = listNewFct.begin(); itLNF != listNewFct.end(); itLNF++)
	    {
	      it = this->vecItvFunc_->insert(it, *itLNF);
	      it++;
	    }

	  //it++;
	}
      else if(!trouveDeb)
	{
	  if(curItv->isDefinedAt(fbinf))
	    {
	      trouveDeb = true;

	      // check if fbsup exactly equals curItv->bsup_
	      if(fbsup == curItv->bsup_)
		{
		  trouveFin = true;

		  bool eraseFirst = false;
		  if(fbinf == curItv->binf_)
		    {
		      eraseFirst = true;
		      it = this->vecItvFunc_->erase(it);
		      delete curItv;
		    }
		  else
		    {
		      // we move on the next interval to push the repacement 
		      // function after the current one
		      it++;
		    }

		  // insert all the intermediate functions
		  list<ItvFunction*>::iterator itLNF;
		  for(itLNF = listNewFct.begin(); itLNF != listNewFct.end(); itLNF++)
		    {
		      it = this->vecItvFunc_->insert(it, *itLNF);
		      it++;
		    }
		  
		  // we change the upper bound of the current interval
		  // to stop before the new intervals added
		  if(!eraseFirst)
		    curItv->bsup_ = fbinf;
		}

	      // check if the input function is completely defined
	      // inside of the current interval
	      else if(fbsup < curItv->bsup_)
		{
		  // case where the input function is completely
		  // defined inside of the current interval : 
		  // cut the current interval in two parts and
		  // insert the input function between them
		  
		  trouveFin = true;
		  
		  double oldBInf = curItv->binf_;

		  // define the last interval
		  curItv->binf_ = fbsup;

		  // insert the same function for the first interval
		  if(oldBInf != fbinf)
		    {
		      ItvFunction* prevItv = new ItvFunction(oldBInf,fbinf,curItv->fct_->copy());
		      it = this->vecItvFunc_->insert(it, prevItv);
		      it++;
		    }

		  // insert all the intermediate functions
		  list<ItvFunction*>::iterator itLNF;
		  for(itLNF = listNewFct.begin(); itLNF != listNewFct.end(); itLNF++)
		    {
		      it = this->vecItvFunc_->insert(it, *itLNF);
		      it++;
		    }
		}
	      else // (fbsup > curItv->bsup_)
		{
		  list<ItvFunction*>::iterator nextit = it;
		  nextit++;

		  // check if the current interval is the last interval
		  if(nextit == this->vecItvFunc_->end())
		    {
		      trouveFin = true;

		      // we change the end of the current interval and
		      // add the new functions after
		      curItv->bsup_ = fbinf;

		      list<ItvFunction*>::iterator itLNF;
		      for(itLNF = listNewFct.begin(); itLNF != listNewFct.end(); itLNF++)
			{
			  this->vecItvFunc_->push_back(*itLNF);
			}
		    }
		  else
		    {
		      // case where the input function overlap one or 
		      // several intervals : we change the end of the
		      // current interval, add the input function
		      // after and try to find where it stops
		      if(curItv->binf_ == fbinf)
			{
			  // delete the current interval as it is reduce to 0
			  it = this->vecItvFunc_->erase(it);
			}
		      else
			{
			  // modify the upper bound of the current interval
			  curItv->bsup_ = fbinf;
			  it++;
			}
		      
		      // insert all the intermediate functions
		      list<ItvFunction*>::iterator itLNF;
		      for(itLNF = listNewFct.begin(); itLNF != listNewFct.end(); itLNF++)
			{
			  it = this->vecItvFunc_->insert(it, *itLNF);
			  it++;
			}
			  
// 			  // ???????
// 			  ItvFunction* itv = new ItvFunction(fbinf,fbsup,f);
// 			  this->vecItvFunc_->insert(it, itv);
			
		    }
		}
	    }
	  else 
	    {
	      it++;
	    }
	}
      else
	{
	  // check if fbsup exactly equals curItv->bsup_
	  if(fbsup == curItv->bsup_)
	    {
	      trouveFin = true;
	      
	      //curItv->fct_ = f;
	      it = this->vecItvFunc_->erase(it);
	      delete curItv;
	      
	      //this->display();
	    }

	  // check if the input function stops 
	  // during the current interval
	  else if(fbsup < curItv->bsup_)
	    {
	      // in that case we just modify the end of the current interval
	      trouveFin = true;
	      curItv->binf_ = fbsup;
	    }
	  else // (fbsup > curItv->bsup_)
	    {
	      // in that case, the current interval is completely
	      // overlaped. we delete it from the function
	      it = this->vecItvFunc_->erase(it);
	      delete curItv;
	      if(it == this->vecItvFunc_->end()) trouveFin = true;
	    }
	}
    }
}

// change one or several intervals of the function by f
void MultiItvFunction::changeItv(double bi,double bs,Function* f)
{
//   cout << "///////////////////////////////////////////////////////////////" << endl;
//   cout << "avant  -> bi : " << bi << "  bs : " << bs << endl;
//   this->display();

  if(bi >= bs) return;

  if(f == NULL) return;
  double fbinf = bi;
  double fbsup = bs;

//   cout << "changeItv -> f : " << endl;
//   f->display();
  
  bool trouveDeb = false;
  bool trouveFin = false;
  //int i=0;
  list<ItvFunction*>::iterator it = this->vecItvFunc_->begin();

  // search where the input function must be inserted
  while(!trouveFin && it != this->vecItvFunc_->end())
    {
      // get back the function in the current interval
      ItvFunction* curItv = *it; //this->vecItvFunc_->at(i);
      
      if(it == this->vecItvFunc_->begin() && fbinf <= curItv->binf_)
	{
	  trouveDeb = true;
	  
	  ItvFunction* itv = new ItvFunction(fbinf,fbsup,f->copy());
	  it = this->vecItvFunc_->insert(it, itv);

	  //i++;
	  it++;
	}
      else if(!trouveDeb)
	{
	  if(curItv->isDefinedAt(fbinf))
	    {
	      trouveDeb = true;

	      // check if fbsup exactly equals curItv->bsup_
	      if(fbsup == curItv->bsup_)
		{
//  		  cout << "2" << endl;
		  trouveFin = true;

// 		  cout << "curItv : " << endl;
// 		  curItv->display();

		  ItvFunction* prevItv = new ItvFunction(curItv->binf_,fbinf,curItv->fct_);
		  it = this->vecItvFunc_->insert(it, prevItv);
		  
		  curItv->binf_ = fbinf;
		  curItv->fct_ = f->copy();
		}

	      // check if the input function is completely defined
	      // inside of the current interval
	      else if(fbsup < curItv->bsup_)
		//if(curItv->isDefinedAt(fbsup))
		{
//  		  cout << "3" << endl;
		  // case where the input function is completely
		  // defined inside of the current interval : 
		  // cut the current interval in two parts and
		  // insert the input function between them
		  
		  trouveFin = true;
		  
		  ItvFunction* itv = new ItvFunction(fbinf,fbsup,f->copy());
		  it = this->vecItvFunc_->insert(it, itv);

		  ItvFunction* prevItv = new ItvFunction(curItv->binf_,fbinf,curItv->fct_->copy());
		  it = this->vecItvFunc_->insert(it, prevItv);
		  
		  curItv->binf_ = fbsup;
		}
	      else // (fbsup > curItv->bsup_)
		{
		  list<ItvFunction*>::iterator nextit = it;
		  nextit++;

		  // check if the current interval is the last interval
		  if(nextit == this->vecItvFunc_->end())
		    {
//  		      cout << "41" << endl;
		      trouveFin = true;

		      // we change the end of the current interval and
		      // add the new function after
		      curItv->bsup_ = fbinf;

		      ItvFunction* itv = new ItvFunction(fbinf,fbsup,f->copy());
		      this->vecItvFunc_->push_back(itv);

		    }
		  else
		    {
		      // case where the input function overlap one or 
		      // several intervals : we change the end of the
		      // current interval, add the input function
		      // after and try to find where it stops

		      if(curItv->binf_ == fbinf)
			{
//  			  cout << "421" << endl;
			  curItv->bsup_ = fbsup;
			  curItv->fct_ = f->copy();

			  // move on the next interval
			  it++;
			}
		      else
			{
//  			  cout << "422" << endl;
			  curItv->bsup_ = fbinf;
			  it++;
			  
			  ItvFunction* itv = new ItvFunction(fbinf,fbsup,f->copy());
			  it = this->vecItvFunc_->insert(it, itv);
			  
			  // move on the next interval
			  it++;

			  //this->display();
			}
		    }
		  
		}
	    }
	  else 
	    {
//  	      cout << "5" << endl;
	      //i++;
	      it++;
	    }
	}
      else
	{
	  // check if fbsup exactly equals curItv->bsup_
	  if(fbsup == curItv->bsup_)
	    {
//  	      cout << "6" << endl;
	      trouveFin = true;

	      //curItv->fct_ = f;
	      it = this->vecItvFunc_->erase(it);
	      delete curItv;

	      //this->display();
	    }

	  // check if the input function stops 
	  // during the current interval
	  else if(fbsup < curItv->bsup_)
	    //if(curItv->isDefinedAt(fbsup))
	    {
//  	      cout << "7" << endl;
	      // in that case we insert the input function
	      // here and modify the end of the current interval 
	      trouveFin = true;
	      
// 	      ItvFunction* itv = new ItvFunction(fbinf,fbsup,f);
// 	      this->vecItvFunc_->insert(it, itv);
	      
	      curItv->binf_ = fbsup;
	    }
	  else // (fbsup > curItv->bsup_)
	    {
//  	      cout << "8" << endl;
	      // in that case, the current interval is completely
	      // overlaped. we delete it from the function

	      //cout << "before deletion" << endl;
	      //this->display();
	      //cout << "delete an interval : [" << (*it)->binf_ << ";" << (*it)->bsup_ << "]" << endl;
	      //cout << "next interval : [" << (*(it+1))->binf_ << ";" << (*(it+1))->bsup_ << "]" << endl;
	      
	      it = this->vecItvFunc_->erase(it);
	      //cout << "erasing done" << endl;
	      //cout << "*it after deletion : [" << (*it)->binf_ << ";" << (*it)->bsup_ << "]" << endl;

	      delete curItv;

	      //cout << "end delete" << endl;

	      if(it == this->vecItvFunc_->end()) trouveFin = true;
	    }
	}
    }

  // debugging process : check the consistency of the 
  // piecewise function after the change of intervals
//   if(!this->checkItvConsistency())
//     {
//       cout << "MultiItvFunction::changeItv(fct) : An unconsistency has been found" << endl;
//       cout << "bi : " << bi << endl;
//       cout << "bs : " << bs << endl;
//       cout << "bound : " << endl;
//       f->display();
//       cout << "apres : " << endl;
//       this->display();
//       cout << "///////////////////////////////////////////////////////////////" << endl;
//       exit(-1);
//     }
//   else 
//     {
//       cout << "apres : " << endl;
//       this->display();
//       cout << "MultiItvFunction::changeItv(fct) : Consistency OK" << endl;
//     }

  //cout << "exit changeItv" << endl;

//   cout << "apres" << endl;
//   this->display();

//   cout << "FFFFFFIIIIINNNNN changeItv(double bi,double bs,Function* f)" << endl;
}

// add a minimal bound over the function
void MultiItvFunction::addMinBound(double bmin,std::list<ItvFunction>* lRes, bool debug)
{
//   // define the list of bounded intervals
//   list<ItvFunction>* lRes = new list<ItvFunction>;

  // define the function f(t) = bmin
  Polynomial fmin(bmin);

//   cout << "fmin : " << endl;
//   fmin.display();

  // search for the roots of the current function
  vector<double> roots;
  int nbRoots = this->realRoots(roots,bmin);

//   if(debug)
//     {
//       cout << "nbRoots : " << nbRoots << endl;
//       cout << "roots : "; displayVectDble(&roots);
//       cout << endl;
//     }

  if(nbRoots == 0)
    {      
      // if no roots have been found, check if the
      // function is not totaly out of the bound
      if(this->getVal(0.0) < bmin)
	{
	  // in that case, change all the function by f(t) = bmin
// 	  double oldBInf = 0.0;
// 	  double oldBSup = INFINITY;
	  double oldBInf = this->binf_;
	  double oldBSup = this->bsup_;

	  list<ItvFunction*>::iterator it;
	  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
	    {
	      delete *it;
	    }
	  this->vecItvFunc_->clear();

	  ItvFunction* newItv = new ItvFunction(oldBInf,oldBSup,new Polynomial(fmin));
	  this->vecItvFunc_->push_back(newItv);
	}
      else
	{
	  // nothing to do
	}
    }
  else 
    {
      // test if the part before the first root is
      // over or under the bound
      if(roots[0]-EPS < this->getBInf())
	{
// 	  cout << "minB 1" << endl;
	  
	  if(this->getVal(roots[0]+EPS) > bmin)
	    {
// 	      cout << "minB 11" << endl;
	      // replace the interval before the first root
	      if(roots[0] != this->getBInf())
		{
		  this->changeItv(this->getBInf(),roots[0],&fmin);
		  if(lRes != NULL) lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));
		}
	      
	      // change even intervals between roots
	      for(int i=1; i<nbRoots; i+=2)
		{
		  if(i == nbRoots-1)
		    {
		      this->changeItv(roots[i],this->getBSup(),&fmin);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		    }
		  else
		    {
		      this->changeItv(roots[i],roots[i+1],&fmin);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		    }
		}
	    }
	  else
	    {
// 	      cout << "minB 12" << endl;
	      // change odd intervals between roots
	      for(int i=0; i<nbRoots; i+=2)
		{
		  if(i == nbRoots-1)
		    {
		      this->changeItv(roots[i],this->getBSup(),&fmin);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		    }
		  else
		    {
		      this->changeItv(roots[i],roots[i+1],&fmin);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		    }
		}
	    }
	}
      else if(this->getVal(roots[0]-EPS) < bmin)
	{
// 	  cout << "minB 2" << endl;
	  // replace the interval before the first root
	  this->changeItv(this->getBInf(),roots[0],&fmin);
	  if(lRes != NULL) lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));

	  // change even intervals between roots
	  for(int i=1; i<nbRoots; i+=2)
	    {
	      if(i == nbRoots-1) 
		{
		  this->changeItv(roots[i],this->getBSup(),&fmin);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		}
	      else 
		{
		  this->changeItv(roots[i],roots[i+1],&fmin);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		}
	    }
	}
      else
	{
// 	  cout << "minB 3" << endl;

	  // change odd intervals between roots
	  for(int i=0; i<nbRoots; i+=2)
	    {
	      if(i == nbRoots-1) 
		{
// 		  cout << "minB 31" << endl;
		  this->changeItv(roots[i],this->getBSup(),&fmin);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		}
	      else 
		{
// 		  cout << "minB 32" << endl;
		  this->changeItv(roots[i],roots[i+1],&fmin);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		}
	    }
	}
    }

//   // return the list of bounded intervals
//   return lRes;
}

// add a maximal bound over the function
void MultiItvFunction::addMaxBound(double bmax,std::list<ItvFunction>* lRes)
{
//   // define the list of bounded intervals
//   list<ItvFunction>* lRes = new list<ItvFunction>;

  // define the function f(t) = bmin
  Polynomial fmax(bmax);

  // search for the roots of the current function
  vector<double> roots;
  int nbRoots = this->realRoots(roots,bmax);

  if(nbRoots == 0)
    {
      // if no roots have been found, check if the
      // function is not totaly out of the bound
      if(this->getVal(0.0) > bmax)
	{
	  // in that case, change all the function by f(t) = bmax
	  double oldBInf = this->binf_;
	  double oldBSup = this->bsup_;

// 	  cout << "==========================================================================" << endl;
// 	  cout << "oldBInf : " << oldBInf << "  oldBSup : " << oldBSup << endl;
// 	  this->display(); 
// 	  exit(1);
// 	  cout << "==========================================================================" << endl;

	  list<ItvFunction*>::iterator it;
	  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
	    {
	      delete *it;
	    }
	  this->vecItvFunc_->clear();

	  ItvFunction* newItv = new ItvFunction(oldBInf,oldBSup,new Polynomial(fmax));
	  this->vecItvFunc_->push_back(newItv);
	}
      else
	{
	  // nothing to do
	}
    }
  else 
    {
      // test if the part before the first root is
      // over or under the bound
      if(roots[0]-EPS < this->getBInf())
	{
	  if(this->getVal(roots[0]+EPS) < bmax)
	    {
	      // replace the interval before the first root
	      if(roots[0] != this->getBInf()) 
		{
		  this->changeItv(this->getBInf(),roots[0],&fmax);
		  if(lRes != NULL) lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));
		}
	      
	      // change even intervals between roots
	      for(int i=1; i<nbRoots; i+=2)
		{
		  if(i == nbRoots-1) 
		    {
		      this->changeItv(roots[i],this->getBSup(),&fmax);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		    }
		  else 
		    {
		      this->changeItv(roots[i],roots[i+1],&fmax);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		    }
		}
	    }
	  else
	    {
	      // change odd intervals between roots
	      for(int i=0; i<nbRoots; i+=2)
		{
		  if(i == nbRoots-1) 
		    {
		      this->changeItv(roots[i],this->getBSup(),&fmax);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		    }
		  else 
		    {
		      this->changeItv(roots[i],roots[i+1],&fmax);
		      if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		    }
		}
	    }
	}
      else if(this->getVal(roots[0]-EPS) > bmax)
	{
	  // replace the interval before the first root
	  this->changeItv(this->getBInf(),roots[0],&fmax);
	  if(lRes != NULL) lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));

	  // change even intervals between roots
	  for(int i=1; i<nbRoots; i+=2)
	    {
	      if(i == nbRoots-1) 
		{
		  this->changeItv(roots[i],this->getBSup(),&fmax);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		}
	      else 
		{
		  this->changeItv(roots[i],roots[i+1],&fmax);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		}
	    }
	}
      else
	{
	  // change odd intervals between roots
	  for(int i=0; i<nbRoots; i+=2)
	    {
	      if(i == nbRoots-1) 
		{
		  this->changeItv(roots[i],this->getBSup(),&fmax);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
		}
	      else 
		{
		  this->changeItv(roots[i],roots[i+1],&fmax);
		  if(lRes != NULL) lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
		}
	    }
	}
    }

//   // return the list of bounded intervals
//   return lRes;
}

// // add a minimal bound over the function
// std::list<ItvFunction>* MultiItvFunction::addMinBound2(MultiItvFunction* fMinB)
// {
//   // define the list of bounded intervals
//   list<ItvFunction>* lRes = new list<ItvFunction>;

//   MultiItvFunction fdiff(*this);
//   fdiff -= (*fMinB);

//   // search for the roots of the difference function
//   vector<double> roots;
//   int nbRoots = fdiff.realRoots(roots,0.0);

//   if(nbRoots == 0)
//     {
//       // if no roots have been found, check if the
//       // function is not totaly out of the bound
//       if(this->getVal(0.0) < fMinB->getVal(0.0))
// 	{
// 	  // in that case, change all the function by f(t) = bmax
// 	  this->clearItv();

// 	  std::list<ItvFunction*>::iterator it2;
// 	  for(it2 = fMinB->vecItvFunc_->begin(); it2 != fMinB->vecItvFunc_->end(); it2++)
// 	    {
// 	      ItvFunction* newItv = new ItvFunction(**it2);
// 	      this->vecItvFunc_->push_back(newItv);
// 	    }
// 	  this->binf_ = fMinB->getBInf();
// 	  this->bsup_ = fMinB->getBSup();
// 	}
//       else
// 	{
// 	  // nothing to do
// 	}
//     }

//   else if(nbRoots == 1)
//     {
//       // compute an epsilon value as great as possible 
//       // to avoid numerical computations errors
//       double eps;
//       if(this->getBSup() > roots[0] + 1000000.0)
// 	eps = 500000.0;
//       else
// 	eps = 0.5 * (this->getBSup() - roots[0]);

//       if(this->getVal(roots[0]+eps) > fMinB->getVal(roots[0]+eps))
// 	{
// 	  // replace the interval before the root
// 	  if(roots[0] != this->getBInf())
// 	    {
// 	      this->changeItv(this->getBInf(),roots[0],fMinB);
// 	      lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));
// 	    }
// 	}
//       else
// 	{
// 	  // replace the interval after the root
// 	  if(roots[0] != this->getBSup())
// 	    {
// 	      this->changeItv(roots[0],this->getBSup(),fMinB);
// 	      lRes->push_back(ItvFunction(roots[0],this->getBSup(),NULL));
// 	    }	  
// 	}
//     }

//   else // (nbRoots >= 2)
//     {
//       double eps = 0.5 * (roots[1] - roots[0]);

//       if(this->getVal(roots[0]+eps) > fMinB->getVal(roots[0]+eps))
// 	{
// 	  // replace the interval before the first root
// 	  if(roots[0] != this->getBInf())
// 	    {
// 	      this->changeItv(this->getBInf(),roots[0],fMinB);
// 	      lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));	      
// 	    }
	      
// 	  // change even intervals between roots
// 	  for(int i=1; i<nbRoots; i+=2)
// 	    {
// 	      if(i == nbRoots-1) 
// 		{
// 		  this->changeItv(roots[i],this->getBSup(),fMinB);
// 		  lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
// 		}
// 	      else 
// 		{
// 		  this->changeItv(roots[i],roots[i+1],fMinB);
// 		  lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
// 		}
// 	    }
// 	}
//       else
// 	{
// 	  // change odd intervals between roots
// 	  for(int i=0; i<nbRoots; i+=2)
// 	    {
// 	      if(i == nbRoots-1)
// 		{
// 		  this->changeItv(roots[i],this->getBSup(),fMinB);
// 		  lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
// 		}
// 	      else 
// 		{
// 		  this->changeItv(roots[i],roots[i+1],fMinB);
// 		  lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
// 		}
// 	    }
// 	}
//     }

//   // return the list of bounded intervals
//   return lRes;
// }

// // add a maximal bound over the function
// std::list<ItvFunction>* MultiItvFunction::addMaxBound2(MultiItvFunction* fMaxB)
// {
//   // define the list of bounded intervals
//   list<ItvFunction>* lRes = new list<ItvFunction>;

//   MultiItvFunction fdiff(*fMaxB);
//   fdiff -= (*this);
  
//   // search for the roots of the difference function
//   vector<double> roots;
//   int nbRoots = fdiff.realRoots(roots,0.0);

//   cout << "nbRoots : " << nbRoots << endl;
//   cout << "roots : "; displayVectDble(&roots);
  
//   if(nbRoots == 0)
//     { 
//       // if no roots have been found, check if the
//       // function is not totaly out of the bound
//       if(this->getVal(0.0) > fMaxB->getVal(0.0))
// 	{
// 	  // in that case, change all the function by f(t) = bmax
// 	  this->clearItv();

// 	  std::list<ItvFunction*>::iterator it2;
// 	  for(it2 = fMaxB->vecItvFunc_->begin(); it2 != fMaxB->vecItvFunc_->end(); it2++)
// 	    {
// 	      ItvFunction* newItv = new ItvFunction(**it2);
// 	      this->vecItvFunc_->push_back(newItv);
// 	    }
// 	  this->binf_ = fMaxB->getBInf();
// 	  this->bsup_ = fMaxB->getBSup();
// 	}
//       else
// 	{
// 	  // nothing to do
// 	}
//     }
  
//   else if(nbRoots == 1)
//     {
//       // compute an epsilon value as great as possible 
//       // to avoid numerical computations errors
//       double eps;
//       if(this->getBSup() > roots[0] + 1000000.0)
// 	eps = 500000.0;
//       else
// 	eps = 0.5 * (this->getBSup() - roots[0]);

//       cout << "this->getBSup() : " << this->getBSup() << endl;
//       cout << "eps : " << eps << endl;
//       cout << "this->getVal(roots[0]+eps) : " << this->getVal(roots[0]+eps) << endl;
//       cout << "fMaxB->getVal(roots[0]+eps) : " << fMaxB->getVal(roots[0]+eps) << endl;

//       if(this->getVal(roots[0]+eps) < fMaxB->getVal(roots[0]+eps))
// 	{
// 	  cout << "case 1" << endl;
// 	  // replace the interval before the root
// 	  if(roots[0] != this->getBInf())
// 	    {
// 	      this->changeItv(this->getBInf(),roots[0],fMaxB);
// 	      lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));
// 	    }
// 	}
//       else
// 	{
// 	  cout << "case 2" << endl;
// 	  // replace the interval after the root
// 	  if(roots[0] != this->getBSup())
// 	    {
// 	      this->changeItv(roots[0],this->getBSup(),fMaxB);
// 	      lRes->push_back(ItvFunction(roots[0],this->getBSup(),NULL));
// 	    }	  
// 	}
//     }
  
//   else // (nbRoots >= 2)
//     {
//       double eps = 0.5 * (roots[1] - roots[0]);
    
//       if(this->getVal(roots[0]+eps) < fMaxB->getVal(roots[0]+eps))
// 	{
// 	  // replace the interval before the first root
// 	  if(roots[0] != this->getBInf()) 
// 	    {
// 	      this->changeItv(this->getBInf(),roots[0],fMaxB);
// 	      lRes->push_back(ItvFunction(this->getBInf(),roots[0],NULL));
// 	    }
	  
// 	  // change even intervals between roots
// 	  for(int i=1; i<nbRoots; i+=2)
// 	    {
// 	      if(i == nbRoots-1) 
// 		{
// 		  this->changeItv(roots[i],this->getBSup(),fMaxB);
// 		  lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
// 		}
// 	      else 
// 		{
// 		  this->changeItv(roots[i],roots[i+1],fMaxB);
// 		  lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
// 		}
// 	    }
// 	}
//       else
// 	{
// 	  // change odd intervals between roots
// 	  for(int i=0; i<nbRoots; i+=2)
// 	    {
// 	      if(i == nbRoots-1) 
// 		{
// 		  this->changeItv(roots[i],this->getBSup(),fMaxB);
// 		  lRes->push_back(ItvFunction(roots[i],this->getBSup(),NULL));
// 		}
// 	      else 
// 		{
// 		  this->changeItv(roots[i],roots[i+1],fMaxB);
// 		  lRes->push_back(ItvFunction(roots[i],roots[i+1],NULL));
// 		}
// 	    }
// 	}      
//     }

//   // return the list of bounded intervals
//   return lRes;
// }

// add a minimal bound over the function, return the bounded intervals
void MultiItvFunction::addMinBound(MultiItvFunction* fMinB,std::list<ItvFunction>* lRes)
{
//   // define the list of bounded intervals
//   list<ItvFunction>* lRes = new list<ItvFunction>;

  MultiItvFunction fdiff(*this);
  fdiff -= (*fMinB);

//   MultiItvFunction fdiff(*fMinB);
//   fdiff -= (*this);
  
  // search for the roots of the difference function
  vector<double> roots;
  int nbRoots = fdiff.realRoots(roots,0.0);
  
  vector<double> rootBoundDisc;
  
  // add the inferior and superior bounds
  rootBoundDisc.push_back(this->getBInf());
  rootBoundDisc.push_back(this->getBSup());
  
  // add the roots
  for(int i=0; i<nbRoots; i++)
    rootBoundDisc.push_back(roots[i]);
  
  // add the definition intervals of the bound
  vector<double>* chgPoints = this->getListChangePoints();
  for(int i=0; i<chgPoints->size(); i++)
    rootBoundDisc.push_back((*chgPoints)[i]);  
  delete chgPoints;
  
  // sort the vector
  sortVectDble(rootBoundDisc);
  // delete double values
  delDoubleValues(rootBoundDisc);
  
//   cout << "binf : " << this->getBInf() << endl;
//   cout << "bsup : " << this->getBSup() << endl;
//   cout << "roots : ";
//   displayVectDble(&roots);
//   cout << "chgPoints : ";
//   displayVectDble(&chgPoints);
//   cout << "rootBoundDisc : ";
//   displayVectDble(&rootBoundDisc);
  
  // check between each couple of values if the interval has to be changed
  int N = rootBoundDisc.size();
  for(int i=0; i<N-1; i++)
    {
      // get back the inferior and superior bounds
      // Warning : these values must not be used to define the bounds
      double vmin = rootBoundDisc[i];
      double vmax = rootBoundDisc[i+1];
      if(vmin == -INFINITY) vmin = -999999999.0;
      if(vmax == INFINITY) vmax = 999999999.0;
      
      double valThis = this->getVal(0.5*(vmin + vmax));
      double valBound = fMinB->getVal(0.5*(vmin + vmax));

      if(valThis < valBound)
	{
	  // case where the multiItvFunction is greater 
	  // than the maximal bound : we change the interval
	  this->changeItv(rootBoundDisc[i],rootBoundDisc[i+1],fMinB);
	  if(lRes != NULL) lRes->push_back(ItvFunction(rootBoundDisc[i],rootBoundDisc[i+1],NULL));	  
	}
    }

//   return lRes;  
}

// add a maximal bound over the function, return the bounded intervals
void MultiItvFunction::addMaxBound(MultiItvFunction* fMaxB,std::list<ItvFunction>* lRes)
{
//   // define the list of bounded intervals
//   list<ItvFunction>* lRes = new list<ItvFunction>;

  MultiItvFunction fdiff(*fMaxB);
  fdiff -= (*this);
  
  // search for the roots of the difference function
  vector<double> roots;
  int nbRoots = fdiff.realRoots(roots,0.0);
  
  vector<double> rootBoundDisc;
  
  // add the inferior and superior bounds
  rootBoundDisc.push_back(this->getBInf());
  rootBoundDisc.push_back(this->getBSup());
  
  // add the roots
  for(int i=0; i<nbRoots; i++)
    rootBoundDisc.push_back(roots[i]);
  
  // add the definition intervals of the bound
  vector<double>* chgPoints = this->getListChangePoints();
  for(int i=0; i<chgPoints->size(); i++)
    rootBoundDisc.push_back((*chgPoints)[i]);  
  delete chgPoints;
  
  // sort the vector
  sortVectDble(rootBoundDisc);
  // delete double values
  delDoubleValues(rootBoundDisc);
  
//   cout << "binf : " << this->getBInf() << endl;
//   cout << "bsup : " << this->getBSup() << endl;
//   cout << "roots : ";
//   displayVectDble(&roots);
//   cout << "chgPoints : ";
//   displayVectDble(&chgPoints);
//   cout << "rootBoundDisc : ";
//   displayVectDble(&rootBoundDisc);
  
  // check between each couple of values if the interval has to be changed
  int N = rootBoundDisc.size();
  for(int i=0; i<N-1; i++)
    {
      // get back the inferior and superior bounds
      // Warning : these values must not be used to define the bounds
      double vmin = rootBoundDisc[i];
      double vmax = rootBoundDisc[i+1];
      if(vmin == -INFINITY) vmin = -999999999.0;
      if(vmax == INFINITY) vmax = 999999999.0;
      
      double valThis = this->getVal(0.5*(vmin + vmax));
      double valBound = fMaxB->getVal(0.5*(vmin + vmax));

      if(valThis > valBound)
	{
	  // case where the multiItvFunction is greater 
	  // than the maximal bound : we change the interval
	  this->changeItv(rootBoundDisc[i],rootBoundDisc[i+1],fMaxB);
	  if(lRes != NULL) lRes->push_back(ItvFunction(rootBoundDisc[i],rootBoundDisc[i+1],NULL));	  
	}
    }

//   return lRes;
}

// get the list of intervals between bounds bmin,bmax
list<ItvFunction*> MultiItvFunction::getListItvBetween(double bmin,double bmax)
{
//   cout << "getListItvBetween(" << bmin << "," << bmax << ")" << endl;
//   this->display();

  // initialize the new list
  list<ItvFunction*> listRes; // = new list<ItvFunction*>;
  
  // search for the first interval of the current
  // function which contains the minimal bound bmin
  list<ItvFunction*>::iterator it =  this->vecItvFunc_->begin();
  bool bminReached = false;
  bool bmaxReached = false;
  
  // check if bmin is out of the definition domain of the current function
  if(bmin < this->getBInf()) bminReached = true;

  while(!bmaxReached && it != this->vecItvFunc_->end())
    {
      // searching for the minimal bound bmin
      if(!bminReached)
	{
	  // test if the minimal bound has been reached
	  if(bmin <= (*it)->bsup_)
	    {
	      bminReached = true;

	      // check if [bmin;bmax] is completely inside of the current interval
	      if(bmax <= (*it)->bsup_)
		{
		  bmaxReached = true;

		  // add the current interval and define it between bmin and bmax
		  ItvFunction* newItv = new ItvFunction(bmin,bmax,(*it)->fct_->copy());
		  listRes.push_back(newItv);
		}
	      else //if(bmin != (*it)->bsup_)
		{
		  // add the current interval and define it between bmin and itv->bmax_
		  ItvFunction* newItv = new ItvFunction(bmin,(*it)->bsup_,(*it)->fct_->copy());
		  listRes.push_back(newItv);
		}
	    }
	}
      // searching for the maximal bound bmax
      else
	{
	  // test if the maximal bound has been reached
	  if(bmax <= (*it)->bsup_) 
	    {
	      bmaxReached = true;
	      
	      // add the current interval and define it between itv->bmin_ and bmax
	      ItvFunction* newItv = new ItvFunction((*it)->binf_,bmax,(*it)->fct_->copy());
	      listRes.push_back(newItv);
	    }
	  else
	    {
	      // add the current interval and define it between itv->bmin_ and itv->bmax_
	      ItvFunction* newItv = new ItvFunction((*it)->binf_,(*it)->bsup_,(*it)->fct_->copy());
	      listRes.push_back(newItv);
	    }
	}
      it++;
    }

//   cout << "Intervals found : " << endl;
//   list<ItvFunction*>::iterator itLNF11;
//   for(itLNF11 = listRes->begin(); itLNF11 != listRes->end(); itLNF11++)
//     {
//       (*itLNF11)->display();
//     }
//   cout << "End getListItvBetween" << endl;
  
  return listRes;
}

// get the list of value at which the definition of the function change
std::vector<double>* MultiItvFunction::getListChangePoints()
{
  vector<double>* vres = new vector<double>;
  list<ItvFunction*>::iterator it = vecItvFunc_->begin();
  
  // check if the multiitv function contain at least one interval
  if(it != vecItvFunc_->end())
    {
      while(it != vecItvFunc_->end())
	{
	  vres->push_back((*it)->binf_);
	  it++;
	}
      it--;
      vres->push_back((*it)->bsup_);
    }
  return vres;
}

/**
 *  \brief shift the time bounds of all time intervals
 *  \param dt : the shifting time value
 */
void MultiItvFunction::shiftTime(double dt)
{
  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      ItvFunction* curItv = *it;
      curItv->shiftTime(dt);
    }
}

/**
 *  \brief concatenate two piecewise polynomial functions
 *  \param f : the input fonction to concatenate
 */
void MultiItvFunction::concat(const MultiItvFunction& f)
{
  // check if the start time of the second fuction corresponds to the end time of the current one
  if(this->getBSup() != f.getBInf())
    {
      cerr << "MultiItvFunction::concat : Unconsistent starting and endin times of both functions." << endl;
      exit(1);
    }

  // add each interval of the second function to the current one
  list<ItvFunction*>::iterator it;
  for(it = f.vecItvFunc_->begin(); it != f.vecItvFunc_->end(); it++)
    {
      this->vecItvFunc_->push_back(new ItvFunction(**it));
    }
}


/**************** Utility functions ***************/

// display the polynomial
void MultiItvFunction::display() const
{
 //  int nbitv = this->vecItvFunc_->size();
//   for(int i=0; i<nbitv; i++)
//     {
//       cout << "[" << this->vecItvFunc_->at(i)->binf_ << ";" << this->vecItvFunc_->at(i)->bsup_ << "] \\cap ";
//       Function* curFct = this->vecItvFunc_->at(i)->fct_;
//       curFct->display();
//     }
  
  if(vecItvFunc_ != NULL)
    {
      list<ItvFunction*>::iterator it;
      for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
	{
	  ItvFunction* curItv = *it;
	  cout << "[" << curItv->binf_ << ";" << curItv->bsup_ << "] \\cap ";
	  Function* curFct = curItv->fct_;
	  if(curFct != NULL)
	    curFct->display();
	  else cout << "function is nul" << endl;
	}
    }
  else
    {
      cout << "vecItvFunc_ is nul" << endl;
    }
}

// shift the function on Y axis from value v
void MultiItvFunction::shiftY(double v)
{
//   int nbitv = this->vecItvFunc_->size();
//   for(int i=0; i<nbitv; i++)
//     {
//       this->vecItvFunc_->at(i)->fct_->shiftY(v);
//     }

  list<ItvFunction*>::iterator it;
  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      ItvFunction* curItv = *it;
      curItv->fct_->shiftY(v);
    }
}

// return false if there is an unconsistency in the intervals defined
bool MultiItvFunction::checkItvConsistency() const
{
  if(vecItvFunc_ == NULL)
    {
      cout << "MultiItvFunction::checkItvConsistency() : Warning. vecItvFunc_ is nul" << endl;
      return false;
    }

  list<ItvFunction*>::iterator it;
  list<ItvFunction*>::iterator previt = this->vecItvFunc_->begin();

  for(it = this->vecItvFunc_->begin(); it != this->vecItvFunc_->end(); it++)
    {
      // get back the current interval
      ItvFunction* curItv = *it;
      // get back the bounds of the interval
      double minB = curItv->binf_;
      double maxB = curItv->bsup_;
      // check the consistency of the bounds
      //if(minB >= maxB)
      if(minB > maxB)
	{
	  cout << "MultiItvFunction::checkItvConsistency() : Warning. minB >= maxB" << endl;
	  cout << "minB : " << minB << "  maxB : " << maxB << endl;
	  return false;  
	}
      // check the consistency with the previous interval
      if(it != this->vecItvFunc_->begin())
	{
	  if((*previt)->bsup_ != minB)
	    {
	      cout << "MultiItvFunction::checkItvConsistency() : Warning. (*previt)->bsup_ != minB" << endl;
	      cout << "(*previt)->bsup_ : " << (*previt)->bsup_ << "  minB : " << minB << endl;
	      return false;	      
	    }
	}
      previt = it;
    }
  return true;
}

/**
 *  \brief search for the discontinuities of the function
 *  \param discPoints : resulting discontinuity points
 */
void MultiItvFunction::getVecDiscPoints(std::vector<Disc_Point>& discPoints) const
{
  list<ItvFunction*>::iterator it = vecItvFunc_->begin();
  
  // check if the multiitv function contain at least one interval
  if(it != vecItvFunc_->end())
    {
      while(it != vecItvFunc_->end())
	{
	  it++;
	  if(it != vecItvFunc_->end())
	    {
	      double t = (*it)->binf_;
	      double v = (*it)->fct_->getVal(t);

	      Disc_Point dp;
	      dp.t = t;
	      dp.v = v;

	      discPoints.push_back(dp);
	    }
	}
    }
}
