#ifndef MPC_PSTRAJ_HPP
#define MPC_PSTRAJ_HPP

#include "motion/inc/STrajectory.hpp"
#include "motion/inc/PCTraj.hpp"
#include "math/inc/MultiItvFunction.hpp"

#include "system/inc/ManaSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class PSTraj ******************/
                                    /**************************************************/

  /** \class PSTraj
   *  \brief Piecewise polynomial state trajectory followed by a system
   */
  class PSTraj : public STrajectory
  {
  protected : 
    traj_t type_;                                     /**< type of the state trajectory */
    int nbS_;                                         /**< dimension of the state vector (Size of tabFunc_. Warning : this table does not contains quadratic profiles) */
    MultiItvFunction** tabFunc_;                      /**< table containing the piecewise parametric profiles of the functions defining the state trajectory */
    StateTime* initST_;                               /**< initial state-time of the trajectory */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    PSTraj();

    /**
     *  \brief init constructor
     *  \param dim : dimension of the vector containing polynomial profiles described by the state trajectory (do not include quadratic profiles)
     *  \param startTime : the starting time of the state trajectory
     *  \param endTime : the ending time of the state trajectory
     */
    PSTraj(int dim,double startTime=0.0,double endTime=-1.0);

  public : 
    /**
     *  \brief copy constructor
     */
    PSTraj(const PSTraj& traj);

    /**
     *  \brief destructor
     */
    virtual ~PSTraj();

    /**
     *  \brief return a copy of the current state trajectory
     *  \return a copy of the current state trajectory
     *
     *  Return a new instance of PSTraj
     */
    virtual PSTraj* copy() const = 0;


    /************** Getters and setters ***************/

  protected : 
    /**
     *  \brief return the dimension of vector containing the polynomial profiles described by the state trajectory (do not include quadratic profiles)
     *  \return the dimension of the control vector
     */
    inline int nbS() const { return this->nbS_; }

  public : 
    /**
     *  \brief return the type of the trajectory
     *  \return the type of the trajectory
     */
    inline traj_t type() const { return this->type_; }

    /**
     *  \brief return the start time of the control trajectory
     *  \return the start time of the control trajectory
     */
    double startTime() const;

    /**
     *  \brief return the end time of the control trajectory
     *  \return the end time of the control trajectory
     */
    double endTime() const;


    /******** Control trajectory modifications ********/

  public : 
    /**
     *  \brief concatenate two trajectories
     *  \param traj : the trajectory to concatenate to the current one
     */
    virtual void concat(const PSTraj& traj);

    /**
     *  \brief shift the trajectory by time t
     *  \param t : the time to shift the trajectory
     */
    virtual void shift(double t);


    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief sample the state trajectory
     *  \return a sampled state trajectory
     *  \param nb : the number of samples
     */
    virtual STraj* sample(int nb=20) const = 0;

    /**
     *  \brief sample the state trajectory
     *  \return a sampled state trajectory
     *  \param dt : the time interval between two samples
     */
    virtual STraj* sample(double dt) const = 0;

    /**
     *  \brief diplay the control trajectory
     */
    virtual void display() const = 0;
  };


                                    /**************************************************/
                                    /**************** Class DD2PSTraj ****************/
                                    /**************************************************/

#define DD2_INDX 0 
#define DD2_INDY 1 
#define DD2_INDTH 2 
#define DD2_INDV 3 
#define DD2_INDW 4 

  /** \class DD2PSTraj
   *  \brief Piecewise polynomial state trajectory followed by a differential drive robotic system
   */
  class DD2PSTraj : public PSTraj
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    DD2PSTraj();

  public : 
    /**
     *  \brief copy constructor
     */
    DD2PSTraj(const DD2PSTraj& traj);

    /**
     *  \brief init constructor (from a state and a control trajectory)
     *  \param s : the initial state
     *  \param ctraj : the input control trajectory
     *  \param minV, maxV : bounds over the linear velocity
     *  \param minW, maxW : bounds over the angular velocity
     */
    DD2PSTraj(const DD2State& s, const DD2PCTraj& ctraj, 
	       double minV=-INFINITY, double maxV=INFINITY, 
	       double minW=-INFINITY, double maxW=INFINITY);

    /**
     *  \brief destructor
     */
    virtual ~DD2PSTraj();

    /**
     *  \brief return a copy of the current state trajectory
     *  \return a copy of the current state trajectory
     *
     *  Return a new instance of PSTraj
     */
    virtual DD2PSTraj* copy() const;


//     /**************** Quadrature tools ****************/

//   private : 
//     /*
//      *  \brief Quadrature tools...
//      */  
//     void gauleg(const double x1, const double x2, double x[], double w[], int n) const;

//     /*
//      *  \brief Quadrature tools...
//      */
//     double qgaus(double (DD2PSTraj::*func)(const double) const, const double a, const double b) const;


    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief return the initial state-time of the trajectory
     *  \return the initial state-time of the trajectory
     */
    inline DD2StateT initST() const { return (*((DD2StateT*)(this->initST_))); }

    /**
     *  \brief getter : first state of the trajectory
     *  \return the first state of the trajectory
     */
    virtual DD2StateT firstStateT() const;

    /**
     *  \brief getter : last state of the trajectory
     *  \return the last state of the trajectory
     */
    virtual DD2StateT lastStateT() const;

    /**
     *  \brief getter : get the state-time defined at time t
     *  \param t : the input time
     *  \return return the state defined at time t
     */
    virtual DD2StateT at(double t) const;

  private : 
    /*
     *  \brief function to integrate to obtain the X(t) function
     *  \param t : the time variable
     *  \return V(t) * sin(Th(t))
     */
    double itgX(double t) const;

    /*
     *  \brief function to integrate to obtain the Y(t) function
     *  \param t : the time variable
     *  \return V(t) * cos(Th(t))
     */
    double itgY(double t) const;

  public : 
    /**
     *  \brief return the position x reached at time t
     *  \param t : the input time
     *  \return the position x reached at time t
     */
    double x(double t) const;

    /**
     *  \brief return the position y reached at time t
     *  \param t : the input time
     *  \return the position y reached at time t
     */
    double y(double t) const;

    /**
     *  \brief return the position th reached at time t
     *  \param t : the input time
     *  \return the position th reached at time t
     */
    double th(double t) const;

    /**
     *  \brief return the linear velocity reached at time t
     *  \param t : the input time
     *  \return the linear velocity v reached at time t
     */
    double v(double t) const;

    /**
     *  \brief return the angular velocity w reached at time t
     *  \param t : the input time
     *  \return the angular velocity w reached at time t
     */
    double w(double t) const;

    /**
     *  \brief return a pointer to the linear velocity function
     *  \return a pointer to the linear velocity function
     */
    inline const MultiItvFunction* fctTh() const { return ((MultiItvFunction*)(tabFunc_[DD2_INDTH])); }

    /**
     *  \brief return a pointer to the linear velocity function
     *  \return a pointer to the linear velocity function
     */
    inline const MultiItvFunction* fctVel() const { return ((MultiItvFunction*)(tabFunc_[DD2_INDV])); }

    /**
     *  \brief return a pointer to the linear velocity function
     *  \return a pointer to the linear velocity function
     */
    inline const MultiItvFunction* fctAVel() const { return ((MultiItvFunction*)(tabFunc_[DD2_INDW])); }


    /**************** Boundary functions **************/

  public : 
    /**
     *  \brief add a minimal and a maximal bound over the linear velocity profile
     *  \param minV : the minimal bound
     *  \param maxV : the maximal bound
     */
    void addVelBounds(double minV,double maxV);

    /**
     *  \brief add a minimal and a maximal bound over the angular velocity profile
     *  \param minAV : the minimal bound
     *  \param maxAV : the maximal bound
     */
    void addAVelBounds(double minAV,double maxAV);


    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief sample the state trajectory
     *  \param nb : the number of samples
     *  \return a sampled state trajectory
     *
     *  A new instance of DD2STraj is returned
     */
    virtual DD2STraj* sample(int nb=20) const;

    /**
     *  \brief sample the state trajectory
     *  \param dt : the time interval between two samples
     *  \return a sampled state trajectory
     *
     *  A new instance of DD2STraj is returned
     */
    virtual DD2STraj* sample(double dt) const;

    /**
     *  \brief diplay the control trajectory
     */
    virtual void display() const;
  };

};

#define ManaPSTraj DD2PSTraj

#endif // MPC_PSTRAJ_HPP
