#include "GeomPath.hpp"
using namespace mpc;

#include "Utils.hpp"
#include <cmath>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /***************** Class GPrimPath ****************/
                                    /**************************************************/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
GPrimPath::GPrimPath()
{
  this->initCfg_ = DPMConfig();
  this->length_  = 0;
  this->type_    = ND_GPT;
}

/**
 *  \brief copy constructor
 *
 *  Only used by subclasses
 */
GPrimPath::GPrimPath(const GPrimPath& p)
{
  this->initCfg_ = p.initCfg();
  this->length_  = p.getLength();
  this->type_    = p.type();
}

/**
 *  \brief destructor
 */
GPrimPath::~GPrimPath()
{
}

/************** Getters and setters ***************/

/**
 *  \brief return a pointer to a copy of the current primitive path
 *  \return a pointer to a copy of the current primitive path
 */
GPrimPath* GPrimPath::copy() const
{
  GPrimPath* res;
  switch(this->type())
    {
    case ND_GPT : 
      {
	cout << "GPrimPath::copy : Undefined geometrical path type!" << endl;
	exit(1);
      }
    case STR_LINE : 
      res = new StrLinePath(*((StrLinePath*)this));
      break;
    case CIRC_BOW : 
      cout << "GPrimPath::copy : Not implemented yet!" << endl;
      exit(1);
      break;
    case CLOTH_BOW : 
      cout << "GPrimPath::copy : Not implemented yet!" << endl;
      exit(1);
      break;
    }
  return res;
}


                                    /**************************************************/
                                    /***************** Class GeomPath *****************/
                                    /**************************************************/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
GeomPath::GeomPath()
{
  this->listPrim_ = new std::list<GPrimPath*>;
}

/**
 *  \brief copy constructor
 *
 *  Only used by subclasses
 */
GeomPath::GeomPath(const GeomPath& p)
{
  this->listPrim_ = new std::list<GPrimPath*>;
}

/**
 *  \brief destructor
 */
GeomPath::~GeomPath()
{
  this->clear();
  delete listPrim_;
}

/************** Getters and setters ***************/

/**
 *  \brief return the initial configuration of the path
 *  \return the initial configuration of the path
 */
DPMConfig GeomPath::initCfg() const
{
  if(this->size() == 0)
    {
      return DPMConfig();
    }
  else
    {
      return this->at(0).initCfg();
    }
}

/**
 *  \brief return the goal configuration of the path
 *  \return the goal configuration of the path
 */
DPMConfig GeomPath::goalCfg() const
{
  if(this->size() == 0)
    {
      return DPMConfig();
    }
  else
    {
      return this->at(this->size()-1).goalCfg();
    }
}

/**
 *  \brief return the length of the path
 *  \return the length of the path
 */
double GeomPath::getLength() const
{
  double res = 0;
  int N = this->size();
  for(int i=0; i<N; i++)
    {
      res += this->at(i).getLength();
    }
  return res;
}

/**
 *  \brief return the i-th primitive of the path
 *  \param i : index of the primitive to return
 *  \return the i-th primitive of the path
 *
 *  If possible, avoid to use this function (costly function)
 *  Use iterators instead
 */
const GPrimPath& GeomPath::at(int i) const
{
  std::list<GPrimPath*>::iterator it = listPrim_->begin();
  for(int j=0; j<i; j++) it++;
  return **it;
}


/**************** Path modification ***************/

/**
 *  \brief append a new primitive to the path
 *  \param c : the new primitive to append
 */
void GeomPath::append(const GPrimPath& p)
{
  GPrimPath* newP = p.copy();
  listPrim_->push_back(newP);
}

/**
 *  \brief prepend a new primitive to the path
 *  \param c : the the primitive to prepend
 */
void GeomPath::prepend(const GPrimPath& p)
{
  GPrimPath* newP = p.copy();
  listPrim_->push_front(newP);
}

/**
 *  \brief concatenate a path to the current one
 *  \param p : the input path to concatenate
 */
void GeomPath::concat(const GeomPath& p)
{
  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}

/**
 *  \brief reverse the current path
 */
void GeomPath::reverse()
{
  // construct the reverse list in a temporary list
  std::list<GPrimPath*>* tmpList = new std::list<GPrimPath*>;
  std::list<GPrimPath*>::iterator it = listPrim_->begin();

  int N = this->size();
  for(int i=0; i<N; i++,it++)
    {
      tmpList->push_front(*it);
    }
  
  // replace the current list by a temporary list
  this->listPrim_->clear();
  delete(this->listPrim_);
  this->listPrim_ = tmpList;
}

/**
 *  \brief clear the list of primitives
 */
void GeomPath::clear()
{
  std::list<GPrimPath*>::iterator it;

  for(it = listPrim_->begin(); it != listPrim_->end(); it++)
    {
      delete (*it);
    }
  listPrim_->clear();
}

    
/**************** Utility functions ***************/

/**
 *  \brief diplay the path
 */
void GeomPath::display() const
{
  int N = this->size();
  cout << "Geometrical path : " << endl;
  for(int i=0; i<N; i++)
    {
      this->at(i).display();
    }
  cout << endl;
}

/**
 *  \brief display in an openGL viewer
 */
void GeomPath::toGL() const
{
  cout << "GeomPath::toGL : Not implemented yet!" << endl;
  exit(1);
}
  
/**
 *  \brief return the barycenter of the path in the workspace
 */
PositionXY GeomPath::center() const
{
  cout << "GeomPath::center : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief compute an approximative radius of the set of all configurations of the path
 *  \return an approximative radius of the set of all configurations of the path
 */
double GeomPath::radius() const
{
  cout << "GeomPath::radius : Not implemented yet!" << endl;
  exit(1);
}

/************* Input/Output functions *************/

/**
 *  \brief load the path from an input file
 *  \param nomfic : name of the input file
 */
bool GeomPath::load(char* nomfic)
{
  cout << "GeomPath::load : Not implemented yet!" << endl;
  exit(1);
}
  
/**
 *  \brief save the path in an output file
 *  \param nomfic : name of the output file
 */
bool GeomPath::save(char* nomfic) const
{
  cout << "GeomPath::save : Not implemented yet!" << endl;
  exit(1);
}


                                    /**************************************************/
                                    /**************** Class StrLinePath ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
StrLinePath::StrLinePath() : GPrimPath()
{
  this->type_ = STR_LINE;
}

/**
 *  \brief copy constructor
 */
StrLinePath::StrLinePath(const StrLinePath& p) : GPrimPath(p)
{
}
  
/**
 *  \brief copy constructor
 */
StrLinePath::StrLinePath(const DPMConfig& initCfg,double dist)
{
  this->initCfg_ = initCfg;
  this->length_  = dist;  
  this->type_    = STR_LINE;
}

/**
 *  \brief copy constructor
 */
StrLinePath::StrLinePath(const PMConfig& initCfg,const PMConfig& goalCfg) 
{
  PMConfig vecICGC(goalCfg.x() - initCfg.x(),goalCfg.y() - initCfg.y());
  double dir = atan2(vecICGC.y(),vecICGC.x());
  double norm = vecICGC.norm();
  
  if(norm == 0)
    {
      this->initCfg_ = DPMConfig(initCfg.x(),initCfg.y(),0);
      this->length_  = norm;
      this->type_    = STR_LINE;  
    }
  else
    {
      this->initCfg_ = DPMConfig(initCfg.x(),initCfg.y(),dir);
      this->length_  = norm;
      this->type_    = STR_LINE;  
    }
}

/**
 *  \brief destructor
 */
StrLinePath::~StrLinePath()
{
}


/************** Getters and setters ***************/

/**
 *  \brief return the goal configuration of the path
 *  \return the goal configuration of the path
 */
DPMConfig StrLinePath::goalCfg() const
{
  double x = initCfg().x() + getLength()*cos(initCfg().th());
  double y = initCfg().y() + getLength()*sin(initCfg().th());
  double th = initCfg().th();
  return DPMConfig(x,y,th);
}


/**************** Path modification ***************/

/**
 *  \brief reverse the current path
 */
void StrLinePath::reverse()
{
  DPMConfig goal = this->goalCfg();
  goal.th() = normalizedAngle(goal.th() + M_PI);
  this->initCfg_ = goal;
}


/**************** Utility functions ***************/

/**
 *  \brief diplay the path
 */
void StrLinePath::display() const
{
  cout << "Straight line path : (";
  cout << "initCfg : "; initCfg().display();
  cout << "   goalCfg : "; goalCfg().display();
  cout << "   length : " << this->getLength() << ") ";
  cout << endl;
}

/**
 *  \brief display in an openGL viewer
 */
void StrLinePath::toGL() const
{
  cout << "StrLinePath::toGL : Not implemented yet!" << endl;
  exit(1);
}
  
/**
 *  \brief return the barycenter of the path in the workspace
 */
PositionXY StrLinePath::center() const
{
  double x = initCfg().x() + 0.5*getLength()*cos(initCfg().th());
  double y = initCfg().y() + 0.5*getLength()*sin(initCfg().th());
  return PositionXY(x,y);
}

/**
 *  \brief compute an approximative radius of the set of all configurations of the path
 *  \return an approximative radius of the set of all configurations of the path
 */
double StrLinePath::radius() const
{
  return (0.5*this->getLength());
}

/************* Input/Output functions *************/

/**
 *  \brief load the path from an input file
 *  \param nomfic : name of the input file
 */
bool StrLinePath::load(char* nomfic)
{
  cout << "StrLinePath::load : Not implemented yet!" << endl;
  exit(1);
}
  
/**
 *  \brief save the path in an output file
 *  \param nomfic : name of the output file
 */
bool StrLinePath::save(char* nomfic) const
{
  cout << "StrLinePath::save : Not implemented yet!" << endl;
  exit(1);
}
