#include "motion/inc/PCTraj.hpp"
using namespace mpc;

#include <iostream>
#include <cstring>
using namespace std;

                                    /**************************************************/
                                    /****************** Class PCTraj ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PCTraj::PCTraj() : CTrajectory()
{
  this->type_ = ND_CTRAJ;
  this->nbC_ = 0;
  this->tabFunc_ = NULL;
}

/**
 *  \brief init constructor
 *  \param dim : dimension of teh control vector
 *  \param startTime : the starting time of the control trajectory
 *  \param endTime : the ending time of the control trajectory
 */
PCTraj::PCTraj(int dim) : CTrajectory()
{
  this->type_ = ND_CTRAJ;
  this->nbC_ = dim;
  this->tabFunc_ = new MultiItvFunction* [nbC_];
  for(int i=0; i<nbC_; i++)
    {
//       Polynomial poly(0.0);
//       this->tabFunc_[i] = new MultiItvFunction(&poly, startTime, endTime);
      this->tabFunc_[i] = NULL;
    }
}

/**
 *  \brief copy constructor
 */
PCTraj::PCTraj(const PCTraj& traj) : CTrajectory(traj)
{
  this->type_ = traj.type_;
  this->nbC_ = traj.nbC_;
  this->tabFunc_ = new MultiItvFunction* [nbC_];
  for(int i=0; i<nbC_; i++)
    {
      this->tabFunc_[i] = new MultiItvFunction(*(traj.tabFunc_[i]));
    }
}

/**
 *  \brief destructor
 */
PCTraj::~PCTraj()
{
  if(this->tabFunc_ != NULL) 
    {
      for(int i=0; i<nbC_; i++)
	{
	  if(tabFunc_[i] != NULL) delete tabFunc_[i];
	}
      delete [] tabFunc_;
    }
}


/************** Getters and setters ***************/

/**
 *  \brief return the start time of the control trajectory
 *  \return the start time of the control trajectory
 */
double PCTraj::startTime() const
{
  if(nbC_ <= 0) return -INFINITY;
  else return tabFunc_[0]->getBInf();
}

/**
 *  \brief return the end time of the control trajectory
 *  \return the end time of the control trajectory
 */
double PCTraj::endTime() const
{
  if(nbC_ <= 0) return INFINITY;
  else return tabFunc_[0]->getBSup();
}


/******** Control trajectory modifications ********/

/**
 *  \brief concatenate two trajectories
 *  \param traj : the trajectory to concatenate to the current one
 */
void PCTraj::concat(const PCTraj& traj)
{
  // check if the type of both trajectories is the same
  if(this->type_ != traj.type_)
    {
      cerr << "PCTraj::concat : the type of the input control trajectory differs from the current one." << endl;
      throw std::string("PCTraj::concat : the type of the input control trajectory differs from the current one.");
    }

  // check if the final time of the current trajectory is finite
  if(this->endTime() == INFINITY)
    {
      cerr << "PCTraj::concat : the final time of the current trajectory is infinite." << endl;
      throw std::string("PCTraj::concat : the final time of the current trajectory is infinite.");
    }

  // build a copy of the input control trajectory
  PCTraj* trajCopy = traj.copy();

  if(trajCopy == NULL)
    {
      cerr << "PCTraj::concat : the copy of the input control trajectory is null." << endl;
      throw std::string("PCTraj::concat : the copy of the input control trajectory is null.");
    }

  trajCopy->shift(this->endTime() - trajCopy->startTime());
  
  // concatenate each component of the control trajectory
  for(int i=0; i<nbC_; i++)
    {
      this->tabFunc_[i]->concat(*(trajCopy->tabFunc_[i]));
    }
  
  // delete the input control trajectory copy
  delete trajCopy;
}

/**
 *  \brief shift the trajectory by time t
 *  \param t : the time to shift the trajectory
 */
void PCTraj::shift(double t)
{
  for(int i=0; i<nbC_; i++)
    {
      this->tabFunc_[i]->shiftTime(t);
    }
}


                                    /**************************************************/
                                    /**************** Class DD2PCTraj ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2PCTraj::DD2PCTraj() : PCTraj(4)
{
  this->type_ = DD2_CTRAJ;

  double startTime = 0;
  double endTime = -1;

  for(int i=0; i<nbC_; i++)
    {
      Polynomial poly(0.0);
      this->tabFunc_[i] = new MultiItvFunction(&poly, startTime, endTime);
    }
}

/**
 *  \brief copy constructor
 */
DD2PCTraj::DD2PCTraj(const DD2PCTraj& traj) : PCTraj(traj)
{
}

/**
 *  \brief init constructor (from a constant control)
 *  \param v : the constant linear velocity control
 *  \param w : the constant angular velocity control
 *  \param t : the duration of the control
 */
DD2PCTraj::DD2PCTraj(double v, double w, double t) : PCTraj()
{
  this->type_ = DD2_CTRAJ;
  this->nbC_ = 4;
  this->tabFunc_ = new MultiItvFunction* [nbC_];

  Polynomial polyV(v);
  Polynomial polyW(w);
  
  this->tabFunc_[0] = new MultiItvFunction(&polyV, 0.0, t);
  this->tabFunc_[1] = new MultiItvFunction(&polyW, 0.0, t);
  this->tabFunc_[2] = fctVel()->derivation();
  this->tabFunc_[3] = fctAVel()->derivation();
}

/**
 *  \brief init constructor (from a linear control)
 *  \param v0 : the initial linear velocity
 *  \param a : the constant linear acceleration
 *  \param w0 : the initial angular velocity
 *  \param mu : the constant angular acceleration
 *  \param t : the duration of the control
 */
DD2PCTraj::DD2PCTraj(double v0, double a, double w0, double mu, double t) : PCTraj(4)
{
  this->type_ = DD2_CTRAJ;

  Polynomial polyV(a,v0);
  Polynomial polyW(mu,w0);
  
  this->tabFunc_[0] = new MultiItvFunction(&polyV, 0.0, t);
  this->tabFunc_[1] = new MultiItvFunction(&polyW, 0.0, t);
  this->tabFunc_[2] = fctVel()->derivation();
  this->tabFunc_[3] = fctAVel()->derivation();  
}

/**
 *  \brief init constructor (from two polynomial functions)
 *  \param polyV : the polynomial function representing the linear velocity control
 *  \param polyW : the polynomial function representing the angular velocity control
 *  \param t : the duration of the control
 */
DD2PCTraj::DD2PCTraj(const Polynomial& polyV, const Polynomial& polyW, double t) : PCTraj(4)
{
  this->type_ = DD2_CTRAJ;

  this->tabFunc_[0] = new MultiItvFunction(&polyV, 0.0, t);
  this->tabFunc_[1] = new MultiItvFunction(&polyW, 0.0, t);  
  this->tabFunc_[2] = fctVel()->derivation();
  this->tabFunc_[3] = fctAVel()->derivation();
}

/**
 *  \brief init constructor (from two polynomial acceleration functions and two initial velocities)
 *  \param v0 : the initial linear velocity
 *  \param polyAcc : the polynomial function representing the linear acceleration control
 *  \param w0 : the initial angular velocity
 *  \param polyAAcc : the polynomial function representing the angular acceleration control
 *  \param t : the duration of the control
 */
DD2PCTraj::DD2PCTraj(double v0, const Polynomial& polyAcc, double w0, const Polynomial& polyAAcc, double t) : PCTraj(4)
{
  this->type_ = DD2_CTRAJ;

  this->tabFunc_[2] = new MultiItvFunction(&polyAcc, 0.0, t);
  this->tabFunc_[3] = new MultiItvFunction(&polyAAcc, 0.0, t);  
  this->tabFunc_[0] = fctAcc()->integration(0.0,v0);
  this->tabFunc_[1] = fctAAcc()->integration(0.0,w0);
}

/**
 *  \brief init constructor (from the coefficients of the linear and angular velocity profiles)
 *  \param vel : vector containing the coefficient of the piecewise polynomial linear velocity profile
 *  \param avel : vector containing the coefficient of the piecewise polynomial angular velocity profile
 */
DD2PCTraj::DD2PCTraj(const std::vector<std::vector<double> >& vecVel, const std::vector<std::vector<double> >& vecAVel) : PCTraj(4)
{
  this->type_ = DD2_CTRAJ;

  this->tabFunc_[0] = new MultiItvFunction(vecVel);
  this->tabFunc_[1] = new MultiItvFunction(vecAVel);
  this->tabFunc_[2] = fctVel()->derivation();
  this->tabFunc_[3] = fctAVel()->derivation();
}

/**
 *  \brief destructor
 */
DD2PCTraj::~DD2PCTraj()
{
}

/**
 *  \brief return a copy of the current control trajectory
 *  \return a copy of the current control trajectory
 *
 *  Return a new instance of PCTraj
 */
DD2PCTraj* DD2PCTraj::copy() const
{
  DD2PCTraj* newTraj = new DD2PCTraj(*this);
  return newTraj;
}


/************** Getters and setters ***************/

/**
 *  \brief getter : first control of the trajectory
 *  \return the first control of the trajectory
 */
DD2Control DD2PCTraj::firstCtrl() const
{
  DD2Control ctrl(this->v(this->startTime()),
		   this->w(this->startTime()));
  return ctrl;
}

/**
 *  \brief getter : last control of the trajectory
 *  \return the last control of the trajectory
 */
DD2Control DD2PCTraj::lastCtrl() const
{
  DD2Control ctrl(this->v(this->endTime()),
		   this->w(this->endTime()));
  return ctrl;
}

/**
 *  \brief getter : get the control defined at time t
 *  \param t : the input time
 *  \return return the control defined at time t
 */
DD2Control DD2PCTraj::at(double t) const
{
  DD2Control ctrl(this->v(t),
		   this->w(t));
  return ctrl;
}

/**
 *  \brief return the linear velocity value to apply at time t
 *  \param t : the input time
 *  \return the linear velocity value to apply at time t
 */
double DD2PCTraj::v(double t) const
{
  return fctVel()->getVal(t);
}

/**
 *  \brief return the angular velocity value to apply at time t
 *  \param t : the input time
 *  \return the angular velocity value to apply at time t
 */
double DD2PCTraj::w(double t) const
{
  return fctAVel()->getVal(t);
}

/**
 *  \brief return the linear acceleration value to apply at time t
 *  \param t : the input time
 *  \return the linear acceleration value to apply at time t
 */
double DD2PCTraj::acc(double t) const
{
  return fctAcc()->getVal(t);
}

/**
 *  \brief return the angular acceleration value to apply at time t
 *  \param t : the input time
 *  \return the angular acceleration value to apply at time t
 */
double DD2PCTraj::aacc(double t) const
{
  return fctAAcc()->getVal(t);
}


// /**************** State computation ***************/

// /**
//  *  \brief return the state reached from s at time t after application of the control trajectory
//  *  \param s0 : the initial state from which starts the control trajectory
//  *  \param t : the time at which the output state is evaluated
//  *  \param sys : the robotic system constraints
//  *  \return the state reached from s at time t after application of the control trajectory
//  */
// DD2State DD2PCTraj::getStateReachedFrom(const DD2State& s0, double t, const DD2Sys* sys) const
// {
// }

// /**
//  *  \brief return the state reached from s at time t after application of the control trajectory
//  *  \param s0 : the initial state from which starts the control trajectory
//  *  \param t : the time at which the output state is evaluated
//  *  \param sys : the robotic system constraints
//  *  \return the state reached from s at time t after application of the control trajectory
//  */
// DD2StateT DD2PCTraj::getStateReachedFrom(const DD2StateT& s0, double t, const DD2Sys* sys) const
// {
// }


/**************** Boundary functions **************/

/**
 *  \brief add a minimal and a maximal bound over the linear acceleration profile
 *  \param minV : the minimal bound
 *  \param maxV : the maximal bound
 */
void DD2PCTraj::addAccBounds(double minA,double maxA)
{
  // add the minimal bound
  tabFunc_[2]->addMinBound(minA);
  // add the maximal bound
  tabFunc_[2]->addMaxBound(maxA);

  // get back the starting time and the initial velocity
  double t0 = this->startTime();
  double v0 = this->v(t0);

  // recompute the linear velocity profile if needed
  if(tabFunc_[0] != NULL) delete tabFunc_[0];
  tabFunc_[0] = fctAcc()->integration(t0,v0);
}

/**
 *  \brief add a minimal and a maximal bound over the angular acceleration profile
 *  \param minAV : the minimal bound
 *  \param maxAV : the maximal bound
 */
void DD2PCTraj::addAAccBounds(double minAA,double maxAA)
{
  // add the minimal bound
  tabFunc_[3]->addMinBound(minAA);
  // add the maximal bound
  tabFunc_[3]->addMaxBound(maxAA);

  // get back the starting time and the initial velocity
  double t0 = this->startTime();
  double w0 = this->w(t0);

  // recompute the angular velocity profile if needed
  if(tabFunc_[1] != NULL) delete tabFunc_[1];
  tabFunc_[1] = fctAAcc()->integration(t0,w0);
}

/**
 *  \brief add a minimal and a maximal bound over the linear velocity profile
 *  \param minV : the minimal bound
 *  \param maxV : the maximal bound
 */
void DD2PCTraj::addVelBounds(double minV,double maxV)
{
  // add the minimal bound
  tabFunc_[0]->addMinBound(minV);
  // add the maximal bound
  tabFunc_[0]->addMaxBound(maxV);

  // recompute the linear acceleration profile if needed
  if(tabFunc_[2] != NULL) delete tabFunc_[2];
  tabFunc_[2] = fctVel()->derivation();
}

/**
 *  \brief add a minimal and a maximal bound over the angular velocity profile
 *  \param minAV : the minimal bound
 *  \param maxAV : the maximal bound
 */
void DD2PCTraj::addAVelBounds(double minAV,double maxAV)
{
  // add the minimal bound
  tabFunc_[1]->addMinBound(minAV);
  // add the maximal bound
  tabFunc_[1]->addMaxBound(maxAV);

  // recompute the angular acceleration profile if needed
  if(tabFunc_[3] != NULL) delete tabFunc_[3];
  tabFunc_[3] = fctAVel()->derivation();
}


/**************** Utility functions ***************/

/**
 *  \brief sample the velocity control trajectory
 *  \return a sampled velocity control trajectory
 *  \param nb : the number of samples
 */
DD2CTraj* DD2PCTraj::sampleVelTraj(int nb) const
{
  // check if the input samping number is correct
  if(nb < 2) return NULL;
  // check if the trajectory profile has been defined
  double startTime = this->startTime();
  double endTime = this->endTime();
  if(endTime - startTime < 0) return NULL;

  // compute the sampling time used
  double dt = (endTime - startTime)/((double)(nb-1));

  // initialize the new trajectory
  DD2CTraj* traj = new DD2CTraj;
  
  for(int i=0; i<nb; i++)
    {
      double curT = startTime + i*dt;
      
      // compute the current velocity control
      DD2Control ctrl = this->at(curT);
      
      // append it to the sampled trajectory
      ((DD2CTraj*)traj)->append(DD2ControlT(ctrl.v(),ctrl.w(),curT));
    }
  
  return traj;
}

/**
 *  \brief sample the velocity control trajectory
 *  \return a sampled velocity control trajectory
 *  \param dt : the time interval between two samples
 */
DD2CTraj* DD2PCTraj::sampleVelTraj(double dt) const
{
  double startTime = this->startTime();
  double endTime = this->endTime();

  int nb = ((int)((endTime - startTime) / dt)) + 1;
  return this->sampleVelTraj(nb);
}

/**
 *  \brief sample the acceleration control trajectory
 *  \return a sampled acceleration control trajectory
 *  \param nb : the number of samples
 */
DD2AccCTraj* DD2PCTraj::sampleAccTraj(int nb) const
{
  // check if the input samping number is correct
  if(nb < 2) return NULL;
  // check if the trajectory profile has been defined
  double startTime = this->startTime();
  double endTime = this->endTime();
  if(endTime - startTime < 0) return NULL;

  // compute the sampling time used
  double dt = (endTime - startTime)/((double)(nb-1));

  // initialize the new trajectory
  DD2AccCTraj* traj = new DD2AccCTraj;
  
  for(int i=0; i<nb; i++)
    {
      double curT = startTime + i*dt;
      
      // compute the current acceleration control
      double acc = this->acc(curT);
      double aacc = this->aacc(curT);
      
      // append it to the sampled trajectory
      ((DD2AccCTraj*)traj)->append(DD2AccControlT(acc,aacc,curT));
    }
  
  return traj;
}

/**
 *  \brief sample the acceleration control trajectory
 *  \return a sampled acceleration control trajectory
 *  \param dt : the time interval between two samples
 */
DD2AccCTraj* DD2PCTraj::sampleAccTraj(double dt) const
{
  double startTime = this->startTime();
  double endTime = this->endTime();

  int nb = ((int)((endTime - startTime) / dt)) + 1;
  return this->sampleAccTraj(nb);
}

/**
 *  \brief sample the control trajectory
 *  \return a sampled control trajectory
 *  \param nb : the number of samples
 */
DD2CTraj* DD2PCTraj::sample(int nb) const
{
  return this->sampleVelTraj(nb);
}

/**
 *  \brief sample the control trajectory
 *  \return a sampled control trajectory
 *  \param dt : the time interval between two samples
 */
DD2CTraj* DD2PCTraj::sample(double dt) const
{
  return this->sampleVelTraj(dt);
}

/**
 *  \brief diplay the control trajectory
 */
void DD2PCTraj::display() const
{
  cout << "Linear Velocity V(t) : " << endl;
  tabFunc_[0]->display();
  cout << "Angular Velocity W(t) : " << endl;
  tabFunc_[1]->display();
  cout << "Linear Acceleration A(t) : " << endl;
  tabFunc_[2]->display();
  cout << "Angular Acceleration MU(t) : " << endl;
  tabFunc_[3]->display();
}

/**
 *  \brief return the polynomial coefficients of the linear and angular velocity profiles
 *  \param coefPolyVel : the resulting coefficients of the polynomial linear velocity profiles
 *  \param coefPolyAVel : the resulting coefficients of the polynomial angular velocity profiles
 */
void DD2PCTraj::getPolyCoef(vector<vector<double> >& coefPolyVel, vector<vector<double> >& coefPolyAVel) const
{
//   int resCoefVel = this->fctVel()->getPolyCoef(coefPolyVel, binf, bsup);
//   int resCoefAVel = this->fctAVel()->getPolyCoef(coefPolyAVel, binf, bsup);
  int resCoefVel = this->fctVel()->getPolyCoef(coefPolyVel);
  int resCoefAVel = this->fctAVel()->getPolyCoef(coefPolyAVel);

  if(resCoefVel != 0 || resCoefAVel != 0) cerr << "DD2PCTraj::getPolyCoef : Warning. the polynomial coefficient retrieval failed" << endl;
}

