#include "motion/inc/PSTraj.hpp"
using namespace mpc;

#include "math/inc/CosFunc.hpp"
#include "math/inc/SinFunc.hpp"
#include "math/inc/MultFunc.hpp"
#include "math/inc/ItgFunc.hpp"

                                    /**************************************************/
                                    /****************** Class PSTraj ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PSTraj::PSTraj() : STrajectory()
{
  type_ = ND_TRAJ;
  nbS_ = 0;
  tabFunc_ = NULL;
  initST_ = NULL;
}

/**
 *  \brief init constructor
 *  \param dim : dimension of the vector containing polynomial profiles described by the state trajectory (do not include quadratic profiles)
 *  \param startTime : the starting time of the state trajectory
 *  \param endTime : the ending time of the state trajectory
 */
PSTraj::PSTraj(int dim,double startTime,double endTime) : STrajectory()
{
  this->type_ = ND_TRAJ;
  this->nbS_ = dim;
  this->tabFunc_ = new MultiItvFunction* [nbS_];
  for(int i=0; i<nbS_; i++)
    {
      Polynomial poly(0.0);
      this->tabFunc_[i] = new MultiItvFunction(&poly, startTime, endTime);
    }
  this->initST_ = NULL;
}

/**
 *  \brief copy constructor
 */
PSTraj::PSTraj(const PSTraj& traj) : STrajectory(traj)
{
  this->type_ = traj.type_;
  this->nbS_ = traj.nbS_;
  this->tabFunc_ = new MultiItvFunction* [nbS_];
  for(int i=0; i<nbS_; i++)
    {
      this->tabFunc_[i] = traj.tabFunc_[i]->copy();
    }
  this->initST_ = NULL;
}

/**
 *  \brief destructor
 */
PSTraj::~PSTraj()
{
  if(this->tabFunc_ != NULL) 
    {
      for(int i=0; i<nbS_; i++)
	{
	  if(tabFunc_[i] != NULL) delete tabFunc_[i];
	}
      delete [] tabFunc_;
    }
  if(initST_ != NULL) delete initST_;
}


/************** Getters and setters ***************/

/**
 *  \brief return the start time of the control trajectory
 *  \return the start time of the control trajectory
 */
double PSTraj::startTime() const
{
  if(nbS_ <= 0) return -INFINITY;
  else return tabFunc_[0]->getBInf();
}

/**
 *  \brief return the end time of the control trajectory
 *  \return the end time of the control trajectory
 */
double PSTraj::endTime() const
{
  if(nbS_ <= 0) return INFINITY;
  else return tabFunc_[0]->getBSup();
}


/******** Control trajectory modifications ********/

/**
 *  \brief concatenate two trajectories
 *  \param traj : the trajectory to concatenate to the current one
 */
void PSTraj::concat(const PSTraj& traj)
{
  // check if the type of both trajectories is the same
  if(this->type_ != traj.type_)
    {
      cerr << "PSTraj::concat : the type of the input state trajectory differs from the current one." << endl;
      throw std::string("PSTraj::concat : the type of the input state trajectory differs from the current one.");
    }

  // check if the final time of the current trajectory is finite
  if(this->endTime() == INFINITY)
    {
      cerr << "PSTraj::concat : the final time of the current trajectory is infinite." << endl;
      throw std::string("PSTraj::concat : the final time of the current trajectory is infinite.");
    }

  // build a copy of the input control trajectory
  PSTraj* trajCopy = traj.copy();

  if(trajCopy == NULL)
    {
      cerr << "PSTraj::concat : the copy of the input state trajectory is null." << endl;
      throw std::string("PSTraj::concat : the copy of the input state trajectory is null.");
    }

  trajCopy->shift(this->endTime() - trajCopy->startTime());
  
  // concatenate each component of the control trajectory
  for(int i=0; i<nbS_; i++)
    {
      this->tabFunc_[i]->concat(*(trajCopy->tabFunc_[i]));
    }
  
  // delete the input control trajectory copy
  delete trajCopy;
}

/**
 *  \brief shift the trajectory by time t
 *  \param t : the time to shift the trajectory
 */
void PSTraj::shift(double t)
{
  for(int i=0; i<nbS_; i++)
    {
      this->tabFunc_[i]->shiftTime(t);
    }
}


                                    /**************************************************/
                                    /**************** Class DD2PSTraj ****************/
                                    /**************************************************/

/**
 *  \brief default constructor
 */
DD2PSTraj::DD2PSTraj() : PSTraj(5)
{
  this->type_ = DD2_TRAJ;
  this->initST_ = new DD2StateT(0,0,0,0,0,0);
}

/**
 *  \brief copy constructor
 */
DD2PSTraj::DD2PSTraj(const DD2PSTraj& traj) : PSTraj(traj)
{
  this->initST_ = new DD2StateT(*((DD2StateT*)(traj.initST_)));
}

/**
 *  \brief init constructor (from a state and a control trajectory)
 *  \param s : the initial state
 *  \param ctraj : the input control trajectory
 *  \param minV, maxV : bounds over the linear velocity
 *  \param minW, maxW : bounds over the angular velocity
 */
DD2PSTraj::DD2PSTraj(const DD2State& s, const DD2PCTraj& ctraj, 
		       double minV, double maxV, 
		       double minW, double maxW)
{
  double startTime = ctraj.startTime();
  double endTime = ctraj.endTime();

  this->type_ = DD2_TRAJ;
  this->nbS_ = 5;
  this->initST_ = new DD2StateT(s,startTime);
  
  this->tabFunc_ = new MultiItvFunction* [nbS_];

  // V
  this->tabFunc_[DD2_INDV] = new MultiItvFunction(*(ctraj.fctVel()));
  // add the bounds over the linear velocity function if needed
  if(minV != -INFINITY) tabFunc_[DD2_INDV]->addMinBound(minV);
  if(maxV != INFINITY) tabFunc_[DD2_INDV]->addMinBound(maxV);

  // W
  this->tabFunc_[DD2_INDW] = new MultiItvFunction(*(ctraj.fctAVel()));
  // add the bounds over the angular velocity function if needed
  if(minW != -INFINITY) tabFunc_[DD2_INDW]->addMinBound(minW);
  if(maxW != INFINITY) tabFunc_[DD2_INDW]->addMinBound(maxW);

  // Th
  this->tabFunc_[DD2_INDTH] = fctAVel()->integration(0.0,s.th());;

  // X
  ItgFunc itgFuncX(new MultFunc(ctraj.fctVel()->copy(), new CosFunc(tabFunc_[DD2_INDTH]->copy() ) ), initST().x());
  this->tabFunc_[DD2_INDX] = new MultiItvFunction(&itgFuncX);

//   cout << "tabFunc_[DD2_INDX] : " << endl;
//   tabFunc_[DD2_INDX]->display();

  // Y
  ItgFunc itgFuncY(new MultFunc(ctraj.fctVel()->copy(), new SinFunc(tabFunc_[DD2_INDTH]->copy() ) ), initST().y());
  this->tabFunc_[DD2_INDY] = new MultiItvFunction(&itgFuncY);
//   cout << "tabFunc_[DD2_INDY] : " << endl;
//   tabFunc_[DD2_INDY]->display();

//   cout << "____________________________ DD2PSTraj::DD2PSTraj ____________________________" << endl;
//   this->display();
//   cout << "________________________________________________________________________________" << endl;
}

/**
 *  \brief destructor
 */
DD2PSTraj::~DD2PSTraj()
{
}

/**
 *  \brief return a copy of the current state trajectory
 *  \return a copy of the current state trajectory
 *
 *  Return a new instance of PSTraj
 */
DD2PSTraj* DD2PSTraj::copy() const
{
  DD2PSTraj* newTraj = new DD2PSTraj(*this);
  return newTraj;
}


/************** Getters and setters ***************/

/**
 *  \brief getter : first state of the trajectory
 *  \return the first state of the trajectory
 */
DD2StateT DD2PSTraj::firstStateT() const
{
  double t = this->startTime();

  DD2StateT statet(this->x(t),
		    this->y(t),
		    this->th(t),
		    this->v(t),
		    this->w(t),
		    t);
  return statet;
}

/**
 *  \brief getter : last state of the trajectory
 *  \return the last state of the trajectory
 */
DD2StateT DD2PSTraj::lastStateT() const
{
  double t = this->endTime();

  DD2StateT statet(this->x(t),
		    this->y(t),
		    this->th(t),
		    this->v(t),
		    this->w(t),
		    t);
  return statet;
}

/**
 *  \brief getter : get the state-time defined at time t
 *  \param t : the input time
 *  \return return the state defined at time t
 */
DD2StateT DD2PSTraj::at(double t) const
{
  DD2StateT statet(this->x(t),
		    this->y(t),
		    this->th(t),
		    this->v(t),
		    this->w(t),
		    t);
  return statet;
}

// /**
//  *  \brief Quadrature tools...
//  */  
// void DD2PSTraj::gauleg(const double x1, const double x2, double x[], double w[], int n) const
// {
//     const double EPS=1.0e-14;
//     int m,j,i;
//     double z1,z,xm,xl,pp,p3,p2,p1;
    
//     m=(n+1)/2;
//     xm=0.5*(x2+x1);
//     xl=0.5*(x2-x1);
//     for (i=0;i<m;i++) {
//         z=cos(3.141592654*(i+0.75)/(n+0.5));
//         do {
//             p1=1.0;
//             p2=0.0;
//             for (j=0;j<n;j++) {
//                 p3=p2;
//                 p2=p1;
//                 p1=((2.0*j+1.0)*z*p2-j*p3)/(j+1);
//             }
//             pp=n*(z*p1-p2)/(z*z-1.0);
//             z1=z;
//             z=z1-p1/pp;
//         } while (fabs(z-z1) > EPS);
//         x[i]=xm-xl*z;
//         x[n-1-i]=xm+xl*z;
//         w[i]=2.0*xl/((1.0-z*z)*pp*pp);
//         w[n-1-i]=w[i];
//     }
// }

// /**
//  *  \brief Quadrature tools...
//  */
// double DD2PSTraj::qgaus(double (DD2PSTraj::*func)(const double) const, const double a, const double b) const
// {
//   static const int n = 20;
//   double x[n];
//   double w[n];
//   int j;
//   double s;
  
//   gauleg(a, b, x, w, n);      
//   s=0;
//   for (j=0;j<n;j++) {        
//     s += w[j]*((this->*func)(x[j]));
//     }
//   return s;
// }

/*
 *  \brief function to integrate to obtain the X(t) function
 *  \param t : the time variable
 *  \return V(t) * sin(Th(t))
 */
double DD2PSTraj::itgX(double t) const
{
  return v(t) * cos(th(t));
}

/*
 *  \brief function to integrate to obtain the Y(t) function
 *  \param t : the time variable
 *  \return V(t) * cos(Th(t))
 */
double DD2PSTraj::itgY(double t) const
{
  return v(t) * sin(th(t));
}

/**
 *  \brief return the position x reached at time t
 *  \param t : the input time
 *  \return the position x reached at time t
 */
double DD2PSTraj::x(double t) const
{
//   double t_0 = this->startTime();
//   return initST().x() + qgaus(&DD2PSTraj::itgX,t_0,t);

  return tabFunc_[DD2_INDX]->getVal(t);
}

/**
 *  \brief return the position y reached at time t
 *  \param t : the input time
 *  \return the position y reached at time t
 */
double DD2PSTraj::y(double t) const
{
//   double t_0 = this->startTime();
//   return initST().y() + qgaus(&DD2PSTraj::itgY,t_0,t);

  return tabFunc_[DD2_INDY]->getVal(t);
}

/**
 *  \brief return the position th reached at time t
 *  \param t : the input time
 *  \return the position th reached at time t
 */
double DD2PSTraj::th(double t) const
{
  return tabFunc_[DD2_INDTH]->getVal(t);
}

/**
 *  \brief return the linear velocity reached at time t
 *  \param t : the input time
 *  \return the linear velocity v reached at time t
 */
double DD2PSTraj::v(double t) const
{
  return tabFunc_[DD2_INDV]->getVal(t);
}

/**
 *  \brief return the angular velocity w reached at time t
 *  \param t : the input time
 *  \return the angular velocity w reached at time t
 */
double DD2PSTraj::w(double t) const
{
  return tabFunc_[DD2_INDW]->getVal(t);
}


/**************** Boundary functions **************/

/**
 *  \brief add a minimal and a maximal bound over the linear velocity profile
 *  \param minV : the minimal bound
 *  \param maxV : the maximal bound
 */
void DD2PSTraj::addVelBounds(double minV,double maxV)
{
  // add the minimal bound
  tabFunc_[DD2_INDV]->addMinBound(minV);
  // add the maximal bound
  tabFunc_[DD2_INDV]->addMaxBound(maxV);
}

/**
 *  \brief add a minimal and a maximal bound over the angular velocity profile
 *  \param minAV : the minimal bound
 *  \param maxAV : the maximal bound
 */
void DD2PSTraj::addAVelBounds(double minAV,double maxAV)
{
  // add the minimal bound
  tabFunc_[DD2_INDW]->addMinBound(minAV);
  // add the maximal bound
  tabFunc_[DD2_INDW]->addMaxBound(maxAV);
}


/**************** Utility functions ***************/

/**
 *  \brief sample the state trajectory
 *  \param nb : the number of samples
 *  \return a sampled state trajectory
 *
 *  A new instance of DD2STraj is returned
 */
DD2STraj* DD2PSTraj::sample(int nb) const
{
  // check if the input samping number is correct
  if(nb < 2) return NULL;
  // check if the trajectory profile has been defined
  double startTime = this->startTime();
  double endTime = this->endTime();
  if(endTime - startTime < 0) return NULL;

  // initialize the new trajectory
  DD2STraj* traj = new DD2STraj;

  // compute the sampling time used
  double dt = (endTime - startTime)/((double)(nb-1));

  for(int i=0; i<nb; i++)
    {
      // compute the current state-time
      DD2StateT statet;
      if(i == 0) statet = this->at(startTime);
      else if(i == nb-1) statet = this->at(endTime);
      else statet = this->at(startTime + i*dt);

      // append it to the sampled trajectory
      traj->append(statet);
    }
  return traj;
}

/**
 *  \brief sample the state trajectory
 *  \param dt : the time interval between two samples
 *  \return a sampled state trajectory
 *
 *  A new instance of DD2STraj is returned
 */
DD2STraj* DD2PSTraj::sample(double dt) const
{
  double startTime = this->startTime();
  double endTime = this->endTime();

  int nb = ((int)((endTime - startTime) / dt)) + 1;
  return this->sample(nb);
}

/**
 *  \brief diplay the control trajectory
 */
void DD2PSTraj::display() const
{
  cout << "Position X(t) : " << endl;
//   cout << "X_0 + Int_{t_0}^t V(t) cos(Th(t)) dt" << endl;
  tabFunc_[DD2_INDX]->display();
  cout << "Position Y(t) : " << endl;
//   cout << "Y_0 + Int_{t_0}^t V(t) sin(Th(t)) dt" << endl;
  tabFunc_[DD2_INDY]->display();
  cout << "Heading Th(t) : " << endl;
  tabFunc_[DD2_INDTH]->display();
  cout << "Linear Velocity V(t) : " << endl;
  tabFunc_[DD2_INDV]->display();
  cout << "Angular Velocity W(t) : " << endl;
  tabFunc_[DD2_INDW]->display();
}
