#include "Path.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

#include <fstream>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /******************* Class Path *******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Path::Path()
{
  this->type_ = ND_PATH;
}

/**
 *  \brief destructor
 */
Path::~Path()
{
}


                                    /**************************************************/
                                    /******************* Class CPath ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CPath::CPath() : Path()
{
}

/**
 *  \brief destructor
 */
CPath::~CPath()
{
  this->clear();
}


/************** Getters and setters ***************/

/**
 *  \brief compute the length of the path
 *  \return the length of the path
 */
double CPath::getLength() const
{
  double res=0;
  int N = this->size();
  for(int i=0; i<N-1; i++)
    {
      PositionXY p1 = this->at(i).posXY();
      PositionXY p2 = this->at(i+1).posXY();

      double dx = p2.x() - p1.x();
      double dy = p2.y() - p1.y();
      double dist = sqrt(dx*dx + dy*dy);
      res += dist;
    }
  return res;
}

/**
 *  \brief compute the length of the path up to the n-th node
 *  \param n : the input index of node
 *  \return the length of the path up to the n-th node
 */
double CPath::getLengthUpTo(int n) const
{
  double res=0;
  int N = this->size();
  
  // check if the input index is available
  if(n > N-1) n = N-1;

  for(int i=0; i<n; i++)
    {
      PositionXY p1 = this->at(i).posXY();
      PositionXY p2 = this->at(i+1).posXY();

      double dx = p2.x() - p1.x();
      double dy = p2.y() - p1.y();
      double dist = sqrt(dx*dx + dy*dy);
      res += dist;
    }
  return res;
}

/**
 *  \brief compute the length of the path between the a-th and the b-th node
 *  \param a : the first index of node
 *  \param b : the second index of node
 *  \return the length of the path between the a-th and the b-th node
 */
double CPath::getLengthBetween(int a,int b) const
{
  double res=0;
  int N = this->size();
  
  // check if the input index is available
  if(b > N-1) b = N-1;

  for(int i=a; i<b; i++)
    {
      PositionXY p1 = this->at(i).posXY();
      PositionXY p2 = this->at(i+1).posXY();

      double dx = p2.x() - p1.x();
      double dy = p2.y() - p1.y();
      double dist = sqrt(dx*dx + dy*dy);
      res += dist;
    }
  return res;
}

/**
 *  \brief return true if the path is empty
 *  \return true if the path is empty
 */
bool CPath::empty() const
{
  return listCfg_.empty();
}


/******** Configurations list modification ********/

/**
 *  \brief reverse the current path
 */
void CPath::reverse()
{
  cout << "Path::reverse : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief clear the list of configurations
 */
void CPath::clear()
{
  int N = this->size();
  for(int i=0; i<N; i++)
    {
      delete (this->listCfg_[i]);
    }
  this->listCfg_.clear();
}

  
/**************** Utility functions ***************/

/**
 *  \brief diplay the path
 */
void CPath::display() const
{
  cout << "Path : " << endl;
  int N = this->size();
  for(int i=0; i<N-1; i++)
    {
      this->at(i).display();
      cout << " -> ";
    }
  if(N != 0) this->at(N-1).display();
  cout << endl;
}

/**
 *  \brief display in an openGL viewer
 */
void CPath::toGL() const
{
  glBegin(GL_POINTS);
  int N = this->size();
  for(int i=0; i<N; i++)
    {
      PositionXY pos = this->at(i).posXY();
      glVertex3f(pos.x(),pos.y(),0);
    }
  glEnd();
}

/**
 *  \brief return the barycenter of the path in the workspace
 */
PositionXY CPath::center() const
{
  if(this->size() == 0)
    return PositionXY();
  else
    {
      PositionXY pres;
      int N = this->size();
      for(int i=0; i<N; i++)
	{
	  PositionXY pos = this->at(i).posXY();
	  pres += pos;
	}
      return PositionXY(pres.x()/((double)N),pres.y()/((double)N));
    }
}

/**
 *  \brief compute an approximative radius of the set of all configurations of the path
 *  \return an approximative radius of the set of all configurations of the path
 */
double CPath::radius() const
{
  if(this->size() == 0)
    return 0;
  else
    {
      // compute the center of the path
      PositionXY center = this->center();

      // search for the further configuration of the path from its center
      int N = this->size();
      double minSqDist = -1;
      for(int i=0; i<N; i++)
	{
	  // get the position of the current configuration
	  PositionXY pos = this->at(i).posXY();
	  double dx = center.x() - pos.x();
	  double dy = center.y() - pos.y();
	  double sqDist = dx*dx + dy*dy;

	  if(i == 0 || sqDist < minSqDist) minSqDist = sqDist;
	}

      return sqrt(minSqDist);
    }
}


                                    /**************************************************/
                                    /****************** Class PMPath ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PMPath::PMPath()
{
  this->type_ = PM_PATH;
}

/**
 *  \brief copy constructor
 *  \param t : the path to copy
 */
PMPath::PMPath(const PMPath& p)
{
  this->type_ = PM_PATH;

  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}

/**
 *  \brief destructor
 */
PMPath::~PMPath()
{
}


/************** Getters and setters ***************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
const PMConfig& PMPath::at(int i) const
{
  return *((PMConfig*)(this->listCfg_.at(i)));
}

/**
 *  \brief getter : middle configuration of the path
 *  \return the middle configuration of the path
 */
const PMConfig& PMPath::middleCfg() const
{
  if(this->size() != 0)
    return this->at( ((int) (0.5*this->size())) );
  else
    throw string("Empty path");
}

/**
 *  \brief getter : first configuration of the path
 *  \return the first configuration of the path
 */
const PMConfig& PMPath::firstCfg() const
{
  if(this->size() != 0)
    return this->at(0);
  else
    throw string("Empty path");
}

/**
 *  \brief getter : last configuration of the path
 *  \return the last configuration of the path
 */
const PMConfig& PMPath::lastCfg() const
{
  if(this->size() != 0)
    return this->at(this->size()-1);
  else
    throw string("Empty path");
}


/******** Configurations list modification ********/

/**
 *  \brief append a new configuration to the path
 *  \param c : the configuration to append
 */
void PMPath::append(const PMConfig& c)
{
  PMConfig* newC = new PMConfig(c);
  this->listCfg_.push_back(newC);
}

/**
 *  \brief prepend a new configuration to the path
 *  \param c : the configuration to prepend
 */
void PMPath::prepend(const PMConfig& c)
{
  vector <Config*>::iterator it = this->listCfg_.begin();
  this->listCfg_.insert(it,new PMConfig(c));
}

/**
 *  \brief concatenate a path to the current one
 *  \param p : the input path to concatenate
 */
void PMPath::concat(const PMPath& p)
{
  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}


/******************** Operators *******************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
PMConfig& PMPath::operator[](int i)
{
  return *((PMConfig*)(this->listCfg_[i]));
}


/************** Path transformations **************/

/**
 *  \brief shift the path coordinates on the plane
 *  \param dx : shift on the X axis
 *  \param dy : shift on the Y axis
 */
void PMPath::shiftXY(double dx, double dy)
{
  int N = this->size();

  for(int i=0; i<N; i++)
    {
      (*this)[i].x() += dx;
      (*this)[i].y() += dy;
    }
}


/**************** Utility functions ***************/

// /**
//  *  \brief diplay the path
//  */
// void PMPath::display() const
// {  
// }

/*
 *  \brief search for the minimal distance from the path to the input position
 *  \param (x,y) : the input position
 *  \return the resulting minimal distance
 */
double PMPath::minDistFrom(double x, double y)
{
  int N = this->size();

  double minDist = 999999999;
  for(int i=0; i<N-1; i++)
    {
      PMConfig curCfg = this->at(i);
      PMConfig nextCfg = this->at(i+1);

      PositionXY p1(curCfg.x(), curCfg.y());
      PositionXY p2(nextCfg.x(), nextCfg.y());

      double curMinDist = mpc::closeDistSgmt(PositionXY(x,y), p1, p2);

      if(i == 0 || curMinDist < minDist)
	{
	  minDist = curMinDist;
	}
    }
  return minDist;
}

/*
 *  \brief get back the index of the input coord in a grid
 *  \param (nbCols, nbLines) : number of columns and lines of the image
 *  \param (xOrigin, yOrigin) : coordinates of the top left corner of the image
 *  \param (xScale, yScale) : scales of the image (length between two cells of the image)
 *  \param (x,y) : input coordinates (in the world)
 *  \param (i,j) : output coordinates in the image (column, line)
 */
void PMPath::getIndexPos(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale, float x,float y,int& i,int& j)
{
  i = nbLines - 1 - (int)round((y - yOrigin) / yScale);
  j = (int)round((x - xOrigin) / xScale);
}

/**
 *  \brief draw the path on a rgb image
 *  \param image : the input image
 *  \param (xOrigin, yOrigin) : coordinates of the top left corner of the image
 *  \param (xScale, yScale) : scales of the image (length between two cells of the image)
 */
void PMPath::drawOnRGBImage(RGBImage& image, float xOrigin, float yOrigin, float xScale, float yScale, const RGBColor& color) const
{
  int N = this->size();

  for(int i=0; i<N-1; i++)
    {
      // get back the successive configurations of the path
      PMConfig curCfg = this->at(i);
      PMConfig nextCfg = this->at(i+1);

      // get back the coordinates of both config in the image
      int i1, j1, i2, j2;
      PMPath::getIndexPos(image.nbCols(), image.nbLines(), xOrigin, yOrigin, xScale, yScale, curCfg.x(), curCfg.y(), i1, j1);
      PMPath::getIndexPos(image.nbCols(), image.nbLines(), xOrigin, yOrigin, xScale, yScale, nextCfg.x(), nextCfg.y(), i2, j2);

      // draw a line between both end points on the image
      image.drawLine(i1, j1, i2, j2, color);
    }
}

/**
 *  \brief get the boundaries of the path
 *  \param minX : the minimal X coordinate
 *  \param maxX : the maximal X coordinate
 *  \param minY : the minimal Y coordinate
 *  \param maxY : the maximal Y coordinate
 *  \return 0 if the process succeeded, -1 otherwise (if the path is empty)
 */
int PMPath::getBounds(float& minX, float& maxX, float& minY, float& maxY) const
{
  int N = this->size();
  if(N == 0) return -1;

  for(int i=0; i<N-1; i++)
    {
      PMConfig curCfg = this->at(i);
      if(i == 0)
	{
	  minX = curCfg.x();
	  maxX = curCfg.x();
	  minY = curCfg.y();
	  maxY = curCfg.y();
	}
      else
	{
	  if(curCfg.x() < minX) minX = curCfg.x();
	  if(curCfg.x() > maxX) maxX = curCfg.x();
	  if(curCfg.y() < minY) minY = curCfg.y();
	  if(curCfg.y() > maxY) maxY = curCfg.y();
	}
    }
  return 0;
}


/************* Input/Output functions *************/

/**
 *  \brief load the path from an input file
 *  \param nomfic : name of the input file
 */
bool PMPath::load(char* nomfic)
{
  // delete the old path if exists
  if(this->size() != 0)
    {
      this->clear();
    }

  // create an input file
  ifstream file;
  
  // open the file
  file.open(nomfic);

  if(file.fail())
    {
      return false;
    }

  // read all configurations of the path
  while(!file.eof())
    {
      double x,y;
      file >> x;
      file >> y;
      PMConfig cfg(x,y);
      this->append(cfg);
    }

  // close the file
  file.close();

  return true;
}

/**
 *  \brief save the path in an output file
 *  \param nomfic : name of the output file
 */
bool PMPath::save(char* nomfic) const
{
  // create an output pile
  ofstream file;

  // open the output file
  file.open(nomfic);

  // write all configurations of the path
  for(int i=0; i< this->size(); i++)
    {
      file << this->at(i).x();
      file << " ";
      file << this->at(i).y();
      file << endl;
    }

  // close the file
  file.close();
  
  return true;
}

                                    /**************************************************/
                                    /***************** Class DPMPath ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DPMPath::DPMPath()
{
  this->type_ = DPM_PATH;
}

/**
 *  \brief copy constructor
 *  \param t : the path to copy
 */
DPMPath::DPMPath(const DPMPath& p)
{
  this->type_ = DPM_PATH;

  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}

/**
 *  \brief destructor
 */
DPMPath::~DPMPath()
{
}


/************** Getters and setters ***************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
const DPMConfig& DPMPath::at(int i) const
{
  return *((DPMConfig*)(this->listCfg_.at(i)));
}

/**
 *  \brief getter : middle configuration of the path
 *  \return the middle configuration of the path
 */
const DPMConfig& DPMPath::middleCfg() const
{
  if(this->size() != 0)
    return this->at( ((int) (0.5*this->size())) );
  else
    throw string("Empty path");
}

/**
 *  \brief getter : first configuration of the path
 *  \return the first configuration of the path
 */
const DPMConfig& DPMPath::firstCfg() const
{
  if(this->size() != 0)
    return this->at(0);
  else
    throw string("Empty path");
}

/**
 *  \brief getter : last configuration of the path
 *  \return the last configuration of the path
 */
const DPMConfig& DPMPath::lastCfg() const
{
  if(this->size() != 0)
    return this->at(this->size()-1);
  else
    throw string("Empty path");
}


/******** Configurations list modification ********/

/**
 *  \brief append a new configuration to the path
 *  \param c : the configuration to append
 */
void DPMPath::append(const DPMConfig& c)
{
  DPMConfig* newC = new DPMConfig(c);
  this->listCfg_.push_back(newC);
}

/**
 *  \brief prepend a new configuration to the path
 *  \param c : the configuration to prepend
 */
void DPMPath::prepend(const DPMConfig& c)
{
  vector <Config*>::iterator it = this->listCfg_.begin();
  this->listCfg_.insert(it,new DPMConfig(c));
}

/**
 *  \brief concatenate a path to the current one
 *  \param p : the input path to concatenate
 */
void DPMPath::concat(const DPMPath& p)
{
  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}

/**
 *  \brief return an inverted path from the current one
 *  \return an inverted path from the current one
 */
DPMPath* DPMPath::invert() const
{
  DPMPath* inverted_path = new DPMPath();

  int N = this->size();
  for(int i=N-1; i>=0; i--)
    {
      DPMConfig cfg = this->at(i);
      cfg.th() = normalizedAngle(cfg.th() + M_PI);

      inverted_path->append(cfg);
    }
  
  return inverted_path;
}


/******************** Operators *******************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
DPMConfig& DPMPath::operator[](int i)
{
  return *((DPMConfig*)(this->listCfg_[i]));
}


/**************** Utility functions ***************/

// /**
//  *  \brief diplay the path
//  */
// void PMPath::display() const
// {  
// }


/************** Path transformations **************/

/**
 *  \brief lateral shift of the path coordinates on the plane
 *  \param dist : shift distance
 */
void DPMPath::shiftLat(double dist)
{
  // check if the path is not empty
  if(this->empty()) return;

  // get back the initial heading of the path
  DPMConfig cfg0 = this->at(0);
  double th0 = cfg0.th();

  // compute the shift of the path on X and Y axes
  double dx = dist * -sin(th0);
  double dy = dist * cos(th0);

  this->shiftXY(dx,dy);
}

/**
 *  \brief rotate the path around its origin
 *  \param ang : rotation angle
 */
void DPMPath::rotate(double ang)
{
  // check if the path is not empty
  if(this->empty()) return;

  // get back the initial heading of the path
  DPMConfig cfg0 = this->at(0);
  double th0 = cfg0.th();

  int N = this->size();

  for(int i=0; i<N; i++)
    {
      DPMConfig cfg = this->at(i);
  
      double dx = cfg.x() - cfg0.x();
      double dy = cfg.y() - cfg0.y();
      double dist = sqrt(dx*dx + dy*dy);
      double dir = atan2(dy,dx);

      if(dist != 0)
	{
	  (*this)[i].x() = cfg0.x() + dist * cos(normalizedAngle(dir + ang));
	  (*this)[i].y() = cfg0.y() + dist * sin(normalizedAngle(dir + ang));
	}
      (*this)[i].th() = normalizedAngle((*this)[i].th() + ang);
    }
}


/************* Input/Output functions *************/

/**
 *  \brief load the path from an input file
 *  \param nomfic : name of the input file
 */
bool DPMPath::load(char* nomfic)
{
  // delete the old path if exists
  if(this->size() != 0)
    {
      this->clear();
    }

  // create an input file
  ifstream file;
  
  // open the file
  file.open(nomfic);

  if(file.fail())
    {
      return false;
    }

  // read all configurations of the path
  while(!file.eof())
    {
      double x,y,th;
      file >> x;
      file >> y;
      file >> th;
      DPMConfig cfg(x,y,th);
      this->append(cfg);
    }

  // close the file
  file.close();

  return true;
}

/**
 *  \brief save the path in an output file
 *  \param nomfic : name of the output file
 */
bool DPMPath::save(char* nomfic) const
{
  // create an output pile
  ofstream file;

  // open the output file
  file.open(nomfic);

  // write all configurations of the path
  for(int i=0; i< this->size(); i++)
    {
      file << this->at(i).x();
      file << " ";
      file << this->at(i).y();
      file << " ";
      file << this->at(i).th();
      file << endl;
    }

  // close the file
  file.close();
  
  return true;
}

/**
 *  \brief load the path (for a robufast car) from an input file
 *  \param nomfic : name of the input file
 */
bool DPMPath::loadRobufastPath(char* nomfic)
{
  cout << "DPMPath::loadRobufastPath deb" << endl;

  // delete the old path if exists
  if(this->size() != 0)
    {
      this->clear();
    }

  // create an input file
  ifstream file;
  
  // open the file
  file.open(nomfic, ios::in);

  if(file.fail())
    {
      cerr << "DPMPath::loadRobufastPath : Unable to load file" << nomfic << endl;
      return false;
    }

      cerr << "DPMPath::loadRobufastPath : File found" << nomfic << endl;

  // read all configurations of the path
  while(!file.eof())
    {
      double longitudeLambert, latitudeLambert, vitesseGpsRtk, CapGpsRtk, VitesseAngYaw, time;
      double angleAV,angleAR,VelRR,VelRL,VelFR,VelFL, unknown, IntRR,IntRL,IntFR,IntFL;

      file >> longitudeLambert;
      file >> latitudeLambert;
      file >> vitesseGpsRtk;
      file >> CapGpsRtk;
      file >> VitesseAngYaw;
      file >> time;
      file >> angleAV; 
      file >> angleAR; 
      file >> VelRR; 
      file >> VelRL; 
      file >> VelFR; 
      file >> VelFL; 
      file >> unknown; 
      file >> IntRR; 
      file >> IntRL; 
      file >> IntFR; 
      file >> IntFL;

      DPMConfig cfg(longitudeLambert, latitudeLambert, normalizedAngle(CapGpsRtk));
      this->append(cfg);

      cout << "new Cfg : "; cfg.display(); cout << endl;
    }

  // close the file
  file.close();

  return true;
}


                                    /**************************************************/
                                    /**************** Class XYThS_Path ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
XYThS_Path::XYThS_Path()
{
  this->type_ = XYTHS_PATH;
}

/**
 *  \brief copy constructor
 *  \param t : the path to copy
 */
XYThS_Path::XYThS_Path(const XYThS_Path& p)
{
  this->type_ = XYTHS_PATH;

  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}

/**
 *  \brief copy constructor (from a DPMPath)
 *  \param t : the path to copy
 */
XYThS_Path::XYThS_Path(const DPMPath& p)
{
  this->type_ = XYTHS_PATH;

  int N = p.size();
  double s = 0;
  DPMConfig prevCfg;

  if(N != 0)
    {
      DPMConfig initDpmCfg = p.at(0);
      XYThS_Config initCfg(initDpmCfg.x(), initDpmCfg.y(), initDpmCfg.th(), 0);
      prevCfg = initDpmCfg;

      this->append(initCfg);
    }

  for(int i=1; i<N; i++)
    {
      // get back the current configuration
      DPMConfig dpmCfg = p.at(i);

      // compute the current curve abscica
      double dx = dpmCfg.x() - prevCfg.x();
      double dy = dpmCfg.y() - prevCfg.y();
      double ds = sqrt(dx*dx + dy*dy);
      s += ds;

      // append the new configuration to the current path
      XYThS_Config cfg(dpmCfg.x(), dpmCfg.y(), dpmCfg.th(), s);
      this->append(cfg);

      // replace the previous configuration
      prevCfg = dpmCfg;
    }
}


/**
 *  \brief destructor
 */
XYThS_Path::~XYThS_Path()
{
}


/************** Getters and setters ***************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
const mpc::XYThS_Config& XYThS_Path::at(int i) const
{
  return *((XYThS_Config*)(this->listCfg_.at(i)));
}

/**
 *  \brief getter : middle configuration of the path
 *  \return the middle configuration of the path
 */
const mpc::XYThS_Config& XYThS_Path::middleCfg() const
{
  if(this->size() != 0)
    return this->at( ((int) (0.5*this->size())) );
  else
    throw string("Empty path");
}

/**
 *  \brief getter : first configuration of the path
 *  \return the first configuration of the path
 */
const mpc::XYThS_Config& XYThS_Path::firstCfg() const
{
  if(this->size() != 0)
    return this->at(0);
  else
    throw string("Empty path");
}

/**
 *  \brief getter : last configuration of the path
 *  \return the last configuration of the path
 */
const mpc::XYThS_Config& XYThS_Path::lastCfg() const
{
  if(this->size() != 0)
    return this->at(this->size()-1);
  else
    throw string("Empty path");
}

/**
 *  \brief put in cfg the configuration of the path given at curve abscica s
 *  \param s : the input curve abscica
 *  \param cfg : the resulting configuration
 */
void XYThS_Path::getCfgAtCAbs(double s, XYThS_Config& cfg) const
{
  // check if the path vector is not null
  if(this->size() == 0)
    {
      cfg = XYThS_Config();
      return;
    }
  
  // check if the curve abscica belongs to the path
  if(s < 0)
    {
      // return the initial state of the path
      cfg = this->at(0);
    }
  else if(s >= this->at(this->size()-1).s())
    {
      // return the final configuration of the path
      cfg = this->at(this->size()-1);
    }
  else
    {
      // general case : search for both configurations bounding the curve abscica
      bool trouve = false;
      int i=1;
      while(!trouve)
	{
	  if(s < this->at(i).s()) trouve = true;
	  else i++;
	}

      // get back both configurations bounding the curve absica
      XYThS_Config prevCfg = this->at(i-1);
      XYThS_Config nextCfg = this->at(i);

      // compute a linear interpolation between both configurations
      double coef = (s - prevCfg.s()) / (nextCfg.s() - prevCfg.s());

      cfg = (prevCfg * (1.0 - coef)) + (nextCfg * coef);
    }
}


/******** Configurations list modification ********/

/**
 *  \brief append a new configuration to the path
 *  \param c : the configuration to append
 */
void XYThS_Path::append(const mpc::XYThS_Config& c)
{
  XYThS_Config* newC = new XYThS_Config(c);
  this->listCfg_.push_back(newC);
}

/**
 *  \brief prepend a new configuration to the path
 *  \param c : the configuration to prepend
 */
void XYThS_Path::prepend(const mpc::XYThS_Config& c)
{
  vector <Config*>::iterator it = this->listCfg_.begin();
  this->listCfg_.insert(it,new XYThS_Config(c));
}

/**
 *  \brief concatenate a path to the current one
 *  \param p : the input path to concatenate
 */
void XYThS_Path::concat(const XYThS_Path& p)
{
  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}


/******************** Operators *******************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
mpc::XYThS_Config& XYThS_Path::operator[](int i)
{
  return *((XYThS_Config*)(this->listCfg_[i]));
}


                                    /**************************************************/
                                    /****************** Class CLPath ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CLPath::CLPath()
{
  this->type_ = CL_PATH;
}

/**
 *  \brief copy constructor
 *  \param t : the path to copy
 */
CLPath::CLPath(const CLPath& p)
{
  this->type_ = CL_PATH;

  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}

/**
 *  \brief destructor
 */
CLPath::~CLPath()
{
  this->clear();
}


/************** Getters and setters ***************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
const CLConfig& CLPath::at(int i) const
{
  return *((CLConfig*)(this->listCfg_.at(i)));
}

/**
 *  \brief getter : middle configuration of the path
 *  \return the middle configuration of the path
 */
const CLConfig& CLPath::middleCfg() const
{
  if(this->size() != 0)
    return this->at( ((int) (0.5*this->size())) );
  else
    throw string("Empty path");
}

/**
 *  \brief getter : first configuration of the path
 *  \return the first configuration of the path
 */
const CLConfig& CLPath::firstCfg() const
{
  if(this->size() != 0)
    return this->at(0);
  else
    throw string("Empty path");
}

/**
 *  \brief getter : last configuration of the path
 *  \return the last configuration of the path
 */
const CLConfig& CLPath::lastCfg() const
{
  if(this->size() != 0)
    return this->at(this->size()-1);
  else
    throw string("Empty path");
}


/******** Configurations list modification ********/

/**
 *  \brief append a new configuration to the path
 *  \param c : the configuration to append
 */
void CLPath::append(const CLConfig& c)
{
  CLConfig* newC = new CLConfig(c);
  this->listCfg_.push_back(newC);
}

/**
 *  \brief prepend a new configuration to the path
 *  \param c : the configuration to prepend
 */
void CLPath::prepend(const CLConfig& c)
{
  vector <Config*>::iterator it = this->listCfg_.begin();
  this->listCfg_.insert(it,new CLConfig(c));
}

/**
 *  \brief concatenate a path to the current one
 *  \param p : the input path to concatenate
 */
void CLPath::concat(const CLPath& p)
{
  int N = p.size();
  for(int i=0; i<N; i++)
    {
      this->append(p.at(i));
    }
}


/******************** Operators *******************/

/**
 *  \brief return the i-th element of the path
 *  \param i : index of the element to return
 *  \return the i-th element of the path
 */
CLConfig& CLPath::operator[](int i)
{
  return *((CLConfig*)(this->listCfg_[i]));
}


/**************** Utility functions ***************/

// /**
//  *  \brief diplay the path
//  */
// void CLPath::display() const
// {
// }


/************* Input/Output functions *************/

/**
 *  \brief load the path from an input file
 *  \param nomfic : name of the input file
 */
bool CLPath::load(char* nomfic)
{
  // delete the old path if exists
  if(this->size() != 0)
    {
      this->clear();
    }

  // create an input file
  ifstream file;
  
  // open the file
  file.open(nomfic);

  if(file.fail())
    {
      return false;
    }

  // read all configurations of the path
  while(!file.eof())
    {
      double x,y,th;
      file >> x;
      file >> y;
      file >> th;
      CLConfig cfg(x,y,th);
      this->append(cfg);
    }

  // close the file
  file.close();

  return true;
}
  
/**
 *  \brief save the path in an output file
 *  \param nomfic : name of the output file
 */
bool CLPath::save(char* nomfic) const 
{
  // create an output pile
  ofstream file;

  // open the output file
  file.open(nomfic);

  // write all configurations of the path
  for(int i=0; i< this->size(); i++)
    {
      file << this->at(i).x();
      file << " ";
      file << this->at(i).y();
      file << " ";
      file << this->at(i).th();
      file << endl;
    }

  // close the file
  file.close();
  
  return true;
}
