#ifndef PFPATHPLANNER_HPP
#define PFPATHPLANNER_HPP

#include "planner/inc/PathPlanner.hpp"
#include "common/inc/Config.hpp"

namespace mpc
{

  // Different search position types
  typedef enum {
    OBS  = -1, // obstacle position
    FRE  =  0, // free position
    VIS  =  1, // visited position
    REA  =  2  // reachable position
  } sp_t;

  class PFPathPlanner : public PathPlanner
  {
    
  public : 
    /********* Element for the path planner search *********/
    class PPElem 
    {
    private : 
      PMConfig cfg_;                    // current configuration
      int lg_;                          // length of the path already gone through
      double est_;                      // estimation of the rest of the path to go through
      PPElem* prev_;                    // pointer to the previous element
          
    public : 
      /************* Constructors and destructors ************/
      PPElem();                         // default constructor
      PPElem(const PPElem& e);          // copy constructor
      PPElem(const PMConfig& c,int l,double e,PPElem* prev); // init constructor
      PPElem(PPElem* from,const PMConfig& cur,const PMConfig& to,Map* pfm); // init constructor
      ~PPElem();                        // destructor
              
      /*********************** Getters ***********************/
      const PMConfig& getCfg() const;   // return the current configuration
      int getLg() const;                // return the current length
      double getEst() const;            // return the current estimation
      PPElem* getPrev() const;          // return the previous element
      double getHeuristic() const;      // return the current heuristic

      /****************** Utility functions ******************/
      void display(Map* pfMap);         // display the pp element
    };
    
  protected : 
    vector <PFPathPlanner::PPElem*> *visitedCfg_; // list of all configuration which have already benn visited
    vector <PFPathPlanner::PPElem*> *nextCfg_;    // list of all reachable configurations from the visited configs
    //Map* map;                                    // map on which is computed the path
    Map* pfMap_;                                 // potential field map

    sp_t** stateMap;                             // map of the state of each position during the path search
    PPElem*** elemMap;                           // map of all reachable or visited configurations

  public : 
    /************* Constructors and destructors ************/
    PFPathPlanner();                          // default constructor
    //PFPathPlanner(const PFPathPlanner& pp);   // copy constructor
    PFPathPlanner(Map* pfm);                  // init constructor
    PFPathPlanner(Map* map,Map* pfm);         // init constructor
    virtual ~PFPathPlanner();                 // destructor
      
    /********************** Accessors **********************/

    /**
     *  \brief return the initial configuration
     *  \return the initial configuration
     */
    inline virtual const PMConfig& initCfg() const { return *((PMConfig*)initCfg_); }

    /**
     *  \brief return the goal configuration
     *  \return the goal configuration
     */
    inline virtual const PMConfig& goalCfg() const { return *((PMConfig*)goalCfg_); }

    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual const PMPath* getPath() const { return ((PMPath*)(this->path_)); }

  protected : 
    /**
     *  \brief return the initial configuration
     *  \return the initial configuration
     */
    inline virtual PMConfig& initCfg() { return *((PMConfig*)initCfg_); }

    /**
     *  \brief return the goal configuration
     *  \return the goal configuration
     */
    inline virtual PMConfig& goalCfg() { return *((PMConfig*)goalCfg_); }

    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual PMPath* path() { return ((PMPath*)(this->path_)); }
  
    /******************* Path Computing ********************/

  public : 
    //virtual int computePath();                // compute the path from initCfg to goalCfg on map
    virtual int computePath(const PMConfig& c1,const PMConfig& c2); // compute the path from c1 to c2 on map


  protected : 
    void searchPathInit();                    // initialize the path search
    PPElem* searchPathBody();                 // search the path from initCfg to goalCfg (Body)
    // return a pointer on the last PPElem of the path
    bool searchPath();                        // search the path from initCfg to goalCfg
    void buildPath(PPElem* e);                // build the path when the goal cfg has been reached
      
    void addReachableElem(PPElem* e);         // add a pp element to the list of reachable elements
    void addListReachableElem(vector <PPElem*>* l); // add a list of pp element to the list 
    // of reachable elements
    PPElem* getBestPPElem();                  // get back the best pp element from the list of reachable
    // elements and add it to the list of visited configurations
    void smoothPath();                        // smooth the path built
  
     
    /****************** Path Computing (V2) ****************/

  protected : 
    int searchPathInitV2(const PMConfig& c1,const PMConfig& c2); // initialize the path search
    PPElem* searchPathBodyV2();               // search the path from initCfg to goalCfg (Body)
    // return a pointer on the last PPElem of the path
    bool searchPathV2(const PMConfig& c1,const PMConfig& c2); // search the path from initCfg to goalCfg

    void addReachableElemV2(PPElem* e);       // add a pp element to the list of reachable elements
    void addListReachableElemV2(vector <PPElem*>* l); // add a list of pp element to the list 
    // of reachable elements
    PPElem* getBestPPElemV2();                // get back the best pp element from the list of reachable
    // elements and add it to the list of visited configurations

    /*************** Reachability functions ****************/

  protected : 
    vector<PMConfig>* reachableCfg(const PMConfig& cfg);
    // search not still visited reachable configuration
    // from the config cfg
    vector <PFPathPlanner::PPElem*>* reachablePPElem(PFPathPlanner::PPElem* e);
    // search not still visited reachable pp elements
    // from the p element e
    bool stillVisited(const PMConfig& cfg);     // verify if the config cfg has already been visited
    bool stillReachable(const PMConfig& cfg);   // verify if the config cfg is already in the reachability list
      
    /*************** Reachability functions (V2) ***********/

  protected : 
    vector<PMConfig>* reachableCfgV2(const PMConfig& cfg);
    // search not still visited reachable configuration
    // from the config cfg
    vector <PFPathPlanner::PPElem*>* reachablePPElemV2(PFPathPlanner::PPElem* e);
    // search not still visited reachable pp elements
    // from the p element e  
    bool alreadyVisitedV2(const PMConfig& cfg);   // verify if the config cfg has already been visited
    bool alreadyReachableV2(const PMConfig& cfg); // verify if the config cfg is already in the reachability list

    /**************** Heuristics functions *****************/
  
  protected : 
    static double estimation(const PMConfig& c1,const PMConfig& c2,Map* pfm);
    // return an estimation to go from c1 to c2 on map pfm

    /****************** Utility functions ******************/

  public : 
    void displayVisitedCfg() const;           // display the visited cfg list
    void displayNextCfg() const;              // display the reachable cfg list
    
  };

};
#endif

