#ifndef MPC_PATHPLANNER_HPP
#define MPC_PATHPLANNER_HPP

#include "motion/inc/Path.hpp"
#include "environment/inc/Environment.hpp"
#include "system/inc/System.hpp"
#include "common/inc/Config.hpp"

namespace mpc
{
  /**
   * \enum res_pp_t
   * \brief Enumeration of the different path planner status
   *
   * PP_NOT_STARTED : path planner not started
   * ERR_CONFIGS    : wrong init or goal configurations
   * UNDEF_GPT_TYP  : undefined geom path type
   * ERR_PF_MAP     : wrong potential field map
   * ERR_STD_MAP    : wrong standard map
   * ERR_GOAL_POS   : the goal configuration is not available or free 
   * ERR_INIT_POS   : the init configuration is not available or free
   * NO_PATH        : no path has been found
   * PATH_CPTD      : the path has been computed
   */
  typedef enum {
    PP_NOT_STARTED = -7,
    ERR_CONFIGS    = -6,
    UNDEF_GPT_TYP  = -5,
    ERR_PF_MAP     = -4,
    ERR_STD_MAP    = -3,
    ERR_GOAL_POS   = -2,
    ERR_INIT_POS   = -1,
    NO_PATH        =  0,
    PATH_CPTD      =  1
  } res_pp_t;

                                    /**************************************************/
                                    /**************** Class PathPlanner ***************/
                                    /**************************************************/

  /** \class PathPlanner
   *  \brief Path planner
   */
  class PathPlanner
  {
  protected : 
    Path *path_;                             /**< the path which will be computed */
    Environment *env_;                        /**< the environment on which the path is computed */
    System* sys_;                             /**< the robotic system handled */
    Config* initCfg_;                         /**< the initial configuration */
    Config* goalCfg_;                         /**< the goal configuration */
    res_pp_t status_;                         /**< the path planner status */

    /************* Constructors and destructors ************/
  protected : 
    /**
     *  \brief default constructor
     */
    PathPlanner();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     */
    PathPlanner(System* sys,Environment* e);

    //   /**
    //    *  \brief init constructor
    //    *  \param sys : robotic system handled
    //    *  \param m : map of the environment in which the system evolve
    //    */
    //   PathPlanner(System* sys,Map* m);

    /**
     *  \brief destructor
     */
    virtual ~PathPlanner();
  
    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual const Path* getPath() const { return this->path_; }

    /**
     *  \brief return the current environment
     *  \return the current environment
     */
    inline const Environment* env() const { return this->env_; }

    /**
     *  \brief return the current map
     *  \return the current map
     */
    inline const Map* map() const { return this->env()->map(); }

    /**
     *  \brief return the status of the path planner
     *  \return the status of the path planner
     */
    inline res_pp_t status() const { return this->status_; }

    /**
     *  \brief return the initial configuration
     *  \return the initial configuration
     */
    virtual const Config& initCfg() const = 0;

    /**
     *  \brief return the goal configuration
     *  \return the goal configuration
     */
    virtual const Config& goalCfg() const = 0;

  protected : 
    /**
     *  \brief return the current environment
     *  \return the current environment
     */
    inline Environment* env() { return this->env_; }

    /**
     *  \brief return the current map
     *  \return the current map
     */
    inline Map* map() { return this->env()->map(); }

    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual Path* path() { return this->path_; }

    /**
     *  \brief return the initial configuration
     *  \return the initial configuration
     */
    virtual Config& initCfg() = 0;

    /**
     *  \brief return the goal configuration
     *  \return the goal configuration
     */
    virtual Config& goalCfg() = 0;

  /******************* Path Computing ********************/

//   /**
//    *  \brief compute a path between two configurations
//    *  \param cfg1 : the initial configuration
//    *  \param cfg2 : the final configuration
//    *  \return the path planning status
//    *       ERR_CONFIGS    : wrong init or goal configurations
//    *       UNDEF_GPT_TYP  : undefined geom path type
//    *       ERR_PF_MAP     : wrong potential field map
//    *       ERR_STD_MAP    : wrong standard map
//    *       ERR_GOAL_POS   : the goal configuration is not available or free 
//    *       ERR_INIT_POS   : the init configuration is not available or free
//    *       NO_PATH        : no path has been found
//    *       PATH_CPTD      : the path has been computed
//    */
//   virtual int computePath(const Config& cfg1,const Config& cfg2) = 0;
  };

};








// class PathPlanner
// {
//     protected : 
//         Path *path;                             // the path which will be computed
//         Map *map;                               // the map on which the path is computed
//         Environment *env_;                      // the environment on which the path is computed
//         ConfigR* initCfg;                       // the initial configuration
//         ConfigR* goalCfg;                       // the goal configuration
        
//     public : 
//         /************* Constructors and destructors ************/
//         PathPlanner();                                           // default constructor
// //         PathPlanner(const PathPlanner& pp);                      // copy constructor
//         PathPlanner(Environment* e,const Config& cfg1,const Config& cfg2); // init constructor
//         virtual ~PathPlanner();                                  // destructor
        
//         /********************** Accessors **********************/
//         Path* getPath() const;                                   // return the current path
//         Map* getMap() const;                                     // return the current map
//         void setMap(Map* m);                                     // set the current map
//         Environment* getEnv() const;                             // return the current environment
//         void setEnv(Environment* e);                             // set the current environment
//         ConfigR* getInitCfg() const;                             // return the initial configuration
//         void setInitCfg(ConfigR* c);                             // set the initial configuration
//         ConfigR* getGoalCfg() const;                             // return the goal configuration
//         void setGoalCfg(ConfigR* c);                             // set the goal configuration

//         /******************* Path Computing ********************/
//         virtual int computePath() = 0;                           // compute the path from initCfg to goalCfg on map
        
// };

#endif

