#ifndef MPC_SIMPLESTRAJPLANNER_HPP
#define MPC_SIMPLESTRAJPLANNER_HPP

#include "trajGen/inc/TrajGen.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

namespace mpc
{
                                    /**************************************************/
                                    /************ Class SimpleSTrajPlanner ************/
                                    /**************************************************/

  /** \class SimpleSTrajPlanner
   *  \brief Simple state trajectory planner (no collision checking)
   */
  class SimpleSTrajPlanner
  {
  protected : 
    System* sys_;                                          /**< the system used */

    TrajGen* trajgen_;                                     /**< trajectory generator tool used to find a control trajectory to the goal */

    double coefVMax_;                                      /**< Maximal velocity coefficient */
    double coefAMax_;                                      /**< Maximal acceleration coefficient */

    int debug_;                                            /**< debug level of the controller */
    PSTraj* straj_;                                        /**< state trajectory currently followed */

    boost::mutex mutexTraj_;                               /**< mutex used to prevent control and state trajectories access conflicts */


    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Only used by subclasses
     */
    SimpleSTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the robotic system handled
     *  \param coefVMax : maximal velocity coefficient
     *  \param coefAMax : maximal acceleration coefficient
     *  \param debug : the debugging level
     *  
     *  coefVMax \in [0;1]
     *  coefAMax \in [0;1]
     */
    SimpleSTrajPlanner(System* sys, double coefVMax=0.8, double coefAMax=0.8, int debug=0);

    /**
     *  \brief destructor
     */
    virtual ~SimpleSTrajPlanner();
  

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    virtual const System* sys() const = 0;


    /*************** Trajectory planning **************/

  protected : 
    /**
     *  \brief Initialize the standard parameters of the trajectory generator
     *  \param dbg : the trajectory generator debugging mode
     */
    virtual void initTrajGen(int dbg=0);

    /**
     *  \brief compute the time needed to execute the trajectory between s1 and s2
     *  \param p1 : the initial position
     *  \param v1 : the initial speed
     *  \param p2 : the goal position
     *  \param v2 : the goal speed
     */
    double computeTrajTime(const PositionXYTh& p1, double v1, const PositionXYTh& p2, double v2) const;


    /***************** Debug / Display ****************/

  public : 
    /**
     *  \brief return a copy of the planned state trajectory
     *  \return a copy of the planned state trajectory
     * 
     *  A new instance of a subclass of PSTraj is returned
     */
    virtual const PSTraj* getPSTraj() = 0;

    /**
     *  \brief return a sampled planned state trajectory
     *  \param sampling : the sampling of the sate trajectory
     *  \return a copy of a sampled planned state trajectory
     * 
     *  A new instance of a subclass of PSTraj is returned
     */
    virtual const STraj* getSTraj(int sampling=50) = 0;
  };

};

#endif // MPC_SIMPLESTRAJPLANNER_HPP
