#include "RoadMapTP.hpp"
using namespace mpc;

#include "Utils.hpp"
#include "CLTrajGen.hpp"
// #include "CLTrajGen2.hpp"
#include "WCTrajGen.hpp"
// #include "WCTrajGen2.hpp"

#define DT_STATE 0.6

                                    /**************************************************/
                                    /***************** Class RoadMapTP ****************/
                                    /**************************************************/

double RoadMapTP::maxSpeedCoef_ = 0.8;
double RoadMapTP::maxAccCoef_ = 0.8;

double RoadMapTP::minDObsCoef_ = 2.0;
double RoadMapTP::minLengthTraj_ = 100.0;
double RoadMapTP::maxLengthTraj_ = 300.0;

/************* Constructors and destructors ************/

/**
 *  \brief default constructor
 */
RoadMapTP::RoadMapTP() : TrajPlanner()
{
  this->rmpp_ = NULL;
  this->timeStep_ = 0.5;
  this->type_     = ROADMAP_TP;
}

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e : environment in which the system evolve
 */
RoadMapTP::RoadMapTP(System* sys,Environment* e) : TrajPlanner(sys,e)
{
  this->rmpp_ = NULL;
  this->timeStep_ = 0.5;
  this->type_     = ROADMAP_TP;
}

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e : environment in which the system evolve
 *  \param rm : roadmap defined over the environment
 */
RoadMapTP::RoadMapTP(System* sys,Environment* e,RoadMap* rm) : TrajPlanner(sys,e)
{
  this->rmpp_ = new RoadMapPP(sys,e,rm);
  this->timeStep_ = 0.5;
  this->type_     = ROADMAP_TP;
}

/**
 *  \brief destructor
 */
RoadMapTP::~RoadMapTP()
{
  if(rmpp_ != NULL) delete rmpp_;
}


/********************** Accessors **********************/

/**
 *  \brief set the maximum speed coefficient
 *  \param msc : the new maximum speed coefficient
 */
void RoadMapTP::setMSC(double msc)
{
  if(msc > 0 && msc <= 1)
    {
      RoadMapTP::maxSpeedCoef_ = msc;
    }
}

/**
 *  \brief set the maximum acceleration coefficient
 *  \param mac : the new maximum acceleration coefficient
 */
void RoadMapTP::setMAC(double mac)
{
  if(mac > 0 && mac <= 1)
    {
      RoadMapTP::maxAccCoef_ = mac;
    }
}


/******************** Initialization *******************/

/**
 *  \brief reinit the planner and the roadmap if needed
 */
void RoadMapTP::reinit()
{
  // reinit the roadmap path planner
  if(rmpp_ != NULL) rmpp_->reinit();
}


                                    /**************************************************/
                                    /**************** Class CLRoadMapTP ***************/
                                    /**************************************************/

/************* Constructors and destructors ************/

/**
 *  \brief default constructor
 */
CLRoadMapTP::CLRoadMapTP() : RoadMapTP()
{
}

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e : environment in which the system evolve
 */
CLRoadMapTP::CLRoadMapTP(CarLikeSys* sys,Environment* e) : RoadMapTP(sys,e)
{
}

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e : environment in which the system evolve
 *  \param rm : roadmap defined over the environment
 */
CLRoadMapTP::CLRoadMapTP(CarLikeSys* sys,Environment* e,RoadMap* rm) : RoadMapTP(sys,e,rm)
{
//   cout << "CLRoadMapTP::CLRoadMapTP" << endl;
//   cout << "sys : " << endl;
//   this->sys()->display();
//   cout << "rm : " << endl;
//   rm->display();
}

/**
 *  \brief destructor
 */
CLRoadMapTP::~CLRoadMapTP()
{
}


/********************** Accessors **********************/

/**
 *  \brief return the current trajectory
 *  \return the current trajectory
 */
const CLSTraj* CLRoadMapTP::getTraj() const
{
  if(this->status_ == TRAJ_CPTD) return ((CLSTraj*)(this->traj_));
  else return NULL;
}


/**************** Trajectory Computing *****************/

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been computed
 *            ERR_ROBOT     : wrong robot features
 *            ERR_ENV       : wrong environment
 *            ERR_ENV_MAP   : wrong static obstacles map
 *            ERR_ENV_OBS   : wrong dynamic obstacles list
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int CLRoadMapTP::computeTrajV2(const CLState& s1,const CLState& s2)
{
//   cout << "s1 : "; s1.display(); cout << endl;
//   cout << "s2 : "; s2.display(); cout << endl;

  // reinit the trajectory if needed
  if(this->traj_ != NULL) delete this->traj_;
  this->traj_ = new CLSTraj();

  PMConfig cfg1(s1.x(),s1.y());
  PMConfig cfg2(s2.x(),s2.y());

  int resPP = rmpp_->computePath(cfg1,cfg2);

  if(resPP == PATH_CPTD)
    {
      // retrieve the path computed
      PMPath path(*(rmpp_->getPath()));

      std::list<double>* vecMD = rmpp_->vecMinDist();

      // compute its length
      double pathLength = path.getLength();

      bool fin = false;

      CLStateT curST(s1.x(),s1.y(),s1.th(),s1.ph(),s1.s(),0);
      double curLength = 0;
      int indCurCfg = 0;

      while(!fin)
	{
	  cout << "//////////////////////////////////////////////////" << endl;
	  cout << "curST : "; curST.display(); cout << endl;

	  CLStateT targetST;
	  int indTargetCfg;
	  double targetLength;

	  // select the target state to reach
	  this->selectTargetST(path,pathLength,vecMD,s2,
			       curST,indCurCfg,curLength,
			       targetST,indTargetCfg,targetLength);

	  cout << "targetST : "; targetST.display(); cout << endl;
// 	  this->traj()->append(targetST);
	  
	  // try to generate a trajectory between curST and targetST
	  CLStateT sR;
	  this->genTraj(curST,targetST,sR,*(this->traj()));

	  cout << "sR : "; sR.display(); cout << endl;
	  
	  // check if the target state has been reached
	  if(indTargetCfg == path.size()-1)
	    {
	      fin = true;
	    }
	  else
	    {
	      // replace the current state by the target state
	      curST = sR;
	      //curST = targetST;
	      indCurCfg = indTargetCfg;
	      curLength = targetLength;
	    }
	  cout << "//////////////////////////////////////////////////" << endl;
	}

      this->status_ = TRAJ_CPTD;
      return TRAJ_CPTD;
    }
  else
    {
      return NO_PATH_LDD;
    }
}

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been computed
 *            ERR_ROBOT     : wrong robot features
 *            ERR_ENV       : wrong environment
 *            ERR_ENV_MAP   : wrong static obstacles map
 *            ERR_ENV_OBS   : wrong dynamic obstacles list
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int CLRoadMapTP::computeTraj(const CLState& s1,const CLState& s2)
{
  // reinit the trajectory if needed
  if(this->traj_ != NULL) delete this->traj_;
  this->traj_ = new CLSTraj();

  PMConfig cfg1(s1.x(),s1.y());
  PMConfig cfg2(s2.x(),s2.y());

  int resPP = rmpp_->computePath(cfg1,cfg2);

  if(resPP == PATH_CPTD)
    {
      // retrieve the path computed
      PMPath path(*(rmpp_->getPath()));

      std::list<double>* vecMD = rmpp_->vecMinDist();

      // compute its length
      double pathLength = path.getLength();

      bool fin = false;

      CLStateT curST(s1.x(),s1.y(),s1.th(),s1.ph(),s1.s(),0);
      double curLength = 0;
      int indCurCfg = 0;

      // compute the waypoints of the trajectory
      CLSTraj WPtraj;
      while(!fin)
	{
	  CLStateT targetST;
	  int indTargetCfg;
	  double targetLength;

	  // select the target state to reach
	  this->selectTargetST(path,pathLength,vecMD,s2,
			       curST,indCurCfg,curLength,
			       targetST,indTargetCfg,targetLength);
	  
	  // add the target state found to the list of waypoints
	  WPtraj.append(targetST);

	  // check if the target state has been reached
	  if(indTargetCfg == path.size()-1)
	    {
	      fin = true;
	    }
	  else
	    {
	      // replace the current state by the target state
	      curST = targetST;
	      indCurCfg = indTargetCfg;
	      curLength = targetLength;
	    }
	}

      // generate a trajectory between each waypoint
      CLStateT curST2(s1.x(),s1.y(),s1.th(),s1.ph(),s1.s(),0);
      for(int i=0; i<WPtraj.size(); i++)
	{
	  // modify the heading of the target state depending of the current state and the next waypoint
	  CLStateT targetST2 = WPtraj.at(i);
	  if(i == WPtraj.size()-1)
	    {
	      CLStateT prevCfgTgt = curST2;
	      double dxPT = targetST2.x() - prevCfgTgt.x();
	      double dyPT = targetST2.y() - prevCfgTgt.y();
	      double dirVecPT = atan2(dyPT,dxPT);
	      targetST2.th() = normalizedAngle(dirVecPT);
	    }
	  else
	    {
	      CLStateT prevCfgTgt = curST2;
	      CLStateT nextCfgTgt = WPtraj.at(i+1);
	      
	      double dxPT = targetST2.x() - prevCfgTgt.x();
	      double dyPT = targetST2.y() - prevCfgTgt.y();
	      double dirVecPT = atan2(dyPT,dxPT);
	      double dxTN = nextCfgTgt.x() - targetST2.x();
	      double dyTN = nextCfgTgt.y() - targetST2.y();
	      double dirVecTN = atan2(dyTN,dxTN);
	      
	      double diffAngle = computeMinDiffAngle(normalizedAngle(dirVecTN),normalizedAngle(dirVecPT));
	      targetST2.th() = normalizedAngle(dirVecPT + 0.5*diffAngle);
	    }

	  curST2.ph() = 0;

	  // try to generate a trajectory between curST and targetST
	  CLStateT sR;
	  this->genTraj(curST2,targetST2,sR,*(this->traj()));

// 	  cout << "curST2 : "; curST2.display(); cout << "  targetST2 : "; targetST2.display(); cout << "  sR : "; sR.display(); cout << endl; 

	  curST2 = sR;
	}

      this->status_ = TRAJ_CPTD;
      return TRAJ_CPTD;
    }
  else
    {
      return NO_PATH_LDD;
    }
}

/**
 *  \brief select the target state-time to reach
 *  \param path : the path computed on the roadmap
 *  \param pathLength : the total length of the path
 *  \param vecMD : list of minimal distances to obstacles for all configurations of the path computed
 *  \param s2 : the goal state of the trajectory
 *  \param curST : the current state-time
 *  \param indCurCfg : index of the current configuration of the path we suppose to have already reached
 *  \param curLength : length of the path up to the current configuration
 *  \param targetST : the resulting target state-time
 *  \param indTargetCfg : index of the target configuration of the path
 *  \param targetLength : length of the path up to the target configuration
 */
void CLRoadMapTP::selectTargetST(const PMPath& path,double pathLength,
				 std::list<double>* vecMD,const CLState& s2,
				 const CLStateT& curST,int indCurCfg,double curLength,
				 CLStateT& targetST,int& indTargetCfg,double& targetLength)
{
  // get back the minimal distance to obstacles of the current configuration
  std::list<double>::iterator it = vecMD->begin();
  for(int i=0; i<indCurCfg; i++) it++;
  double curMinDist = *it;

  // compute the distance we would like to travel
  double distTrav = curMinDist * minDObsCoef_;
  if(distTrav < RoadMapTP::minLengthTraj_) distTrav = RoadMapTP::minLengthTraj_;
  else if(distTrav > RoadMapTP::maxLengthTraj_) distTrav = RoadMapTP::maxLengthTraj_;

//   cout << "CLRoadMapTP::selectTargetST 2" << endl;

  // search the current configuration of the path already reached
  PMConfig curCfg = path.at(indCurCfg);

//   cout << "CLRoadMapTP::selectTargetST 3" << endl;

  // search the target configuration of the path to reach
  indTargetCfg = indCurCfg + 1;
  bool found = false;
  double distCurTgtCfg = 0;  
  PMConfig targetCfg;
  PMConfig prevCfg = curCfg;

//   cout << "CLRoadMapTP::selectTargetST 4" << endl;

  while(!found && indTargetCfg != path.size()-1)
    {
      // get back the current target configuration
      targetCfg = path.at(indTargetCfg);

      // compute the distance between the target configuration and its previous configuration
      double dx1 = targetCfg.x() - prevCfg.x();
      double dy1 = targetCfg.y() - prevCfg.y();
      double distPrevTgtCfg = sqrt(dx1*dx1 + dy1*dy1);

      distCurTgtCfg += distPrevTgtCfg;
      if(distCurTgtCfg >= distTrav) found = true;
      else 
	{
	  // compute the distance between the current state and the current target configuration
	  double dx = targetCfg.x() - curST.x();
	  double dy = targetCfg.y() - curST.y();
	  double dist = sqrt(dx*dx + dy*dy);
	  
	  if(dist >= distTrav) found = true;
	  else 
	    {
	      indTargetCfg++;
	      prevCfg = targetCfg;
	    }
	}
    }

//   cout << "CLRoadMapTP::selectTargetST 5" << endl;

//   // check if the target configuration is not too close to the last configuration of the path
  targetLength = curLength + distCurTgtCfg;
  if(pathLength - targetLength < RoadMapTP::minLengthTraj_)
    {
      indTargetCfg = path.size()-1;
      targetLength = pathLength;
      targetCfg = path.at(path.size()-1);
    }

//   cout << "CLRoadMapTP::selectTargetST 6" << endl;

  /// compute the target state-time to reach

  // compute the heading of the last configuration
  double targetHead;
  if(indTargetCfg == path.size()-1)
    {
      PMConfig prevCfgTgt = path.at(indTargetCfg-1);
      double dxPT = targetCfg.x() - prevCfgTgt.x();
      double dyPT = targetCfg.y() - prevCfgTgt.y();
      double dirVecPT = atan2(dyPT,dxPT);

      targetHead = normalizedAngle(dirVecPT);
      cout << "targetHead : " << targetHead << endl;

//       double dx2 = targetCfg.x() - curST.x();
//       double dy2 = targetCfg.y() - curST.y();
//       double dirVec = atan2(dy2,dx2);
//       double diffAngle = computeMinDiffAngle(normalizedAngle(dirVec),normalizedAngle(curST.th()));
//       targetHead = normalizedAngle(curST.th() + 2.0*diffAngle);
    }
  else
    {
      PMConfig prevCfgTgt = path.at(indTargetCfg-1);
      PMConfig nextCfgTgt = path.at(indTargetCfg+1);

      double dxPT = targetCfg.x() - prevCfgTgt.x();
      double dyPT = targetCfg.y() - prevCfgTgt.y();
      double dirVecPT = atan2(dyPT,dxPT);
      double dxTN = nextCfgTgt.x() - targetCfg.x();
      double dyTN = nextCfgTgt.y() - targetCfg.y();
      double dirVecTN = atan2(dyTN,dxTN);

      double diffAngle = computeMinDiffAngle(normalizedAngle(dirVecTN),normalizedAngle(dirVecPT));
      targetHead = normalizedAngle(dirVecPT + 0.5*diffAngle);

      cout << "dirVecPT : " << dirVecPT << "  dirVecTN : " << dirVecTN << "  targetHead : " << targetHead << endl;
    }

//   cout << "CLRoadMapTP::selectTargetST 7" << endl;

  // compute the minimal distance to stop the system at the current speed
  double distStop = 0;
  if(curST.s() > s2.s())
    {
      double dtV = (curST.s() - s2.s()) / (maxAccCoef_*sys_->getAMax());
      distStop = curST.s()*dtV - 0.5*maxAccCoef_*sys_->getAMax()*dtV*dtV;
    }

  // compute the remaining distance of the path to travel
  double remainDist = pathLength - curLength;

  // check if we need to decelerate now to be able to reach the final speed
  bool dec = false;
  if(distStop > remainDist) dec = true;

  // compute the estimated distance to travel between the current state and the target configuration
  double distTravel = (distCurTgtCfg)*1.2;

//   cout << "CLRoadMapTP::selectTargetST 8" << endl;

  // compute the final speed to reach
  double targetSpd;
  double DT;
  if(dec)
    {
      double coefSpd = (distCurTgtCfg) / (pathLength - curLength);
      targetSpd = coefSpd*s2.s() + (1.0-coefSpd)*curST.s();
      cout << "deceleration control; coefSpd : " << coefSpd << endl;
    }
  else
    {
      double dtV = ((maxSpeedCoef_*sys_->getVMax()) - curST.s()) / (maxAccCoef_*sys_->getAMax());
      double distAccMaxSpd = curST.s()*dtV + 0.5*maxAccCoef_*sys_->getAMax()*dtV*dtV;
      if(distAccMaxSpd < distTravel)
	{
	  cout << "acceleration (can reach the max speed)" << endl;
	  targetSpd = maxSpeedCoef_*sys_->getVMax();
	}
      else
	{
	  cout << "acceleration (cannot reach the max speed)" << endl;
	  double coefSpd = distTravel / distAccMaxSpd;
	  targetSpd = curST.s() + coefSpd*(maxSpeedCoef_*sys_->getVMax() - curST.s());
	}
    }

  cout << "targetSpd : " << targetSpd << endl;

  double midSpd = 0.5*(curST.s()+targetSpd);
  DT = distTravel / midSpd;

  cout << "midSpd : " << midSpd << endl;
  cout << "distTravel : " << distTravel << endl;
  cout << "DT : " << DT << endl;

//   if(DT > 5.0) 
//     {
//       cout << "Attention!!! -> DT : " << DT << endl;
//       DT = 5.0;
//     }

  if(indTargetCfg == path.size()-1)
    {
      targetST.x()  = s2.x();
      targetST.y()  = s2.y();
      //targetST.th() = s2.th();
      targetST.th() = targetHead;
      targetST.s()  = s2.s();
      targetST.ph() = s2.ph();
      targetST.t()  = curST.t() + DT;
    }
  else
    {
      targetST.x()  = targetCfg.x();
      targetST.y()  = targetCfg.y();
      targetST.th() = targetHead;
      targetST.s()  = targetSpd;
      targetST.ph() = 0;
      targetST.t()  = curST.t() + DT;
    }

//   cout << "CLRoadMapTP::selectTargetST 10" << endl;
}

/**
 *  \brief try to generate a trajectory between two state-times
 *  \param s1 : the initial state-time
 *  \param s2 : the goal state-time
 *  \param sR : the final state-time reached
 *  \param trajRes : the resulting discrete trajectory
 */
void CLRoadMapTP::genTraj(const CLStateT& s1,const CLStateT& s2,CLStateT& sR,CLSTraj& traj)
{
  CLTrajGen trajgen(sys());
  trajgen.fixFinalXY(true);
  trajgen.fixFinalTH(true);
  trajgen.fixFinalV(false);
  trajgen.fixFinalPH(false);

  trajgen.setDBG(0);
  CLTrajGen::setMaxNbStep(30);

  double dt = s2.t() - s1.t();

//   cout << "//////////////////////////////////////////////////////////"  << endl;
//   cout << "s1 : "; s1.display(); cout << "  s2 : "; s2.display(); cout << endl;
//   cout << "system : " << endl; sys()->display();

  int res = trajgen.ComputeTrajectory(s1,s2,(double**)NULL);

//   cout << "//////////////////////////////////////////////////////////"  << endl;

  double x = trajgen.funcX(dt);
  double y = trajgen.funcY(dt); 
  double th = trajgen.funcTheta(dt);
  double v = trajgen.funcVel(dt);
  double om = trajgen.funcAVel(dt);
  double ph;
  if(v != 0.0) ph = atan(sys()->getLength() * om / v);
  else ph = 0.0;

  sR = CLStateT(x,y,th,ph,v,s1.t() + dt);

  int nbT = ((int)(dt / RoadMapTP::timeStep_)) + 1;
  double realTS = dt / nbT;

  for(int i=0; i<=nbT; i++)
    {
      double dt1 = i*realTS;

      double x1 = trajgen.funcX(dt1);
      double y1 = trajgen.funcY(dt1); 
      double th1 = trajgen.funcTheta(dt1);
      double v1 = trajgen.funcVel(dt1);
      double om1 = trajgen.funcAVel(dt1);
      double ph1;
      if(v1 != 0.0) ph1 = atan(sys()->getLength() * om1 / v1);
      else ph1 = 0.0;

//       cout << "x1 : " << x1 << "  y1 : " << y1 << "  th1 : " << th1 << "  v1 : " << v1 << "  om1 : " << om1 << "  ph1 : " << ph1 << "  dt1 : " << dt1 << endl;

      CLStateT newST(x1,y1,th1,ph1,v1,s1.t() + dt1);

      traj.append(newST);
    }
}


                                    /**************************************************/
                                    /**************** Class WCRoadMapTP ***************/
                                    /**************************************************/

/**
 *  \brief default constructor
 */
WCRoadMapTP::WCRoadMapTP() : RoadMapTP()
{
}

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e : environment in which the system evolve
 */
WCRoadMapTP::WCRoadMapTP(WheelChairSys* sys,Environment* e) : RoadMapTP(sys,e)
{
}

// /**
//  *  \brief init constructor
//  *  \param sys : robotic system handled
//  *  \param rm : roadmap defined over the environment
//  */
// WCRoadMapTP::WCRoadMapTP(WheelChairSys* sys,RoadMap* rm) : RoadMapTP(sys,rm)
// {
// }

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e : environment in which the system evolve
 *  \param rm : roadmap defined over the environment
 */
WCRoadMapTP::WCRoadMapTP(WheelChairSys* sys,Environment* e,RoadMap* rm) : RoadMapTP(sys,e,rm)
{
  this->env_ = e;
}


/**
 *  \brief destructor
 */
WCRoadMapTP::~WCRoadMapTP()
{
}


/********************** Accessors **********************/

/**
 *  \brief return the current trajectory
 *  \return the current trajectory
 */
const WCSTraj* WCRoadMapTP::getTraj() const
{
  if(this->status_ == TRAJ_CPTD) return ((WCSTraj*)(this->traj_));
  else return NULL;
}


/**************** Trajectory Computing *****************/

/**
 *  \brief compute the trajectory from s1 to s2 (old version without collision checking)
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been loaded
 *            ERR_ROBOT     : wrong robot features
 *            ERR_ENV       : wrong environment
 *            ERR_ENV_MAP   : wrong static obstacles map
 *            ERR_ENV_OBS   : wrong dynamic obstacles list
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int WCRoadMapTP::computeTrajV3(const WCState& s1,const WCState& s2)
{
  // reinit the trajectory if needed
  if(this->traj_ != NULL) delete this->traj_;
  this->traj_ = new WCSTraj();

  PMConfig cfg1(s1.x(),s1.y());
  PMConfig cfg2(s2.x(),s2.y());

  int resPP = rmpp_->computePath(cfg1,cfg2);

  if(resPP == PATH_CPTD)
    {
      // retrieve the path computed
      PMPath path(*(rmpp_->getPath()));

      std::list<double>* vecMD = rmpp_->vecMinDist();

      // compute its length
      double pathLength = path.getLength();

      bool fin = false;

      WCStateT curST(s1.x(),s1.y(),s1.th(),s1.s(),s1.om(),0);
      double curLength = 0;
      int indCurCfg = 0;

      // compute the waypoints of the trajectory
      WCSTraj WPtraj;
      while(!fin)
	{
	  WCStateT targetST;
	  int indTargetCfg;
	  double targetLength;

	  // select the target state to reach
	  this->selectTargetST(path,pathLength,vecMD,s2,
			       curST,indCurCfg,curLength,
			       targetST,indTargetCfg,targetLength);
 	  this->computeTargetST(path,curST,indCurCfg,indTargetCfg,s2,targetST);
	  
	  // add the target state found to the list of waypoints
	  WPtraj.append(targetST);

	  // check if the target state has been reached
	  if(indTargetCfg == path.size()-1)
	    {
	      fin = true;
	    }
	  else
	    {
	      // replace the current state by the target state
	      curST = targetST;
	      indCurCfg = indTargetCfg;
	      curLength = targetLength;
	    }
	}
      
      cout << "list of waypoints : " << endl;
      WPtraj.display();

      // generate a trajectory between each waypoint
      WCStateT curST2(s1.x(),s1.y(),s1.th(),s1.s(),s1.om(),0);
      for(int i=0; i<WPtraj.size(); i++)
	{
	  // modify the heading of the target state depending of the current state and the next waypoint
	  WCStateT targetST2 = WPtraj.at(i);
	  if(i == WPtraj.size()-1)
	    {
	      WCStateT prevCfgTgt = curST2;
	      double dxPT = targetST2.x() - prevCfgTgt.x();
	      double dyPT = targetST2.y() - prevCfgTgt.y();
	      double dirVecPT = atan2(dyPT,dxPT);
	      targetST2.th() = normalizedAngle(dirVecPT);
	    }
	  else
	    {
	      WCStateT prevCfgTgt = curST2;
	      WCStateT nextCfgTgt = WPtraj.at(i+1);
	      
	      double dxPT = targetST2.x() - prevCfgTgt.x();
	      double dyPT = targetST2.y() - prevCfgTgt.y();
	      double dirVecPT = atan2(dyPT,dxPT);
	      double dxTN = nextCfgTgt.x() - targetST2.x();
	      double dyTN = nextCfgTgt.y() - targetST2.y();
	      double dirVecTN = atan2(dyTN,dxTN);
	      
	      double diffAngle = computeMinDiffAngle(normalizedAngle(dirVecTN),normalizedAngle(dirVecPT));
	      targetST2.th() = normalizedAngle(dirVecPT + 0.5*diffAngle);
	    }

 	  curST2.om() = 0;
	  //this->traj()->append(targetST2);

	  cout << "curST2 : "; curST2.display(); cout << "targetST2 : "; targetST2.display(); cout << endl; 

	  // try to generate a trajectory between curST and targetST
	  WCStateT sR;
	  this->genTraj(curST2,targetST2,sR,*(this->traj()));

	  //curST2 = targetST2;
	  curST2 = sR;
	}

      this->status_ = TRAJ_CPTD;
      return TRAJ_CPTD;
    }
  else
    {
      return NO_PATH_LDD;
    }
}

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been loaded
 *            ERR_ROBOT     : wrong robot features
 *            ERR_ENV       : wrong environment
 *            ERR_ENV_MAP   : wrong static obstacles map
 *            ERR_ENV_OBS   : wrong dynamic obstacles list
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int WCRoadMapTP::computeTrajV2(const WCState& s1,const WCState& s2)
{
  // reinit the trajectory if needed
  if(this->traj_ != NULL) delete this->traj_;
  this->traj_ = new WCSTraj();

  PMConfig cfg1(s1.x(),s1.y());
  PMConfig cfg2(s2.x(),s2.y());

  int resPP = rmpp_->computePath(cfg1,cfg2);

  if(resPP == PATH_CPTD)
    {
      // retrieve the path computed
      PMPath path(*(rmpp_->getPath()));

      std::list<double>* vecMD = rmpp_->vecMinDist();

      // compute its length
      double pathLength = path.getLength();

      bool fin = false;         // true if the last waypoint has been reached
      bool noSol = false;       // true if no collision-free solution has been found

      WCStateT curST(s1.x(),s1.y(),s1.th(),s1.s(),s1.om(),0);
      double curLength = 0;
      int indCurCfg = 0;

      // compute the waypoints of the trajectory
      WCSTraj WPtraj;
      while(!fin && !noSol)
	{
	  WCStateT targetST;
	  int indTargetCfg;
	  double targetLength;

	  // select the target state to reach
	  this->selectTargetST(path,pathLength,vecMD,s2,
			       curST,indCurCfg,curLength,
			       targetST,indTargetCfg,targetLength);
	  
// 	  // add the target state found to the list of waypoints
// 	  WPtraj.append(targetST);

	  // try to compute a trajectory between the current state and the target state found
	  bool trajFound = false;
	  WCStateT sR;

	  while(!trajFound && indTargetCfg > indCurCfg)
	    {
	      // compute the target state from the index of position selected
	      this->computeTargetST(path,curST,indCurCfg,indTargetCfg,s2,targetST);

	      WCStateT curST2(curST.x(),curST.y(),curST.th(),curST.s(),curST.om(),curST.t());
	      curST2.om() = 0;
	      
	      cout << "indCurCfg : " << indCurCfg << "  indTargetCfg : " << indTargetCfg;
	      cout << "  curST2 : "; curST2.display(); cout << "  targetST : "; targetST.display(); 

	      // try to generate a trajectory between curST and targetST
	      WCSTraj curTraj;
	      this->genTraj(curST2,targetST,sR,curTraj);

	      cout << "  sR : "; sR.display();
	      cout << endl;
	      
	      // check if the trajectory computed is free of collisions
	      if(mpc::staticCollision(curTraj,env_))
		{
		  // a collision occurs, try with the previous target node
		  indTargetCfg--;
		  //cout << "collision found!" << endl;
		}
	      else 
		{
		  trajFound = true;
		  // append the trajectory computed to the total trajectory
		  this->traj()->concat(curTraj);
		  //cout << "no collision" << endl;
		}
	    }

	  // check if a trajectory has been found
	  if(!trajFound)
	    {
	      return NO_TRAJ;
	    }

	  // check if the target state has been reached
	  if(indTargetCfg == path.size()-1)
	    {
	      fin = true;
	      this->status_ = TRAJ_CPTD;
	      return TRAJ_CPTD;
	    }
	  else
	    {
	      // replace the current state by the target state
	      //curST = targetST;
	      curST = sR;
	      indCurCfg = indTargetCfg;
	      curLength = path.getLengthUpTo(indTargetCfg);
	    }
	}
    }
  else
    {
      return NO_PATH_LDD;
    }
}

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been loaded
 *            ERR_ROBOT     : wrong robot features
 *            ERR_ENV       : wrong environment
 *            ERR_ENV_MAP   : wrong static obstacles map
 *            ERR_ENV_OBS   : wrong dynamic obstacles list
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int WCRoadMapTP::computeTraj(const WCState& s1,const WCState& s2)
{
  // reinit the trajectory if needed
//   if(this->traj_ != NULL) delete this->traj_;
  this->traj_ = new WCSTraj();

  PMConfig cfg1(s1.x(),s1.y());
  PMConfig cfg2(s2.x(),s2.y());

  int resPP = rmpp_->computePath(cfg1,cfg2);

  if(resPP == PATH_CPTD)
    {
      // retrieve the path computed
      PMPath path(*(rmpp_->getPath()));

      std::list<double>* vecMD = rmpp_->vecMinDist();

      // compute its length
      double pathLength = path.getLength();

      bool fin = false;         // true if the last waypoint has been reached
      bool noSol = false;       // true if no collision-free solution has been found

      double dtDeb = 0;
      double dtFin = 0;

      WCStateT curST(s1.x(),s1.y(),s1.th(),s1.s(),s1.om(),0);

      int N = path.size();
      for(int i=0; i<N; i++)
	{
	  // get back the current position of the path
	  PMConfig curP = path.at(i);

	  WCStateT curS;
	  if(i == 0)
	    {
	      // compute the initial state
	      curS = WCStateT(s1.x(),s1.y(),s1.th(),s1.s(),s1.om(),0);
	      this->traj()->append(curS);

	      // compute the required heading to go to the next position
	      PMConfig next = path.at(1);
	      double nextTH = atan2(next.y()-s1.y(),next.x()-s1.x());

	      // compute the rotation angle to head in the direction of the next position
	      double minDiffA = computeMinDiffAngle(nextTH,s1.th());

	      // compute the number of additional states that must be added
	      int nbAddS = ((int)(fabs(minDiffA) / (0.25*M_PI))) + 1;
	      double deltaAngle = minDiffA / nbAddS;

	      // compute each state to add
	      for(int j=0; j<nbAddS; j++)
		{
		  WCStateT addS(s1.x(),s1.y(),s1.th() + (j+1)*deltaAngle,s1.s(),s1.om(),(j+1)*DT_STATE);
		  this->traj()->append(addS);
		}

	      // compute the additional time required to reach to good heading
	      dtDeb = nbAddS*DT_STATE;
	    }
	  else if(i == N-1)
	    {
	      // get back the previous and next positions
	      PMConfig prev = path.at(i-1);	      
	      double th = atan2(s2.y()-prev.y(),s2.x()-prev.x());

	      // compute the final state-time
	      curS = WCStateT(s2.x(),s2.y(),th,s2.s(),s2.om(),dtDeb + N*DT_STATE);
	      this->traj()->append(curS);
	    }
	  else
	    {
	      // get back the previous and next positions
	      PMConfig prev = path.at(i-1);
	      PMConfig next = path.at(i+1);

	      double x = curP.x();
	      double y = curP.y();
	      double th = atan2(next.y()-prev.y(),next.x()-prev.x());
	      curS = WCStateT(x,y,th,0,0,dtDeb + i*DT_STATE);
	      this->traj()->append(curS);
	    }
	}

      // adjust the time and speed of each state-time of the trajectory
      this->adjustTimeSpeed(s1,s2,pathLength);
      try
	{
	  // resample the trajectory
	  this->resampleTraj();
	}
      catch (string str)
	{
	  cout << "Exception: " << str << endl;
	  exit(1);
	}

      this->status_ = TRAJ_CPTD;
      return TRAJ_CPTD;      
    }
  else
    {
      return NO_PATH_LDD;
    }
  
}

/**
 *  \brief select the target state-time to reach
 *  \param path : the path computed on the roadmap
 *  \param pathLength : the total length of the path
 *  \param vecMD : list of minimal distances to obstacles for all configurations of the path computed
 *  \param s2 : the goal state of the trajectory
 *  \param curST : the current state-time
 *  \param indCurCfg : index of the current configuration of the path we suppose to have already reached
 *  \param curLength : length of the path up to the current configuration
 *  \param targetST : the resulting target state-time
 *  \param indTargetCfg : index of the target configuration of the path
 *  \param targetLength : length of the path up to the target configuration
 */
void WCRoadMapTP::selectTargetST(const PMPath& path,double pathLength,
		    std::list<double>* vecMD,const WCState& s2,
		    const WCStateT& curST,int indCurCfg,double curLength,
		    WCStateT& targetST,int& indTargetCfg,double& targetLength)
{
  // get back the minimal distance to obstacles of the current configuration
  std::list<double>::iterator it = vecMD->begin();
  for(int i=0; i<indCurCfg; i++) it++;
  double curMinDist = *it;

  // compute the distance we would like to travel
  double distTrav = curMinDist * minDObsCoef_;
  if(distTrav < RoadMapTP::minLengthTraj_) distTrav = RoadMapTP::minLengthTraj_;
  else if(distTrav > RoadMapTP::maxLengthTraj_) distTrav = RoadMapTP::maxLengthTraj_;

  // search the current configuration of the path already reached
  PMConfig curCfg = path.at(indCurCfg);

  // search the target configuration of the path to reach
  indTargetCfg = indCurCfg + 1;
  bool found = false;
  double distCurTgtCfg = 0;  
  PMConfig targetCfg;
  PMConfig prevCfg = curCfg;

  while(!found && indTargetCfg != path.size()-1)
    {
      // get back the current target configuration
      targetCfg = path.at(indTargetCfg);

      // compute the distance between the target configuration and its previous configuration
      double dx1 = targetCfg.x() - prevCfg.x();
      double dy1 = targetCfg.y() - prevCfg.y();
      double distPrevTgtCfg = sqrt(dx1*dx1 + dy1*dy1);

      distCurTgtCfg += distPrevTgtCfg;
      if(distCurTgtCfg >= distTrav) found = true;
      else 
	{
	  // compute the distance between the current state and the current target configuration
	  double dx = targetCfg.x() - curST.x();
	  double dy = targetCfg.y() - curST.y();
	  double dist = sqrt(dx*dx + dy*dy);
	  
	  if(dist >= distTrav) found = true;
	  else 
	    {
	      indTargetCfg++;
	      prevCfg = targetCfg;
	    }
	}
    }

//   // check if the target configuration is not too close to the last configuration of the path
  targetLength = curLength + distCurTgtCfg;
  if(pathLength - targetLength < RoadMapTP::minLengthTraj_)
    {
      indTargetCfg = path.size()-1;
      targetLength = pathLength;
      targetCfg = path.at(path.size()-1);
    }

  /// compute the target state-time to reach

//   // compute the heading of the last configuration
//   double targetHead;
//   if(indTargetCfg == path.size()-1)
//     {
//       PMConfig prevCfgTgt = path.at(indTargetCfg-1);
//       double dxPT = targetCfg.x() - prevCfgTgt.x();
//       double dyPT = targetCfg.y() - prevCfgTgt.y();
//       double dirVecPT = atan2(dyPT,dxPT);

//       targetHead = normalizedAngle(dirVecPT);
//       cout << "targetHead : " << targetHead << endl;
//     }
//   else
//     {
//       PMConfig prevCfgTgt = path.at(indTargetCfg-1);
//       PMConfig nextCfgTgt = path.at(indTargetCfg+1);

//       double dxPT = targetCfg.x() - prevCfgTgt.x();
//       double dyPT = targetCfg.y() - prevCfgTgt.y();
//       double dirVecPT = atan2(dyPT,dxPT);
//       double dxTN = nextCfgTgt.x() - targetCfg.x();
//       double dyTN = nextCfgTgt.y() - targetCfg.y();
//       double dirVecTN = atan2(dyTN,dxTN);

//       double diffAngle = computeMinDiffAngle(normalizedAngle(dirVecTN),normalizedAngle(dirVecPT));
//       targetHead = normalizedAngle(dirVecPT + 0.5*diffAngle);
//     }

//   // compute the minimal distance to stop the system at the current speed
//   double distStop = 0;
//   if(curST.s() > s2.s())
//     {
//       double dtV = (curST.s() - s2.s()) / (maxAccCoef_*sys_->getAMax());
//       distStop = curST.s()*dtV - 0.5*maxAccCoef_*sys_->getAMax()*dtV*dtV;
//     }

//   // compute the remaining distance of the path to travel
//   double remainDist = pathLength - curLength;

//   // check if we need to decelerate now to be able to reach the final speed
//   bool dec = false;
//   if(distStop > remainDist) dec = true;

//   // compute the estimated distance to travel between the current state and the target configuration
//   double distTravel = (distCurTgtCfg)*1.2;

//   // compute the final speed to reach
//   double targetSpd;
//   double DT;
//   if(dec)
//     {
//       double coefSpd = (distCurTgtCfg) / (pathLength - curLength);
//       targetSpd = coefSpd*s2.s() + (1.0-coefSpd)*curST.s();
//     }
//   else
//     {
//       double dtV = ((maxSpeedCoef_*sys_->getVMax()) - curST.s()) / (maxAccCoef_*sys_->getAMax());
//       double distAccMaxSpd = curST.s()*dtV + 0.5*maxAccCoef_*sys_->getAMax()*dtV*dtV;
//       if(distAccMaxSpd < distTravel)
// 	{
//  	  targetSpd = maxSpeedCoef_*sys_->getVMax();
// 	}
//       else
// 	{
// 	  double coefSpd = distTravel / distAccMaxSpd;
// 	  targetSpd = curST.s() + coefSpd*(maxSpeedCoef_*sys_->getVMax() - curST.s());
// 	}
//     }

//   double midSpd = 0.5*(curST.s()+targetSpd);
//   DT = distTravel / midSpd;

//   if(indTargetCfg == path.size()-1)
//     {
//       targetST.x()  = s2.x();
//       targetST.y()  = s2.y();
//       //targetST.th() = s2.th();
//       targetST.th() = targetHead;
//       targetST.s()  = s2.s();
//       targetST.om() = s2.om();
//       targetST.t()  = curST.t() + DT;
//     }
//   else
//     {
//       targetST.x()  = targetCfg.x();
//       targetST.y()  = targetCfg.y();
//       targetST.th() = targetHead;
//       targetST.s()  = targetSpd;
//       targetST.om() = 0;
//       targetST.t()  = curST.t() + DT;
//     }
}


/**
 *  \brief compute the target state-time to reach from its position and the previous state-time
 *  \param path : the path computed on the roadmap
 *  \param curST : the current state-time
 *  \param indCurCfg : index of the current configuration of the path we suppose to have already reached
 *  \param indTargetCfg : index of the target configuration of the path
 *  \param s2 : the goal state of the trajectory  
 *  \param targetST : the resulting target state-time
 */
void WCRoadMapTP::computeTargetST(const PMPath& path,const WCStateT& curST,int indCurCfg,int indTargetCfg,const WCState& s2,WCStateT& targetST)
{
  // get back the position of the target state to compute
  PMConfig targetCfg = path.at(indTargetCfg);
  // compute the length of the path
  double pathLength = path.getLength();

  // compute the heading of the last configuration
  double targetHead;
  if(indTargetCfg == path.size()-1)
    {
      PMConfig prevCfgTgt = path.at(indTargetCfg-1);
      double dxPT = targetCfg.x() - prevCfgTgt.x();
      double dyPT = targetCfg.y() - prevCfgTgt.y();
      double dirVecPT = atan2(dyPT,dxPT);

      targetHead = normalizedAngle(dirVecPT);
//       cout << "targetHead : " << targetHead << endl;

//       double dx2 = targetCfg.x() - curST.x();
//       double dy2 = targetCfg.y() - curST.y();
//       double dirVec = atan2(dy2,dx2);
//       double diffAngle = computeMinDiffAngle(normalizedAngle(dirVec),normalizedAngle(curST.th()));
//       targetHead = normalizedAngle(curST.th() + 2.0*diffAngle);
    }
  else
    {
      PMConfig prevCfgTgt = path.at(indTargetCfg-1);
      PMConfig nextCfgTgt = path.at(indTargetCfg+1);

      double dxPT = targetCfg.x() - prevCfgTgt.x();
      double dyPT = targetCfg.y() - prevCfgTgt.y();
      double dirVecPT = atan2(dyPT,dxPT);
      double dxTN = nextCfgTgt.x() - targetCfg.x();
      double dyTN = nextCfgTgt.y() - targetCfg.y();
      double dirVecTN = atan2(dyTN,dxTN);

      double diffAngle = computeMinDiffAngle(normalizedAngle(dirVecTN),normalizedAngle(dirVecPT));
      targetHead = normalizedAngle(dirVecPT + 0.5*diffAngle);
    }

  // compute the minimal distance to stop the system at the current speed
  double distStop = 0;
  if(curST.s() > s2.s())
    {
      double dtV = (curST.s() - s2.s()) / (maxAccCoef_*sys_->getAMax());
      distStop = curST.s()*dtV - 0.5*maxAccCoef_*sys_->getAMax()*dtV*dtV;
    }

  // compute the remaining distance of the path to travel
  double curLength = path.getLengthUpTo(indCurCfg);
  double remainDist = pathLength - curLength;

  // check if we need to decelerate now to be able to reach the final speed
  bool dec = false;
  if(distStop > remainDist) dec = true;

  // compute the estimated distance to travel between the current state and the target configuration
  double distCurTgtCfg = path.getLengthBetween(indCurCfg,indTargetCfg);
  double distTravel = (distCurTgtCfg)*1.2;

  // compute the final speed to reach
  double targetSpd;
  double DT;
  if(dec)
    {
      double coefSpd = (distCurTgtCfg) / (pathLength - curLength);
      targetSpd = coefSpd*s2.s() + (1.0-coefSpd)*curST.s();
//        cout << "deceleration control; coefSpd : " << coefSpd << endl;
    }
  else
    {
      double dtV = ((maxSpeedCoef_*sys_->getVMax()) - curST.s()) / (maxAccCoef_*sys_->getAMax());
      double distAccMaxSpd = curST.s()*dtV + 0.5*maxAccCoef_*sys_->getAMax()*dtV*dtV;
      if(distAccMaxSpd < distTravel)
	{
//  	  cout << "acceleration (can reach the max speed)" << endl;
	  targetSpd = maxSpeedCoef_*sys_->getVMax();
	}
      else
	{
//  	  cout << "acceleration (cannot reach the max speed)" << endl;
	  double coefSpd = distTravel / distAccMaxSpd;
	  targetSpd = curST.s() + coefSpd*(maxSpeedCoef_*sys_->getVMax() - curST.s());
	}
    }

  cout << "curST.s() : " << curST.s() << "  targetSpd : " << targetSpd;
  double midSpd = 0.5*(curST.s()+targetSpd);
  DT = distTravel / fabs(midSpd);
  cout << "  midSpd : " << midSpd << "  DT : " << DT << endl;

  if(indTargetCfg == path.size()-1)
    {
      targetST.x()  = s2.x();
      targetST.y()  = s2.y();
      //targetST.th() = s2.th();
      targetST.th() = targetHead;
      targetST.s()  = s2.s();
      targetST.om() = s2.om();
      targetST.t()  = curST.t() + DT;
    }
  else
    {
      targetST.x()  = targetCfg.x();
      targetST.y()  = targetCfg.y();
      targetST.th() = targetHead;
      targetST.s()  = targetSpd;
      targetST.om() = 0;
      targetST.t()  = curST.t() + DT;
    }
}

/**
 *  \brief try to generate a trajectory between two state-times
 *  \param s1 : the initial state-time
 *  \param s2 : the goal state-time
 *  \param sR : the final state-time reached
 *  \param trajRes : the resulting discrete trajectory
 */
void WCRoadMapTP::genTraj(const WCStateT& s1,const WCStateT& s2,WCStateT& sR,WCSTraj& traj)
{
  WCTrajGen trajgen(sys());
  trajgen.fixFinalXY(true);
  trajgen.fixFinalTH(true);
  trajgen.fixFinalV(false);
  trajgen.fixFinalOM(false);

  trajgen.setUseVelCst(true);
  trajgen.setUseAccCst(false);
  trajgen.setUseAVelCst(false);
  trajgen.setUseAAccCst(false);

  trajgen.setDBG(0);
  WCTrajGen::setMaxNbStep(30);
  
//   // init the initial state vector
//   Vec_DP x0(6);
//   x0[0] = s1.x();
//   x0[1] = s1.y();
//   x0[2] = s1.th();
//   x0[3] = s1.s();
//   x0[4] = s1.om();
//   x0[5] = 0.0;
  
//   // init the final state vector
//   Vec_DP xf(6);
//   xf[0] = s2.x();
//   xf[1] = s2.y();
//   xf[2] = s2.th();
//   xf[3] = s2.s();
//   xf[4] = s2.om();
//   xf[5] = s2.t() - s1.t();
  
  // init the constraints vector
//   Vec_DP constraints(4); 
//   constraints[0] = sys()->getAMax();
//   constraints[1] = sys()->getAAMax();
//   constraints[2] = sys()->getVMax();
//   constraints[3] = sys()->getAVMax();

  int res = trajgen.TrajGen::ComputeTrajectory(s1,s2,(double**)NULL);

  double dt = s2.t() - s1.t();
  double x = trajgen.funcX(dt);
  double y = trajgen.funcY(dt); 
  double th = trajgen.funcTheta(dt);
  double v = trajgen.funcVel(dt);
  double om = trajgen.funcAVel(dt);

  sR = WCStateT(x,y,th,v,om,s1.t() + dt);

  int nbT = ((int)(dt / RoadMapTP::timeStep_)) + 1;
  double realTS = dt / nbT;

  for(int i=0; i<=nbT; i++)
    {
      double dt1 = i*realTS;

      double x1 = trajgen.funcX(dt1);
      double y1 = trajgen.funcY(dt1); 
      double th1 = trajgen.funcTheta(dt1);
      double v1 = trajgen.funcVel(dt1);
      double om1 = trajgen.funcAVel(dt1);

      WCStateT newST(x1,y1,th1,v1,om1,s1.t() + dt1);

      traj.append(newST);
    }
}

/**
 *  \brief adjust the time and speed of the trajectory once it has been computed
 *  \param sInit : the initial state
 *  \param sGoal : the final state
 *  \param pathLength : the length of the path
 *
 *  The Trajectory results in a three steps velocity profile
 *  - A first step of acceleration up to the maximal speed
 *  - A second step with constant velocity (this part may be empty)
 *  - A last part of deceleration down to the final speed
 */
void WCRoadMapTP::adjustTimeSpeed(const WCState& sInit,const WCState& sGoal,double pathLength)
{
  // get back the initial and final states of the trajectory

  // compute the maximal acceleration and speed used for the planning
  double maxAcc = sys()->getAMax() * RoadMapTP::maxAccCoef_;
  double maxSpd = sys()->getVMax() * RoadMapTP::maxSpeedCoef_;

  // compute the time and length of the path needed to reach the maximal speed from the initial speed
  double t1,l1;
  minTDtoReachSpd(sInit.s(),maxSpd,maxAcc,t1,l1);

  // compute the time and length of the path needed to reach the final speed from the maximal speed
  double t2,l2;
  minTDtoReachSpd(maxSpd,sGoal.s(),maxAcc,t2,l2);
  
  // check if the maximal speed may be reached
  if(l1 + l2 < pathLength)
    {
      // compute the length of the intermediate part (traveled at maximal speed)
      double lInt = pathLength - (l1+l2);
      // compute the associated time
      double tInt = lInt/maxSpd;

//       double curLength = 0;
      double curT = 0;
      int step = 0;             // step = 0 -> acceleration phase
                                //        1 -> constant maximal speed
                                //        2 -> deceleration phase
      double tStep1 = t1;
      double tStep2 = t1+tInt;

      cout << "t1 : " << t1 << "  t2 : " << t2 << "  tInt : " << tInt << "  tStep1 : " << tStep1 << "  tStep2 : " << tStep2 << endl;

      int N = traj_->getSize();

      for(int i=0; i<N; i++)
	{
	  // handle the case of the initial state
	  if(i == 0)
	    {
	      (*(this->traj()))[N-1].t() = 0.0;
	    }
	  // handle the case of the final state
	  else if(i == N-1)
	    {
	      (*(this->traj()))[N-1].t() = t1+tInt+t2;
	    }
	  else
	    {
	      double curDT=0;
	      double finalV=0;
// 	      double curLg=0;

	      // get back the current and the previous state-times
	      WCStateT prev = traj()->at(i-1);
	      WCStateT cur = traj()->at(i);

	      // check for the reached step
	      if(curT < tStep1)
		{
		  step = 0;
		  // compute the time to reach the current state from the previous one
		  WCRoadMapTP::elapsedTime_AccPhase(prev,cur,prev.s(),maxAcc,tStep1-curT,maxSpd,curDT,finalV);
		}
	      else if(curT < tStep2)
		{
		  step = 1;
		  // compute the time to reach the current state from the previous one
		  WCRoadMapTP::elapsedTime_CstPhase(prev,cur,maxSpd,tStep2-curT,maxAcc,curDT,finalV);
		}
	      else
		{
		  step = 2;
		  // compute the time to reach the current state from the previous one
		  WCRoadMapTP::elapsedTime_DecPhase(prev,cur,prev.s(),maxAcc,curDT,finalV);
		}

	      (*(this->traj()))[i].t() = traj()->at(i-1).t() + curDT;
	      (*(this->traj()))[i].s() = finalV;

// 	      curLength += curLg;
	      curT += curDT;
	    }
	}
    }
  else
    {
      cerr << "WCRoadMapTP::adjustTimeSpeed : case not handled yet!" << endl;
//       exit(1);
    }
}

/**
 *  \brief compute the elapsed time between two states during the acceleration phase
 *  \param s1 : the initial state-time
 *  \param s2 : the final state-time
 *  \param v1 : the initial speed
 *  \param amax : the maximal acceleration
 *  \param tcst : the remaining time before the constant speed phase
 *  \param vmax : the maximal speed
 *  \param finalDT : the resulting elapsed time
 *  \param finalV : the speed of the final state
 */
void WCRoadMapTP::elapsedTime_AccPhase(const WCStateT& s1,const WCStateT& s2,double v1,double amax,double tcst,double vmax,double& finalDT,double& finalV)
{
  // compute the distance between positions p1 and p2
  double dx = s2.x() - s1.x();
  double dy = s2.y() - s1.y();
  double dist = sqrt(dx*dx+dy*dy);

  double delta = v1*v1 +2.0*amax*dist;
  double dt = (-v1 + sqrt(delta))/amax;

  // check if the maximal speed hasn't been reached
  if(dt > tcst) 
    {
      double dAtCst = dist - (v1*tcst + 0.5*tcst*tcst*amax);
      double tAtCst = dAtCst / vmax;
      finalDT = (tcst + tAtCst);
      finalV = vmax;
    }
  else 
    {
      finalDT = dt;
      finalV = v1 + finalDT*amax;
    }
  cout << "WCRoadMapTP::elapsedTime_AccPhase("; s1.display(); cout << ","; s2.display(); cout << ", v1:" << v1 
       << ", amax:" << amax << ", tcst:" << tcst << ", vmax:" << vmax << ", finalDT:" << finalDT << ", finalV:" << finalV << ")" << endl;
}

/**
 *  \brief compute the elapsed time between two states during the constant speed phase
 *  \param s1 : the initial state-time
 *  \param s2 : the final state-time
 *  \param vmax : the maximal speed
 *  \param tdec : the remaining time before the deceleration phase
 *  \param amax : the maximal acceleration
 *  \param finalDT : the resulting elapsed time
 *  \param finalV : the speed of the final state
 */
void WCRoadMapTP::elapsedTime_CstPhase(const WCStateT& s1,const WCStateT& s2,double vmax,double tdec,double amax,double& finalDT,double& finalV)
{
  // compute the distance between positions p1 and p2
  double dx = s2.x() - s1.x();
  double dy = s2.y() - s1.y();
  double dist = sqrt(dx*dx+dy*dy);

  double dt = dist / vmax;

  if(dt > tdec)
    {
      double dAtDec = dist - (vmax*tdec);

      double delta = vmax*vmax -2.0*amax*dAtDec;
      double tAtDec = (-vmax + sqrt(delta))/(-amax);
      if(tAtDec < 0) throw string("WCRoadMapTP::elapsedTime_CstPhase : tAtDec < 0");

      finalDT = (tdec + tAtDec);
      finalV = vmax - tAtDec*amax;
    }
  else 
    {
      finalDT = dt;
      finalV = vmax;
    }
}

/**
 *  \brief compute the elapsed time between two states during the acceleration phase
 *  \param s1 : the initial state-time
 *  \param s2 : the final state-time
 *  \param v1 : the initial speed
 *  \param amax : the maximal acceleration
 *  \param finalDT : the resulting elapsed time
 *  \param finalV : the speed of the final state
 */
void WCRoadMapTP::elapsedTime_DecPhase(const WCStateT& s1,const WCStateT& s2,double v1,double amax,double& finalDT,double& finalV)
{
  // compute the distance between positions p1 and p2
  double dx = s2.x() - s1.x();
  double dy = s2.y() - s1.y();
  double dist = sqrt(dx*dx+dy*dy);

  double delta = v1*v1 -2.0*amax*dist;
  double dt = (-v1 + sqrt(delta))/(-amax);
  if(dt < 0) throw string("WCRoadMapTP::elapsedTime_DecPhase : dt < 0");

  finalDT = dt;
  finalV = v1 - finalDT*amax;
}

/**
 *  \brief resample the trajectory
 */
void WCRoadMapTP::resampleTraj()
{
  // check if the trajectory has been correctly computed
  if(this->traj_ == NULL || this->traj()->getSize() == 0) 
    {
      //cerr << "WCRoadMapTP::resampleTraj : The trajectory hasn't been computed!" << endl;
      throw string("WCRoadMapTP::resampleTraj : The trajectory hasn't been computed!");
    }

  // get back the initial and final state-times of the trajectory
  WCStateT sInit = this->traj()->at(0);
  WCStateT sGoal = this->traj()->at(this->traj()->getSize()-1);

  // compute the elapsed time between the initial and final states
  double totalTime = sGoal.t() - sInit.t();
  // compute the number of expected state-time in the trajectory 
  // (depending of the expected time interval between each state-time)
  int nbST = ((int)(totalTime / timeStep_))+1;
  // compute the real time interval used
  double dt = totalTime / ((double)nbST);

  // save the trajectory that we want to resample
  WCSTraj* oldTraj = this->traj();
  // create the new trajectory
  this->traj_ = new WCSTraj();

  // sample the new trajectory from the pevious one
  for(int i=0; i<=nbST; i++)
    {
      WCStateT curST;
      oldTraj->getStateAtTime(i*dt,curST);
      this->traj()->append(curST);
    }

  // delete the old trajectory
  delete oldTraj;
}
