#include "CL_SSLattice.hpp"
#include "CLTrajGen.hpp"
using namespace mpc;

#include "gnuplot/inc/gnuplot_i.hpp"
#include "common/inc/Utils.hpp"

                                    /**************************************************/
                                    /**************** Class CL_SSLattice **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CL_SSLattice::CL_SSLattice() : SSLattice()
{
  this->ctrlSize_ = 2;

  tabPosTrajX_ = NULL;
  tabPosTrajY_ = NULL;
  tabPosTrajT_ = NULL;
  tabPosTrajV_ = NULL;
  tabPosTrajSt_ = NULL;
}

/**
 *  \brief init constructor
 *  \param sys : the car-like system for which the state space lattice is built
 */
CL_SSLattice::CL_SSLattice(CarLikeSys* sys) : SSLattice(sys)
{
  this->ctrlSize_ = 2;

  tabPosTrajX_ = NULL;
  tabPosTrajY_ = NULL;
  tabPosTrajT_ = NULL;
  tabPosTrajV_ = NULL;
  tabPosTrajSt_ = NULL;
}

/**
 *  \brief destructor
 */
CL_SSLattice::~CL_SSLattice()
{
  if(tabPosTrajX_ != NULL) delete tabPosTrajX_;
  if(tabPosTrajY_ != NULL) delete tabPosTrajY_;
  if(tabPosTrajT_ != NULL) delete tabPosTrajT_;
  if(tabPosTrajV_ != NULL) delete tabPosTrajV_;
  if(tabPosTrajSt_ != NULL) delete tabPosTrajSt_;
}


/************** Lattice construction **************/

/**
 *  \brief allocate a new state-time and copy the input one
 *  \param s : input state-time
 *  \return a pointer to the copy of the input state-time
 */
CLStateT* CL_SSLattice::copyStateT(const StateTime& s)
{
  return new CLStateT(((CLStateT*)(&s))->x(),
		      ((CLStateT*)(&s))->y(),
		      ((CLStateT*)(&s))->th(),
		      ((CLStateT*)(&s))->ph(),
		      ((CLStateT*)(&s))->s(),
		      ((CLStateT*)(&s))->t());
}

/**
 *  \brief generate a trajectory to compute the state reached by s by ct
 *  \param st : the initial state-time of the trajectory
 *  \param ct : the control-time applied on the system
 *  \return the final state-time reached
 */
CLStateT* CL_SSLattice::genTraj(const StateTime* st,const ControlTime* ct)
{
  cout << "st : "; st->display(); 
  cout << "  ct : "; ct->display(); cout << endl;

  // generate the trajectory
  CLTrajGen trajgen(sys());

  trajgen.ComputeFromCtrl(*st,*ct);
  CLStateT finalST = trajgen.getCLStateTReachedAt(ct->t());
  finalST.t() += st->t();

  // fill the trajectory vectors if needed
  if(displayTraj_)
    {
      for(int i=0; i<trajSp_; i++)
	{
	  double t = ct->t() * ((double)i)/((double)(trajSp_-1));

	  (*tabPosTrajX_)[I_] = trajgen.funcX(t);
	  (*tabPosTrajY_)[I_] = trajgen.funcY(t);
	  (*tabPosTrajT_)[I_] = st->t() + t;
	  (*tabPosTrajV_)[I_] = trajgen.funcVel(t);
	  (*tabPosTrajSt_)[I_] = trajgen.funcSteer(t);
	  I_++;
	}
    }
  
  // return  the final state-time of the trajectory
  return new CLStateT(finalST);
}

/**
 *  \brief sample the available controls
 *  \param ctrlSp : number of samples used for each control
 *  \param dt : the time of application of each control
 *  \return the control sampling vector
 */
void CL_SSLattice::sampleControl(int nbCtrlSp,double dt)
{
  // reinit the control sampling if needed
  this->deleteCtrlSp();
  ctrlSp_ = new std::vector<ControlTime*>;

  // init the bounds of the control
  double minAcc = -sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minStSpd = -sys()->getMaxSteerSpd();
  double maxStSpd = sys()->getMaxSteerSpd();

  // check if a correct value of nbCtrlSp is given
  if(nbCtrlSp < 2) nbCtrlSp = 2;

  for(int i=0; i<nbCtrlSp; i++)
    {
      double acc = minAcc + ((((double)i) / ((double)(nbCtrlSp-1)))*(maxAcc-minAcc));

      for(int j=0; j<nbCtrlSp; j++)
	{
	  double stSpd = minStSpd + ((((double)j) / ((double)(nbCtrlSp-1)))*(maxStSpd-minStSpd));

	  // compute the new control-time
	  CLControlT* ctrlt = new CLControlT(acc,stSpd,dt);
	  
	  // append it to the vector of control-times
	  ctrlSp_->push_back(ctrlt);
	}
    }
}


/************** Feasibility functions *************/

/**
 *  \brief compute the available error between two state-times
 *  \return the available error between two state-times
 */
StateTime* CL_SSLattice::computeAvError(const StateTime* s1,const StateTime* s2) const
{
  CLStateT clS1(*((CLStateT*)(s1)));
  CLStateT clS2(*((CLStateT*)(s2)));

  double dt = curDT_;

  double maxV = max(clS1.s(),clS2.s());
  double maxPH = max(fabs(clS1.ph()),fabs(clS2.ph()));

  StateTime* sres = new CLStateT(0.01* curDT_ * maxV,
				 0.01* curDT_ * maxV,
				 0.1* curDT_ * maxV * tan(maxPH) / sys()->getLength(),
				 0.1* curDT_ * sys()->getMaxSteerSpd(),
				 0.1* curDT_ * sys()->getAMax(),
				 0.0);
  return sres;
}

/**
 *  \brief compute the difference between two state-times
 *  \return the difference between two state-times
 *
 *  A new instance of a state-time is computed, it must be deleted
 */
StateTime* CL_SSLattice::diffStateT(const StateTime* s1,const StateTime* s2) const
{
  cout << "deb diffStateT" << endl;
  cout << "s1 : "; s1->display(); cout << endl;
  cout << "s2 : "; s2->display(); cout << endl;

  CLStateT clS1(*((CLStateT*)(s1)));
  CLStateT clS2(*((CLStateT*)(s2)));

  cout << "clS1 : "; clS1.display(); cout << endl;
  cout << "clS2 : "; clS2.display(); cout << endl;
  cout << "2" << endl;

  StateTime* sres = new CLStateT(fabs(clS1.x() - clS2.x()),
				 fabs(clS1.y() - clS2.y()),
				 fabs(computeMinDiffAngle(clS1.th(),clS2.th())),
				 fabs(clS1.ph() - clS2.ph()),
				 fabs(clS1.s() - clS2.s()),
				 fabs(clS1.t() - clS2.t()));
  return sres;
}


/**************** Utility functions ***************/

/**
 *  \allocate the trajectory vectors
 *  \param nb : the number of trajectories to allocate
 */
void CL_SSLattice::allocTrajVec(int nb)
{
  int size = trajSp_ * nb;
  
  tabPosTrajX_ = new vector<double>(size);
  tabPosTrajY_ = new vector<double>(size);
  tabPosTrajT_ = new vector<double>(size);
  tabPosTrajV_ = new vector<double>(size);
  tabPosTrajSt_ = new vector<double>(size);

  // reinit the index of vectors
  I_ = 0;
}


/**************** Display functions ***************/

/**
 *  \brief draw the lattice in a gnuplot window
 *  \return 0 if the process succeed, -1 else
 */
int CL_SSLattice::drawGP() const
{
  // check if the lattice has been defined
  if(tree_ == NULL || tabPosTrajX_ == NULL || tabPosTrajY_ == NULL || tabPosTrajV_ == NULL)
    {
      return -1;
    }

//   for(unsigned int i=0; i<tabPosTrajX_->size(); i++)
//     {
//       cout << "X:" << tabPosTrajX_->at(i) << "  Y:" << tabPosTrajY_->at(i) << "  V:" << tabPosTrajV_->at(i) << "  T:" << tabPosTrajT_->at(i) << endl;
//     }

  // draw the XxYxT profile
  Gnuplot gXYT;
  gXYT.reset_plot();
  gXYT.cmd("set linestyle 1");
  gXYT.set_style("points").plot_xyz(*tabPosTrajX_,*tabPosTrajY_,*tabPosTrajT_,"SS Lattice (XxYxT)");  

//   for(unsigned int i=0; i<tabPosTrajX_->size(); i++)
//     {
//       int ind = i;
//     }

  // draw the XxYxV profile
  Gnuplot gXYV;
  gXYV.reset_plot();
  gXYV.cmd("set linestyle 1");
  gXYV.set_style("points").plot_xyz(*tabPosTrajX_,*tabPosTrajY_,*tabPosTrajV_,"SS Lattice (XxYxV)");
  
//   for(unsigned int i=0; i<tabPosTrajX_->size(); i++)
//     {
//       int ind = i;
//     }
  
  wait_for_key();
}
