#include "SS_SSLattice.hpp"
#include "SSTrajGen.hpp"
using namespace mpc;

#include "gnuplot/inc/gnuplot_i.hpp"
#include "common/inc/Utils.hpp"

                                    /**************************************************/
                                    /**************** Class SS_SSLattice **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SS_SSLattice::SS_SSLattice() : SSLattice()
{
  this->ctrlSize_ = 2;

  tabPosTrajX_ = NULL;
  tabPosTrajY_ = NULL;
  tabPosTrajT_ = NULL;
  tabPosTrajVX_ = NULL;
  tabPosTrajVY_ = NULL;
}

/**
 *  \brief init constructor
 *  \param sys : the spaceship system for which the state space lattice is built
 */
SS_SSLattice::SS_SSLattice(SpaceShipSys* sys) : SSLattice(sys)
{
  this->ctrlSize_ = 2;

  tabPosTrajX_ = NULL;
  tabPosTrajY_ = NULL;
  tabPosTrajT_ = NULL;
  tabPosTrajVX_ = NULL;
  tabPosTrajVY_ = NULL;
}

/**
 *  \brief destructor
 */
SS_SSLattice::~SS_SSLattice()
{
  if(tabPosTrajX_ != NULL) delete tabPosTrajX_;
  if(tabPosTrajY_ != NULL) delete tabPosTrajY_;
  if(tabPosTrajT_ != NULL) delete tabPosTrajT_;
  if(tabPosTrajVX_ != NULL) delete tabPosTrajVX_;
  if(tabPosTrajVY_ != NULL) delete tabPosTrajVY_;
}

/************** Lattice construction **************/

/**
 *  \brief allocate a new state-time and copy the input one
 *  \param s : input state-time
 *  \return a pointer to the copy of the input state-time
 */
SSStateT* SS_SSLattice::copyStateT(const StateTime& s)
{
  return new SSStateT(((SSStateT*)(&s))->x(),
		      ((SSStateT*)(&s))->y(),
		      ((SSStateT*)(&s))->th(),
		      ((SSStateT*)(&s))->sx(),
		      ((SSStateT*)(&s))->sy(),
		      ((SSStateT*)(&s))->om(),
		      ((SSStateT*)(&s))->t());
}

/**
 *  \brief generate a trajectory to compute the state reached by s by ct
 *  \param st : the initial state-time of the trajectory
 *  \param ct : the control-time applied on the system
 *  \return the final state-time reached
 */
SSStateT* SS_SSLattice::genTraj(const StateTime* st,const ControlTime* ct)
{
  // generate the trajectory
  SSTrajGen trajgen(sys());
  trajgen.ComputeFromCtrl(*st,*ct);
  SSStateT finalST = trajgen.getSSStateTReachedAt(ct->t());
  finalST.t() += st->t();

  // fill the trajectory vectors if needed
  if(displayTraj_)
    {
      for(int i=0; i<trajSp_; i++)
	{
	  double t = ct->t() * ((double)i)/((double)(trajSp_-1));

	  (*tabPosTrajX_)[I_] = trajgen.funcX(t);
	  (*tabPosTrajY_)[I_] = trajgen.funcY(t);
	  (*tabPosTrajT_)[I_] = st->t() + t;
	  (*tabPosTrajVX_)[I_] = trajgen.funcVelX(t);
	  (*tabPosTrajVY_)[I_] = trajgen.funcVelY(t);
	  I_++;
	}
    }
  
  // return  the final state-time of the trajectory
  return new SSStateT(finalST);
}

/**
 *  \brief sample the available controls
 *  \param ctrlSp : number of samples used for each control
 *  \param dt : the time of application of each control
 *  \return the control sampling vector
 */
void SS_SSLattice::sampleControl(int nbCtrlSp,double dt)
{
  // reinit the control sampling if needed
  this->deleteCtrlSp();
  ctrlSp_ = new std::vector<ControlTime*>;

  // init the bounds of the control
  double minAcc = 0.0; //-sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minAAcc = -sys()->getAThMax();
  double maxAAcc = sys()->getAThMax();

  // check if a correct value of nbCtrlSp is given
  if(nbCtrlSp < 3) nbCtrlSp = 3;

  for(int i=0; i<nbCtrlSp; i++)
    {
      double acc = minAcc + ((((double)i) / ((double)(nbCtrlSp-1)))*(maxAcc-minAcc));
	  
      for(int j=0; j<nbCtrlSp; j++)
	{
	  double aacc = minAAcc + ((((double)j) / ((double)(nbCtrlSp-1)))*(maxAAcc-minAAcc));
	  
	  // compute the new control-time
	  SSControlT* ctrlt = new SSControlT(acc,aacc,dt);
	  
	  // append it to the vector of control-times
	  ctrlSp_->push_back(ctrlt);
	}
    }
}

/************** Feasibility functions *************/

/**
 *  \brief compute the available error between two state-times
 *  \return the available error between two state-times
 */
StateTime* SS_SSLattice::computeAvError(const StateTime* s1,const StateTime* s2) const
{
  SSStateT ssS1(*((SSStateT*)(s1)));
  SSStateT ssS2(*((SSStateT*)(s2)));

  double dt = curDT_;

  double maxVX = max(ssS1.sx(),ssS2.sx());
  double maxVY = max(ssS1.sy(),ssS2.sy());
  double maxOM = max(fabs(ssS1.om()),fabs(ssS2.om()));

  StateTime* sres = new SSStateT(0.01* curDT_ * maxVX,
				 0.01* curDT_ * maxVY,
				 0.1* curDT_ * maxOM,
				 0.1* curDT_ * sys()->getAXMax(),
				 0.1* curDT_ * sys()->getAYMax(),
				 0.1* curDT_ * sys()->getAThMax(),
				 0.0);
  return sres;
}

/**
 *  \brief compute the difference between two state-times
 *  \return the difference between two state-times
 *
 *  A new instance of a state-time is computed, it must be deleted
 */
StateTime* SS_SSLattice::diffStateT(const StateTime* s1,const StateTime* s2) const
{
  SSStateT ssS1(*((SSStateT*)(s1)));
  SSStateT ssS2(*((SSStateT*)(s2)));

  StateTime* sres = new SSStateT(fabs(ssS1.x() - ssS2.x()),
				 fabs(ssS1.y() - ssS2.y()),
				 fabs(computeMinDiffAngle(ssS1.th(),ssS2.th())),
				 fabs(ssS1.sx() - ssS2.sx()),
				 fabs(ssS1.sy() - ssS2.sy()),
				 fabs(computeMinDiffAngle(ssS1.om(),ssS2.om())),
				 fabs(ssS1.t() - ssS2.t()));
  return sres;
}

//   /**
//    *  \brief check if the input state-time is reachable according to the sslattice
//    *  \param s : the input state to evaluate
//    */
//   virtual bool isReachable(const StateTime& s) const;


/**************** Utility functions ***************/

/**
 *  \allocate the trajectory vectors
 *  \param nb : the number of trajectories to allocate
 */
void SS_SSLattice::allocTrajVec(int nb)
{
  int size = trajSp_ * nb;
  
  tabPosTrajX_ = new vector<double>(size);
  tabPosTrajY_ = new vector<double>(size);
  tabPosTrajT_ = new vector<double>(size);
  tabPosTrajVX_ = new vector<double>(size);
  tabPosTrajVY_ = new vector<double>(size);

  // reinit the index of vectors
  I_ = 0;
}


/**************** Display functions ***************/

/**
 *  \brief draw the lattice in a gnuplot window
 *  \return 0 if the process succeed, -1 else
 */
int SS_SSLattice::drawGP() const
{
  // check if the lattice has been defined
  if(tree_ == NULL || tabPosTrajX_ == NULL || tabPosTrajY_ == NULL || tabPosTrajT_ == NULL || tabPosTrajVX_ == NULL || tabPosTrajVY_ == NULL)
    {
      return -1;
    }

  // draw the XxYxT profile
  Gnuplot gXYT;
  gXYT.reset_plot();
  gXYT.cmd("set linestyle 1");
  gXYT.set_style("points").plot_xyz(*tabPosTrajX_,*tabPosTrajY_,*tabPosTrajT_,"SS Lattice (XxYxT)");  

  // draw the XxYxV profile
  Gnuplot gXYV;
  gXYV.reset_plot();
  gXYV.cmd("set linestyle 1");
  gXYV.set_style("points").plot_xyz(*tabPosTrajVX_,*tabPosTrajVY_,*tabPosTrajT_,"SS Lattice (VXxVYxT)");  
  
  wait_for_key();
}
