#include "WC_SSLattice.hpp"
#include "WCTrajGen.hpp"
using namespace mpc;

#include "gnuplot/inc/gnuplot_i.hpp"
#include "common/inc/Utils.hpp"

                                    /**************************************************/
                                    /**************** Class WC_SSLattice **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
WC_SSLattice::WC_SSLattice() : SSLattice()
{
  this->ctrlSize_ = 2;

  tabPosTrajX_ = NULL;
  tabPosTrajY_ = NULL;
  tabPosTrajT_ = NULL;
  tabPosTrajV_ = NULL;
  tabPosTrajAV_ = NULL;
}

/**
 *  \brief init constructor
 *  \param sys : the wheelchair system for which the state space lattice is built
 */
WC_SSLattice::WC_SSLattice(WheelChairSys* sys) : SSLattice(sys)
{
  this->ctrlSize_ = 2;

  tabPosTrajX_ = NULL;
  tabPosTrajY_ = NULL;
  tabPosTrajT_ = NULL;
  tabPosTrajV_ = NULL;
  tabPosTrajAV_ = NULL;
}

/**
 *  \brief destructor
 */
WC_SSLattice::~WC_SSLattice()
{
  if(tabPosTrajX_ != NULL) delete tabPosTrajX_;
  if(tabPosTrajY_ != NULL) delete tabPosTrajY_;
  if(tabPosTrajT_ != NULL) delete tabPosTrajT_;
  if(tabPosTrajV_ != NULL) delete tabPosTrajV_;
  if(tabPosTrajAV_ != NULL) delete tabPosTrajAV_;
}


/************** Lattice construction **************/

/**
 *  \brief allocate a new state-time and copy the input one
 *  \param s : input state-time
 *  \return a pointer to the copy of the input state-time
 */
WCStateT* WC_SSLattice::copyStateT(const StateTime& s)
{
  return new WCStateT(((WCStateT*)(&s))->x(),
		      ((WCStateT*)(&s))->y(),
		      ((WCStateT*)(&s))->th(),
		      ((WCStateT*)(&s))->s(),
		      ((WCStateT*)(&s))->om(),
		      ((WCStateT*)(&s))->t());
}

/**
 *  \brief generate a trajectory to compute the state reached by s by ct
 *  \param st : the initial state-time of the trajectory
 *  \param ct : the control-time applied on the system
 *  \return the final state-time reached
 */
WCStateT* WC_SSLattice::genTraj(const StateTime* st,const ControlTime* ct)
{
  // generate the trajectory
  WCTrajGen trajgen(sys());
  trajgen.ComputeFromCtrl(*st,*ct);
  WCStateT finalST = trajgen.getWCStateTReachedAt(ct->t());
  finalST.t() += st->t();

  // fill the trajectory vectors if needed
  if(displayTraj_)
    {
      for(int i=0; i<trajSp_; i++)
	{
	  double t = ct->t() * ((double)i)/((double)(trajSp_-1));

	  (*tabPosTrajX_)[I_] = trajgen.funcX(t);
	  (*tabPosTrajY_)[I_] = trajgen.funcY(t);
	  (*tabPosTrajT_)[I_] = st->t() + t;
	  (*tabPosTrajV_)[I_] = trajgen.funcVel(t);
	  (*tabPosTrajAV_)[I_] = trajgen.funcAVel(t);
	  I_++;
	}
    }
  
  // return  the final state-time of the trajectory
  return new WCStateT(finalST);
}

/**
 *  \brief sample the available controls
 *  \param ctrlSp : number of samples used for each control
 *  \param dt : the time of application of each control
 *  \return the control sampling vector
 */
void WC_SSLattice::sampleControl(int nbCtrlSp,double dt)
{
  // reinit the control sampling if needed
  this->deleteCtrlSp();
  ctrlSp_ = new std::vector<ControlTime*>;

  // init the bounds of the control
  //double minAcc = -sys()->getAMax();
  double minAcc = 0.0;
  double maxAcc = sys()->getAMax();
  double minAAcc = -sys()->getAAMax();
  double maxAAcc = sys()->getAAMax();

  // check if a correct value of nbCtrlSp is given
  if(nbCtrlSp < 2) nbCtrlSp = 2;

  for(int i=0; i<nbCtrlSp; i++)
    {
      double acc = minAcc + ((((double)i) / ((double)(nbCtrlSp-1)))*(maxAcc-minAcc));

      for(int j=0; j<nbCtrlSp; j++)
	{
	  double aacc = minAAcc + ((((double)j) / ((double)(nbCtrlSp-1)))*(maxAAcc-minAAcc));

	  // compute the new control-time
	  WCControlT* ctrlt = new WCControlT(acc,aacc,dt);
	  
	  // append it to the vector of control-times
	  ctrlSp_->push_back(ctrlt);
	}
    }
}


/************** Feasibility functions *************/

/**
 *  \brief compute the available error between two state-times
 *  \return the available error between two state-times
 */
StateTime* WC_SSLattice::computeAvError(const StateTime* s1,const StateTime* s2) const
{
  WCStateT wcS1(*((WCStateT*)(s1)));
  WCStateT wcS2(*((WCStateT*)(s2)));

  double dt = curDT_;

  double maxV = max(wcS1.s(),wcS2.s());
  double maxAV = max(fabs(wcS1.om()),fabs(wcS2.om()));

  StateTime* sres = new WCStateT(curDT_ * maxV,
				 curDT_ * maxV,
				 curDT_ * maxAV,
				 curDT_ * sys()->getAMax(),
				 curDT_ * sys()->getAAMax(),
				 0.0);
  return sres;
}

/**
 *  \brief compute the difference between two state-times
 *  \return the difference between two state-times
 *
 *  A new instance of a state-time is computed, it must be deleted
 */
StateTime* WC_SSLattice::diffStateT(const StateTime* s1,const StateTime* s2) const
{
  WCStateT wcS1(*((WCStateT*)(s1)));
  WCStateT wcS2(*((WCStateT*)(s2)));

  StateTime* sres = new WCStateT(fabs(wcS1.x() - wcS2.x()),
				 fabs(wcS1.y() - wcS2.y()),
				 fabs(computeMinDiffAngle(wcS1.th(),wcS2.th())),
				 fabs(wcS1.s() - wcS2.s()),
				 fabs(wcS1.om() - wcS2.om()),
				 fabs(wcS1.t() - wcS2.t()));
  return sres;
}


/**************** Utility functions ***************/

/**
 *  \allocate the trajectory vectors
 *  \param nb : the number of trajectories to allocate
 */
void WC_SSLattice::allocTrajVec(int nb)
{
  int size = trajSp_ * nb;
  
  tabPosTrajX_ = new vector<double>(size);
  tabPosTrajY_ = new vector<double>(size);
  tabPosTrajT_ = new vector<double>(size);
  tabPosTrajV_ = new vector<double>(size);
  tabPosTrajAV_ = new vector<double>(size);

  // reinit the index of vectors
  I_ = 0;
}


/**************** Display functions ***************/

/**
 *  \brief draw the lattice in a gnuplot window
 *  \return 0 if the process succeed, -1 else
 */
int WC_SSLattice::drawGP() const
{
  // check if the lattice has been defined
  if(tree_ == NULL || tabPosTrajX_ == NULL || tabPosTrajY_ == NULL || tabPosTrajV_ == NULL)
    {
      return -1;
    }

  // draw the XxYxT profile
  Gnuplot gXYT;
  gXYT.reset_plot();
  gXYT.cmd("set linestyle 1");
  gXYT.set_style("points").plot_xyz(*tabPosTrajX_,*tabPosTrajY_,*tabPosTrajT_,"SS Lattice (XxYxT)");  

  // draw the XxYxV profile
  Gnuplot gXYV;
  gXYV.reset_plot();
  gXYV.cmd("set linestyle 1");
  gXYV.set_style("points").plot_xyz(*tabPosTrajX_,*tabPosTrajY_,*tabPosTrajV_,"SS Lattice (XxYxV)");
  
  wait_for_key();
}
