#ifndef SYSTEM_HPP
#define SYSTEM_HPP

//#ifdef HUGR
//#include "hugr.hpp"
//#endif

namespace mpc
{
  /**
   * \enum system_t
   * \brief Enumeration of the different system types
   *
   * ND_SYS         : undefined system
   * POINTMASS_SYS  : point mass system
   * CARLIKE_SYS    : car-like system
   * ARM_SYS        : manipulator arm system
   * SPACESHIP_SYS  : spaceship system
   * DIFFDRIVE_SYS  : differential drive system
   * WHEELCHAIR_SYS : wheelchair system
   * MANA_SYS       : Mana system
   * DIFFDRIVE2_SYS : differential drive system
   */
  typedef enum {
    ND_SYS             =  0,
    POINTMASS_SYS      =  1,
    CARLIKE_SYS        =  2,
    ARM_SYS            =  3,
    SPACESHIP_SYS      =  4,
    DIFFDRIVE_SYS      =  5,
    WHEELCHAIR_SYS     =  6,
    MANA_SYS           =  7,
    DIFFDRIVE2_SYS      =  8
  } system_t;

                                    /**************************************************/
                                    /****************** Class System ******************/
                                    /**************************************************/

  /** \class System
   * \brief Generic robotic system class
   *
   *  Each robotic system shall derivate from this class
   */
  class System
  {
  protected : 
    system_t type_;                   /**< type of the robotic system */
  
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    System();
    /**
     *  \brief destructor
     */
    virtual ~System();

    /***************** Initialization *****************/

    //   /**
    //    *  \brief init the system from an input file
    //    */
    //   void init(char* filename) = 0;

    /************** Getters and setters ***************/
  
    /**
     *  \brief getter : type of the robotic system
     *  \return the type of the robotic system
     */
    virtual system_t getType() const;
    /**
     *  \brief setter : type of the robotic system
     *  set the type of the robotic system
     *  \param t : the new type of the robotic system
     */
    virtual void setType(const system_t& t);

    /**
     *  \brief getter : radius of the robotic system
     *  return the radius of an approximative circle including the robot
     *  \return radius of the robotic system
     */
    virtual double getRadius() const = 0;
  
    /**
     *  \brief getter : maximal speed
     *  return the maximal speed wrt. its environment
     *  \return maximal speed
     */
    virtual double getVMax() const = 0;
    /**
     *  \brief getter : maximal acceleration
     *  return the maximal acceleration wrt. its environment
     *  \return acceleration speed
     */
    virtual double getAMax() const = 0;

    /*************** Display functions ****************/

    /**
     *  \brief display the features of the robot
     */
    virtual void display() const = 0;

    /**
     *  \brief convert the system into a string
     *  \return the resulting string
     */
    virtual const char* toString() const = 0;

    /**
     *  \brief draw the robotic system in an openGL viewer
     */
    virtual void toGL() const = 0;
  };

};
#endif
