#include "CarLikeSys.hpp"
using namespace mpc;

#include <cmath>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <fstream>
using namespace std;

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
namespace po = boost::program_options;
namespace fs = boost::filesystem;

/********** Constructors and destructors **********/

// default constructor
CarLikeSys::CarLikeSys() : System()
{
  this->length_       = 0;
  this->vmax_         = 0;
  this->amax_         = 0;
  this->maxSteering_  = 0;
  this->maxSteerSpd_  = 0;
  this->maxCurvature_ = 0;
  this->maxCurvSpd_   = 0;
  this->type_         = CARLIKE_SYS;
  this->canMoveBack_  = false;
}

/**
 *  \brief init constructor
 *  \param cfg_filename : configuration filename
 */
CarLikeSys::CarLikeSys(char* cfg_filename)
{
  this->length_       = 0;
  this->vmax_         = 0;
  this->amax_         = 0;
  this->maxSteering_  = 0;
  this->maxSteerSpd_  = 0;
  this->maxCurvature_ = 0;
  this->maxCurvSpd_   = 0;
  this->type_         = CARLIKE_SYS;
  this->canMoveBack_  = false;

  this->loadFromCfgFile(cfg_filename);
}

// destructor
CarLikeSys::~CarLikeSys(){}


/************** Getters and setters ***************/

// return the length of the robot
double CarLikeSys::getLength() const
{
  return this->length_;
}

// set the length of the robot
void CarLikeSys::setLength(const double& l)
{
  this->length_ = l;
}

// return the radius of the circle including the robot
double CarLikeSys::getRadius() const
{
  return this->radius_;
}

// set the radius of the circle including the robot
void CarLikeSys::setRadius(const double& r)
{
  this->radius_ = r;
}

// return the maximal speed
double CarLikeSys::getVMax() const
{
  return this->vmax_;
}

// set the maximal speed
void CarLikeSys::setVMax(const double& v)
{
  this->vmax_ = v;
}

// return the maximal acceleration
double CarLikeSys::getAMax() const
{
  return this->amax_;
}

// set the maximal acceleration
void CarLikeSys::setAMax(const double& a)
{
  this->amax_ = a;
}

// return the maximal steering
double CarLikeSys::getMaxSteering() const
{
  return this->maxSteering_;
}

// set the maximal steering
void CarLikeSys::setMaxSteering(const double& s)
{
  // set the maximal steering
  this->maxSteering_ = s;
  // set the maximal curvature
  if(this->length_ > 0)
    {
      this->maxCurvature_ = (tan(this->maxSteering_) / this->length_);
    }
}

// return the maximal steering speed
double CarLikeSys::getMaxSteerSpd() const
{
  return this->maxSteerSpd_;
}

// set the maximal steering speed
void CarLikeSys::setMaxSteerSpd(const double& ss)
{
  // set the maximal steering speed
  this->maxSteerSpd_ = ss;
  // set the maximal curvature speed
  if(this->length_ > 0 && this->maxSteering_ != 0)
    {
      this->maxCurvSpd_ = (this->maxSteerSpd_ * (1.0 + pow(tan(this->maxSteering_),2)) / this->length_);
    }
}

// return the maximal curvature
double CarLikeSys::getMaxCurvature() const
{
  //  this->maxCurvature_;
  if(this->length_ < 0)
    {
      // case where the length of the robot hasnt been defined
      return 0;
    }
  else
    {
      // case where the length of the robot is not null
      return (tan(this->maxSteering_) / this->length_);
    }
}

// set the maximal curvature
void CarLikeSys::setMaxCurvature(const double& c)
{
  this->maxCurvature_ = c;
}

// return the maximal curvature speed
double CarLikeSys::getMaxCurvSpd() const
{
//   //  return this->maxCurvSpd_;
//   if(this->length_ <= 0)
//     {
//       // case where the length of the robot hasnt been defined
//       return 0;
//     }
//   else
//     {
//       // case where the length of the robot is not null
//       return (this->maxSteerSpd_ * (1.0 + pow(tan(this->maxSteering_),2)) / this->length_);
//     }
  return this->getMaxCurvature() * 0.2;
}

// set the maximal curvature speed
void CarLikeSys::setMaxCurvSpd(const double& cs)
{
  this->maxCurvSpd_ = cs;
}


/************* Initialization functions **************/

/**
 *  \brief init the car-like system from a configuration file
 *  \param cfg_filename : the configuration filename
 */
void CarLikeSys::loadFromCfgFile(char* cfg_filename)
{
  double maxSteeringDeg;
  double maxSteerSpdDeg;

  po::options_description config_files("Configuration Files"); // Configuration files
  config_files.add_options() 
    ("CarLikeSys.length", po::value<double>(&length_), "")
    ("CarLikeSys.vmax", po::value<double>(&vmax_), "")
    ("CarLikeSys.amax", po::value<double>(&amax_), "")
    ("CarLikeSys.maxSteering", po::value<double>(&maxSteeringDeg), "")
    ("CarLikeSys.maxSteerSpd", po::value<double>(&maxSteerSpdDeg), "")
    ("CarLikeSys.canMoveBack", po::value<bool>(&canMoveBack_), "") 
    ;

  po::variables_map vm_config;
  fs::path cfg_path(fs::initial_path());
#if BOOST_FILESYSTEM_VERSION == 2
  cfg_path = fs::system_complete(fs::path(cfg_filename,fs::native));
  string cfg_string = cfg_path.native_file_string();
#elif BOOST_FILESYSTEM_VERSION == 3
  cfg_path = fs::system_complete(fs::path(cfg_filename));
  string cfg_string = cfg_path.native();
#else
  cfg_path = fs::system_complete(fs::path(cfg_filename,fs::native));
  string cfg_string = cfg_path.native_file_string();
#endif
  if ( !fs::exists( cfg_path ) )
    {
      cout << "DEM_Ifc : Unable to load the poster names. Configuration file not found" << endl;
      return;
    }
  // Open configuration file
  ifstream ifs((const char *)cfg_string.c_str());
  store(parse_config_file(ifs, config_files), vm_config);
  notify(vm_config);

  maxSteering_ = maxSteeringDeg / 180.0 * M_PI;
  maxSteerSpd_ = maxSteerSpdDeg / 180.0 * M_PI;
}


/*************** Display functions ****************/

// display the features of the robot
void CarLikeSys::display() const
{
  cout << "CarLikeSys : " << endl;
  cout << "\t Length : " << length_ << endl;
  cout << "\t Speed Max : " << vmax_ << endl;
  cout << "\t Acc Max : " << amax_ << endl;
  cout << "\t Steer Max : " << maxSteering_ << endl;
  cout << "\t SteerSpd Max : " << maxSteerSpd_ << endl;
  cout << "\t Curvature Max : " << getMaxCurvature() << endl;
  cout << "\t CurvatureSpd Max : " << getMaxCurvSpd() << endl;
  cout<< endl;
}

/**
 *  \brief draw the car-like system in an openGL viewer
 */
void CarLikeSys::toGL() const
{
  double heading = 0;
  double robotRadius = 0.5*length_;
  const GLfloat blueMaterial[]   = {0.,0.,1.,1.};
  const GLfloat paleblueMaterial[] = {0.1,0.1,0.7,1.};
  const GLfloat redMaterial[]    = {1.,0.,0.,1.};

  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);  

  // draw the body of the robot
  glBegin(GL_QUADS);
  glVertex3f(1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading + 0.5*M_PI),
	     1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading + 0.5*M_PI),
	     0.0);
  glVertex3f(1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading - 0.5*M_PI),
	     1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading - 0.5*M_PI),
	     0.0);
  glVertex3f(-1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading - 0.5*M_PI),
	     -1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading - 0.5*M_PI),
	     0.0);
  glVertex3f(-1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading + 0.5*M_PI),
	     -1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading + 0.5*M_PI),
	     0.0);
  
  glEnd();
  
  // draw the wheels of the robot
  glMaterialfv(GL_FRONT, GL_AMBIENT, paleblueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, paleblueMaterial);
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading + 0.5*M_PI),
	       1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 0.1*robotRadius, 0.3*robotRadius, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading - 0.5*M_PI),
	       1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 0.1*robotRadius, 0.3*robotRadius, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(-1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading - 0.5*M_PI),
	       -1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 0.1*robotRadius, 0.3*robotRadius, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(-1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading + 0.5*M_PI),
	       -1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 0.1*robotRadius, 0.3*robotRadius, 1);
  glPopMatrix();
  
  // draw the orientation of the robot
  glLineWidth(((GLfloat)(3.0)));
  glBegin(GL_LINES);
  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
  glVertex3f(0.0,0.0,0.0);
  glVertex3f(2.2*robotRadius*cos(heading), 
	     2.2*robotRadius*sin(heading), 
	     0.0);
  glEnd();
  glLineWidth(((GLfloat)(1.0)));
}

/**
 *  \brief convert the system into a string
 *  \return the resulting string
 */
const char* CarLikeSys::toString() const
{
  cout << "CarLikeSys::toString : Not implemented yet!" << endl;
  exit(1);
}
