#ifndef MPC_CCTRAJFOLLOW_HPP
#define MPC_CCTRAJFOLLOW_HPP

#include "trajFollow/inc/TrajFollow.hpp"
#include "system/inc/CarLikeSys.hpp"
// #include "CLTrajGen2.hpp"
#include "trajGen/inc/CLTrajGen.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class CCTrajFollow ***************/
                                    /**************************************************/

  /** \class CCTrajFollow
   *  \brief Trajectory following process for a cycab system
   */
  class CCTrajFollow : public TrajFollow
  {
  protected : 
    CLStateT lastStateToReach_;                     /**< variable to stock the last state we tried to reach */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Should not be used
     */
    CCTrajFollow();
  
  public : 
    /**
     *  \brief init constructor
     *  \param traj : the trajectory to follow
     */
    CCTrajFollow(CarLikeSys* sys,CLSTraj* traj);
  
    /**
     *  \brief destructor
     */
    virtual ~CCTrajFollow();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    inline virtual const CarLikeSys* sys() const { return ((CarLikeSys*)system_); }

    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    inline virtual const CLSTraj* traj() const { return ((CLSTraj*)traj_); }

    /**
     *  \brief return the last state we tried to reach (with the trajectory generator)
     *  \return the last state we tried to reach
     */
    inline virtual const CLStateT& lastStateToReach() const { return this->lastStateToReach_; }

  protected : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    inline virtual CLTrajGen* trajgen() const { return ((CLTrajGen*)trajgen_); }

    /**
     *  \brief return the last state-time registered to compute the trajectory following process
     *  \return the last state-time registered to compute the trajectory following process
     *
     *  The time if this state-time is relative to the time of trajectory we follow
     */
    inline const CLStateT* curRegST() const { return ((CLStateT*)(this->curRegST_)); }
  
    /*************** Localization update **************/

  public :   
    /**
     *  \brief selection of the state of the trajectory to reach
     *  \param curS : the input current state-time
     *  \param goalS : the resulting state-time to reach
     */
    virtual void selectStateToReach(const CLStateT& curS,CLStateT& goalS);

    /**
     *  \brief Update the localization of the robot
     *  \param curS : current state-time updated
     *  \return 0 if the trajectory generation successed, -1 else
     */
    virtual int updateLoc(const CLStateT& curS);


    /***************** Control request ****************/
  
  protected : 
    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param t : time at which is evaluated the control
     *  \param ctrl : resulting control
     */
    virtual void getControl(double t,CCControl& ctrl);

    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param t : time at which is evaluated the control
     *  \param dt : duration of the application of the control
     *  \param ctrl : resulting control
     */
    virtual void getControl(double t,double dt,CCControl& ctrl);

  public : 
    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param curS : current state-time updated
     *  \param t : time at which is evaluated the control
     *  \param dt : duration of the application of the control
     *  \param ctrl : resulting control
     *  \return 0 if the trajectory generation successed, -1 else
     *
     *  Synchronous localization update and control request
     */
    int getControl(const CLStateT& curS,double t,double dt,CCControl& ctrl);


    /**************** Trajectory update ***************/

    /*
     *  \brief update the trajectory wrt. the maximal allowed speed
     *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
     */
    virtual void updateTrajVelProfile(float coefMaxVel);


    /**************** Utility functions ***************/

    /*
     *  \brief compute the minimal breaking distance of the robot wrt. its current state
     *  \return the minimal breaking distance of the robot
     */
    virtual float minBreakDist() const;

    /*
     *  \brief compute the minimal breaking time needed by the robot wrt. its current state
     *  \return the minimal breaking time needed by the robot
     */
    virtual float minBreakTime() const;

    /*
     *  \brief get back the last time of the trajectory
     *  \return the last time of the trajectory
     */
    double lastTimeTraj();


    /**************** Display functions ***************/
  
  public : 
    /**
     *  \brief display the trajectory computed by the trajectory generator
     */
    virtual void displayTrajGen();
  };

};
#endif // MPC_CCTRAJFOLLOW_HPP
