#ifndef MPC_TRAJFOLLOW_HPP
#define MPC_TRAJFOLLOW_HPP

#include "motion/inc/STrajectory.hpp"
#include "common/inc/ControlTime.hpp"
#include "trajGen/inc/TrajGen.hpp"
#include "system/inc/System.hpp"
#include "common/inc/Timer.hpp"
#include "environment/inc/DistObsMap.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#define TF_DFLT_MIN_DIST_TO_OBS 0.5
#define TF_DFLT_MAX_DIST_TO_OBS 5.0
#define TF_DFLT_MIN_SPEED_COEF 0.1

#define TF_DFLT_MAX_DIST_TGT_REACHED 0.5
#define TF_DFLT_MAX_DIST_WP_REACHED 2.5
#define TF_DFLT_MAX_ANGLE_WP_REACHED 0.75*M_PI

#define TF_TIME_HORIZON_MODE 1
#define TF_DIST_HORIZON_MODE 2

namespace mpc
{
  class TrajGen;
  class System;

                                    /**************************************************/
                                    /**************** Class TrajFollow ****************/
                                    /**************************************************/

  /** \class TrajFollow
   *  \brief Trajectory following process
   */
  class TrajFollow
  {
  protected : 
    STraj* traj_;                                     /*< the trajectory the current process try to follow */
    StateTime* curRegST_;                             /*< the last state-time registered to compute the trajectory following process */
    TrajGen* trajgen_;                                /*< the trajectory generator used to determine the control to apply */
    System* system_;                                  /*< the system used */
    double timeMeeting_;                              /*< time at which we must ensure the robotic system has met the trajectory */
    double lastDuration_;                             /*< duration of the last following process step */
    bool endTraj_;                                    /*< true if the end of the trajectory has been reached */

    double dtReach_;                                  /*< delta time at which we try to reach the reference trajectory */
    double distReach_;                                /*< delta distance at which we try to reach the reference trajectory */
    int horizonMode_;                                 /*< mode to select the state to reach on the trajectory */

    Timer timer_;                                     /*< timer used to date the states of the followed trajectory */

    DistObsMap* distmap_;                             /*< distance to obstacles map */
    boost::mutex mutexDOM_;                           /*< mutex used to prevent DOM access conflicts */
    boost::mutex mutexTraj_;                          /*< mutex used to prevent DOM access conflicts */

    bool tooCloseToObs_;                              /*< true if the robot is too close to obstacles to move */

  public : 
    static float minDistToObstacles_;                 /*< minimal distance that the robot must keep to obstacles */
    static float maxDistToObstacles_;                 /*< maximal distance to obstacles up to which the robot's speed is reduced  */
    static float minCoefSpeed_;                       /*< minimal coefficient used to reduce the robot's speed while it is not too close to the obstacles  */

    static float maxDistTgtReached_;                  /*< maximal distance from the goal position until which it is considered as reached */
    static float maxDistWpReached_;                   /*< maximal distance from a waypoint until which it is considered as reached */
    static float maxAngleWpReached_;                  /*< maximal angle (in deg.) in the front of a waypoint until which it is considered as reached */
    

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Only used by subclasses
     */
    TrajFollow();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the robotic system handeled
     *  \param traj : the trajectory to follow
     */
    TrajFollow(System* sys,STraj* traj);

    /**
     *  \brief destructor
     */
    virtual ~TrajFollow();


    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    virtual const System* sys() const = 0;
  
    /**
     *  \brief return the time of the last internal update
     *  \return time of the last internal update
     */
    virtual double lastTimeUp() const;

    /**
     *  \brief return a reference to the delta time used for trajectory generation
     *  \return reference to the delta time used for trajectory generation
     */
    inline double& dtReach() { return this->dtReach_; }

    /**
     *  \brief return the delta time used for trajectory generation
     *  \return delta time used for trajectory generation
     */
    inline double dtReach() const { return this->dtReach_; }

    /**
     *  \brief return a reference to the delta dist used for trajectory generation
     *  \return reference to the delta dist used for trajectory generation
     */
    inline double& distReach() { return this->distReach_; }

    /**
     *  \brief return the delta dist used for trajectory generation
     *  \return the delta dist used for trajectory generation
     */
    inline double distReach() const { return this->distReach_; }

    /**
     *  \brief return a reference to the mode to select the state to reach on the trajectory
     *  \return a reference to the mode to select the state to reach on the trajectory
     */
    inline int& horizonMode() { return this->horizonMode_; }

    /**
     *  \brief return the mode to select the state to reach on the trajectory
     *  \return the mode to select the state to reach on the trajectory
     */
    inline int horizonMode() const { return this->horizonMode_; }

    /**
     *  \brief return a reference to the time at which we must ensure the robotic system has met the trajectory
     *  \return reference to the time at which we must ensure the robotic system has met the trajectory
     */
    inline double& timeMeeting() { return this->timeMeeting_; }

    /**
     *  \brief return the time at which we must ensure the robotic system has met the trajectory
     *  \return the time at which we must ensure the robotic system has met the trajectory
     */
    inline double timeMeeting() const { return this->timeMeeting_; }

    /**
     *  \brief return the duration of the last following process step
     *  \return duration of the last following process step
     */
    inline double lastDuration() const { return this->lastDuration_; }

    /**
     *  \brief return the associated trajectory
     *  \return the associated trajectory
     */
    inline STraj* getTraj() const { return this->traj_; }

    /**
     *  \brief update the trajectory we want to follow
     *  \param traj : the trajectory we want to follow
     */
    inline void setTraj(STraj* traj) { this->traj_ = traj; }

    /**
     *  \brief return the last state we tried to reach (with the trajectory generator)
     *  \return the last state we tried to reach
     */
    virtual const StateTime& lastStateToReach() const = 0;

    /**
     *  \brief return true if the end (final time) of the trajectory has been reached
     *  \return true if the end (final time) of the trajectory has been reached
     */
    inline bool endTraj() const { return this->endTraj_; }


    //   /*************** Localization update **************/

    //   /**
    //    *  \brief selection of the state of the trajectory to reach
    //    */
    //   virtual void selectStateToReach(StateR& goalS) const = 0;
  
    //   /**
    //    *  \brief Update the localization of the robot
    //    *  \param curS : current state-time updated
    //    */
    //   virtual void updateLoc(const StateR& curS) = 0;


    //   /***************** Control request ****************/
  
    //   /**
    //    *  \brief Request for a new control to apply (at a specified time)
    //    *  \param t : time at which is evaluated the control
    //    *  \param ctrl : resulting control
    //    */
    //   virtual void getControl(double t,Control& ctrl) = 0;

    //   /**
    //    *  \brief Request for a new control to apply (at a specified time)
    //    *  \param t : time at which is evaluated the control
    //    *  \param dt : duration of the application of the control
    //    *  \param ctrl : resulting control
    //    */
    //   virtual void getControl(double t,double dt,Control& ctrl) = 0;


    /*************** Environment update ***************/

  public : 
    /**
     *  \brief update an input distance to obstacles map
     *  \param distmap : the input distmap
     *  \return 0 if the process succeed, -1 else
     */
    int updateDistMap(const DistObsMap* distmap);

  protected : 
    /*
     *  \brief search for the states (poses) of the trajectory that the system is to take until it stops after a breaking maneuver
     *  \return a vector containing the successive poses
     */
    vector<PositionXYT> posesBreakingManeuver();

    /*
     *  \brief compute a coefficient to apply on the maximal speed wrt. the maximal distance to obstacles and to the goal location
     *  \return compute a coefficient to apply on the maximal speed wrt. the environment
     */
    float computeCoefMaxSpeedWrtEnv();

    /**************** Trajectory update ***************/

    /*
     *  \brief update the trajectory wrt. the maximal allowed speed
     *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
     */
    virtual void updateTrajVelProfile(float coefMaxVel) = 0;

    /**************** Utility functions ***************/

    /*
     *  \brief compute the minimal breaking distance of the robot wrt. its current state
     *  \return the minimal breaking distance of the robot
     */
    virtual float minBreakDist() const = 0;

    /*
     *  \brief get back the last time of the trajectory
     *  \return the last time of the trajectory
     */
    double lastTimeTraj();

    /**************** Display functions ***************/

  public :  
    /**
     *  \brief display the trajectory computed by the trajectory generator
     */
    virtual void displayTrajGen() = 0;

    /****************** I/O functions *****************/

  public : 
    /**
     *  \brief return true if the trajectory following process is running
     *  \return true if the trajectory following process is running
     */
    bool isRunning() const;

    /**
     *  \brief start the trajectory following process
     */
    void start();

    /**
     *  \brief stop the trajectory following process
     */
    void stop();

    /**
     *  \brief start the trajectory following process
     *  \param traj : the trajectory to follow
     */
    void reinit(STraj* traj=NULL);
  };

};
#endif // MPC_TRAJFOLLOW_HPP
