#include "CCTrajFollow.hpp"
#include "UtilsTG.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

                                    /**************************************************/
                                    /*************** Class CCTrajFollow ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/
  
/*
 *  \brief default constructor
 *  Should not be used
 */
CCTrajFollow::CCTrajFollow() : TrajFollow()
{
}
  
/*
 *  \brief init constructor
 *  \param traj : the trajectory to follow
 */
CCTrajFollow::CCTrajFollow(CarLikeSys* sys,CLSTraj* traj) : TrajFollow(sys,traj)
{
//   this->trajgen_ = new CLTrajGen2();
  this->trajgen_ = new CLTrajGen(sys);
  this->trajgen_->setMaxResidual(0.05 * this->sys()->getLength());
}
  
/*
 *  \brief destructor
 */
CCTrajFollow::~CCTrajFollow()
{
}
  
  
/*************** Localization update **************/
  
/*
 *  \brief selection of the state of the trajectory to reach
 *  \param curS : the input current state-time
 *  \param goalS : the resulting state-time to reach
 */
void CCTrajFollow::selectStateToReach(const CLStateT& curS,CLStateT& goalS)
{
  // get back the time of the current state
  double curT = this->lastTimeUp();

  // compute the time of the state to reach
  double goalT = curT + this->dtReach();

  // check if the trajectory must already be reached at goalT
  if(goalT <= this->timeMeeting()) goalT = this->timeMeeting();

  // get back the state of the trajectory at time t
  ((CLSTraj*)traj_)->getStateAtTime(goalT,goalS);

  // if the time is too short to generate the trajectory, set it to a minimal time
  if(goalS.t() - curT < this->lastDuration_) goalS.t() = (curT + this->lastDuration_);

  // compute the distance between the init position and the goal position
  double dx = goalS.x() - curS.x();
  double dy = goalS.y() - curS.y();
  double dist = sqrt(dx*dx + dy*dy);
  // if the distance between the current position and the goal position is too great,
  // set the time to the minimal time needed to travel this distance
  double minTravT = minTravelTime(dist,curS.s(),((CarLikeSys*)system_)->getVMax(),((CarLikeSys*)system_)->getAMax());
  if(goalS.t() - curT < minTravT)
    {
//       cerr << "CCTrajFollow::selectStateToReach : time too short" << endl;
      goalS.t() = (curT + minTravT);
    }
  else
    {
//       cerr << "CCTrajFollow::selectStateToReach : time OK" << endl;
      // if the distance between the current position and the goal position is too short,
      // set the time to the maximal time needed to travel this distance
      double maxTravT = maxTravelTime(dist,curS.s(),((CarLikeSys*)system_)->getVMax(),((CarLikeSys*)system_)->getAMax());
      if(goalS.t() - curT > maxTravT)
	{
// 	  cerr << "CCTrajFollow::selectStateToReach : time too great" << endl;
	  goalS.t() = (curT + maxTravT);
	}
    }
  
  // stock the goal state-time computed
  this->lastStateToReach_ = goalS;
  this->lastStateToReach_.t() += curS.t();
}

/*
 *  \brief Update the localization of the robot
 *  \param curS : current state-time 
 *  \return 0 if the trajectory generation successed, -1 else
 */
int CCTrajFollow::updateLoc(const CLStateT& curS)
{
  // update the new state reached
  if(this->curRegST_ != NULL) delete this->curRegST_;
  this->curRegST_ = new CLStateT(curS);

  // select the goal state on the trajectory we want to reach
  CLStateT goalS;
  this->selectStateToReach(curS,goalS);

//   cout << "curS : "; curS.display(); cout << "  goalS : "; goalS.display(); cout << endl;

//   int res = ((CLTrajGen2*)trajgen_)->ComputeTrajectory(curS,goalS,sys(),NULL);
  int res = ((CLTrajGen*)trajgen_)->ComputeTrajectory(curS,goalS,(double**)NULL);

  //trajgen_->displayProfiles();
  //trajgen_->displayAllFct();

//   cout << "final residual : " << trajgen_->finalResidual_ << endl;

  if(res != 0)
    {
      // What to do it the trajectory generation failed ? 
      // Problem to think about quickly...
      // Should keep the best result : So nothing to do?
    }

  return res;
}


/***************** Control request ****************/
  
/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param ctrl : resulting control
 */
void CCTrajFollow::getControl(double t,CCControl& ctrl)
{
  double dt = t - this->lastTimeUp();
  
//   cout << "t = " << t << "  lastTimeUp() = " << lastTimeUp() << "  dt = " << dt << endl;

  // get back the acceleration control to apply at time t
//   double v = ((CLTrajGen2*)trajgen_)->funcVel(dt);
//   double ph = ((CLTrajGen2*)trajgen_)->funcSteer(dt);
  double v = ((CLTrajGen*)trajgen_)->funcVel(dt);
  double ph = ((CLTrajGen*)trajgen_)->funcSteer(dt);

  ctrl = CCControl(v,ph);
}

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 */
void CCTrajFollow::getControl(double t,double dt,CCControl& ctrl)
{
  // A linearization over both initial and final values is 
  // the straightforward solution, over the function may be 
  // uncontinuous... What to do in that case?
  // Ensure bound over control derivatives?
  // Simple method currently used

  // compute the distance between the current state and the final state of the trajectory
  double dx = traj()->lastST().x() - curRegST()->x();
  double dy = traj()->lastST().y() - curRegST()->y();
  double dist = sqrt(dx*dx + dy*dy);
  
  if(t > traj_->lastST().t()) endTraj_ = true;

  // check if the final time of the trajectory has been already reached
//   cout << "t : " << t << "  traj_->lastST().t() : " << traj_->lastST().t() << endl;
  if(t > traj_->lastST().t() && dist < 0.2 * this->sys()->getLength())
    {
      ctrl = CCControl(0,0);
    }
  else
    {
      this->lastDuration_ = dt;
      
      CCControl ctrl1,ctrl2;
      //   cout << "getControl1(t=" << t << ")" << endl;
      this->getControl(t,ctrl1);
      //   cout << "getControl2(t=" << t << ", dt=" << dt << ")" << endl;
      this->getControl(t+dt,ctrl2);
      
//       cout << "ctrl1 : (v=" << ctrl1.v() << ",ph=" << ctrl1.ph() << ") t=" << t;
//       cout << "  ctrl2 : (v=" << ctrl2.v() << ",ph=" << ctrl2.ph() << ") t=" << t+dt << endl;
      
      ctrl = CCControl(0.5*(ctrl1.v() + ctrl2.v()),
		       0.5*(ctrl1.ph() + ctrl2.ph()));
    }
}

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param curS : current state-time updated
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 *  \return 0 if the trajectory generation successed, -1 else
 *
 *  Synchronous localization update and control request
 */
int CCTrajFollow::getControl(const CLStateT& curS,double t,double dt,CCControl& ctrl)
{
  // update the localization
  int res = this->updateLoc(curS);

  // return the control to apply
  this->getControl(t,dt,ctrl);

  return res;
}


/**************** Trajectory update ***************/

/*
 *  \brief update the trajectory wrt. the maximal allowed speed
 *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
 */
void CCTrajFollow::updateTrajVelProfile(float coefMaxVel)
{
//   // check if a localization has already been registered
//   if(this->curRegST() == NULL) return;

//   CL_PVDTrajPlanner planner((CarLikeSys*)system_);

//   float maxVel = coefMaxVel * sys()->getVMax();
//   float startTime = this->curRegST()->t();
//   float startVel = this->curRegST()->v();
//   float endTime = startTime + COEF_TRAJ_UPDATE * this->minBreakTime();
// //   cout << "CLTrajFollow::updateTrajVelProfile -> maxVel : " << maxVel << "  startTime : " << startTime << "  startVel : " << startVel << "  endTime : " << endTime << endl;

//   // replan a part of the trajectory
//   boost::mutex::scoped_lock lock(mutexTraj_);
//   int resReplanning = planner.replanTraj((CLSTraj*)traj_, maxVel, startTime, startVel, endTime);

//   return;

  cerr << "CCTrajFollow::updateTrajVelProfile : Not implemented yet!" << endl;
  exit(1);
}


/**************** Utility functions ***************/

/*
 *  \brief compute the minimal breaking distance of the robot wrt. its current state
 *  \return the minimal breaking distance of the robot
 */
float CCTrajFollow::minBreakDist() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->s();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)dres);
}

/*
 *  \brief compute the minimal breaking time needed by the robot wrt. its current state
 *  \return the minimal breaking time needed by the robot
 */
float CCTrajFollow::minBreakTime() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->s();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)tres);
}

/*
 *  \brief get back the last time of the trajectory
 *  \return the last time of the trajectory
 */
double CCTrajFollow::lastTimeTraj()
{
  boost::mutex::scoped_lock lock(mutexTraj_);

  return traj_->lastST().t();
}


/**************** Display functions ***************/
  
/*
 *  \brief display the trajectory computed by the trajectory generator
 */
void CCTrajFollow::displayTrajGen()
{
  int N = 30;
  glLineWidth(((GLfloat)(2.0)));

  // check if the trajectory generator is available
  if(trajgen_ != NULL && trajgen_->finalNbSteps_ != -1)
    {
      // sample the trajectory
      trajgen()->SaveTrajectory(N);

      glBegin(GL_LINE_STRIP);
      for(int i=0; i<N; i++)
	{
// 	  double X = ((CLTrajGen2*)trajgen_)->posX->at(i);
// 	  double Y = ((CLTrajGen2*)trajgen_)->posY->at(i);
// 	  double T;
// 	  if(curRegST_ != NULL) T = ((CLTrajGen2*)trajgen_)->timeT->at(i) + curRegST_->t();
// 	  else T = ((CLTrajGen2*)trajgen_)->timeT->at(i);

	  double X = ((CLTrajGen*)trajgen_)->posX->at(i);
	  double Y = ((CLTrajGen*)trajgen_)->posY->at(i);
	  double T;
	  if(curRegST_ != NULL) T = ((CLTrajGen*)trajgen_)->timeT->at(i) + curRegST_->t();
	  else T = ((CLTrajGen*)trajgen_)->timeT->at(i);

	  glVertex3f(X,Y,T);
	}
      glEnd();
    }
}
