#include "CLTrajFollow.hpp"
#include "UtilsTG.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

                                    /**************************************************/
                                    /*************** Class CLTrajFollow ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 *  Should not be used
 */
CLTrajFollow::CLTrajFollow() : TrajFollow()
{
}

/*
 *  \brief init constructor
 *  \param traj : the trajectory to follow
 */
CLTrajFollow::CLTrajFollow(CarLikeSys* sys,CLSTraj* traj) : TrajFollow(sys,traj)
{
  this->trajgen_ = new CLTrajGen(sys);
}

/*
 *  \brief destructor
 */
CLTrajFollow::~CLTrajFollow()
{
}


/*************** Localization update **************/


/*
 *  \brief selection of the state of the trajectory to reach
 */
void CLTrajFollow::selectStateToReach(CLStateT& goalS) const
{
  // get back the time of the current state
  double curT = this->lastTimeUp();

  // compute the time of the state to reach
  double goalT = curT + this->dtReach();

  // check if the trajectory must already be reached at goalT
  if(goalT <= this->timeMeeting()) goalT = this->timeMeeting();

  // get back the state of the trajectory at time t
  ((CLSTraj*)traj_)->getStateAtTime(goalT,goalS);

  // if the time is too short to generate the trajectory, set it to a minimal time
  if(goalS.t() - curT < this->lastDuration_) goalS.t() = (curT + this->lastDuration_);
}

/*
 *  \brief Update the localization of the robot
 *  \param curS : current state-time updated
 */
int CLTrajFollow::updateLoc(const CLStateT& curS)
{
  // update the new state reached
  if(this->curRegST_ != NULL) delete this->curRegST_;
  this->curRegST_ = new CLStateT(curS);

  /// trajectory generation initialisation

  // get back the length of the car-like system
  double L = ((CarLikeSys*)system_)->getLength();

  // select the goal state on the trajectory we want to reach
  CLStateT goalS;
  this->selectStateToReach(goalS);

  // convert both initial and final states into state vectors for the trajectory generation process
//   Vec_DP x0,xf;
//   CLStateT_2_CLVecDP(curS,goalS,x0,xf,L);
//   cout << "x0 : (" << x0[0] << "," << x0[1] << "," << x0[2] << "," << x0[3] << "," << x0[4] << "," << x0[5] << ")  " 
//        << "xf : (" << xf[0] << "," << xf[1] << "," << xf[2] << "," << xf[3] << "," << xf[4] << "," << xf[5] << ")" << endl;

  // init the constraints from the car-like system
//   Vec_DP constraints;
//   CLSys_2_CLVecCstt(*((CarLikeSys*)(this->system_)),constraints);

  // init a nul set of parameters to use the precomputed initial guess
  Vec_DP* parameters = NULL;
  
  int res = ((CLTrajGen*)trajgen_)->ComputeTrajectory(curS,goalS,&parameters);
  
  if(res != 0)
    {
      // What to do it the trajectory generation failed ? 
      // Problem to think about quickly...
      // Should keep the best result : So nothing to do?
    }

  return res;
}


/***************** Control request ****************/

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param ctrl : resulting control
 */
void CLTrajFollow::getControl(double t,CLControl& ctrl)
{
  double dt = t - this->lastTimeUp();
  
  // get back the acceleration control to apply at time t
  double acc = ((CLTrajGen*)trajgen_)->funcAcc(dt);
  double ss = ((CLTrajGen*)trajgen_)->funcSteerSpd(dt);

  ctrl = CLControl(acc,ss);
}

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 */
void CLTrajFollow::getControl(double t,double dt,CLControl& ctrl)
{
  // A linearization over both initial and final values is 
  // the straightforward solution, over the function may be 
  // uncontinuous... What to do in that case?
  // Ensure bound over control derivatives?
  // Simple method currently used
  
  if(t > traj_->lastST().t()) endTraj_ = true;

  this->lastDuration_ = dt;

  CLControl ctrl1,ctrl2;
  this->getControl(t,ctrl1);
  this->getControl(t+dt,ctrl2);
  
  ctrl = CLControl(0.5*(ctrl1.acc() + ctrl2.acc()),
		   0.5*(ctrl1.ss() + ctrl2.ss()));
}

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param curS : current state-time updated
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 *
 *  Synchronous localization update and control request
 */
int CLTrajFollow::getControl(const CLStateT& curS,double t,double dt,CLControl& ctrl)
{
  // update the localization
  int res = this->updateLoc(curS);

  // return the control to apply
  this->getControl(t,dt,ctrl);

  return res;
}


/**************** Trajectory update ***************/

/*
 *  \brief update the trajectory wrt. the maximal allowed speed
 *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
 */
void CLTrajFollow::updateTrajVelProfile(float coefMaxVel)
{
//   // check if a localization has already been registered
//   if(this->curRegST() == NULL) return;

//   CL_PVDTrajPlanner planner((CarLikeSys*)system_);

//   float maxVel = coefMaxVel * sys()->getVMax();
//   float startTime = this->curRegST()->t();
//   float startVel = this->curRegST()->v();
//   float endTime = startTime + COEF_TRAJ_UPDATE * this->minBreakTime();
// //   cout << "CLTrajFollow::updateTrajVelProfile -> maxVel : " << maxVel << "  startTime : " << startTime << "  startVel : " << startVel << "  endTime : " << endTime << endl;

//   // replan a part of the trajectory
//   boost::mutex::scoped_lock lock(mutexTraj_);
//   int resReplanning = planner.replanTraj((CLSTraj*)traj_, maxVel, startTime, startVel, endTime);

//   return;

  cerr << "CLTrajFollow::updateTrajVelProfile : Not implemented yet!" << endl;
  exit(1);
}


/**************** Utility functions ***************/

/*
 *  \brief compute the minimal breaking distance of the robot wrt. its current state
 *  \return the minimal breaking distance of the robot
 */
float CLTrajFollow::minBreakDist() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->s();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)dres);
}

/*
 *  \brief compute the minimal breaking time needed by the robot wrt. its current state
 *  \return the minimal breaking time needed by the robot
 */
float CLTrajFollow::minBreakTime() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->s();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)tres);
}

/*
 *  \brief get back the last time of the trajectory
 *  \return the last time of the trajectory
 */
double CLTrajFollow::lastTimeTraj()
{
  boost::mutex::scoped_lock lock(mutexTraj_);

  return traj_->lastST().t();
}


/**************** Display functions ***************/

/*
 *  \brief display the trajectory computed by the trajectory generator
 */
void CLTrajFollow::displayTrajGen()
{
  int N = 30;
  glLineWidth(((GLfloat)(2.0)));

  // check if the trajectory generator is available
  if(trajgen_ != NULL)
    {
      // sample the trajectory
      trajgen()->SaveTrajectory(N);

      glBegin(GL_LINE_STRIP);
      for(int i=0; i<N; i++)
	{
	  double X = ((CLTrajGen*)trajgen_)->posX->at(i);
	  double Y = ((CLTrajGen*)trajgen_)->posY->at(i);
	  double T;
	  if(curRegST_ != NULL) T = ((CLTrajGen*)trajgen_)->timeT->at(i) + curRegST_->t();
	  else T = ((CLTrajGen*)trajgen_)->timeT->at(i);

	  glVertex3f(X,Y,T);
	}
      glEnd();
    }
}
