#include "trajFollow/inc/ManaTrajFollow.hpp"
using namespace mpc;

#include "planner/inc/PVDTrajPlanner.hpp"

#include <iostream>
using namespace std;

#define TF_COEF_TRAJ_UPDATE 2.0
#define TF_COEF_TIME_BREAK 1.5

                                    /**************************************************/
                                    /************** Class ManaTrajFollow **************/
                                    /**************************************************/

ManaStateT lastStateToReach_;                     /**< variable to stock the last state we tried to reach */

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 *  Should not be used
 */
ManaTrajFollow::ManaTrajFollow() : TrajFollow()
{
  pctraj_ = NULL;
  old_pctraj_ = NULL;
  trajConnected_ = false;
}
  
/**
 *  \brief init constructor
 *  \param traj : the trajectory to follow
 */
ManaTrajFollow::ManaTrajFollow(ManaSys* sys,ManaSTraj* traj) : TrajFollow(sys,traj)
{
  pctraj_ = NULL;
  old_pctraj_ = NULL;
  trajConnected_ = false;

  trajgen_ = new ManaTrajGen(sys);
  trajgen_->setMaxResidual(0.05 * this->sys()->getLength());

  trajgen()->fixFinalXY(true);
  trajgen()->fixFinalTH(true);
  trajgen()->fixFinalV(true);
  trajgen()->fixFinalOM(true);

//   trajgen()->setUseVelCst(false);
//   trajgen()->setUseAccCst(false);
//   trajgen()->setUseAVelCst(false);
//   trajgen()->setUseAAccCst(false);
  trajgen()->setUseVelCst(true);
  trajgen()->setUseAccCst(true);
  trajgen()->setUseAVelCst(true);
  trajgen()->setUseAAccCst(true);
}

/**
 *  \brief destructor
 */
ManaTrajFollow::~ManaTrajFollow()
{
  if(pctraj_ != NULL) delete pctraj_;
  if(old_pctraj_ != NULL) delete old_pctraj_;
}


/*************** Getters and Setters **************/

/**
 *  \brief return the last state-time of the followed trajectory
 *  \return the last state-time of the followed trajectory
 */
ManaStateT ManaTrajFollow::goalStateT() const
{
  if(this->traj_ != NULL) return this->traj()->lastST();
  else
    {
      cerr << "ManaTrajFollow::goalStateT() : No followed trajectory has been defined. Cannot return its final state-time." << endl;
      return ManaStateT();
    }
}


/*************** Localization update **************/
  
/**
 *  \brief selection of the state of the trajectory to reach
 */
void ManaTrajFollow::selectStateToReach(const ManaStateT& curS,ManaStateT& goalS)
{
  // get back the time of the current state
  double curT = this->lastTimeUp();

  // compute the time of the state to reach
  double goalT;
  if(this->horizonMode() == TF_TIME_HORIZON_MODE)
    {
      goalT = curT + this->dtReach();

      boost::mutex::scoped_lock lock(mutexTraj_);

      // get back the state of the trajectory at time goalT
      ((ManaSTraj*)traj_)->getStateAtTime(goalT,goalS);
    }
  else // if(this->horizonMode() == TF_DIST_HORIZON_MODE)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);

      // get back the state of the trajectory given from curS after traveling a distance distReach
      ((ManaSTraj*)traj_)->getStateAtDist(curS.t(), this->distReach(), goalS);
   }

  // if the time is too short to generate the trajectory, set it to a minimal time
  if(goalS.t() - curT < this->lastDuration_) goalS.t() = (curT + this->lastDuration_);

  // compute the distance between the init position and the goal position
  double dx = goalS.x() - curS.x();
  double dy = goalS.y() - curS.y();
  double dist = sqrt(dx*dx + dy*dy);
  // if the distance between the current position and the goal position is too great,
  // set the time to the minimal time needed to travel this distance
  double minTravT = minTravelTime(dist,curS.v(),((ManaSys*)system_)->getVMax(),((ManaSys*)system_)->getAMax());
  if(goalS.t() - curT < minTravT)
    {
//       cerr << "ManaTrajFollow::selectStateToReach : time too short" << endl;
      goalS.t() = (curT + minTravT);
    }
  else
    {
      // if the distance between the current position and the goal position is too short,
      // set the time to the maximal time needed to travel this distance                                  ------------------------------------   does it really works? 
      double maxTravT = maxTravelTime(dist,curS.s(),((WheelChairSys*)system_)->getVMax(),((WheelChairSys*)system_)->getAMax());
      if(goalS.t() - curT > maxTravT)
	{
//  	  cerr << "WCTrajFollow::selectStateToReach : time too great" << endl;
	  goalS.t() = (curT + maxTravT);
	}
    }
  
  // stock the goal state-time computed
  this->lastStateToReach_ = goalS;
  this->lastStateToReach_.t() += curS.t();

}

// /**
//  *  \brief Update the localization of the robot
//  *  \param curS : current state-time updated
//  *  \return 0 if the trajectory generation successed, -1 else
//  */
// int ManaTrajFollow::updateLoc(const ManaStateT& curS)
// {
//   // update the new state reached
//   if(this->curRegST_ != NULL) delete this->curRegST_;
//   this->curRegST_ = new ManaStateT(curS);

//   // select the goal state on the trajectory we want to reach
//   ManaStateT goalS;
//   this->selectStateToReach(curS,goalS);

//   // save the previously computed trajectory
//   if(old_pctraj_ != NULL) delete old_pctraj_;
//   if(pctraj_ != NULL) 
//     {
//       old_pctraj_ = pctraj_;
//       pctraj_ = NULL;
//     }

//   // generate the trajectory to the goal
//   int res = ((ManaTrajGen*)trajgen_)->genTraj(curS,goalS);

//   if(res == 0)
//     {
//       // save the resulting control trajectory
//       pctraj_ = ((ManaTrajGen*)trajgen_)->getPCTraj();

//       trajConnected_ = true;
//     }
//   else
//     {
//       // What to do it the trajectory generation failed ? 
//       // Problem to think about quickly...
//       // Should keep the best result : So nothing to do?

//       // check if a connected trajectory was found at the previous step
//       if(trajConnected_)
// 	{
// 	  // keep the previous connected trajectory
// 	  pctraj_ = old_pctraj_;
// 	  old_pctraj_ = NULL;
// 	}
//       else
// 	{
// 	  // use the new generated trajectory even if it didn't reach the goal
// 	  pctraj_ = ((ManaTrajGen*)trajgen_)->getPCTraj();

// 	  trajConnected_ = false;
// 	}
//     }

//   return res;
// }

/**
 *  \brief Update the localization of the robot
 *  \param curS : current state-time updated
 *  \return 0 if the trajectory generation successed, -1 else
 */
int ManaTrajFollow::updateLoc(const ManaStateT& curS)
{
  // update the new state reached
  if(this->curRegST_ != NULL) delete this->curRegST_;
  this->curRegST_ = new ManaStateT(curS);

  // save the previously computed trajectory
  if(old_pctraj_ != NULL) delete old_pctraj_;
  if(pctraj_ != NULL) 
    {
      old_pctraj_ = pctraj_;
      pctraj_ = NULL;
    }

  // check if the robot can move (if it is not too close to obstacles)
  if(tooCloseToObs_)
    {
      // set a null control trajectory as input of the robot
      pctraj_ = new ManaPCTraj(0.0, 0.0, 999999999.9);

      return 0;
    }

  // select the goal state on the trajectory we want to reach
  ManaStateT goalS;
  this->selectStateToReach(curS,goalS);

  // generate the trajectory to the goal
  int res = ((ManaTrajGen*)trajgen_)->genTraj(curS,goalS);

  if(res == 0)
    {
      // save the resulting control trajectory
      pctraj_ = ((ManaTrajGen*)trajgen_)->getPCTraj();

      trajConnected_ = true;
    }
  else
    {
      // What to do it the trajectory generation failed ? 
      // Problem to think about quickly...
      // Should keep the best result : So nothing to do?

      // save the resulting control trajectory
      pctraj_ = ((ManaTrajGen*)trajgen_)->getPCTraj();

      trajConnected_ = false;
    }

  return res;
}

/***************** Control request ****************/

/**
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param ctrl : resulting control
 */
void ManaTrajFollow::getControl(double t,ManaControl& ctrl)
{
  double dt = t - this->lastTimeUp();
  
  // get back the acceleration control to apply at time t
  double vel = 0; 
  double avel = 0;
  if(pctraj_ != NULL)
    {
      vel = pctraj_->v(dt);
      avel = pctraj_->w(dt);
    }

  // check if these controls respect the motion constraints
  double maxVel = this->sys()->getVMax();
  double maxAVel = this->sys()->getAVMax();
  if(vel > maxVel) vel = maxVel;
  else if(vel < -maxVel) vel = -maxVel;
  if(avel > maxAVel) avel = maxAVel;
  else if(avel < -maxAVel) avel = -maxAVel;

  ctrl.v() = vel;
  ctrl.w() = avel;
}

/**
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 */
void ManaTrajFollow::getControl(double t,double dt,ManaControl& ctrl)
{
  // A linearization over both initial and final values is 
  // the straightforward solution, over the function may be 
  // uncontinuous... What to do in that case?
  // Ensure bound over control derivatives?
  // Simple method currently used
  
//   if(t > traj_->lastST().t())
  if(t > this->lastTimeTraj())
    {
      ctrl.v() = 0;
      ctrl.w() = 0;
      endTraj_ = true;
    }
  else
    {
      this->lastDuration_ = dt;
      
      ManaControl ctrl1,ctrl2;
      this->getControl(t,ctrl1);
      this->getControl(t+dt,ctrl2);
      
//       ctrl.v() = 0.5*(ctrl1.v() + ctrl2.v());
//       ctrl.w() = 0.5*(ctrl1.w() + ctrl2.w());

      ctrl.v() = 0.0*ctrl1.v() + 1.0*ctrl2.v();
      ctrl.w() = 0.0*ctrl1.w() + 1.0*ctrl2.w();
    }
}

/**
 *  \brief Request for a new control to apply (at a specified time)
 *  \param curS : current state-time updated
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 *  \return 0 if the trajectory generation succeeded, -1 else
 *
 *  Synchronous localization update and control request
 */
int ManaTrajFollow::getControl(const ManaStateT& curS,double t,double dt,ManaControl& ctrl)
{
  // update the localization
  int res = this->updateLoc(curS);

  // return the control to apply
  this->getControl(t,dt,ctrl);

  return res;
}

/**
 *  \brief Request for a new control to apply (at a specified time)
 *  \param curS : current state-time updated
 *  \param ctrl : resulting control
 *  \return 0 if the trajectory generation succeeded, -1 else
 *
 *  Synchronous localization update and control request
 */
int ManaTrajFollow::getControl(const ManaState& curS,ManaControl& ctrl)
{
  // check if the trajectory following process has been started
  if(this->isRunning())
    {
      // get back teh current time
      double t = timer_.totalElapsedTime();

      // update the localization
      int res = this->updateLoc(ManaStateT(curS,t));
      
      // return the control to apply
      this->getControl(t,0.5,ctrl);

      return res;
    }
  else return -1;
}

/**
 *  \brief return a control trajectory to apply on the robot
 *  \param curS : current state-time updated
 *  \return resulting control trajectory
 *
 *  Synchronous localization update and control request
 */
ManaPCTraj* ManaTrajFollow::getCTraj(const ManaState& curS)
{
  // check if the trajectory following process has been started
  if(this->isRunning())
    {
      // get back the current time
      double t = timer_.totalElapsedTime();

      // update the localization
      int res = this->updateLoc(ManaStateT(curS,t));

      // check if the end time or the goal have been reached
//       if(t > traj_->lastST().t()) 
      if(t > this->lastTimeTraj() || this->goalReached(ManaStateT(curS,t), this->goalStateT()))
	endTraj_ = true;

      // get back the computed control trajectory
      if(pctraj_ == NULL) return NULL;
      else
	{
	  ManaPCTraj* resPCTraj = new ManaPCTraj(*pctraj_);

	  // get back the minimum time to break
	  float time_dec = min(this->minBreakTime(), this->dtReach());

	  if(t > this->lastTimeTraj() - TF_COEF_TIME_BREAK*time_dec)
	    {
	      float coef_time_break;
	      if(t > this->lastTimeTraj()) coef_time_break = 0;
	      else coef_time_break = (this->lastTimeTraj() - t) / (TF_COEF_TIME_BREAK*this->minBreakTime());

	      // add additional bounds on the control force the robot to slow at the end of the trajectory
// 	      resPCTraj->addAccBounds(-sys()->getAMax(), coef_time_break*sys()->getAMax());
// 	      resPCTraj->addAAccBounds(-coef_time_break*sys()->getAAMax(), coef_time_break*sys()->getAAMax());
	      resPCTraj->addVelBounds(0.0, coef_time_break*sys()->getVMax());
	      resPCTraj->addAVelBounds(-coef_time_break*sys()->getAVMax(), coef_time_break*sys()->getAVMax());
	    }

	  return resPCTraj;
	}
  }
  else return NULL;
}


/**************** Trajectory update ***************/

/*
 *  \brief update the trajectory wrt. the maximal allowed speed
 *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
 */
void ManaTrajFollow::updateTrajVelProfile(float coefMaxVel)
{
  // check if a localization has already been registered
  if(this->curRegST() == NULL) return;

  DD2_PVDTrajPlanner planner((DiffDrive2Sys*)system_);

  float maxVel = coefMaxVel * sys()->getVMax();
  float startTime = this->curRegST()->t();
  float startVel = this->curRegST()->v();
  float endTime = startTime + TF_COEF_TRAJ_UPDATE * this->minBreakTime();
//   cout << "ManaTrajFollow::updateTrajVelProfile -> maxVel : " << maxVel << "  startTime : " << startTime << "  startVel : " << startVel << "  endTime : " << endTime << endl;

  // replan a part of the trajectory
  boost::mutex::scoped_lock lock(mutexTraj_);
  int resReplanning = planner.replanTraj((DD2STraj*)traj_, maxVel, startTime, startVel, endTime);

  return;
}


/**************** Utility functions ***************/

/*
 *  \brief compute the minimal breaking distance of the robot wrt. its current state
 *  \return the minimal breaking distance of the robot
 */
float ManaTrajFollow::minBreakDist() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->v();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)dres);
}

/*
 *  \brief compute the minimal breaking time needed by the robot wrt. its current state
 *  \return the minimal breaking time needed by the robot
 */
float ManaTrajFollow::minBreakTime() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->v();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)tres);
}

/*
 *  \brief return true if the goal has been reached
 *  \param curStateT : the current state-time of the robot
 *  \param goalStateT : the goal state-time to reach
 *  \return true if the goal has been reached
 */
bool ManaTrajFollow::goalReached(const ManaStateT& curStateT, const ManaStateT& goalStateT) const
{
  if(curStateT.t() > goalStateT.t() - dtReach())
    {
      // compute the distance between the current and goal positions
      float dx = curStateT.x() - goalStateT.x();
      float dy = curStateT.y() - goalStateT.y();
      float sqDist = dx*dx + dy*dy;
      
      float goalToCur_head = atan2(dy,dx);
      double diffAngle = fabs(computeMinDiffAngle(goalToCur_head, goalStateT.th()));

      if(sqDist < TrajFollow::maxDistTgtReached_ * TrajFollow::maxDistTgtReached_)
	{
	  return true;
	}
      else if(sqDist < TrajFollow::maxDistWpReached_ * TrajFollow::maxDistWpReached_ && diffAngle < TrajFollow::maxAngleWpReached_)
	{
	  return true;
	}
      else return false;
    }
  else return false;
}


/**************** Display functions ***************/

/**
 *  \brief display the trajectory computed by the trajectory generator
 */
void ManaTrajFollow::displayTrajGen()
{
  cerr << "ManaTrajFollow::displayTrajGen : Not implemented yet!" << endl;
  exit(1);
}
