#include "WCTrajFollow.hpp"
#include "UtilsTG.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

                                    /**************************************************/
                                    /*************** Class CLTrajFollow ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 *  Should not be used
 */
WCTrajFollow::WCTrajFollow() : TrajFollow()
{
}

/*
 *  \brief init constructor
 *  \param traj : the trajectory to follow
 */
WCTrajFollow::WCTrajFollow(WheelChairSys* sys,WCSTraj* traj) : TrajFollow(sys,traj)
{
  trajgen_ = new WCTrajGen(sys);
  trajgen_->setMaxResidual(0.05 * this->sys()->getLength());

  trajgen()->fixFinalXY(true);
  trajgen()->fixFinalTH(true);
  trajgen()->fixFinalV(true);
  trajgen()->fixFinalOM(true);

  trajgen()->setUseVelCst(false);
  trajgen()->setUseAccCst(false);
  trajgen()->setUseAVelCst(false);
  trajgen()->setUseAAccCst(false);
}

/*
 *  \brief destructor
 */
WCTrajFollow::~WCTrajFollow()
{
}


/*************** Localization update **************/


/*
 *  \brief selection of the state of the trajectory to reach
 */
void WCTrajFollow::selectStateToReach(const WCStateT& curS,WCStateT& goalS)
{
  // get back the time of the current state
  double curT = this->lastTimeUp();

  // compute the time of the state to reach
  double goalT = curT + this->dtReach();

  // check if the trajectory must already be reached at goalT
  if(goalT <= this->timeMeeting()) goalT = this->timeMeeting();

  // get back the state of the trajectory at time t
  ((WCSTraj*)traj_)->getStateAtTime(goalT,goalS);

  // if the time is too short to generate the trajectory, set it to a minimal time
  if(goalS.t() - curT < this->lastDuration_) goalS.t() = (curT + this->lastDuration_);

//   // stock the goal state-time computed
//   this->lastStateToReach_ = goalS;

  // compute the distance between the init position and the goal position
  double dx = goalS.x() - curS.x();
  double dy = goalS.y() - curS.y();
  double dist = sqrt(dx*dx + dy*dy);
  // if the distance between the current position and the goal position is too great,
  // set the time to the minimal time needed to travel this distance
  double minTravT = minTravelTime(dist,curS.s(),((WheelChairSys*)system_)->getVMax(),((WheelChairSys*)system_)->getAMax());
  if(goalS.t() - curT < minTravT)
    {
//       cerr << "WCTrajFollow::selectStateToReach : time too short" << endl;
      goalS.t() = (curT + minTravT);
    }
  else
    {
//       cerr << "WCTrajFollow::selectStateToReach : time OK" << endl;
      // if the distance between the current position and the goal position is too short,
      // set the time to the maximal time needed to travel this distance
      double maxTravT = maxTravelTime(dist,curS.s(),((WheelChairSys*)system_)->getVMax(),((WheelChairSys*)system_)->getAMax());
      if(goalS.t() - curT > maxTravT)
	{
//  	  cerr << "WCTrajFollow::selectStateToReach : time too great" << endl;
	  goalS.t() = (curT + maxTravT);
	}
    }
  
  // stock the goal state-time computed
  this->lastStateToReach_ = goalS;
  this->lastStateToReach_.t() += curS.t();
}

/*
 *  \brief Update the localization of the robot
 *  \param curS : current state-time updated
 */
int WCTrajFollow::updateLoc(const WCStateT& curS)
{
  // update the new state reached
  if(this->curRegST_ != NULL) delete this->curRegST_;
  this->curRegST_ = new WCStateT(curS);

  // select the goal state on the trajectory we want to reach
  WCStateT goalS;
  this->selectStateToReach(curS,goalS);

//   int res = ((WCTrajGen*)trajgen_)->ComputeTrajectory(curS,goalS,(double**)NULL);
  int res = ((WCTrajGen*)trajgen_)->genTraj(curS,goalS);
  
  if(res != 0)
    {
      // What to do it the trajectory generation failed ? 
      // Problem to think about quickly...
      // Should keep the best result : So nothing to do?
    }

  return res;
}


/***************** Control request ****************/

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param ctrl : resulting control
 */
void WCTrajFollow::getControl(double t,WCControl& ctrl)
{
  double dt = t - this->lastTimeUp();
  
  // get back the acceleration control to apply at time t
  double vel = ((WCTrajGen*)trajgen_)->funcVel(dt);
  double avel = ((WCTrajGen*)trajgen_)->funcAVel(dt);

  // check if these controls respect the motion constraints
  double maxVel = this->sys()->getVMax();
  double maxAVel = this->sys()->getAVMax();
  if(vel > maxVel) vel = maxVel;
  else if(vel < -maxVel) vel = -maxVel;
  if(avel > maxAVel) avel = maxAVel;
  else if(avel < -maxAVel) avel = -maxAVel;

  ctrl = WCControl(vel,avel);
}

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 */
void WCTrajFollow::getControl(double t,double dt,WCControl& ctrl)
{
  // A linearization over both initial and final values is 
  // the straightforward solution, over the function may be 
  // uncontinuous... What to do in that case?
  // Ensure bound over control derivatives?
  // Simple method currently used
  
  // compute the distance between the current state and the final state of the trajectory
  double dx = traj()->lastST().x() - curRegST()->x();
  double dy = traj()->lastST().y() - curRegST()->y();
  double dist = sqrt(dx*dx + dy*dy);
  
  // check if the final time of the trajectory has been already reached
  //cout << "t : " << t << "  traj_->lastST().t() : " << traj_->lastST().t() << endl;
  
  //if(t > traj_->lastST().t() && dist < 0.2 * this->sys()->getLength())
  if(t > traj_->lastST().t())
    {
      ctrl = WCControl(0,0);
      endTraj_ = true;
    }
  else
    {
      this->lastDuration_ = dt;
      
      WCControl ctrl1,ctrl2;
      this->getControl(t,ctrl1);
      this->getControl(t+dt,ctrl2);
      
      ctrl = WCControl(0.5*(ctrl1.v() + ctrl2.v()),
		       0.5*(ctrl1.av() + ctrl2.av()));
    }
}

/*
 *  \brief Request for a new control to apply (at a specified time)
 *  \param curS : current state-time updated
 *  \param t : time at which is evaluated the control
 *  \param dt : duration of the application of the control
 *  \param ctrl : resulting control
 *
 *  Synchronous localization update and control request
 */
int WCTrajFollow::getControl(const WCStateT& curS,double t,double dt,WCControl& ctrl)
{
  // update the localization
  int res = this->updateLoc(curS);

  // return the control to apply
  this->getControl(t,dt,ctrl);

  return res;
}


/**************** Trajectory update ***************/

/*
 *  \brief update the trajectory wrt. the maximal allowed speed
 *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
 */
void WCTrajFollow::updateTrajVelProfile(float coefMaxVel)
{
//   // check if a localization has already been registered
//   if(this->curRegST() == NULL) return;

//   CL_PVDTrajPlanner planner((CarLikeSys*)system_);

//   float maxVel = coefMaxVel * sys()->getVMax();
//   float startTime = this->curRegST()->t();
//   float startVel = this->curRegST()->v();
//   float endTime = startTime + COEF_TRAJ_UPDATE * this->minBreakTime();
// //   cout << "CLTrajFollow::updateTrajVelProfile -> maxVel : " << maxVel << "  startTime : " << startTime << "  startVel : " << startVel << "  endTime : " << endTime << endl;

//   // replan a part of the trajectory
//   boost::mutex::scoped_lock lock(mutexTraj_);
//   int resReplanning = planner.replanTraj((CLSTraj*)traj_, maxVel, startTime, startVel, endTime);

//   return;

  cerr << "WCTrajFollow::updateTrajVelProfile : Not implemented yet!" << endl;
  exit(1);
}


/**************** Utility functions ***************/

/*
 *  \brief compute the minimal breaking distance of the robot wrt. its current state
 *  \return the minimal breaking distance of the robot
 */
float WCTrajFollow::minBreakDist() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->s();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)dres);
}

/*
 *  \brief compute the minimal breaking time needed by the robot wrt. its current state
 *  \return the minimal breaking time needed by the robot
 */
float WCTrajFollow::minBreakTime() const
{
  double tres;
  double dres;

  double curVel = this->curRegST()->s();
  double maxVel = this->sys()->getVMax();
  double maxAcc = this->sys()->getAMax();

  minTDtoReachSpd(curVel, 0, maxAcc, tres, dres);

  return ((float)tres);
}

/*
 *  \brief get back the last time of the trajectory
 *  \return the last time of the trajectory
 */
double WCTrajFollow::lastTimeTraj()
{
  boost::mutex::scoped_lock lock(mutexTraj_);

  return traj_->lastST().t();
}


/**************** Display functions ***************/

/*
 *  \brief display the trajectory computed by the trajectory generator
 */
void WCTrajFollow::displayTrajGen()
{
  int N = 30;
  glLineWidth(((GLfloat)(2.0)));

  // check if the trajectory generator is available
  if(trajgen_ != NULL)
    {
      // sample the trajectory
      trajgen()->SaveTrajectory(N);

      glBegin(GL_LINE_STRIP);
      for(int i=0; i<N; i++)
	{
	  double X = ((WCTrajGen*)trajgen_)->posX->at(i);
	  double Y = ((WCTrajGen*)trajgen_)->posY->at(i);
	  double T;
	  if(curRegST_ != NULL) T = ((WCTrajGen*)trajgen_)->timeT->at(i) + curRegST_->t();
	  else T = ((WCTrajGen*)trajgen_)->timeT->at(i);

	  glVertex3f(X,Y,T);
	}
      glEnd();
    }
}
