#ifndef MPC_DDTRAJGEN_HPP
#define MPC_DDTRAJGEN_HPP

#include "trajGen/inc/TrajGen.hpp"
#include "math/inc/MultiItvFunction.hpp"
#include "system/inc/DiffDriveSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class DDTrajGen ****************/
                                    /**************************************************/

  /** \class DDTrajGen
   *  \brief Trajectory generator TIJI for a differential drive system
   */
  class DDTrajGen : public TrajGen
  {
  public :
    double maxWAcc;                                    /**< maximal angular acceleration of the wheels */
    double maxWVel;                                    /**< maximal angular velocity of the wheels */

    MultiItvFunction* fctWLAcc_;                       /**< angular acceleration profile of the left wheel */
    MultiItvFunction* fctWRAcc_;                       /**< angular acceleration profile of the right wheel */
    MultiItvFunction* fctWLVel_;                       /**< angular velocity profile of the left wheel */
    MultiItvFunction* fctWRVel_;                       /**< angular velocity profile of the right wheel */

    MultiItvFunction* fctVel_;                         /**< linear velocity profile of the robot */
    MultiItvFunction* fctAVel_;                        /**< angular velocity profile of the robot */
    MultiItvFunction* fctTH_;                          /**< heading profile of the robot */

    Polynomial* fctWLAccWB_;                           /**< angular acceleration profile of the left wheel */
    Polynomial* fctWRAccWB_;                           /**< angular acceleration profile of the right wheel */
    Polynomial* fctWLVelWB_;                           /**< angular velocity profile of the left wheel */
    Polynomial* fctWRVelWB_;                           /**< angular velocity profile of the right wheel */

    Polynomial* fctVelWB_;                             /**< linear velocity profile of the robot */
    Polynomial* fctAVelWB_;                            /**< angular velocity profile of the robot */
    Polynomial* fctTHWB_;                              /**< heading profile of the robot */

    //   int lastItvAccBdd_;                                // last interval of the acceleration, linear
    //   int lastItvAVelBdd_;                               // -1 = minB,  0 = not bounded,  1 = maxB
    //   int lastItvAAccBdd_;

    double A1_;
    double A2_;
    double R_;
    double UnSurR_;

  public : 
    double finalErrX_;                                 /**< final error over X found */
    double finalErrY_;                                 /**< final error over Y found */
    double finalErrTH_;                                /**< final error over TH found */

    vector<double> *posX;                              /**< vector containing a sampling of the position X */
    vector<double> *posY;                              /**< vector containing a sampling of the position Y */
    vector<double> *headTH;                            /**< vector containing a sampling of the heading TH */
    vector<double> *velV;                              /**< vector containing a sampling of the linear velocity V */
    vector<double> *avelOM;                            /**< vector containing a sampling of the angular velocity OM */
    vector<double> *timeT;                             /**< vector containing a sampling of the time T */

    /********** Constructors and destructors **********/
  
    /**
     *  \brief default constructor
     */  
    DDTrajGen();

    /**
     *  \brief destructor
     */  
    virtual ~DDTrajGen();

    /*************** Getters and Setters **************/


    /********* Optimization related routines **********/

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void ComputeIG(Vec_DP** parameters);

    /**
     *  \brief compute the error between the state reached and the state we want to reach
     */
    virtual void ComputeG();

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
     */
    void ComputeJacobianG_FD();

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeJacobianG();

    double fct1(const double t) const;
    double fct2(const double t) const;

    double itgdXdB1(const double t) const;
    double itgdXdC1(const double t) const;
    double itgdXdD1(const double t) const;
    double itgdXdB2(const double t) const;
    double itgdXdC2(const double t) const;
    double itgdXdD2(const double t) const;
    double itgdYdB1(const double t) const;
    double itgdYdC1(const double t) const;
    double itgdYdD1(const double t) const;
    double itgdYdB2(const double t) const;
    double itgdYdC2(const double t) const;
    double itgdYdD2(const double t) const;

    /**
     *  \brief compute the residual between the goal state and the stae reached
     *  \return the residual between the goal state and the stae reached
     */
    virtual double ComputeResidual();

    /**************** Quadrature tools ****************/
  
    /**
     *  \brief Quadrature tools...
     */
    double qgaus(double (DDTrajGen::*func)(const double) const, const double a, const double b) const;

    /*********** State profiles computations **********/

    /**
     *  \brief compute the angular acceleration profile of the left wheel
     */
    void computeFctWLAcc();

    /**
     *  \brief compute the angular acceleration profile of the right wheel
     */
    void computeFctWRAcc();

    /**
     *  \brief compute the angular velocity profile of the left wheel
     */
    void computeFctWLVel();

    /**
     *  \brief compute the angular velocity profile of the right wheel
     */
    void computeFctWRVel();

    /**
     *  \brief compute the linear velocity profile of the robot
     */
    void computeFctVel();

    /**
     *  \brief compute the angular velocity profile of the robot
     */
    void computeFctAVel();

    /**
     *  \brief compute the heading profile of the robot
     */
    void computeFctTH();

    /**
     *  \brief compute all profiles
     */
    void computeAllFct();

  
    /********** State Parameters Computation **********/  

    /**
     *  \brief get the value of the angular acceleration profile of the left wheel at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular acceleration profile of the left wheel at time t
     */
    double funcWLAcc(const double t) const;

    /**
     *  \brief get the value of the angular acceleration profile of the right wheel at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular acceleration profile of the right wheel at time t
     */
    double funcWRAcc(const double t) const;

    /**
     *  \brief get the value of the angular velocity profile of the left wheel at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular velocity profile of the left wheel at time t
     */
    double funcWLVel(const double t) const;

    /**
     *  \brief get the value of the angular velocity profile of the right wheel at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular velocity profile of the right wheel at time t
     */
    double funcWRVel(const double t) const;

    /**
     *  \brief get the value of the linear velocity profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the linear velocity profile of the robot at time t
     */
    double funcVel(const double t) const;

    /**
     *  \brief get the value of the angular velocity profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular velocity profile of the robot at time t
     */
    double funcAVel(const double t) const;

    /**
     *  \brief get the value of the heading profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the heading profile of the robot at time t
     */
    double funcTheta(const double t) const;

    /**
     *  \brief integration function for the position X profile computation
     *  \param t : the time at which the function is evaluated
     *  \return the value at time t of the integration function
     */  
    double funcItgX(const double t) const;

    /**
     *  \brief integration function for the position Y profile computation
     *  \param t : the time at which the function is evaluated
     *  \return the value at time t of the integration function
     */  
    double funcItgY(const double t) const;

    /**
     *  \brief get the value of the position X profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the position X profile at time t
     */
    double funcX(const double t) const;

    /**
     *  \brief get the value of the position Y profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the position Y profile at time t
     */
    double funcY(const double t) const;

    /*************** Forward Solution *****************/
  
    /**
     *  \brief compute the state reached by the current set of parameters
     */
    virtual void ForwardSolution();

    /***************** Initialization *****************/

    /**
     *  \brief set the initial guess of parameters
     *  \param b,c,d : coefficients of the polynomial right wheel acceleration profile
     *  \param b2,c2,d2 : coefficients of the polynomial left wheel acceleration profile
     *  \param t : final time to compute the trajectory (time of the final state)
     */
    void setInitialGuess(double b1,double c1,double d1,double b2,double c2,double d2,double t);

    /**
     *  \brief set the initial state, final state and constraints vectors
     *  \param x0 : initial state
     *  \param xf : final state
     *  \param constraints : constraints defined over the system
     *
     *  constraints[0] : max angular acceleration of the wheels
     *  constraints[1] : max linear acceleration of the wheels
     *  constraints[2] : radius of the robot
     */
    virtual void setInitialParam(Vec_DP& x0,Vec_DP& xf,Vec_DP& constraints);

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void setInitialGuess(double* parameters);

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void setInitialConfig(double* x0,double* constraints);

    /************* Trajectory Generation **************/

    /**
     *  \brief init a vector of the constraints defined over the system
     *  \param sys : the robotic system features
     *  \param vcstt : the resulting vector of constraints
     */
    virtual void initCstt(System* sys,Vec_DP& vcstt);

    //   /**
    //    *  \brief generate a trajectory by constraints optimiation
    //    *  \param x0 : initial state
    //    *  \param xf : final state
    //    *  \param constraints : constraints defined over the system
    //    *  \param parameters : initial guess of parameters
    //    *
    //    *  x0 = (X0,Y0,TH0,V0,OM0,T0)
    //    *  xf = (Xf,Yf,THf,Vf,OMf,Tf)
    //    *  constraints = (maxWAcc,maxWVel,R)
    //    */
    //   virtual int ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP constraints,Vec_DP** parameters);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sys : constraints defined over the system
     *  \param param : initial guess of parameters
     */
    inline virtual void ComputeFromIG(const StateTime& s0,const System* sys,double* param) 
    { this->ComputeFromIG(*((DDStateT*)(&s0)),(DiffDriveSys*)sys,param); }

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param statet : initial state-time
     *  \param sys : constraints defined over the robotic system
     *  \param ctrlt : control-time applied on the system
     */
    virtual void ComputeFromCtrl(const StateTime& statet,const System* sys,const ControlTime& ctrlt);

    /**************** Output Functions ****************/

    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (w,y,th,omr,oml,t)
     */
    virtual void getStateTReachedAt(double* rstatet,double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (w,y,th,omr,oml,t)
     */
    virtual void getFinalStateTReached(double* rstatet) const;

    /**
     *  \brief save a sampling of the trajectory profile
     *  \param nbp : number of sampling points
     */
    virtual void SaveTrajectory(int nbp);

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     */
    virtual void traceOn(Map *map,int nbp);

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     *  \param initPos : initial position
     */  
    virtual void traceOn(Map *map,int nbp,const PositionXYTh& initPos);

    /**
     *  \brief return a sample trajectory from the parameterized one
     *  \param ts : the sampling time step
     *  \return a sample trajectory from the parameterized one
     *
     *  A new instance of sample trajectory is returned here
     */
    virtual DDSTraj* sampledTraj(double ts) const;

    /*************** Utility Functions ****************/

    /**
     *  \brief display both wheels velocity, linear and angular velocity, heading and position profiles
     */
    void displayProfiles();

    /**
     *  \brief display all functions
     */
    virtual void displayAllFct();
  };

};
#endif // MPC_DDTRAJGEN_HPP

