#ifndef MPC_INITIALGUESS
#define MPC_INITIALGUESS

#include "trajGen/inc/TrajGen.hpp"

#include "environment/inc/Map.hpp"
#include "common/inc/Position.hpp"
#include "system/inc/System.hpp"

#include <list>

#define RAND_FIRSTIG 0
#define SC_FIRSTIG 1
#define SCP_FIRSTIG 2
#define RSTAR_FIRSTIG 3
#define LOADFROMFILE_FIRSTIG 4

namespace mpc
{
  class TrajGen;

  /**
   * \enum system_t
   * \brief Enumeration of the different system types
   *
   * ND_N2R         : undefined N2R
   * IG_N2R         : N2R defined during the initial step
   * EXPLORED_N2R   : explored N2R
   * REACHED_N2R    : reached N2R
   * USED_N2R       : reached N2R whose neighbours are already reached too
   */
  typedef enum {
    ND_N2R             =  0,
    IG_N2R             =  1,
    EXPLORED_N2R       =  2,
    REACHED_N2R        =  3,
    USED_N2R           =  4
  } n2r_t;


                                    /**************************************************/
                                    /******************** Class N2R *******************/
                                    /**************************************************/

  /** \class N2R
   *  \brief Class representing a node to reach during the learning process of the initial guess
   */
  class N2R
  {
  protected : 
    int dimP_;                                     /**< dimension of the parameters vector */
    double* param_;                                /**< initial guess of parameters */
    StateTime* initST_;                            /**< initial state-time */
    StateTime* goalST_;                            /**< state-time reached from a position (x,y,th) = (0,0,0) */

    int* pos_;                                     /**< position of the node on the grid */
    std::list<N2R*>::iterator it_;                 /**< iterator allowing to access directly to the position of the node in the explored or reached list */

    //   std::list<N2R*>* lnear_;                       /**< list of candidate nodes to use the IG parameters from */

    n2r_t status_;                                 /**< status of the N2R during the learning process (not defined, explored, reached) */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    N2R();

  public : 
    /**
     *  \brief destructor
     */
    virtual ~N2R();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the dimension of the parametric control vector
     *  \return the dimension of the parametric control vector
     */
    inline int dimP() const { return this->dimP_; }
  
    /**
     *  \bref return the vector of parameters computed
     *  \return the vector of parameters computed or NULL if no one has been defined now
     */
    inline const double* params() const { return this->param_; }

    /**
     *  \brief return the status of the current node
     *  \return the status of the current node
     */
    inline n2r_t status() const { return this->status_; }

    /**
     *  \brief return the i-th parameter
     *  \param i : number of the parameter we want to get back
     *  \return i-th parameter
     */
    double param(int i) const;

    /**
     *  \brief return the final state-time reached
     */
    virtual const StateTime& initST() const = 0;

    /**
     *  \brief return the final state-time reached
     */
    virtual const StateTime& goalST() const = 0;

    /**
     *  \brief get back the position of the N2R in the associated grid
     */
    int* pos() const;

    //   /**
    //    *  \brief add a new candidate node to try to use the IG parameters from
    //    *  \param n : the new candidate node
    //    */
    //   void addNearNode(N2R* n);

    //   /**
    //    *  \brief clear the list of candiates nodes
    //    */
    //   void clearNearNodes();

    /**
     *  \brief set the current node as reached and define its parameters
     *  \param param : the parameterized control
     *  
     *  The node must have been set to explored before
     */
    void setAsReached(double* param);

    /**
     *  \brief set the current node as used
     *
     *  The node must have been set to reached before
     */
    void setAsUsed();

    /**
     *  \brief append the current node in a list
     *  \param it : the list in which the node must be appended
     */
    void appendInList(std::list<N2R*>* list);

    /**
     *  \brief remove the current node from the input list
     *  \param list : the list from which the node must be removed
     *
     *  The user must guarantee that the current node belongs to the input list
     */
    void removeFromList(std::list<N2R*>* list);

    //   /**
    //    *  \brief return a random neighbour of the current node
    //    *  \return a random neighbour of the current node if one exists, NULL else
    //    */
    //   virtual N2R* getRandNeighbour() const;

    //   /**
    //    *  \brief return the nearest neighbour of the current node
    //    *  \return the nearest neighbour of the current node if one exists, NULL else
    //    */
    //   virtual N2R* getNearNeighbour() const;

    /**************** Display functions ***************/

    /**
     *  \brief display the current node
     */
    void display() const;


    /**************** Utility functions ***************/

    /**
     *  \brief save a list of parameterized trajectories into an output file
     *  \param list : the list of trajectories to save
     *  \param filename : the path name of the output file
     *  \return 0 if the process succeed, -1 else
     *
     *  The features saved are the initial state-time, the final state-time and the parameterized control
     */
    static int saveListN2R(std::list<N2R*>* list,const char* filename);

    /**
     *  \brief return a weighted distance between the current node and the input node
     *  \param n2 : the second node
     *  \return a weighted distance between both nodes
     */
    virtual double computeWDistTo(N2R* n2) const = 0;
  };


                                    /**************************************************/
                                    /****************** Class N2RGrid *****************/
                                    /**************************************************/

  /** \class N2RGrid
   *  \brief Class representing a grid containing the nodes to reach during the learning process
   */
  class N2RGrid
  {
  protected : 
    N2R* *tabGrid_;                               /**< table representation of the grid */
    std::list<N2R*>* listN2R_;                    /**< list used to access quicky the nodes to 
						     reach at a step of the learning process */
    std::list<N2R*>* listNReached_;               /**< list used to access quicky the nodes already
						     reached at a step of the learning process */
    std::list<N2R*>* listNUsed_;                  /**< list used to access quicky the nodes already
						     used (reached and whose neighbours are already 
						     reached too) at a step of the learning process */

    int n_;                                       /**< dimension of the grid space */
    int* sampling_;                               /**< sampling of all dimensions of the grid */

    double* minSampS_;                            /**< minimal value of each state component */
    double* maxSampS_;                            /**< maximal value of each state component */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    N2RGrid();

  public : 
    /**
     *  \brief init constructor
     *  \param n : dimension of the grid space
     *  \param sampling : sampling of all dimensions of the grid
     *  \param minSampS : minimal values taken by the sampling
     *  \param maxSampS : maximal values taken by the sampling
     */
    N2RGrid(int n,int* sampling,double* minSampS,double* maxSampS);

    /**
     *  \brief destructor
     */
    virtual ~N2RGrid();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the dimension of the grid sampling vector
     */
    inline int dimS() const { return n_; }

    /**
     *  \brief return the size of the grid (computed from the sampling dimensions)
     *  \return the size of the grid
     */
    int size() const;

    //   /**
    //    *  \brief return the list of nodes to reach
    //    *  \return a pointer to the list of nodes to reach
    //    */
    //   inline std::list<N2R*>* listN2R() const { return this->listN2R_; }

    //   /**
    //    *  \brief return the list of reached nodes
    //    *  \return a pointer to the list of reached nodes
    //    */
    //   inline std::list<N2R*>* listNReached() const { return this->listNReached_; }

    //   /**
    //    *  \brief return the list of used nodes
    //    *  \return a pointer to the list of used nodes
    //    */
    //   inline std::list<N2R*>* listNUsed() const { return this->listNUsed_; }

    /**
     *  \brief return the list of nodes to reach
     *  \return a copy of the list of nodes to reach
     */
    std::list<N2R*> listN2R() const;

    /**
     *  \brief return the list of reached nodes
     *  \return a copy of the list of reached nodes
     */
    std::list<N2R*> listNReached() const;

    /**
     *  \brief return the list of used nodes
     *  \return a copy of the list of used nodes
     */
    std::list<N2R*> listNUsed() const;

    /********* Access/Modification of the grid ********/

  public : 
    /**
     *  \brief init the grid from the current sampling parameters
     */
    void initGrid();

    /**
     *  \brief clear the grid
     */
    void clearGrid();
  
    /**
     *  \brief return an element of the grid at coord
     *  \param coord : coordinates of the N2R
     *  \return the element of the grid at coord, NULL if the input coordinates are outside of the grid
     */
    N2R* gridAt(int* coord) const;

    /**
     *  \brief search for reachable nodes on the grid around a reached node (and mark them as reachable on the grid)
     *  \param node : the input already reached node
     *  \return a list containing the reachable nodes on the grid around the input node
     *
     *  An new instance of list is allocated and returned here
     */
    virtual std::list<N2R*>* reachableNodesAround(N2R* node) = 0;

    /**
     *  \brief search for already reached nodes on the grid around a node to reach
     *  \param node : the input node to reach
     *  \return a list containing the already reached nodes on the grid around the input node
     *
     *  An new instance of list is allocated and returned here
     */
    virtual std::list<N2R*>* reachedNodesAround(N2R* node) const = 0;

    /**
     *  \brief select randomly an already reached node around the input node
     *  \param node : the input node
     *  \param lfig : list of first initial guesses
     */
    N2R* selectReachedNodeAround(N2R* node,std::list<N2R*>* lfig=NULL) const;

    /**
     *  \brief update the list of reachable nodes by considering the nodes already reached
     */
    void updateN2RfromReached();

    /**
     *  \brief update the list of reachable nodes from the initial guess of trajectories
     *  \param lig : the list of nodes provided by the initial guess
     */
    void updateN2RfromIG(std::list<N2R*>& lig);

    /**
     *  \brief set the node at position coord as reached
     *  \param coord : the input position in the grid of the reached node
     *  \param param : the parameterized control
     *
     *  The vector of parameterized control will be copied
     */
    void setNodeAsReached(int* coord,double* param);

    /**
     *  \brief set the input node as used
     *  \param node : the input node
     */
    void setNodeAsUsed(N2R* node);

  protected : 
    /**
     *  \brief return a reference to an element of the grid at coord
     *  \param coord : coordinates of the N2R
     *  \return a reference to the element of the grid at coord
     */
    N2R*& gridAt(int* coord);  

    /**
     *  \brief create a N2R from its coordinates in the grid and a node from which it can be derivated
     *  \param coord : coordinates of the N2R
     *  \param node : node from which the N2R may be reachable
     *
     *  An new instance of N2R is allocated here
     *  Put directly the new node in the grid
     */
    virtual N2R* createExploredN2R(int* coord,N2R* node) = 0;

    /**
     *  \brief create a N2R from a trajectory already computed (its init and goal state-times and the parameterized control used)
     *  \param coord : coordinates of the N2R
     *  \param node : node from which the N2R may be reachable
     *
     *  An new instance of N2R is allocated here
     *  Both state-times and the control parameterization are copied here
     */
    virtual N2R* createIGN2R(const StateTime* initST,const StateTime* goalST,double* param) = 0;

  public : 
    /**
     *  \brief create a reached node from an input file thanks to its position and parameters
     *  \param pos : the position of the node in the grid
     *  \param params : the parameterized control defining the trajectory
     *  \return a pointer to the created node
     */
    virtual N2R* createReachedNode(int* coord,double* params) = 0;

    /**
     *  \brief note a N2R of the grid as possibly reachable from node
     *  \param coord : coordinates of the N2R
     *  \param node : node from which the N2R may be reachable
     *  \return a pointer to the N2R
     */
    N2R* addN2Rfrom(int* coord,N2R* node);

    /**
     *  \brief compute the number of nodes noted as "nodes to reach" on the grid
     *  \return the number of nodes noted as "nodes to reach" on the grid
     */
    virtual int N2RonGrid() const = 0;

    /**
     *  \brief compute the number of nodes noted as reached on the grid
     *  \return the number of nodes noted as reached on the grid
     */
    virtual int NReachedonGrid() const = 0;

    /**
     *  \brief compute the number of nodes noted as used on the grid
     *  \return the number of nodes noted as used on the grid
     */
    virtual int NUsedonGrid() const = 0;

    /**
     *  \brief compute the number of nodes noted as "nodes to reach" in the list
     *  \return the number of nodes noted as "nodes to reach" in the list
     */
    inline int N2RinList() const { return this->listN2R_->size(); }

    /**
     *  \brief compute the number of nodes noted as reached in the list
     *  \return the number of nodes noted as reached in the list
     */
    inline int NReachedinList() const { return this->listNReached_->size(); }

    /**
     *  \brief compute the number of nodes noted as used in the list
     *  \return the number of nodes noted as used in the list
     */
    inline int NUsedinList() const { return this->listNUsed_->size(); }


    /******** Access/Modification of the lists ********/

  protected : 
    /**
     *  \brief append the input node to the list of used nodes
     *  \param node : the input node
     */
    void appendNodeToLUsed(N2R* node);

    /**
     *  \brief append the input node to the list of reached nodes
     *  \param node : the input node
     */
    void appendNodeToLReached(N2R* node);

    /**
     *  \brief append the input node to the list of explored nodes (nodes to reach)
     *  \param node : the input node
     */
    void appendNodeToLN2R(N2R* node);

    /**
     *  \brief remove the input node from the list of reached nodes
     *  \param node : the input node
     */
    void removeNodeFromLReached(N2R* node);

    /**
     *  \brief remove the input node from the list of explored nodes (nodes to reach)
     *  \param node : the input node
     */
    void removeNodeFromLN2R(N2R* node);

  public : 
    /**
     *  \brief clear the list of nodes to reach
     */
    void clearListN2R();

    /**
     *  \brief display the list of explored nodes (nodes to reach)
     */
    void displayLN2R() const;

    /**
     *  \brief display the list of reached nodes
     */
    void displayLReached() const;

    /**
     *  \brief display the list of used nodes
     */
    void displayLUsed() const;

    /**
     *  \brief display the nodes set as reached on the grid
     */
    virtual void displayReachedOnGrid() const = 0;

    /**
     *  brief display the sampling of the grid
     */
    virtual void displayGridSampling() const = 0;
  };


                                    /**************************************************/
                                    /**************** Class InitialGuess **************/
                                    /**************************************************/

  /** \class InitialGuess
   *  \brief Class to compute and stock initial guess of parameters
   */
  class InitialGuess
  {
  protected : 
    int dimS_;                                    /**< dimension of the state-time vector */
    int dimP_;                                    /**< dimension of the parameter vector */

    N2RGrid* grid_;                               /**< the grid describing the sampling of the state-time space */

    int dimSampS_;                                /**< dimension of the sampling vector */

    int* sampling_;                               /**< sampling over each state component */
    double* minSampS_;                            /**< minimal value of each state component */
    double* maxSampS_;                            /**< maximal value of each state component */

    int firstIGMode_;                             /**< Mode to determine the first set of initial guesses 
						     (0=random, 1=sampled control, 2=sampled param control, 3=rstar, 4=load from file) */
    char* FIGfilename_;                           /**< Filename of the first IG parameters (used only by the FIG "load from file" mode) */

    TrajGen* trajGen_;                            /**< trajectory generator used to learn the base */
    System* sys_;                                 /**< the robotic system used */

    int minNNTSR_;                                /**< minimal number of new target state-times that must 
						     be reached to continue the learning process */

    int debug_;                                   /**< debugging mode (0=no debug, 1=save trajectories of final state-times reached,
						     2=save trajectories of intermediate steps) */

    vector<double>* tabX_;                        /**< vector containing the positions X of the generated trajectories */
    vector<double>* tabY_;                        /**< vector containing the positions Y of the generated trajectories */
    vector<double>* tabV_;                        /**< vector containing the velocities V of the generated trajectories */
    vector<double>* tabT_;                        /**< vector containing the times T of the generated trajectories */

    double maxTGResidual_;                        /**< the maximal residual of the trajectory generation process */
    int maxTGSteps_;                              /**< the maximal number of convergence steps of the trajectory generation process */


    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */  
    InitialGuess();
  
  public : 
    /**
     *  \brief init constructor
     */  
    InitialGuess(System* sys);

    /**
     *  \brief destructor
     */  
    virtual ~InitialGuess();

    /*************** Getters and Setters **************/

  protected : 
  public : 
    /**
     *  \brief return the robotic system features
     *  \return the robotic system features
     */
    virtual const System* sys() const = 0;

    //   /**
    //    *  \brief set the mode used to determine the first initial guess 
    //    *  \param mode : the new mode (0=random, 1=sampled control, 2=sampled param control, 3=rstar)
    //    *  \param options : options of the different modes
    //    *   - SC_FIRSTIG : table containing the sampling over the control
    //    *   - LOADFROMFILE_FIRSTIG : input filename path
    //    */
    //   void setFirstIGMode(int mode,const void* nomfic = NULL);

    /**
     *  \brief use the "load from file" mode to determine the first initial guess 
     */
    void use_LFF_FIGMode(const char* filename);

    //   /**
    //    *  \brief return a vector of parameters selected randomly from the constraints
    //    *  \param t : time the control generated from the set of parameters will last
    //    *  \return vector of parameters selected randomly
    //    */
    //   virtual double* getRandParam(double t) const = 0;

  public : 
    /**
     *  \brief set the debugging mode
     *  \param mode : the new debugging mode
     *
     *  (0=no debug, 1=save traj of final state-times reached, 2=save traj of intermediate steps)
     */
    inline void setDebug(int mode) { this->debug_ = mode; }

    /**
     *  \brief set the minimal number of new target state-times 
     that must be reached to continue the learning process
     *  \param nb : the new required number
     */
    inline void setMinNNTSR(int nb) { this->minNNTSR_ = nb; }

    /***************** Initialization *****************/

  protected : 
    /**
     *  \brief set the default state sampling parameters
     */
    virtual void setDfltSampS() = 0;

    /**
     *  \brief set the default state sampling min/max values
     */
    virtual void setDfltMinMaxSampS() = 0;

    /**
     *  \brief initialize the sampling grid
     */
    virtual void initGrid() = 0;

    /**
     *  \brief clear the sampling grid
     */
    virtual void clearGrid();

  public : 
    /**
     *  \brief set the trajectory generation maximal residual
     *  \param maxRes : the new maximal residual
     */
    void setMaxTGResidual(double maxRes) { this->maxTGResidual_ = maxRes; }

    /**
     *  \brief set the maximal number of convergence steps of the trajectory generation
     *  \param maxSteps : the new maximal number of convergence steps of the trajectory generation
     */
    void setMaxTGSteps(int maxSteps) { this->maxTGSteps_ = maxSteps; }

    /******************** Learning ********************/

  protected : 
    /**
     *  \brief compute random initial guesses for all samples of the state space
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeRandFirstIG() = 0;

    /**
     *  \brief compute a regular sampled control to determine the first initial guess
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeSCFirstIG() = 0;

    /**
     *  \brief compute a regular sampled parameterization of the control to determine the first initial guess
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeSCPFirstIG() = 0;

    /**
     *  \brief compute a reachable star to determine the first initial guess
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeRStarFirstIG() = 0;

    /**
     *  \brief load the first initial guess from a file
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* loadFirstIGfromFile() = 0;

    /**
     *  \brief compute the first initial guess from general functions
     *  \return the resulting list of first initial guesses
     */
    std::list<N2R*>* computeFirstIG();

    /**
     *  \brief determine the list of potentially reachable nodes
     *  \return the resulting list of potentially reachable nodes
     */
    std::list<N2R*> determineNodesToReach();

    /**
     *  \brief determine the list of potentially reachable nodes from a list
     *  \param lig : the input list of nodes
     *  \return the resulting list of potentially reachable nodes
     */
    std::list<N2R*> determineNodesToReach(std::list<N2R*>& lig);

    /**
     *  \brief try to generate a trajectory to the input node n2r thanks to the parameters of the node near
     *  \param n2r : the input node to reach
     *  \param near : the nearest node chosen as initial guess to converge to n2r
     *  \param param : the resulting parametrized control used to reach the goal defined if the trajectory generation succeed
     *  \return 0 if the node n2r has been reached, -1 else
     *
     *  The node n2r if updated if it has been reached
     */
    virtual int genTrajTo(N2R* n2r,N2R* near,double* param,bool argh=false) const = 0;

  public : 
    /**
     *  \brief learn the initial guess for the sample set
     */
    virtual void learnIG();


    /********************* Query **********************/

  public : 
    //   /**
    //    *  \brief query the initial guess base
    //    *  \param is : the initial state-time
    //    *  \param is : the goal state-time
    //    *  \return the best set of parameters found
    //    */
    //   virtual double* queryIG(const CLStateT& is,const CLStateT& gs) const = 0;


    /************* Input/Output Functions *************/

    /**
     *  \brief load the sampling of initial guesses from an input file
     *  \param nomfic : the name of the input file
     *  \return 0 if the process succeed, -1 else
     */
    virtual int load(char* nomfic);

    /**
     *  \brief save the sampling of initial guesses into an output file
     *  \param nomfic : the name of the output file
     *  \return 0 if the process succeed, -1 else
     */
    virtual int save(char* nomfic);


    /*************** Utility Functions ****************/
  
  protected : 
    /**
     *  \brief sample the trajectory defined on the node into tables of position and time
     *  \param node : the input node containing a trajectory
     *  \param tabX : the resulting table containing the position X values
     *  \param tabY : the resulting table containing the position Y values
     *  \param tabT : the resulting table containing the time T values
     *  \param n : the number of elements defined into tables
     *
     *  Tables must be allocated first
     */
    virtual void sampleTrajNode(N2R* node,double* tabX,double* tabY,double *tabT,int n) const = 0;

    /**
     *  \brief sample the target state-times into tables of position and time (WxT)
     *  \param tabX : the resulting table containing the position X values
     *  \param tabY : the resulting table containing the position Y values
     *  \param tabT : the resulting table containing the time T values
     *
     *  Tables must be allocated first
     */
    virtual void sampleTgtST(double* tabX,double* tabY,double *tabT) const;

    /**
     *  \brief list all nodes of the grid
     *  \return a pointer to a vector containing all nodes of the grid
     */
    virtual std::vector<N2R*>* listAllNodes() const = 0;

    /**
     *  \brief list all reached nodes of the grid
     *  \return a pointer to a vector containing all reached nodes of the grid
     */
    virtual std::vector<N2R*>* listAllReachedNodes() const = 0;

  public : 
    /**
     *  \brief compute the dimension of the grid (number of target state-times we try to reach)
     *  \return the dimension of the grid (number of target state-times we try to reach)
     */
    virtual int computeNbTargetST() const;

    /**
     *  \brief display the base of initial guesses
     */
    virtual void display() const;

    /**
     *  \brief display in a gnuplot window the trajectories of the input list
     *  \param l : list of the trajectories to display
     */
    virtual void drawGPTraj(std::list<N2R*> l);

    /**
     *  \brief display in a gnuplot window the status of each node of the grid
     */
    virtual void drawGPGrid();

    //   /**
    //    *  \brief display in a gnuplot window the generated trajectories of the grid
    //    */
    //   virtual void drawGPTraj() = 0;
  };

};
#endif // MPC_INITIALGUESS
