#ifndef MPC_UTILSTG_HPP
#define MPC_UTILSTG_HPP

#include "common/inc/Position.hpp"
#include "common/inc/StateTime.hpp"
#include "math/inc/nrtypes.h"
#include "system/inc/CarLikeSys.hpp"
#include "system/inc/WheelChairSys.hpp"
#include "common/inc/Control.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /************** Conversion functions **************/
                                    /**************************************************/

  /**
   *  \brief Convert car-like states into vectors for the trajectory generation process
   *  \param initS : initial car-like state
   *  \param goalS : goal car-like state
   *  \param initTS : resulting initial state table
   *  \param goalTS : resulting goal state table
   *
   *  initTS = (X0,Y0,TH0,PH0,V0,T0)
   *  goalTS = (Xf,Yf,THf,PHf,Vf,Tf)
   */
  void CLStateT_2_CL2tabD(const CLStateT& initS,const CLStateT& goalS,double** initTS,double** goalTS);

  /**
   *  \brief Convert a car-like state into a vector for the trajectory generation process
   *  \param initS : initial car-like state
   *  \param initTS : resulting initial state table
   *
   *  initTS = (X0,Y0,TH0,PH0,V0,T0)
   */
  void CLStateT_1_CL2tabD(const CLStateT& initS,double** initTS,double L);

  /**
   *  \brief Convert a car-like state into a vector for the trajectory generation process
   *  \param initS : initial car-like state
   *  \param initTS : resulting initial state table
   *
   *  initTS = (X0,Y0,TH0,OM0,V0,T0)
   */
  void CLStateT_1_CLtabD(const CLStateT& initS,double** initTS,double L);

  /**
   *  \brief Convert car-like states into vectors for the trajectory generation process
   *  \param initS : initial car-like state
   *  \param goalS : goal car-like state
   *  \param initVS : resulting initial state vector
   *  \param goalVS : resulting goal state vector
   *  \param L : length of the car-like vehicle
   *
   *  initVS = (X0,Y0,TH0,OM0,V0,T0)
   *  goalVS = (Xf,Yf,THf,OMf,Vf,Tf)
   */
  void CLStateT_2_CLVecDP(const CLStateT& initS,const CLStateT& goalS,Vec_DP& initVS,Vec_DP& goalVS,double L);

  /**
   *  \brief Convert wheelchair states into vectors for the trajectory generation process
   *  \param initS : initial wheelchair state
   *  \param goalS : goal wheelchair state
   *  \param initVS : resulting initial state vector
   *  \param goalVS : resulting goal state vector
   *
   *  initVS = (X0,Y0,TH0,V0,OM0,T0)
   *  goalVS = (Xf,Yf,THf,Vf,OM0,Tf)
   */
  void WCStateT_2_WCVecDP(const WCStateT& initS,const WCStateT& goalS,Vec_DP& initVS,Vec_DP& goalVS);

  /**
   *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
   *  \param sys : input car-like system
   *  \param cstt : output vector of constraints
   *
   *  constraints = (maxCurv,maxAcc,maxVel,robotLength,maxSteer,maxSteerSpd)
   */
  void CLSys_2_CLVecCstt(const CarLikeSys& sys,Vec_DP& cstt);

  /**
   *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
   *  \param sys : input car-like system
   *  \param cstt : output vector of constraints
   *
   *  constraints = (maxAcc,maxVel,robotLength,maxSteer,maxSteerSpd)
   */
  void CLSys_2_CL3VecCstt(const CarLikeSys& sys,Vec_DP& cstt);

  /**
   *  \brief Convert a wheelchair system into a vector of constraints for the trajectory generation process
   *  \param sys : input wheelchair system
   *  \param cstt : output vector of constraints
   *
   *  constraints = (maxAcc,maxAAcc,maxVel,maxAVel)
   */
  void WCSys_2_WCVecCstt(const WheelChairSys& sys,Vec_DP& cstt);

  /**
   *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
   *  \param sys : input car-like system
   *  \param cstt : output vector of constraints
   *
   *  constraints = (maxAcc,maxVel,maxSteerSpd,maxSteer,robotLength)
   */
  void CLSys_2_CL2TabCstt(const CarLikeSys& sys,double** cstt);

  /**
   *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
   *  \param sys : input car-like system
   *  \param cstt : output vector of constraints
   *
   *  constraints = (maxAcc,maxVel,maxSteerSpd,maxSteer,robotLength)
   */
  void CLSys_2_CL3TabCstt(const CarLikeSys& sys,double** cstt);

  // /**
  //  *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
  //  *  \param ctrl : input car-like control
  //  *  \param t : input time
  //  *  \param param : output parameters IG for the trajectory generation process
  //  *  \param s : the state on which the control is applied
  //  *  \param L : input length of the car-like
  //  */
  // void CLctrlt_2_CLIG(const CLControl& ctrl,double t,double** param,const StateS6D& s,double L);

  /**
   *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
   *  \param ctrl : input car-like control
   *  \param t : input time
   *  \param param : output parameters IG for the trajectory generation process
   *  \param s : the state on which the control is applied
   *  \param L : input length of the car-like
   */
  void CLctrlt_2_CLIG(const CLControl& ctrl,double t,double** param,const CLStateT& s,double L);

  /**
   *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
   *  \param ctrl : input car-like control
   *  \param t : input time
   *  \param param : output parameters IG for the trajectory generation process
   */
  void CLctrlt_2_CL2IG(const CLControl& ctrl,double t,double** param);

  /**
   *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
   *  \param ctrl : input cycab control
   *  \param t : input time
   *  \param param : output parameters IG for the trajectory generation process
   *  \param s : the state on which the control is applied
   *  \param L : input length of the car-like
   */
  void CCctrlt_2_CL3IG(const CCControl& ctrl,double t,double** param,const CLStateT& s,double L);

  /**
   *  \brief Convert a wheelchair control into a vector of parameters IG for the trajectory generation process
   *  \param ctrl : input wheelchair control
   *  \param t : input time
   *  \param param : output parameters IG for the trajectory generation process
   */
  void WCctrlt_2_WCIG(const WCControl& ctrl,double t,double** param);


                                    /**************************************************/
                                    /***************** Bends functions ****************/
                                    /**************************************************/

  /**
   *  \brief compute the center of curvature of a left bend follow by a car-like
   *  \param p : initial position of the car-like
   *  \param maxSt : maximal steering angle of the car-like
   *  \param L : length of the car-like
   *  \return the center of curvature of a left bend follow by a car-like
   */
  PositionXY leftCurvCenter(PositionXYTh p,double maxSt,double L);

  /**
   *  \brief compute the center of curvature of a right bend follow by a car-like
   *  \param p : initial position of the car-like
   *  \param maxSt : maximal steering angle of the car-like
   *  \param L : length of the car-like
   *  \return the center of curvature of a right bend follow by a car-like
   */
  PositionXY rightCurvCenter(PositionXYTh p,double maxSt,double L);

};
#endif
