#ifndef MPC_WCINITIALGUESS
#define MPC_WCINITIALGUESS

#include "trajGen/inc/InitialGuess.hpp"
#include "system/inc/WheelChairSys.hpp"
#include "trajGen/inc/WCTrajGen.hpp"

#include <list>

namespace mpc
{
class WCTrajGen;

                                    /**************************************************/
                                    /******************* Class WC_N2R *****************/
                                    /**************************************************/

  /** \class WC_N2R
   *  \brief Class representing a node to reach during the learning process of the initial guess (for a wheelchair)
   */
  class WC_N2R : public N2R
  {
  public : 
    static double wX_;                            /**< weight used to compute the distance between two nodes (on X) */
    static double wY_;                            /**< weight used to compute the distance between two nodes (on Y) */
    static double wTH_;                           /**< weight used to compute the distance between two nodes (on TH) */
    static double wVI_;                           /**< weight used to compute the distance between two nodes (on VI) */
    static double wVG_;                           /**< weight used to compute the distance between two nodes (on VG) */
    static double wT_;                            /**< weight used to compute the distance between two nodes (on T) */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    WC_N2R();

  public : 
    /**
     *  \brief init constructor (of an ig node)
     *  \param initST : the initial state of the trajectory
     *  \param goalST : the final state to reached
     *  \param param : the parametric representation of the control
     */
    WC_N2R(const WCStateT& initST,const WCStateT& goalST,double* param);

    /**
     *  \brief init constructor (of an explored node)
     *  \param initST : the initial state of the trajectory
     *  \param goalST : the final state to reached
     *  \param i,j,k,l,m,n : position of the node in the grid
     *  \param node : the node from which the current one may be derivated
     */
    WC_N2R(const WCStateT& initST,const WCStateT& goalST,
	   int i,int j,int k,int l,int m,int n,N2R* node);

    /**
     *  \brief init constructor (of a reached node)
     *  \param initST : the initial state of the trajectory
     *  \param goalST : the final state to reached
     *  \param i,j,k,l,m,n : position of the node in the grid
     *  \param param : the parametric representation of the control
     */
    WC_N2R(const WCStateT& initST,const WCStateT& goalST,
	   int i,int j,int k,int l,int m,int n,double* param);

    /**
     *  \brief destructor
     */
    ~WC_N2R();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the final state-time reached
     */
    virtual const WCStateT& initST() const { return *((WCStateT*)initST_); }

    /**
     *  \brief return the final state-time reached
     */
    virtual const WCStateT& goalST() const { return *((WCStateT*)goalST_); }

    //   /**
    //    *  \brief return a random neighbour of the current node
    //    *  \return a random neighbour of the current node if one exists, NULL else
    //    */
    //   inline virtual WC_N2R* getRandNeighbour() const { return ((WC_N2R*)(this->N2R::getRandNeighbour())); }

    //   /**
    //    *  \brief return the nearest neighbour of the current node
    //    *  \return the nearest neighbour of the current node if one exists, NULL else
    //    */
    //   virtual WC_N2R* getNearNeighbour() const;

    /**************** Utility functions ***************/

    /**
     *  \brief return a weighted distance between the current node and the input node
     *  \param n2 : the second node
     *  \return a weighted distance between both nodes
     */
    inline double computeWDistTo(N2R* n2) const { return this->computeWDistTo((WC_N2R*)n2); }

    /**
     *  \brief return a weighted distance between the current node and the input node
     *  \param n2 : the second node
     *  \return a weighted distance between both nodes
     */
    double computeWDistTo(WC_N2R* n2) const;
  };


                                    /**************************************************/
                                    /***************** Class WC_N2RGrid ***************/
                                    /**************************************************/

  /** \class WC_N2RGrid
   *  \brief Class representing a grid containing the nodes to reach during the learning process (for a wheelchair)
   */
  class WC_N2RGrid : public N2RGrid
  {
  protected : 

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    WC_N2RGrid();

  public : 
    /**
     *  \brief init constructor
     */
    WC_N2RGrid(int* sampling,double* minSampS,double* maxSampS);

    /********* Access/Modification of the grid ********/

  public : 
    /**
     *  \brief return an element of the grid at pos = (i,j,k,l,m,n)
     *  \param i : index of the time dimension
     *  \param j : index of the initial velocity dimension
     *  \param k : index of the goal velocity dimension
     *  \param l : index of the x position dimension
     *  \param m : index of the y position dimension
     *  \param n : index of the heading dimension
     *  \return the element of the grid at pos = (i,j,k,l,m,n), NULL if the input coordinates are outside of the grid
     */
    N2R* gridAt(int i,int j,int k,int l,int m,int n) const;

    /**
     *  \brief return an element of the grid at pos = (i,j,k,l,m,n)
     *  \param i : index of the time dimension
     *  \param j : index of the initial velocity dimension
     *  \param k : index of the goal velocity dimension
     *  \param l : index of the x position dimension
     *  \param m : index of the y position dimension
     *  \param n : index of the heading dimension
     *  \return the element of the grid at pos = (i,j,k,l,m,n), NULL if the input coordinates are outside of the grid
     */
    N2R*& gridAt(int i,int j,int k,int l,int m,int n);

  protected :
    /**
     *  \brief compute the initial state-time of a cell of the grid
     *  \param i : index of the time dimension
     *  \param j : index of the initial velocity dimension
     *  \param k : index of the goal velocity dimension
     *  \param l : index of the x position dimension
     *  \param m : index of the y position dimension
     *  \param n : index of the heading dimension
     *  \param initST : resulting initial state-time
     */
    void computeInitST(int i,int j,int k,int l,int m,int n,
		       WCStateT& initST) const;

    /**
     *  \brief compute the goal state-times of a cell of the grid
     *  \param i : index of the time dimension
     *  \param j : index of the initial velocity dimension
     *  \param k : index of the goal velocity dimension
     *  \param l : index of the x position dimension
     *  \param m : index of the y position dimension
     *  \param n : index of the heading dimension
     *  \param goalST : resulting goal state-time
     */
    void computeGoalST(int i,int j,int k,int l,int m,int n,
		       WCStateT& goalST) const;

    /**
     *  \brief search for reachable nodes on the grid around a reached node (and mark them as reachable on the grid)
     *  \param node : the input already reached node
     *  \return a list containing the reachable nodes on the grid around the input node
     *
     *  An new instance of list is allocated and returned here
     */
    std::list<N2R*>* reachableNodesAround(N2R* node);

    /**
     *  \brief search for already reached nodes on the grid around a node to reach
     *  \param node : the input node to reach
     *  \return a list containing the already reached nodes on the grid around the input node
     *
     *  An new instance of list is allocated and returned here
     */
    std::list<N2R*>* reachedNodesAround(N2R* node) const;

    /**
     *  \brief create a N2R from its coordinates and a a node from which it can be derivated
     *  \param coord : coordinates of the N2R
     *  \param node : node from which the N2R may be reachable
     *
     *  An new instance of N2R is allocated here
     *  Put directly the new node in the grid
     */
    virtual N2R* createExploredN2R(int* coord,N2R* node);

    /**
     *  \brief create a N2R from its coordinates and a a node from which it can be derivated
     *  \param coord : coordinates of the N2R
     *  \param node : node from which the N2R may be reachable
     *
     *  An new instance of N2R is allocated here
     *  Both state-times and the control parameterization are copied here
     */
    virtual N2R* createIGN2R(const StateTime* initST,const StateTime* goalST,double* param);

    /**
     *  \brief create a reached node from an input file thanks to its position and parameters
     *  \param pos : the position of the node in the grid
     *  \param params : the parameterized control defining the trajectory
     *  \return a pointer to the created node
     */
    virtual N2R* createReachedNode(int* coord,double* params);

  public : 
    /**
     *  \brief compute the number of nodes noted as "nodes to reach" on the grid
     *  \return the number of nodes noted as "nodes to reach" on the grid
     */
    int N2RonGrid() const;

    /**
     *  \brief compute the number of nodes noted as reached on the grid
     *  \return the number of nodes noted as reached on the grid
     */
    int NReachedonGrid() const;

    /**
     *  \brief compute the number of nodes noted as used on the grid
     *  \return the number of nodes noted as used on the grid
     */
    int NUsedonGrid() const;

    /**
     *  \brief display the nodes set as reached on the grid
     */
    void displayReachedOnGrid() const;

    /**
     *  brief display the sampling of the grid
     */
    void displayGridSampling() const;

    //   /**
    //    *  \brief search for the nearest defined node of the grid to compute a trajectory between both input state-times
    //    *  \param is : the input init state-time
    //    *  \param gs : the input goal state-time
    //    *  \return the nearest defined node of the grid to compute a trajectory between both input state-times
    //    */
    //   N2R* nearestNodeToComputeTrajBetween(const WCStateT& is,const WCStateT& gs) const;

    /**
     *  \brief search for the nearest defined node of the grid to compute a trajectory between both input state-times (2nd version)
     *  \param is : the input init state-time
     *  \param gs : the input goal state-time
     *  \return the nearest defined node of the grid to compute a trajectory between both input state-times
     */
    N2R* nearestNodeToComputeTrajBetweenV2(const WCStateT& is,const WCStateT& gs) const;
  };

                                    /**************************************************/
                                    /*************** Class WCInitialGuess *************/
                                    /**************************************************/

  /** \class WCInitialGuess
   *  \brief Class to compute and stock initial guess of parameters (for a wheelchair system)
   */
  class WCInitialGuess : public InitialGuess
  {
  protected : 
    int SC_AccSp_;                                /**< number of samples of the acceleration control for the SC_FIG mode */
    int SC_AAccSp_;                               /**< number of samples of the angular acceleration control for the SC_FIG mode */

    /********** Constructors and destructors **********/

  protected :   
    /**
     *  \brief default constructor
     */  
    WCInitialGuess();
  
  public : 
    /**
     *  \brief init constructor
     *  \param sys : the wheelchair features
     */  
    WCInitialGuess(WheelChairSys* sys);

    /**
     *  \brief init constructor
     *  \param sys : the wheelchair features
     *  \param tmin : the minimal time
     *  \param tmax : the maximal time
     *  \param spg_t : the time sampling
     *  \param spg_vi : the initial speed sampling
     *  \param spg_vg : the goal speed sampling
     *  \param spg_dx : the position sampling
     *  \param spg_dy : the position sampling
     *  \param spg_dth : the heading sampling
     */
    WCInitialGuess(WheelChairSys* sys,
		   int tmin,int tmax,
		   int spg_t,int spg_vi,int spg_vg,int spg_dx,int spg_dy,int spg_dth);

    /**
     *  \brief destructor
     */  
    virtual ~WCInitialGuess();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the robotic system features
     *  \return the robotic system features
     */
    inline virtual const WheelChairSys* sys() const { return ((WheelChairSys*)sys_); }

    /**
     *  \brief use the sampled control mode to determine the first initial guess 
     *  \param sampAcc : number of samples of the acceleration control
     *  \param sampStSpd : number of samples of the angular acceleration control
     */
    void use_SC_FIGMode(double sampAcc,double sampAAcc);

    /***************** Initialization *****************/

  protected : 
    /**
     *  \brief initialize the sampling of the state space
     *  \param spg_t : sampling over the time
     *  \param spg_vi : sampling over the initial speed
     *  \param spg_vg : sampling over the goal speed
     *  \param spg_dx : sampling over the position x
     *  \param spg_dy : sampling over the position y
     *  \param spg_dth : sampling over the heading
     */
    void initSampS(int spg_t,int spg_vi,int spg_vg,int spg_dx,int spg_dy,int spg_dth);

    /**
     *  \brief set the default state sampling parameters
     */
    virtual void setDfltSampS();

    /**
     *  \brief set the default state sampling min/max values
     */
    virtual void setDfltMinMaxSampS();

    /**
     *  \brief initialize the sampling grid
     */
    virtual void initGrid();


    /******************** Learning ********************/

  protected : 
    /**
     *  \brief compute a random initial guess for a sample of the state space
     *  \param ig : the resulting initial guess (must be allocated)
     *  \param t : time the control generated from the set of parameters will last
     */
    virtual void computeRandFirstIG(double* ig,double t);

    /**
     *  \brief compute random initial guesses for all samples of the state space
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeRandFirstIG();

    /**
     *  \brief compute a regular sampled control to determine the first initial guess
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeSCFirstIG();

    /**
     *  \brief compute a regular sampled parameterization of the control to determine the first initial guess
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeSCPFirstIG();

    //   /**
    //    *  \brief fake initialization
    //    *  \return the resulting list of first initial guesses
    //    */
    //   virtual std::list<N2R*>* computeSCPFirstIG2();

    /**
     *  \brief compute a  to determine the first initial guess
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* computeRStarFirstIG();

    /**
     *  \brief load the first initial guess from a file
     *  \return the resulting list of first initial guesses
     */
    virtual std::list<N2R*>* loadFirstIGfromFile();

    /**
     *  \brief try to generate a trajectory to the input node n2r thanks to the parameters of the node near
     *  \param n2r : the input node to reach
     *  \param near : the nearest node chosen as initial guess to converge to n2r
     *  \param param : the resulting parametrized control used to reach the goal defined if the trajectory generation succeed
     *  \return 0 if the node n2r has been reached, -1 else
     *
     *  The node n2r if updated if it has been reached
     */
    int genTrajTo(N2R* n2r,N2R* near,double* param,bool argh=false) const;


    // public : 
    //   /**
    //    *  \brief learn the initial guess for the sample set
    //    */
    //   virtual void learnIG();

    /********************* Query **********************/

  public : 
    /**
     *  \brief query the initial guess base
     *  \param is : the initial state-time
     *  \param is : the goal state-time
     *  \return the best set of parameters found
     */
    virtual const double* queryIG(const WCStateT& is,const WCStateT& gs) const;


    /*************** Utility Functions ****************/
  
  protected : 
    /**
     *  \brief sample the trajectory defined on the node into tables of position and time
     *  \param node : the input node containing a trajectory
     *  \param tabX : the resulting table containing the position X values
     *  \param tabY : the resulting table containing the position Y values
     *  \param tabT : the resulting table containing the time T values
     *  \param n : the number of elements defined into tables
     *
     *  Tables must be allocated first
     */
    virtual void sampleTrajNode(N2R* node,double* tabX,double* tabY,double *tabT,int n) const;

    /**
     *  \brief list all nodes of the grid
     *  \return a pointer to a vector containing all nodes of the grid
     */
    virtual std::vector<N2R*>* listAllNodes() const;

    /**
     *  \brief list all reached nodes of the grid
     *  \return a pointer to a vector containing all reached nodes of the grid
     */
    virtual std::vector<N2R*>* listAllReachedNodes() const;

  public : 
    //   /**
    //    *  \brief display in a gnuplot window the generated trajectories of the grid
    //    */
    //   virtual void drawGPTraj();

    //   /**
    //    *  \brief display in a gnuplot window the status of each node of the grid
    //    */
    //   virtual void drawGPGrid();
  };

};
#endif // MPC_WCINITIALGUESS
