#include <boost/program_options.hpp>
namespace po = boost::program_options;

#include <iostream>
#include <fstream>
#include <iterator>
#include <math.h>
using namespace std;

#include "utilsLuis.h"
#include "nrtypes.h"
#include "modelFInt.hpp"
#include "gnuplot_i.hpp"
#include "Utils.hpp"

#include <time.h>
#include <sys/time.h>

// Parameters
#define S      (*(carModel.vecParameters_p))[6]

int main (int ac, char* av[]) {    
  try {
    ModelFInt carModel;
    Vec_DP vecConstraints(carModel.dLength);
    Vec_DP vecState_x0(carModel.xLength + 1);
    Vec_DP vecState_xf(carModel.xLength + 1);    	    	        
    
    // Group of options that will be allowed only on command line
    po::options_description generic("Generic options");
    generic.add_options()
      ("version,v", "print version string")
      ("help", "produce help message")    
      ;
    
    // Group of options that will be allowed both on command line and in
    // config file
    po::options_description config("Configuration");
    config.add_options()
      ("xinitial", po::value<double>(&vecState_x0[0])->default_value(0.0), 
       "Intial State X")
      ("yinitial", po::value<double>(&vecState_x0[1])->default_value(0.0), 
       "Intial State Y")            
      ("thetainitial", po::value<double>(&vecState_x0[2])->default_value(0.0), 
       "Intial State Theta")            
      ("curvatureinitial", po::value<double>(&vecState_x0[3])->default_value(0.0), 
       "Intial State Curvature")
      ("velinitial", po::value<double>(&vecState_x0[4])->default_value(0.0), 
       "Intial State Velocity")
      ("timeinitial", po::value<double>(&vecState_x0[5])->default_value(0.0), 
       "Intial State Time")
      ("xfinal", po::value<double>(&vecState_xf[0])->default_value(4.0), 
       "Final State X")
      ("yfinal", po::value<double>(&vecState_xf[1])->default_value(3.0), 
       "Final State Y")            
      ("thetafinal", po::value<double>(&vecState_xf[2])->default_value(M_PI), 
       "Final State Theta")            
      ("curvaturefinal", po::value<double>(&vecState_xf[3])->default_value(1.0), 
       "Final State Curvature")
      ("velfinal", po::value<double>(&vecState_xf[4])->default_value(1.0), 
       "Final State Velocity")
      ("timefinal", po::value<double>(&vecState_xf[5])->default_value(15.0), 
       "Final State Time")
      ("maxcurvature", po::value<double>(&vecConstraints[0])->default_value(1.4), 
       "Maximum Curvature")
      ("maxacceleration", po::value<double>(&vecConstraints[1])->default_value(0.3), 
       "Maximum Acceleration")
      ("maxvelocity", po::value<double>(&vecConstraints[2])->default_value(29.0), 
       "Maximum Velocity")
      ("robotlength", po::value<double>(&vecConstraints[3])->default_value(10.0), 
       "Robot Length")
      ("maxsteering", po::value<double>(&vecConstraints[4])->default_value(0.25*M_PI), 
       "Maximal Steering")
      ("maxsteerspd", po::value<double>(&vecConstraints[5])->default_value(0.02*M_PI), 
       "Maximal Steering Speed")
      ;
    
    // Hidden options, will be allowed both on command line and
    // in config file, but will not be shown to the user.
    po::options_description hidden("Hidden options");
    hidden.add_options()
      ("input-file", po::value< vector<string> >(), "input file")
      ;
    
    po::options_description cmdline_options;
    cmdline_options.add(generic).add(config).add(hidden);
    
    po::options_description config_file_options;
    config_file_options.add(config).add(hidden);
    
    po::options_description visible("Allowed options");
    visible.add(generic).add(config);
    
    po::positional_options_description p;
    p.add("input-file", -1);
    
    po::variables_map vm;
    store(po::command_line_parser(ac, av).
	  options(cmdline_options).positional(p).run(), vm);
    
    ifstream ifs("../etc/trajectory_generation.cfg");
    store(parse_config_file(ifs, config_file_options), vm);
    notify(vm);
    
    if (vm.count("help")) {
      cout << visible << "\n";
      return 0;
    }
    
    if (vm.count("version")) {
      cout << "Trajectory Generation, version 1.0\n";
      return 0;
    }
    
    if (vm.count("input-file"))
      {
	cout << "Input files are: " 
	     << vm["input-file"].as< vector<string> >() << "\n";
      }
    
    cout << "Initial State (x,y,theta,curvature,vel,time) is (" <<  vecState_x0[0] << 
      "," << vecState_x0[1] <<  "," << vecState_x0[2] << 
      "," << vecState_x0[3] << "," << vecState_x0[4] << "," << vecState_x0[5] << ")\n";    
    cout << "Final State (x,y,theta,curvature,vel,time) is (" <<  vecState_xf[0] << 
      "," << vecState_xf[1] <<  "," << vecState_xf[2] << 
      "," << vecState_xf[3] << "," << vecState_xf[4] << "," << vecState_xf[5] << ")\n";
    
    // init options

    carModel.setMaxNbStep(40);
    carModel.setMaxResidual(0.5);
    
    carModel.setUseVelCst(true);
    carModel.setUseAccCst(true);
    carModel.setUseAVelCst(true);
    carModel.setUseAAccCst(false);
    
    carModel.setDBG(2);

    // initialize timer
    
    struct timeval t1,t2;
    gettimeofday(&t1,NULL);

    //Compute trajectory
    
    carModel.ComputeTrajectory(vecState_x0,vecState_xf,vecConstraints,NULL);
    
    // stop timer
    
    gettimeofday(&t2,NULL);

    // compute the time passed
    
    int elapsedTs = t2.tv_sec - t1.tv_sec;
    int elapsedTus = t2.tv_usec - t1.tv_usec;
    double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
    cout << "Elapsed time : " << elapsedT << endl;
    
//     //Computing Curvature
//     vector<double> curvature;
//     vector<double> distanceK;
//     for(int j=0;j<=30;j++){
//       curvature.push_back((double)carModel.funcCurvatureConstrained(j*(S/30)));
//       distanceK.push_back((j*(S/30)));                
//     }    
//     Gnuplot gCurvature;
//     gCurvature.reset_plot();
//     gCurvature.set_style("lines").plot_xy(distanceK,curvature,"Curvature");        
    
    //Computing Angular Velocity
    vector<double> angvel;
    vector<double> distanceK;
    for(int j=0;j<=30;j++){
      angvel.push_back((double)carModel.funcAVel(j*(S/30)));                
      distanceK.push_back((j*(S/30)));
    }
    Gnuplot gAngVel;
    gAngVel.reset_plot();
    gAngVel.set_style("lines").plot_xy(distanceK,angvel,"Angular Velocity");
    
    //Computing Position
    carModel.SaveTrajectory();
    Gnuplot gPosition;
    gPosition.reset_plot();
    gPosition.set_style("lines").plot_xy(*(carModel.posX),*(carModel.posY),"Position");
    
    //Computing Velocity
    vector<double> vel;
    vector<double> boundVel;
    vector<double> distanceV;
    for(int j=0;j<=30;j++){
      vel.push_back((double)carModel.funcVel(j*(S/30)));
      boundVel.push_back(vecConstraints[2]);
      distanceV.push_back((j*(S/30)));                
    }    
    Gnuplot gVel;
    gVel.reset_plot();
    gVel.set_style("lines").plot_xy(distanceV,vel,"Velocity");
    gVel.set_style("lines").plot_xy(distanceV,boundVel,"Velocity bound");
    
    //Computing Acceleration
    vector<double> acc;
    vector<double> boundAcc;
    vector<double> distanceA;
    for(int j=0;j<=30;j++){
      acc.push_back((double)carModel.funcAcc(j*(S/30)));
      boundVel.push_back(vecConstraints[3]);
      distanceA.push_back((j*(S/30)));	        
    }    
    Gnuplot gAcc;
    gAcc.reset_plot();
    gAcc.set_style("lines").plot_xy(distanceA,acc,"Acceleration");
    //gAcc.set_style("lines").plot_xy(distanceA,boundAcc,"Acceleration bound");
    
    //Computing Heading2
    vector<double> dir2;
    vector<double> distance2;
    for(int j=0;j<=30;j++){
      double thtmp = normalizedAngle((double)carModel.funcTheta(j*(S/30.0)));
      double disttmp = (j*(S/30.0));
      dir2.push_back(thtmp);
      distance2.push_back(disttmp);
    } 
    Gnuplot gTH2;
    gTH2.reset_plot();
    gTH2.set_style("lines").plot_xy(distance2,dir2,"Heading2");

    // Compute steering
    vector<double> steer;
    vector<double> distanceSt;
    for(int j=0;j<=30;j++){
      double st = (double)carModel.funcSteer(j*(S/30.0));
      double dSt = (j*(S/30.0));
      steer.push_back(st);
      distanceSt.push_back(dSt);
    } 
    Gnuplot gSt;
    gSt.reset_plot();
    gSt.set_style("lines").plot_xy(distanceSt,steer,"Steering Angle");
    

    wait_for_key();                                      
  }
  catch(exception& e)
    {
      cout << e.what() << "\n";
      return 1;
    }    
  return 0;    
}

