#include "TrajGen.hpp"
using namespace mpc;

#include "InvMatrix.hpp"

#define DFLT_DERIV_VAL -999999999.99

                                    /**************************************************/
                                    /****************** Class TrajGen *****************/
                                    /**************************************************/

int TrajGen::DBG_ = 1;

int TrajGen::maxNbStep_ = 40;
double TrajGen::maxResidual_ = 0.5;

// bool TrajGen::timeFixed_ = true;
bool TrajGen::variableCoef_ = true;


/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */  
TrajGen::TrajGen()
{
  ig_ = NULL;

  ctraj_ = NULL;
  straj_ = NULL;
  
//   COEF_COR_ = 0.5;
  COEF_COR_ = 0.1;
  FDpartDeriv_ = false;

  tcStatus_ = TIME_CSTD;
  minTime_ = 0.0;
  maxTime_ = 999999999.9;

  finalResidual_ = 0.05;
  finalNbSteps_ = 20;

  matInvJacobianG = NULL;
  vecParameters_p = NULL;

  weightsCF_ = NULL;
  CFtype_ = SQUARE_CF;
}

/*
 *  \brief init constructor
 *  \param sys : the constraints of the robotic system
 */  
TrajGen::TrajGen(const System* sys)
{
  sys_ = sys;

  ig_ = NULL;

  ctraj_ = NULL;
  straj_ = NULL;
  
//   COEF_COR_ = 0.5;
  COEF_COR_ = 0.1;
  FDpartDeriv_ = false;

  tcStatus_ = TIME_CSTD;
  minTime_ = 0.0;
  maxTime_ = 999999999.9;

  finalResidual_ = 0.05;
  finalNbSteps_ = 20;

  matInvJacobianG = NULL;
  vecParameters_p = NULL;

  weightsCF_ = NULL;
  CFtype_ = SQUARE_CF;
}

/*
 *  \brief destructor
 */  
TrajGen::~TrajGen()
{
  if(ctraj_ != NULL) delete ctraj_;
  if(straj_ != NULL) delete straj_;

  if(matInvJacobianG != NULL) delete matInvJacobianG;
  if(vecParameters_p != NULL) 
    {
      delete vecParameters_p;
      vecParameters_p = NULL;
    }
  if(weightsCF_ != NULL) delete [] weightsCF_;
}

/*************** Getters and Setters **************/

/**
 *  \brief set d(S_i) / d(p_j) 
 *  \param i : index of the state vector in the hessian matrix
 *  \param j : index of the parameter vector in the hessian matrix
 *  \param val : the new value
 */
void TrajGen::setJacobianS(int i,int j,double val)
{
  vecJacobianS_[i*pLengthFree() + j] = val;
}

/**
 *  \brief return a value of the jacobian of the state vector
 *  \param i : index of the state vector in the jacobian matrix
 *  \param j : index of the parameter vector in the jacobian matrix
 *  \return d(S_i) / d(p_j) 
 */
double TrajGen::getJacobianS(int i,int j)
{
  return vecJacobianS_[i*pLengthFree() + j];
}

/**
 *  \brief set d^2(S_i) / (d(p_j) d(p_k)) 
 *  \param i : index of the state vector in the hessian matrix
 *  \param j,k : indexes of parameters vectors in the hessian matrix
 *  \param val : the new value
 */
void TrajGen::setHessianS(int i,int j,int k,double val)
{
  matHessianS_[k][i*pLengthFree() + j] = val;
}

/**
 *  \brief return a value of the hessian of the state vector
 *  \param i : index of the state vector in the hessian matrix
 *  \param j,k : indexes of parameters vectors in the hessian matrix
 *  \return d^2(S_i) / (d(p_j) d(p_k))
 */
double TrajGen::getHessianS(int i,int j,int k)
{
  return matHessianS_[k][i*pLengthFree() + j];
}

/**
 *  \brief return a value of the inverse hessian of the state vector
 *  \param i : index of the state vector in the inverse hessian matrix
 *  \param j,k : indexes of parameters vectors in the inverse hessian matrix
 *  \return (d^2(S_i) / (d(p_j) d(p_k)))^{-1}
 */
double TrajGen::getInvHessianS(int i,int j,int k)
{
  return invHessianS_[i*pLengthFree() + j][k];
}

/**
 *  \brief getter : return the maximum number of steps
 *  \return maximum number of steps allowed to generate the trajectory
 */
int TrajGen::getMaxNbStep()
{
  return maxNbStep_;
}

/*
 *  \brief setter : set the maximum number of steps
 *  \param n : maximum number of steps allowed to generate the trajectory
 */
void TrajGen::setMaxNbStep(const int n)
{
  maxNbStep_ = n;
}

/*
 *  \brief getter : return the maximal residual allowed
 *  \return maximal residual allowed to stop the trajectory generation process
 */
double TrajGen::getMaxResidual()
{
  return maxResidual_;
}

/*
 *  \brief setter : set the maximal residual allowed
 *  \param r : maximal residual allowed to stop the trajectory generation process
 */
void TrajGen::setMaxResidual(const double r)
{
  maxResidual_ = r;
}

/*
 *  \brief getter : return the debugging level
 *  \return debugging level
 */
int TrajGen::getDBG()
{
  return DBG_;
}

/*
 *  \brief setter : set the debugging level
 *  \param dbg : debugging level
 */
void TrajGen::setDBG(const int dbg)
{
  DBG_ = dbg;
}

/*
 *  \brief set the time constraint status
 *  \param the time constraint status
 */
void TrajGen::setTcStatus(timeCstt_t tcS,double minT,double maxT)
{
  this->tcStatus_ = tcS;
  this->minTime_ = minT;
  this->maxTime_ = maxT;

  vecCorrParameters_p = Vec_DP(pLengthFree());
  matJacobianG = Mat_DP(gLength,pLengthFree());
  //matInvJacobianG = Mat_DP(pLengthFree(),gLength);
}

/*
 *  \brief return the number of free parameters
 *  \return the number of free parameters
 */
int TrajGen::pLengthFree() const
{
  if (tcStatus_ == TIME_FIXED) return pLength-1;
  else return pLength;
}

/*
 *  \brief return the final parameters
 *  \return the final parameters
 *
 *  params must be allocated first
 */
void TrajGen::finalParams(double* params) const
{
  if(vecParameters_p != NULL)
    {
      for(int i=0; i<pLength; i++)
	params[i] = (*vecParameters_p)[i];
    }
}


/********* Optimization related routines **********/

void TrajGen::ComputeIG(Vec_DP** parameters)
{
  // delete the old parameters
  if(vecParameters_p != NULL) delete vecParameters_p;

  // allocate the parameters vector
  if(parameters == NULL)
    {
      vecParameters_p = new Vec_DP(pLength);
    }
  else if(*parameters == NULL)
    {
      *parameters = new Vec_DP(pLength);
      vecParameters_p = *parameters;
    }
  else
    {
      vecParameters_p = *parameters;
      return;
    }

  // check if an initial guess database exists
  int resIG = -1;
  if(ig_ != NULL)
    {
      resIG = this->ComputeIGfromDB();
    }

  if(resIG == -1)
    {
      this->ComputeDfltIG();
    }
}

/*
 *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
 */
void TrajGen::ComputeJacobianG_FD()
{
  // initialize the partial derivatives
//   for(int i=0; i<gLength; i++)
//     for(int j=0; j<pLength-1; j++)
//       matJacobianG[i][j] = DFLT_DERIV_VAL;
  for(int i=0; i<gLength; i++)
    for(int j=0; j<pLengthFree(); j++)
      matJacobianG[i][j] = DFLT_DERIV_VAL;

  // save the state reached without any modifications over the parameters
  this->computeAllFct();
  double* oldStateT = new double[xLength];
  this->getFinalStateTReached(oldStateT);

//   cout << "old param : "; this->displayParameters(); 
//   cout << " -> old rstatet : "; this->displayStateReached(); cout << endl;

//   for(int i=0; i<pLength-1; i++)
  for(int i=0; i<pLengthFree(); i++)
    {
      // compute the forward method again with a change over the i-th parameter
      (*vecParameters_p)[i] += E;
      this->computeAllFct();
      
//       cout << "param : "; this->displayParameters(); 
//       cout << " -> rstatet : "; this->displayStateReached(); cout << endl;

      // compute the new state-time reached after modification of the i-th parameter
      double* newStateT = new double[xLength];
      this->getFinalStateTReached(newStateT);

//       cout << "i : " << i << " -> newStateT : (";
//       for(int j=0; j<xLength; j++) cout << newStateT[j] << " ";
//       cout << ")" << endl;

      // compute the partial derivatives wrt. the i-th parameter
      for(int j=0; j<gLength; j++)
	matJacobianG[j][i] = (newStateT[j] - oldStateT[j]) * UnSurE;
      
      // restore the current value of parameters
      (*vecParameters_p)[i] -= E;

      // delete the new state-time reached
      delete newStateT;
    }
//   cout << "Jacobian : " << endl;
//   this->displayJacobian();

  // restore the initial profiles of the state functions
  this->computeAllFct();

  // delete all vectors used
  delete oldStateT;
}

/*
 *  \brief invert the partial derivatives matrix
 */
void TrajGen::ComputeInvJacobianS()
{
  //this->displayJacobian();

  if(matInvJacobianG != NULL) delete matInvJacobianG;
  matInvJacobianG = new Mat_DP(pLengthFree(),gLength);

  int M2 = matInvJacobianG->nrows();
  int N2 = matInvJacobianG->ncols();
  
  for(int i=0; i<M2; i++) for(int j=0; j<N2; j++) (*matInvJacobianG)[i][j] = 0;

  invMatDP(matJacobianG, *matInvJacobianG);
}

/**
 *  \brief set the vector representing the jacobian of S
 */
void TrajGen::setVecJacobianS()
{
  int M = gLength;
  int N = pLengthFree();

  for(int i=0; i<M; i++)
    for(int j=0; j<N; j++)
      setJacobianS(i,j,matJacobianG[i][j]);
}

/**
 *  \brief invert the hessian matrix
 */
void TrajGen::ComputeInvHessianS()
{
  int M = gLength*pLengthFree();
  int N = pLengthFree();

  for(int i=0; i<M; i++)
      for(int j=0; j<N; j++)
	invHessianS_[i][j] = 0;

  invMatDP(matHessianS_, invHessianS_);
}

/*
 *  \brief compute the correction to apply to the parameters
 */
void TrajGen::ComputeCorrection()
{
  for(int i=0; i<pLengthFree();i++)
    {
      vecCorrParameters_p[i] = 0;
      
      for(int j=0;j<gLength;j++)
	{
	  vecCorrParameters_p[i] += (-1.0)*vecConstraints_g[j]*(*matInvJacobianG)[i][j]; 
	}	    
    }
}

/*
 *  \brief apply the correction to the set of parameters
 */
void TrajGen::ApplyCorrection()
{
  if (variableCoef_)
    {
      // update the correction coefficient
      if(residual_ <= oldResidual_)
	{
	  nbDeacreaseStep_++;
	  
	  if(nbDeacreaseStep_ >= 10) coefCor_ = 1.0;
	  else if(nbDeacreaseStep_ >= 9) coefCor_ = 0.8;
	  else if(nbDeacreaseStep_ >= 8) coefCor_ = 0.7;
	  else if(nbDeacreaseStep_ >= 7) coefCor_ = 0.6;
	  else if(nbDeacreaseStep_ >= 6) coefCor_ = 0.5;
	  else if(nbDeacreaseStep_ >= 5) coefCor_ = 0.4;
	  else if(nbDeacreaseStep_ >= 3) coefCor_ = 0.3;
	}
      else
	{
	  coefCor_ = COEF_COR_;
	  nbDeacreaseStep_ = 0;
	}
    }
  
  for(int i=0; i<pLength-1;i++)
    (*vecParameters_p)[i] += coefCor_*vecCorrParameters_p[i];
  if(tcStatus_ == TIME_FREE)
    {
      (*vecParameters_p)[pLength-1] += coefCor_*vecCorrParameters_p[pLength-1];
      if((*vecParameters_p)[pLength-1] < 0.0) (*vecParameters_p)[pLength-1] = 0.0;
    }
  else if(tcStatus_ == TIME_CSTD)
    {
      (*vecParameters_p)[pLength-1] += 1.0*vecCorrParameters_p[pLength-1];
      if((*vecParameters_p)[pLength-1] < minTime_) (*vecParameters_p)[pLength-1] = minTime_;
      if((*vecParameters_p)[pLength-1] > maxTime_) (*vecParameters_p)[pLength-1] = maxTime_;
    }
}

/*
 *  \brief compute the residual between the goal state and the stae reached
 *  \return the residual between the goal state and the stae reached
 */
double TrajGen::ComputeResidual() const
{
  double sqRes = 0;
  for(int i=0; i<gLength; i++)
    {
      sqRes += (vecConstraints_g[i] * vecConstraints_g[i]);
    }

  return sqrt(sqRes);
}

/*
 *  \brief save the current vector of parameters into the vector of best parameters
 */
void TrajGen::SaveBestParams()
{
  bestResidual_ = residual_;
  bestCNT_ = CNT_;

  for(int i=0; i<pLengthFree(); i++)
    {
      vecBestParam_[i] = (*vecParameters_p)[i];
    }
}

/*
 *  \brief load the vector of best parameters into the current vector of parameters
 */
void TrajGen::LoadBestParams()
{
  for(int i=0; i<pLengthFree(); i++)
    {
      (*vecParameters_p)[i] = vecBestParam_[i];
    }
}


/**************** Quadrature tools ****************/

/*
 *  \brief Quadrature tools...
 */  
void TrajGen::gauleg(const double x1, const double x2, double x[], double w[], int n) const
{
    const double EPS=1.0e-14;
    int m,j,i;
    double z1,z,xm,xl,pp,p3,p2,p1;
    
    m=(n+1)/2;
    xm=0.5*(x2+x1);
    xl=0.5*(x2-x1);
    for (i=0;i<m;i++) {
        z=cos(3.141592654*(i+0.75)/(n+0.5));
        do {
            p1=1.0;
            p2=0.0;
            for (j=0;j<n;j++) {
                p3=p2;
                p2=p1;
                p1=((2.0*j+1.0)*z*p2-j*p3)/(j+1);
            }
            pp=n*(z*p1-p2)/(z*z-1.0);
            z1=z;
            z=z1-p1/pp;
        } while (fabs(z-z1) > EPS);
        x[i]=xm-xl*z;
        x[n-1-i]=xm+xl*z;
        w[i]=2.0*xl/((1.0-z*z)*pp*pp);
        w[n-1-i]=w[i];
    }
}

/*
 *  \brief Quadrature tools...
 */  
void TrajGen::gauleg2(const long double x1, const long double x2, long double x[], long double w[], int n) const
{
    const long double EPS=1.0e-14;
    int m,j,i;
    long double z1,z,xm,xl,pp,p3,p2,p1;
    
    m=(n+1)/2;
    xm=0.5*(x2+x1);
    xl=0.5*(x2-x1);
    for (i=0;i<m;i++) {
        z=cos(3.141592654*(i+0.75)/(n+0.5));
        do {
            p1=1.0;
            p2=0.0;
            for (j=0;j<n;j++) {
                p3=p2;
                p2=p1;
                p1=((2.0*j+1.0)*z*p2-j*p3)/(j+1);
            }
            pp=n*(z*p1-p2)/(z*z-1.0);
            z1=z;
            z=z1-p1/pp;
        } while (fabs(z-z1) > EPS);
        x[i]=xm-xl*z;
        x[n-1-i]=xm+xl*z;
        w[i]=2.0*xl/((1.0-z*z)*pp*pp);
        w[n-1-i]=w[i];
    }
}


/***************** Initialization *****************/

/**
 *  \brief set the initial and final state
 *  \param x0 : initial state
 *  \param xf : final state
 */
void TrajGen::setStateT(Vec_DP& x0,Vec_DP& xf)
{
 //Set initial, final states and curvature constraint
  vecState_x0 = x0;
  vecState_xf = xf;
}

/**
 *  \brief set the initial and final states
 *  \param x0 : initial state
 *  \param xf : final state
 */
void TrajGen::setStateT(const double* x0,const double* xf)
{
  vecState_x0 = Vec_DP(xLength);
  for(int i=0; i<xLength; i++) vecState_x0[i] = x0[i];
  vecState_xf = Vec_DP(xLength);
  for(int i=0; i<xLength; i++) vecState_xf[i] = xf[i];
}


/************* Trajectory Generation **************/

/*
 *  \brief initialize the trajectory generation process
 */
void TrajGen::initTrajGen()
{
  gLength = initFixedSParam();
//   cout << "gLength : " << gLength << endl;
//   cout << "pLengthFree() : " << pLengthFree() << endl;

  vecConstraints_g = Vec_DP(gLength);
  vecCorrParameters_p = Vec_DP(pLengthFree());
  matJacobianG = Mat_DP(gLength,pLengthFree());

  this->residual_        = 1000000;
  this->oldResidual_     = 0;
  this->coefCor_         = this->COEF_COR_;
  this->nbDeacreaseStep_ = 0;
  this->vecBestParam_    = Vec_DP(pLengthFree());
  
  this->bestResidual_    = -1;
  this->CNT_             = 0;
  this->bestCNT_         = -1;
}

/*
 *  \brief generate a trajectory by constraints optimiation
 *  \param x0 : initial state
 *  \param xf : final state
 *  \param parameters : initial guess of parameters
 *  \return 0 if a solution has been found, -1 else
 */
int TrajGen::ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP** parameters)
{
  // set the system parameters
  setSysParams(sys_);

  // initialize the trajectory generation process
  initTrajGen();

  //Set the initial and final states
  setStateT(x0,xf);

  if(DBG_ >= 1) { cout << "trajGen : "; this->displayInitState(); cout << " -> "; this->displayGoalState(); cout << endl; }

//   cout << "init and goal states : " << endl;
//   displayInitState();
//   displayGoalState();

  //Heuristic guess
  ComputeIG(parameters);
  if(DBG_ >= 1) { cout << "Initial Guess :  "; this->displayParameters(); cout << endl; }

  // compute the state reached by the initial guess of parameters
  ForwardSolution();
  if(DBG_ >= 1) { cout << "Simulated state = "; this->displayStateReached(); cout << endl; }
  if(DBG_ >= 3) { this->displayProfiles(); }

  // compute the error between the goal state and the reached state
  ComputeG();
  if(DBG_ >= 1) { cout << "Error = "; this->displayError(); cout << endl; }

  // compute the first residual 
  residual_ = this->ComputeResidual();
  if(DBG_ >= 1) { cout << "step : " << CNT_ << "\t residual : " << residual_ << endl; }

  while(residual_ > maxResidual_)
    {
      // save the best parameters
      if(CNT_ == 0 || residual_ < bestResidual_)
	this->SaveBestParams();
      
      // increase the number of optimization steps
      CNT_++;
      oldResidual_ = residual_;

      // compute the partial derivatives of the state wrt the parameters
      if(FDpartDeriv_ == false) ComputeJacobianS();
      else ComputeJacobianG_FD();
      if(DBG_ >= 4) { this->displayJacobian(); }

      // invert the jabobian matrix
      ComputeInvJacobianS();
      if(DBG_ >= 4) { this->displayInvJacobian(); }

      //Compute the correction
      this->ComputeCorrection();
      if(DBG_ >= 1) { cout << "Correction = "; this->displayCorrection(); cout << endl; }      

      //Apply the correction
      this->ApplyCorrection();
      if(DBG_ >= 1) { cout << "Parameters :  "; this->displayParameters(); cout << endl; }

      // compute the new state reached by the forward solution
      ForwardSolution();
      if(DBG_ >= 1) { cout << "Simulated state = "; this->displayStateReached(); cout << endl; }
      if(DBG_ >= 3) { this->displayProfiles(); }

      // compute the error between the goal state and the reached state
      ComputeG();
      if(DBG_ >= 1) { cout << "Error = "; this->displayError(); cout << endl; }

      // compute the current residual
      residual_ = this->ComputeResidual();
      if(DBG_ >= 1) { cout << "step : " << CNT_ << "\t residual : " << residual_ << endl; }

      if(CNT_ >= maxNbStep_ && residual_ > maxResidual_)
	{
	  if(DBG_ >= 1) { cout << "  No traj found in " << maxNbStep_ << " iterations" << endl; }

	  // get back the best parameters found
	  LoadBestParams();

	  // compute the forward method with these parameters
	  ForwardSolution();

	  // compute the associated error
	  ComputeG();

	  // get back errors on X, Y and TH
	  finalResidual_ = bestResidual_;
	  finalNbSteps_ = CNT_;

	  if(DBG_ >= 1) {
	    cout << "Best Solution = "; this->displayParameters(); cout << endl;
	    cout << "Best Residual : " << bestResidual_ << "   Best CNT : " << bestCNT_ << endl;
	  }
	  return -1;
	}

    }

  if(DBG_ >= 1) { cout << "  Nb iterations : " << CNT_ << endl; }
  if(DBG_ == 2) { this->displayProfiles(); }
  finalResidual_ = residual_;
  finalNbSteps_ = CNT_;

  return 0;
}

// /*
//  *  \brief generate a trajectory by constraints optimiation
//  *  \param x0 : initial state
//  *  \param xf : final state
//  *  \param constraints : constraints defined over the system
//  *  \param parameters : initial guess of parameters (NULL if no initial guess is defined)
//  *  \return 0 if a solution has been found, -1 else
//  */
// int TrajGen::ComputeTrajectory(double* x0,double* xf,double* constraints,double** parameters)
// {
//   // init the init state
//   Vec_DP vx0(xLength);
//   for(int i=0; i<xLength; i++) vx0[i] = x0[i];

//   // init the goal state
//   Vec_DP vxf(xLength);
//   for(int i=0; i<xLength; i++) vxf[i] = xf[i];

//   // init the constraints
//   Vec_DP vconstraints(xLength);
//   for(int i=0; i<dLength; i++) vconstraints[i] = constraints[i];

//   // init the parameters (IG)
//   Vec_DP** vparameters = new Vec_DP*;
//   if(*parameters == NULL) *vparameters = NULL;
//   else
//     {
//       *vparameters = new Vec_DP(pLength);
//       for(int i=0; i<pLength; i++) (**vparameters)[i] = (*parameters)[i];
//     }

//   // generate the trajectory
//   int res = this->ComputeTrajectory(vx0,vxf,vconstraints,vparameters);

//   // get back the parameters used to generate the resulting trajectory
//   if(*parameters == NULL) *parameters = new double[pLength];
//   for(int i=0; i<pLength; i++) (*parameters)[i] = (**vparameters)[i];
//   delete *vparameters;
//   delete vparameters;

//   return res;
// }

/*
 *  \brief generate a trajectory by constraints optimiation
 *  \param s0 : initial state
 *  \param sf : final state
 *  \param parameters : initial guess of parameters (NULL if no initial guess is defined)
 *  \return 0 if a solution has been found, -1 else
 */
int TrajGen::ComputeTrajectory(const StateTime& s0,const StateTime& sf,double** parameters)
{
  // init the init state
  Vec_DP vx0(xLength);
  for(int i=0; i<xLength; i++) vx0[i] = s0.at(i); // ------------------------------ !!!!!!!!!!!!!!!!!!!! ne marche pas pour le car-like!!!!!!!!

  // init the goal state
  Vec_DP vxf(xLength);
  for(int i=0; i<xLength; i++) vxf[i] = sf.at(i); // ------------------------------ !!!!!!!!!!!!!!!!!!!! ne marche pas pour le car-like!!!!!!!!

//   // init the constraints
//   Vec_DP vconstraints(xLength);
//   initCstt(sys,vconstraints);

  // init the parameters (IG)
  Vec_DP** vparameters = new Vec_DP*;
  if(parameters == NULL || *parameters == NULL) *vparameters = NULL;
  else
    {
      *vparameters = new Vec_DP(pLength);
      for(int i=0; i<pLength; i++) (**vparameters)[i] = (*parameters)[i];
    }

  // generate the trajectory
  int res = this->ComputeTrajectory(vx0,vxf,vparameters);

  // get back the parameters used to generate the resulting trajectory
  if(parameters != NULL)
    {
      if(*parameters == NULL) *parameters = new double[pLength];
      for(int i=0; i<pLength; i++) (*parameters)[i] = (**vparameters)[i];
    }
//   delete *vparameters;
//   delete vparameters;

  return res;
}

/*
 *  \brief generate a trajectory by constraints optimiation
 *  \param x0 : initial state
 *  \param constraints : constraints defined over the system
 *  \param parameters : initial guess of parameters
 */
void TrajGen::ComputeFromIG(const double* x0,const double* parameters)
{
  // initialize the system parameters
  this->setSysParams(sys_);

  // initialize the trajectory generation process
  this->initTrajGen();

  // initialize the configuration
  this->setInitialConfig(x0);

  // set the initial guess of parameters
  this->setInitialGuess(parameters);

  // compute the state reached by the initial guess of parameters
  this->ForwardSolution();
  if(DBG_ >= 1) { cout << "Simulated state = "; this->displayStateReached(); cout << endl; }
  if(DBG_ >= 2) { this->displayProfiles(); }
}


/****** Trajectory Generation (2nd version) ******/

/**
 *  \brief compute the cost function
 *  \return the current value of the cost function
 */
double TrajGen::costFunction() const
{
  double res=0;
  if(CFtype_ == SIMPLE_CF)
    {
      for(int i=0; i<gLength; i++) res += fabs(vecConstraints_g[i] * weightsCF_[i]);
    }
  else if(CFtype_ == SQUARE_CF)
    {
      for(int i=0; i<gLength; i++) res += vecConstraints_g[i] * vecConstraints_g[i] * weightsCF_[i];
    }
  else // if(CFtype_ == OLD_CF)
    {
      res = this->ComputeResidual();
    }
  return res;
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void TrajGen::ComputeIG(const double* p)
{
  // delete the old parameters
  if(vecParameters_p != NULL) delete vecParameters_p;
  vecParameters_p = new Vec_DP(pLength);

  if(p != NULL)
    {
      for(int i=0; i<pLength; i++) (*vecParameters_p)[i] = p[i];
      return;
    }

  // check if an initial guess database exists
  int resIG = -1;
  if(ig_ != NULL)
    {
      resIG = this->ComputeIGfromDB();
    }

  if(resIG == -1)
    {
      this->ComputeDfltIG();
    }
}

// /**
//  *  \brief compute the jacobian of the cost function wrt parameters p
//  */
// void TrajGen::computeJacobianJ_1()
// {
//   for(int i=0; i<pLengthFree(); i++)
//     {
//       JacobianJ_[i] = 0;
//       for(int j=0; j<gLength; j++)
// 	{
// 	  JacobianJ_[i] += (weightsCF_[j]*matJacobianG[j][i]);
// 	}
//     }
// }

// /**
//  *  \brief compute the jacobian of the cost function wrt parameters p
//  */
// void TrajGen::computeJacobianJ_2()
// {
//   for(int i=0; i<pLengthFree(); i++)
//     {
//       JacobianJ_[i] = 0;
//       for(int j=0; j<gLength; j++)
// 	{
// 	  JacobianJ_[i] += (2.0*weightsCF_[j]*matJacobianG[j][i]*vecConstraints_g[j]);
// 	}
//     }
// }

// /**
//  *  \brief compute the hessian of the cost function wrt parameters p
//  */
// void TrajGen::computeHessianJ_1()
// {
//   for(int i=0; i<pLengthFree(); i++)
//     {
//       for(int j=0; j<pLengthFree(); j++)
// 	{
// 	  HessianJ_[i][j] = 0;
// 	  for(int k=0; k<gLength; k++)
// 	    {
// 	      HessianJ_[i][j] += (weightsCF_[j]*HessianS_[k][i][j]);
// 	    }
// 	}
//     }
// }

// /**
//  *  \brief compute the hessian of the cost function wrt parameters p
//  */
// void TrajGen::computeHessianJ_2()
// {
//   for(int i=0; i<pLengthFree(); i++)
//     {
//       for(int j=0; j<pLengthFree(); j++)
// 	{
// 	  HessianJ_[i][j] = 0;
// 	  for(int k=0; k<gLength; k++)
// 	    {
// 	      HessianJ_[i][j] += (2.0*weightsCF_[j]* (HessianS_[k][i][j]*vecConstraints_g[j] + matJacobianG[k][i]*matJacobianG[k][j]) );
// 	    }
// 	}
//     }
// }

// /**
//  *  \brief compute the jacobian of the cost function wrt parameters p
//  */
// void TrajGen::computeJacobianJ()
// {
//   cout << "weights : ";
//   for(int i=0; i<gLength; i++) cout << weightsCF_[i] << " ";
//   cout << endl;

//   if(CFtype_ == SIMPLE_CF) 
//     computeJacobianJ_1();
//   else // if(CFtype_ == SQUARE_CF)
//     computeJacobianJ_2();
// }

// /**
//  *  \brief compute the hessian of the cost function wrt parameters p
//  */
// void TrajGen::computeHessianJ()
// {
//   if(CFtype_ == SIMPLE_CF) 
//     computeHessianJ_1();
//   else // if(CFtype_ == SQUARE_CF)
//     computeHessianJ_2();
// }

// /**
//  *  \brief compute the inverse vector of the jacobian of the cost function wrt parameters p
//  */
// void TrajGen::computeInvJacobianJ()
// {
//   for(int i=0; i<pLengthFree(); i++)
//     {
//       if(JacobianJ_[i] != 0) invJacobianJ_[i] = 1.0 / JacobianJ_[i];
//       else invJacobianJ_[i] = 0.0;
//     }
// }

// /**
//  *  \brief compute the inverse matrix of the hessian of the cost function wrt parameters p
//  */
// void TrajGen::computeInvHessianJ()
// {
//   int M2 = invHessianJ_.nrows();
//   int N2 = invHessianJ_.ncols();
  
//   for(int i=0; i<M2; i++) for(int j=0; j<N2; j++) invHessianJ_[i][j] = 0;
//   invMatDP(HessianJ_, invHessianJ_);
// }

/**
 *  \brief compute the correction to apply over the set of parameters
 */
void TrajGen::computeNewCorrection()
{
  if (variableCoef_)
    {
      // update the correction coefficient
      if(residual_ <= oldResidual_)
	{
	  nbDeacreaseStep_++;

	  if(nbDeacreaseStep_ >= 4) coefCor_ = 1.0;
	  else if(nbDeacreaseStep_ >= 3) coefCor_ = 0.7;
	  else if(nbDeacreaseStep_ >= 2) coefCor_ = 0.5;
	  else if(nbDeacreaseStep_ >= 1) coefCor_ = 0.35;
	}
      else
	{
	  coefCor_ = COEF_COR_;
	  nbDeacreaseStep_ = 0;
	}
    }

//   cout << "variableCoef_ : " << variableCoef_ << "  coefCor_ : " << coefCor_ << "  COEF_COR_ : " << COEF_COR_ << "  nbDeacreaseStep_ : " << nbDeacreaseStep_ << endl;

  for(int i=0; i<pLengthFree();i++)
    {
      vecCorrParameters_p[i] = 0;
      
      for(int j=0;j<gLength;j++)
	{
 	  vecCorrParameters_p[i] += (-1.0)*coefCor_*vecConstraints_g[j]*(*matInvJacobianG)[i][j];
// 	  vecCorrParameters_p[i] += coefCor_*vecConstraints_g[j]*(*matInvJacobianG)[i][j];
	}
    }
}

// /**
//  *  \brief compute the correction to apply over the set of parameters
//  */
// void TrajGen::computeNewCorrection()
// {
//   if (variableCoef_)
//     {
//       // update the correction coefficient
//       if(residual_ <= oldResidual_)
// 	{
// 	  nbDeacreaseStep_++;

// 	  if(nbDeacreaseStep_ >= 4) coefCor_ = 1.0;
// 	  else if(nbDeacreaseStep_ >= 3) coefCor_ = 0.7;
// 	  else if(nbDeacreaseStep_ >= 2) coefCor_ = 0.5;
// 	  else if(nbDeacreaseStep_ >= 1) coefCor_ = 0.35;
// 	}
//       else
// 	{
// 	  coefCor_ = COEF_COR_;
// 	  nbDeacreaseStep_ = 0;
// 	}
//     }

//   for(int i=0; i<pLengthFree();i++)
//     {
//       vecCorrParameters_p[i] = 0;
      
//       for(int j=0;j<gLength;j++)
// 	{
// 	  vecCorrParameters_p[i] += -1.0*coefCor_*vecJacobianS_[j]*invHessianS_[j][i];
// 	}
//     }
// }

/**
 *  \brief apply the correction to the set of parameters
 */
void TrajGen::applyNewCorrection()
{ 
  for(int i=0; i<pLength-1;i++)
    (*vecParameters_p)[i] += vecCorrParameters_p[i];
  if(tcStatus_ == TIME_FREE)
    {
      (*vecParameters_p)[pLength-1] += vecCorrParameters_p[pLength-1];
      if((*vecParameters_p)[pLength-1] < 0.0) (*vecParameters_p)[pLength-1] = 0.0;
    }
  else if(tcStatus_ == TIME_CSTD)
    {
      (*vecParameters_p)[pLength-1] += 1.0*vecCorrParameters_p[pLength-1];
      if((*vecParameters_p)[pLength-1] < minTime_) (*vecParameters_p)[pLength-1] = minTime_;
      if((*vecParameters_p)[pLength-1] > maxTime_) (*vecParameters_p)[pLength-1] = maxTime_;
    }
}

/**
 *  \brief initialize the trajectory generation process (for the new trajectory generation process)
 */
void TrajGen::initNewTrajGen()
{
  gLength = initFixedSParam();

  if(weightsCF_ != NULL) delete [] weightsCF_;
  weightsCF_ = new double[gLength];
  initWeights();

  vecConstraints_g = Vec_DP(gLength);
  vecCorrParameters_p = Vec_DP(pLengthFree());

//   JacobianJ_ = Vec_DP(pLengthFree());
//   invJacobianJ_ = Vec_DP(pLengthFree());
//   HessianJ_ = Mat_DP(pLengthFree(),pLengthFree());
//   invHessianJ_ = Mat_DP(pLengthFree(),pLengthFree());

  matJacobianG = Mat_DP(gLength,pLengthFree());
//   JacobianS_.init(gLength,pLengthFree());
//   HessianS_.init(gLength,pLengthFree(),pLengthFree());

  vecJacobianS_ = Vec_DP(gLength*pLengthFree());
  matHessianS_ = Mat_DP(pLengthFree(),gLength*pLengthFree());
  invHessianS_ = Mat_DP(gLength*pLengthFree(),pLengthFree());

  this->residual_        = 1000000;
  this->oldResidual_     = 1000000;
  this->coefCor_         = this->COEF_COR_;
  this->nbDeacreaseStep_ = 0;
  this->vecBestParam_    = Vec_DP(pLengthFree());
  
  this->bestResidual_    = -1;
  this->CNT_             = 0;
  this->bestCNT_         = -1;
}

/**
 *  \brief generate a trajectory between two states
 *  \param x0 : initial state-time
 *  \param xf : initial state-time
 *  \param p : initial guess of parameters
 */
int TrajGen::genTraj(const double* x0,const double* xf,const double* p)
{
  // set the system parameters
  setSysParams(sys_);

  // initialize the trajectory generation process
  initNewTrajGen();

  //Set the initial and final states
  setStateT(x0,xf);

  if(DBG_ >= 1) { cout << "trajGen : "; this->displayInitState(); cout << " -> "; this->displayGoalState(); cout << endl; }

  //Heuristic guess
  ComputeIG(p);
  if(DBG_ >= 1) { cout << "Initial Guess :  "; this->displayParameters(); cout << endl; }

  // compute the state reached by the initial guess of parameters
  ForwardSolution();
  if(DBG_ >= 1) { cout << "Simulated state = "; this->displayStateReached(); cout << endl; }
  if(DBG_ >= 3) { this->displayProfiles(); }
//   if(DBG_ >= 2) { this->SaveTrajectory(100); }

  // compute the error between the goal state and the reached state
  ComputeG();
  if(DBG_ >= 1) { cout << "Error = "; this->displayError(); cout << endl; }

  // compute the first residual 
  residual_ = this->costFunction();
  oldResidual_ = residual_;
  if(DBG_ >= 1) { cout << "step : " << CNT_ << "\t residual : " << residual_ << endl; }

  while(residual_ > maxResidual_)
    {
      // save the best parameters
      if(CNT_ == 0 || residual_ < bestResidual_)
	this->SaveBestParams();
      
      // increase the number of optimization steps
      CNT_++;

      // compute the state jacobian
      ComputeJacobianS();
      if(DBG_ >= 4) { this->displayJacobian(); }

//       if(CFtype_ == OLD_CF)
// 	{
      // invert the state jacobian matrix
      ComputeInvJacobianS();
      if(DBG_ >= 4) { this->displayInvJacobian(); }
// 	}

//       if(CFtype_ == SIMPLE_CF || CFtype_ == SQUARE_CF)
// 	{
// 	  // compute the jacobian of the cost function
// 	  computeJacobianJ();
// // 	  if(DBG_ >= 4) 
// 	  { this->displayJacobianJ(); }
	  
// 	  // invert the jacobian
// 	  computeInvJacobianJ();
// // 	  if(DBG_ >= 4) 
// 	    { this->displayInvJacobianJ(); }

      // set the vector representation of the jacobian
//       setVecJacobianS();

//       // compute the hessian of the state vector
//       ComputeHessianS();
//       if(DBG_ >= 4) this->displayHessianS();
      
//       // compute the inverse matrix of the hessian of the state vector
//       ComputeInvHessianS();
//       if(DBG_ >= 4) this->displayInvHessianS();

//       // compute the hessian of the cost function
//       computeHessianJ();
//       if(DBG_ >= 4) 
// 	{ this->displayHessianJ(); }
      
//       // invert the hessian
//       computeInvHessianJ();
//       if(DBG_ >= 4) 
// 	{ this->displayInvHessianJ(); }

      //Compute the correction
      this->computeNewCorrection();
      if(DBG_ >= 1) { cout << "Correction = "; this->displayCorrection(); cout << endl; }

      //Apply the correction
      this->applyNewCorrection();
      if(DBG_ >= 1) { cout << "Parameters :  "; this->displayParameters(); cout << endl; }

      // compute the new state reached by the forward solution
      ForwardSolution();
      if(DBG_ >= 1) { cout << "Simulated state = "; this->displayStateReached(); cout << endl; }
      if(DBG_ >= 3) { this->displayProfiles(); }
//       if(DBG_ >= 2) { this->SaveTrajectory(100); }

      // compute the error between the goal state and the reached state
      ComputeG();
      if(DBG_ >= 1) { cout << "Error = "; this->displayError(); cout << endl; }

      // save the old residual
      oldResidual_ = residual_;

      // compute the current residual
      residual_ = this->costFunction();
      if(DBG_ >= 1) { cout << "step : " << CNT_ << "\t residual : " << residual_ << endl; }

      if(CNT_ >= maxNbStep_ && residual_ > maxResidual_)
	{
	  if(DBG_ >= 1) { cout << "  No traj found in " << maxNbStep_ << " iterations" << endl; }

	  // get back the best parameters found
	  LoadBestParams();

	  // compute the forward method with these parameters
	  ForwardSolution();

	  // compute the associated error
	  ComputeG();

	  // get back errors on X, Y and TH
	  finalResidual_ = bestResidual_;
	  finalNbSteps_ = CNT_;

	  if(DBG_ >= 1) {
	    cout << "Best Solution = "; this->displayParameters(); cout << endl;
	    cout << "Best Residual : " << bestResidual_ << "   Best CNT : " << bestCNT_ << endl;
	  }
	  return -1;
	}
      
    }
  
  if(DBG_ >= 1) { cout << "  Nb iterations : " << CNT_ << endl; }
  if(DBG_ == 2) { this->displayProfiles(); }
  finalResidual_ = residual_;
  finalNbSteps_ = CNT_;

  return 0;
}

/**
 *  \brief generate a trajectory between two states
 *  \param s0 : initial state-time
 *  \param sf : initial state-time
 *  \param p : initial guess of parameters
 */
int TrajGen::genTraj(const StateTime& s0,const StateTime& sf,const double* p)
{
  double x0[xLength];
  double xf[xLength];
  InitGoalST2Tab(s0,sf,x0,xf);

  return this->TrajGen::genTraj(x0,xf,p);
}


/*************** Utility Functions ****************/

/*
 *  \brief display the jacobian matrix
 */
void TrajGen::displayJacobian()
{
//   int pl = 0;
//   if (tcStatus_ == TIME_FIXED) pl = pLength-1;
//   else pl = pLength;

  cout << "JacobianG : " << endl;
  for(int i=0; i<gLength; i++)
    {
//       for(int j=0; j<pl; j++)
      for(int j=0; j<pLengthFree(); j++)
	{
	  cout << matJacobianG[i][j] << " ";
	}
      cout << endl;
    }
  cout << endl;
}

/*
 *  \brief display the inverse of the jacobian matrix
 */
void TrajGen::displayInvJacobian()
{
//   int pl = 0;
//   if (tcStatus_ == TIME_FIXED) pl = pLength-1;
//   else pl = pLength;

  cout << "InvJacobianG : " << endl;
//   for(int i=0; i<pl; i++)
  for(int i=0; i<pLengthFree(); i++)
    {
      for(int j=0; j<gLength; j++)
	{
	  cout << matInvJacobianG[i][j] << " ";
	}
      cout << endl;
    }
  cout << endl;
}

/*
 *  \brief display the set of parameters
 */
void TrajGen::displayParameters()
{
  cout << "(";
  for(int i=0; i<pLength-1; i++)
    {
      cout << (*vecParameters_p)[i] << ",";
    }
  if(pLength!=0) cout << (*vecParameters_p)[pLength-1];
  cout << ")";
}

/*
 *  \brief display the initial state
 */
void TrajGen::displayInitState()
{
  cout << "(";
  for(int i=0; i<xLength-1; i++)
    {
      cout << vecState_x0[i] << ",";
    }
  if(xLength!=0) cout << vecState_x0[xLength-1];
  cout << ")";
}

/*
 *  \brief display the goal state
 */
void TrajGen::displayGoalState()
{
  cout << "(";
  for(int i=0; i<xLength-1; i++)
    {
      cout << vecState_xf[i] << ",";
    }
  if(xLength!=0) cout << vecState_xf[xLength-1];
  cout << ")";
}

/*
 *  \brief display the state reached at the current step
 */
void TrajGen::displayStateReached()
{
  int N;
  if(tcStatus_ == TIME_FIXED) N = xLength-1;
  else N = xLength;

  cout << "(";
  for(int i=0; i<N-1; i++)
    {
      cout << vecState_xsim[i] << ",";
    }
  if(N>0) cout << vecState_xsim[N-1];
  cout << ")";
}

/*
 *  \brief display the correction computed
 */
void TrajGen::displayCorrection()
{
//   int pl = 0;
//   if (tcStatus_ == TIME_FIXED) pl = pLength-1;
//   else pl = pLength;
  
//   for(int i=0; i<pl-1; i++)
  cout << "(";
  for(int i=0; i<pLengthFree()-1; i++)
    {
      cout << vecCorrParameters_p[i] << ",";
    }
  //if(pl!=0) cout << vecCorrParameters_p[pl-1];
  if(pLengthFree()>0) cout << vecCorrParameters_p[pLengthFree()-1];
  cout << ")";
}

/*
 *  \brief display the error between the goal state and the reached state
 */
void TrajGen::displayError()
{
  cout << "(";
  for(int i=0; i<gLength-1; i++)
    {
      cout << vecConstraints_g[i] << ",";
    }
  if(gLength!=0) cout << vecConstraints_g[gLength-1];
  cout << ")";
}

// /*
//  *  \brief display the jacobian of the cost function
//  */
// void TrajGen::displayJacobianJ()
// {
//   cout << "JacobianJ : (";
//   for(int i=0; i<pLengthFree()-1; i++)
//     {
//       cout << JacobianJ_[i] << ",";
//     }
//   if(pLengthFree()!=0) cout << JacobianJ_[pLengthFree()-1];
//   cout << ")" << endl;
// }

// /*
//  *  \brief display the inverse of the jacobian of the cost function
//  */
// void TrajGen::displayInvJacobianJ()
// {
//   cout << "invJacobianJ : (";
//   for(int i=0; i<pLengthFree()-1; i++)
//     {
//       cout << invJacobianJ_[i] << ",";
//     }
//   if(pLengthFree()!=0) cout << invJacobianJ_[pLengthFree()-1];
//   cout << ")" << endl;
// }

// /*
//  *  \brief display the hessian of the cost function
//  */
// void TrajGen::displayHessianJ()
// {
//   for(int i=0; i<pLengthFree(); i++)
//     {
//       for(int j=0; j<pLengthFree(); j++)
// 	{
// 	  cout << HessianJ_[i][j] << " ";
// 	}
//       cout << endl;
//     }
// }

// /*
//  *  \brief display the inverse of the hessian of the cost function
//  */
// void TrajGen::displayInvHessianJ()
// {
//   for(int i=0; i<pLengthFree(); i++)
//     {
//       for(int j=0; j<pLengthFree(); j++)
// 	{
// 	  cout << invHessianJ_[i][j] << " ";
// 	}
//       cout << endl;
//     }
// }

/**
 *  \brief display the hessian of the state vector
 */
void TrajGen::displayHessianS()
{
  cout << "HessianS : " << endl;
  for(int i=0; i<gLength; i++)
    {
      for(int j=0; j<pLengthFree(); j++)
	{
	  for(int k=0; k<pLengthFree(); k++)
	    cout << getHessianS(i,j,k) << " ";
	  cout << endl;
	}
      cout << endl;
    }
}

/**
 *  \brief display the inverse of the hessian of the state vector
 */
void TrajGen::displayInvHessianS()
{
  int M = gLength*pLengthFree();
  int N = pLengthFree();

  cout << "InvHessianS : " << endl;
  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  cout << invHessianS_[i][j] << " ";
	}
      cout << endl;
    }
}
