#include "trajSel/inc/ManaTrajSel.hpp"
using namespace mpc;

#include "gnuplot/inc/TrajPlot.hpp"

#include <iostream>
using namespace std;

#define TS_ACC_SAMPLING_MODE 0 
#define TS_JERK_SAMPLING_MODE 1 

#define COEF_BREAK_DIST 1.5

                                    /**************************************************/
                                    /**************** Class ManaTrajSel ***************/
                                    /**************************************************/

double ManaTrajSel::TS_WG = 1.0;                               /**< Goal position heuritic weight */
double ManaTrajSel::TS_WO = 0.2;                               /**< Obstacles heuritic weight */
double ManaTrajSel::TS_WH = 1.0;                               /**< Goal heading heuritic weight */

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
ManaTrajSel::ManaTrajSel() : TrajSel()
{
  this->spgAcc_ = 5;
  this->spgAAcc_ = 5;
  this->spgJerk_ = 3;
  this->spgAJerk_ = 3;

  this->modeCtrlSpg_ = TS_ACC_SAMPLING_MODE;

  this->coefAcc_ = 0.3;
  this->coefAAcc_ = 0.8;
  this->coefJerk_ = 2.0;
  this->coefAJerk_ = 3.0;
}

/**
 *  \brief default constructor
 *  \param sys : the robotic system constraints
 *  \param spgAcc : sampling over the linear acceleration control
 *  \param spgAAcc : sampling over the angular acceleration control
 */
ManaTrajSel::ManaTrajSel(ManaSys* sys, int spgAcc, int spgAAcc) : TrajSel(sys)
{
  this->spgAcc_ = spgAcc;
  this->spgAAcc_ = spgAAcc;
  this->spgJerk_ = 0;
  this->spgAJerk_ = 0;

  this->modeCtrlSpg_ = TS_ACC_SAMPLING_MODE;

  this->coefAcc_ = 0.3;
  this->coefAAcc_ = 0.8;
  this->coefJerk_ = 2.0;
  this->coefAJerk_ = 3.0;
}

/**
 *  \brief default constructor
 *  \param sys : the robotic system constraints
 *  \param spgAcc : sampling over the linear acceleration control
 *  \param spgAAcc : sampling over the angular acceleration control
 *  \param spgJerk : sampling over the linear jerk control
 *  \param spgAJerk : sampling over the angular jerk control
 *  \param coefAcc : coefficient of the linear acceleration control used
 *  \param coefAAcc : coefficient of the angular acceleration control used
 *  \param coefJerk : coefficient of the linear jerk control used
 *  \param coefAJerk : coefficient of the angular jerk control used
 */
ManaTrajSel::ManaTrajSel(ManaSys* sys, int spgAcc, int spgAAcc, int spgJerk, int spgAJerk, 
			 double coefAcc, double coefAAcc, double coefJerk, double coefAJerk) : TrajSel(sys)
{
  this->spgAcc_ = spgAcc;
  this->spgAAcc_ = spgAAcc;
  this->spgJerk_ = spgJerk;
  this->spgAJerk_ = spgAJerk;

  this->modeCtrlSpg_ = TS_JERK_SAMPLING_MODE;

  this->coefAcc_ = coefAcc;
  this->coefAAcc_ = coefAAcc;
  this->coefJerk_ = coefJerk;
  this->coefAJerk_ = coefAJerk;
}

/**
 *  \brief destructor
 */
ManaTrajSel::~ManaTrajSel()
{
}


/**************** Vector management ***************/

/*
 *  \brief return a pointer the the i-th candidate control trajectory
 *  \return a pointer the the i-th candidate control trajectory
 */
const ManaPCTraj* ManaTrajSel::getCandCTraj(int i) const
{
  return ((ManaPCTraj*)(vecPCTraj_.at(i)));
}

/*
 *  \brief return a pointer the the i-th candidate state trajectory
 *  \return a pointer the the i-th candidate state trajectory
 */
const ManaPSTraj* ManaTrajSel::getCandSTraj(int i) const
{
  return ((ManaPSTraj*)(vecPSTraj_.at(i)));
}


/************* Trajectory computation *************/

/*
 *  \brief compute the set of control trajectories
 *  \param distToGoal : distance from the current position of the robot to the goal location
 *  \param distToObs : distance from the current position of the robot to the closest obstacle
 *  \return the number of computed control trajectories
 */
int ManaTrajSel::computeCTrajSet(double distToGoal, double distToObs)
{
  if(modeCtrlSpg_ == TS_ACC_SAMPLING_MODE) return this->computeCTrajSetAcc();
  else if(modeCtrlSpg_ == TS_JERK_SAMPLING_MODE) return this->computeCTrajSetJerk(distToGoal, distToObs);
  else
    {
      cerr << "ManaTrajSel::computeCTrajSet : Unknown control sampling mode!" << endl;
      exit(1);
    }
}

/*
 *  \brief compute the set of control trajectories (with acceleration control)
 *  \return the number of computed control trajectories
 */
int ManaTrajSel::computeCTrajSetAcc()
{
  // get back the constraints of the robotic system
  double minAcc = -sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minAAcc = -sys()->getAAMax();
  double maxAAcc = sys()->getAAMax();

  double minVel = 0;
  double maxVel = sys()->getVMax();
  double minAVel = -sys()->getAVMax();
  double maxAVel = sys()->getAVMax();

//   cout << "minAcc : " << minAcc << endl;
//   cout << "maxAcc : " << maxAcc << endl;
//   cout << "minAAcc : " << minAAcc << endl;
//   cout << "maxAAcc : " << maxAAcc << endl;
//   cout << "minVel : " << minVel << endl;
//   cout << "maxVel : " << maxVel << endl;
//   cout << "minAVel : " << minAVel << endl;
//   cout << "maxAVel : " << maxAVel << endl;

  double minAccUsed = minAcc *coefAcc_;
  double maxAccUsed = maxAcc *coefAcc_;
  double minAAccUsed = minAAcc *coefAAcc_;
  double maxAAccUsed = maxAAcc *coefAAcc_;

  double stepAcc = (maxAccUsed - minAccUsed) / ((double) (spgAcc_-1));
  double stepAAcc = (maxAAccUsed - minAAccUsed) / ((double) (spgAAcc_-1));
  
  // get back the last saved robotic state
  ManaState curS = *(this->loc());

  for(int i=0; i<spgAcc_; i++)
    {
      // compute the current linear acceleration control
      double acc = minAccUsed + i * stepAcc;

      for(int j=0; j<spgAAcc_; j++)
	{
	  // compute the current angular acceleration control
	  double aacc = minAAccUsed + j * stepAAcc;

	  if(curS.v() != 0 || acc >= 0)
	    {
	      // create a new candidate trajectory from a sampling of the available input command
	      ManaPCTraj* ctraj = new ManaPCTraj(curS.v(), acc, curS.w(), aacc, TS_TIME_TRAJ);
	      
// 	      cout << "traj ->  v : " << curS.v() << "  acc : " << acc << "  w : " << curS.w() << "  aacc : " << aacc << endl;
	      
	      // add all motion constraints over this trajectory
	      ctraj->addAccBounds(minAcc,maxAcc);
	      ctraj->addAAccBounds(minAAcc,maxAAcc);
	      ctraj->addVelBounds(minVel,maxVel);
	      ctraj->addAVelBounds(minAVel,maxAVel);
	      
	      // add it to the set of candidate trajectories
	      this->addCandCTraj(ctraj);
	    }

	}
    }
}

/*
 *  \brief compute the set of control trajectories (with jerk control)
 *  \param distToGoal : distance from the current position of the robot to the goal location
 *  \param distToObs : distance from the current position of the robot to the closest obstacle
 *  \return the number of computed control trajectories
 */
int ManaTrajSel::computeCTrajSetJerk(double distToGoal, double distToObs)
{
  // compute the traveled distance to break from the maximal speed of the robot
  double timeBreak;
  double distBreak;
  mpc::minTDtoReachSpd(sys()->getVMax(), 0, sys()->getAMax(), timeBreak, distBreak);
//   cout << "distBreak : " << distBreak << endl;
//   cout << "COEF_BREAK_DIST * distBreak : " << COEF_BREAK_DIST * distBreak << endl;

  double coefMinVel_BD = 0.1;
  double coefMaxVel_BD = 1.0;
  double coefMaxVel_BD_tmp = 1.0;
  // reduce the robot's speed to break when arriving close to the goal position
  if(distToGoal < COEF_BREAK_DIST * distBreak) coefMaxVel_BD = coefMinVel_BD + ((coefMaxVel_BD - coefMinVel_BD) * distToGoal / (COEF_BREAK_DIST * distBreak));
//   cout << "distToGoal : " << distToGoal << "   COEF_BREAK_DIST * distBreak : " << COEF_BREAK_DIST * distBreak << "   coefMaxVel_BD : " << coefMaxVel_BD << endl;

  // reduce the robot's speed to break when arriving close to an obstacle
  if(distToObs < TrajSel::TS_DIST_OBS_DECREASED_SPD)
    {
      coefMaxVel_BD_tmp = coefMinVel_BD + ((coefMaxVel_BD - coefMinVel_BD) * distToObs / TrajSel::TS_DIST_OBS_DECREASED_SPD);
      if(coefMaxVel_BD_tmp < coefMaxVel_BD) coefMaxVel_BD = coefMaxVel_BD_tmp;
    }

  // get back the constraints of the robotic system
  double minAcc = -sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minAAcc = -sys()->getAAMax();
  double maxAAcc = sys()->getAAMax();

  double minVel = 0;
  double maxVel = sys()->getVMax() * coefMaxVel_BD;
//   double maxVel = sys()->getVMax();
  double minAVel = -sys()->getAVMax();
  double maxAVel = sys()->getAVMax();

  double minAccUsed = minAcc * coefAcc_;
  double maxAccUsed = maxAcc * coefAcc_;
  double minAAccUsed = minAAcc * coefAAcc_;
  double maxAAccUsed = maxAAcc * coefAAcc_;

  double minJerkUsed = -coefJerk_ * maxAccUsed / TS_TIME_TRAJ;
  double maxJerkUsed = coefJerk_ * maxAccUsed / TS_TIME_TRAJ;
  double minAJerkUsed = -coefAJerk_ * maxAAccUsed / TS_TIME_TRAJ;
  double maxAJerkUsed = coefAJerk_ * maxAAccUsed / TS_TIME_TRAJ;

  double stepAcc = (maxAccUsed - minAccUsed) / ((double) (spgAcc_-1));
  double stepAAcc = (maxAAccUsed - minAAccUsed) / ((double) (spgAAcc_-1));

  double stepJerk = (maxJerkUsed - minJerkUsed) / ((double) (spgJerk_-1));
  double stepAJerk = (maxAJerkUsed - minAJerkUsed) / ((double) (spgAJerk_-1));

  // get back the last saved robotic state
  ManaState curS = *(this->loc());

  for(int i=0; i<spgAcc_; i++)
    {
      // compute the current linear acceleration control
      double acc = minAccUsed + i * stepAcc;

      for(int j=0; j<spgAAcc_; j++)
	{
	  // compute the current angular acceleration control
	  double aacc = minAAccUsed + j * stepAAcc;

	  for(int k=0; k<spgJerk_; k++)
	    {
	      // compute the current linear jerk control
	      double jerk = minJerkUsed + k * stepJerk;

	      for(int l=0; l<spgAJerk_; l++)
		{
		  // compute the current angular jerk control
		  double ajerk = minAJerkUsed + l * stepAJerk;

		  if(curS.v() != 0 || acc >= 0 || jerk >=0)
		    {
		      // create a new candidate trajectory from a sampling of the available input command
		      ManaPCTraj* ctraj = new ManaPCTraj(curS.v(), Polynomial(jerk, acc), curS.w(), Polynomial(ajerk, aacc), TS_TIME_TRAJ);

// 		      cout << "traj ->  v : " << curS.v() << "  acc : " << acc << "  w : " << curS.w() << "  aacc : " << aacc << "  jerk : " << jerk << "  ajerk : " << ajerk << endl;
	      
		      // add all motion constraints over this trajectory
		      ctraj->addAccBounds(minAcc,maxAcc);
		      ctraj->addAAccBounds(minAAcc,maxAAcc);
		      ctraj->addVelBounds(minVel,maxVel);
		      ctraj->addAVelBounds(minAVel,maxAVel);
	      
		      // add it to the set of candidate trajectories
		      this->addCandCTraj(ctraj);
		    }
		  
		}
	    }
	}
    }
}

/*
 *  \brief compute the set of state trajectories
 *  \return the number of computed state trajectories
 */
int ManaTrajSel::computeSTrajSet()
{
  // get back the last saved robotic state
  ManaState curS = *(this->loc());

  int N = nbCandTraj();
  for(int i=0; i<N; i++)
    {
      // compute the candidate state trajectories from the current control trajectory
      ManaPSTraj* straj = new ManaPSTraj(curS, *(this->getCandCTraj(i)));

      // add it to the vector of candidate state trajectories
      this->addCandSTraj(straj);
    }
}


/************* Heurisctics computation ************/

/*
 *  \brief compute the heuristic of an input trajectory
 *  \param curS : the current state of the robot
 *  \param targetPos : the target location
 *  \param ctraj : the input control trajectory
 *  \param straj : the input state trajectory
 */
float ManaTrajSel::computeH(const ManaState& curS, const PositionXY& targetPos, 
	       const ManaPCTraj* ctraj, const ManaPSTraj* straj) const
{
  // get back the final state of the trajectory
  ManaStateT finalS = straj->lastStateT();
//   cout << "finalS : "; finalS.display();
  ManaStateT finalS_Obs = straj->at(TS_TIME_TRAJ_OBS_FREE);
//   cout << "finalS_Obs : "; finalS_Obs.display();

  float HG = this->computeHG(curS, finalS, finalS_Obs, targetPos, straj);
  float HO = this->computeHO(curS, finalS_Obs, straj);
  float HH = this->computeHH(finalS, targetPos);
  float H_static = this->computeH_static(curS, finalS);

//   cout << "HG : " << HG << "  HO : " << HO << "  HH : " << HH << "  H : " << (HG + HO + HH) << endl;
//   cout << "    HG : " << HG << "  HO : " << HO << "  H : " << (HG + HO) << endl;
  
//   return (HG + HO + HH);
  return (HG + HO + H_static);
//   return HO;
}

/*
 *  \brief compute the heuristic of an input trajectory : distance to the goal
 *  \param curS : the current state of the robot
 *  \param finalS : the final state of a trajectory
 *  \param targetPos : the target location
 *  \param straj : the input state trajectory
 */
float ManaTrajSel::computeHG(const ManaState& curS, const ManaStateT& finalS, const ManaStateT& finalS_Obs, const PositionXY& targetPos, const ManaPSTraj* straj) const
{
  // compute the distance from the final state of the current trajectory to the goal state
  double dx1 = finalS.x() - targetPos.x();
  double dy1 = finalS.y() - targetPos.y();
  double dist = sqrt(dx1*dx1 + dy1*dy1);

  // compute the distance from the current state to the goal state
  double dxCT = curS.x() - targetPos.x();
  double dyCT = curS.y() - targetPos.y();
  double distCT = sqrt(dxCT*dxCT + dyCT*dyCT);

  // compute the distance from the current state to the goal state
  double dxCF = curS.x() - finalS.x();
  double dyCF = curS.y() - finalS.y();
  double distCF = sqrt(dxCF*dxCF + dyCF*dyCF);

  if(distCT < distCF)
    {
      // compute the closest position of the current candidate trajectory to the goal position
      double minSqDist = 999999999.9;

      double dt = TS_TIME_TRAJ / ((double)(TS_SAMPLING_CAND_TRAJ-1.0));

      for(int i=0; i<TS_SAMPLING_CAND_TRAJ; i++)
	{
	  // get back the current position along the trajectory
	  ManaStateT statet = straj->at(i * dt);

	  double curDX = statet.x() - targetPos.x();
	  double curDY = statet.y() - targetPos.y();
	  double curSqDist = curDX*curDX + curDY*curDY;

	  if(curSqDist < minSqDist) minSqDist = curSqDist;
	}

      return TS_WG * (sqrt(minSqDist) / distCT);
    }
  else
    { 
      //   return TS_WG * (dist1 / dist2) * DIST_OBS_MAX;
      return TS_WG * (dist / distCT);
    }
}

/*
 *  \brief compute the heuristic of an input trajectory : distance to obstacles
 *  \param wp : waypoint (state-time) of a trajectory
 *  \param coefVel : velocity coeficient used to prevent high velocities when the robot is close to obstacles
 */
float ManaTrajSel::computeHO_pos(const ManaStateT& wp, double coefVel) const
{
  // search for the distance to obstacles of the current position on the distmap
  obs2D_cell_t cellSt; 
  float dist; 
  PositionXY near;
  distMap_->getInfo(wp.x(), wp.y(), cellSt, dist, near);
  
  // compute the heuristic
  if(dist <= 0)
    return INFINITY;
  if(dist < TS_DIST_OBS_MAX)
    return ((coefVel * TS_DIST_OBS_MAX) / dist);
  else if(dist < 2.0*TS_DIST_OBS_MAX)
    return (coefVel * (1.0 - (dist - TS_DIST_OBS_MAX)/TS_DIST_OBS_MAX));
  else
    return 0;
}

/*
 *  \brief compute the heuristic of an input trajectory : distance to obstacles
 *  \param curS : the current state of the robot
 *  \param finalS : the final state of a trajectory
 *  \param straj : the input state trajectory
 */
float ManaTrajSel::computeHO(const ManaState& curS, const ManaStateT& finalS_Obs, const ManaPSTraj* straj) const
{
//   double minVel = 0;
//   double maxVel = sys()->getVMax();
//   double minAcceptedVel = 0.3*sys()->getVMax();

//   double vmoy = 0.5 * (curS.v() + finalS.v());
  double coefVel = TS_WO;
//   if(vmoy < minAcceptedVel) coefVel = TS_WO;
//   else coefVel = TS_WO + ((1.0-TS_WO)*((fabs(vmoy) - minAcceptedVel) / (maxVel - minAcceptedVel)));

//   cout << "    vmoy : " << vmoy << "  coefVel : " << coefVel;

//   // sample the input state trajectory
//   ManaSTraj* sampledTraj = straj->sample(SAMPLING_CAND_TRAJ);

  double dt = TS_TIME_TRAJ_OBS_FREE / ((double)(TS_SAMPLING_CAND_TRAJ-1.0));

  // search for the minimal distance to obstacles along the trajectory
  double minDist = 999999999.9;

  float sumHO = 0;

  for(int i=0; i<TS_SAMPLING_CAND_TRAJ; i++)
    {
      // get back the current position along the trajectory
      ManaStateT statet = straj->at(i * dt);
//       sampledTraj->at(i);

      float curHO = this->computeHO_pos(statet, coefVel);
      if(curHO == INFINITY) return INFINITY;
      else sumHO += curHO;
    }

  return (sumHO / TS_SAMPLING_CAND_TRAJ);
}

/*
 *  \brief compute the heuristic of an input trajectory : final heading
 *  \param finalS : the final state of a trajectory
 *  \param targetPos : the target location
 */
float ManaTrajSel::computeHH(const ManaStateT& finalS, const PositionXY& targetPos) const
{
  // compute the angle from the final position of the trajectory to the goal one
  double dx1 = targetPos.x() - finalS.x();
  double dy1 = targetPos.y() - finalS.y();
  double a1 = atan2(dy1,dx1);

  // get back the heading of the final state
  double a2 = finalS.th();

  double diffHead = fabs(computeMinDiffAngle(a1,a2));

  // compute the heuristic
  if(diffHead < 0.5*M_PI)
    return 0;
  else
    return ((diffHead - 0.5*M_PI) / (0.5*M_PI)) * TS_WH;
}

/*
 *  \brief compute an heuristic that penalize trajectory with no motion
 *  \param curS : the current state of the robot
 *  \param finalS : the final state of a trajectory
 */
float ManaTrajSel::computeH_static(const ManaState& curS, const ManaStateT& finalS) const
{
  if(curS.v() == 0 && finalS.v() == 0) return 1.0;
  else return 0.0;
}


/************** Trajectory selection **************/

/*
 *  \brief compute heuristic of the trajectory corresponding to index i
 *  \param i : the input index
 *  \return heuristic corresponding to the i-th trajectory
 */
float ManaTrajSel::computeH(int i) const
{
  // get back the current state
  ManaState curS = *(this->loc());

  // get back the target postion
  PositionXY targetPos = this->targetPos_;

  // get back the i-th candidate control trajectory
  const ManaPCTraj* ctraj = this->getCandCTraj(i);

  // get back the i-th candidate state trajectory
  const ManaPSTraj* straj = this->getCandSTraj(i);

  // compute the heuristic
  float H = this->computeH(curS, targetPos, ctraj, straj);
  return H;
}

/**
 *  \brief compute the best trajectory to follow
 *  \return 0 if the process succeedm -1 else
 *
 *  If the input dem is not null, update the dem
 *  Do not use this option if you want asynchronous trajectory selesction and DEM update
 */
int ManaTrajSel::computeBestTraj(const ManaState& curS, const PositionXY& tgt, const DEM* dem)
{
//   cout << "ManaTrajSel::computeBestTraj 1" << endl;
  // update the current state
  this->updateLoc(curS);

//   cout << "ManaTrajSel::computeBestTraj 2" << endl;
  this->startTimerTrajCpted();

//   cout << "ManaTrajSel::computeBestTraj 3" << endl;
  // update the target position
  this->updateTgt(tgt);

//   cout << "ManaTrajSel::computeBestTraj 4" << endl;
  // update the DEM if needed
  if(dem != NULL) this->updateDEM(dem);

//   cout << "ManaTrajSel::computeBestTraj 5" << endl;
  // compute the best trajectory
  return this->computeBestTrajectory();
}


/****************** I/O functions *****************/

/**
 *  \brief return a copy of the best control trajectory found
 *  \return a copy of the best control trajectory found, or NULL if the process failed
 */
ManaPCTraj* ManaTrajSel::getBestCTraj() const
{
  if(bestPCTraj_ == NULL) return NULL;

  ManaPCTraj* resTraj = new ManaPCTraj(*bestCTraj());
  resTraj->shift(dateTraj_.tv_sec + 0.000001*dateTraj_.tv_usec);

  return resTraj;
}

/**
 *  \brief return a copy of the best state trajectory found
 *  \return a copy of the best state trajectory found, or NULL if the process failed
 */
ManaPSTraj* ManaTrajSel::getBestSTraj() const
{
  if(bestPSTraj_ == NULL) return NULL;

  ManaPSTraj* resTraj = new ManaPSTraj(*bestSTraj());
  resTraj->shift(dateTraj_.tv_sec + 0.000001*dateTraj_.tv_usec);

  return resTraj;
}

/**
 *  \brief return a dated copy of the best control trajectory found
 *  \param pctraj : the resulting copy of the best control trajectory found, or NULL if the process failed
 *  \param date : the date at which the trajectory has been registered
 */
void ManaTrajSel::getBestCTraj(ManaPCTraj*& pctraj, struct timeval& date) const
{
  if(bestPCTraj_ == NULL) 
    {
      pctraj = NULL;
      date.tv_sec = -1;
      date.tv_usec = -1;
    }
  else
    {
      pctraj = new ManaPCTraj(*bestCTraj());
      date.tv_sec = dateTraj_.tv_sec;
      date.tv_usec = dateTraj_.tv_usec;
    }
}

/**
 *  \brief return a dated copy of the best state trajectory found
 *  \param pstraj : the resulting copy of the best state trajectory found, or NULL if the process failed
 *  \param date : the date at which the trajectory has been registered
 */
void ManaTrajSel::getBestSTraj(ManaPSTraj*& pstraj, struct timeval& date) const
{
  if(bestPSTraj_ == NULL) 
    {
      pstraj = NULL;
      date.tv_sec = -1;
      date.tv_usec = -1;
    }
  else
    {
      pstraj = new ManaPSTraj(*bestSTraj());
      date.tv_sec = dateTraj_.tv_sec;
      date.tv_usec = dateTraj_.tv_usec;
    }
}

/**
 *  \brief return the best control to apply
 *  \return the best control to apply
 */
ManaControl ManaTrajSel::getCtrl()
{
  if(bestPCTraj_ == NULL) return ManaControl(0,0);
  else
    {
      boost::mutex::scoped_lock lock(mutexTimers_);

      // get back the current time
      double curT = timerTrajUsed_->totalElapsedTime();
      
      // return the control to apply at the current time
      ManaControl ctrl = bestCTraj()->at(curT + 0.5);

//       cout << "==============================================================================" << endl;
//       cout << "Best traj : " << endl;
//       bestCTraj()->display();
//       cout << "------------------------------------------------------------------------------" << endl;
//       cout << "curT : " << curT << endl;
//       cout << "Control : "; ctrl.display(); cout << endl;
//       cout << "==============================================================================" << endl;

//       return bestCTraj()->at(curT);
      return ctrl;
    }
}


/**************** Display functions ***************/

/**
 *  \brief display all the candidate trajectories in a gnuplot viewer
 */
void ManaTrajSel::displayGP_CandTraj() const
{
  Gnuplot gPath;
  gPath.reset_plot();

  int N = this->nbCandTraj();
  cout << "N : " << N << endl;

  for(int i=0; i<N; i++)
    {
      TrajPlot::plotDD2Path(gPath, *(this->getCandSTraj(i)), "", 50);
    }

  Gnuplot plotVel;
  plotVel.reset_plot();
  Gnuplot plotAVel;
  plotAVel.reset_plot();
  Gnuplot plotAcc; 
  plotAcc.reset_plot();
  Gnuplot plotAAcc;
  plotAAcc.reset_plot();

  for(int i=0; i<N; i++)
    {
      if (i == 0) TrajPlot::plotDD2PCTraj(plotVel, plotAVel, plotAcc, plotAAcc, *(this->getCandCTraj(i)), this->sys(), 50);
      else TrajPlot::plotDD2PCTraj(plotVel, plotAVel, plotAcc, plotAAcc, *(this->getCandCTraj(i)), NULL, 50);
    }

//   for(int i=0; i<N; i++)
//     {
//       this->getCandCTraj(i)->display();
//       TrajPlot::plotDD2PCTraj(*(this->getCandCTraj(i)), this->sys(), true, 50);
//     }

  wait_for_key();
}
