#ifndef MPC_CONTROLLER_HPP
#define MPC_CONTROLLER_HPP

#include "common/inc/ControlTime.hpp"
#include "common/inc/StateTime.hpp"
#include "common/inc/Timer.hpp"
//using namespace mpc;

#include "hugr.hpp"
//namespace hugr;

#include <string>

#include "hugr/HCommon/inc/HCommon.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

namespace mpc
{
  /**
   * \enum controllerState_t
   * \brief Enumeration of the different controller state types
   *
   * CTRLER_NULL : controller not allocated
   * CTRLER_DISCONNECTED : disconnected controller
   * CTRLER_STOPPED : stopped controller
   * CTRLER_STARTED : started controller
   */
  typedef enum {
    CTRLER_NULL          = -1,
    CTRLER_DISCONNECTED  =  0,
    CTRLER_STOPPED       =  1,
    CTRLER_STARTED       =  2
  } controllerStatus_t;

  /**
   * \enum controllerType_t
   * \brief Enumeration of the different controller types
   *
   * CTRLER_NULL : controller not allocated
   * CTRLER_DISCONNECTED : disconnected controller
   * CTRLER_STOPPED : stopped controller
   * CTRLER_STARTED : started controller
   */
  typedef enum {
    TF2_CTRLER         = -2,
    ND_CTRLER          = -1,
    SIMPLE_CTRLER      =  0,
    TRAJFOLLOW_CTRLER  =  1,
    TEDDY_CTRLER       =  2
  } controller_t;

                                    /**************************************************/
                                    /**************** Class Controller ****************/
                                    /**************************************************/

  /** \class Controller
   * \brief Simple controller for manual driving
   */
  class Controller
  {
  protected : 
    Timer timer_;                                /**< timer used to send a control at constant time intervals */
    double dt_;                                  /**< the interval of time to send controls (in sec.) */
    double dtToWait_;                            /**< the time to wait after sending a control */
    ControlTime* controlt_;                      /**< the current control send */

    controllerStatus_t ctrlerSt_;                /**< the controller state (disconnected, connected and stopped, connected and started) */
    controller_t ctrlerType_;                    /**< the robot's controller type */

    hugr::Store* store_;                         /**< the hugr store used to control the robot */
    hugr::VariableId controlId_;                 /**< store id for the control applied */
    std::string nameRb_;                         /**< the name of the controlled robot */
    std::string commandName_;                    /**< the name of the command variable in the hugr store */

    boost::thread* threadCtrl_;                  /**< the thread used to execute the control process */

    boost::mutex* mutexLoc_;                     /**< mutex used to prevent localization access conflicts */
    boost::mutex* mutexCtrl_;                    /**< mutex used to prevent control access conflicts */
    boost::mutex* mutexTraj_;                    /**< mutex used to prevent trajectory access conflicts */
    boost::mutex* mutexEnv_;                     /**< mutex used to prevent environment model access conflicts */

    int debug_;                                  /**< the debug level used */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    Controller();

  public : 
    /**
     *  \brief init constructor
     */
    Controller(hugr::Store* store,const std::string& name,
	       boost::mutex* mutexLoc,boost::mutex* mutexCtrl,
	       boost::mutex* mutexTraj,boost::mutex* mutexEnv);

    /**
     *  \brief destructor
     */
    virtual ~Controller();

    //   /***************** Initialization *****************/

    //   /**
    //    *  \brief initialize the controller execution
    //    */
    //   virtual void initExec() = 0;
  
    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : return the time intervals between two send controls
     *  \return the time intervals between two send controls
     */
    inline double dt() const { return this->dt_; }

    /**
     *  \brief setter : return a reference to the time intervals between two send controls
     *  \return a reference to the time intervals between two send controls
     */
    inline double& dt() { return this->dt_; }

    /**
     *  \brief getter : return the current control-time applied
     *  \return the current control-time applied
     */
    inline const ControlTime* controlt() { return this->controlt_; }

    /**
     *  \brief return the status of the controller
     *  \return the status of the controller
     */
    inline controllerStatus_t status() const { return this->ctrlerSt_; }

    /**
     *  \brief return the robot's controller type
     *  \return the robot's controller type
     */
    inline controller_t ctrlerType() { return this->ctrlerType_; }

    /**
     *  \brief return the current time of the supposed reached waypoint of the trajectory
     *  \return the current time of the supposed reached waypoint of the trajectory
     */
    inline virtual double curTimeOnTraj() const { return 0; }

    /**
     *  \brief return the state of the trajectory followed (if exists) at absolute time t
     *  \param t : the input absolute time
     *  \param res : the resulting state of the trajectory followed at absolute time t if it exists, 0 else
     */
    inline virtual void getStateAtAbsTime(double t,StateTime& res) const = 0;

    /**
     *  \brief return the debug level used
     *  \return the debug level used
     */
    inline int debug() const { return this->debug_; }

    /**
     *  \brief return the debug level used
     *  \return the debug level used
     */
    inline virtual void setDebug(int dbg) { this->debug_ = dbg; }

    /******** hugr store connection functions *********/

  public : 
    /**
     *  \brief connect the controller to the hugr store
     *  \return 0 if the connection to the hugr store has been correctly established, -1 else
     */
    virtual int connect() = 0;


    /********************* Actions ********************/

  protected : 
    /**
     *  \brief compute the control to apply from the time t during a time dt
     *  \param t : the initial time of application of the control
     *  \param dt : the duration of the application of the control
     *
     *  The resulting control is put in controlt_ to be send  
     */
    virtual void computeControl(double t,double dt) = 0;

    /**
     *  \brief send a control to the hugr store
     *  \return 0 if the control has been sent successfully, -1 else
     */
    virtual int sendControl() const = 0;

    /**
     *  \brief execute the control loop
     */
    virtual void loop();

  public : 
    /**
     *  \brief reinit the controller
     */
    virtual void reinit() = 0;

    /**
     *  \brief start the timer
     *  \return 0 if the starting process succeed, -1 else
     */
    virtual int start();

    /**
     *  \brief stop the timer
     *  \return 0 if the stoping process succeed, -1 else
     */
    virtual int stop();

    /**
     *  \brief update the input control
     *  \param ctrl : the new input control
     *  \return 0 if the update process succeed
     */
    virtual int updateCtrl(const Control& c) { return 0; }

  public : 
    /**
     *  \brief update the localization of the trajectory follower
     *  \param curST : the current state-time of the robot
     */
    virtual void updateStateT(const StateTime& curST) = 0;

    /*************** Display functions ****************/

  public : 
    /**
     *  \brief draw the controller features in an openGL viewer
     */
    virtual void toGL() const {}
  };

};
#endif  // MPC_CONTROLLER_HPP


#ifndef MPC_WCCONTROLLER_HPP
#define MPC_WCCONTROLLER_HPP

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class WCController ***************/
                                    /**************************************************/

  /** \class WCController
   * \brief Simple controller for manual driving for a wheelchair
   */
  class WCController : public Controller
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    WCController();

  public : 
    /**
     *  \brief init constructor
     */
    WCController(hugr::Store* store,const std::string& name);

    /**
     *  \brief destructor
     */
    virtual ~WCController();

    //   /***************** Initialization *****************/

    // public : 
    //   /**
    //    *  \brief initialize the controller execution
    //    */
    //   virtual void initExec() {}

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : return the current control-time applied
     *  \return the current control-time applied
     */
    inline const WCControlT* controlt() const { return ((WCControlT*)(this->controlt_)); }

    /**
     *  \brief return the state of the trajectory followed (if exists) at absolute time t
     *  \param t : the input absolute time
     *  \param res : the state of the trajectory followed at absolute time t if it exists, 0 else
     */
    inline virtual void getStateAtAbsTime(double t,StateTime& res) const { }

    /******** hugr store connection functions *********/

  public : 
    /**
     *  \brief connect the controller to the hugr store
     *  \return 0 if the connection to the hugr store has been correctly established, -1 else
     */
    virtual int connect();

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief compute the control to apply from the time t during a time dt
     *  \param t : the initial time of application of the control
     *  \param dt : the duration of the application of the control
     *
     *  The resulting control is put in controlt_ to be send  
     */
    virtual void computeControl(double t,double dt);

    /**
     *  \brief send a control to the hugr store
     *  \return 0 if the control has been sent successfully, -1 else
     */
    virtual int sendControl() const;
  
    /**
     *  \brief update the input control
     *  \param ctrl : the new input control
     *  \return 0 if the update process succeed
     */
    virtual int updateCtrl(const Control& c);

  public : 
    /**
     *  \brief reinit the controller
     */
    virtual void reinit();

    /**
     *  \brief update the localization of the trajectory follower
     *  \param curST : the current state-time of the robot
     */
    virtual void updateStateT(const StateTime& curST) {}
  };

};
#endif // MPC_WCCONTROLLER_HPP


#ifndef MPC_WCCONTROLLER2_HPP
#define MPC_WCCONTROLLER2_HPP

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class WCController2 **************/
                                    /**************************************************/

  /** \class WCController2
   * \brief Simple controller for manual driving for a wheelchair
   */
  class WCController2 : public Controller
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    WCController2();

  public : 
    /**
     *  \brief init constructor
     */
    WCController2(hugr::Store* store,const std::string& name);

    /**
     *  \brief destructor
     */
    virtual ~WCController2();

    //   /***************** Initialization *****************/

    // public : 
    //   /**
    //    *  \brief initialize the controller execution
    //    */
    //   virtual void initExec() {}

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : return the current control-time applied
     *  \return the current control-time applied
     */
    inline const WCControlT* controlt() const { return ((WCControlT*)(this->controlt_)); }

    /**
     *  \brief return the state of the trajectory followed (if exists) at absolute time t
     *  \param t : the input absolute time
     *  \param res : the state of the trajectory followed at absolute time t if it exists, 0 else
     */
    inline virtual void getStateAtAbsTime(double t,StateTime& res) const { }

    /******** hugr store connection functions *********/

  public : 
    /**
     *  \brief connect the controller to the hugr store
     *  \return 0 if the connection to the hugr store has been correctly established, -1 else
     */
    virtual int connect();

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief compute the control to apply from the time t during a time dt
     *  \param t : the initial time of application of the control
     *  \param dt : the duration of the application of the control
     *
     *  The resulting control is put in controlt_ to be send  
     */
    virtual void computeControl(double t,double dt);

    /**
     *  \brief send a control to the hugr store
     *  \return 0 if the control has been sent successfully, -1 else
     */
    virtual int sendControl() const;

  public : 
    /**
     *  \brief reinit the controller
     */
    virtual void reinit();

    /**
     *  \brief update the localization of the trajectory follower
     *  \param curST : the current state-time of the robot
     */
    virtual void updateStateT(const StateTime& curST) {}
  };

};
#endif // MPC_WCCONTROLLER2_HPP


#ifndef MPC_CCCONTROLLER_HPP
#define MPC_CCCONTROLLER_HPP

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class CCController ***************/
                                    /**************************************************/

  /** \class CCController
   * \brief Simple controller for manual driving for a cycab
   */
  class CCController : public Controller
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    CCController();

  public : 
    /**
     *  \brief init constructor
     */
    CCController(hugr::Store* store,const std::string& name);

    /**
     *  \brief destructor
     */
    virtual ~CCController();

    //   /***************** Initialization *****************/

    // public : 
    //   /**
    //    *  \brief initialize the controller execution
    //    */
    //   virtual void initExec() {}

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : return the current control-time applied
     *  \return the current control-time applied
     */
    inline const CCControlT* controlt() const { return ((CCControlT*)(this->controlt_)); }

    /**
     *  \brief return the state of the trajectory followed (if exists) at absolute time t
     *  \param t : the input absolute time
     *  \param res : the state of the trajectory followed at absolute time t if it exists, 0 else
     */
    inline virtual void getStateAtAbsTime(double t,StateTime& res) const { }

    /******** hugr store connection functions *********/

  public : 
    /**
     *  \brief connect the controller to the hugr store
     *  \return 0 if the connection to the hugr store has been correctly established, -1 else
     */
    virtual int connect();

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief compute the control to apply from the time t during a time dt
     *  \param t : the initial time of application of the control
     *  \param dt : the duration of the application of the control
     *
     *  The resulting control is put in controlt_ to be send  
     */
    virtual void computeControl(double t,double dt);

    /**
     *  \brief send a control to the hugr store
     *  \return 0 if the control has been sent successfully, -1 else
     */
    virtual int sendControl() const;
  
    /**
     *  \brief update the input control
     *  \param ctrl : the new input control
     *  \return 0 if the update process succeed
     */
    virtual int updateCtrl(const Control& c);

  public : 
    /**
     *  \brief reinit the controller
     */
    virtual void reinit();

    /**
     *  \brief update the localization of the trajectory follower
     *  \param curST : the current state-time of the robot
     */
    virtual void updateStateT(const StateTime& curST) {}
  };

};
#endif // MPC_CCCONTROLLER_HPP
