#ifndef MPC_HUGR_CC_TFD_HPP
#define MPC_HUGR_CC_TFD_HPP

#include "hugr/hugrController/inc/Hugr_CCTrajFollow.hpp"
#include "hugr/hugrDeformer/inc/Hugr_CLTrajDeform.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class Hugr_CC_TFD ***************/
                                    /**************************************************/

  /** \class Hugr_CC_TFD
   * \brief Trajectory following process to control a cycab coupled with a trajectory deformer
   */
  class Hugr_CC_TFD : public Hugr_CCTrajFollow
  {
  protected : 
    Hugr_CLTrajDeform* deformer_;                     /**< the trajectory deformer used */

    /********** Constructors and destructors **********/
  protected : 
    /**
     *  \brief default constructor
     */
    Hugr_CC_TFD();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the system constraints of the wheelchair
     *  \param env : the environment in which the system evolve
     *  \param traj : the trajectory to follow
     *  \param store : the hugr store on which to send the control data
     *  \param name : the name of the cycab controlled
     *  \param mutexLoc : the mutex used to prevent localization conflicts
     *  \param mutexCtrl : the mutex used to prevent control conflicts
     */
    Hugr_CC_TFD(CarLikeSys* sys,Environment* env,CLSTraj* traj,hugr::Store* store,const std::string& name,
		boost::mutex* mutexLoc,boost::mutex* mutexCtrl,boost::mutex* mutexTraj,boost::mutex* mutexEnv);

    /**
     *  \brief destructor
     */
    virtual ~Hugr_CC_TFD();
  
    /************** Getters and setters ***************/

    /**
     *  \brief return the debug level used
     *  \return the debug level used
     */
    virtual void setDebug(int dbg);

    /**
     *  \brief set the deformation parameters
     *  \param k_int : internal forces coefficient
     *  \param k_ext : external forces coefficient (dynamic obstacles)
     *  \param k_stext : external forces coefficient (static obstacles)
     *  \param dyn_med : maximal effective distance (dynamic obstacles)
     *  \param st_med : maximal effective distance (static obstacles)
     *  \param dyn_med : maximal temporal effective distance
     */
    void setTeddyParams(double k_int,double k_ext,double k_stext,double dyn_med,double st_med,double tmp_med);

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief execute the control loop
     */
    void loop();

  public : 
    /**
     *  \brief start the controller
     *  \return 0 if the starting process succeed, -1 else
     */
    int start();

    /**
     *  \brief stop the controller
     *  \return 0 if the stoping process succeed, -1 else
     */
    int stop();

    /*************** Display functions ****************/

  public : 
    /**
     *  \brief update the trajectory deformation process visualization
     *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
     *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
     *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
     *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
     */
    void updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs);
  };

};
#endif // MPC_HUGR_CC_TFD_HPP
