#include "Hugr_CC_TFD.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /**************** Class Hugr_CC_TFD ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
Hugr_CC_TFD::Hugr_CC_TFD() : Hugr_CCTrajFollow()
{
  this->deformer_ = NULL;
  this->ctrlerType_ = TEDDY_CTRLER;
}

/*
 *  \brief init constructor
 *  \param sys : the system constraints of the cycab
 *  \param env : the environment in which the system evolve
 *  \param traj : the trajectory to follow
 *  \param store : the hugr store on which to send the control data
 *  \param name : the name of the cycab controlled
 *  \param mutexLoc : the mutex used to prevent localization conflicts
 *  \param mutexCtrl : the mutex used to prevent control conflicts
 */
Hugr_CC_TFD::Hugr_CC_TFD(CarLikeSys* sys,Environment* env,CLSTraj* traj,hugr::Store* store,const std::string& name,
	    boost::mutex* mutexLoc,boost::mutex* mutexCtrl,boost::mutex* mutexTraj,boost::mutex* mutexEnv)
  : Hugr_CCTrajFollow(sys,traj,store,name,mutexLoc,mutexCtrl,mutexTraj,mutexEnv)
{
  this->deformer_ = new Hugr_CLTrajDeform(traj,env,sys,&timer_,mutexTraj,mutexLoc,mutexEnv);
  this->ctrlerType_ = TEDDY_CTRLER;
}

/*
 *  \brief destructor
 */
Hugr_CC_TFD::~Hugr_CC_TFD()
{
  if(deformer_ != NULL) delete deformer_;
}


/************** Getters and setters ***************/

/*
 *  \brief return the debug level used
 *  \return the debug level used
 */
void Hugr_CC_TFD::setDebug(int dbg)
{
  // update the local debugging level
  this->debug_ = dbg;
  
  // update the deformer debugging level
  if(deformer_ != NULL) deformer_->setDebug(dbg);
}

/**
 *  \brief set the deformation parameters
 *  \param k_int : internal forces coefficient
 *  \param k_ext : external forces coefficient (dynamic obstacles)
 *  \param k_stext : external forces coefficient (static obstacles)
 *  \param dyn_med : maximal effective distance (dynamic obstacles)
 *  \param st_med : maximal effective distance (static obstacles)
 *  \param dyn_med : maximal temporal effective distance
 */
void Hugr_CC_TFD::setTeddyParams(double k_int,double k_ext,double k_stext,double dyn_med,double st_med,double tmp_med)
{
  if(deformer_ != NULL)
    {
      deformer_->setForcesCoef(k_int,k_ext,k_stext,0.5,2.0,1.0,10.0);
      deformer_->setMaxEffectDist(dyn_med,tmp_med,st_med);
    }
}

  
/********************* Actions ********************/

/*
 *  \brief execute the control loop
 */
void Hugr_CC_TFD::loop()
{
  while(ctrlerSt_ == CTRLER_STARTED)
    {
      try
	{
	  // compute the control to apply from the time t during a time dt
	  this->computeControl(timer_.totalElapsedTime(),dt_);
	  
	  // send the control to the hugr store
	  this->sendControl();
	  
	  // compute the elapsed time since the last control sent
	  double elapsedT = timer_.elapsedTime();
	  
	  if(dtToWait_ == -1)
	    {
	      dtToWait_ = dt_ - elapsedT;
	      if(dtToWait_ > 0) usleep((int)(1000000.0*(dtToWait_)));
	      else usleep((int)(1000000.0*(dt_)));
	    }
	  else
	    {
	      dtToWait_ = (dtToWait_ - dt_ + elapsedT);
	      if(dtToWait_ > 0) usleep((int)(1000000.0*(dtToWait_)));
	      else usleep((int)(1000000.0*(dt_)));
	    }
	}
      catch(string str)
	{
	  cout << "Hugr_CC_TFD::loop -> Exception : " << str << endl;
	  throw 2;
	}
    }

  deformer_->stop();
}

/*
 *  \brief start the controller
 *  \return 0 if the starting process succeed, -1 else
 */
int Hugr_CC_TFD::start()
{
  switch(ctrlerSt_)
    {
    case CTRLER_NULL : 
    case CTRLER_DISCONNECTED : 
      {
	cerr << "Hugr_CC_TFD::start : Warning. The controller is disconnected." << endl;
	return -1;
      }
    case CTRLER_STOPPED : 
      {
	ctrlerSt_ = CTRLER_STARTED;
	timer_.start();
	threadCtrl_ = new boost::thread(boost::bind(&Hugr_CC_TFD::loop,this));
	deformer_->start();
	return 0;
      }
    case CTRLER_STARTED : 
      {
	// nothing to do
	cerr << "Hugr_CC_TFD::start : Controller already started" << endl;
	return 0;
      }
    default : 
      {
	cerr << "Hugr_CC_TFD::start : Unknown controller status" << endl;
	return -1;
      }
    }
}

/*
 *  \brief stop the controller
 *  \return 0 if the stoping process succeed, -1 else
 */
int Hugr_CC_TFD::stop()
{
  switch(ctrlerSt_)
    {
    case CTRLER_NULL : 
    case CTRLER_DISCONNECTED : 
      {
	cerr << "Hugr_CC_TFD::stop : Warning. The controller is disconnected." << endl;
	return -1;
      }
    case CTRLER_STOPPED : 
      {
	// nothing to do
	return 0;
      }
    case CTRLER_STARTED : 
      {
	ctrlerSt_ = CTRLER_STOPPED;
	dtToWait_ = -1;
	timer_.stop();
	if(threadCtrl_ != NULL) delete threadCtrl_;
	deformer_->stop();
	return 0;
      }
    }
}


/*************** Display functions ****************/

/**
 *  \brief update the trajectory deformation process visualization
 *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
 *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
 *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
 *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
 */
void Hugr_CC_TFD::updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs)
{
  if(deformer_ != NULL) deformer_->updateTeddyView(dispFExt,dispStFExt,dispNearDynObs,dispNearStObs);
}
