#include "Hugr_WCTrajFollow.hpp"
using namespace mpc;

#include "WheelchairSimulator.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <utime.h>

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /************* Class Hugr_WCTrajFollow ************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
Hugr_WCTrajFollow::Hugr_WCTrajFollow() : Controller(), WCTrajFollow()
{
   this->curTimeOnTraj_ = 0;
   this->ctrlerType_ = TRAJFOLLOW_CTRLER;
}

/*
 *  \brief init constructor
 *  \param sys : the system constraints of the wheelchair
 *  \param traj : the trajectory to follow
 *  \param store : the hugr store on which to send the control data
 *  \param name : the name of the cycab controlled
 *  \param mutexLoc : the mutex used to prevent localization conflicts
 *  \param mutexCtrl : the mutex used to prevent control conflicts
 *  \param mutexTraj : the mutex used to prevent trajectory conflicts
 *  \param mutexEnv : the mutex used to prevent environment conflicts
 */
Hugr_WCTrajFollow::Hugr_WCTrajFollow(WheelChairSys* sys,WCSTraj* traj,
				     hugr::Store* store,const std::string& name,
				     boost::mutex* mutexLoc,boost::mutex* mutexCtrl,
				     boost::mutex* mutexTraj,boost::mutex* mutexEnv)
  : Controller(store,name,mutexLoc,mutexCtrl,mutexTraj,mutexEnv), WCTrajFollow(sys,traj)
{
   this->curTimeOnTraj_ = 0;
   this->ctrlerType_ = TRAJFOLLOW_CTRLER;
}

/*
 *  \brief destructor
 */
Hugr_WCTrajFollow::~Hugr_WCTrajFollow()
{
}


/************** Getters and setters ***************/

/*
 *  \brief return the state of the trajectory followed (if exists) at absolute time t
 *  \param t : the input absolute time
 *  \param res : the resulting state of the trajectory followed at absolute time t if it exists, 0 else
 */
void Hugr_WCTrajFollow::getStateAtAbsTime(double t,StateTime& res) const
{
  double relativeTime = t - absRbST_.t() + curRbST_.t();
//   cout << "t : " << t << "  absRbST_.t() : " << absRbST_.t() << "  curRbST_.t() : " << curRbST_.t() << "  relativeTime : " << relativeTime << endl;
  this->traj()->getStateAtTime(relativeTime,*((WCStateT*)(&res)));
  res.t() = t;
//   cout << "res : "; res.display(); cout << endl;
}


/******** hugr store connection functions *********/

/*
 *  \brief connect the controller to the hugr store
 *  \int return 0 if the connection to the hugr store has been correctly established, -1 else
 */
int Hugr_WCTrajFollow::connect()
{
  if(store_ == NULL)
    {
      cout << "Hugr_WCTrajFollow::connect : Warning. The HUGR store hasn't been defined." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }
  
  // connect the hugr store if it has not been already done
  store_->connect();

  // get back the wheelchair command id
//   std::string str(nameRb_);
  try
    {
//       str += "_command";
//       controlId_ = store_->lookupVariable(str.c_str());
      controlId_ = store_->lookupVariable(commandName_.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "Hugr_WCTrajFollow::connect : Warning. " << commandName_.c_str() << " cannot be found in the HUGR store." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }

  ctrlerSt_ = CTRLER_STOPPED;
  return 0;
}


/********************* Actions ********************/

/*
 *  \brief compute the control to apply from the time t during a time dt
 *  \param t : the initial time of application of the control
 *  \param dt : the duration of the application of the control
 *
 *  The resulting control is put in controlt_ to be send  
 */
void Hugr_WCTrajFollow::computeControl(double t,double dt)
{ 
  // lock the localization mutex to prevent loc conflicts
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajFollow::computeControl : waiting for mutex Loc" << endl;
  boost::mutex::scoped_lock lockLoc(*mutexLoc_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajFollow::computeControl : mutex Loc locked" << endl;

//   // lock the control mutex to prevent control conflicts
//   boost::mutex::scoped_lock lockCtrl(*mutexCtrl_);
  
  // get the last saved state-time
  WCStateT curST(curRbST_);

  // compute the time elapsed from the previous control request
  double DT = timer_.elapsedTime();

  // set the current time as the time registered by the timer
  curTimeOnTraj_ = timer_.totalElapsedTime();
//   cout << "curTimeOnTraj_ : " << curTimeOnTraj_ << "  ";
  //curST.t() = curTimeOnTraj_;

  WCControl ctrl;

//   int res = this->getControl(curST,
// 			     curST.t() + this->lastDuration_,
// 			     this->lastDuration_,
// 			     ctrl);
  int res = this->getControl(curST,
			     curTimeOnTraj_,
			     DT,
			     ctrl);

  if(isnan(ctrl.v()) || isnan(ctrl.av()))
    {
      cerr << "WCTrajFollow::computeControl : Warning. Input control is not a number!" << endl;
      ((WCControlT*)(this->controlt_))->v() = 0.0;
      ((WCControlT*)(this->controlt_))->av() = 0.0;
      ((WCControlT*)(this->controlt_))->t() = dt;
    }
  else
    {
      ((WCControlT*)(this->controlt_))->v() = ctrl.v();
      ((WCControlT*)(this->controlt_))->av() = ctrl.av();
      ((WCControlT*)(this->controlt_))->t() = dt;
    }

  // modify the state of the robot to consider the actual velocities
  curRbST_.s() = ctrl.v();
  curRbST_.om() = ctrl.av();

  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajFollow::computeControl : mutex Loc released" << endl;
}

/*
 *  \brief send a control to the hugr store
 *  \return 0 if the control has been sent successfully, -1 else
 */
int Hugr_WCTrajFollow::sendControl() const
{
  if(store_ == NULL)
    {
      cerr << "WCController::sendControl : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }

  WheelchairCommand WC_Command;
  WC_Command.v = this->controlt()->v();
  WC_Command.omega = this->controlt()->av();

  try
    {
      store_->writeVariable(controlId_, WC_Command);

//       cout << "control : "; this->controlt()->display(); cout << endl;
    }
  catch(runtime_error e)
    {
      cerr << "WCController::sendControl : Warning. The control ID has not been found in the HUGR store." << endl;
      return -1;
    }
  return 0;
}

/*
 *  \brief execute the control loop
 */
void Hugr_WCTrajFollow::loop()
{
//   cout << "Hugr_WCTrajFollow::loop" << endl;
  while(ctrlerSt_ == CTRLER_STARTED)
    {
      // compute the control to apply from the time t during a time dt
      this->computeControl(timer_.totalElapsedTime(),dt_);

      // send the control to the hugr store
      this->sendControl();

      // compute the elapsed time since the last control sent
      double elapsedT = timer_.elapsedTime();

      if(dtToWait_ == -1)
	{
	  dtToWait_ = dt_ - elapsedT;
	  if(dtToWait_ > 0) usleep((int)(1000000.0*(dtToWait_)));
	  else usleep((int)(1000000.0*(dt_)));
	}
      else
	{
	  dtToWait_ = (dtToWait_ - dt_ + elapsedT);
	  if(dtToWait_ > 0) usleep((int)(1000000.0*(dtToWait_)));
	  else usleep((int)(1000000.0*(dt_)));
	}

      // check if the end of the trajectory has been reached
      if(this->endTraj()) 
	{ 
	  this->stop(); 
// 	  cout << "endTraj!" << endl; 
	}
//       else cout << "Pas endTraj!" << endl;
      //if(this->endTraj()) ctrlerSt_ = CTRLER_STOPPED;
    }
}

/*
 *  \brief update the localization of the trajectory follower
 *  \param curST : the current state-time of the robot
 */
void Hugr_WCTrajFollow::updateStateT(const StateTime& curST)
{
  // lock the localization mutex to prevent loc conflicts
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajFollow::updateStateT : waiting for mutex Loc" << endl;
  boost::mutex::scoped_lock lock(*mutexLoc_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajFollow::updateStateT : mutex Loc locked" << endl;

  // get the position of the input state
  PositionXYTh pos = curST.posXYTh();

  curRbST_.x() = pos.x();
  curRbST_.y() = pos.y();
  curRbST_.th() = normalizedAngle(pos.th());
  curRbST_.t() = timer_.totalElapsedTime();

  absRbST_.x() = pos.x();
  absRbST_.y() = pos.y();
  absRbST_.th() = normalizedAngle(pos.th());
  absRbST_.t() = curST.t();

//   if(curRbST_ != NULL) delete curRbST_;
//   curRbST_ = new WCStateT(*((WCStateT*)(&curST)));
//   ((WCStateT*)curRbST_)->th() = normalizedAngle(((WCStateT*)curRbST_)->th());
//   ((WCStateT*)curRbST_)->t() = timer_.totalElapsedTime();

//   if(absRbST_ != NULL) delete absRbST_;
//   absRbST_ = new WCStateT(*((WCStateT*)(&curST)));
//   ((WCStateT*)absRbST_)->th() = normalizedAngle(((WCStateT*)absRbST_)->th());

  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajFollow::updateStateT : mutex Loc released" << endl;
}

/*
 *  \brief reinit the controller
 */
void Hugr_WCTrajFollow::reinit()
{
  this->curTimeOnTraj_ = 0;
  timer_.reinit();
}
