#ifndef MPC_HUGR_WCTRAJDEFORM_HPP
#define MPC_HUGR_WCTRAJDEFORM_HPP

#include "deformation/inc/WCTrajDeform3.hpp"

#include "hugr/HCommon/inc/HCommon.hpp"
#include "hugr/hugrEnvironment/inc/HEnvironment.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

namespace mpc
{
                                    /**************************************************/
                                    /************* Class Hugr_WCTrajDeform ************/
                                    /**************************************************/

  /** \class Hugr_WCTrajDeform
   *  \brief Trajectory deformer TEDDY to control a wheelchair using HUGR
   */
  class Hugr_WCTrajDeform : public WCTrajDeform3
  {
  protected :
    WCSTraj* trajRes_;                            /**< the resulting deformed trajectory */

    HEnvironment* envReg_;                        /**< last snapshot of the environment registered by the perception process */

    Timer* timer_;                                /**< timer used to date the input data (localization, environment perception) */

    deformerStatus_t status_;                     /**< status of the deformation process */

    boost::thread* threadDeform_;                 /**< the thread used to execute the deformation process */

    //   boost::mutex* mutexTraj_;                     /**< mutex used to prevent trajectory access conflicts */
    boost::mutex* mutexLoc_;                      /**< mutex used to prevent localization access conflicts */
    boost::mutex* mutexEnv_;                      /**< mutex used to prevent environment model access conflicts */

    int trajResSize_;                             /**< size of the last trajectory registered */
    PositionXYT **tableNearStateCpy_;             /**< table of nearest states of dynamic obstacles from the states of the trajectory */
    PositionXYT **tableNearStStateCpy_;           /**< table of nearest states of static obstacles from the states of the trajectory */
    WCStateT *tableFIntCpy_;                      /**< table of the internal forces */
    WCStateT *tableFExtCpy_;                      /**< table of the external forces (due to dynamic obstacles) */
    WCStateT *tableStFExtCpy_;                    /**< table of the external forces (due to static obstacles) */

    bool dispFExt_;                               /**< true if the external forces (from dynamic obstacles) must be displayed */
    bool dispStFExt_;                             /**< true if the external forces (from static obstacles) must be displayed */
    bool dispNearDynObs_;                         /**< true if the nearest dynamic obstacles must be displayed */
    bool dispNearStObs_;                          /**< true if the nearest static obstacles must be displayed */

    int nbDeform_;                                /**< number of deformation proceeded */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     */
    Hugr_WCTrajDeform();

  public : 
    /**
     *  \brief init constructor
     *  \param traj : the input trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \parma sys : the constraints defined over the robotic system
     *  \param timer : timer of the associated controller
     *  \param mutexTraj : mutex used to prevent trajectory access conflicts
     *  \param mutexLoc : mutex used to prevent localization access conflicts
     *  \param mutexEnv : mutex used to prevent environment model access conflicts
     */
    Hugr_WCTrajDeform(WCSTraj* traj,HEnvironment* env,WheelChairSys* sys,
		      Timer* timer,boost::mutex* mutexTraj,
		      boost::mutex* mutexLoc,boost::mutex* mutexEnv);

    //   /**
    //    *  \brief init constructor
    //    *  \param traj : the initial trajectory to deform
    //    *  \param env : the environment in which the system evolve
    //    *  \param sys : the features of the robot system used
    //    *  \param timer : timer of the associated controller
    //    *  \param mutexTraj : mutex used to prevent trajectory access conflicts
    //    *  \param mutexLoc : mutex used to prevent localization access conflicts
    //    *  \param mutexEnv : mutex used to prevent environment model access conflicts
    //    *  \param typeS : pointer to the table of state types (for simulation display)
    //    *  \param tns : pointer to the table of nearest states (for simulation display)
    //    *  \param tfi : pointer to the table of internal forces (for simulation display)
    //    *  \param tfe : pointer to the table of external forces (for simulation display)
    //    */
    //   Hugr_WCTrajDeform(WCSTraj* traj,Environment* env,WheelChairSys* sys,
    // 		    Timer* timer,boost::mutex* mutexTraj,
    // 		    boost::mutex* mutexLoc,boost::mutex* mutexEnv,
    // 		    int** typeS,PositionXYT*** tns,
    // 		    StateTime** tfi,StateTime** tfe);

    /**
     *  \brief destructor
     */
    virtual ~Hugr_WCTrajDeform();

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief return the status of the deformation process
     *  \return the status of the deformation process
     */
    inline deformerStatus_t status() const { return this->status_; }

    /**
     *  \brief return the resulting deformed trajectory
     *  \return the resulting deformed trajectory
     */
    inline WCSTraj* trajRes() const { return ((WCSTraj*)trajRes_); }

    /**
     *  \brief return the trajectory handled by the deformation process
     *  \return the trajectory handled by the deformation process
     */
    virtual WCSTraj* traj() { return ((WCSTraj*)(this->WCTrajDeform3::traj_)); }

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief init the trajectory to deform
     *
     *  Copy the old resulting trajectory to perform the deformation process on it
     */
    virtual void initTraj();

    /**
     *  \brief update the environment model registered by the perception process
     *
     *  Take a snapshot of the current environment to perform the deformation process with it
     */
    virtual void initEnv();

    /**
     *  \brief execute the control loop
     */
    void loop();

  public : 
    /**
     *  \brief start the timer
     *  \return 0 if the starting process succeed, -1 else
     */
    int start();

    /**
     *  \brief stop the timer
     *  \return 0 if the stoping process succeed, -1 else
     */
    int stop();

    /**
     *  \brief update the trajectory followed by the robot with the deformed trajectory
     */
    virtual void updateTraj();

    /**
     *  \brief update the environment model registered by the perception process
     *  \param env : the new snapshot of the environment
     */
    inline void updateEnv(HEnvironment* env) { this->envReg_ = env; }

    /*************** Display functions ****************/

  public : 
    /**
     *  \brief update the trajectory deformation process visualization
     *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
     *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
     *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
     *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
     */
    void updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs);

    /**
     *  \label get back informations about the deformation process
     *  \param nbDeform : the number of deformations proceeded
     *  \param avgDeformDur : the average deformation duration
     *  \param DPS : the average number of deformations per second
     */
    void getDeformInfo(int& nbDeform,double& avgDeformDur,double& DPS);

    /**
     *  \brief draw the controller features in an openGL viewer
     */
    virtual void toGL();

    //   /******************* Resampling *******************/
  
    // protected : 
    //   /**
    //    *  \brief resample the trajectory
    //    *  \param curStep : index of the node already reached
    //    */
    //   inline virtual void resample(int curStep) { return this->WCTrajDeform::resample(curStep); }

    //   /*************** Deformation process **************/

    // protected : 
    //   /**
    //    *  \brief init the deformation and check the temporal space between two consecutive states
    //    *  \param curStep : index of the node of the trajectory that has been already reached
    //    */
    //   inline virtual void initDeform(int curStep) { return this->WCTrajDeform::initDeform(curStep); }
  
    //   /**
    //    *  \brief finish the deformation
    //    *  \param time : the current time
    //    *  \param curStep : index of the node of the trajectory that has been already reached
    //    */
    //   inline virtual int finishDeform(double time,int curStep) { return this->WCTrajDeform::finishDeform(time,curStep); }

    // public : 
    //   /**
    //    *  \brief deform the current trajectory
    //    *  \param time : the current time
    //    *  \param curStep : index of the node of the trajectory that has been already reached
    //    *  \return 0 if the trajectory generation process succeed without any problem
    //    *  \return -1 if no input trajectory has been defined
    //    *  \return -2 if a collision may occur during the prediction horizon
    //    *  \return -3 if the connectivity of the trajectory has been lost
    //    */
    //   inline virtual int deform(double time, int curStep) { return this->WCTrajDeform::deform(time,curStep); }

    //   /**
    //    *  \brief deform the current trajectory
    //    *  \param time : the current time
    //    *  \return 0 if the trajectory generation process succeed without any problem
    //    *  \return -1 if no input trajectory has been defined
    //    *  \return -2 if a collision may occur during the prediction horizon
    //    *  \return -3 if the connectivity of the trajectory has been lost
    //    *  \return -4 if the final time of the trajectory has been already past
    //    */
    //   inline virtual int deform(double time) { return this->WCTrajDeform::deform(time); }
  };

};
#endif // MPC_HUGR_WCTRAJDEFORM_HPP
