#include "Hugr_CLTrajDeform.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /************* Class Hugr_CLTrajDeform ************/
                                    /**************************************************/

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 */
Hugr_CLTrajDeform::Hugr_CLTrajDeform() : CLTrajDeform()
{
  trajRes_ = NULL;

  envReg_ = NULL;

  timer_ = NULL;

  threadDeform_ = NULL;

  mutexTraj_ = NULL;
  mutexLoc_ = NULL;
  mutexEnv_ = NULL;

  status_ = DEFORMER_STOPPED;
}

/**
 *  \brief init constructor
 *  \param traj : the input trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \parma sys : the constraints defined over the robotic system
 *  \param timer : timer of the associated controller
 *  \param mutexTraj : mutex used to prevent trajectory access conflicts
 *  \param mutexLoc : mutex used to prevent localization access conflicts
 *  \param mutexEnv : mutex used to prevent environment model access conflicts
 */
Hugr_CLTrajDeform::Hugr_CLTrajDeform(CLSTraj* traj,Environment* env,CarLikeSys* sys,
				     Timer* timer,boost::mutex* mutexTraj,
				     boost::mutex* mutexLoc,boost::mutex* mutexEnv) : 
  CLTrajDeform(NULL,NULL,sys)
{
  //this->trajRes_ = new CLSTraj(*traj);
  this->trajRes_ = traj;

  envReg_ = env;

  timer_ = timer;

  threadDeform_ = NULL;

  mutexTraj_ = mutexTraj;
  mutexLoc_ = mutexLoc;
  mutexEnv_ = mutexEnv;

  status_ = DEFORMER_STOPPED;
}

/**
 *  \brief destructor
 */
Hugr_CLTrajDeform::~Hugr_CLTrajDeform()
{
}


/********************* Actions ********************/

/**
 *  \brief init the trajectory to deform
 *
 *  Copy the old resulting trajectory to perform the deformation process on it
 */
void Hugr_CLTrajDeform::initTraj()
{
  // lock a mutex to prevent trajectory accesses
  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::initTraj : waiting for mutex Traj" << endl;
  boost::mutex::scoped_lock lock(*mutexTraj_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::initTraj : mutex Traj locked" << endl;

  if(trajRes_ == NULL)
    {
      cerr << "Hugr_CLTrajDeform::initTraj. Warning : The trajectory to copy is empty" << endl;
      return;
    }

  if(this->traj_ != NULL) delete traj_;
  this->CLTrajDeform::traj_ = new CLSTraj(*((CLSTraj*)trajRes_));

  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::initTraj : mutex Traj released" << endl;
}

/**
 *  \brief update the environment model registered by the perception process
 *
 *  Take a snapshot of the current environment to perform the deformation process with it
 */
void Hugr_CLTrajDeform::initEnv()
{
  if(this->envReg_ == NULL)
    {
      cerr << "Hugr_CLTrajDeform::initEnv. Warning : The environment to copy is empty" << endl;
      return;
    }

  // lock a mutex to prevent trajectory accesses
  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::initEnv : waiting for mutex Env" << endl;
  boost::mutex::scoped_lock lock(*mutexEnv_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::initEnv : mutex Env locked" << endl;

  if(this->env_ != NULL) delete env_;
  this->env_ = new Environment(*envReg_);

  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::initEnv : mutex Env released" << endl;
}

/**
 *  \brief execute the control loop
 */
void Hugr_CLTrajDeform::loop()
{
  while(this->status_ == DEFORMER_STARTED)
    {
      if(this->traj() == NULL) cout << "Hugr_CLTrajDeform::loop : traj() null" << endl;
      else cout << "Hugr_CLTrajDeform::loop : traj() not null" << endl;
      
      if(trajRes_ == NULL) cout << "Hugr_CLTrajDeform::loop : trajRes_ null" << endl;
      else cout << "Hugr_CLTrajDeform::loop : trajRes_ not null" << endl;

      // init the trajectory to deform
      this->initTraj();

      // replace the current environment model by the last 
      // snapshot registered by the perception process
      if(env_ != NULL) delete env_;
      env_ = envReg_;

      // retrieve the current time wrt the trajectory followed
      double time = timer_->totalElapsedTime();

      // deform the trajectory
      int res = this->deform(time);

      switch(res)
	{
	case 0 : // case where the deformation process ended successfully
	  // update the result trajectory
	  this->updateTraj();
	  break;
	case -1 : // case where no input trajectory has been defined
	  cerr << "Hugr_CLTrajDeform::loop. No input trajectory has been defined." << endl;
	  this->status_ = DEFORMER_STOPPED;
	  exit(1);
	  break;
	case -2 : // case where a collision may occur during the prediction horizon
	  cerr << "Hugr_CLTrajDeform::loop. Warning : A collision may occur during the prediction horizon." << endl;
	  break;
	case -3 : // case where the connectivity of the trajectory has been lost
	  cerr << "Hugr_CLTrajDeform::loop. Warning : The connectivity of the trajectory has been lost." << endl;
	  break;
	case -4 : // case where the final time of the trajectory has been already past
	  cout << "Hugr_CLTrajDeform::loop. End of the deformation process." << endl;
	  this->status_ = DEFORMER_STOPPED;
	  break;
	}
    }
}

/**
 *  \brief start the timer
 *  \return 0 if the starting process succeed, -1 else
 */
int Hugr_CLTrajDeform::start()
{
  switch(status_)
    {
    case DEFORMER_NULL : 
    case DEFORMER_DISCONNECTED : 
      {
	cerr << "Hugr_CLTrajDeform::start : Warning. The deformation process is disconnected. Should never appear..." << endl;
	return -1;
      }
    case DEFORMER_STOPPED : 
      {
	status_ = DEFORMER_STARTED;
	threadDeform_ = new boost::thread(boost::bind(&Hugr_CLTrajDeform::loop,this));
	return 0;
      }
    case DEFORMER_STARTED : 
      {
	// nothing to do
	cerr << "Hugr_CLTrajDeform::start : Deformation process already started" << endl;
	return 0;
      }
    default : 
      {
	cerr << "Hugr_CLTrajDeform::start : Unknown deformation process status" << endl;
	return -1;
      }
    }
}

/**
 *  \brief stop the timer
 *  \return 0 if the stoping process succeed, -1 else
 */
int Hugr_CLTrajDeform::stop()
{
  switch(status_)
    {
    case DEFORMER_NULL : 
    case DEFORMER_DISCONNECTED : 
      {
	cerr << "Hugr_CLTrajDeform::stop : Warning. The deformation process is disconnected. Should never appear..." << endl;
	return -1;
      }
    case DEFORMER_STOPPED : 
      {
	// nothing to do
	return 0;
      }
    case DEFORMER_STARTED : 
      {
	status_ = DEFORMER_STOPPED;
	if(threadDeform_ != NULL) delete threadDeform_;
	return 0;
      }
    }
}

/**
 *  \brief update the trajectory followed by the robot with the deformed trajectory
 */
void Hugr_CLTrajDeform::updateTraj()
{
  // block the access to the trajectory during its reinitialization
  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::updateTraj : waiting for mutex Traj" << endl;
  boost::mutex::scoped_lock lock(*mutexTraj_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::updateTraj : mutex Traj locked" << endl;

//   if(trajRes_ != NULL) delete trajRes_;
//   trajRes_ = traj();

  if(trajRes_ != NULL)
    {
      trajRes_->clear();
      trajRes_->concat(*traj());
    }

  if(MUTEX_DEBUG == 1) cout << "Hugr_CLTrajDeform::updateTraj : mutex Traj released" << endl;
}


/*************** Display functions ****************/

/**
 *  \brief update the trajectory deformation process visualization
 *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
 *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
 *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
 *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
 */
void Hugr_CLTrajDeform::updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs)
{
  cerr << "Hugr_CLTrajDeform::updateTeddyView : Not implemented yet!" << endl;
  exit(1);
}
