#include "Hugr_WCTrajDeform.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

                                    /**************************************************/
                                    /************* Class Hugr_WCTrajDeform ************/
                                    /**************************************************/

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 */
Hugr_WCTrajDeform::Hugr_WCTrajDeform() : WCTrajDeform3()
{
  trajRes_ = NULL;

  envReg_ = NULL;

  timer_ = NULL;

  threadDeform_ = NULL;

  mutexTraj2_ = NULL;
  mutexLoc_ = NULL;
  mutexEnv_ = NULL;

  status_ = DEFORMER_STOPPED;

  trajResSize_ = 0;
  tableNearStateCpy_ = NULL;
  tableNearStStateCpy_ = NULL;
  tableFIntCpy_ = NULL;
  tableFExtCpy_ = NULL;
  tableStFExtCpy_ = NULL;

  dispFExt_ = false;
  dispStFExt_ = false;
  dispNearDynObs_ = false;
  dispNearStObs_ = false;

  nbDeform_ = 0;
}

/**
 *  \brief init constructor
 *  \param traj : the input trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \parma sys : the constraints defined over the robotic system
 *  \param timer : timer of the associated controller
 *  \param mutexTraj : mutex used to prevent trajectory access conflicts
 *  \param mutexLoc : mutex used to prevent localization access conflicts
 *  \param mutexEnv : mutex used to prevent environment model access conflicts
 */
Hugr_WCTrajDeform::Hugr_WCTrajDeform(WCSTraj* traj,HEnvironment* env,WheelChairSys* sys,
				     Timer* timer,boost::mutex* mutexTraj,
				     boost::mutex* mutexLoc,boost::mutex* mutexEnv)
  : WCTrajDeform3(NULL,NULL,sys)
{
  //this->trajRes_ = new WCSTraj(*traj);
  this->trajRes_ = traj;

//   if(this->traj() == NULL) cout << "Hugr_WCTrajDeform::Hugr_WCTrajDeform : WCTrajDeform::traj_ null" << endl;
//   else cout << "Hugr_WCTrajDeform::Hugr_WCTrajDeform : WCTrajDeform::traj_ not null" << endl;

//   if(trajRes_ == NULL) cout << "Hugr_WCTrajDeform::Hugr_WCTrajDeform : trajRes_ null" << endl;
//   else cout << "Hugr_WCTrajDeform::Hugr_WCTrajDeform : trajRes_ not null" << endl;

//   trajRes_ = NULL;

  envReg_ = env;

  timer_ = timer;

  threadDeform_ = NULL;

  mutexTraj2_ = mutexTraj;
  mutexLoc_ = mutexLoc;
  mutexEnv_ = mutexEnv;

  status_ = DEFORMER_STOPPED;

  trajResSize_ = 0;
  tableNearStateCpy_ = NULL;
  tableNearStStateCpy_ = NULL;
  tableFIntCpy_ = NULL;
  tableFExtCpy_ = NULL;
  tableStFExtCpy_ = NULL;

  dispFExt_ = false;
  dispStFExt_ = false;
  dispNearDynObs_ = false;
  dispNearStObs_ = false;

  nbDeform_ = 0;
}

/**
 *  \brief destructor
 */
Hugr_WCTrajDeform::~Hugr_WCTrajDeform()
{
}


/********************* Actions ********************/

/**
 *  \brief init the trajectory to deform
 *
 *  Copy the old resulting trajectory to perform the deformation process on it
 */
void Hugr_WCTrajDeform::initTraj()
{
  // lock a mutex to prevent trajectory accesses
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::initTraj : waiting for mutex Traj" << endl;
  boost::mutex::scoped_lock lock(*mutexTraj2_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::initTraj : mutex Traj locked" << endl;

  if(trajRes_ == NULL)
    {
      cerr << "Hugr_WCTrajDeform::initTraj. Warning : The trajectory to copy is empty" << endl;
      if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::initTraj : mutex Traj released2" << endl;
      return;
    }

  if(this->traj_ != NULL) delete traj_;
  this->WCTrajDeform3::traj_ = new WCSTraj(*((WCSTraj*)trajRes_));

  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::initTraj : mutex Traj released" << endl;
}

/**
 *  \brief update the environment model registered by the perception process
 *
 *  Take a snapshot of the current environment to perform the deformation process with it
 */
void Hugr_WCTrajDeform::initEnv()
{
  if(this->envReg_ == NULL)
    {
      cerr << "Hugr_WCTrajDeform::initEnv. Warning : The environment to copy is empty" << endl;
      return;
    }

  // lock a mutex to prevent trajectory accesses
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::initEnv : waiting for mutex Env" << endl;
  boost::mutex::scoped_lock lock(*mutexEnv_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::initEnv : mutex Env locked" << endl;

  if(this->env_ != NULL) delete env_;
  this->env_ = new Environment(*envReg_);
//   this->env_->displayNbObs();

  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::initEnv : mutex Env released" << endl;
}

/**
 *  \brief execute the control loop
 */
void Hugr_WCTrajDeform::loop()
{
  while(this->status_ == DEFORMER_STARTED)
    {
      try
	{
	  //       if(this->traj() == NULL) cout << "Hugr_WCTrajDeform::loop : traj() null" << endl;
	  //       else cout << "Hugr_WCTrajDeform::loop : traj() not null" << endl;
	  
	  //       if(trajRes_ == NULL) cout << "Hugr_WCTrajDeform::loop : trajRes_ null" << endl;
	  //       else cout << "Hugr_WCTrajDeform::loop : trajRes_ not null" << endl;
	  
	  // init the trajectory to deform
	  this->initTraj();
	  
	  // init the environment used to do it
	  this->initEnv();
	  
	  //       if(envReg_ == NULL) cout << "envReg_ null" << endl;
	  //       else cout << "envReg_ not null" << endl;
	  
	  //       // replace the current environment model by the last 
	  //       // snapshot registered by the perception process
	  //       if(env_ != NULL) delete env_;
	  //       env_ = envReg_;
	  
	  // retrieve the current time wrt the trajectory followed
	  double time = timer_->totalElapsedTime();
	  
	  //       if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::loop : waiting for mutex Traj" << endl;
	  //       boost::mutex::scoped_lock lock(*mutexTraj_);
	  //       if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::loop : mutex Traj locked" << endl;
	  
	  // deform the trajectory
	  int res = this->deform(time,envReg_->getTime());

	  // increase the number of proceeded deformations
	  nbDeform_++;
	  
	  switch(res)
	    {
	    case 0 : // case where the deformation process ended successfully
	      // update the result trajectory
	      this->updateTraj();
	      break;
	    case -1 : // case where no input trajectory has been defined
	      cerr << "Hugr_WCTrajDeform::loop. No input trajectory has been defined." << endl;
	      this->status_ = DEFORMER_STOPPED;
	      exit(1);
	      break;
	    case -2 : // case where a collision may occur during the prediction horizon
	      cerr << "Hugr_WCTrajDeform::loop. Warning : A collision may occur during the prediction horizon." << endl;
	      break;
	    case -3 : // case where the connectivity of the trajectory has been lost
	      cerr << "Hugr_WCTrajDeform::loop. Warning : The connectivity of the trajectory has been lost." << endl;
	      break;
	    case -4 : // case where the final time of the trajectory has been already past
	      cout << "Hugr_WCTrajDeform::loop. End of the deformation process." << endl;
	      this->status_ = DEFORMER_STOPPED;
	      break;
	    }
	}
      catch(string str)
	{
	  cout << "Hugr_WCTrajDeform::loop -> Exception : " << str << endl;
	  throw 1;
	}

      if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::loop : mutex Traj released" << endl;
    }
  cout << "Hugr_WCTrajDeform::loop : Out of deformer loop" << endl;
}

/**
 *  \brief start the timer
 *  \return 0 if the starting process succeed, -1 else
 */
int Hugr_WCTrajDeform::start()
{
  switch(status_)
    {
    case DEFORMER_NULL : 
    case DEFORMER_DISCONNECTED : 
      {
	cerr << "Hugr_WCTrajDeform::start : Warning. The deformation process is disconnected. Should never appear..." << endl;
	return -1;
      }
    case DEFORMER_STOPPED : 
      {
	status_ = DEFORMER_STARTED;
	threadDeform_ = new boost::thread(boost::bind(&Hugr_WCTrajDeform::loop,this));
	return 0;
      }
    case DEFORMER_STARTED : 
      {
	// nothing to do
	cerr << "Hugr_WCTrajDeform::start : Deformation process already started" << endl;
	return 0;
      }
    default : 
      {
	cerr << "Hugr_WCTrajDeform::start : Unknown deformation process status" << endl;
	return -1;
      }
    }
}

/**
 *  \brief stop the timer
 *  \return 0 if the stoping process succeed, -1 else
 */
int Hugr_WCTrajDeform::stop()
{
  switch(status_)
    {
    case DEFORMER_NULL : 
    case DEFORMER_DISCONNECTED : 
      {
	cerr << "Hugr_WCTrajDeform::stop : Warning. The deformation process is disconnected. Should never appear..." << endl;
	return -1;
      }
    case DEFORMER_STOPPED : 
      {
	// nothing to do
	return 0;
      }
    case DEFORMER_STARTED : 
      {
	status_ = DEFORMER_STOPPED;
	if(threadDeform_ != NULL) delete threadDeform_;
	return 0;
      }
    }
}

/**
 *  \brief update the trajectory followed by the robot with the deformed trajectory
 */
void Hugr_WCTrajDeform::updateTraj()
{
  // block the access to the trajectory during its reinitialization
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::updateTraj : waiting for mutex Traj" << endl;
  boost::mutex::scoped_lock lock(*mutexTraj2_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::updateTraj : mutex Traj locked" << endl;

//   if(trajRes_ != NULL) delete trajRes_;
//   trajRes_ = traj();

  if(trajRes_ != NULL)
    {
      // reinit the trajectory
      trajRes_->clear();
      trajRes_->concat(*traj());

      if(debug_ >= 1)
	{
	  // block the access to the displaying tables
	  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : waiting for mutex Tables" << endl;
	  boost::mutex::scoped_lock lock(mutexTables_);
	  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : mutex Tables locked" << endl;

	  // delete the old debug table values
	  if(tableNearStateCpy_ != NULL) 
	    {
	      for(int i=0; i<trajResSize_; i++) delete tableNearStateCpy_[i];
	      delete [] tableNearStateCpy_;
	    }
	  if(tableNearStStateCpy_ != NULL) 
	    {
	      for(int i=0; i<trajResSize_; i++) delete tableNearStStateCpy_[i];
	      delete [] tableNearStStateCpy_;
	    }
	  if(tableFIntCpy_ != NULL) delete [] tableFIntCpy_;
	  if(tableFExtCpy_ != NULL) delete [] tableFExtCpy_;
	  if(tableStFExtCpy_ != NULL) delete [] tableStFExtCpy_;
	  
	  // update the debug tables
	  trajResSize_ = this->traj()->getSize();

	  tableNearStateCpy_ = (new PositionXYT* [trajResSize_]);
	  tableNearStStateCpy_ = (new PositionXYT* [trajResSize_]);
	  tableFIntCpy_ = new WCStateT[trajResSize_];
	  tableFExtCpy_ = new WCStateT[trajResSize_];
	  tableStFExtCpy_ = new WCStateT[trajResSize_];

	  for(int i=0; i<trajResSize_; i++)
	    {
	      if(tableNearState_[i] == NULL) tableNearStateCpy_[i] = NULL;
	      else tableNearStateCpy_[i] = new PositionXYT(*(tableNearState_[i]));
	      if(tableNearStState_[i] == NULL) tableNearStStateCpy_[i] = NULL;
	      else tableNearStStateCpy_[i] = new PositionXYT(*(tableNearStState_[i]));
// 	      tableFIntCpy_[i] = *((WCStateT*)(&((*tableFInt)[i])));
 	      tableFExtCpy_[i] = tableFExt_[i];
 	      tableStFExtCpy_[i] = tableStFExt_[i];
	    }
	}
    }

  if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::updateTraj : mutex Traj released" << endl;
}

/*************** Display functions ****************/

/**
 *  \brief update the trajectory deformation process visualization
 *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
 *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
 *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
 *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
 */
void Hugr_WCTrajDeform::updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs)
{
  dispFExt_ = dispFExt;
  dispStFExt_ = dispStFExt;
  dispNearDynObs_ = dispNearDynObs;
  dispNearStObs_ = dispNearStObs;
}

/**
 *  \label get back informations about the deformation process
 *  \param nbDeform : the number of deformations proceeded
 *  \param avgDeformDur : the average deformation duration
 *  \param DPS : the average number of deformations per second
 */
void Hugr_WCTrajDeform::getDeformInfo(int& nbDeform,double& avgDeformDur,double& DPS)
{
  double time = timer_->totalElapsedTime();

  nbDeform = nbDeform_;
  if(time != 0 && nbDeform_ != 0) 
    {
      avgDeformDur = time / nbDeform_;
      DPS = 1.0 / avgDeformDur;
    }
  else 
    {
      avgDeformDur = 0.0;
      DPS = 0.0;
    }
}

/**
 *  \brief draw the controller features in an openGL viewer
 */
void Hugr_WCTrajDeform::toGL()
{
//   cout << "Hugr_WCTrajDeform::toGL() -> debug_ : " << debug_ << "  status_ : " << status_ << "  trajResSize_ : " << trajResSize_ << endl;
  if(debug_ >= 1 && status_ == DEFORMER_STARTED)
    {
//       // block the access to the trajectory during its reinitialization
//       if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : waiting for mutex Traj" << endl;
//       boost::mutex::scoped_lock lock(*mutexTraj2_);
//       if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : mutex Traj locked" << endl;

      // block the access to the displaying tables
      if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : waiting for mutex Tables" << endl;
      boost::mutex::scoped_lock lock(mutexTables_);
      if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : mutex Tables locked" << endl;

      double time = timer_->totalElapsedTime();
      double timeEnv = envReg_->getTime();

      try
	{
	  // display the external forces from dynamic obstacles
	  if(dispFExt_ && tableFExtCpy_ != NULL)
	    {
	      const GLfloat orangeMaterial[]={1.,1.,0.,1.};
	      glMaterialfv(GL_FRONT, GL_AMBIENT, orangeMaterial);
	      glMaterialfv(GL_FRONT, GL_DIFFUSE, orangeMaterial);
	      
	      int N = trajRes_->size();
	      
	      for(int i=0; i<N; i++)
		{
		  glBegin(GL_LINES);
		  glVertex3f(trajRes_->at(i).x(), 
			     trajRes_->at(i).y(), 
			     trajRes_->at(i).t() - time + timeEnv);
		  glVertex3f(trajRes_->at(i).x() + 10 * tableFExtCpy_[i].x(),
			     trajRes_->at(i).y() + 10 * tableFExtCpy_[i].y(),
			     trajRes_->at(i).t() + 10 * tableFExtCpy_[i].t() - time + timeEnv);
		  glEnd();
		}
	    }
	  
// 	  // display the external forces from static obstacles
// 	  if(dispStFExt_ && tableStFExtCpy_ != NULL)
// 	    {
// 	      const GLfloat cyanMaterial[]={0.,1.,1.,1.};
// 	      glMaterialfv(GL_FRONT, GL_AMBIENT, cyanMaterial);
// 	      glMaterialfv(GL_FRONT, GL_DIFFUSE, cyanMaterial);

// 	      int N = trajRes_->size();
	      
// 	      for(int i=0; i<N; i++)
// 		{
// 		  glBegin(GL_LINES);
// 		  glVertex3f(trajRes_->at(i).x(), 
// 			     trajRes_->at(i).y(), 
// 			     trajRes_->at(i).t() - time + timeEnv);
// 		  glVertex3f(trajRes_->at(i).x() + 10 * tableStFExtCpy_[i].x(),
// 			     trajRes_->at(i).y() + 10 * tableStFExtCpy_[i].y(),
// 			     trajRes_->at(i).t() + 10 * tableStFExtCpy_[i].t() - time + timeEnv);
// 		  glEnd();
// 		}	      
// 	    }

	  // display the nearest dynamic obstacles from the state-times of the trajectory
	  if(dispNearDynObs_ && tableNearStateCpy_ != NULL)
	    {
	      const GLfloat blackMaterial[]={0.,0.,0.,1.};
	      glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
	      glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);

	      int nbDef = 0;

	      int N = trajRes_->size();
 	      //for(int i=0; i<trajResSize_; i++)
	      for(int i=0; i<N; i++)
		{
		  if(tableNearStateCpy_[i] != NULL)
		    {
		      glBegin(GL_LINES);
		      glColor3f(0.,0.,0.);
 		      glVertex3f(trajRes_->at(i).x(), 
				 trajRes_->at(i).y(), 
				 trajRes_->at(i).t() - time + timeEnv);
 		      glVertex3f(tableNearStateCpy_[i]->x(), 
				 tableNearStateCpy_[i]->y(), 
				 tableNearStateCpy_[i]->t());
		      glEnd();
		      
 		      //cout << "state : "; trajRes_->at(i).display(); cout << "  nearstate : "; tableNearStateCpy_[i]->display(); cout << endl;
		      
		      nbDef++;
		    }
 		  //else cout << "tableNearStateCpy_[" << i << "] NULL" << endl;
		}
	      
 	      //cout << "nbDef : " << nbDef << endl;
	    }
	  
// 	  // display the nearest static obstacles from the state-times of the trajectory
// 	  if(dispNearStObs_ && tableNearStStateCpy_ != NULL)
// 	    {
// 	      const GLfloat greyMaterial[]={0.6,0.6,0.6,1.};
// 	      glMaterialfv(GL_FRONT, GL_AMBIENT, greyMaterial);
// 	      glMaterialfv(GL_FRONT, GL_DIFFUSE, greyMaterial);

// 	      int N = trajRes_->size();
//  	      //for(int i=0; i<trajResSize_; i++)
// 	      for(int i=0; i<N; i++)
// 		{
// 		  if(tableNearStStateCpy_[i] != NULL)
// 		    {
// 		      glBegin(GL_LINES);
// 		      glColor3f(0.6,0.6,0.6);
//  		      glVertex3f(trajRes_->at(i).x(), 
// 				 trajRes_->at(i).y(), 
// 				 trajRes_->at(i).t() - time + timeEnv);
//  		      glVertex3f(tableNearStStateCpy_[i]->x(), 
// 				 tableNearStStateCpy_[i]->y(), 
// 				 tableNearStStateCpy_[i]->t() - time + timeEnv);
// 		      glEnd();
// 		    }
// 		}
// 	    }
	}
      catch(string str)
	{
	  cout << "Hugr_WCTrajDeform::toGL -> Exception : " << str << endl;
	  return;
	}

//       if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : mutex Traj released" << endl;
      if(MUTEX_DEBUG == 1) cout << "Hugr_WCTrajDeform::toGL : mutex Tables released" << endl;
    }
}
