// #include "hugrSlammot/src/DynamicObstacle.hpp"
// #include "hugrSlammot/src/SlammotState.hpp"
// #include "hugrSlammot/src/SlammotState.hpp"
#include "hugrSlammot/DynamicObstacle.hpp"
#include "hugrSlammot/SlammotState.hpp"
#include "hugrSlammot/SlammotState.hpp"
#include "slammot/occgrid.hpp"
#include "hugrSlammot/OccgridSerializer.hpp"

#include "HEnvironment.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /*************** Class HEnvironment ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
HEnvironment::HEnvironment() : Environment()
{
  this->roadmap_ = NULL;
  this->autobuildRM_ = false;
  this->dispRM_ = false;

  this->timer_ = NULL;
  this->dt_ = 0.05;

  this->status_ = ENVUP_DISCONNECTED;

  this->store_ = NULL;
  lObsId_ = -1;
  OGId_ = -1;
  SlammotStateId_ = -1;

  distFilter_ = 0.1;
  securDist_ = 1.0;

  OGRequest_ = OGR_OFF;
  OGRdt_ = 5.0;
  OGRelapsedT_ = 0.0;
  OGRtotElapsedT_ = 0.0;

  this->threadEnv_ = NULL;
  this->mutexEnv_ = NULL;
}

/**
 *  \brief init constructor
 */
HEnvironment::HEnvironment(hugr::Store* store,Timer* timer,boost::mutex* mutexEnv) : Environment()
{
  this->roadmap_ = NULL;
  this->autobuildRM_ = false;
  this->dispRM_ = false;

  this->timer_ = timer;
  this->dt_ = 0.05;

  this->status_ = ENVUP_DISCONNECTED;

  this->store_ = store;
  lObsId_ = -1;
  OGId_ = -1;
  SlammotStateId_ = -1;

  distFilter_ = 0.1;
  securDist_ = 1.0;

  OGRequest_ = OGR_OFF;
  OGRdt_ = 5.0;
  OGRelapsedT_ = 0.0;
  OGRtotElapsedT_ = 0.0;

  this->threadEnv_ = NULL;
  this->mutexEnv_ = mutexEnv;
}

/**
 *  \brief destructor
 */
HEnvironment::~HEnvironment()
{
  if(this->roadmap_ != NULL) delete roadmap_;
  if(this->threadEnv_ != NULL) delete threadEnv_;
}

/******** hugr store connection functions *********/
  
/**
 *  \brief connect the environment updater to the hugr store
 *  \return 0 if the connection to the hugr store has been correctly established, -1 else
 */
int HEnvironment::connect()
{
  if(store_ == NULL)
    {
      cerr << "HEnvironment::connect : Warning. The HUGR store hasn't been defined." << endl;
      status_ = ENVUP_DISCONNECTED;
      return -1;
    }

  // connect the hugr store if it has not been already done
  store_->connect();

  bool trouveDynObs = true;
  bool trouveOccGrid = true;
  bool trouveSlammotState = true;

  // get back the list of dynamic obstacles id
  try
    {
      lObsId_ = store_->lookupVariable("DynamicObstacles");
    }
  catch(runtime_error e)
    {
      cerr << "HEnvironment::connect : Warning. " << "DynamicObstacles" << " cannot be found in the HUGR store." << endl;
      trouveDynObs = false;
//       status_ = ENVUP_DISCONNECTED;
//       return -1;
    }

  // get back the occupancy grid id
  try
    {
      OGId_ = store_->lookupVariable("SlammotMap");
    }
  catch(runtime_error e)
    {
      cerr << "HEnvironment::connect : Warning. " << "SlammotMap" << " cannot be found in the HUGR store." << endl;
      trouveOccGrid = false;
    }

  // get back the slammot state id
  try
    {
      SlammotStateId_ = store_->lookupVariable("SlammotState");
    }
  catch(runtime_error e)
    {
      cerr << "HEnvironment::connect : Warning. " << "SlammotState" << " cannot be found in the HUGR store." << endl;
      trouveSlammotState = false;
    }

  if(trouveDynObs == false && (trouveOccGrid == false || trouveSlammotState == false))
    {
      status_ = ENVUP_DISCONNECTED;
      return -1;
    }
  else
    {
      status_ = ENVUP_STOPPED;
      return 0;
    }
}


/********************* Actions ********************/

/**
 *  \brief update the Detection and Tracking of Moving Objects
 *  \return 0 if the update process succeed, -1 else
 */
int HEnvironment::updateDaTMO()
{
  DynamicObstacleList listObsHugr;
  try
    {
      store_->readVariable(lObsId_, listObsHugr, true);
    }
  catch(runtime_error e)
    {
      cerr << "HEnvironment::updateSickScan : Warning. The sick scan ID has not been found in the HUGR store." << endl;
      return -1;
    }
  
  DynamicObstacleList::iterator it;

  //  cout << "//////////////////////////////////////////////////" << endl;
  for(it = listObsHugr.begin(); it != listObsHugr.end(); it++)
    {
      // get informations about the current obstacle
      double t = timer_->totalElapsedTime(); // get the current time
      double rad = this->securDist_;         // obstacle radius
      
      // compute the coordinates of the obstacle in the global frame
      PositionXYTh posObs(it->x,it->y,0);
      PositionXYT post(posObs.x(),posObs.y(),t);
      // compute the velocity vector of the obstacle in the global frame
      SpeedXY spd(it->vx,it->vy);
//       spd.rotate(-0.5*M_PI);
 
      DiskObs2D obs(post,spd,rad);
      if(!(this->isFalsePositive(obs))) this->appendObs2D(obs);
//       this->appendObs2D(obs);

//       cout << "cfg : (" << it->x << "," << it->y << ")  spd : (" << it->vx << "," << it->vy << ")   ";
//       cout << "cfg2 : "; post.display(); cout << " spd2 : "; spd.display(); cout << endl; //cout << "   frame : "; pos_.display(); cout << endl;

      if(DATMO_DEBUG)
	{
	  Obs2D* curObs = this->getObs2D(this->getNbObs2D()-1);
	  cout << "frame:"; frame_.display(); 
	  cout << " LocalObs:"; this->getLocalObs2D(this->getNbObs2D()-1).display();
	  cout << " ObsInF:"; curObs->display();
	  delete curObs;
	}
    }  
  //  cout << "//////////////////////////////////////////////////" << endl;
  return 0;
}

/**
 *  \brief update the occupancy grid given by the SLAM process
 *  \return 0 if the update process succeed, -1 else
 */
int HEnvironment::updateOG()
{
  // check if an occupancy grid request must be send
  bool ogr;
  switch(OGRequest_)
    {
    case OGR_OFF : 
      ogr = false;
      break;
    case OGR_ON : 
      ogr = true;
      break;
    case OGR_ONESHOT : 
      ogr = true;
      OGRequest_ = OGR_OFF;
      break;
    case OGR_TIMED : 
      // get the current total elapsed time
      double time = timer_->totalElapsedTime();
      // compute the elapsed time since the last request
      OGRelapsedT_ += (time - OGRtotElapsedT_);
      // update the new total elapsed time
      OGRtotElapsedT_ = time;

      // check if an occupancy grid request must be send
      if(OGRelapsedT_ >= OGRdt_)
	{
	  OGRelapsedT_ -= OGRdt_;
	  ogr = true;
	}
      else ogr = false;
      break;
    }

  if(ogr)
    {
      SlammotState ss = SLAMMOT_MAP_ONESHOT;
      try
	{
	  // send a request for the occupancy grid
	  store_->writeVariable(SlammotStateId_, ss);

	  // wait for the occupancy grid
	  OccupancyGrid occgrid(GRID_ROWS,GRID_COLS);

	  store_->readVariable(OGId_, occgrid, true);
	}
      catch(runtime_error e)
	{
	  cerr << "HEnvironment::updateOG : Warning. Unable to send the occupancy grid request." << endl;
	  return -1;
	}

      cout << "HEnvironment::updateOG : Occupancy grid loaded..." << endl;
    }

  return 0;
}

/**
 *  \brief update the environment thanks to the SLAMMOT process
 *  \return 0 if the update process succeed, -1 else
 */
int HEnvironment::updateEnv()
{
  // block the access to the environment during its construction
  if(MUTEX_DEBUG == 1) cout << "HEnvironment::updateEnv : waiting for mutex Env" << endl;
  boost::mutex::scoped_lock lock(*mutexEnv_);
  if(MUTEX_DEBUG == 1) cout << "HEnvironment::updateEnv : mutex Env locked" << endl;
  
  this->clearObs2D();

  // update the time value of the environment's frame
//   this->frame_.t() = timer_->totalElapsedTime();

  int resUpDaTMO = 0;
  if(lObsId_ != -1) resUpDaTMO = this->updateDaTMO();

  int resUpOG = 0;
  if(OGId_ != -1 && SlammotStateId_ != -1) resUpOG = this->updateOG();      
  
  if(MUTEX_DEBUG == 1) cout << "HEnvironment::updateEnv : mutex Env released" << endl;
  return (resUpDaTMO + resUpOG);
}

/**
 *  \brief execute the control loop
 */
void HEnvironment::loop()
{
  while(status_ == ENVUP_STARTED)
    {
      int resUpEnv = this->updateEnv();

      if(resUpEnv == 0 && status_ == ENVUP_STARTED)
	{
	  usleep((int)(1000000.0*(dt_)));
	}
      else return;
    }
}

/**
 *  \brief init the detection and tracking of moving objects process
 *  \param distFilter : maximal distance from static obstacles to consider the dynamic obstacles perceived as noise
 *  \param securDist : minimal distance kept to dynamic obstacles
 *  \return 0 if the initialization succeed
 *  \return -1 if the DATMO has been already started
 *  \return -2 if the environment perception tool has not been defined
 *  \return -3 if an unknown error occured
 */
int HEnvironment::initDATMO(double distFilter,double securDist)
{
  switch(status_)
    {
    case ENVUP_NULL : 
      {
	cerr << "HEnvironment::initDATMO : Warning. The environment perception tool has not been defined." << endl;
	return -2;
      }
    case ENVUP_DISCONNECTED : 
    case ENVUP_STOPPED : 
      {
	this->distFilter_ = distFilter;
	this->securDist_ = securDist;
	return 0;
      }
    case ENVUP_STARTED : 
      {
	cerr << "HEnvironment::initDATMO : Warning. The environment perception tool must be stopped first." << endl;
	return -1;
      }
    default : 
      return -3;
    }
}

/**
 *  \brief update the location from which the perception process acts
 */
void HEnvironment::updatePose(const PositionXYTh pos) 
{
  this->frame_.x() = pos.x();
  this->frame_.y() = pos.y();
//   this->frame_.th() = (pos.th() - (0.5*M_PI));
  this->frame_.th() = pos.th();

  //cout << "frame : "; frame_.display(); cout << endl;
}

/**
 *  \brief start the timer
 *  \return 0 if the starting process succeed, -1 else
 */
int HEnvironment::start()
{
  switch(status_)
    {
    case ENVUP_NULL : 
    case ENVUP_DISCONNECTED : 
      {
	cerr << "HEnvironment::start : Warning. The environment perception tool is disconnected." << endl;
	return -1;
      }
    case ENVUP_STOPPED : 
      {
	status_ = ENVUP_STARTED;
// 	timer_.start();
	threadEnv_ = new boost::thread(boost::bind(&HEnvironment::loop,this));
	return 0;
      }
    case ENVUP_STARTED : 
      {
	// nothing to do
	return 0;
      }
    default : 
      return -1;
    }
}

/**
 *  \brief stop the timer
 *  \return 0 if the stoping process succeed, -1 else
 */
int HEnvironment::stop()
{
  switch(status_)
    {
    case ENVUP_NULL : 
    case ENVUP_DISCONNECTED : 
      {
	cerr << "HEnvironment::stop : Warning. The environment perception tool is disconnected." << endl;
	return -1;
      }
    case ENVUP_STOPPED : 
      {
	// nothing to do
	return 0;
      }
    case ENVUP_STARTED : 
      {
	status_ = ENVUP_STOPPED;
// 	timer_.stop();
	if(threadEnv_ != NULL) delete threadEnv_;
	return 0;
      }
    default : 
      return -1;
    }
}

/**
 *  \brief provide knowledge of the static environment to the robot
 *  \param filename : filename of the input static environment
 *  \param size : the obstacles are incresed by this size
 *  \param RMfilename : filename of an input roadmap alredy built on the environment
 *  \return 0 if the process succeed
 *          -1 if the file has not been found
 *          -4 if an unknown error occured
 */
int HEnvironment::addStaticEnv(const char* filename,double size,const char* RMfilename)
{
  int res = this->loadSegmentedEnv(filename,size);

  if(res == 0)
    {
      if(RMfilename != NULL) this->loadRoadmap(RMfilename);
      else if(this->autobuildRM()) this->buildRoadmap();
    }

  return res;
}

/**
 *  \brief clear the static environment
 *  \return 0 if the process succeed, -1 else
 *          -4 if an unknown error occured
 */
int HEnvironment::clearStaticEnv()
{
  // clear the static environment
  this->clearFixedObs2D();

  // clear the associated roadmap if needed
  if(roadmap_ != NULL) 
    {
      delete roadmap_;
      roadmap_ = NULL;
    }

  return 0;
}

/**
 *  \brief launch the roadmap building on the static environment
 *  \return 0 if the process succeed, -1 else
 */
int HEnvironment::buildRoadmap()
{
  if(roadmap_ != NULL) delete roadmap_;
  roadmap_ = new DRoadMap(this);
  roadmap_->setNbNodesNeeded(500);
  roadmap_->setMaxDistNgh(0.1*this->getRadius());
  roadmap_->construct();
  return 0;
}

/**
 *  \brief load the roadmap from an input file
 *  \param filename : the input filename
 *  \return 0 if the process succeed, -1 else
 */
int HEnvironment::loadRoadmap(const char* filename)
{
  roadmap_ = new DRoadMap(this);
  int res = roadmap_->load(filename);
  
  // check if the roadmap has been successfully loaded
  if(res != 0 && roadmap_ != NULL)
    {
      delete roadmap_;
      roadmap_ = NULL;
    }
  return res;
}

/**
 *  \brief save the current roadmap
 *  \param filename : the output filename
 *  \return 0 if the process succeed, -1 else
 */
int HEnvironment::saveRoadmap(const char* filename)
{
  // check if the roadmap has been built
  if(roadmap_ == NULL) return -1;

  // save the roadmap into the output file
  int res = roadmap_->save(filename);
  return res;
}


/************** Obstacles filtering ***************/

/**
 *  \brief check if the input obstacles may be a false positive
 *  \param obs : the input obstacle
 *  \return true if the input obstacle is considered as a false positive
 */
bool HEnvironment::isFalsePositive(const DiskObs2D& obs)
{
  DiskObs2D* obsInF = obs.copyFromFrame(frame_);

  // get back the registered position of the obstacle
  PositionXYThT pos = obsInF->lastPosT();

  // check if this position is close to the static obstacles
  PositionXY near;
  double dist;
  bool obstruct = this->staticObstruct(PositionXY(pos.x(),pos.y()),near,dist);

  // delete the copy of the obstacle
  delete obsInF;

  // check if the position of the dynamic osbtacle is inside or too close to the static obstacles
  if(obstruct || dist < distFilter_) return true;
  else return false;
}


/*************** Display functions ****************/

/**
 *  \brief draw the environment in an openGL viewer
 *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
 */
void HEnvironment::toGL(bool dispFM) const
{
  // draw the environment
  this->Environment::toGL(dispFM,timer_->totalElapsedTime());

  // draw the roadmap if needed
  if(dispRM_ && roadmap_ != NULL)
    {
      roadmap_->toGL(timer_->totalElapsedTime());
    }
}
