#include "WheelChair.hpp"
using namespace mpc;

#include "SimpleTrajPlanner.hpp"
#include "RoadMapTP.hpp"
#include "Hugr_WCTrajFollow.hpp"
#include "Hugr_WC_TFD.hpp"

                                    /**************************************************/
                                    /**************** Class WheelChair ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
WheelChair::WheelChair() : Robot()
{
}

/*
 *  \brief default constructor
 *  \param name : the name of the robot
 *  \param store : the hugr store containing informations about the robot
 */
WheelChair::WheelChair(std::string name,hugr::Store* store) : Robot(name,store)
{
  this->sys_ = new WheelChairSys();
  this->statet_ = new WCStateT();
  //this->ctrl_ = new WCControl();

  // init the store variable names
  
  this->WC_poseName_ = string("wheelchair_pose");
  this->WC_commandName_ = string("wheelchair_command");

  // use a simple controller
  this->controller_ = new WCController(store,WC_commandName_);

  // init a simple planner
  this->initPlanner(SIMPLE_TP);

  // update the system from the hugr store
  int resUp = this->updateSys();
  if(resUp != 0)
    {
      cerr << "WheelChair::WheelChair : Warning. A problem occured during the system update." << endl;
    }
}

/*
 *  \brief default constructor
 *  \param store : the hugr store containing informations about the robot
 *  \param namePos : the name of wheelchair's pose in the hugr store
 *  \param nameCmd : the name of wheelchair's ommand in the hugr store
 */
WheelChair::WheelChair(hugr::Store* store,std::string namePos,std::string nameCmd) : Robot(string("wheelchair"),store)
{
  this->sys_ = new WheelChairSys();
  this->statet_ = new WCStateT();
  //this->ctrl_ = new WCControl();

  // init the store variable names
  
  this->WC_poseName_ = namePos;
  this->WC_commandName_ = nameCmd;

  // use a simple controller
  this->controller_ = new WCController(store,WC_commandName_);

  // init a simple planner
  this->initPlanner(SIMPLE_TP);

  // update the system from the hugr store
  int resUp = this->updateSys();
  if(resUp != 0)
    {
      cerr << "WheelChair::WheelChair : Warning. A problem occured during the system update." << endl;
    }
}

// /*
//  *  \brief default constructor
//  *  \param name : the name of the robot
//  *  \param sys : the robotic system configuration of the robot
//  */
// WheelChair::WheelChair(std::string name,WheelChairSys* sys) : Robot(name,sys)
// {
//   this->statet_ = new WCStateT();
//   //this->ctrl_ = new WCControl();
// }

/*
 *  \brief destructor
 */
WheelChair::~WheelChair()
{
}


/************** Getters and setters ***************/

/*
 *  \brief return the current control applied on the wheelchair
 *  \return the current control applied on the wheelchair
 */
const WCControlT* WheelChair::controlt()
{
  if(MUTEX_DEBUG == 1) cout << "WheelChair::controlt : waiting for mutex Ctrl" << endl;
  boost::mutex::scoped_lock lock(mutexCtrl_);
  if(MUTEX_DEBUG == 1) cout << "WheelChair::controlt : mutex Ctrl locked" << endl;

  return this->controller()->controlt();

  if(MUTEX_DEBUG == 1) cout << "WheelChair::controlt : mutex Traj released" << endl;
}

/*
 *  \brief return true if the robot is moving
 *  \return true if the robot is moving
 */
bool WheelChair::isMoving()
{
  const WCControlT* ctrl = this->controlt();
  if(ctrl != NULL) { cout << "ctrl : "; ctrl->display(); cout << endl; }
  else cout << "ctrl NULL" << endl;
  return (ctrl != NULL && (ctrl->v() != 0 || ctrl->av() != 0));
}

/*
 *  \brief stop the robot motion
 */
void WheelChair::stopMoving()
{
  // check if the robot is moving
  bool moving = this->isMoving();
  if(moving)
    {
      // stop the controller
      this->stopCtrler();

      // stop the robot motion
      this->initCtrler(SIMPLE_CTRLER);
      this->updateCtrl(WCControl(0,0));
      this->startCtrler();
    }
}


/*************** Hugr store reading ***************/
  
#ifdef HUGR
/*
 *  \brief update the system configuration from the HUGR store
 *  \return 0 if the update succeed
 *         -1 if the hugr store has not been defined
 *         -2 if the wheelchair pose has not been found in the store
 *         -3 if the wheelchair command has not been found in the store
 */
int WheelChair::updateSys()
{
  if(store_ == NULL)
    {
      cerr << "WheelChair::updateSys : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }

//   // get back the wheelchair constants
//   std::string str(name_);
//   try
//     {
//       str += "_constants";
//       WC_constantsId = store_->lookupVariable(str.c_str());
//     }
//   catch(runtime_error e)
//     {
//       cerr << "WheelChair::updateSys : Warning. " << str.c_str() << "cannot be found in the hugr store" << endl;
//       return -1;
//     }
  
//   WheelchairConstants WC_constants;
//   store_->readVariable(WC_constantsId, WC_constants, false);

//   // update the system configuration of the wheelchair
//   this->sys()->setLength(WC_constants.length);
//   this->sys()->setWidth(WC_constants.width);
//   this->sys()->setWheelRad(WC_constants.wheel_radius);
//   this->sys()->setWheelBase(WC_constants.wheel_base);
//   this->sys()->setVMax(WC_constants.max_speed);
//   this->sys()->setAMax(WC_constants.max_acc);
//   this->sys()->setAVMax(WC_constants.max_rotation_speed);
//   this->sys()->setAAMax(WC_constants.max_rotation_acc);
//   this->sys()->setBrakingDist(WC_constants.braking_dist);

  // update the system configuration of the wheelchair
  WheelchairConstants WC_constants = wheelchair_default_constants;
  this->sys()->setLength(WC_constants.length);
  this->sys()->setWidth(WC_constants.width);
  this->sys()->setWheelRad(WC_constants.wheel_radius);
  this->sys()->setWheelBase(WC_constants.wheel_base);
  this->sys()->setVMax(WC_constants.max_speed);
  this->sys()->setAMax(WC_constants.max_acc);
  this->sys()->setAVMax(WC_constants.max_rotation_speed);
  this->sys()->setAAMax(WC_constants.max_rotation_acc);
  this->sys()->setBrakingDist(WC_constants.braking_dist);

  // display the system configuration
  this->sys()->display();

  // connect the controller to the hugr store
  int resConnecCtrler = this->controller_->connect();
  if(resConnecCtrler != 0)
    {
      cerr << "WheelChair::updateSys : Warning. The wheelchair cannot be controled." << endl;
      return -1;
    }

  // connect the localizer to the hugr store
  // just check here if the pose is defined
  int resConnecLoczer = this->updatePose();
  if(resConnecLoczer != 0)
    {
      cerr << "WheelChair::updateSys : Warning. The wheelchair cannot be localized." << endl;
      this->loczerStatus_ = LOCZER_DISCONNECTED;
      return -1;
    }
  this->loczerStatus_ = LOCZER_STOPPED;
  
  return 0;
}

/*
 *  \brief update the robot pose from the HUGR store
 */
int WheelChair::updatePose()
{
  if(store_ == NULL)
    {
      cerr << "WheelChair::updatePose : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }

  // get back the wheelchair pose id
  try
    {
      WC_poseId = store_->lookupVariable(WC_poseName_.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "WheelChair::updatePose : Warning. " << WC_poseName_.c_str() << " cannot be found in the hugr store" << endl;
      return -1;
    }
  
  WheelchairPose WC_pose;
  store_->readVariable(WC_poseId, WC_pose, false);

  // update the state of the system with the pose found
  this->statet()->x() = WC_pose.position.x;
  this->statet()->y() = WC_pose.position.y;
  this->statet()->th() = WC_pose.position.theta;
  this->statet()->t() = timer_->totalElapsedTime();

//   cout << "state : "; this->statet()->display(); cout << endl;

  // update the state registered in the controller if needed
  if(controller_ != NULL)
    this->controller_->updateStateT(*(statet()));

  // update the state registered in the environment perception process
  if(env_ != NULL)
    this->env_->updatePose(PositionXYTh(WC_pose.position.x,WC_pose.position.y,normalizedAngle(WC_pose.position.theta)));

  return 0;
}

// /**
//  *  \brief send the control to apply on the robot to the HUGR store
//  */
// int WheelChair::sendControl()
// {
//   if(controller_ == NULL) return -1;
//   else return controller_->sendControl();
// }


/************* Localization functions *************/

/*
 *  \brief compute the current deviation between the localization of the robot and the trajectory followed
 *  \param trajSt : the current state-time of the trajectory (set during the process)
 *  \param loc : the current localization state-time (set during the process)
 *  \param dev : the current deviation between the localization of the robot and the trajectory followed (set during the process)
 *  \param ctrlt : the current control applied on the robot
 *
 *  Every input parameters must be previously allocated
 *  Their value are set during the deviation computation
 */
void WheelChair::computeCurDeviation(StateTime* trajSt,StateTime* loc,StateTime* dev,ControlTime* ctrlt)
{
  WCStateT* WCtrajSt = ((WCStateT*) trajSt);
  WCStateT* WCloc = ((WCStateT*) loc);
  WCStateT* WCdev = ((WCStateT*) dev);
  WCControlT* WCctrlt = ((WCControlT*) ctrlt);

  double time = this->statet()->t();

  WCctrlt->v() = this->controlt()->v();
  WCctrlt->av() = this->controlt()->av();
  WCctrlt->t() = this->controlt()->t();

  WCloc->x()  = this->statet()->x();
  WCloc->y()  = this->statet()->y();
  WCloc->th() = this->statet()->th();
  WCloc->s()  = this->statet()->s();
  WCloc->om() = this->statet()->om();
  WCloc->t()  = this->statet()->t();

  // get back the state of the trajectry followed at the localization time
  if(this->traj() != NULL) 
    {
      //this->traj()->getStateAtTime(time,*((WCStateT*)trajSt));
      this->controller()->getStateAtAbsTime(time,*WCtrajSt);

      cout << "traj not null... t : " << time << "  loc : "; loc->display(); cout << "  trajSt : "; trajSt->display();

      WCdev->x()  = WCloc->x() - WCtrajSt->x();
      WCdev->y()  = WCloc->y() - WCtrajSt->y();
      WCdev->th() = mpc::computeMinDiffAngle(WCloc->th(),WCtrajSt->th());
      WCdev->s()  = WCloc->s() - WCtrajSt->s();
      WCdev->om() = WCloc->om() - WCtrajSt->om();
      WCdev->t()  = WCloc->t();

      cout << "  dev : "; dev->display(); cout << endl;
    }
  else
    {
      cout << "traj null" << endl;
      WCtrajSt->x() = 0;
      WCtrajSt->y() = 0;
      WCtrajSt->th() = 0;
      WCtrajSt->s() = 0;
      WCtrajSt->om() = 0;
      WCtrajSt->t() = 0;

      WCdev->x() = 0;
      WCdev->y() = 0;
      WCdev->th() = 0;
      WCdev->s() = 0;
      WCdev->om() = 0;
      WCdev->t() = 0;
    }
}


/*************** Control functions ****************/

/*
 *  \brief change the controller type
 *  \param type : the new controller type
 *  \return 0 if the initialization succeed
 *  \return -1 if the current controller is running
 *  \return -2 if such a controller has already been initialized
 *  \return -3 if the input controller type is wrong or unknown
 *  \return -4 if an unknown problem occured
 *  \return -5 if the connection to the store cannot be established
 *  \return -6 if a required input of the controller is missing
 */
int WheelChair::initCtrler(controller_t type)
{
  // check if the current controller (if it exists) can be replaced
  if(this->controller_ != NULL)
    {
      if(this->controller_->ctrlerType() == type && type != TRAJFOLLOW_CTRLER)
	{
	  // nothing t do : such a controller has been already initialized
	  return -2;
	}
      else switch(this->controller_->status())
	{
	case CTRLER_STARTED : 
	  // cannot change the controller : the current controller is running
	  return -1;
	case CTRLER_DISCONNECTED : 
	case CTRLER_STOPPED : 
	  {
// 	    if(MUTEX_DEBUG == 1) cout << "WheelChair::initCtrler : waiting for mutex Loc" << endl;
// 	    boost::mutex::scoped_lock lockLoc(mutexLoc_);
// 	    if(MUTEX_DEBUG == 1) cout << "WheelChair::initCtrler : mutex Loc locked" << endl;

	    // delete the current controller
// 	    cout << "argh1" << endl;
	    //delete this->controller_;
	    this->controller_ = NULL;
// 	    cout << "argh2" << endl;
// 	    if(MUTEX_DEBUG == 1) cout << "WheelChair::initCtrler : mutex Loc released" << endl;
	  }
	  break;
	default : 
	  return -4;
	}
    }

  // initialize a new controller
  switch(type)
    {
    case SIMPLE_CTRLER : 
      this->controller_ = new WCController(this->store_,this->WC_commandName_);
      break;
    case TF2_CTRLER : 
      this->controller_ = new WCController2(this->store_,this->WC_commandName_);
      break;
    case TRAJFOLLOW_CTRLER : 
      this->controller_ = new Hugr_WCTrajFollow(sys(),traj(),store_,WC_commandName_,&mutexLoc_,&mutexCtrl_,&mutexTraj_,&mutexEnv_);
      break;
    case TEDDY_CTRLER : 
      if(traj() == NULL)
	{
	  cerr << "WheelChair::initCtrler : traj() null" << endl;
	  this->controller_ = NULL;
	  return -6;
	}
      this->controller_ = new Hugr_WC_TFD(sys(),env(),traj(),store_,WC_commandName_,&mutexLoc_,&mutexCtrl_,&mutexTraj_,&mutexEnv_);
      break;
    default : 
      cerr << "WheelChair::initCtrler : Warning. Wrong or unknown controller type. Such a controller cannot be defined." << endl;
      return -3;
    }

  // connect it to the hugr store
  int resConnec = this->controller_->connect();
  if(resConnec != 0)
    {
      cerr << "WheelChair::initCtrler : Warning. The Connection of the controller to the store cannot be established." << endl;
      return -5;
    }
  return 0;
}
#endif


/*************** Planner functions ****************/

/*
 *  \brief change the planner type
 *  \param type : the new planner type
 *  \return 0 if the initialization succeed
 *  \return -2 if such a planner has already been initialized
 *  \return -3 if the input planner type is wrong or unknown
 *  \return -4 if a roadmap is required
 *  \return -5 if the environment has not been initialized
 *  \return -6 if an unknown problem occured
 */
int WheelChair::initPlanner(trajPlanner_t type)
{
  // check if the current controller (if it exists) can be replaced
  if(this->planner_ != NULL)
    {
      if(this->planner_->type() == type)
	{
	  // nothing t do : such a planner has been already initialized
	  return -2;
	}
      else
	{
	  // delete the current planner
	  delete this->planner_;
	}
    }

  // initialize a new planner
  switch(type)
    {
    case SIMPLE_TP : 
      this->planner_ = new WCSimpleTrajPlanner(sys(),NULL);
      break;
//     case PVD_TP : 
//       this->planner_ = new 
//       break;
    case ROADMAP_TP : 
      // check if the environment has been initialized
      if(env_ == NULL) return -5;
      // check if a roadmap is available
      if(env_->roadmap() == NULL) return -4;
      // set the properties of the planner
//       ((RoadMapTP*)planner_)->setTimeStep(0.1);
      RoadMapTP::setMSC(0.7);
      RoadMapTP::setMAC(0.7);
      RoadMapTP::minLengthTraj_ = 0.5;
      RoadMapTP::maxLengthTraj_ = 1.5;
      // plan the trajectory on the roadmap
      planner_ = new WCRoadMapTP(sys(),env_,env_->roadmap());
      break;
    default : 
      cerr << "WheelChair::initPlanner : Warning. Wrong or unknown planner type. Such a planner cannot be defined." << endl;
      return -3;
    }
  return 0;
}

/*
 *  \brief reset the trajectory followed
 *  \param traj : the new trajectory
 */
void WheelChair::resetTraj(const STrajectory* traj)
{
  cout << "RESET TRAJ!" << endl;
  // stop the motion of the robot
  cout << "Stop Moving" << endl;
  this->stopMoving();

  // stop and reinit the controller
  cout << "Stop Controller" << endl;
  this->stopCtrler();
  cout << "Reinit Controller" << endl;
  this->resetCtrler();

  // reset the new trajectory followed
  //if(traj_ != NULL) delete traj_;                       ////////////////////////////////// NON??????????????? A TESTER
  cout << "Reinit Trajectory" << endl;
  traj_ = new WCSTraj(*((WCSTraj*)traj));
}

/*
 *  \brief start the control process of the robot
 *  \param goalS : the goal state
 *  \param t : the final time at which the goal state must be reached
 *  \param dt : time step between each state-time of the trajectory
 *  \param coefSpd : maximal speed coefficient
 *  \param coefAcc : maximal acceleration coefficient
 *  \return 0 if the controller starting process succeed
 *  \return -1 if no trajectory has been found
 *  \return -2 if the planner hasn't been initialized
 *  \return -3 if the planner type is not handled
 *  \return -4 if a roadmap is required
 *  \return -5 if the environment has not been initialized
 *  \return -6 if the robot is still moving
 */
int WheelChair::planTraj(const WCState& goalS,double t,double dt,double coefSpd,double coefAcc)
{
  // check if the planner has been initialized
  if(this->planner_ == NULL)
    {
      cerr << "WheelChair::planTraj : The planner hasn't been initialized..." << endl;
      return -2;
    }

  // check if the the robot is still in motion
  if(this->isMoving())
    {
      cerr << "WheelChair::planTraj : The robot is still moving. Cannot plan a new path." << endl;
      return -6;
    }

  // plan the trajectory between the current state of the robot and the goal state
  int res;
  switch(this->planner_->type())
    {
    case SIMPLE_TP : 
      res = ((WCSimpleTrajPlanner*)planner_)->computeTraj(statet()->toState(),goalS,t,dt);
      break;
//     case PVD_TP : 
//       break;
    case ROADMAP_TP : 
      ((WCRoadMapTP*)planner_)->reinit();
      ((WCRoadMapTP*)planner_)->setTimeStep(dt);
      ((WCRoadMapTP*)planner_)->setMSC(coefSpd);
      ((WCRoadMapTP*)planner_)->setMAC(coefAcc);
      res = ((WCRoadMapTP*)planner_)->computeTraj(statet()->toState(),goalS);
      break;
    default : 
      cerr << "WheelChair::planTraj : Wrong or unknown planner type." << endl;
      return -3;
    }

  // check if the trajectory has been correctly computed
  if(res == TRAJ_CPTD)
    {
      // lock a mutex to prevent trajectory accesses
      if(MUTEX_DEBUG == 1) cout << "WheelChair::planTraj : waiting for mutex Traj" << endl;
      boost::mutex::scoped_lock lock(mutexTraj_);
      if(MUTEX_DEBUG == 1) cout << "WheelChair::planTraj : mutex Traj locked" << endl;

      // set the computed trajectory as the current trajectory followed
      this->resetTraj(planner_->getTraj());
//       //if(traj_ != NULL) delete traj_;                                     ////////////////////////////////// NON??????????????? A TESTER
//       traj_ = new WCSTraj(*((WCSTraj*)(planner_->getTraj())));
//       //       // init the current time of the supposed reached waypoint of the trajectory
//       //       this->curTimeOnTraj_ = 0;

      if(MUTEX_DEBUG == 1) cout << "WheelChair::planTraj : mutex Traj released" << endl;
      return 0;
    }
  else
    {
      cerr << "WheelChair::planTraj : No trajectory has been found" << endl;
      return -1;
    }
}


/************* Deformation functions **************/

/**
 *  \brief set the deformation parameters
 *  \param k_int : internal forces coefficient
 *  \param k_ext : external forces coefficient (dynamic obstacles)
 *  \param k_stext : external forces coefficient (static obstacles)
 *  \param dyn_med : maximal effective distance (dynamic obstacles)
 *  \param st_med : maximal effective distance (static obstacles)
 *  \param dyn_med : maximal temporal effective distance
 */
void WheelChair::setTeddyParams(double k_int,double k_ext,double k_stext,double dyn_med,double st_med,double tmp_med)
{
  // check if a deformation process has been initialized
  if(this->ctrlerType() == TEDDY_CTRLER) ((Hugr_WC_TFD*)controller_)->setTeddyParams(k_int,k_ext,k_stext,dyn_med,st_med,tmp_med);
}

/**
 *  \label get back informations about the deformation process
 *  \param nbDeform : the number of deformations proceeded
 *  \param avgDeformDur : the average deformation duration
 *  \param DPS : the average number of deformations per second
 */
void WheelChair::getDeformInfo(int& nbDeform,double& avgDeformDur,double& DPS)
{
  if(this->ctrlerType() == TEDDY_CTRLER) ((Hugr_WC_TFD*)controller_)->getDeformInfo(nbDeform,avgDeformDur,DPS);
  else 
    {
      nbDeform = 0;
      avgDeformDur = 0;
      DPS = 0;
    }
}


/*************** Display functions ****************/

/**
 *  \brief update the trajectory deformation process visualization
 *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
 *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
 *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
 *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
 */
void WheelChair::updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs)
{
  // check if a deformation process has been initialized
  if(this->ctrlerType() == TEDDY_CTRLER) ((Hugr_WC_TFD*)controller_)->updateTeddyView(dispFExt,dispStFExt,dispNearDynObs,dispNearStObs);
}

/*
 *  \brief display the trajectory followed by the robot
 */
void WheelChair::trajGL()
{
  // check if the trajectory has been defined
  if(traj_ != NULL)
    {
      // lock a mutex to prevent trajectory accesses
      if(MUTEX_DEBUG == 1) cout << "WheelChair::trajGL : waiting for mutex Traj" << endl;
      boost::mutex::scoped_lock lock(mutexTraj_);
      if(MUTEX_DEBUG == 1) cout << "WheelChair::trajGL : mutex Traj locked" << endl;

      // copy the current trajectory
      WCSTraj traj(*(this->traj()));

      // get back the current time of the robot
      double curT = this->statet()->t();
//       cout << "curT : " << curT;

      // get back the time reached on the trajectory
      double trajT = this->curTimeOnTraj();
//       cout << "  trajT : " << trajT << endl;

      // shift the trajectory to the current time
      traj.shift(curT-trajT);

      // display the trajectory in an openGL viewer
      traj.toGL();

      // if the trajectory is deformed, display the 
      // deformation features in  an openGL viewer
      if(this->ctrlerType() == TEDDY_CTRLER)
	{
	  ((Hugr_WC_TFD*)controller_)->toGL();
	}

      if(MUTEX_DEBUG == 1) cout << "WheelChair::trajGL : mutex Traj released" << endl;
    }
}
