#include "TrajGenIfc.hpp"

#include <time.h>
#include <sys/time.h>

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
namespace po = boost::program_options;
namespace fs = boost::filesystem;

#include "environment/inc/Map.hpp"
#include "gnuplot/inc/gnuplot_i.hpp"
#include "common/inc/Utils.hpp"

#include "sslattice/inc/CL_SSLattice.hpp"
#include "sslattice/inc/WC_SSLattice.hpp"
#include "sslattice/inc/SS_SSLattice.hpp"

#include "trajGen/inc/CLInitialGuess.hpp"
#include "trajGen/inc/WCInitialGuess.hpp"

#include "gnuplot/inc/TrajPlot.hpp"

#define THRESHOLD 200
#define INCREASING 20

// different system used
#define CARLIKE_SYS 0
#define SPACESHIP_SYS 1
// #define DIFFDRIVE_SYS 2
#define WHEELCHAIR_SYS 3
// #define CARLIKE2_SYS 4
// #define WHEELCHAIR2_SYS 5
// #define PM3D_SYS 6
// #define MANIP_ARM1_SYS 7
// #define SPACESHIP2_SYS 8
#define MANA_SYS 9

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// default constructor (only used by subclasses)
TrajGenIfc::TrajGenIfc(){}

// init constructor (creation of the window)
TrajGenIfc::TrajGenIfc(FXApp* a):FXMainWindow(a,"Trajectory Generation",NULL,NULL,DECOR_ALL,0,0,950,650)
{
  /***************************************************************************************************************
   ********************************************** initialisation *************************************************
   ***************************************************************************************************************/

  /*******************************************************************
   ************************* system components ***********************
   *******************************************************************/

  // the robotic system handled
  sys_ = NULL;

  // the trajectory generator
  trajGenerator = NULL;

  // the state space lattice
  lattice_ = NULL;

  // initial guess of parameters database and learning module
  ig_ = NULL;

  // initial and final state-times
  initST_ = NULL;
  goalST_ = NULL;

  // the map used
  map = NULL;
  wtI = 0;
  htI = 0;

  tabPosTrajX.clear();
  tabPosTrajY.clear();
  tabPosTrajZ.clear();
  tabPosTrajT.clear();
  tabPosTrajV.clear();
  tabPosTrajVX.clear();
  tabPosTrajVY.clear();
  tabPosTrajVZ.clear();
  tabPosTrajTmpX = NULL;
  tabPosTrajTmpY = NULL;
  tabPosTrajTmpZ = NULL;
  tabPosTrajTmpT = NULL;
  tabPosTrajTmpV = NULL;
  tabPosTrajTmpVX = NULL;
  tabPosTrajTmpVY = NULL;
  tabPosTrajTmpVZ = NULL;

  this->initSampleTraj();

  /*******************************************************************
   *********************** interface components **********************
   *******************************************************************/

  genTrajMethod = 1;
  genTrajMethod_target.connect(genTrajMethod);

  filename = FXString("../map/greatEmptyMap.pgm");
  filename_target.connect(filename);

  // init and goal configurations
  x1 = 0;
  y1 = 0;
  theta1 = 0;
  kp1 = 0;
  //v1 = 0.01;
  v1 = 0;
  t1 = 0;
  x2 = 0;
  y2 = 0;
  theta2 = 0;
  kp2 = 0;
  v2 = 0;
  //t2 = 25;
  t2 = 5;

  vx1 = 0.01;
  vy1 = 0.01;
  om1 = 0.0;
  vx2 = 0;
  vy2 = 0;
  om2 = 0;

  x1_target.connect(x1);
  y1_target.connect(y1);
  theta1_target.connect(theta1);
  kp1_target.connect(kp1);
  v1_target.connect(v1);
  t1_target.connect(t1);
  x2_target.connect(x2);
  y2_target.connect(y2);
  theta2_target.connect(theta2);
  kp2_target.connect(kp2);
  v2_target.connect(v2);
  t2_target.connect(t2);

  om1_target.connect(om1);
  vx1_target.connect(vx1);
  vy1_target.connect(vy1);

  om2_target.connect(om2);
  vx2_target.connect(vx2);
  vy2_target.connect(vy2);
  
  rb_Scale = 100.0;
  rb_L = 20.0;
  rb_VMAX = 64.0;
  rb_AMAX = 7.0;
  rb_KMAX = 45.0;
  rb_OMAX = 45.0;
//   rb_AVMAX = 0.5*M_PI;
//   rb_AAMAX = 0.10*M_PI;

  rb_Scale_target.connect(rb_Scale);
  rb_L_target.connect(rb_L);
  rb_VMAX_target.connect(rb_VMAX);
  rb_AMAX_target.connect(rb_AMAX);
  rb_KMAX_target.connect(rb_KMAX);
  rb_OMAX_target.connect(rb_OMAX);

//   rb_AVMAX_target.connect(rb_AVMAX);
//   rb_AAMAX_target.connect(rb_AAMAX);

  WC_AccMax = 3.0;
  WC_AAccMax = 0.1*M_PI;
  WC_VelMax = 15.0;
  WC_AVelMax = 0.3*M_PI;

  WC_AccMax_target.connect(WC_AccMax);
  WC_AAccMax_target.connect(WC_AAccMax);
  WC_VelMax_target.connect(WC_VelMax);
  WC_AVelMax_target.connect(WC_AVelMax);

  SS_AccMax = 15.0;
  SS_AAccMax = 0.1*M_PI;
  SS_VelMax = 60.0;
  SS_AVelMax = 0.3*M_PI;

  SS_AccMax_target.connect(SS_AccMax);
  SS_AAccMax_target.connect(SS_AAccMax);
  SS_VelMax_target.connect(SS_VelMax);
  SS_AVelMax_target.connect(SS_AVelMax);

  debugLevel = 1;
  maxNbStep = 20;
  maxResidual = 0.5;

  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;

  canMoveBackCL = true;            // true if the car-like can move backward
  dispCurvCirc = false;

  useVelCstWC = true;
  useAccCstWC = true;
  useAVelCstWC = true;
  useAAccCstWC = false;

  canMoveBackWC = true;            // true if the wheelchair can move backward

  debugLevel_target.connect(debugLevel);
  maxNbStep_target.connect(maxNbStep);
  maxResidual_target.connect(maxResidual);

  nbBranches = 18;
  minAngRS = -180.0;
  maxAngRS = 180.0;
  minDistRS = 80.0;
  maxDistRS = 360.0;
  nbItvDistRS = 6;
  trajByBranch = 5;
  diffAngle = 50.0;

  nbBranches_target.connect(nbBranches);
  minAngRS_target.connect(minAngRS);
  maxAngRS_target.connect(maxAngRS);
  minDistRS_target.connect(minDistRS);
  maxDistRS_target.connect(maxDistRS);
  nbItvDistRS_target.connect(nbItvDistRS);
  trajByBranch_target.connect(trajByBranch);
  diffAngle_target.connect(diffAngle);

  drawUncompleteTraj = false;
  drawTargetPos = false;
  
  initColor = FXRGB(255,0,0);
  goalColor = FXRGB(0,0,255);

  optSys = 0;
  optSys_target.connect(optSys);

  dd_WVMAX = 8.0*M_PI;
  dd_WAMAX = 1.5*M_PI;
  dd_RAD = 10.0;

  dd_WVMAX_target.connect(dd_WVMAX);
  dd_WAMAX_target.connect(dd_WAMAX);
  dd_RAD_target.connect(dd_RAD);

  timeCsttStatus = 0; // 0 -> fixed;  1 -> free;  2 -> constrained
  minTime = t2 * 0.8;
  maxTime = t2 * 1.2;

  timeCsttStatus_target.connect(timeCsttStatus);
  minTime_target.connect(minTime);
  maxTime_target.connect(maxTime);

  SSL_height = 3;
  SSL_nbCS = 3;
  SSL_height_target.connect(SSL_height);
  SSL_nbCS_target.connect(SSL_nbCS);

  SSL_dispTraj = true;
  SSL_trajSp = 30;
  SSL_trajSp_target.connect(SSL_trajSp);

  // test the reachability of the final state
  // 0 : Never  1 : Only when the target is not reached  2 : Always
  testReachability = 0;
  testReachability_target.connect(testReachability);


  x1_3D = 200.0;
  y1_3D = 200.0;
  z1_3D = 200.0;
  vx1_3D = -10.0;
  vy1_3D = 10.0;
  vz1_3D = -10.0;
  t1_3D = 0.0;

  x2_3D = 400.0;
  y2_3D = 400.0;
  z2_3D = 400.0;
  vx2_3D = 20.0;
  vy2_3D = 20.0;
  vz2_3D = -20.0;
  t2_3D = 15.0;

  PM3D_AX = 12.0;
  PM3D_AY = 12.0;
  PM3D_AZ = 12.0;
  PM3D_VX = 70.0;
  PM3D_VY = 70.0;
  PM3D_VZ = 70.0;

  x1_3D_target.connect(x1_3D);
  y1_3D_target.connect(y1_3D);
  z1_3D_target.connect(z1_3D);
  vx1_3D_target.connect(vx1_3D);
  vy1_3D_target.connect(vy1_3D);
  vz1_3D_target.connect(vz1_3D);
  t1_3D_target.connect(t1_3D);

  x2_3D_target.connect(x2_3D);
  y2_3D_target.connect(y2_3D);
  z2_3D_target.connect(z2_3D);
  vx2_3D_target.connect(vx2_3D);
  vy2_3D_target.connect(vy2_3D);
  vz2_3D_target.connect(vz2_3D);
  t2_3D_target.connect(t2_3D);

  MA_TH1 = M_PI * 0.1;
  MA_TH2 = -M_PI * 0.4;
  MA_TH3 = -M_PI * 0.3;
  MA_OM1 = 0.0;
  MA_OM2 = 0.0;
  MA_OM3 = 0.0;

  MA_AAcc1 = M_PI * 0.05;
  MA_AAcc2 = M_PI * 0.05;
  MA_AAcc3 = M_PI * 0.05;
  MA_AVel1 = M_PI * 0.25;
  MA_AVel2 = M_PI * 0.25;
  MA_AVel3 = M_PI * 0.25;

  PM3D_AX_target.connect(PM3D_AX);
  PM3D_AY_target.connect(PM3D_AY);
  PM3D_AZ_target.connect(PM3D_AZ);
  PM3D_VX_target.connect(PM3D_VX);
  PM3D_VY_target.connect(PM3D_VY);
  PM3D_VZ_target.connect(PM3D_VZ);

  MA_TH1_target.connect(MA_TH1);
  MA_TH2_target.connect(MA_TH2);
  MA_TH3_target.connect(MA_TH3);
  MA_OM1_target.connect(MA_OM1);
  MA_OM2_target.connect(MA_OM2);
  MA_OM3_target.connect(MA_OM3);

  MA_AAcc1_target.connect(MA_AAcc1);
  MA_AAcc2_target.connect(MA_AAcc2);
  MA_AAcc3_target.connect(MA_AAcc3);
  MA_AVel1_target.connect(MA_AVel1);
  MA_AVel2_target.connect(MA_AVel2);
  MA_AVel3_target.connect(MA_AVel3);

  
  // initial guess learning components
  IG_debugMode = 1;
  IG_minNNTSR = 5;

  IG_debugMode_target.connect(IG_debugMode);
  IG_minNNTSR_target.connect(IG_minNNTSR);

  IGfilename = FXString("../ig/WC_IG.dat");
  IGfilename_target.connect(IGfilename);

  IG_tmin = 0.0;
  IG_tmax = 5.0;
  IG_tmin_target.connect(IG_tmin);
  IG_tmax_target.connect(IG_tmax);

  useIG = false;
  useFDPD = false;
  saveRStar = true;

  IG_1stParams = SC_FIRSTIG;                       // initialization mode of the IG learning process
                                                   // (RAND_FIRSTIG, SC_FIRSTIG, SCP_FIRSTIG, RSTAR_FIRSTIG, LOADFROMFILE_FIRSTIG)
  IG_1stParams_target.connect(IG_1stParams);

  IG1stParamsName = FXString("../ig/WC_RStar.ltp");
  IG1stParamsName_target.connect(IG1stParamsName);

  CL_sampAcc = 8;               // sampling of the acceleration control of a car-like for IG learning
  CL_sampStSpd = 8;             // sampling of the steering speed control of a car-like for IG learning
  WC_sampAcc = 8;               // sampling of the linear acceleration control of a wheelchair for IG learning
  WC_sampAAcc = 8;              // sampling of the angular acceleration control of a wheelchair for IG learning
  SS_sampAcc = 8;               // sampling of the linear acceleration control of a spaceship for IG learning
  SS_sampAAcc = 8;              // sampling of the angular acceleration control of a spaceship for IG learning

  CL_sampAcc_target.connect(CL_sampAcc);
  CL_sampStSpd_target.connect(CL_sampStSpd);
  WC_sampAcc_target.connect(WC_sampAcc);
  WC_sampAAcc_target.connect(WC_sampAAcc);
  SS_sampAcc_target.connect(SS_sampAcc);
  SS_sampAAcc_target.connect(SS_sampAAcc);

  CL_samp_t = 9;                // sampling of the time for the CL initial guess learning
  CL_samp_vi = 9;               // sampling of the initial speed for the CL initial guess learning
  CL_samp_vg = 9;               // sampling of the final speed for the CL initial guess learning
  CL_samp_dx = 20;              // sampling of the position for the CL initial guess learning
  CL_samp_dy = 20;              // sampling of the position for the CL initial guess learning
  CL_samp_dth = 12;             // sampling of the heading for the CL initial guess learning

  WC_samp_t = 9;                // sampling of the time for the WC initial guess learning
  WC_samp_vi = 9;               // sampling of the initial speed for the WC initial guess learning
  WC_samp_vg = 9;               // sampling of the final speed for the WC initial guess learning
  WC_samp_dx = 20;              // sampling of the position for the WC initial guess learning
  WC_samp_dy = 20;              // sampling of the position for the WC initial guess learning
  WC_samp_dth = 12;             // sampling of the heading for the WC initial guess learning

  CL_samp_t_target.connect(CL_samp_t);
  CL_samp_vi_target.connect(CL_samp_vi);
  CL_samp_vg_target.connect(CL_samp_vg);
  CL_samp_dx_target.connect(CL_samp_dx);
  CL_samp_dy_target.connect(CL_samp_dy);
  CL_samp_dth_target.connect(CL_samp_dth);

  WC_samp_t_target.connect(WC_samp_t);
  WC_samp_vi_target.connect(WC_samp_vi);
  WC_samp_vg_target.connect(WC_samp_vg);
  WC_samp_dx_target.connect(WC_samp_dx);
  WC_samp_dy_target.connect(WC_samp_dy);
  WC_samp_dth_target.connect(WC_samp_dth);

  // final state constraints
  CL_XYf_fixed = true;             // true if the final position of the car-like is 
  CL_THf_fixed = true;             // true if the final heading of the car-like is fixed
  CL_PHf_fixed = true;             // true if the final steering angle of the car-like is fixed
  CL_Vf_fixed = true;              // true if the final velocity of the car-like is fixed

  WC_XYf_fixed = true;             // true if the final position of the wheelchair is fixed
  WC_THf_fixed = true;             // true if the final heading of the wheelchair is fixed
  WC_OMf_fixed = true;             // true if the final angular velocity of the wheelchair is fixed
  WC_Vf_fixed = true;              // true if the final linear velocity of the wheelchair is fixed

  SS_XYf_fixed = true;             // true if the final position of the spaceship is fixed
  SS_THf_fixed = true;             // true if the final heading of the spaceship is fixed
  SS_OMf_fixed = true;             // true if the final angular velocity of the spaceship is fixed
  SS_Vf_fixed = true;              // true if the final linear velocity of the spaceship is fixed


  /***************************************************************************************************************
   ********************************** Initialization from configuration files ************************************
   ***************************************************************************************************************/

  // init the gui main parameters
  this->initTrajGenIfcFromFile("../cfg/TrajGenIfc.cfg");

  // init the wheelchair system
  this->initWCSysFromFile("../cfg/wheelchair.cfg");

  // init the car-like system
  this->initCLSysFromFile("../cfg/carlike.cfg");
  
  // init the spaceship system
  this->initSSSysFromFile("../cfg/spaceship.cfg");


  /***************************************************************************************************************
   ************************************************ main frame ***************************************************
   ***************************************************************************************************************/

  contents=new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  
  /***************************************************************************************************************
   *********************************************** viewer frame **************************************************
   ***************************************************************************************************************/

  // LEFT pane to contain the canvas
  imageFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the canvas
  new FXLabel(imageFrame,"Viewer",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(imageFrame,SEPARATOR_GROOVE|LAYOUT_FILL_X);

  // Make image widget
  imageview=new FXImageView(imageFrame,NULL,this,ID_IMAGE,LAYOUT_FILL_X|LAYOUT_FILL_Y);

  msgBox = new FXLabel(imageFrame,"No map loaded. Click on \"Load map\" to load a new map.",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  /***************************************************************************************************************
   ********************************************** commands frame *************************************************
   ***************************************************************************************************************/

  // RIGHT pane for the buttons
  buttonFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the buttons
  new FXLabel(buttonFrame,"Commands",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Tab book with switchable panels
  FXTabBook* panels=new FXTabBook(buttonFrame,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);


  /***************************************************************************************************************
   ************************************ car-like system configuration tab ****************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Car-like");

  FXVerticalFrame *tabSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXRadioButton(tabSysCfg,"Car-like system configuration :",&optSys_target,FXDataTarget::ID_OPTION + CARLIKE_SYS, ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
//   new FXRadioButton(tabSysCfg,"Car-like system config 2 :",&optSys_target,FXDataTarget::ID_OPTION+4,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matSysCfg=new FXMatrix(tabSysCfg,3,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSysCfg,"");
  new FXLabel(matSysCfg,"Scale :");
  new FXTextField(matSysCfg,6,&rb_Scale_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  new FXLabel(matSysCfg,"");
  new FXLabel(matSysCfg,"Robot Length:");
  new FXTextField(matSysCfg,6,&rb_L_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXCheckButton *velCstBtn = new FXCheckButton(matSysCfg,"",this,ID_VEL_CST);
  velCstBtn->setCheck(useVelCst,false);
  new FXLabel(matSysCfg,"Max velocity:");
  new FXTextField(matSysCfg,6,&rb_VMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXCheckButton *accCstBtn = new FXCheckButton(matSysCfg,"",this,ID_ACC_CST);
  accCstBtn->setCheck(useAccCst,false);
  new FXLabel(matSysCfg,"Max acceleration:");
  new FXTextField(matSysCfg,6,&rb_AMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXCheckButton *avelCstBtn = new FXCheckButton(matSysCfg,"",this,ID_AVEL_CST);
  avelCstBtn->setCheck(useAVelCst,false);
  new FXLabel(matSysCfg,"Max steering angle:");
  new FXTextField(matSysCfg,6,&rb_KMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXCheckButton *aaccCstBtn = new FXCheckButton(matSysCfg,"",this,ID_AACC_CST);
  aaccCstBtn->setCheck(useAAccCst,false);
  new FXLabel(matSysCfg,"Max steering speed:");
  new FXTextField(matSysCfg,6,&rb_OMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXCheckButton *canMoveBackCLBtn = new FXCheckButton(tabSysCfg,"Can move backward",this,ID_CMBACK_CL);
  canMoveBackCLBtn->setCheck(canMoveBackCL,false);

  // Horizontal divider line
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame* frameState = new FXHorizontalFrame(tabSysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matInitState=new FXMatrix(frameState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(matInitState,"Init configuration :");
  FXMatrix *matInitCfg=new FXMatrix(matInitState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitCfg,"X1 :");
  new FXTextField(matInitCfg,9,&x1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"Y1 :");
  new FXTextField(matInitCfg,9,&y1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"TH1 :");
  new FXTextField(matInitCfg,9,&theta1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"PH1 :");
  new FXTextField(matInitCfg,9,&kp1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"V1 :");
  new FXTextField(matInitCfg,9,&v1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"T1 :");
  new FXTextField(matInitCfg,9,&t1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   // Horizontal divider line
//   new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matGoalState=new FXMatrix(frameState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(matGoalState,"Goal configuration :");
  FXMatrix *matGoalCfg=new FXMatrix(matGoalState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matGoalCfg,"X2 :");
  FXCheckButton *cl_xf_chb = new FXCheckButton(matGoalCfg,"X2 :",this,ID_CL_XYF_FIXED);
  cl_xf_chb->setCheck(true,false);  
  new FXTextField(matGoalCfg,9,&x2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfg,"Y2 :");
//   FXCheckButton *cl_yf_chb = new FXCheckButton(matGoalCfg,"Y2 :",this,ID_CL_XYF_FIXED);
//   cl_yf_chb->setCheck(true,false);  
  new FXTextField(matGoalCfg,9,&y2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfg,"Th2 :");
  FXCheckButton *cl_thf_chb = new FXCheckButton(matGoalCfg,"TH2 :",this,ID_CL_THF_FIXED);
  cl_thf_chb->setCheck(true,false);  
  new FXTextField(matGoalCfg,9,&theta2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfg,"Kp2 :");
  FXCheckButton *cl_phf_chb = new FXCheckButton(matGoalCfg,"PH2 :",this,ID_CL_PHF_FIXED);
  cl_phf_chb->setCheck(true,false);  
  new FXTextField(matGoalCfg,9,&kp2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfg,"V2 :");
  FXCheckButton *cl_vf_chb = new FXCheckButton(matGoalCfg,"V2 :",this,ID_CL_VF_FIXED);
  cl_vf_chb->setCheck(true,false);  
  new FXTextField(matGoalCfg,9,&v2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfg,"T2 :");
  new FXTextField(matGoalCfg,9,&t2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabSysCfg,"IG learning control sampling:");
  FXMatrix *matCL_IGCS=new FXMatrix(tabSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matCL_IGCS,"Acceleration:");
  new FXTextField(matCL_IGCS,4,&CL_sampAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCL_IGCS,"Steer Speed:");
  new FXTextField(matCL_IGCS,4,&CL_sampStSpd_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabSysCfg,"Maneuvers Computation:");

  // Button to compute the reachable star
  FXCheckButton *curvcirc_chb = new FXCheckButton(tabSysCfg,"display min curvature circles",this,ID_DISP_CURVCIRC);
  curvcirc_chb->setCheck(dispCurvCirc,false);  
  new FXButton(tabSysCfg,"Compute &Maneuvers",NULL,this,ID_CPT_MNVER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  


  /***************************************************************************************************************
   ************************************ spaceship system configuration tab ****************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Spaceship");

  FXVerticalFrame *tabSSSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXRadioButton(tabSSSysCfg,"Space ship system configuration :",&optSys_target,FXDataTarget::ID_OPTION + SPACESHIP_SYS, ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
//   new FXRadioButton(tabSSSysCfg,"Space ship system config 2 :",&optSys_target,FXDataTarget::ID_OPTION+8,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matSSSysCfg=new FXMatrix(tabSSSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSSSysCfg,"Max velocity:");
  new FXTextField(matSSSysCfg,6,&SS_VelMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSSSysCfg,"Max acceleration:");
  new FXTextField(matSSSysCfg,6,&SS_AccMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSSSysCfg,"Max angular vel:");
  new FXTextField(matSSSysCfg,6,&SS_AVelMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSSSysCfg,"Max angular acc:");
  new FXTextField(matSSSysCfg,6,&SS_AAccMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  // Horizontal divider line
  new FXHorizontalSeparator(tabSSSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame* frameSSState = new FXHorizontalFrame(tabSSSysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matInitSSState=new FXMatrix(frameSSState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(matInitSSState,"Init configuration :");
  FXMatrix *matInitCfgSS=new FXMatrix(matInitSSState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitCfgSS,"X1 :");
  new FXTextField(matInitCfgSS,9,&x1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgSS,"Y1 :");
  new FXTextField(matInitCfgSS,9,&y1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgSS,"TH1 :");
  new FXTextField(matInitCfgSS,9,&theta1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgSS,"OM1 :");
  new FXTextField(matInitCfgSS,9,&om1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgSS,"VX1 :");
  new FXTextField(matInitCfgSS,9,&vx1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgSS,"VY1 :");
  new FXTextField(matInitCfgSS,9,&vy1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgSS,"T1 :");
  new FXTextField(matInitCfgSS,9,&t1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   // Horizontal divider line
//   new FXHorizontalSeparator(tabSSSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matGoalSSState=new FXMatrix(frameSSState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(matGoalSSState,"Goal configuration :");
  FXMatrix *matGoalCfgSS=new FXMatrix(matGoalSSState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matGoalCfgSS,"X2 :");
  FXCheckButton *ss_xf_chb = new FXCheckButton(matGoalCfgSS,"X2 :",this,ID_SS_XYF_FIXED);
  ss_xf_chb->setCheck(true,false);  
  new FXTextField(matGoalCfgSS,9,&x2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfgSS,"Y2 :");
  new FXTextField(matGoalCfgSS,9,&y2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgSS,"Th2 :")
  FXCheckButton *ss_thf_chb = new FXCheckButton(matGoalCfgSS,"TH2 :",this,ID_SS_THF_FIXED);
  ss_thf_chb->setCheck(true,false);
  new FXTextField(matGoalCfgSS,9,&theta2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgSS,"om2 :");
  FXCheckButton *ss_omf_chb = new FXCheckButton(matGoalCfgSS,"OM2 :",this,ID_SS_OMF_FIXED);
  ss_omf_chb->setCheck(true,false);
  new FXTextField(matGoalCfgSS,9,&om2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgSS,"VX2 :");
  FXCheckButton *ss_vf_chb = new FXCheckButton(matGoalCfgSS,"VX2 :",this,ID_SS_VF_FIXED);
  ss_vf_chb->setCheck(true,false);
  new FXTextField(matGoalCfgSS,9,&vx2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfgSS,"VY2 :");
  new FXTextField(matGoalCfgSS,9,&vy2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfgSS,"T2 :");
  new FXTextField(matGoalCfgSS,9,&t2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabSSSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabSSSysCfg,"IG learning control sampling:");
  FXMatrix *matSS_IGCS=new FXMatrix(tabSSSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSS_IGCS,"Linear Acceleration:");
  new FXTextField(matSS_IGCS,4,&SS_sampAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSS_IGCS,"Angular Acceleration:");
  new FXTextField(matSS_IGCS,4,&SS_sampAAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


  /***************************************************************************************************************
   ******************************* differential drive  system configuration tab **********************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"DiffDrive");

  FXVerticalFrame *tabDDSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXHorizontalFrame* frameDDSysCfg = new FXHorizontalFrame(tabDDSysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

//   FXVerticalFrame *frameDDSysCfg1=new FXVerticalFrame(frameDDSysCfg,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   new FXRadioButton(frameDDSysCfg1,"Diff drive sys config :",&optSys_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

//   FXMatrix *matDDSysCfg=new FXMatrix(frameDDSysCfg1,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matDDSysCfg,"Max wheel vel:");
//   new FXTextField(matDDSysCfg,6,&dd_WVMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXLabel(matDDSysCfg,"Max wheel acc:");
//   new FXTextField(matDDSysCfg,6,&dd_WAMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXLabel(matDDSysCfg,"Radius:");
//   new FXTextField(matDDSysCfg,6,&dd_RAD_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXVerticalFrame *frameDDSysCfg2=new FXVerticalFrame(frameDDSysCfg,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);

  new FXRadioButton(frameDDSysCfg2,"Wheelchair sys config :",&optSys_target,FXDataTarget::ID_OPTION + WHEELCHAIR_SYS, ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
//   new FXRadioButton(frameDDSysCfg2,"Wheelchair sys config2 :",&optSys_target,FXDataTarget::ID_OPTION+5,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(frameDDSysCfg2,"MANA sys config :",&optSys_target,FXDataTarget::ID_OPTION + MANA_SYS, ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matDDSysCfg2=new FXMatrix(frameDDSysCfg2,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *accCstBtnWC = new FXCheckButton(matDDSysCfg2,"Max acc",this,ID_ACC_WC_CST);
  accCstBtnWC->setCheck(useAccCstWC,false);
  new FXTextField(matDDSysCfg2,6,&WC_AccMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  FXCheckButton *aaccCstBtnWC = new FXCheckButton(matDDSysCfg2,"Max aacc",this,ID_AACC_WC_CST);
  aaccCstBtnWC->setCheck(useAAccCstWC,false);
  new FXTextField(matDDSysCfg2,6,&WC_AAccMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  FXCheckButton *velCstBtnWC = new FXCheckButton(matDDSysCfg2,"Max vel",this,ID_VEL_WC_CST);
  velCstBtnWC->setCheck(useVelCstWC,false);
  new FXTextField(matDDSysCfg2,6,&WC_VelMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  FXCheckButton *avelCstBtnWC = new FXCheckButton(matDDSysCfg2,"Max avel",this,ID_AVEL_WC_CST);
  avelCstBtnWC->setCheck(useAVelCstWC,false);
  new FXTextField(matDDSysCfg2,6,&WC_AVelMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXCheckButton *canMoveBackWCBtn = new FXCheckButton(frameDDSysCfg2,"Can move backward",this,ID_CMBACK_WC);
  canMoveBackWCBtn->setCheck(canMoveBackWC,false);

//   new FXLabel(matDDSysCfg2,"Max acc:");
//   new FXTextField(matDDSysCfg2,6,&WC_AccMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matDDSysCfg2,"Max aacc:");
//   new FXTextField(matDDSysCfg2,6,&WC_AAccMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matDDSysCfg2,"Max vel:");
//   new FXTextField(matDDSysCfg2,6,&WC_VelMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matDDSysCfg2,"Max avel:");
//   new FXTextField(matDDSysCfg2,6,&WC_AVelMax_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  // Horizontal divider line
  new FXHorizontalSeparator(tabDDSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame* frameDDState = new FXHorizontalFrame(tabDDSysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matInitDDState=new FXMatrix(frameDDState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(matInitDDState,"Init configuration :");
  FXMatrix *matInitCfgDD=new FXMatrix(matInitDDState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitCfgDD,"X1 :");
  new FXTextField(matInitCfgDD,9,&x1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgDD,"Y1 :");
  new FXTextField(matInitCfgDD,9,&y1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgDD,"TH1 :");
  new FXTextField(matInitCfgDD,9,&theta1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgDD,"OM1 :");
  new FXTextField(matInitCfgDD,9,&om1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgDD,"V1 :");
  new FXTextField(matInitCfgDD,9,&v1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfgDD,"T1 :");
  new FXTextField(matInitCfgDD,9,&t1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   // Horizontal divider line
//   new FXHorizontalSeparator(tabSSSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matGoalDDState=new FXMatrix(frameDDState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(matGoalDDState,"Goal configuration :");
  FXMatrix *matGoalCfgDD=new FXMatrix(matGoalDDState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matGoalCfgDD,"X2 :");
  FXCheckButton *dd_xf_chb = new FXCheckButton(matGoalCfgDD,"X2 :",this,ID_DD_XYF_FIXED);
  dd_xf_chb->setCheck(true,false);  
  new FXTextField(matGoalCfgDD,9,&x2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfgDD,"Y2 :");
  new FXTextField(matGoalCfgDD,9,&y2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgDD,"TH2 :");
  FXCheckButton *dd_thf_chb = new FXCheckButton(matGoalCfgDD,"TH2 :",this,ID_DD_THF_FIXED);
  dd_thf_chb->setCheck(true,false);
  new FXTextField(matGoalCfgDD,9,&theta2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgDD,"OM2 :");
  FXCheckButton *dd_omf_chb = new FXCheckButton(matGoalCfgDD,"OM2 :",this,ID_DD_OMF_FIXED);
  dd_omf_chb->setCheck(true,false);
  new FXTextField(matGoalCfgDD,9,&om2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgDD,"V2 :");
  FXCheckButton *dd_vf_chb = new FXCheckButton(matGoalCfgDD,"V2 :",this,ID_DD_VF_FIXED);
  dd_vf_chb->setCheck(true,false);
  new FXTextField(matGoalCfgDD,9,&v2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfgDD,"T2 :");
  new FXTextField(matGoalCfgDD,9,&t2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabSSSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabDDSysCfg,"IG learning control sampling:");
  FXMatrix *matWC_IGCS=new FXMatrix(tabDDSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matWC_IGCS,"Linear Acceleration:");
  new FXTextField(matWC_IGCS,4,&WC_sampAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matWC_IGCS,"Angular Acceleration:");
  new FXTextField(matWC_IGCS,4,&WC_sampAAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


//   /***************************************************************************************************************
//    ************************ Manipulator arm (Simplified PUMA) system configuration tab ***************************
//    ***************************************************************************************************************/

//   new FXTabItem(panels,"S. PUMA");

//   FXVerticalFrame *tabMA1SysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   new FXRadioButton(tabMA1SysCfg,"Manip Arm system configuration :",&optSys_target,FXDataTarget::ID_OPTION+7,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

//   FXMatrix *matMA1SysCfg=new FXMatrix(tabMA1SysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matMA1SysCfg,"Max avel 1:");
//   new FXTextField(matMA1SysCfg,6,&MA_AVel1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matMA1SysCfg,"Max avel 2:");
//   new FXTextField(matMA1SysCfg,6,&MA_AVel2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matMA1SysCfg,"Max avel 3:");
//   new FXTextField(matMA1SysCfg,6,&MA_AVel3_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXLabel(matMA1SysCfg,"Max aacc 1:");
//   new FXTextField(matMA1SysCfg,6,&MA_AAcc1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matMA1SysCfg,"Max aacc 2:");
//   new FXTextField(matMA1SysCfg,6,&MA_AAcc2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matMA1SysCfg,"Max aacc 3:");
//   new FXTextField(matMA1SysCfg,6,&MA_AAcc3_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


//   // Horizontal divider line
//   new FXHorizontalSeparator(tabMA1SysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

//   FXHorizontalFrame* frameMA1State = new FXHorizontalFrame(tabMA1SysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

//   FXMatrix *matInitMA1State=new FXMatrix(frameMA1State,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   new FXLabel(matInitMA1State,"Init configuration :");
//   FXMatrix *matInitCfgMA1=new FXMatrix(matInitMA1State,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matInitCfgMA1,"TH1 :");
//   new FXTextField(matInitCfgMA1,9,&MA_TH1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgMA1,"TH2 :");
//   new FXTextField(matInitCfgMA1,9,&MA_TH2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgMA1,"TH3 :");
//   new FXTextField(matInitCfgMA1,9,&MA_TH3_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXLabel(matInitCfgMA1,"OM1 :");
//   new FXTextField(matInitCfgMA1,9,&MA_OM1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgMA1,"OM2 :");
//   new FXTextField(matInitCfgMA1,9,&MA_OM2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgMA1,"OM3 :");
//   new FXTextField(matInitCfgMA1,9,&MA_OM3_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgMA1,"T1 :");
//   new FXTextField(matInitCfgMA1,9,&t1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   FXMatrix *matGoalMA1State=new FXMatrix(frameMA1State,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   new FXLabel(matGoalMA1State,"Goal configuration :");
//   FXMatrix *matGoalCfgMA1=new FXMatrix(matGoalMA1State,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matGoalCfgMA1,"X2 :");
//   new FXTextField(matGoalCfgMA1,9,&x2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgMA1,"Y2 :");
//   new FXTextField(matGoalCfgMA1,9,&y2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgMA1,"Z2 :");
//   new FXTextField(matGoalCfgMA1,9,&z2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

// //   new FXLabel(matGoalCfgMA1,"VX2 :");
// //   new FXTextField(matGoalCfgMA1,9,&vx2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
// //   new FXLabel(matGoalCfgMA1,"VY2 :");
// //   new FXTextField(matGoalCfgMA1,9,&vy2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
// //   new FXLabel(matGoalCfgMA1,"VZ2 :");
// //   new FXTextField(matGoalCfgMA1,9,&vz2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgMA1,"T2 :");
//   new FXTextField(matGoalCfgMA1,9,&t2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


//   /***************************************************************************************************************
//    ********************************** point mass 3D system configuration tab *************************************
//    ***************************************************************************************************************/

//   new FXTabItem(panels,"PM3D");

//   FXVerticalFrame *tabPMSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   new FXRadioButton(tabPMSysCfg,"Point Mass 3D system configuration :",&optSys_target,FXDataTarget::ID_OPTION+6,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

//   FXMatrix *matPMSysCfg=new FXMatrix(tabPMSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matPMSysCfg,"Max vel X:");
//   new FXTextField(matPMSysCfg,6,&PM3D_VX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matPMSysCfg,"Max vel Y:");
//   new FXTextField(matPMSysCfg,6,&PM3D_VY_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matPMSysCfg,"Max vel Z:");
//   new FXTextField(matPMSysCfg,6,&PM3D_VZ_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXLabel(matPMSysCfg,"Max acc X:");
//   new FXTextField(matPMSysCfg,6,&PM3D_AX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matPMSysCfg,"Max acc Y:");
//   new FXTextField(matPMSysCfg,6,&PM3D_AY_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matPMSysCfg,"Max acc Z:");
//   new FXTextField(matPMSysCfg,6,&PM3D_AZ_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


//   // Horizontal divider line
//   new FXHorizontalSeparator(tabPMSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

//   FXHorizontalFrame* framePMState = new FXHorizontalFrame(tabPMSysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

//   FXMatrix *matInitPMState=new FXMatrix(framePMState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   new FXLabel(matInitPMState,"Init configuration :");
//   FXMatrix *matInitCfgPM=new FXMatrix(matInitPMState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matInitCfgPM,"X1 :");
//   new FXTextField(matInitCfgPM,9,&x1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgPM,"Y1 :");
//   new FXTextField(matInitCfgPM,9,&y1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgPM,"Z1 :");
//   new FXTextField(matInitCfgPM,9,&z1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXLabel(matInitCfgPM,"VX1 :");
//   new FXTextField(matInitCfgPM,9,&vx1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgPM,"VY1 :");
//   new FXTextField(matInitCfgPM,9,&vy1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgPM,"VZ1 :");
//   new FXTextField(matInitCfgPM,9,&vz1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matInitCfgPM,"T1 :");
//   new FXTextField(matInitCfgPM,9,&t1_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   FXMatrix *matGoalPMState=new FXMatrix(framePMState,1,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   new FXLabel(matGoalPMState,"Goal configuration :");
//   FXMatrix *matGoalCfgPM=new FXMatrix(matGoalPMState,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matGoalCfgPM,"X2 :");
//   new FXTextField(matGoalCfgPM,9,&x2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgPM,"Y2 :");
//   new FXTextField(matGoalCfgPM,9,&y2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgPM,"Z2 :");
//   new FXTextField(matGoalCfgPM,9,&z2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXLabel(matGoalCfgPM,"VX2 :");
//   new FXTextField(matGoalCfgPM,9,&vx2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgPM,"VY2 :");
//   new FXTextField(matGoalCfgPM,9,&vy2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgPM,"VZ2 :");
//   new FXTextField(matGoalCfgPM,9,&vz2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalCfgPM,"T2 :");
//   new FXTextField(matGoalCfgPM,9,&t2_3D_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


  /***************************************************************************************************************
   *********************************************** options tab ***************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Options");

  FXVerticalFrame *tabOptions=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matOptPrec=new FXMatrix(tabOptions,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptPrec,"Speed / Precision :");
  new FXLabel(matOptPrec,"");
  new FXLabel(matOptPrec,"Max number of steps:");
  new FXTextField(matOptPrec,6,&maxNbStep_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptPrec,"Max residual:");
  new FXTextField(matOptPrec,6,&maxResidual_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  FXMatrix *matOptDbg=new FXMatrix(tabOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptDbg,"Debugging mode :");
  new FXRadioButton(matOptDbg,"0 (no debug)",&debugLevel_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"1 (store trajectories)",&debugLevel_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"2 (disp final profiles)",&debugLevel_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"3 (disp itd steps)",&debugLevel_target,FXDataTarget::ID_OPTION+3,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matTimeCsttStatus=new FXMatrix(tabOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTimeCsttStatus,"Time constraint:");
  new FXRadioButton(matTimeCsttStatus,"Fixed",&timeCsttStatus_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matTimeCsttStatus,"Free",&timeCsttStatus_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matTimeCsttStatus,"Constrained",&timeCsttStatus_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  FXMatrix *matTimeCstt=new FXMatrix(matTimeCsttStatus,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTimeCstt,"Min time:");
  new FXTextField(matTimeCstt,6,&minTime_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTimeCstt,"Max time:");
  new FXTextField(matTimeCstt,6,&maxTime_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matTestReachty=new FXMatrix(tabOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTestReachty,"Reachability test:");
  new FXRadioButton(matTestReachty,"Never",&testReachability_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matTestReachty,"When net reached",&testReachability_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matTestReachty,"Always",&testReachability_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matUseIG=new FXMatrix(tabOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  FXCheckButton *useIGBtn = new FXCheckButton(matUseIG,"Use IG database",this,ID_USE_IG);
  useIGBtn->setCheck(useIG,false);
  FXCheckButton *useFDPDBtn = new FXCheckButton(matUseIG,"Use PD part. derivatives",this,ID_USE_FDPD);
  useIGBtn->setCheck(useFDPD,false);


  /***************************************************************************************************************
   *********************************************** Commands tab **************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Commands");

  FXVerticalFrame *tabCommands=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);  

  /***************************************************************************************************************
   ******************************************* Reachable Star subtab *********************************************
   ***************************************************************************************************************/

//   new FXTabItem(panels,"RStar");

  FXVerticalFrame *tabRStar=new FXVerticalFrame(tabCommands,FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   FXVerticalFrame *tabRStar=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabRStar,"RStar Computation:");

  new FXHorizontalSeparator(tabRStar,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matParamRStar=new FXMatrix(tabRStar,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matParamRStar,"Nb branches:");
  new FXTextField(matParamRStar,6,&nbBranches_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Min angle:");
  new FXTextField(matParamRStar,6,&minAngRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Max angle:");
  new FXTextField(matParamRStar,6,&maxAngRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Min distance:");
  new FXTextField(matParamRStar,6,&minDistRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Max distance:");
  new FXTextField(matParamRStar,6,&maxDistRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Nb itv dist:");
  new FXTextField(matParamRStar,6,&nbItvDistRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Traj by branch:");
  new FXTextField(matParamRStar,6,&trajByBranch_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Diff angle:");
  new FXTextField(matParamRStar,6,&diffAngle_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);  

  FXMatrix *matSaveRStar=new FXMatrix(tabRStar,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  FXCheckButton *saveRStarBtn = new FXCheckButton(matSaveRStar,"Save RStar Trajectories",this,ID_SAVE_RSTAR);
  saveRStarBtn->setCheck(saveRStar,false);
  FXMatrix *matSaveRStarFilename=new FXMatrix(matSaveRStar,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXTextField(matSaveRStarFilename,35,&IG1stParamsName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);
  new FXButton(matSaveRStarFilename,"...",NULL,this,ID_SAVE_RSTAR_FILENAME,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  

  // Button to compute the reachable star
  new FXButton(tabRStar,"Compute &Reachable Star",NULL,this,ID_DRAW_RSTAR,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  

  /***************************************************************************************************************
   ******************************************** Sampling trees subtab ********************************************
   ***************************************************************************************************************/

  FXVerticalFrame *tabLattice=new FXVerticalFrame(tabCommands,FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   FXVerticalFrame *tabLattice=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabLattice,"Sampling trees Params:");

  new FXHorizontalSeparator(tabLattice,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame* frameOptLattice = new FXHorizontalFrame(tabLattice,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matParamLattice=new FXMatrix(frameOptLattice,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matParamLattice,"Tree height:");
  new FXTextField(matParamLattice,6,&SSL_height_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamLattice,"Nb ctrl sampling:");
  new FXTextField(matParamLattice,6,&SSL_nbCS_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matDispLattice=new FXMatrix(frameOptLattice,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matDispLattice,"Traj sampling:");
  new FXTextField(matDispLattice,6,&SSL_trajSp_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  FXCheckButton *dispLatticeBtn = new FXCheckButton(matDispLattice,"Display lattice",this,ID_DISP_LATTICE_CB);
  dispLatticeBtn->setCheck(SSL_dispTraj,false);

  // Button to compute the reachable star
  new FXButton(tabLattice,"Compute &Sampling Tree",NULL,this,ID_DRAW_SSLATTICE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(tabLattice,"D&raw Gnuplot STree",NULL,this,ID_DRAW_GPLATTICE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(tabLattice,"&Gen Traj to STree Nodes",NULL,this,ID_GENTRAJ_SSLNODES,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  /***************************************************************************************************************
   *********************************************** Drawing subtab ************************************************
   ***************************************************************************************************************/

  FXVerticalFrame *tabDrawing=new FXVerticalFrame(tabCommands,FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   FXVerticalFrame *tabLattice=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *drawUncompTrajBtn = new FXCheckButton(tabDrawing,"draw uncomplete traj.",this,ID_DRAW_UNCOMP);
  drawUncompTrajBtn->setCheck(drawUncompleteTraj,false);
  FXCheckButton *drawTgtPosBtn = new FXCheckButton(tabDrawing,"draw target position",this,ID_DRAW_TGTPOS);
  drawTgtPosBtn->setCheck(drawTargetPos,false);

  // Button to display the reachable star in a gnuplot window
  new FXButton(tabDrawing,"&Draw Gnuplot2D Star",NULL,this,ID_DRAW_GP2D,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to empty the table of trajectory displayed
  new FXButton(tabDrawing,"&Empty Traj Tab",NULL,this,ID_EMPTY_VECTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ********************************************* IG Learning subtab **********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"IG");

  FXVerticalFrame *tabIG=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);  

  FXVerticalFrame *tabIGLearn=new FXVerticalFrame(tabIG,FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabIGLearn,"IG Learning Params:");

  new FXHorizontalSeparator(tabIGLearn,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matIGFirstParams=new FXMatrix(tabIGLearn,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matIGFirstParams,"First parameters initialization");
  new FXRadioButton(matIGFirstParams,"Sampling over the control space",&IG_1stParams_target,FXDataTarget::ID_OPTION+SC_FIRSTIG,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matIGFirstParams,"Load from file",&IG_1stParams_target,FXDataTarget::ID_OPTION+LOADFROMFILE_FIRSTIG,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matFIGfilename=new FXMatrix(matIGFirstParams,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXTextField(matFIGfilename,35,&IG1stParamsName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);
  new FXButton(matFIGfilename,"...",NULL,this,ID_IG_1STPARAM,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matIGLearnOpt=new FXMatrix(tabIGLearn,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  FXMatrix *matIGL_Time=new FXMatrix(matIGLearnOpt,4,MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matIGL_Time,"t_min:");
  new FXTextField(matIGL_Time,4,&IG_tmin_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matIGL_Time,"t_max:");
  new FXTextField(matIGL_Time,4,&IG_tmax_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matIGL_NNTSR=new FXMatrix(matIGLearnOpt,2,MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matIGL_NNTSR,"Stop when nb st reached <");
  new FXTextField(matIGL_NNTSR,4,&IG_minNNTSR_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matIGL_Debug=new FXMatrix(matIGLearnOpt,4,MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matIGL_Debug,"Debug level:");
  new FXRadioButton(matIGL_Debug,"0",&IG_debugMode_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matIGL_Debug,"1",&IG_debugMode_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matIGL_Debug,"2",&IG_debugMode_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);  

  new FXButton(tabIGLearn,"&Learn Initial Guess",NULL,this,ID_LEARN_IG,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matLoadIG=new FXMatrix(tabIGLearn,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);  
  new FXTextField(matLoadIG,35,&IGfilename_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);
  new FXButton(matLoadIG,"...",NULL,this,ID_IG_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load a map from a file
  new FXButton(tabIGLearn,"&Load IG Database",NULL,this,ID_LOAD_IG,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load a map from a file
  new FXButton(tabIGLearn,"&Display IG Database",NULL,this,ID_DISP_IG,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ********************************************* general commands ************************************************
   ***************************************************************************************************************/

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matGenTraj=new FXMatrix(buttonFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);  
  new FXRadioButton(matGenTraj,"ComputeTraj",&genTrajMethod_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matGenTraj,"genTraj",&genTrajMethod_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  // Button to plan the path
  new FXButton(buttonFrame,"&Generate the Trajectory",NULL,this,ID_GEN_TRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matNameMap=new FXMatrix(buttonFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);  
  new FXTextField(matNameMap,35,&filename_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);
  new FXButton(matNameMap,"...",NULL,this,ID_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load a map from a file
  new FXButton(buttonFrame,"&Load Map from file",NULL,this,ID_LOAD_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Exit button
  new FXButton(buttonFrame,"&Exit",NULL,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
}

// destructor
TrajGenIfc::~TrajGenIfc()
{
  // the robotic system handled
  if(sys_ != NULL) delete sys_;

  // the trajectory generator
  if(trajGenerator != NULL) delete trajGenerator;

  // the state space lattice
  if(lattice_ != NULL) delete lattice_;

  // initial guess of parameters database and learning module
  if(ig_ != NULL) delete ig_;

  // initial and final state-times
  if(initST_ != NULL) delete initST_;
  if(goalST_ != NULL) delete goalST_;

  // the map used
  if(map != NULL) delete map;

  if(tabPosTrajTmpX != NULL) delete tabPosTrajTmpX;
  if(tabPosTrajTmpY != NULL) delete tabPosTrajTmpY;
  if(tabPosTrajTmpZ != NULL) delete tabPosTrajTmpZ;
  if(tabPosTrajTmpT != NULL) delete tabPosTrajTmpT;
  if(tabPosTrajTmpV != NULL) delete tabPosTrajTmpV;
  if(tabPosTrajTmpVX != NULL) delete tabPosTrajTmpVX;
  if(tabPosTrajTmpVY != NULL) delete tabPosTrajTmpVY;
  if(tabPosTrajTmpVZ != NULL) delete tabPosTrajTmpVZ;
}

// Create and initialize
void TrajGenIfc::create(){

  // Create the windows
  FXMainWindow::create();

  // Make the main window appear
  show(PLACEMENT_SCREEN);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long TrajGenIfc::loadMapFromFile(FXObject*,FXSelector,void*)
{
  cout << "Load Map from file" << endl;
//   FXFileDialog open(this,"Load map (PGM Image)");
//   open.setFilename(filename);
//   if(open.execute()){
//     filename=open.getFilename();
  bool loaded = loadimage(filename);
  if(loaded)
    {
      msgBox->setText("A map has been Loaded. Click left to select init configuration and click right to select target configuration");

      // reinit vectors to stock the computed trajectories
      this->emptyTrajVec();
    }
  else
    FXMessageBox::error(this,MBOX_OK,"Error while Loading the Map","The image filename doesn't exist.");
//   }
  return 1;
}

long TrajGenIfc::changeMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Load map (PGM Image)");
  open.setFilename(filename);
  if(open.execute()){
    filename=open.getFilename();
  }
  return 1;
}

long TrajGenIfc::genTraj(FXObject*,FXSelector,void*)
{
  cout << "///////////////////////////////////////////////////////////////////////////// " << endl;
  cout << "///////////////////////////////////////////////////////////////////////////// " << endl;
  cout << "///////////////////////////////////////////////////////////////////////////// " << endl;

  if(map == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","No map has been loaded");
      return -1;
    }

//   vector<double>* subTabPosTrajX;
//   vector<double>* subTabPosTrajY;
//   vector<double>* subTabPosTrajT;
//   vector<double>* subTabPosTrajV;
  
  try 
    {
//       if(debugLevel >= 1)
// 	{
// 	  subTabPosTrajX = new vector<double>;
// 	  subTabPosTrajY = new vector<double>;
// 	  subTabPosTrajT = new vector<double>;
// 	  subTabPosTrajV = new vector<double>;	  
// 	}

      FXString reachStatus("");

      // init the timer
      struct timeval tv1,tv2;
      gettimeofday(&tv1,NULL);

      // init the robotic system
      this->initSys();

      // init the trajectory generator
      this->initTrajGen();

      // init the initial and final state-times
      this->initIST();
      this->initGST();

      cout << "initST_ : "; initST_->display(); cout << endl;
      cout << "goalST_ : "; goalST_->display(); cout << endl;

//       switch(optSys)
// 	{
// 	case CARLIKE_SYS : this->initCarLikeSys(vecState_x0,vecState_xf,vecConstraints); break;
// // 	case CARLIKE2_SYS : this->initCarLikeSys2(vecState_x0,vecState_xf,vecConstraints); break;
// // 	case SPACESHIP_SYS : this->initSpaceShipSys(vecState_x0,vecState_xf,vecConstraints); break;
// // 	case SPACESHIP2_SYS : this->initSpaceShipSys(vecState_x0,vecState_xf,vecConstraints); break;
// // 	case DIFFDRIVE_SYS : this->initDiffDriveSys(vecState_x0,vecState_xf,vecConstraints); break;
// // 	case WHEELCHAIR_SYS : this->initWheelChairSys(vecState_x0,vecState_xf,vecConstraints); break;
// // 	case WHEELCHAIR2_SYS : this->initWheelChairSys(vecState_x0,vecState_xf,vecConstraints); break;
// 	}
      
//       cout << "vecState_x0[0] : " << vecState_x0[0] << endl;
//       cout << "vecState_x0[1] : " << vecState_x0[1] << endl;
//       cout << "vecState_x0[2] : " << vecState_x0[2] << endl;
//       cout << "vecState_x0[3] : " << vecState_x0[3] << endl;
//       cout << "vecState_x0[4] : " << vecState_x0[4] << endl;
//       cout << "vecState_x0[5] : " << vecState_x0[5] << endl;
//       cout << "vecState_x0[6] : " << vecState_x0[6] << endl;

//       cout << "trajGenerator->xLength : " << trajGenerator->xLength << endl;
//       cout << "(" << vecState_x0[0] << "," << vecState_x0[1] << "," << vecState_x0[2] << "," << vecState_x0[3] << "," 
// 	   << vecState_x0[4] << "," << vecState_x0[5] << "," << vecState_x0[6] << ")" << endl;
//       cout << "(" << vecState_xf[0] << "," << vecState_xf[1] << "," << vecState_xf[2] << "," << vecState_xf[3] << "," 
// 	   << vecState_xf[4] << "," << vecState_xf[5] << "," << vecState_xf[6] << ")" << endl;

      // init the time constraint of the trajectory generator
      int resITC = initTimeCstt();
      if(resITC != 0) return -1;

      //Compute trajectory
      int res;
      if(genTrajMethod == 0) res = trajGenerator->ComputeTrajectory(*initST_,*goalST_,NULL);
      else res = trajGenerator->genTraj(*initST_,*goalST_);

      // stop the timer
      gettimeofday(&tv2,NULL);
      
      // compute the time passed
      int elapsedTs = tv2.tv_sec - tv1.tv_sec;
      int elapsedTus = tv2.tv_usec - tv1.tv_usec;
      double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
      cout << "Elapsed time : " << elapsedT << endl;

      if(res == 0)
	{
	  // case where the trajectory generation succeed
	  if(debugLevel >= 1)
	    {
	      // trace the trajectory on the map
	      trajGenerator->traceOn(map,300);
	      
	      // compute a sample trajectory to display in a gnuplot viewer
	      this->sampleTraj(trajGenerator);
	      this->stockSampleTraj();

	      // saving the map
	      map->save("/tmp/ResultMap_traj.pgm");
	      
	      // load the map in the canvas
	      this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));
	    }

	  if(testReachability == 2 && lattice_ != NULL)
	    {
// 	      // test the reachability of the final state
// 	      if(testReachty(*goalST_)) reachStatus = FXString(" (reachable)");
// 	      else reachStatus = FXString(" (unreachable)");
	      reachStatus = FXString(" ()");
	    }
	  
	  FXString str("Trajectory generation succeed in ");
	  str += FXStringVal(elapsedT,2,FALSE);
	  str += FXString(" sec after ");
	  str += FXStringVal((FXdouble)trajGenerator->finalNbSteps_,0,FALSE);
	  str += FXString(" steps.");
	  str += reachStatus;
	  msgBox->setText(str);
	}
      else if(res == -1)
	{
	  // case where the trajectory generation didn't reach the goal : 
	  // trace the better trajectory found during the generation

	  // load the better parameters found during generation
	  // ...

	  if(debugLevel >= 1)
	    {
	      // trace the trajectory on the map
	      trajGenerator->traceOn(map,300);
	      
	      // compute a sample trajectory to display in a gnuplot viewer
	      this->sampleTraj(trajGenerator);
	      this->stockSampleTraj();
	      
	      // saving the map
	      map->save("/tmp/ResultMap_traj.pgm");

	      // load the map in the canvas
	      this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));
	    }

	  if(testReachability >= 1 && lattice_ != NULL)
	    {
// 	      // test the reachability of the final state
// 	      if(testReachty(*goalST_)) reachStatus = FXString(" (reachable)");
// 	      else reachStatus = FXString(" (unreachable)");
	      reachStatus = FXString(" ()");
	    }
	  
	  FXString str("The goal has not been reach after ");
	  str += FXStringVal((FXdouble)CLTrajGen::getMaxNbStep(),0,FALSE);
	  str += FXString(" steps. Elapsed time : ");
	  str += FXStringVal(elapsedT,2,FALSE);
	  str += FXString(" sec.");
	  str += reachStatus;
	  msgBox->setText(str);
	}
      else // (res == -2)
	{
// 	  if(testReachability >= 1 && lattice_ != NULL)
// 	    {
// 	      // test the reachability of the final state
// 	      if(testReachty(*goalST_)) reachStatus = FXString(" (reachable)");
// 	      else reachStatus = FXString(" (unreachable)");
// 	    }
// 	  else
// 	    {
// 	      if(lattice_ == NULL) cout << "lattice null" << endl;
// 	      else cout << "lattice not null" << endl;
// 	      cout << "testReachability : " << testReachability << endl;
// 	    }

 	  FXMessageBox::error(this,MBOX_OK,"Error during Trajectory Generation","Trajectory generation failed.");
	  FXString str("Trajectory generation failed after ");
	  str += FXStringVal((FXdouble)trajGenerator->finalNbSteps_,0,FALSE);
	  str += FXString(" steps.");
	  str += reachStatus;
	  msgBox->setText(str);
	}
    }
  catch(exception& e)
    {
      cout << e.what() << "\n";
      return 1;
    } 
  return 0;
}

long TrajGenIfc::onLeftMouseDown(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // grab the image window
  imageview->grab();

  // the left button is pressed
  mlbFlag = true;

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init position on the map
  dc.setForeground(initColor);
  dc.drawEllipse(ev->win_x - 2, ev->win_y - 2, 4, 4);

  // set the coordinates of init position
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  x1 = xM;
  y1 = yM;
  return 1;
}

long TrajGenIfc::onLeftMouseUp(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // ungrab the image window
  imageview->ungrab();

  // the left button is unpressed
  mlbFlag = false;
  
  // compute the init direction
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  // compute the vector between init position and the new position
  int dx = (int)(xM - x1);
  int dy = (int)(yM - y1);
  
  Vec2D vect(dx,dy);
  Vec vectN = vect.normalized();
  Vec2D vectN20(vectN.getVal(0)*20.0f,vectN.getVal(1)*20.0f);

  int xTgtMap = (int)(x1 + vectN20.getVal(0));
  int yTgtMap = (int)(y1 + vectN20.getVal(1));

  // compute the direction from the vector
  double ang = atan2(dy,dx);
  theta1 = ang / M_PI * 180.0f;

  // draw the direction on the map
  int x1Frame=0,y1Frame=0;
  coordMapToCoordFrame((int)x1, (int)y1, x1Frame, y1Frame);

  int xTgtFrame=0,yTgtFrame=0;
  coordMapToCoordFrame(xTgtMap, yTgtMap, xTgtFrame, yTgtFrame);

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init direction on the map
  dc.setForeground(initColor);
  dc.drawLine(x1Frame, y1Frame, xTgtFrame, yTgtFrame);

//   this->loadimage(filename);

  return 1;
}

long TrajGenIfc::onRightMouseDown(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // grab the image window
  imageview->grab();

  // the right button is pressed
  mrbFlag = true;

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init position on the map
  dc.setForeground(goalColor);
  dc.drawEllipse(ev->win_x - 2, ev->win_y - 2, 4, 4);

  // set the coordinates of init position
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  x2 = xM;
  y2 = yM;
  return 1;
}

long TrajGenIfc::onRightMouseUp(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // ungrab the image window
  imageview->ungrab();

  // the right button is unpressed
  mrbFlag = false;

  // compute the goal direction
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  // compute the vector between init position and the new position
  int dx = (int)(xM - x2);
  int dy = (int)(yM - y2);
  
  Vec2D vect(dx,dy);
  Vec vectN = vect.normalized();
  Vec2D vectN20(vectN.getVal(0)*20.0f,vectN.getVal(1)*20.0f);

  int xTgtMap = (int)(x2 + vectN20.getVal(0));
  int yTgtMap = (int)(y2 + vectN20.getVal(1));

  // compute the direction from the vector
  double ang = atan2(dy,dx);
  theta2 = ang / M_PI * 180.0f;

  // draw the direction on the map
  int x2Frame=0,y2Frame=0;
  coordMapToCoordFrame((int)x2, (int)y2, x2Frame, y2Frame);

  int xTgtFrame=0,yTgtFrame=0;
  coordMapToCoordFrame(xTgtMap, yTgtMap, xTgtFrame, yTgtFrame);

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init direction on the map
  dc.setForeground(goalColor);
  dc.drawLine(x2Frame, y2Frame, xTgtFrame, yTgtFrame);

  return 1;
}

long TrajGenIfc::onMouseMove(FXObject*,FXSelector,void*)
{
  return 1;
}

long TrajGenIfc::displayBProfiles(FXObject*,FXSelector,void*)
{
  return 1;
}

long TrajGenIfc::displayWBProfiles(FXObject*,FXSelector,void*)
{
  return 1;
}

long TrajGenIfc::displayAllProfiles(FXObject*,FXSelector,void*)
{
  return 1;
}

long TrajGenIfc::updateUseVelCst(FXObject*,FXSelector,void*)
{
  // update the use of linear velocity constraints
  if(useVelCst == false)
    useVelCst = true;
  else
    useVelCst = false;
  return 1;
}

long TrajGenIfc::updateUseAccCst(FXObject*,FXSelector,void*)
{
  // update the use of linear acceleration constraints
  if(useAccCst == false)
    useAccCst = true;
  else
    useAccCst = false;
  return 1;
}

long TrajGenIfc::updateUseAVelCst(FXObject*,FXSelector,void*)
{
  // update the use of angular velocity constraints
  if(useAVelCst == false)
    useAVelCst = true;
  else
    useAVelCst = false;
  return 1;
}

long TrajGenIfc::updateUseAAccCst(FXObject*,FXSelector,void*)
{
  // update the use of angular acceleration constraints
  if(useAAccCst == false)
    useAAccCst = true;
  else
    useAAccCst = false;
  return 1;
}

long TrajGenIfc::updateUseVelCstWC(FXObject*,FXSelector,void*)
{
  // update the use of linear velocity constraints
  if(useVelCstWC == false)
    useVelCstWC = true;
  else
    useVelCstWC = false;
  return 1;
}

long TrajGenIfc::updateUseAccCstWC(FXObject*,FXSelector,void*)
{
  // update the use of linear acceleration constraints
  if(useAccCstWC == false)
    useAccCstWC = true;
  else
    useAccCstWC = false;
  return 1;
}

long TrajGenIfc::updateUseAVelCstWC(FXObject*,FXSelector,void*)
{
  // update the use of angular velocity constraints
  if(useAVelCstWC == false)
    useAVelCstWC = true;
  else
    useAVelCstWC = false;
  return 1;
}

long TrajGenIfc::updateUseAAccCstWC(FXObject*,FXSelector,void*)
{
  // update the use of angular acceleration constraints
  if(useAAccCstWC == false)
    useAAccCstWC = true;
  else
    useAAccCstWC = false;
  return 1;
}

long TrajGenIfc::drawRStar(FXObject*,FXSelector,void*)
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
//     case CARLIKE2_SYS : 
    case SPACESHIP_SYS : 
//     case SPACESHIP2_SYS : 
//     case DIFFDRIVE_SYS : 
    case WHEELCHAIR_SYS : 
//     case WHEELCHAIR2_SYS : 
//     case PM3D_SYS : 
    case MANA_SYS : 
      this->drawRStar2D();
      break;
    }

  return 1;
}

long TrajGenIfc::updateDrawUncompleteTraj(FXObject*,FXSelector,void*)
{
  // update the drawing of uncomplete trajectories
  if(drawUncompleteTraj == false)
    drawUncompleteTraj = true;
  else
    drawUncompleteTraj = false;
  return 1;
}

long TrajGenIfc::updateDrawTargetPos(FXObject*,FXSelector,void*)
{
  // update the drawing of uncomplete trajectories
  if(drawTargetPos == false)
    drawTargetPos = true;
  else
    drawTargetPos = false;
  return 1;  
}

long TrajGenIfc::drawGnuplot2D(FXObject*,FXSelector,void*)
{
//   cout << "tabPosTraj.size() : " << tabPosTraj.size() << endl;

//   Gnuplot gPosition;
//   gPosition.reset_plot();

//   if(drawTargetPos) gPosition.set_style("points").plot_xy(tabPosReachedX,tabPosReachedY,"");
  
//   for(unsigned int i=0; i<tabPosTraj.size(); i+=4)
//     {      
//       if(tabPosTraj[i] != NULL && tabPosTraj[i+1] != NULL && tabPosTraj[i]->size() != 0 && tabPosTraj[i+1]->size() != 0)
//  	{
// 	  gPosition.cmd("set linestyle 1");
// 	  //gPosition.cmd("set linestyle 1 lt 1 lw 6");
// 	  gPosition.set_style("lines").plot_xy(*(tabPosTraj[i]),*(tabPosTraj[i+1]),"");
//  	}
//     }

//   Gnuplot gPosition2;
//   gPosition2.reset_plot();
  
//   if(drawTargetPos) gPosition2.set_style("points").plot_xyz(tabPosReachedX,tabPosReachedY,tabPosReachedT,"");
  
//   for(unsigned int i=0; i<tabPosTraj.size(); i+=4)
//     {      
//       if(tabPosTraj[i] != NULL && tabPosTraj[i+1] != NULL && tabPosTraj[i]->size() != 0 && tabPosTraj[i+1]->size() != 0)
//  	{
// 	  gPosition2.cmd("set linestyle 1");
// 	  //gPosition.cmd("set linestyle 1 lt 1 lw 6");
// 	  gPosition2.set_style("lines").plot_xyz(*(tabPosTraj[i]),*(tabPosTraj[i+1]),*(tabPosTraj[i+2]),"");
//  	}
//     }

//   Gnuplot gPosition3;
//   gPosition3.reset_plot();
  
//   if(drawTargetPos) gPosition3.set_style("points").plot_xyz(tabPosReachedX,tabPosReachedY,tabPosReachedV,"");
  
//   for(unsigned int i=0; i<tabPosTraj.size(); i+=4)
//     {      
//       if(tabPosTraj[i] != NULL && tabPosTraj[i+1] != NULL && tabPosTraj[i]->size() != 0 && tabPosTraj[i+1]->size() != 0)
//  	{
// 	  gPosition3.cmd("set linestyle 1");
// 	  //gPosition.cmd("set linestyle 1 lt 1 lw 6");
// 	  gPosition3.set_style("lines").plot_xyz(*(tabPosTraj[i]),*(tabPosTraj[i+1]),*(tabPosTraj[i+3]),"");
//  	}
//     }
  
//   wait_for_key();
  return 1;
}

long TrajGenIfc::drawGnuplot3D(FXObject*,FXSelector,void*)
{
  // ??????????????????????
  
  wait_for_key();
  return 1;
}

long TrajGenIfc::drawGnuplot(FXObject*,FXSelector,void*)
{
  this->drawGP();
  return 1;
}

long TrajGenIfc::emptyVecTraj(FXObject*,FXSelector,void*)
{
  this->emptyTrajVec();
  return 1;
}

long TrajGenIfc::drawSSLattice(FXObject*,FXSelector,void*)
{
  cout << "debut drawSSLattice" << endl;

  // check if the map has been defined
  if(map == NULL && debugLevel != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while computing the SSLattice","No map has been loaded");
      return -1;
    }

  // empty the trajectory vectors if needed
  if(debugLevel >= 1)
    {
      this->emptyTrajVec();
    }

  // init the timer
  struct timeval tv1,tv2;
  gettimeofday(&tv1,NULL);

  // init the system constraints
  this->initSys();

  // reinit the state space lattice
  int resInit = this->initSSLattice();
  if(resInit != 0)
    {
      cout << "resInit != 0" << endl;
      return -1;
    }

  // init the initial state
  this->initIST();
  double dt = t2 - t1;

  // compute the state space lattice
  cout << "deb lattice construction" << endl;
  int res = lattice_->construct(*initST_,dt);
  cout << "fin lattice construction" << endl;
  if(res != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while computing the SSLattice","No input system defined!");
      return -1;
    }

  // stop the timer
  gettimeofday(&tv2,NULL);
  
  // compute the time passed
  int elapsedTs = tv2.tv_sec - tv1.tv_sec;
  int elapsedTus = tv2.tv_usec - tv1.tv_usec;
  double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
  cout << "Elapsed time : " << elapsedT << endl; 

  FXString str("State Space Lattice computed in ");
  str += FXStringVal(elapsedT,2,FALSE);
  str += FXString(" sec.");
  msgBox->setText(str);

  cout << "fin drawSSLattice" << endl;

  return 1;
}

long TrajGenIfc::updateDispLattice(FXObject*,FXSelector,void*)
{
  // update the lattice display status
  if(SSL_dispTraj == false)
    SSL_dispTraj = true;
  else
    SSL_dispTraj = false;
  return 1;
}

long TrajGenIfc::drawGPLattice(FXObject*,FXSelector,void*)
{
  if(lattice_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while drawing the SSLattice","The lattice has not been computed!");
      return -1;
    }

  lattice_->drawGP();
  return 1;
}

long TrajGenIfc::genTrajSSLNodes(FXObject*,FXSelector,void*)
{
  //this->emptyTrajVec();
  this->initSampleTraj();

  // check if the lattice has been defined
  if(lattice_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while generating traj to the final states of the lattice","The lattice has not been computed!");
      return -1;
    }

  // get back the number of leaves of the lattice
  int N = lattice_->nbLeaves();

  // statistical values
  double nbTrajTot  = 0;      // total number of trajectories
  double nbTrajCplt = 0;      // number of completed trajectories
  double pcCplt = 0;          // percentage of completed trajectories
  double totFinalNbSteps = 0; // total number of final steps required to converge

  // init the timer
  struct timeval tv1,tv2;
  gettimeofday(&tv1,NULL);

  // initialize the initial state from the gui
  this->initIST();

  cout << "initST : "; initST_->display(); cout << endl;

  // init the system constraints
  this->initSys();

  // init the trajectory generation process
  this->initTrajGen();

  for(int i=0; i<N; i++)
    {
      nbTrajTot++;

      // get back the current leave of the lattice tree
      StateTime* curST = lattice_->leafAt(i);
      this->saveGST(curST);

      cout << "leaf : ";
      cout << "curST : "; curST->display(); cout << endl;

//       // create the new trajectory generator
//       switch(optSys)
// 	{
// 	case CARLIKE_SYS : trajGenerator = new CLTrajGen((CarLikeSys*)sys_); break;
// // 	case CARLIKE2_SYS : trajGenerator = new CLTrajGen((CarLikeSys*)sys_); break;
// // 	case SPACESHIP_SYS : trajGenerator = new SSTrajGen((SpaceShipSys*)sys_); break;
// // 	case SPACESHIP2_SYS : trajGenerator = new SSTrajGen((SpaceShipSys*)sys_); break;
// // 	case DIFFDRIVE_SYS : trajGenerator = new DDTrajGen((DiffDriveSys*)sys_); break;
// // 	case WHEELCHAIR_SYS : trajGenerator = new WCTrajGen((WheelChairSys*)sys_); break;
// // 	case WHEELCHAIR2_SYS : trajGenerator = new WCTrajGen((WheelChairSys*)sys_); break;
// // 	case PM3D_SYS : trajGenerator = new PM3DTrajGen((PointMass3DSys*)sys_); break;
// 	}
      
//       // init options of the trajectory generation
//       trajGenerator->setMaxNbStep(maxNbStep);
//       trajGenerator->setMaxResidual(maxResidual);
      
//       // init the debug level mode
//       trajGenerator->setDBG(debugLevel);

//       if(useIG) trajGenerator->setIGdb(ig_);
//       else trajGenerator->setIGdb(NULL);
      
      // Compute Trajectory
      int res;
      if(genTrajMethod == 0) res = trajGenerator->ComputeTrajectory(*initST_,*curST,NULL);
      else res = trajGenerator->genTraj(*initST_,*curST);

      
      if(res == 0)
	{
	  // case where the trajectory generation succeed
	  nbTrajCplt++;
	  totFinalNbSteps += trajGenerator->finalNbSteps_;

	  if(debugLevel >= 1)
	    {
	      cout << "final state reached "; curST->display(); cout << " after " << trajGenerator->finalNbSteps_ << " steps."<< endl;

	      // trace the trajectory on the map
	      trajGenerator->traceOn(map,300);
	      
	      // compute a sample trajectory to display in a gnuplot viewer  
	      this->sampleTraj(trajGenerator);
	    }
	}
      else
	{
	  if(drawUncompleteTraj && debugLevel >= 1) 
	    {
	      cout << "final state not reached : "; curST->display(); cout << endl;
	      
	      // trace the trajectory on the map
	      trajGenerator->traceOn(map,300);
	      
	      // compute a sample trajectory to display in a gnuplot viewer
	      this->sampleTraj(trajGenerator);
	    }
	}
      cout << "Computed : " << i+1 << "/" << N << " (" << 100.0*((double)(i+1))/((double)N) << "\%)  Complete : " 
	   << nbTrajCplt << "/" << (i+1) << " (" << ((double)nbTrajCplt) / ((double)(i+1)) * 100.0 << "\%)" << endl;
    }

  this->stockSampleTraj();
  
  // stop the timer
  gettimeofday(&tv2,NULL);
  
  // compute the time passed
  int elapsedTs = tv2.tv_sec - tv1.tv_sec;
  int elapsedTus = tv2.tv_usec - tv1.tv_usec;
  double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
  cout << "Elapsed time : " << elapsedT << endl; 

  // compute the average number of steps
  //double avgNbSteps = totFinalNbSteps / nbTrajTot;
  double avgNbSteps = ((double)totFinalNbSteps) / ((double)nbTrajCplt);

  // compute the average time needed by trajectory
  double avgTime = elapsedT / nbTrajTot;
  double avgTPS = 1.0 / avgTime;

  // saving the map
  map->save("/tmp/ResultMap_traj.pgm");
  
  // load the map in the canvas
  this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));

  pcCplt = ((double)nbTrajCplt) / ((double)nbTrajTot) * 100.0;
  
  FXString str("Reachable Star computed in ");
  str += FXStringVal(elapsedT,2,FALSE);
  str += FXString(" sec.  Avg nb steps : ");
  str += FXStringVal(avgNbSteps,2,FALSE);
  str += FXString("  Avg time by traj : ");
  str += FXStringVal(avgTime,4,FALSE);
  str += FXString("  Avg Traj/sec : ");
  str += FXStringVal(avgTPS,2,FALSE);
  str += FXString("  Completed Traj : ");
  str += FXStringVal(nbTrajCplt,0,FALSE);
  str += FXString("/");
  str += FXStringVal(nbTrajTot,0,FALSE);
  str += FXString(" (");
  str += FXStringVal(pcCplt,2,FALSE);
  str += FXString(" %)");
  msgBox->setText(str);

  return 1;
}

long TrajGenIfc::learnIGCmd(FXObject*,FXSelector,void*)
{
  this->learnIG();
  return 1;
}

long TrajGenIfc::changeIGName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Load IG Database (*.dat)");
  open.setFilename(IGfilename);
  if(open.execute()){
    IGfilename=open.getFilename();
  }
  return 1;
}

long TrajGenIfc::loadIG(FXObject*,FXSelector,void*)
{
  // initialize the ig module from the interface
  initIG();

  // check if the ig learning process has been correctly initialized
  if(ig_ == NULL) return -1;

  // load the ig database
  try
    {
      int res = ig_->load((char*)(IGfilename.text()));
      
      if(res != 0)
	{
	  FXMessageBox::error(this,MBOX_OK,"Error loading the IG database","Unknown input file");
	  return -1;
	}

      msgBox->setText("The IG database has been loaded.");
    }
  catch(string s)
    {
      FXString str(s.c_str());
      FXMessageBox::error(this,MBOX_OK,"Error loading the IG database",str.text());
      return -1;
    }
  return 1;
}

long TrajGenIfc::displayIG(FXObject*,FXSelector,void*)
{
  // check if the ig has been defined
  if(ig_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error displaying the IG database","No IG database has been defined");
      return -1;
    }
  
  // display the initial guess database
  ig_->drawGPGrid();
  
  return 1;
}

long TrajGenIfc::updateUseIG(FXObject*,FXSelector,void*)
{
  // update the ig database use
  if(useIG == false)
    useIG = true;
  else
    useIG = false;
  return 1;
}

long TrajGenIfc::updateUseFDPD(FXObject*,FXSelector,void*)
{
  // update the finite difference partial derivative use
  if(useFDPD == false)
    useFDPD = true;
  else
    useFDPD = false;
  return 1;
}

long TrajGenIfc::updateCL_XYf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final car-like position
  if(CL_XYf_fixed == false)
    CL_XYf_fixed = true;
  else
    CL_XYf_fixed = false;
  return 1;
}

long TrajGenIfc::updateCL_THf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final car-like heading
  if(CL_THf_fixed == false)
    CL_THf_fixed = true;
  else
    CL_THf_fixed = false;
  return 1;
}

long TrajGenIfc::updateCL_PHf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final car-like steering angle
  if(CL_PHf_fixed == false)
    CL_PHf_fixed = true;
  else
    CL_PHf_fixed = false;
  return 1;
}

long TrajGenIfc::updateCL_Vf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final car-like velocity
  if(CL_Vf_fixed == false)
    CL_Vf_fixed = true;
  else
    CL_Vf_fixed = false;
  return 1;
}

long TrajGenIfc::updateDD_XYf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final differential drive position
  if(WC_XYf_fixed == false)
    WC_XYf_fixed = true;
  else
    WC_XYf_fixed = false;
  return 1;
}

long TrajGenIfc::updateDD_THf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final differential drive heading
  if(WC_THf_fixed == false)
    WC_THf_fixed = true;
  else
    WC_THf_fixed = false;
  return 1;
}

long TrajGenIfc::updateDD_OMf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final differential drive angular velocity
  if(WC_OMf_fixed == false)
    WC_OMf_fixed = true;
  else
    WC_OMf_fixed = false;
  return 1;
}

long TrajGenIfc::updateDD_Vf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final differential drive linear velocity
  if(WC_Vf_fixed == false)
    WC_Vf_fixed = true;
  else
    WC_Vf_fixed = false;
  return 1;
}

long TrajGenIfc::updateSS_XYf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final spaceship position
  if(SS_XYf_fixed == false)
    SS_XYf_fixed = true;
  else
    SS_XYf_fixed = false;
  return 1;
}

long TrajGenIfc::updateSS_THf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final spaceship heading
  if(SS_THf_fixed == false)
    SS_THf_fixed = true;
  else
    SS_THf_fixed = false;
  return 1;
}

long TrajGenIfc::updateSS_OMf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final spaceship angular velocity
  if(SS_OMf_fixed == false)
    SS_OMf_fixed = true;
  else
    SS_OMf_fixed = false;
  return 1;
}

long TrajGenIfc::updateSS_Vf_fixed(FXObject*,FXSelector,void*)
{
  // update the fixed or free final spaceship linear velocity
  if(SS_Vf_fixed == false)
    SS_Vf_fixed = true;
  else
    SS_Vf_fixed = false;
  return 1;
}

long TrajGenIfc::changeIG1stParamsName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Load IG 1st trajectory parameterizations (*.ltp)");
  open.setFilename(IGfilename);
  if(open.execute()){
    IG1stParamsName=open.getFilename();
  }
  return 1;
}

long TrajGenIfc::updateSaveRStar(FXObject*,FXSelector,void*)
{
  // update the rstar saving option
  if(saveRStar == false)
    saveRStar = true;
  else
    saveRStar = false;
  return 1;
}

long TrajGenIfc::changeSaveRStarName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Save RStar Trajectories (*.ltp)");
  open.setFilename(IGfilename);
  if(open.execute()){
    IG1stParamsName=open.getFilename();
  }
  return 1;
}

long TrajGenIfc::updateCMBackCL(FXObject*,FXSelector,void*)
{
  // update the option to move backward for the car-like
  if(canMoveBackCL == false)
    canMoveBackCL = true;
  else
    canMoveBackCL = false;
  return 1;
}

long TrajGenIfc::updateCMBackWC(FXObject*,FXSelector,void*)
{
  // update the option to move backward for the wheelchair
  if(canMoveBackWC == false)
    canMoveBackWC = true;
  else
    canMoveBackWC = false;
  return 1;
}

long TrajGenIfc::computeCLManeuver(FXObject*,FXSelector,void*)
{
  // init the car-like system
  this->initCarLikeSys();

  // init the car-like initial state-time
  this->initICLST();
  this->initGCLST();

  /// compute the set of state-times to reach from the initial state-time and the car-like features

  // compute the radius of the minimal curvature circle
  double L = ((CarLikeSys*)sys_)->getLength();
  double phiMax = ((CarLikeSys*)sys_)->getMaxSteering();
  double rad = L / tan(phiMax);

  double x_0 = ((CLStateT*)initST_)->x();
  double y_0 = ((CLStateT*)initST_)->y();
  double th_0 = ((CLStateT*)initST_)->th();

  double t_f = ((CLStateT*)goalST_)->t();

  PositionXY tabPos[6];
  tabPos[0] = PositionXY(x_0 - rad*sin(th_0),
			 y_0 + rad*cos(th_0));
  tabPos[1] = PositionXY(x_0 + rad*sin(th_0),
			 y_0 - rad*cos(th_0));
  tabPos[2] = PositionXY(x_0 - 2.0*rad*sin(th_0),
			 y_0 + 2.0*rad*cos(th_0));
  tabPos[3] = PositionXY(x_0 + 2.0*rad*sin(th_0),
			 y_0 - 2.0*rad*cos(th_0));
  tabPos[4] = PositionXY(x_0 - rad*sin(th_0) - rad*cos(th_0),
			 y_0 + rad*cos(th_0) - rad*sin(th_0));
  tabPos[5] = PositionXY(x_0 + rad*sin(th_0) - rad*cos(th_0),
			 y_0 - rad*cos(th_0) - rad*sin(th_0));
//   tabPos[6] = PositionXY(x_0,y_0);
  
  double tabHead[2];
  tabHead[0] = normalizedAngle(th_0 + 0.16666*M_PI);
  tabHead[1] = normalizedAngle(th_0 - 0.16666*M_PI);
//   tabHead[2] = th_0;

  CLStateT tabST[21];
  for(int i=0; i<7; i++)
    {
      for(int j=0; j<3; j++)
	{
	  tabST[i*2+j] = CLStateT(tabPos[i].x(),tabPos[i].y(),tabHead[j],0,0,t_f);
	}
    }

  /// generate a trajectory to reach each goal state-time

  // init the trajectory sampling to display them
  this->initSampleTraj();

  if(map == NULL && debugLevel != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot draw the maneuvers","No map has been loaded");
      return -1;
    }
  
  // statistical values
  double nbTrajTot  = 0;     // total number of trajectories
  double nbTrajCplt = 0;     // number of completed trajectories
  double pcCplt = 0;      // percentage of completed trajectories

  // init the timer
  struct timeval tv1,tv2;
  gettimeofday(&tv1,NULL);

  if(debugLevel >= 1) { this->emptyTrajVec(); }

  double totFinalNbSteps = 0;

  // init the trajectory generation process
  initCLTrajGen();
  trajGenerator->setDBG(debugLevel);

  for(int i=0; i<12; i++)
    {
      nbTrajTot++;

      // init the final state-time
      if(goalST_ != NULL) delete goalST_;
      goalST_ = new CLStateT(tabST[i]);

      tabPosReachedX.push_back(((CLStateT*)goalST_)->x());
      tabPosReachedY.push_back(((CLStateT*)goalST_)->y());
      tabPosReachedV.push_back(((CLStateT*)goalST_)->s());
      tabPosReachedT.push_back(((CLStateT*)goalST_)->t());

      //Compute trajectory
      int res = -2;
//       if(i == 1)
// 	{
	  if(genTrajMethod == 0) res = trajGenerator->ComputeTrajectory(*initST_,*goalST_,NULL);
	  else res = trajGenerator->genTraj(*initST_,*goalST_);
// 	}

      if(res == 0)
	{
	  // case where the trajectory generation succeed
	  nbTrajCplt++;
	  totFinalNbSteps += trajGenerator->finalNbSteps_;
	  
	  if(debugLevel >= 1)
	    {
	      // trace the trajectory on the map
	      trajGenerator->traceOn(map,300);
	      
	      // compute a sample trajectory to display in a gnuplot viewer
	      this->sampleTraj(trajGenerator);
	    }	  
	}
      else if(res == -1)
	{
	  // case where the trajectory generation didn't reach the goal : 
	  // trace the better trajectory found during the generation
	  
	  // trace the trajectory on the map
	  if(drawUncompleteTraj) 
	    {
	      if(debugLevel >= 1)
		{
		  // trace the trajectory on the map
		  trajGenerator->traceOn(map,300);
		  
		  // compute a sample trajectory to display in a gnuplot viewer
		  this->sampleTraj(trajGenerator);
		}	      
	    }
	}
      else // (res == -2)
	{
	  // nothing to do
	}      
    }
  // stock the current branch
  if(debugLevel >= 1) this->stockSampleTraj();

  // stop the timer
  gettimeofday(&tv2,NULL);
  
  // compute the time passed
  int elapsedTs = tv2.tv_sec - tv1.tv_sec;
  int elapsedTus = tv2.tv_usec - tv1.tv_usec;
  double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
  cout << "Elapsed time : " << elapsedT << endl; 

  // compute the average number of steps
  double nbTraj = (nbBranches*trajByBranch*nbItvDistRS);
  //double avgNbSteps = totFinalNbSteps / nbTraj;
  double avgNbSteps = ((double)totFinalNbSteps) / ((double)nbTrajCplt);

  // compute the average time needed by trajectory
  double avgTime = elapsedT / nbTraj;
  double avgTPS = 1.0 / avgTime;

  // saving the map
  map->save("/tmp/ResultMap_traj.pgm");
  
  // load the map in the canvas
  this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));

  pcCplt = ((double)nbTrajCplt) / ((double)nbTrajTot) * 100.0;
  
  FXString str("CL Maneuvers computed in ");
  str += FXStringVal(elapsedT,2,FALSE);
  str += FXString(" sec.  Avg nb steps : ");
  str += FXStringVal(avgNbSteps,2,FALSE);
  str += FXString("  Avg time by traj : ");
  str += FXStringVal(avgTime,4,FALSE);
  str += FXString("  Avg Traj/sec : ");
  str += FXStringVal(avgTPS,2,FALSE);
  str += FXString("  Completed Traj : ");
  str += FXStringVal(nbTrajCplt,0,FALSE);
  str += FXString("/");
  str += FXStringVal(nbTrajTot,0,FALSE);
  str += FXString(" (");
  str += FXStringVal(pcCplt,2,FALSE);
  str += FXString(" %)");
  msgBox->setText(str);

  return 1;
}

long TrajGenIfc::updateDispCurvCirc(FXObject*,FXSelector,void*)
{
  // update the display of the CL min curvature circles
  if(dispCurvCirc == false)
    dispCurvCirc = true;
  else
    dispCurvCirc = false;
  return 1;  
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Other functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 *  \brief init the time constraint of the trajectory generation process
 *  \return 0 if the process succeed
 *  \return -1 else
 */
int TrajGenIfc::initTimeCstt()
{
  if(trajGenerator == NULL) 
    {
      FXMessageBox::error(this,MBOX_OK,"Error during the initialization of the time constraint","The trajectory generator has not been allocated");
      return -1;
    }

  timeCstt_t tcS = TIME_FIXED;
  switch(timeCsttStatus)
    {
    case 0 : tcS = TIME_FIXED; break;
    case 1 : tcS = TIME_FREE; break;
    case 2 : 
      {
	double dt = t2 - t1;
	if(dt < minTime || dt > maxTime)
	  {
	    FXMessageBox::error(this,MBOX_OK,"Error during the initialization of the time constraint","The initial time between both states is not include in the time constraints");
	    return -1;
	  }
	tcS = TIME_CSTD; 
	break;
      }
    }
  trajGenerator->setTcStatus(tcS,minTime,maxTime);
  return 0;
}

/**
 *  \brief init the state space lattice
 *  \return 0 if the process succeed
 *  \return -1 else
 */
int TrajGenIfc::initSSLattice()
{
  // realloc the lattice
  if(lattice_ != NULL) delete lattice_;
  switch(optSys)
    {
    case CARLIKE_SYS : 
//     case CARLIKE2_SYS : 
      lattice_ = new CL_SSLattice((CarLikeSys*)sys_);
      break;
    case SPACESHIP_SYS : 
//     case SPACESHIP2_SYS : 
      lattice_ = new SS_SSLattice((SpaceShipSys*)sys_);
      break;
//     case DIFFDRIVE_SYS : 
//       FXMessageBox::error(this,MBOX_OK,"Error while drawing the SSLattice","The differential drive lattice has not been implemented yet.");
//       lattice_ = NULL;
//       return -1;
    case WHEELCHAIR_SYS : 
//     case WHEELCHAIR2_SYS : 
      lattice_ = new WC_SSLattice((WheelChairSys*)sys_);
      break;

    case MANA_SYS : 
      cerr << "TrajGenIfc::initSSLattice : No MANA SSLattice implemented yet." << endl;
      return -1;
    }
  
  // init the lattice parameters
  lattice_->setHeight(SSL_height);
  lattice_->setNbCS(SSL_nbCS);
  lattice_->setDispTraj(SSL_dispTraj);
  lattice_->setTrajSp(SSL_trajSp);

  return 0;
}

/**
 *  \brief initialize the trajectory generator for a car-like system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 */
void TrajGenIfc::initCarLikeSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = kp1;
  x0[4] = v1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x2;
  xf[1] = y2;
  xf[2] = degToRad(theta2);
  xf[3] = kp2;
  xf[4] = v2;
  xf[5] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(xf[4]);
  tabPosReachedT.push_back(xf[5]);
  
  // init the constraints vector
  cstt[0] = 0.0;
  cstt[1] = rb_AMAX;
  cstt[2] = rb_VMAX;
  cstt[3] = rb_L;
  cstt[4] = degToRad(rb_KMAX);
  cstt[5] = degToRad(rb_OMAX);

  ((CLTrajGen*)trajGenerator)->setUseVelCst(useVelCst);
  ((CLTrajGen*)trajGenerator)->setUseAccCst(useAccCst);
  ((CLTrajGen*)trajGenerator)->setUseAVelCst(useAVelCst);
  ((CLTrajGen*)trajGenerator)->setUseAAccCst(useAAccCst);
}

/**
 *  \brief initialize the trajectory generator for a car-like system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 */
void TrajGenIfc::initCarLikeSys2(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = kp1;
  x0[4] = v1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x2;
  xf[1] = y2;
  xf[2] = degToRad(theta2);
  xf[3] = kp2;
  xf[4] = v2;
  xf[5] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(xf[4]);
  tabPosReachedT.push_back(xf[5]);
  
  // init the constraints vector
  cstt[0] = rb_AMAX;
  cstt[1] = rb_VMAX;
  cstt[2] = rb_L;
  cstt[3] = degToRad(rb_KMAX);
  cstt[4] = degToRad(rb_OMAX);
}

/**
 *  \brief initialize the trajectory generator for a spaceship system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 */
void TrajGenIfc::initSpaceShipSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt)
{
  cout << "x1 : " << x1 << endl;
  cout << "y1 : " << y1 << endl;
  cout << "theta1 : " << theta1 << endl;
  cout << "vx1 : " << vx1 << endl;
  cout << "vy1 : " << vy1 << endl;
  cout << "om1 : " << om1 << endl;
  cout << "t1 : " << t1 << endl;

  cout << "avant" << endl;

  cout << "x0[0] : " << x0[0] << endl;
  cout << "x0[1] : " << x0[1] << endl;
  cout << "x0[2] : " << x0[2] << endl;
  cout << "x0[3] : " << x0[3] << endl;
  cout << "x0[4] : " << x0[4] << endl;
  cout << "x0[5] : " << x0[5] << endl;
  cout << "x0[6] : " << x0[6] << endl;

  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = vx1;
  x0[4] = vy1;
  x0[5] = om1;
  x0[6] = t1;

  cout << "apres" << endl;

  cout << "x0[0] : " << x0[0] << endl;
  cout << "x0[1] : " << x0[1] << endl;
  cout << "x0[2] : " << x0[2] << endl;
  cout << "x0[3] : " << x0[3] << endl;
  cout << "x0[4] : " << x0[4] << endl;
  cout << "x0[5] : " << x0[5] << endl;
  cout << "x0[6] : " << x0[6] << endl;
  
  // init the final state vector
  xf[0] = x2;
  xf[1] = y2;
  xf[2] = degToRad(theta2);
  xf[3] = vx2;
  xf[4] = vy2;
  xf[5] = om2;
  xf[6] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(0.5*(xf[3] + xf[4]));
  tabPosReachedT.push_back(xf[6]);

  // init the constraints vector
  cstt[0] = SS_AAccMax;
  cstt[1] = SS_AccMax;
  cstt[2] = SS_AVelMax;
  cstt[3] = SS_VelMax;
}

/**
 *  \brief initialize the trajectory generator for a differential drive system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 */
void TrajGenIfc::initDiffDriveSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = v1;
  x0[4] = om1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x2;
  xf[1] = y2;
  xf[2] = degToRad(theta2);
  xf[3] = v2;
  xf[4] = om2;
  xf[5] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(xf[3]);
  tabPosReachedT.push_back(xf[5]);
  
  // init the constraints vector
  cstt[0] = dd_WAMAX;
  cstt[1] = dd_WVMAX;
  cstt[2] = dd_RAD;
}

/**
 *  \brief initialize the trajectory generator for a differential drive system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 */
void TrajGenIfc::initWheelChairSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = v1;
  x0[4] = om1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x2;
  xf[1] = y2;
  xf[2] = degToRad(theta2);
  xf[3] = v2;
  xf[4] = om2;
  xf[5] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(xf[3]);
  tabPosReachedT.push_back(xf[5]);
  
  // init the constraints vector
  cstt[0] = WC_AccMax;
  cstt[1] = WC_AAccMax;
  cstt[2] = WC_VelMax;
  cstt[3] = WC_AVelMax;

  ((WCTrajGen*)trajGenerator)->setUseVelCst(useVelCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAccCst(useAccCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAVelCst(useAVelCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAAccCst(useAAccCstWC);
}

/**
 *  \brief initialize the trajectory generator for a Mana system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 */
void TrajGenIfc::initManaSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = v1;
  x0[4] = om1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x2;
  xf[1] = y2;
  xf[2] = degToRad(theta2);
  xf[3] = v2;
  xf[4] = om2;
  xf[5] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(xf[3]);
  tabPosReachedT.push_back(xf[5]);
  
  // init the constraints vector
  cstt[0] = WC_AccMax;
  cstt[1] = WC_AAccMax;
  cstt[2] = WC_VelMax;
  cstt[3] = WC_AVelMax;

  ((ManaTrajGen*)trajGenerator)->setUseVelCst(useVelCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAccCst(useAccCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAVelCst(useAVelCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAAccCst(useAAccCstWC);
}

/**
 *  \brief initialize the trajectory generator for a car-like system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initCarLikeSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = kp1;
  x0[4] = v1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[1] = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  
  xf[3] = kp2;
  xf[4] = v2;
  xf[5] = t2;
  
//   if(j == 0) 
//     {
      tabPosReachedX.push_back(xf[0]);
      tabPosReachedY.push_back(xf[1]);
      tabPosReachedV.push_back(xf[4]);
      tabPosReachedSt.push_back(xf[3]);
      tabPosReachedT.push_back(xf[5]);
//     }
  
  // init the constraints vector
  cstt[0] = 0.0;
  cstt[1] = rb_AMAX;
  cstt[2] = rb_VMAX;
  cstt[3] = rb_L;
  cstt[4] = degToRad(rb_KMAX);
  cstt[5] = degToRad(rb_OMAX);
    
  ((CLTrajGen*)trajGenerator)->setUseVelCst(useVelCst);
  ((CLTrajGen*)trajGenerator)->setUseAccCst(useAccCst);
  ((CLTrajGen*)trajGenerator)->setUseAVelCst(useAVelCst);
  ((CLTrajGen*)trajGenerator)->setUseAAccCst(useAAccCst); 
}

/**
 *  \brief initialize the trajectory generator for a car-like system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initCarLikeSys2(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = kp1;
  x0[4] = v1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[1] = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  
  xf[3] = kp2;
  xf[4] = v2;
  xf[5] = t2;
  
  if(j == 0) 
    {
      tabPosReachedX.push_back(xf[0]);
      tabPosReachedY.push_back(xf[1]);
      tabPosReachedV.push_back(xf[4]);
      tabPosReachedSt.push_back(xf[3]);
      tabPosReachedT.push_back(xf[5]);
    }
  
  // init the constraints vector
  cstt[0] = rb_AMAX;
  cstt[1] = rb_VMAX;
  cstt[2] = rb_L;
  cstt[3] = degToRad(rb_KMAX);
  cstt[4] = degToRad(rb_OMAX);
}

/**
 *  \brief initialize the trajectory generator for a spaceship system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initSpaceShipSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = vx1;
  x0[4] = vy1;
  x0[5] = om1;
  x0[6] = t1;
  
  // init the final state vector
  xf[0] = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[1] = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  xf[3] = vx2;
  xf[4] = vy2;
  xf[5] = om2;
  xf[6] = t2;
  
//   if(j == 0) 
//     {
//       tabPosReachedX.push_back(xf[0]);
//       tabPosReachedY.push_back(xf[1]);
//       tabPosReachedV.push_back(xf[4]);
//       tabPosReachedSt.push_back(xf[3]);
//       tabPosReachedT.push_back(xf[5]);
//     }
  
  // init the constraints vector
  cstt[0] = SS_AAccMax;
  cstt[1] = SS_AccMax;
  cstt[2] = SS_AVelMax;
  cstt[3] = SS_VelMax;
}

/**
 *  \brief initialize the trajectory generator for a differential drive system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initDiffDriveSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = v1;
  x0[4] = om1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[1] = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  xf[3] = v2;
  xf[4] = om2;
  xf[5] = t2;

//   if(j == 0) 
//     {
      tabPosReachedX.push_back(xf[0]);
      tabPosReachedY.push_back(xf[1]);
      tabPosReachedV.push_back(xf[3]);
      tabPosReachedT.push_back(xf[5]);
//     }
  
  // init the constraints vector
  cstt[0] = dd_WAMAX;
  cstt[1] = dd_WVMAX;
  cstt[2] = dd_RAD;
}

/**
 *  \brief initialize the trajectory generator for a differential drive system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initWheelChairSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = v1;
  x0[4] = om1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[1] = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  
  xf[3] = v2;
  xf[4] = om2;
  xf[5] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(xf[3]);
  tabPosReachedT.push_back(xf[5]);
  
  // init the constraints vector
  cstt[0] = WC_AccMax;
  cstt[1] = WC_AAccMax;
  cstt[2] = WC_VelMax;
  cstt[3] = WC_AVelMax;

  ((WCTrajGen*)trajGenerator)->setUseVelCst(useVelCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAccCst(useAccCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAVelCst(useAVelCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAAccCst(useAAccCstWC);
}

/**
 *  \brief initialize the trajectory generator for a Mana system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initManaSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k)
{
  // init the initial state vector
  x0[0] = x1;
  x0[1] = y1;
  x0[2] = degToRad(theta1);
  x0[3] = v1;
  x0[4] = om1;
  x0[5] = t1;
  
  // init the final state vector
  xf[0] = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[1] = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  
  xf[3] = v2;
  xf[4] = om2;
  xf[5] = t2;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedV.push_back(xf[3]);
  tabPosReachedT.push_back(xf[5]);
  
  // init the constraints vector
  cstt[0] = WC_AccMax;
  cstt[1] = WC_AAccMax;
  cstt[2] = WC_VelMax;
  cstt[3] = WC_AVelMax;

  ((ManaTrajGen*)trajGenerator)->setUseVelCst(useVelCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAccCst(useAccCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAVelCst(useAVelCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAAccCst(useAAccCstWC);
}

/**
 *  \brief initialize the trajectory generator for a differential drive system
 *  \param x0 : init state
 *  \param xf : final state
 *  \param cstt : constraints defined over the system
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initPM3DSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k)
{
  // init the initial state vector
  x0[0] = x1_3D;
  x0[1] = y1_3D;
  x0[2] = z1_3D;
  x0[3] = vx1_3D;
  x0[4] = vy1_3D;
  x0[5] = vz1_3D;
  x0[6] = t1_3D;

  //xf[0] = x1_3D + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)j)/(nbBranches-1))*(maxAngRS-minAngRS))) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  //xf[1] = y1_3D + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)j)/(nbBranches-1))*(maxAngRS-minAngRS))) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  //  xf[2] = z1_3D + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)j)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[0] = x1_3D + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[1] = y1_3D + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  xf[2] = z1_3D;
  xf[3] = vx2_3D;
  xf[4] = vy2_3D;
  xf[5] = vz2_3D;
  xf[6] = t2_3D;

  tabPosReachedX.push_back(xf[0]);
  tabPosReachedY.push_back(xf[1]);
  tabPosReachedY.push_back(xf[2]);
  tabPosReachedVX.push_back(xf[3]);
  tabPosReachedVY.push_back(xf[4]);
  tabPosReachedVZ.push_back(xf[5]);
  tabPosReachedT.push_back(xf[6]);

  // init the constraints vector
  cstt[0] = PM3D_AX;
  cstt[1] = PM3D_AY;
  cstt[2] = PM3D_AZ;
  cstt[3] = PM3D_VX;
  cstt[4] = PM3D_VY;
  cstt[5] = PM3D_VZ;
}

/**
 *  \brief initialize the robotic system constraints from the gui
 */
void TrajGenIfc::initSys()
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
      this->initCarLikeSys(); break;
    case SPACESHIP_SYS : 
      this->initSpaceShipSys(); break;
    case WHEELCHAIR_SYS : 
      this->initWheelChairSys(); break;
    case MANA_SYS : 
      this->initManaSys(); break;
    }
}

/**
 *  \brief initialize a car-like system from the gui
 */
void TrajGenIfc::initCarLikeSys()
{
  if(sys_ != NULL) delete sys_;
  sys_ = new CarLikeSys;

  ((CarLikeSys*)sys_)->setLength(rb_L);
  ((CarLikeSys*)sys_)->setRadius(rb_L);
  ((CarLikeSys*)sys_)->setVMax(rb_VMAX);
  ((CarLikeSys*)sys_)->setAMax(rb_AMAX);
//   ((CarLikeSys*)sys_)->setMaxSteering(degToRad(rb_KMAX));
//   ((CarLikeSys*)sys_)->setMaxSteerSpd(degToRad(rb_OMAX));
  ((CarLikeSys*)sys_)->setMaxSteering(rb_KMAX);
  ((CarLikeSys*)sys_)->setMaxSteerSpd(rb_OMAX);

  ((CarLikeSys*)sys_)->setMoveBack(canMoveBackCL);
}

/**
 *  \brief initialize a wheelchair system from the gui
 */
void TrajGenIfc::initWheelChairSys()
{
  if(sys_ != NULL) delete sys_;
  sys_ = new WheelChairSys;
  
  //   sys_->setLength(const double& l);
  //   sys_->setWidth(const double& w);
  //   sys_->setWheelRad(const double& r);
  //   sys_->setWheelBase(const double& r);
  ((WheelChairSys*)sys_)->setVMax(WC_VelMax);
  ((WheelChairSys*)sys_)->setAMax(WC_AccMax);
  ((WheelChairSys*)sys_)->setAVMax(WC_AVelMax);
  ((WheelChairSys*)sys_)->setAAMax(WC_AAccMax);
  //   sys_->setBrakingDist(const double& bd);

  ((WheelChairSys*)sys_)->setMoveBack(canMoveBackWC);
}

/**
 *  \brief initialize a Mana system from the gui
 */
void TrajGenIfc::initManaSys()
{
  if(sys_ != NULL) delete sys_;
  sys_ = new ManaSys;
  
  ((ManaSys*)sys_)->setVMax(WC_VelMax);
  ((ManaSys*)sys_)->setAMax(WC_AccMax);
  ((ManaSys*)sys_)->setAVMax(WC_AVelMax);
  ((ManaSys*)sys_)->setAAMax(WC_AAccMax);

  ((ManaSys*)sys_)->setMoveBack(canMoveBackWC);
}

// /**
//  *  \brief initialize a 3D point mass system from the gui
//  */
// void TrajGenIfc::initPM3DSys()
// {
//   if(sys_ != NULL) delete sys_;
//   sys_ = new PointMass3DSys;

//   ((PointMass3DSys*)sys_)->setVXMax(PM3D_VX);
//   ((PointMass3DSys*)sys_)->setVYMax(PM3D_VY);
//   ((PointMass3DSys*)sys_)->setVZMax(PM3D_VZ);
//   ((PointMass3DSys*)sys_)->setAXMax(PM3D_AX);
//   ((PointMass3DSys*)sys_)->setAYMax(PM3D_AY);
//   ((PointMass3DSys*)sys_)->setAZMax(PM3D_AZ);
// }

/**
 *  \brief initialize a spaceship system from the gui
 */
void TrajGenIfc::initSpaceShipSys()
{
  if(sys_ != NULL) delete sys_;
  sys_ = new SpaceShipSys;

  ((SpaceShipSys*)sys_)->setVXMax(SS_VelMax);
  ((SpaceShipSys*)sys_)->setVYMax(SS_VelMax);
  ((SpaceShipSys*)sys_)->setVThMax(SS_AVelMax);
  ((SpaceShipSys*)sys_)->setAXMax(SS_AccMax);
  ((SpaceShipSys*)sys_)->setAYMax(SS_AccMax);
  ((SpaceShipSys*)sys_)->setAThMax(SS_AAccMax);
}

/**
 *  \brief init the trajectory generation process from the gui
 */
void TrajGenIfc::initTrajGen()
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
      this->initCLTrajGen(); break;
    case SPACESHIP_SYS : 
      this->initSSTrajGen(); break;
    case WHEELCHAIR_SYS : 
      this->initWCTrajGen(); break;
    case MANA_SYS : 
      this->initManaTrajGen(); break;
    }

  // init options of the trajectory generation
  trajGenerator->setMaxNbStep(maxNbStep);
  trajGenerator->setMaxResidual(maxResidual);
  
  // init the debug level mode
  trajGenerator->setDBG(debugLevel);

  // init the IG database use
  if(useIG) trajGenerator->setIGdb(ig_);
  else trajGenerator->setIGdb(NULL);

  // init the finite difference partial derivative
  trajGenerator->setFDpartDeriv(useFDPD);
}

/**
 *  \brief init the trajectory generation process for a car-like system from the gui
 */
void TrajGenIfc::initCLTrajGen()
{
  if(trajGenerator != NULL) delete trajGenerator;
  trajGenerator = new CLTrajGen((CarLikeSys*)sys_);

  ((CLTrajGen*)trajGenerator)->setUseVelCst(useVelCst);
  ((CLTrajGen*)trajGenerator)->setUseAccCst(useAccCst);
  ((CLTrajGen*)trajGenerator)->setUseAVelCst(useAVelCst);
  ((CLTrajGen*)trajGenerator)->setUseAAccCst(useAAccCst);

  ((CLTrajGen*)trajGenerator)->fixFinalXY(CL_XYf_fixed);
  ((CLTrajGen*)trajGenerator)->fixFinalTH(CL_THf_fixed);
  ((CLTrajGen*)trajGenerator)->fixFinalV(CL_Vf_fixed);
  ((CLTrajGen*)trajGenerator)->fixFinalPH(CL_PHf_fixed);
}

/**
 *  \brief init the trajectory generation process for a wheelchair system from the gui
 */
void TrajGenIfc::initWCTrajGen()
{
  if(trajGenerator != NULL) delete trajGenerator;
  trajGenerator = new WCTrajGen((WheelChairSys*)sys_);

  ((WCTrajGen*)trajGenerator)->setUseVelCst(useVelCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAccCst(useAccCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAVelCst(useAVelCstWC);
  ((WCTrajGen*)trajGenerator)->setUseAAccCst(useAAccCstWC);

  ((WCTrajGen*)trajGenerator)->fixFinalXY(WC_XYf_fixed);
  ((WCTrajGen*)trajGenerator)->fixFinalTH(WC_THf_fixed);
  ((WCTrajGen*)trajGenerator)->fixFinalV(WC_Vf_fixed);
  ((WCTrajGen*)trajGenerator)->fixFinalOM(WC_OMf_fixed);
}

/**
 *  \brief init the trajectory generation process for a Mana system from the gui
 */
void TrajGenIfc::initManaTrajGen()
{
  if(trajGenerator != NULL) delete trajGenerator;
  trajGenerator = new ManaTrajGen((ManaSys*)sys_);

  ((ManaTrajGen*)trajGenerator)->setUseVelCst(useVelCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAccCst(useAccCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAVelCst(useAVelCstWC);
  ((ManaTrajGen*)trajGenerator)->setUseAAccCst(useAAccCstWC);

  ((ManaTrajGen*)trajGenerator)->fixFinalXY(WC_XYf_fixed);
  ((ManaTrajGen*)trajGenerator)->fixFinalTH(WC_THf_fixed);
  ((ManaTrajGen*)trajGenerator)->fixFinalV(WC_Vf_fixed);
  ((ManaTrajGen*)trajGenerator)->fixFinalOM(WC_OMf_fixed);
}

/**
 *  \brief init the trajectory generation process for a spaceship system from the gui
 */
void TrajGenIfc::initSSTrajGen()
{
  if(trajGenerator != NULL) delete trajGenerator;
  trajGenerator = new SSTrajGen((SpaceShipSys*)sys_);

  ((SSTrajGen*)trajGenerator)->fixFinalXY(SS_XYf_fixed);
  ((SSTrajGen*)trajGenerator)->fixFinalTH(SS_THf_fixed);
  ((SSTrajGen*)trajGenerator)->fixFinalV(SS_Vf_fixed);
  ((SSTrajGen*)trajGenerator)->fixFinalOM(SS_OMf_fixed);
}

/**
 *  \brief init the initial state-time from the gui
 */
void TrajGenIfc::initIST()
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
      this->initICLST(); break;
    case SPACESHIP_SYS : 
      this->initISSST(); break;
    case WHEELCHAIR_SYS : 
      this->initIWCST(); break;
    case MANA_SYS : 
      this->initIManaST(); break;
    }
}

/**
 *  \brief init the initial car-like state-time from the gui
 */
void TrajGenIfc::initICLST()
{
  if(initST_ != NULL) delete initST_;
  initST_ = new CLStateT(x1,y1,degToRad(theta1),kp1,v1,t1);
}

/**
 *  \brief init the initial spaceship state-time from the gui
 */
void TrajGenIfc::initISSST()
{
  if(initST_ != NULL) delete initST_;
  initST_ = new SSStateT(x1,y1,theta1,vx1,vy1,om1,t1);
}

/**
 *  \brief init the initial wheelchair state-time from the gui
 */
void TrajGenIfc::initIWCST()
{
  if(initST_ != NULL) delete initST_;
  initST_ = new WCStateT(x1,y1,degToRad(theta1),v1,om1,t1);
}

/**
 *  \brief init the initial Mana state-time from the gui
 */
void TrajGenIfc::initIManaST()
{
  if(initST_ != NULL) delete initST_;
  initST_ = new ManaStateT(x1,y1,degToRad(theta1),v1,om1,t1);
}

// /**
//  *  \brief init the initial 3D point mass state-time from the gui
//  */
// void TrajGenIfc::initIPM3DST()
// {
//   if(initST_ != NULL) delete initST_;
//   initST_ = new PM3DStateT(x1_3D,y1_3D,z1_3D,vx1_3D,vy1_3D,vz1_3D,t1);
// }

/**
 *  \brief save the current goal state-time to be able to display it
 *  \param gst : the goal state-time to save
 */
void TrajGenIfc::saveGST(const StateTime* gst)
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
      {
	tabPosReachedX.push_back(((CLStateT*)gst)->x());
	tabPosReachedY.push_back(((CLStateT*)gst)->y());
	tabPosReachedV.push_back(((CLStateT*)gst)->s());
	tabPosReachedT.push_back(((CLStateT*)gst)->t());
	break;
      }
    case SPACESHIP_SYS : 
      {
	tabPosReachedX.push_back(((SSStateT*)gst)->x());
	tabPosReachedY.push_back(((SSStateT*)gst)->y());
	double v = sqrt(vx2*vx2+vy2*vy2);
	tabPosReachedV.push_back(v);
	tabPosReachedT.push_back(((SSStateT*)gst)->t());
	break;
      }
    case WHEELCHAIR_SYS : 
      {
	tabPosReachedX.push_back(((WCStateT*)gst)->x());
	tabPosReachedY.push_back(((WCStateT*)gst)->y());
	tabPosReachedV.push_back(((WCStateT*)gst)->s());
	tabPosReachedT.push_back(((WCStateT*)gst)->t());
	break;
      }
    case MANA_SYS : 
      {
	tabPosReachedX.push_back(((ManaStateT*)gst)->x());
	tabPosReachedY.push_back(((ManaStateT*)gst)->y());
	tabPosReachedV.push_back(((ManaStateT*)gst)->v());
	tabPosReachedT.push_back(((ManaStateT*)gst)->t());
	break;
      }
    }
}

/**
 *  \brief init the goal state-time from the gui
 */
void TrajGenIfc::initGST()
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
      this->initGCLST(); break;
    case SPACESHIP_SYS : 
      this->initGSSST(); break;
    case WHEELCHAIR_SYS : 
      this->initGWCST(); break;
    case MANA_SYS : 
      this->initGManaST(); break;
    }
}

/**
 *  \brief init the goal car-like state-time from the gui
 */
void TrajGenIfc::initGCLST()
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new CLStateT(x2,y2,degToRad(theta2),kp2,v2,t2);

  tabPosReachedX.push_back(((CLStateT*)goalST_)->x());
  tabPosReachedY.push_back(((CLStateT*)goalST_)->y());
  tabPosReachedV.push_back(((CLStateT*)goalST_)->s());
  tabPosReachedT.push_back(((CLStateT*)goalST_)->t());
}

/**
 *  \brief init the goal spaceship state-time from the gui
 */
void TrajGenIfc::initGSSST()
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new SSStateT(x2,y2,theta2,vx2,vy2,om2,t2);

  tabPosReachedX.push_back(((SSStateT*)goalST_)->x());
  tabPosReachedY.push_back(((SSStateT*)goalST_)->y());
  double v = sqrt(vx2*vx2+vy2*vy2);
  tabPosReachedV.push_back(v);
  tabPosReachedT.push_back(((SSStateT*)goalST_)->t());
}

/**
 *  \brief init the goal wheelchair state-time from the gui
 */
void TrajGenIfc::initGWCST()
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new WCStateT(x2,y2,degToRad(theta2),v2,om2,t2);

  tabPosReachedX.push_back(((WCStateT*)goalST_)->x());
  tabPosReachedY.push_back(((WCStateT*)goalST_)->y());
  tabPosReachedV.push_back(((WCStateT*)goalST_)->s());
  tabPosReachedT.push_back(((WCStateT*)goalST_)->t());
}

/**
 *  \brief init the goal Mana state-time from the gui
 */
void TrajGenIfc::initGManaST()
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new ManaStateT(x2,y2,degToRad(theta2),v2,om2,t2);

  tabPosReachedX.push_back(((ManaStateT*)goalST_)->x());
  tabPosReachedY.push_back(((ManaStateT*)goalST_)->y());
  tabPosReachedV.push_back(((ManaStateT*)goalST_)->v());
  tabPosReachedT.push_back(((ManaStateT*)goalST_)->t());
}

// /**
//  *  \brief init the goal 3D point mass state-time from the gui
//  */
// void TrajGenIfc::initGPM3DST()
// {
//   if(goalST_ != NULL) delete goalST_;
//   goalST_ = new PM3DStateT(x2_3D,y2_3D,z2_3D,vx2_3D,vy2_3D,vz2_3D,t2);
// }

/**
 *  \brief init the goal state-time from the gui
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initGST(int i,int j,int k)
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
      this->initGCLST(i,j,k); break;
    case SPACESHIP_SYS : 
      this->initGSSST(i,j,k); break;
    case WHEELCHAIR_SYS : 
      this->initGWCST(i,j,k); break;
    case MANA_SYS : 
      this->initGManaST(i,j,k); break;
    }
}

/**
 *  \brief init the goal car-like state-time from the gui
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initGCLST(int i,int j,int k)
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new CLStateT();

  ((CLStateT*)goalST_)->x() = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  ((CLStateT*)goalST_)->y() = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    ((CLStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    ((CLStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  
  ((CLStateT*)goalST_)->ph() = kp2;
  ((CLStateT*)goalST_)->s() = v2;
  ((CLStateT*)goalST_)->t() = t2;

  tabPosReachedX.push_back(((CLStateT*)goalST_)->x());
  tabPosReachedY.push_back(((CLStateT*)goalST_)->y());
  tabPosReachedV.push_back(((CLStateT*)goalST_)->s());
  tabPosReachedT.push_back(((CLStateT*)goalST_)->t());
}

  /**
   *  \brief init the goal spaceship state-time from the gui
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void TrajGenIfc::initGSSST(int i,int j,int k)
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new SSStateT();

  ((SSStateT*)goalST_)->x() = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  ((SSStateT*)goalST_)->y() = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
  ((SSStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
  ((SSStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
			      - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  ((SSStateT*)goalST_)->sx() = vx2;
  ((SSStateT*)goalST_)->sy() = vy2;
  ((SSStateT*)goalST_)->om() = om2;
  ((SSStateT*)goalST_)->t() = t2;

  tabPosReachedX.push_back(((SSStateT*)goalST_)->x());
  tabPosReachedY.push_back(((SSStateT*)goalST_)->y());
  double sx = ((SSStateT*)goalST_)->sx();
  double sy = ((SSStateT*)goalST_)->sy();
  double v = sqrt(sx*sx+sy*sy);
  tabPosReachedV.push_back(v);
  tabPosReachedT.push_back(((SSStateT*)goalST_)->t());
}

/**
 *  \brief init the goal wheelchair state-time from the gui
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initGWCST(int i,int j,int k)
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new WCStateT();

  ((WCStateT*)goalST_)->x() = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  ((WCStateT*)goalST_)->y() = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    ((WCStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    ((WCStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
					 - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  
  ((WCStateT*)goalST_)->s() = v2;
  ((WCStateT*)goalST_)->om() = om2;
  ((WCStateT*)goalST_)->t() = t2;

  tabPosReachedX.push_back(((WCStateT*)goalST_)->x());
  tabPosReachedY.push_back(((WCStateT*)goalST_)->y());
  tabPosReachedV.push_back(((WCStateT*)goalST_)->s());
  tabPosReachedT.push_back(((WCStateT*)goalST_)->t());
}

/**
 *  \brief init the goal Mana state-time from the gui
 *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
 */
void TrajGenIfc::initGManaST(int i,int j,int k)
{
  if(goalST_ != NULL) delete goalST_;
  goalST_ = new ManaStateT();

  ((ManaStateT*)goalST_)->x() = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  ((ManaStateT*)goalST_)->y() = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
  if(trajByBranch == 1) 
    ((ManaStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
  else 
    ((ManaStateT*)goalST_)->th() = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
					 - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));
  
  ((ManaStateT*)goalST_)->v() = v2;
  ((ManaStateT*)goalST_)->w() = om2;
  ((ManaStateT*)goalST_)->t() = t2;

  tabPosReachedX.push_back(((ManaStateT*)goalST_)->x());
  tabPosReachedY.push_back(((ManaStateT*)goalST_)->y());
  tabPosReachedV.push_back(((ManaStateT*)goalST_)->v());
  tabPosReachedT.push_back(((ManaStateT*)goalST_)->t());
}

//   /**
//    *  \brief init the initial 3D point mass state-time from the gui
//    *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
//    */
//   void TrajGenIfc::initGPM3DST(int i,int j,int k)
// {
// }

/**
 *  \brief sample the trajectory generated to display it in a gnuplot window
 *  \param tg : the trajectory generator used
 */
void TrajGenIfc::sampleTraj(TrajGen* tg)
{
  switch(optSys)
    {
    case CARLIKE_SYS : this->sampleTrajCarLikeSys((CLTrajGen*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
//     case CARLIKE2_SYS : this->sampleTrajCarLikeSys2((CLTrajGen2*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
    case SPACESHIP_SYS : this->sampleTrajSpaceShipSys((SSTrajGen*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
//     case SPACESHIP2_SYS : this->sampleTrajSpaceShipSys2((SSTrajGen2*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
//     case DIFFDRIVE_SYS : this->sampleTrajDiffDriveSys((DDTrajGen*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
    case WHEELCHAIR_SYS : this->sampleTrajWheelChairSys((WCTrajGen*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
//     case WHEELCHAIR2_SYS : this->sampleTrajWheelChairSys2((WCTrajGen2*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
//     case PM3D_SYS : this->sampleTrajPM3DSys((PM3DTrajGen*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpZ,
// 					    tabPosTrajTmpT,tabPosTrajTmpVX,tabPosTrajTmpVY,tabPosTrajTmpVZ); break;
    case MANA_SYS : this->sampleTrajManaSys((ManaTrajGen*)tg,tabPosTrajTmpX,tabPosTrajTmpY,tabPosTrajTmpT,tabPosTrajTmpV); break;
    }
}

/**
 *  \brief initialize the trajectory sampling
 */
void TrajGenIfc::initSampleTraj()
{
  tabPosTrajX.clear();
  tabPosTrajY.clear();
  tabPosTrajZ.clear();
  tabPosTrajT.clear();
  tabPosTrajV.clear();
  tabPosTrajVX.clear();
  tabPosTrajVY.clear();
  tabPosTrajVZ.clear();
  if(tabPosTrajTmpX != NULL) delete tabPosTrajTmpX;
  tabPosTrajTmpX = new vector<double>;
  if(tabPosTrajTmpY != NULL) delete tabPosTrajTmpY;
  tabPosTrajTmpY = new vector<double>;
  if(tabPosTrajTmpZ != NULL) delete tabPosTrajTmpZ;
  tabPosTrajTmpZ = new vector<double>;
  if(tabPosTrajTmpT != NULL) delete tabPosTrajTmpT;
  tabPosTrajTmpT = new vector<double>;
  if(tabPosTrajTmpV != NULL) delete tabPosTrajTmpV;
  tabPosTrajTmpV = new vector<double>;
  if(tabPosTrajTmpVX != NULL) delete tabPosTrajTmpVX;
  tabPosTrajTmpVX = new vector<double>;
  if(tabPosTrajTmpVY != NULL) delete tabPosTrajTmpVY;
  tabPosTrajTmpVY = new vector<double>;
  if(tabPosTrajTmpVZ != NULL) delete tabPosTrajTmpVZ;
  tabPosTrajTmpVZ = new vector<double>;
}

/**
 *  \brief stock the trajectories sampled
 */
void TrajGenIfc::stockSampleTraj()
{
  if(tabPosTrajTmpX != NULL && tabPosTrajTmpX->size() != 0) tabPosTrajX.push_back(tabPosTrajTmpX);
  tabPosTrajTmpX = new vector<double>;
  if(tabPosTrajTmpY != NULL && tabPosTrajTmpY->size() != 0) tabPosTrajY.push_back(tabPosTrajTmpY);
  tabPosTrajTmpY = new vector<double>;
  if(tabPosTrajTmpZ != NULL && tabPosTrajTmpZ->size() != 0) tabPosTrajZ.push_back(tabPosTrajTmpZ);
  tabPosTrajTmpZ = new vector<double>;
  if(tabPosTrajTmpT != NULL && tabPosTrajTmpT->size() != 0) tabPosTrajT.push_back(tabPosTrajTmpT);
  tabPosTrajTmpT = new vector<double>;
  if(tabPosTrajTmpV != NULL && tabPosTrajTmpV->size() != 0) tabPosTrajV.push_back(tabPosTrajTmpV);
  tabPosTrajTmpV = new vector<double>;
  if(tabPosTrajTmpVX != NULL && tabPosTrajTmpVX->size() != 0) tabPosTrajVX.push_back(tabPosTrajTmpVX);
  tabPosTrajTmpVX = new vector<double>;
  if(tabPosTrajTmpVY != NULL && tabPosTrajTmpVY->size() != 0) tabPosTrajVY.push_back(tabPosTrajTmpVY);
  tabPosTrajTmpVY = new vector<double>;
  if(tabPosTrajTmpVZ != NULL && tabPosTrajTmpVZ->size() != 0) tabPosTrajVZ.push_back(tabPosTrajTmpVZ);
  tabPosTrajTmpVZ = new vector<double>;
}

/**
 *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
 *  \param tg : the trajectory generator used
 *  \param tabX : the table to stock X values
 *  \param tabY : the table to stock Y values
 *  \param tabT : the table to stock T values
 *  \param tabV : the table to stock V values
 */
void TrajGenIfc::sampleTrajCarLikeSys(CLTrajGen* tg,
				      vector<double>* tabX,
				      vector<double>* tabY,
				      vector<double>* tabT,
				      vector<double>* tabV )
{
  tg->SaveTrajectory(100);

  for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
  for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
  for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
  for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
  for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
  for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
  for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[j]);
  for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[tg->velV->size() - j -1]);
}

// /**
//  *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
//  *  \param tg : the trajectory generator used
//  *  \param tabX : the table to stock X values
//  *  \param tabY : the table to stock Y values
//  *  \param tabT : the table to stock T values
//  *  \param tabV : the table to stock V values
//  */
// void TrajGenIfc::sampleTrajCarLikeSys2(CLTrajGen2* tg,
// 				       vector<double>* tabX,
// 				       vector<double>* tabY,
// 				       vector<double>* tabT,
// 				       vector<double>* tabV )
// {
//   tg->SaveTrajectory(100);
  
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
//   for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[j]);
//   for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[tg->velV->size() - j -1]);
// }

/**
 *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
 *  \param tg : the trajectory generator used
 */
void TrajGenIfc::sampleTrajSpaceShipSys(SSTrajGen* tg,
				      vector<double>* tabX,
				      vector<double>* tabY,
				      vector<double>* tabT,
				      vector<double>* tabV)
{
  tg->SaveTrajectory(100);
  
  for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
  for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
  for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
  for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
  for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
  for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
  for(unsigned int j=0; j<tg->velVX->size(); j++) 
    {
      double cur_vx = (*(tg->velVX))[j];
      double cur_vy = (*(tg->velVY))[j];
      double cur_v = sqrt(cur_vx*cur_vx + cur_vy*cur_vy);
      tabV->push_back(cur_v);
    }
  for(unsigned int j=0; j<tg->velVX->size(); j++) 
    {
      double cur_vx = (*(tg->velVX))[tg->velVX->size() - j -1];
      double cur_vy = (*(tg->velVY))[tg->velVY->size() - j -1];
      double cur_v = sqrt(cur_vx*cur_vx + cur_vy*cur_vy);
      tabV->push_back(cur_v);
    }
}

// /**
//  *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
//  *  \param tg : the trajectory generator used
//  *  \param tabX : the table to stock X values
//  *  \param tabY : the table to stock Y values
//  *  \param tabT : the table to stock T values
//  *  \param tabV : the table to stock V values
//  */
// void TrajGenIfc::sampleTrajSpaceShipSys2(SSTrajGen2* tg,
// 					 vector<double>* tabX,
// 					 vector<double>* tabY,
// 					 vector<double>* tabT,
// 					 vector<double>* tabV )
// {
//   tg->SaveTrajectory(100);
  
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
//   for(unsigned int j=0; j<tg->velVX->size(); j++) 
//     {
//       double cur_vx = (*(tg->velVX))[j];
//       double cur_vy = (*(tg->velVY))[j];
//       double cur_v = sqrt(cur_vx*cur_vx + cur_vy*cur_vy);
//       tabV->push_back(cur_v);
//     }
//   for(unsigned int j=0; j<tg->velVX->size(); j++) 
//     {
//       double cur_vx = (*(tg->velVX))[tg->velVX->size() - j -1];
//       double cur_vy = (*(tg->velVY))[tg->velVY->size() - j -1];
//       double cur_v = sqrt(cur_vx*cur_vx + cur_vy*cur_vy);
//       tabV->push_back(cur_v);
//     }
// }


// /**
//  *  \brief sample the trajectory generated (for a differential drive system) to display it in a gnuplot window
//  *  \param tg : the trajectory generator used
//  *  \param tabX : the table to stock X values
//  *  \param tabY : the table to stock Y values
//  *  \param tabT : the table to stock T values
//  *  \param tabV : the table to stock V values
//  */
// void TrajGenIfc::sampleTrajDiffDriveSys(DDTrajGen* tg,
// 					vector<double>* tabX,
// 					vector<double>* tabY,
// 					vector<double>* tabT,
// 					vector<double>* tabV )
// {
//   tg->SaveTrajectory(100);
  
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
//   for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[j]);
//   for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[tg->velV->size() - j -1]);
// }

  /**
   *  \brief sample the trajectory generated (for a differential drive system) to display it in a gnuplot window
   *  \param tg : the trajectory generator used
   *  \param tabX : the table to stock X values
   *  \param tabY : the table to stock Y values
   *  \param tabT : the table to stock T values
   *  \param tabV : the table to stock V values
   */
void TrajGenIfc::sampleTrajWheelChairSys(WCTrajGen* tg,
					 vector<double>* tabX,
					 vector<double>* tabY,
					 vector<double>* tabT,
					 vector<double>* tabV )
{
  tg->SaveTrajectory(100);
  
  for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
  for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
  for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
  for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
  for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
  for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
  for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[j]);
  for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[tg->velV->size() - j -1]);
}

/**
 *  \brief sample the trajectory generated (for a mana system) to display it in a gnuplot window
 *  \param tg : the trajectory generator used
 *  \param tabX : the table to stock X values
 *  \param tabY : the table to stock Y values
 *  \param tabT : the table to stock T values
 *  \param tabV : the table to stock V values
 */
void TrajGenIfc::sampleTrajManaSys(ManaTrajGen* tg,
				   vector<double>* tabX,
				   vector<double>* tabY,
				   vector<double>* tabT,
				   vector<double>* tabV )
{
  // sample the trajectory
  int N=100;
  ManaSTraj* sampledTraj = ((ManaTrajGen*)tg)->getSTraj(N);
  
  for(int j=0; j<N; j++)
    {
      // get back the current state-time
      ManaStateT statet = sampledTraj->at(j);

      // put its components in the output tables
      tabX->push_back(statet.x());
      tabY->push_back(statet.y());
      tabT->push_back(statet.t());
      tabV->push_back(statet.v());
    }

  for(int j=N-1; j>=0; j--)
    {
      // get back the current state-time
      ManaStateT statet = sampledTraj->at(j);

      // put its components in the output tables
      tabX->push_back(statet.x());
      tabY->push_back(statet.y());
      tabT->push_back(statet.t());
      tabV->push_back(statet.v());
    }
}


// /**
//  *  \brief sample the trajectory generated (for a differential drive system) to display it in a gnuplot window
//  *  \param tg : the trajectory generator used
//  *  \param tabX : the table to stock X values
//  *  \param tabY : the table to stock Y values
//  *  \param tabT : the table to stock T values
//  *  \param tabV : the table to stock V values
//  */
// void TrajGenIfc::sampleTrajWheelChairSys2(WCTrajGen2* tg,
// 					  vector<double>* tabX,
// 					  vector<double>* tabY,
// 					  vector<double>* tabT,
// 					  vector<double>* tabV )
// {
//   tg->SaveTrajectory(100);
  
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
//   for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[j]);
//   for(unsigned int j=0; j<tg->velV->size(); j++) tabV->push_back((*(tg->velV))[tg->velV->size() - j -1]);
// }

// /**
//  *  \brief sample the trajectory generated (for a 3D point mass system) to display it in a gnuplot window
//  *  \param tg : the trajectory generator used
//  *  \param tabX : the table to stock X values
//  *  \param tabY : the table to stock Y values
//  *  \param tabZ : the table to stock Z values
//  *  \param tabT : the table to stock T values
//  *  \param tabVX : the table to stock VX values
//  *  \param tabVY : the table to stock VY values
//  *  \param tabVZ : the table to stock VZ values
//  */
// void TrajGenIfc::sampleTrajPM3DSys(PM3DTrajGen* tg,
// 				   vector<double>* tabX,
// 				   vector<double>* tabY,
// 				   vector<double>* tabZ,
// 				   vector<double>* tabT,
// 				   vector<double>* tabVX,
// 				   vector<double>* tabVY,
// 				   vector<double>* tabVZ )
// {
//   tg->SaveTrajectory(100);
  
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[j]);
//   for(unsigned int j=0; j<tg->posX->size(); j++) tabX->push_back((*(tg->posX))[tg->posX->size() - j -1]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[j]);
//   for(unsigned int j=0; j<tg->posY->size(); j++) tabY->push_back((*(tg->posY))[tg->posY->size() - j -1]);
//   for(unsigned int j=0; j<tg->posZ->size(); j++) tabZ->push_back((*(tg->posZ))[j]);
//   for(unsigned int j=0; j<tg->posZ->size(); j++) tabZ->push_back((*(tg->posZ))[tg->posZ->size() - j -1]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[j]);
//   for(unsigned int j=0; j<tg->timeT->size(); j++) tabT->push_back((*(tg->timeT))[tg->timeT->size() - j -1]);
//   for(unsigned int j=0; j<tg->velVX->size(); j++) tabVX->push_back((*(tg->velVX))[j]);
//   for(unsigned int j=0; j<tg->velVX->size(); j++) tabVX->push_back((*(tg->velVX))[tg->velVX->size() - j -1]);
//   for(unsigned int j=0; j<tg->velVY->size(); j++) tabVY->push_back((*(tg->velVY))[j]);
//   for(unsigned int j=0; j<tg->velVY->size(); j++) tabVY->push_back((*(tg->velVY))[tg->velVY->size() - j -1]);
//   for(unsigned int j=0; j<tg->velVZ->size(); j++) tabVZ->push_back((*(tg->velVZ))[j]);
//   for(unsigned int j=0; j<tg->velVZ->size(); j++) tabVZ->push_back((*(tg->velVZ))[tg->velVZ->size() - j -1]);
// }

/**
 *  \brief empty the trajectory vectors
 */
void TrajGenIfc::emptyTrajVec()
{
  tabPosTrajX.clear();
  tabPosTrajY.clear();
  tabPosTrajZ.clear();
  tabPosTrajT.clear();
  tabPosTrajV.clear();
  tabPosTrajVX.clear();
  tabPosTrajVY.clear();
  tabPosTrajVZ.clear();

  tabPosReachedX.clear();
  tabPosReachedY.clear();
  tabPosReachedT.clear();
  tabPosReachedV.clear();
  tabPosReachedSt.clear();
}

/**
 *  \brief test the reachability of the final state
 *  \return true if the input state is reachable, false else
 */
bool TrajGenIfc::testReachty(const StateTime& s)
{
  cout << "deb TrajGenIfc::testReachty" << endl;
  cout << "s : "; s.display(); cout << endl;

  switch(optSys)
    {
//     case SPACESHIP_SYS : 
//     case SPACESHIP2_SYS : 
//       cout << "TrajGenIfc::testReachty : Spaceship sslattice not implemented yet!" << endl;
//       return true;
//     case DIFFDRIVE_SYS : 
//       cout << "TrajGenIfc::testReachty : Diffdrive sslattice not implemented yet!" << endl;
//       return true;

//     case WHEELCHAIR_SYS : 
//     case WHEELCHAIR2_SYS : 
//       cout << "TrajGenIfc::testReachty : Wheelchair sslattice not implemented yet!" << endl;
//       return true;

    case CARLIKE_SYS : 
//     case CARLIKE2_SYS : 
      {
	if(lattice_ == NULL)
	  {
	    FXMessageBox::error(this,MBOX_OK,"Error while testing the reachability","The state space lattice has not been computed.");
	    return false;
	  }
	
	return lattice_->isReachable(s);
	break;
      }
    default : 
      return false;
    }
}

/**
 *  \brief test the reachability of the final car-like state
 *  \return true if the input state is reachable, false else
 */
bool TrajGenIfc::testCLReachty(const StateTime& s)
{
  return true;
}

/**
 *  \brief test the reachability of the final wheelchair state
 *  \return true if the input state is reachable, false else
 */
bool TrajGenIfc::testWCReachty(const StateTime& s)
{
  return true;
}

/**
 *  \brief test the reachability of the final Mana state
 *  \return true if the input state is reachable, false else
 */
bool TrajGenIfc::testManaReachty(const StateTime& s)
{
  return true;
}

/**
 *  \brief draw a 2D reachable star
 */
void TrajGenIfc::drawRStar2D()
{
  // init the trajectory sampling to display them
  this->initSampleTraj();

  // init the list of trajectories to save if the saving process is requested
  if(saveRStar) { this->clearListParamTraj(); }

  if(map == NULL && debugLevel != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while drawing the RStar","No map has been loaded");
      return;
    }
  
  // statistical values
  double nbTrajTot  = 0;     // total number of trajectories
  double nbTrajCplt = 0;     // number of completed trajectories
  double pcCplt = 0;      // percentage of completed trajectories

  // init the timer
  struct timeval tv1,tv2;
  gettimeofday(&tv1,NULL);

  if(debugLevel >= 1)
    {
      this->emptyTrajVec();
    }

  double totFinalNbSteps = 0;

//   cout << "nbBranches : " << nbBranches << endl;
//   cout << "trajByBranch : " << trajByBranch << endl;
//   cout << "nbItvDistRS : " << nbItvDistRS << endl;

  // init the robotic system
  initSys();

  // init the trajectory generation process
  initTrajGen();

  // init the initial state-time
  initIST();

  for(int i=0; i<nbBranches; i++)
    {
      for(int j=0; j<trajByBranch; j++)
	{
	  for(int k=0; k<nbItvDistRS; k++)
	    {
	      nbTrajTot++;

	      // init the final state-time
	      initGST(i,j,k);

	      // init the time constraint of the trajectory generator
	      int resITC = initTimeCstt();
	      if(resITC != 0)
		{
		  cout << "AAAAAAAAAAAAAAAAAAAAAAAARRRRRRRRRRRRRGGGGGGGGGGGGGHHHHHHHHHH" << endl;
		  return;
		}

	      //Compute trajectory
	      int res;
	      if(genTrajMethod == 0) res = trajGenerator->ComputeTrajectory(*initST_,*goalST_,NULL);
	      else res = trajGenerator->genTraj(*initST_,*goalST_);
	      
	      if(res == 0)
		{
		  // case where the trajectory generation succeed
		  nbTrajCplt++;
		  totFinalNbSteps += trajGenerator->finalNbSteps_;
		  
		  if(debugLevel >= 1)
		    {
		      // trace the trajectory on the map
		      trajGenerator->traceOn(map,300);
		      
		      // compute a sample trajectory to display in a gnuplot viewer
		      this->sampleTraj(trajGenerator);
		    }

		  if(saveRStar) { this->addParamTrajToSave(); }
		}
	      else if(res == -1)
		{
		  // case where the trajectory generation didn't reach the goal : 
		  // trace the better trajectory found during the generation

		  // trace the trajectory on the map
		  if(drawUncompleteTraj) 
		    {
		      if(debugLevel >= 1)
			{
			  // trace the trajectory on the map
			  trajGenerator->traceOn(map,300);

			  // compute a sample trajectory to display in a gnuplot viewer
			  this->sampleTraj(trajGenerator);
			}

		      if(saveRStar) { this->addParamTrajToSave(); }
		    }
		}
	      else // (res == -2)
		{
		  // nothing to do
		}
	    }
	}
//       if(debugLevel >= 1)
// 	{
// 	  tabPosTraj.push_back(subTabPosTrajX);
// 	  tabPosTraj.push_back(subTabPosTrajY);
// 	  tabPosTraj.push_back(subTabPosTrajT);
// 	  tabPosTraj.push_back(subTabPosTrajV);
// 	}

      // stock the current branch
      if(debugLevel >= 1) this->stockSampleTraj();
    }
  
  // stop the timer
  gettimeofday(&tv2,NULL);
  
  // compute the time passed
  int elapsedTs = tv2.tv_sec - tv1.tv_sec;
  int elapsedTus = tv2.tv_usec - tv1.tv_usec;
  double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
  cout << "Elapsed time : " << elapsedT << endl; 

  // compute the average number of steps
  double nbTraj = (nbBranches*trajByBranch*nbItvDistRS);
  //double avgNbSteps = totFinalNbSteps / nbTraj;
  double avgNbSteps = ((double)totFinalNbSteps) / ((double)nbTrajCplt);

  // compute the average time needed by trajectory
  double avgTime = elapsedT / nbTraj;
  double avgTPS = 1.0 / avgTime;

  // saving the map
  map->save("/tmp/ResultMap_traj.pgm");
  
  // load the map in the canvas
  this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));

  pcCplt = ((double)nbTrajCplt) / ((double)nbTrajTot) * 100.0;
  
  FXString str("Reachable Star computed in ");
  str += FXStringVal(elapsedT,2,FALSE);
  str += FXString(" sec.  Avg nb steps : ");
  str += FXStringVal(avgNbSteps,2,FALSE);
  str += FXString("  Avg time by traj : ");
  str += FXStringVal(avgTime,4,FALSE);
  str += FXString("  Avg Traj/sec : ");
  str += FXStringVal(avgTPS,2,FALSE);
  str += FXString("  Completed Traj : ");
  str += FXStringVal(nbTrajCplt,0,FALSE);
  str += FXString("/");
  str += FXStringVal(nbTrajTot,0,FALSE);
  str += FXString(" (");
  str += FXStringVal(pcCplt,2,FALSE);
  str += FXString(" %)");
  msgBox->setText(str);

  // save the trajectory parameterizations into an output file if requested
  if(saveRStar) { this->saveListParamTraj(IG1stParamsName.text()); }

  return;
}

/**
 *  \brief draw a 3D reachable star
 */
void TrajGenIfc::drawRStar3D()
{
}

/**
 *  \brief draw the set of trajectories computed in a gnuplot viewer
 */
void TrajGenIfc::drawGP()
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
    case SPACESHIP_SYS : 
    case WHEELCHAIR_SYS : 
    case MANA_SYS : 
      this->drawGP_DM2D(); break;
//     case PM3D_SYS : 
//       this->drawGP_PM3D(); break;
//     case MANIP_ARM1_SYS : 
//       this->drawGP_MA(); break;
    }
}

/**
 *  \brief draw the set of trajectories computed for a 2D directed mobile system in a gnuplot viewer
 */
void TrajGenIfc::drawGP_DM2D()
{
  vector<double> LC_X;
  vector<double> LC_Y;
  vector<double> RC_X;
  vector<double> RC_Y;
  if(optSys == CARLIKE_SYS && dispCurvCirc)
    {
      // compute the minimal curvature circles

      // compute the radius of the minimal curvature circles
      double L = ((CarLikeSys*)sys_)->getLength();
      double phiMax = ((CarLikeSys*)sys_)->getMaxSteering();
      double rad = L / tan(phiMax);

      // compute the centers of these circles
      CLStateT ist(*((CLStateT*)initST_));
      PositionXY leftCenter(ist.x() - rad*sin(ist.th()),
			    ist.y() + rad*cos(ist.th()));
      PositionXY rightCenter(ist.x() + rad*sin(ist.th()),
			     ist.y() - rad*cos(ist.th()));

      vector<double> xIGpos, yIGpos;
      xIGpos.push_back(leftCenter.x());
      xIGpos.push_back(rightCenter.x());
      yIGpos.push_back(leftCenter.y());
      yIGpos.push_back(rightCenter.y());

      // compute both circles
      for(int i=0; i<=100; i++)
	{
	  double X = leftCenter.x() + rad*cos(2.0*M_PI*0.01*i);
	  double Y = leftCenter.y() + rad*sin(2.0*M_PI*0.01*i);
	  LC_X.push_back(X);
	  LC_Y.push_back(Y);
	}

      for(int i=0; i<=100; i++)
	{
	  double X = rightCenter.x() + rad*cos(2.0*M_PI*0.01*i);
	  double Y = rightCenter.y() + rad*sin(2.0*M_PI*0.01*i);
	  RC_X.push_back(X);
	  RC_Y.push_back(Y);
	}
      
    }
  
  Gnuplot gPosition;
  gPosition.reset_plot();

  cout << "tabPosReachedX.size() : " << tabPosReachedX.size() << endl;
  cout << "tabPosReachedY.size() : " << tabPosReachedY.size() << endl;
  cout << "tabPosReachedT.size() : " << tabPosReachedT.size() << endl;
  cout << "tabPosReachedV.size() : " << tabPosReachedV.size() << endl;

  if(drawTargetPos) gPosition.set_style("points").plot_xy(tabPosReachedX,tabPosReachedY,"");
  if(optSys == CARLIKE_SYS && dispCurvCirc) 
    {
      gPosition.set_style("lines").plot_xy(LC_X,LC_Y,"");
      gPosition.set_style("lines").plot_xy(RC_X,RC_Y,"");
    }
  
  for(unsigned int i=0; i<tabPosTrajX.size(); i++)
    {      
      gPosition.cmd("set linestyle 1");
      gPosition.set_style("lines").plot_xy(*(tabPosTrajX[i]),*(tabPosTrajY[i]),"");
    }

  Gnuplot gPosition2;
  gPosition2.reset_plot();
  
  if(drawTargetPos) gPosition2.set_style("points").plot_xyz(tabPosReachedX,tabPosReachedY,tabPosReachedT,"");
  
  for(unsigned int i=0; i<tabPosTrajX.size(); i++)
    {
      gPosition2.cmd("set linestyle 1");
      gPosition2.set_style("lines").plot_xyz(*(tabPosTrajX[i]),*(tabPosTrajY[i]),*(tabPosTrajT[i]),"");
    }

  Gnuplot gPosition3;
  gPosition3.reset_plot();
  
  if(drawTargetPos) gPosition3.set_style("points").plot_xyz(tabPosReachedX,tabPosReachedY,tabPosReachedV,"");
  
  for(unsigned int i=0; i<tabPosTrajX.size(); i++)
    {      
      gPosition3.cmd("set linestyle 1");
      gPosition3.set_style("lines").plot_xyz(*(tabPosTrajX[i]),*(tabPosTrajY[i]),*(tabPosTrajV[i]),"");
    }
  
  wait_for_key();
  return;
}

/**
 *  \brief draw the set of trajectories computed for a 2D point mass system in a gnuplot viewer
 */
void TrajGenIfc::drawGP_PM2D()
{
}

/**
 *  \brief draw the set of trajectories computed for a 3D point mass system in a gnuplot viewer
 */
void TrajGenIfc::drawGP_PM3D()
{
  Gnuplot gPosition2;
  gPosition2.reset_plot();
  
  if(drawTargetPos) gPosition2.set_style("points").plot_xyz(tabPosReachedX,tabPosReachedY,tabPosReachedZ,"");
  
  for(unsigned int i=0; i<tabPosTrajX.size(); i++)
    {
      gPosition2.cmd("set linestyle 1");
      gPosition2.set_style("lines").plot_xyz(*(tabPosTrajX[i]),*(tabPosTrajY[i]),*(tabPosTrajZ[i]),"");
    }

  Gnuplot gPosition3;
  gPosition3.reset_plot();
  
  if(drawTargetPos) gPosition3.set_style("points").plot_xyz(tabPosReachedVX,tabPosReachedVY,tabPosReachedVZ,"");
  
  for(unsigned int i=0; i<tabPosTrajVX.size(); i++)
    {      
      gPosition3.cmd("set linestyle 1");
      gPosition3.set_style("lines").plot_xyz(*(tabPosTrajVX[i]),*(tabPosTrajVY[i]),*(tabPosTrajVZ[i]),"");
    }
  
  wait_for_key();
  return;
}

/**
 *  \brief draw the set of trajectories computed in a gnuplot viewer
 */
void TrajGenIfc::drawGP_MA()
{
  cout << "TrajGenIfc::drawGP_MA : Not implemented yet!" << endl;
  FXMessageBox::error(this,MBOX_OK,"TrajGenIfc::drawGP_MA","Not implemented yet!");
}

/**
 *  \brief initialize the IG learning process from the gui
 *  \return 0 if the process succeed, -1 else
 */
int TrajGenIfc::initIG()
{
  // initialize the system considered from the gui
  initSys();

  // reinit the ig learning module
  if(ig_ != NULL) delete ig_;

  switch(optSys)
    {
    case CARLIKE_SYS : 
      ig_ = new CLInitialGuess((CarLikeSys*)sys_,IG_tmin,IG_tmax,CL_samp_t,CL_samp_vi,CL_samp_vg,CL_samp_dx,CL_samp_dy,CL_samp_dth);
      break;
    case WHEELCHAIR_SYS : 
      ig_ = new WCInitialGuess((WheelChairSys*)sys_,IG_tmin,IG_tmax,WC_samp_t,WC_samp_vi,WC_samp_vg,WC_samp_dx,WC_samp_dy,WC_samp_dth);
      break;
    case SPACESHIP_SYS : 
//       ig_ = new SSInitialGuess((SpaceShipSys*)sys_);
//       break;
    case MANA_SYS : 
      FXMessageBox::error(this,MBOX_OK,"TrajGenIfc::initIG","Not implemented yet for Mana system.");      
      return -1;
    default : 
      cout << "TrajGenIfc::initIG : Only implemented yet for CL and WC systems." << endl;
      FXMessageBox::error(this,MBOX_OK,"TrajGenIfc::initIG","Only implemented yet for CL and WC systems.");      
      return -1;
    }

  ig_->setDebug(IG_debugMode);
  ig_->setMinNNTSR(IG_minNNTSR);
  ig_->setMaxTGResidual(maxResidual);
  ig_->setMaxTGSteps(maxNbStep);

  switch(IG_1stParams)
    {
    case SC_FIRSTIG : 
      {
	switch(optSys)
	  {
	  case CARLIKE_SYS : 
	    ((CLInitialGuess*)ig_)->use_SCP_FIGMode(CL_sampAcc,CL_sampStSpd);
// 	    ((CLInitialGuess*)ig_)->initSampS(CL_samp_t,CL_samp_vi,CL_samp_vg,CL_samp_dx,CL_samp_dy,CL_samp_dth);
	    break;
	  case WHEELCHAIR_SYS : 
	    ((WCInitialGuess*)ig_)->use_SC_FIGMode(WC_sampAcc,WC_sampAAcc);
// 	    ((WCInitialGuess*)ig_)->initSampS(WC_samp_t,WC_samp_vi,WC_samp_vg,WC_samp_dx,WC_samp_dy,WC_samp_dth);
	    break;
	  case SPACESHIP_SYS : 
// 	    ((SSInitialGuess*)ig_)->use_SC_FIGMode(SS_sampAcc,SS_sampAAcc);
// 	    break;
	  default : 
	    return -1;
	    break;
	  }
	break;
      }
    case LOADFROMFILE_FIRSTIG : 
      {
	ig_->use_LFF_FIGMode(IG1stParamsName.text());
	break;
      }
    default : 
      cout << "TrajGenIfc::initIG : Only implemented yet for CL and WC systems." << endl;
      FXMessageBox::error(this,MBOX_OK,"TrajGenIfc::initIG","Only implemented yet for CL and WC systems.");      
      return -1;
    }

  return 0;
}

/**
 *  \brief learn the initial guess of parameters
 */
void TrajGenIfc::learnIG()
{
  // initialize the ig module from the interface
  initIG();

  // check if the ig learning process has been correctly initialized
  if(ig_ == NULL) return;

  // launch the learning process
  ig_->learnIG();
}

/**
 *  \brief clear the list of parameterized trajectories to save
 */
void TrajGenIfc::clearListParamTraj()
{
//   std::list<N2R*>::iterator it = listParamTraj.begin();
//   while(it != listParamTraj.end())
//     {
//       /////////////////////////////////////////////////////////////////////////// Seg fault if not commented... Why?
//       if(*it != NULL) delete *it;
//       listParamTraj.pop_front();
//     }
  listParamTraj.clear();
}

/**
 *  \brief save the list of parameterized trajectories
 */
void TrajGenIfc::saveListParamTraj(const char* filename)
{
  int res = N2R::saveListN2R(&listParamTraj,filename);

  if(res != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while saving the RStar trajectories","The saving process failed");
    }
}

/**
 *  \brief add a new parametric trajectory to save to the current list
 */
void TrajGenIfc::addParamTrajToSave()
{
  switch(optSys)
    {
    case CARLIKE_SYS : 
      {
	// get back the initial state-time, final state-time and control parameterization
	CLStateT initST = ((CLTrajGen*)trajGenerator)->getCLStateTReachedAt(0);
	CLStateT finalST = ((CLTrajGen*)trajGenerator)->getFinalCLStateTReached();
	double params[7];
	trajGenerator->finalParams(params);

	// compute the relative coordinates between both state-times
	CLStateT::relativeStateT(initST,finalST);

	// create a node with these features
	CL_N2R* node = new CL_N2R(initST,finalST,params);

	// add it to the list of trajectories to save
	listParamTraj.push_back(node);
	break;
      }
    case WHEELCHAIR_SYS : 
      {
	// get back the initial state-time, final state-time and control parameterization
	WCStateT initST = ((WCTrajGen*)trajGenerator)->getWCStateTReachedAt(0);
	WCStateT finalST = ((WCTrajGen*)trajGenerator)->getFinalWCStateTReached();
	double params[7];
	trajGenerator->finalParams(params);

	// compute the relative coordinates between both state-times
	WCStateT::relativeStateT(initST,finalST);

	// create a node with these features
	WC_N2R* node = new WC_N2R(initST,finalST,params);

	// add it to the list of trajectories to save
	listParamTraj.push_back(node);
	break;
      }
    case SPACESHIP_SYS : 
//       {
// 	// get back the initial state-time, final state-time and control parameterization
// 	SSStateT initST = ((SSTrajGen*)trajGenerator)->getSSStateTReachedAt(0);
// 	SSStateT finalST = ((SSTrajGen*)trajGenerator)->getFinalSSStateTReached();
// 	double params[7];
// 	trajGenerator->finalParams(params);

// 	// compute the relative coordinates between both state-times
// 	SSStateT::relativeStateT(initST,finalST);

// 	// create a node with these features
// 	SS_N2R* node = new SS_N2R(initST,finalST,params);

// 	// add it to the list of trajectories to save
// 	listParamTraj.push_back(node);
// 	break;
//       }

    case MANA_SYS : 

    default : 
      cout << "TrajGenIfc::addParamTrajToSave : Only implemented for car-like, spaceship and wheelchair yet!" << endl;
      return;
    }
}

/**
 *  \brief init the car-like system from an input configuration file
 *  \param filename : the input filename
 */
void TrajGenIfc::initCLSysFromFile(const char* filename)
{
  po::options_description config_files_clsys("Configuration Files"); // Configuration files
  config_files_clsys.add_options() 
    ("CarLikeSys.cl_length", po::value<double>(&rb_L), "car-like length")
    ("CarLikeSys.cl_vmax", po::value<double>(&rb_VMAX), "car-like maximal velocity")
    ("CarLikeSys.cl_amax", po::value<double>(&rb_AMAX), "car-like maximal acceleration")
    ("CarLikeSys.cl_samax", po::value<double>(&rb_KMAX), "car-like maximal steering angle")
    ("CarLikeSys.cl_ssmax", po::value<double>(&rb_OMAX), "car-like maximal steering speed")
    ("CarLikeSys.cl_cmb", po::value<bool>((bool*)(&canMoveBackCL)), "true if the car-like can move backward")

    ("CLTrajGen.cl_use_vmax", po::value<bool>((bool*)(&useVelCst)), "use car-like maximal velocity during trajectory generation")
    ("CLTrajGen.cl_use_amax", po::value<bool>((bool*)(&useAccCst)), "use car-like maximal acceleration during trajectory generation")
    ("CLTrajGen.cl_use_samax", po::value<bool>((bool*)(&useAVelCst)), "use car-like maximal steering angle during trajectory generation")
    ("CLTrajGen.cl_use_ssmax", po::value<bool>((bool*)(&useAAccCst)), "use car-like maximal steering speed during trajectory generation")
    ;

  po::variables_map vm_config_clsys;
  fs::path cfg_path_clsys(fs::initial_path());
  cfg_path_clsys = fs::system_complete(fs::path(filename,fs::native));
  string cfg_string_clsys = cfg_path_clsys.native_file_string();
  if ( !fs::exists( cfg_path_clsys ) )
    {
      cout << "Unable to load the carlike configuration. Configuration file not found" << endl;
      return;
    }
  // Open configuration file
  ifstream ifs_clsys((const char *)cfg_string_clsys.c_str());
  store(parse_config_file(ifs_clsys, config_files_clsys), vm_config_clsys);
  notify(vm_config_clsys);
}

/**
 *  \brief init the wheelchair system from an input configuration file
 *  \param filename : the input filename
 */
void TrajGenIfc::initWCSysFromFile(const char* filename)
{
  po::options_description config_files_wcsys("Configuration Files"); // Configuration files
  config_files_wcsys.add_options() 
    ("WheelChairSys.wc_vmax", po::value<double>(&WC_VelMax), "wheelchair maximal velocity")
    ("WheelChairSys.wc_amax", po::value<double>(&WC_AccMax), "wheelchair maximal acceleration")
    ("WheelChairSys.wc_avmax", po::value<double>(&WC_AVelMax), "wheelchair maximal angular velocity")
    ("WheelChairSys.wc_aamax", po::value<double>(&WC_AAccMax), "wheelchair maximal angular acceleration")
    ("WheelChairSys.wc_cmb", po::value<bool>((bool*)(&canMoveBackWC)), "true if the wheelchair can move backward")

    ("WCTrajGen.wc_use_vmax", po::value<bool>((bool*)(&useVelCstWC)), "use wheelchair maximal velocity during trajectory generation")
    ("WCTrajGen.wc_use_amax", po::value<bool>((bool*)(&useAccCstWC)), "use wheelchair maximal acceleration during trajectory generation")
    ("WCTrajGen.wc_use_avmax", po::value<bool>((bool*)(&useAVelCstWC)), "use wheelchair maximal angular velocity during trajectory generation")
    ("WCTrajGen.wc_use_aamax", po::value<bool>((bool*)(&useAAccCstWC)), "use wheelchair maximal angular acceleration during trajectory generation")
    ;

  po::variables_map vm_config_wcsys;
  fs::path cfg_path_wcsys(fs::initial_path());
  cfg_path_wcsys = fs::system_complete(fs::path(filename,fs::native));
  string cfg_string_wcsys = cfg_path_wcsys.native_file_string();
  if ( !fs::exists( cfg_path_wcsys ) )
    {
      cout << "Unable to load the wheelchair configuration. Configuration file not found" << endl;
      return;
    }
  // Open configuration file
  ifstream ifs_wcsys((const char *)cfg_string_wcsys.c_str());
  store(parse_config_file(ifs_wcsys, config_files_wcsys), vm_config_wcsys);
  notify(vm_config_wcsys);

//   po::options_description config_files_wcig("Configuration Files"); // Configuration files
// //   config_files_wcig.add_options() 
// //     ("WCInitialGuess.wc_fixed_xy", po::value<bool>(&WC_VelMax), "fix the wheelchair final position")
// //     ;

//   po::variables_map vm_config_wcig;
//   fs::path cfg_path_wcig(fs::initial_path());
//   cfg_path_wcig = fs::system_complete(fs::path("../cfg/wc_ig.cfg",fs::native));
//   string cfg_string_wcig = cfg_path_wcig.native_file_string();
//   if ( !fs::exists( cfg_path_wcig ) )
//     {
//       FXMessageBox::error(this,MBOX_OK,"Unable to load the wheelchair configuration","Configuration file not found");
//       return;
//     }
//   // Open configuration file
//   ifstream ifs_wcig((const char *)cfg_string_wcig.c_str());
//   store(parse_config_file(ifs_wcig, config_files_wcig), vm_config_wcig);
//   notify(vm_config_wcig);
}

/**
 *  \brief init the spaceship system from an input configuration file
 *  \param filename : the input filename
 */
void TrajGenIfc::initSSSysFromFile(const char* filename)
{
  po::options_description config_files_sssys("Configuration Files"); // Configuration files
  config_files_sssys.add_options() 
    ("SpaceShipSys.ss_vmax", po::value<double>(&SS_VelMax), "spaceship maximal velocity")
    ("SpaceShipSys.ss_amax", po::value<double>(&SS_AccMax), "spaceship maximal acceleration")
    ("SpaceShipSys.ss_avmax", po::value<double>(&SS_AVelMax), "spaceship maximal angular velocity")
    ("SpaceShipSys.ss_aamax", po::value<double>(&SS_AAccMax), "spaceship maximal angular acceleration")

//     ("SSTrajGen.ss_use_vmax", po::value<bool>((bool*)(&useVelCstSS)), "use spaceship maximal velocity during trajectory generation")
//     ("SSTrajGen.ss_use_amax", po::value<bool>((bool*)(&useAccCstSS)), "use spaceship maximal acceleration during trajectory generation")
//     ("SSTrajGen.ss_use_avmax", po::value<bool>((bool*)(&useAVelCstSS)), "use spaceship maximal angular velocity during trajectory generation")
//     ("SSTrajGen.ss_use_aamax", po::value<bool>((bool*)(&useAAccCstSS)), "use spaceship maximal angular acceleration during trajectory generation")
    ;

  po::variables_map vm_config_sssys;
  fs::path cfg_path_sssys(fs::initial_path());
  cfg_path_sssys = fs::system_complete(fs::path(filename,fs::native));
  string cfg_string_sssys = cfg_path_sssys.native_file_string();
  if ( !fs::exists( cfg_path_sssys ) )
    {
      cout << "Unable to load the spaceship configuration. Configuration file not found" << endl;
      return;
    }
  // Open configuration file
  ifstream ifs_sssys((const char *)cfg_string_sssys.c_str());
  store(parse_config_file(ifs_sssys, config_files_sssys), vm_config_sssys);
  notify(vm_config_sssys);
}

/**
 *  \brief init the gui from an input configuration file
 *  \param filename : the input filename
 */
void TrajGenIfc::initTrajGenIfcFromFile(const char* filename)
{
  string IGfilenameS;

  po::options_description config_files("Configuration Files"); // Configuration files
  config_files.add_options() 
    ("TrajGenIfc.x_0", po::value<double>(&x1), "initial position")
    ("TrajGenIfc.y_0", po::value<double>(&y1), "initial position")
    ("TrajGenIfc.th_0", po::value<double>(&theta1), "initial heading")
    ("TrajGenIfc.t_0", po::value<double>(&t1), "initial time")

    ("TrajGenIfc.v_0", po::value<double>(&v1), "initial linear velocity (carlike/wheelchair)")
    ("TrajGenIfc.ph_0", po::value<double>(&kp1), "initial steering angle (carlike)")
    ("TrajGenIfc.om_0", po::value<double>(&om1), "initial angular velocity (wheelchair/spaceship)")
    ("TrajGenIfc.vx_0", po::value<double>(&vx1), "initial linear velocity (spaceship)")
    ("TrajGenIfc.vy_0", po::value<double>(&vy1), "initial linear velocity (spaceship)")

    ("TrajGenIfc.x_g", po::value<double>(&x2), "goal position")
    ("TrajGenIfc.y_g", po::value<double>(&y2), "goal position")
    ("TrajGenIfc.th_g", po::value<double>(&theta2), "goal heading")
    ("TrajGenIfc.t_g", po::value<double>(&t2), "goal time")

    ("TrajGenIfc.v_g", po::value<double>(&v2), "goal linear velocity (carlike/wheelchair)")
    ("TrajGenIfc.ph_g", po::value<double>(&kp2), "goal steering angle (carlike)")
    ("TrajGenIfc.om_g", po::value<double>(&om2), "goal angular velocity (wheelchair/spaceship)")
    ("TrajGenIfc.vx_g", po::value<double>(&vx2), "goal linear velocity (spaceship)")
    ("TrajGenIfc.vy_g", po::value<double>(&vy2), "goal linear velocity (spaceship)")

    ("TrajGenIfc.tg_debug", po::value<int>(&debugLevel), "Trajectory generation debug level")
    ("TrajGenIfc.max_nbsteps", po::value<int>(&maxNbStep), "maximal number of convergence steps")
    ("TrajGenIfc.max_residual", po::value<double>(&maxResidual), "maximal residual")

    ("TrajGenIfc.cl_fixed_xy", po::value<bool>((bool*)(&CL_XYf_fixed)), "true if the final position of the car-like must be fixed")
    ("TrajGenIfc.cl_fixed_th", po::value<bool>((bool*)(&CL_THf_fixed)), "true if the final heading of the car-like must be fixed")
    ("TrajGenIfc.cl_fixed_v", po::value<bool>((bool*)(&CL_Vf_fixed)), "true if the final velocity of the car-like must be fixed")
    ("TrajGenIfc.cl_fixed_ph", po::value<bool>((bool*)(&CL_PHf_fixed)), "true if the final steering angle of the car-like must be fixed")

    ("TrajGenIfc.wc_fixed_xy", po::value<bool>((bool*)(&WC_XYf_fixed)), "true if the final position of the wheelchair must be fixed")
    ("TrajGenIfc.wc_fixed_th", po::value<bool>((bool*)(&WC_THf_fixed)), "true if the final heading of the wheelchair must be fixed")
    ("TrajGenIfc.wc_fixed_v", po::value<bool>((bool*)(&WC_Vf_fixed)), "true if the final velocity of the wheelchair must be fixed")
    ("TrajGenIfc.wc_fixed_om", po::value<bool>((bool*)(&WC_OMf_fixed)), "true if the final steering angle of the wheelchair must be fixed")

    ("TrajGenIfc.use_igdb", po::value<bool>((bool*)(&useIG)), "true if the initial guess database must be used to generate trajectories")
    ("TrajGenIfc.igdb", po::value<string>(&IGfilenameS), "default initial guess database filename")
    ("TrajGenIfc.ig_tmin", po::value<double>(&IG_tmin), "minimal time of the ig learning")
    ("TrajGenIfc.ig_tmax", po::value<double>(&IG_tmax), "maximal time of the ig learning")

    ("TrajGenIfc.cl_spg_acc", po::value<int>(&CL_sampAcc), "car-like initial guess acceleration sampling")
    ("TrajGenIfc.cl_spg_ss", po::value<int>(&CL_sampStSpd), "car-like initial guess steering speed sampling")

    ("TrajGenIfc.cl_spg_t", po::value<int>(&CL_samp_t), "car-like ig time sampling")
    ("TrajGenIfc.cl_spg_vi", po::value<int>(&CL_samp_vi), "car-like ig time sampling")
    ("TrajGenIfc.cl_spg_vg", po::value<int>(&CL_samp_vg), "car-like ig time sampling")
    ("TrajGenIfc.cl_spg_dx", po::value<int>(&CL_samp_dx), "car-like ig time sampling")
    ("TrajGenIfc.cl_spg_dy", po::value<int>(&CL_samp_dy), "car-like ig time sampling")
    ("TrajGenIfc.cl_spg_dth", po::value<int>(&CL_samp_dth), "car-like ig time sampling")

    ("TrajGenIfc.wc_spg_acc", po::value<int>(&WC_sampAcc), "wheelchair initial guess acceleration sampling")
    ("TrajGenIfc.wc_spg_aacc", po::value<int>(&WC_sampAAcc), "wheelchair initial guess angular acceleration sampling")

    ("TrajGenIfc.wc_spg_t", po::value<int>(&WC_samp_t), "wheelchair ig time sampling")
    ("TrajGenIfc.wc_spg_vi", po::value<int>(&WC_samp_vi), "wheelchair ig time sampling")
    ("TrajGenIfc.wc_spg_vg", po::value<int>(&WC_samp_vg), "wheelchair ig time sampling")
    ("TrajGenIfc.wc_spg_dx", po::value<int>(&WC_samp_dx), "wheelchair ig time sampling")
    ("TrajGenIfc.wc_spg_dy", po::value<int>(&WC_samp_dy), "wheelchair ig time sampling")
    ("TrajGenIfc.wc_spg_dth", po::value<int>(&WC_samp_dth), "wheelchair ig time sampling")
    ;

  po::variables_map vm_config;
  fs::path cfg_path(fs::initial_path());
  cfg_path = fs::system_complete(fs::path(filename,fs::native));
  string cfg_string = cfg_path.native_file_string();
  if ( !fs::exists( cfg_path ) )
    {
      cout << "Unable to load the trajgen gui default parameters. Configuration file not found" << endl;
      return;
    }
  // Open configuration file
  ifstream ifs((const char *)cfg_string.c_str());
  store(parse_config_file(ifs, config_files), vm_config);
  notify(vm_config);

  // post initialization
  IGfilename = FXString(IGfilenameS.c_str());
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Drawing functions /////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void TrajGenIfc::drawInitPos()
{
  // verify that the init position is available
  if(map != NULL && x1 >= 0 && x1 < map->getWidth() && x1 >= 0 && x1 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int xF=0, yF=0;
      coordMapToCoordFrame((int)x1, (int)y1, xF, yF);
      
      // draw the init position on the map
      dc.setForeground(initColor);
      dc.drawEllipse(xF - 2, yF - 2, 4, 4);

      cout << "init position : (" << xF << "," << yF << ")" << endl;
    }
  else
    {
      cout << "Can't draw init position" << endl;
    }
}

void TrajGenIfc::drawInitDir()
{
  // verify that the init position is available
  if(map != NULL && x1 >= 0 && x1 < map->getWidth() && x1 >= 0 && x1 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int x1F=0, y1F=0;
      coordMapToCoordFrame((int)x1, (int)y1, x1F, y1F);
      
      int xTgtM = (int)(x1 + 20.0f * cos(theta1));
      int yTgtM = (int)(y1 + 20.0f * sin(theta1));
      
      // compute the vector of the init position direction in the image frame
      int xTgtF=0, yTgtF=0;
      coordMapToCoordFrame(xTgtM, yTgtM, xTgtF, yTgtF);
      
      // draw the init direction on the map
      dc.setForeground(initColor);
      dc.drawLine(x1F, y1F, xTgtF, yTgtF);

      cout << "init direction : (" << x1F << "," << y1F << ") -> (" << xTgtF << "," << yTgtF << ")" << endl;
    }
  else
    {
      cout << "Can't draw init direction" << endl;
    }
}

void TrajGenIfc::drawInitCfg()
{
  drawInitPos();
  drawInitDir();
}

void TrajGenIfc::drawGoalPos()
{
  // verify that the goal position is available
  if(map != NULL && x2 >= 0 && x2 < map->getWidth() && x2 >= 0 && x2 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int xF=0, yF=0;
      coordMapToCoordFrame((int)x2, (int)y2, xF, yF);
      
      // draw the init position on the map
      dc.setForeground(goalColor);
      dc.drawEllipse(xF - 2, yF - 2, 4, 4);

      cout << "goal position : (" << xF << "," << yF << ")" << endl;
    }
  else
    {
      cout << "Can't draw goal position" << endl;
    }
}

void TrajGenIfc::drawGoalDir()
{
  // verify that the goal position is available
  if(map != NULL && x2 >= 0 && x2 < map->getWidth() && x2 >= 0 && x2 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int x2F=0, y2F=0;
      coordMapToCoordFrame((int)x2, (int)y2, x2F, y2F);
      
      int xTgtM = (int)(x2 + 20.0f * cos(theta2));
      int yTgtM = (int)(y2 + 20.0f * sin(theta2));
      
      // compute the vector of the init position direction in the image frame
      int xTgtF=0, yTgtF=0;
      coordMapToCoordFrame(xTgtM, yTgtM, xTgtF, yTgtF);
      
      // draw the init direction on the map
      dc.setForeground(goalColor);
      dc.drawLine(x2F, y2F, xTgtF, yTgtF);

      cout << "goal direction : (" << x2F << "," << y2F << ") -> (" << xTgtF << "," << yTgtF << ")" << endl;
    }
  else
    {
      cout << "Can't draw goal direction" << endl;
    }
}

void TrajGenIfc::drawGoalCfg()
{
  drawGoalPos();
  drawGoalDir();
}

// Load image
FXbool TrajGenIfc::loadimage(const FXString& file)
{
  FXString ext = FXPath::extension(file);
  FXImage *img = NULL;
  FXImage *old = NULL;

  // delete the old map
  if(map != NULL) delete map;
  map = new Map();

  if(comparecase(ext,"pgm")==0){
    img=new FXPPMImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
  }

  if(img==NULL){
    FXMessageBox::error(this,MBOX_OK,"Error Loading Image","Unsupported type: %s",ext.text());
    return FALSE;
  }

  // Load the map
  if(map->load(file.text()) != 0)
    return false;

  // Load the image
  FXFileStream stream;
  if(stream.open(file,FXStreamLoad)){
    getApp()->beginWaitCursor();
    img->loadPixels(stream);
    stream.close();

    img->create();

    // rescaling the image
    int wtF = imageview->getWidth();
    int htF = imageview->getHeight();

    wtI = img->getWidth();
    htI = img->getHeight();

    double ratioW = ((double) wtI) / ((double) wtF);
    double ratioH = ((double) htI) / ((double) htF);

    if(ratioW > ratioH)
      {
	img->scale((FXint)(((double)img->getWidth())/ratioW), (FXint)(((double)img->getHeight())/ratioW));
      }
    else
      {
	img->scale((FXint)(((double)img->getWidth())/ratioH), (FXint)(((double)img->getHeight())/ratioH));
      }
    
    old=imageview->getImage();
    imageview->setImage(img);
    // if(old != NULL) delete old;                    // !!!!!!!!!!!!!!!!!!!!!!!! Why can't I delete it???
    getApp()->endWaitCursor();

    // display both configuration if they are defined
    drawInitCfg();
    drawGoalCfg();
    }

  return TRUE;
}

// convert frame coordinates to map coordinates
void TrajGenIfc::coordFrameToCoordMap(const int& xF, const int& yF, int& xM, int& yM)
{
  // get back the dimensions of the frame
  int wtF = imageview->getWidth();
  int htF = imageview->getHeight();

  // compute the ratio between frame and map
  double ratioW = ((double) wtI) / ((double) wtF);
  double ratioH = ((double) htI) / ((double) htF);

  if(ratioW > ratioH)
    {
      // compute the shift on y
      int shift = (int)((htF - ((double)htI)/ratioW)/2.0f);

      // compute the coordinates in the map
      xM = (int)(xF * ratioW);
      yM = htI -1 - (int)((yF-shift) * ratioW);
      return;
    }
  else
    {
      // compute the shift on x
      int shift = (int)((wtF - ((double)wtI)/ratioH)/2.0f);

      // compute the coordinates in the map
      xM = (int)((xF-shift) * ratioH);
      yM = htI -1 - (int)(yF * ratioH);
      return;
    }
}

// convert map coordinates to frame coordinates
void TrajGenIfc::coordMapToCoordFrame(const int& xM, const int& yM, int& xF, int& yF)
{
  // get back the dimensions of the frame
  int wtF = imageview->getWidth();
  int htF = imageview->getHeight();

  // compute the ratio between frame and map
  double ratioW = ((double) wtI) / ((double) wtF);
  double ratioH = ((double) htI) / ((double) htF);

  if(ratioW > ratioH)
    {
      // compute the shift on y
      int shift = (int)((htF - ((double)htI)/ratioW)/2.0f);

      // compute the coordinates in the frame
      xF = 1 + (int)(xM / ratioW);
      yF = 1 + (int)(((double)(htI -1 - yM) / ratioW) + shift);
      return;
    }
  else
    {
      // compute the shift on x
      int shift = (int)((wtF - ((double)wtI)/ratioH)/2.0f);

      // compute the coordinates in the frame
      xF = 1 + (int)(xM / ratioH) + shift;
      yF = 1 + (int)(((double)(htI -1 - yM)) / ratioH);
      return;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// Main ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here we begin
int main(int argc,char *argv[]){
  
  // Make application
  FXApp application("Interface","TrajectoryGeneration");

  // Open the display
  application.init(argc,argv);
  
  // Make window
  new TrajGenIfc(&application);
  
  // Create the application's windows
  application.create();
  
  // Run the application
  return application.run();
}
